unit AdvOpt;
{
LForms - utility forms for least squares fitting routines for PGOPHER and other programs
Copyright (C) 2005-15 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
{$I utils.inc}
interface

uses {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
     Graphics, Forms, Controls, StdCtrls, ExtCtrls,
     SysUtils, Classes, NewSpin, NGrid, Grids, AlEdit, EditExtr;

type
  TAdvDlg = class(TForm)
    Label1: TLabel;
    Increment: TFixedSpin;
    Label2: TLabel;
    ScaleChanges: TFixedSpin;
    SVDMat: TNumberGrid;
    Label3: TLabel;
    SVDThresh: TFixedSpin;
    Bevel: TBevel;
    UseStdDev: TCheckBox;
    LinearFit: TCheckBox;
    UseSVD: TCheckBox;
    procedure IncrementIncrement(Sender: TObject; var NewValue: Double;
      LastValue: Double; n: Integer);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormResize(Sender: TObject);
    procedure ScaleChangesIncrement(Sender: TObject; var NewValue: Double;
      LastValue: Double; n: Integer);
  private
    {$IFDEF MSWINDOWS}
    procedure WMGetMinMaxInfo(var Msg: TMessage); message WM_GETMINMAXINFO;
    {$ENDIF}
  end;

var
  AdvDlg: TAdvDlg;

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

procedure TAdvDlg.IncrementIncrement(Sender: TObject; var NewValue: Double;
  LastValue: Double; n: Integer);
begin
  NewValue := LastValue;
  if LastValue = 0 then begin
    NewValue := (Sender as TFixedSpin).MaxValue * 1e-5;
    Exit;
  end;
  while n > 0 do begin
    NewValue := NewValue * 2;
    Dec(n);
  end;
  while n < 0 do begin
    NewValue := NewValue / 2;
    Inc(n);
  end;
end;

procedure TAdvDlg.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  CloseAction := caHide;
end;

procedure TAdvDlg.FormResize(Sender: TObject);
begin
  {$IFDEF LCL_}
  Constraints.MinWidth := UseStdDev.Left + UseStdDev.Width + SVDMat.Left;
  Constraints.MinHeight := SVDMat.Top + 50;
  {$ENDIF}
  SVDMat.AutoDefaultRowHeight;
end;

{$IFDEF MSWINDOWS}
procedure TAdvDlg.WMGetMinMaxInfo(var Msg: TMessage);
var
  w: Integer;
begin
  inherited;
  if (SVDThresh = nil) or (SVDMat = nil) then Exit;
  w := SVDThresh.Left + SVDThresh.Width + + SVDThresh.ExtraWidth + SVDMat.Left;
  Inc(w, 2*GetSystemMetrics(SM_CXFRAME));
  with PMINMAXINFO(Msg.lParam)^ do begin
    {$IFNDEF VER80}
      ptMinTrackSize.x := w;
      ptMinTrackSize.y := SVDMat.Top + 50;
    {$ELSE}
      ptMinTrackSize.x := w;
      ptMinTrackSize.y := SVDMat.Top + 50;
    {$ENDIF}
  end;
  Msg.Result := 0;
end;
{$ENDIF}

procedure TAdvDlg.ScaleChangesIncrement(Sender: TObject;
  var NewValue: Double; LastValue: Double; n: Integer);
begin
  if LastValue = 0 then
    NewValue := 0.25
  else
    IncrementIncrement(Sender, NewValue, LastValue, n);
end;

end.
