unit BaseGraphC;
{$I utils.inc}{ Copyright (C) 2008 C M Western; see AdvOpt.pas for license}
interface

uses
  {$IFNDEF CLX_} Messages, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls,
  {$ELSE} QGraphics, QControls, QForms, QDialogs, QStdCtrls, QExtCtrls, {$ENDIF}
  SysUtils, Classes, Graphs, Split,  PrintPre, ScrlCtrl, MDIChildForm,
  ComCtrls, MathSup, GraphI, GraphM;

type
  TPlottedPoint = record
    x, y: Double;
    Data: TObject;
    Index: Integer;
  end;

  { TBaseGraphChild }

  TBaseGraphChild = class(TMDIChildForm)
    G: TGraphD;
    OC: TGraphD;
    Split: TSplit;
    GraphPrintPreview: TGraphPrintPreview;
    StatusBar: TStatusBar;
    procedure FormResize(Sender: TObject);
    procedure GReplot(Sender: TObject; GG: TGraphM);
    procedure FormDestroy(Sender: TObject);
    procedure OCReplot(Sender: TObject; GG: TGraphM);
    procedure SplitMoved(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  public
    RecordPoints: Boolean;
    PlottedPoints: array [Boolean] of array of TPlottedPoint;
    LastPlottedPoint: array [Boolean] of Integer;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetXLimits(var DPoint: TDPoint);
    procedure CheckYLimits(var DPoint: TDPoint);
    procedure OnPlotPoint(Ax, Ay: Double; Sender: TObject; AIndex: Integer; IsOC: Boolean);
  end;

var
  BaseGraphChild: TBaseGraphChild;

procedure ShowBaseGraphChild;

implementation

uses Math, Types, Axes, BaseDataC;

{$IFNDEF LCL_}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

procedure ShowBaseGraphChild;
begin
  if not Assigned(BaseGraphChild) then
    BaseGraphChild := TBaseGraphChild.Create(Application);
  BaseGraphChild.BringToFront;
end;

procedure TBaseGraphChild.FormResize(Sender: TObject);
begin
  Split.ParentResized;
end;

procedure TBaseGraphChild.GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  GD: TGraphD;
  i, j: Integer;
  s: string;
  IsOC: Boolean;
begin
  IsOC := Sender = OC;
  GD := Sender as TGraphD;
  StatusBar.Panels[0].Text := RoundedStr(GD.LastMouseX, (GD.Xmax-GD.Xmin)/Width);
  StatusBar.Panels[1].Text := RoundedStr(GD.LastMouseY, (GD.Ymax-GD.Ymin)/Height);
  if GD.AreaSelected then
    with GD.SelectedMathArea do begin
      StatusBar.Panels[3].Text := RoundedStr(Right-Left, (GD.Xmax-GD.Xmin)/Width);
      StatusBar.Panels[4].Text := RoundedStr(Bottom-Top, (GD.Ymax-GD.Ymin)/Height);
    end
  else begin
    StatusBar.Panels[3].Text := '';
    StatusBar.Panels[4].Text := '';
  end;
  s := '';
  for i := 0 to Min(LastPlottedPoint[IsOC], High(PlottedPoints[IsOC])) do
    with PlottedPoints[IsOC][i] do
      if GD.CloseToX(GD.LastMouseX, X) and GD.CloseToY(GD.LastMouseY, Y) then
        with MainForm do
          for j := 0 to MDIChildCount - 1 do
            { Check object still exists }
            if (Data = MDIChildren[j]) and (Data is TBaseDataChild) then begin
              s := s + TBaseDataChild(Data).OnGetPointInfo(Index) + ' ';
              Break;
            end;
  StatusBar.Panels[5].Text := s;
end;

procedure TBaseGraphChild.GReplot(Sender: TObject; GG: TGraphM);
var
  Limits: array [0..1] of TDPoint;
  nPoints, ny: Longint;
  i, dy, Line: Integer;
  xp1, xp2, yp1, yp2, yml, ymu: Double;
  InfoSize, ISize: TPoint;
  ShowData, AxesShown: Boolean;
  FOnPlotPoint: TOnPlotPoint;
begin
  nPoints := 0;
  ny := 0;
  SetNan(Limits[1].X);
  if not AxesDlg.xlAuto then
    Limits[0].X := AxesDlg.xl;
  if not AxesDlg.xuAuto then
    Limits[0].Y := AxesDlg.xu;
  with MainForm do
    for i := 0 to MDIChildCount - 1 do
      if MDIChildren[i] is TBaseDataChild then
        TBaseDataChild(MDIChildren[i]).DetermineLimits(Limits, nPoints, ny, AxesDlg.xlAuto, AxesDlg.xuAuto, False, False);
  FOnPlotPoint := nil;
  if GG.IsScreen then
    if RecordPoints then begin
      SetLength(PlottedPoints[False], nPoints);
      FOnPlotPoint := {$IFDEF FPC}@{$ENDIF}OnPlotPoint;
      LastPlottedPoint[False] := -1;
    end else begin
      PlottedPoints[False] := nil;
    end;
  if (nPoints = 0) and (AxesDlg.xlAuto or AxesDlg.xlAuto)
    then Exit;
  SetXLimits(Limits[0]);
  if (AxesDlg.yuAuto or AxesDlg.ylAuto) and AxesDlg.ScaleOnCurve then
    with MainForm do
      for i := 0 to MDIChildCount - 1 do
        if MDIChildren[i] is TBaseDataChild then
          TBaseDataChild(MDIChildren[i]).DetermineLimits(Limits, nPoints, ny, AxesDlg.xlAuto, AxesDlg.xuAuto, True, False);
  CheckYLimits(Limits[1]);
  AxesDlg.SetYLimits(Limits[1]);
  GG.Font := AxesDlg.Font;
  GG.Start;
  GG.Title1(AxesDlg.Title);
  GG.PspaceArg(xp1, xp2, yp1, yp2);
  yml := yp2;
  ymu := yp1;
  if GG.IsPrinter and AxesDlg.PrintFitDetails
    then begin
    dy := Round(GG.CellHeight);
    GG.Justify := taLeftJustify;
    GG.TextI(0, 0, DateToStr(Now));
    InfoSize := Point(0, 2*dy);
    with MainForm do
      for i := 0 to MDIChildCount - 1 do
        if MDIChildren[i] is TBaseDataChild then begin
          ISize := TBaseDataChild(MDIChildren[i]).GetInfoSize(GG);
          if ISize.X > InfoSize.X then InfoSize.X := ISize.X;
          Inc(InfoSize.Y, ISize.Y+dy);
        end;
    Line := GG.Ypixels - 3*dy;
    ShowData := GG.YPixels >= InfoSize.Y;
    with MainForm do
      for i := 0 to MDIChildCount - 1 do
        if MDIChildren[i] is TBaseDataChild then begin
          TBaseDataChild(MDIChildren[i]).ShowInfo(Line, ShowData, GG);
          Dec(Line, dy);
        end;
    xp1 := { xp1 + } (InfoSize.X + 2*GG.AvCharWidth)/GG.XPixels;
  end;
  if (GG.IsClipboard or GG.IsPrinter) and OC.Visible then begin
    yml := (Split.Top + Split.Height div 2) / ClientHeight * (yp1-yp2) + yp2 - yp1/2;
    ymu := yml + yp1;
  end;
  if GG.Visible then begin
    GG.Pspace(xp1, xp2, yp1, yml);
    GG.Style(0);
    GG.Map(Limits[0].X, Limits[0].Y, Limits[1].X, Limits[1].Y);
    GG.LineStyle := AxesDlg.AxisStyle;
    AxesShown := False;
    with MainForm do
      for i := 0 to MDIChildCount - 1 do
        if MDIChildren[i] is TBaseDataChild then
          with TBaseDataChild(MDIChildren[i]) do begin
            if not AxesShown then
              GG.Scales(XLabel, YLabel);
            AxesShown := True;
            PlotData(GG, FOnPlotPoint);
          end;
  end;
  if (GG.IsPrinter or GG.IsClipboard) and OC.Visible then begin
    GG.Pspace(xp1, xp2, ymu, yp2);
    OCReplot(Sender, GG);
  end;
  GG.Finish;
  if Assigned(FOnPlotPoint) then
    SetLength(PlottedPoints[False], LastPlottedPoint[False]+1);
end;

procedure TBaseGraphChild.FormDestroy(Sender: TObject);
begin
  if BaseGraphChild = Self then
    BaseGraphChild := nil;
end;

procedure TBaseGraphChild.SetXLimits(var DPoint: TDPoint);
var
  dx: Double;
begin
  with DPoint do begin
    dx := (Y - X)/20;
    if dx = 0 then dx := 1;
//    if (X >= 0) and (dx > X) then
//      { Special Case - do not span zero if not already included }
//      Problematic, as points and error bars on axis difficult to see
//      X := 0
//    else
      X := X - dx;
    Y := Y + dx;
  end;
  AxesDlg.SetXLimits(DPoint);
end;

procedure TBaseGraphChild.CheckYLimits(var DPoint: TDPoint);
var
  dy: Double;
begin
  with DPoint do begin
    if IsNan(X) then begin
      X := 0;
      Y := 1;
    end;
    dy := (Y - X)/20;
//    if dy = 0 then dy := 1;
//    if (X >= 0) and (dy > X) then
//      X := 0
//    else
      X := X - dy;
    Y := Y + dy;
  end;
  AxesDlg.SetYLimits(DPoint);
end;

procedure TBaseGraphChild.OCReplot(Sender: TObject; GG: TGraphM);
var
  nOC, ny: Longint;
  Limits: array [0..1] of TDPoint;
  i: Integer;
  AxesShown: Boolean;
  FOnPlotPoint: TOnPlotPoint;
begin
  nOC := 0;
  ny := 0;
  if not AxesDlg.xlAuto then
    Limits[0].X := AxesDlg.xl;
  if not AxesDlg.xuAuto then
    Limits[0].Y := AxesDlg.xu;
  with MainForm do
    for i := 0 to MDIChildCount - 1 do
      if MDIChildren[i] is TBaseDataChild then
        TBaseDataChild(MDIChildren[i]).DetermineLimits(Limits, nOC, ny, AxesDlg.xlAuto, AxesDlg.xuAuto, False, True);
  FOnPlotPoint := nil;
  if GG.IsScreen then
    if RecordPoints then begin
      SetLength(PlottedPoints[True], ny);
      FOnPlotPoint := {$IFDEF FPC}@{$ENDIF}OnPlotPoint;
      LastPlottedPoint[True] := -1;
    end else begin
      PlottedPoints[True] := nil;
    end;
  if ny = 0 then Exit;
  SetXLimits(Limits[0]);
  AxesDlg.SetOCLimits(Limits[1]);
  GG.Font := AxesDlg.Font;
  GG.Start;
  GG.Map(Limits[0].X, Limits[0].Y, Limits[1].X, Limits[1].Y);
  GG.LineStyle := AxesDlg.AxisStyle;
  AxesShown := False;
  with MainForm do
    for i := 0 to MDIChildCount - 1 do
      if MDIChildren[i] is TBaseDataChild then
        with TBaseDataChild(MDIChildren[i]) do begin
          if not AxesShown then begin
            GG.Axis(Limits[0].X, Limits[0].Y, 0, True, XLabel);
            GG.Axis(Limits[1].X, Limits[1].Y, Limits[0].X, False, AxesDlg.OCLabel);
          end;
          AxesShown := True;
          PlotOC(GG, FOnPlotPoint);
        end;
  GG.Finish;
  if Assigned(FOnPlotPoint) then
    SetLength(PlottedPoints[True], LastPlottedPoint[True]+1);
end;

procedure TBaseGraphChild.OnPlotPoint(Ax, Ay: Double; Sender: TObject; AIndex: Integer; IsOC: Boolean);
begin
  { TODO: Scope for merging record of points plotted }
  Inc(LastPlottedPoint[IsOC]);
  if LastPlottedPoint[IsOC] > High(PlottedPoints[IsOC]) then
    SetLength(PlottedPoints[IsOC], LastPlottedPoint[IsOC]+1000);
  with PlottedPoints[IsOC][LastPlottedPoint[IsOC]] do begin
    x := Ax;
    y := Ay;
    Data := Sender;
    Index := AIndex;
  end;
end;

{
procedure TGraphChild.GrabBarMove(Sender: TObject);
begin
  if GrabBar.Top < 30 then begin
    OC.Visible := False;
    G.Visible := True;
    GrabBar.Top := 0;
    G.Top := Splitter.Height;
    G.Align := alBottom;
    G.Height := ClientHeight - Splitter.Height;
  end else if ClientHeight - GrabBar.Top - GrabBar.Height < 30 then begin
    G.Visible := False;
    OC.Visible := True;
    GrabBar.Top := ClientHeight - GrabBar.Height;
    OC.Height := ClientHeight - Splitter.Height;
    OC.Top := 0;
    Splitter.Top := OC.Height;
    Splitter.Align := alTop;
  end else begin
    OC.Visible := True;
    G.Visible := True;
    OC.Height := Splitter.Top;
    Splitter.Align := alTop;
    G.Top := Splitter.Top + Splitter.Height;
    G.Align := alClient;
  end;
  EnableAlign;

  FormResize(Self);
end;
}
procedure TBaseGraphChild.SplitMoved(Sender: TObject);
begin
  FormResize(Self);
end;

constructor TBaseGraphChild.Create(AOwner: TComponent);
begin
  inherited;
  AxesDlg.Font := G.Font;
end;

destructor TBaseGraphChild.Destroy;
begin
  if BaseGraphChild = Self then
    BaseGraphChild := nil;
  inherited Destroy;
end;

procedure TBaseGraphChild.FormCreate(Sender: TObject);
begin
  Split.OneWindowVisible(G);
  StatusBar.SimplePanel := False; { DIfferent default in LCL }
end;

//procedure TBaseGraphChild.GMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
//begin
//  G.AreaSelected := False;
//  OC.AreaSelected := False;
//  MainForm.MouseDown(Sender as TGraphD);
//end;
//
//procedure TBaseGraphChild.GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
//begin
//  MainForm.MouseMove(Sender as TGraphD, ssLeft in Shift);
//end;

end.
