unit DatLoader;
{ Copyright (C) 2005 C M Western; see NLFit.pas for license}
{ Simple text/csv file import }
interface

uses SSLoader, TextStrm;

type

  { TDatFileLoader }

  TDatFileLoader = class(TSpreadSheetFileLoader)
  public
    function Load: Boolean; override;
  end;

implementation

uses Math, SysUtils, Classes, MathSup, TabbedLine, UCompat;

{ TDatFileLoader }

function TDatFileLoader.Load: Boolean;
var
  SeparatorCount: TIntegerArray;
  Separator: TSeparator;
  P, Start, LStart, Last: PAnsiChar;
  Buffer: UTF8String;
  Values: array of Double;
  PValues: array of PDouble;
  procedure SetnValues(n: Integer);
  var
    i: Integer;
  begin
    SetLength(Values, n);
    SetLength(PValues, n);
    for i := 0 to n-1 do
      PValues[i] := @Values[i];
  end;
var
  i, nFields, n, nLines, EstnLines, MaxnFields: Integer;
  Text: AnsiString;
  ResizeNeeded: Boolean;
begin
  Result := False;
  if Stream is TCustomMemoryStream then begin
    P := TMemoryStream(Stream).Memory;
    Last := P;
    Inc(P,InitialPosition);
    Inc(Last,Stream.Size);
  end else begin
    { Make a copy }
    if Stream is TStringStream then
      Buffer := TStringStream(Stream).DataString
    else if Stream.Size > InitialPosition then begin
      SetLength(Buffer, Stream.Size-InitialPosition);
      i := Stream.Read(Buffer[1],Length(Buffer));
      if i <> Length(Buffer) then
        FileErr(Format('Only read %d of %d',[i,Length(Buffer)]));
    end else
      Buffer := '';
    P := PAnsiChar(Buffer);
    Last := P + Length(Buffer);
  end;
  Start := P;
  MaxnFields := 0;
  { Autodetect format }
  SetLength(SeparatorCount,3);
  SetnValues(2); { 2 is a reasonable default upper bound, but will grow as required }
  nLines := 0;
  while (P < Last) and (nLines < 10) do begin
    P := TryReadTabbedLine(tlDetect, P, Last, PValues, nFields, SeparatorCount);
    MaxnFields := Max(MaxnFields, nFields);
    if P = nil then
      Break
    else if SkipLineEnding(P) then begin
      if nFields > 0 then
        Inc(nLines);
    end else if nFields = Length(Values) then
      SetnValues(nFields*2) { Not enough space - allocate more and read again }
    else
      SkipString(P, Last, tlDetect);
  end;
  i := SeparatorCount[2] + SeparatorCount[1];
  if ( i > SeparatorCount[0]) and (SeparatorCount[0]+i > 5) then
    if SeparatorCount[2] > SeparatorCount[1] then
      Separator := tlTab
    else
      Separator := tlComma
  else
    Separator := tlSpace;
  ClearSpreadSheet;
  SS.Sheet := 0;
  if P = nil then
    EstnLines := nLines
  else
    EstnLines := nLines*Ceil((Last-Start)/(P-Start));
  if Verbose then
    WriteLn(Log^, nLines,' Lines ', Ord(Separator), ' based on Space ',SeparatorCount[0],' Comma ',SeparatorCount[1],' tab ',SeparatorCount[2], '; Estimated total lines ',EstnLines );
  if EstnLines <= 0 then
    EstnLines := 10;
  ResizeNeeded := True;
  P := Start;
  nLines := 0;
  nFields := 0;
  while (P <> nil) and (P < Last) do begin
    LStart := P;
    if nLines > EstnLines then begin
      EstnLines := Ceil(Max(nLines,EstnLines)*1.25);
      ResizeNeeded := True;
    end;
    if ResizeNeeded then begin
      for i := 0 to MaxnFields-1 do
        SS.MakeAvailable(EstnLines,i);
      SetLength(PValues, MaxnFields);
      ResizeNeeded := False;
    end;
    for i := nFields to MaxnFields-1 do
      PValues[i-nFields] := @SS.Sheets[0][i][nLines];
    for i := MaxnFields-nFields to MaxnFields - 1 do
      PValues[i] := nil;
    P := TryReadTabbedLine(Separator, P, Last, PValues, n, EmptyArray);
    Inc(nFields, n);
    if (Verbose) then begin
      Write(Log^, nLines:10);
      if P = nil then
        Write(Log^, ' end ')
      else
        Write(Log^, Ord(P^):4, ' ');
      for i := 0 to n-1 do
        Write(Log^, ' ', PValues[i]^);
      WriteLn(Log^);
    end;
    if (P = nil) or (P = Last) or SkipLineEnding(P) then begin
      if (nFields > 0) then begin
        Inc(nLines);
        nFields := 0;
      end;
    end else if nFields = MaxnFields then begin
      Inc(MaxnFields);
      ResizeNeeded := True;
      if MaxnFields > 1000 then
        Exit; { Something probably not right }
    end else begin
      { Text field }
      Text := SkipString(P, Last, Separator);
      if Text <> '' then
        SS.AddString(nLines,nFields,Text);
      Inc(nFields);
      { Loop will keep reading rest of line}
    end;
  end;
  SS.Rows := nLines;
  for i := 0 to MaxnFields-1 do
    SetLength(SS.Sheets[0][i], nLines);
  Result := ProcessSpreadSheet;
end;

end.
