unit ExpPlotLoader;
{ Copyright (C) 2005 C M Western; see NLFit.pas for license}
interface

uses Classes, Units, MathSup, pgDOM;

type

  TArrayType = (tyI8, tyI16, tyI32, tyI64, tyU8, tySingle, tyDouble);

  TExperimentalPlotLoader = class;

  TPlotLoaderConnector = class
  private
    FnPlots: Integer;
  protected
    FReportErrors: Boolean;
    FStreamName: string;
    ExperimentalPlotLoader: TExperimentalPlotLoader;
    function GetComment: string; virtual;
    procedure SetComment(const AValue: string); virtual;
    procedure SetnPlots(const Value: Integer); virtual;
  public
    constructor Create(AnObject: TObject; AStreamName: string); virtual;
    function QueryPickDest(var AChannelDest: TIntegerArray; var AnWeights: Integer; Limits: TDPointArray): Integer; virtual;
    procedure ReportError(Msg: string; Fatal: Boolean); virtual;
    procedure SetLineList(Plot: Integer; IsLineList:Boolean); virtual;
    procedure SetIntensityUnits (IntensityUnits: TIntensityUnits); virtual;
    procedure SetUnits (Plot: Integer; Units: TResonanceUnits); virtual;
    procedure SetScale(Scale: Double); virtual;
    procedure SetColour(Plot: Integer; Colour: Integer); virtual;
    procedure SetLength(Plot: Integer; Length: Integer; Weighted, HasFrequency, HasIntensity: Boolean); virtual; abstract;
    procedure SetValue(Plot, ArrayIndex, Row: Integer; Value: Double); virtual; abstract;
    procedure SetCommentN(Plot: Integer; const AValue: string); virtual;
    procedure SetNameN(Plot: Integer; const AValue: string); virtual;
    procedure SetName(const AValue: string); virtual;
    procedure SetBaseName(const AValue: string); virtual;
    procedure ReadAndCopyArray(Plot: Integer; ArrayIndex:Integer; ArrayType: TArrayType; nRead: Longint = -1; Scale: Double = 1; Offset: Double = 0; Stream: TStream = nil); virtual; abstract;
    function NewObject(Index: Integer; Update: Boolean = False):TObject; virtual; abstract;
    function Another(Index: Integer): TPlotLoaderConnector; virtual; abstract;
    procedure CheckAvailableChannels(N, nPoints: Integer; AskAboutWeights: Boolean); virtual; abstract;

    property Comment: string read GetComment write SetComment;
    property ReportErrors: Boolean read FReportErrors;
    property StreamName: string read FStreamName;
    property nPlots: Integer read FnPlots write SetnPlots;
  end;

  TPlotLoaderConnectorClass = class of TPlotLoaderConnector;

  TExperimentalPlotLoaderConnector = class(TPlotLoaderConnector)
  protected
    procedure SetnPlots(const Value: Integer); override;
  public
    Plots: array of TExperimentalPlotLoaderConnector;
    Intensity, Frequency, Weights: TDoubleArray;
    LoadAll: Boolean;
    constructor Create(AnObject: TObject; AStreamName: string); override;
    destructor Destroy; override;
    procedure SetCommentN(Plot: Integer; const AValue: string); override;
    procedure SetNameN(Plot: Integer; const AValue: string); override;
    procedure SetLength(Plot: Integer; Length: Integer; Weighted, HasFrequency, HasIntensity: Boolean); override;
    procedure ReadAndCopyArray(Plot: Integer; ArrayIndex:Integer; ArrayType: TArrayType; nRead: Longint = -1; Scale: Double = 1; Offset: Double = 0; Stream: TStream = nil); override;
    procedure SetValue(Plot, ArrayIndex, Row: Integer; Value: Double); override;
    procedure CheckAvailableChannels(N, nPoints: Integer; AskAboutWeights: Boolean); override;
  end;

  TLoadType = (Only2, TwoOrThree, Only3);

  PText = ^Text;

  { TExperimentalPlotLoader }

  TExperimentalPlotLoader = class
  private
    FConnector: TPlotLoaderConnector;
    FStream: TStream;
    FExt: string;
    ForcePrompt: Boolean;
    FAutoDetectFrequency: Boolean;
    FSwitchingEnds: Boolean;
    FSwapWords: Boolean;
    procedure SetLittleEndian(const Value: Boolean);
    procedure SetSwapWords(const Value: Boolean);
  protected
    LineNumber: Integer;
    FrequencyFactor: Double;
    ChangesUp, ChangesDown, nValues: array of Integer;
    Limits: TDPointArray;
    LastValue: array of Double;
    FLittleEndian: Boolean;
    Log: PText;
    InitialPosition: Int64;
    procedure AddChange(Channel: Integer; Value: Double);
    procedure ReadArray(var a: TDoubleArray; ArrayType: TArrayType; Scale: Double = 1; Offset: Double = 0; From: TStream = nil);
    procedure AddIncrementCountFor(Channel: Integer; nPoints: Integer; ArrayType: TArrayType);
    procedure ReadAndCopyArray(Plot: Integer; ArrayIndex:Integer; ArrayType: TArrayType; nRead: Longint = -1; Scale: Double = 1; Offset: Double = 0; Stream: TStream = nil);
    function MapIndex(var Plot, ArrayIndex: Integer):Boolean;
    procedure SetValue(Plot, ArrayIndex, Row: Integer; Value: Double); virtual;
    procedure SetNameN(Plot, ArrayIndex: Integer; AName: string);
    procedure CheckAvailableChannels(N, nPoints: Integer; AskAboutWeights: Boolean);
    procedure DoAutoDetectFrequency;
    procedure SetnChannels(nChannels:Integer);
    procedure ResetConnector;
    procedure Rewind;
    function Setup: Boolean; virtual;
  public
    ChannelDest: TIntegerArray;
    IndependentFrequencyScales:Boolean;
    ColumnNames: array of string;
    Verbose: Boolean;
    AskAboutWeights: Boolean;
    XMLDocument: TXMLDocument;
    constructor Create(AConnector: TPlotLoaderConnector; AStream: TStream; AExt: string; DefaultUnits: TResonanceUnits; DefaultIntensityUnits: TIntensityUnits);
    destructor Destroy; override;
    function DoLoad(var TheLog: TextFile; DoAskAboutWeights: Boolean = True): Boolean; virtual;
    function Load:Boolean; virtual; abstract;
    procedure SetupChannels(nPlots, nPoints: Integer; HasIndependentFrequencyScales:Boolean);
    procedure FileErr(Msg: string);
    procedure FixEndian(var Value; Length: Integer);
    procedure Read(var a; Count: Integer; From: TStream = nil);
    function NextInt16: Word;
    function NextInt32: Integer;
    function NextSingle: Single;
    function NextDouble: Double;
    function FindNode(Node: TDOMNode; const NodeName: string):TDOMNode;
    property Connector: TPlotLoaderConnector read FConnector;
    property Stream: TStream read FStream;
    property Ext: string read FExt write FExt;
    property LittleEndian: Boolean read FLittleEndian write SetLittleEndian;
    property SwapWords: Boolean read FSwapWords write SetSwapWords;
    property SwitchingEnds: Boolean read FSwitchingEnds;
  end;

  TExperimentalPlotLoaderClass = class of TExperimentalPlotLoader;

procedure RegisterPlotLoader(Loader: TExperimentalPlotLoaderClass; FileFilter: string; FileExt: string);
function TryPlotLoader(var Log: TextFile; ConnectorClass: TPlotLoaderConnectorClass; AnObject: TObject; Stream: TStream; FileName: string; AskAboutWeights, AlwaysPrompt, AutoDetectFrequency: Boolean; DefaultUnits: TResonanceUnits = cm1; DefaultIntensityUnits: TIntensityUnits = Arbitrary):Boolean;

var
  PlotLoaderFilter: string = '';
  PlotLoaders: array of TExperimentalPlotLoaderClass = nil;

implementation

uses { Note individual file loader files must be included here to force inclusion in final exectable }
     OvrLoader, DrvLoader, SPCLoader, SPLoader, PSLoader, DrLoader, SIFLoader,
     JDXLoader, DatLoader, OPJLoader, SPALoader, XLSXLoader, FITSLoader, OpusLoader,
     {$IFDEF FPC} {$IFNDEF NOFPSPREADSHEET} XLSLoader, {$ENDIF}{$ENDIF}
     SysUtils, Math, UCompat, Profile;

var
  PlotLoadTime: string = 'Plot Loading';

function TryPlotLoader(var Log: TextFile; ConnectorClass: TPlotLoaderConnectorClass; AnObject: TObject; Stream: TStream; FileName: string; AskAboutWeights, AlwaysPrompt, AutoDetectFrequency: Boolean; DefaultUnits: TResonanceUnits; DefaultIntensityUnits: TIntensityUnits):Boolean;
var
  Ext: string;
  PlotLoader: TExperimentalPlotLoader;
  i: Integer;
begin
  Result := False;
  Ext := LowerCase(ExtractFileExt('a'+FileName)); { Allows Filename to be simply '.ext' }
  for i := High(PlotLoaders) downto 0 do begin
    PlotLoader := PlotLoaders[i].Create(ConnectorClass.Create(AnObject,FileName), Stream, Ext, DefaultUnits, DefaultIntensityUnits);
    try
      TStart(PlotLoadTime);
      PlotLoader.ForcePrompt := AlwaysPrompt;
      PlotLoader.FAutoDetectFrequency := AutoDetectFrequency;
      Result := PlotLoader.DoLoad(Log, AskAboutWeights);
      TEnd(PlotLoadTime);
      if Result then
        Exit;
    finally
      PlotLoader.Free;
    end;
  end;
end;

procedure RegisterPlotLoader(Loader: TExperimentalPlotLoaderClass; FileFilter: string; FileExt: string);
  procedure SubRegister(Loader: TExperimentalPlotLoaderClass; FileFilter: string; FileExt: string);
  begin
    SetLength(PlotLoaders, Length(PlotLoaders)+1);
    PlotLoaders[Length(PlotLoaders)-1] := Loader;
    {$IFDEF LINUX}
    FileExt := FileExt + ';' + UpperCase(FileExt);
    {$ENDIF}
    PlotLoaderFilter := PlotLoaderFilter + FileFilter + ' (' + FileExt + ')|' + FileExt + '|';
  end;
begin
  { Force ordinary text files to be first on the list, so they will be tried last }
  if Length(PlotLoaders) = 0 then
    SubRegister(TDatFileLoader, 'Text Files', '*.dat;*.csv');
  SubRegister(Loader, FileFilter, FileExt);
end;

{ TExperimentalPlotLoader }

procedure TExperimentalPlotLoader.DoAutoDetectFrequency;
{ Determine frequency channel by looking for the largest number of net changes }
var
  i, FChan, DChan: Integer;
  Ratio: Double;
begin
  if not FAutoDetectFrequency then
    Exit;
  if Verbose then begin
    WriteLn(Log^, '    Changes Up Changes Down     nValues');
    for i := 0 to High(ChangesUp) do
      WriteLn(Log^, i:4, ChangesUp[i]:14, ChangesDown[i]:16, nValues[i]:12);
  end;
  FChan := -1;
  { Look for channel with all changes in one direction }
  for i := 0 to High(ChannelDest) do
    if (ChannelDest[i] >= 0) and ((ChangesUp[i] = 0) xor (ChangesDown[i] = 0)) then
      if FChan >= 0 then begin
        ForcePrompt := True;
        Break; { more than one monotonic channel }
      end else
        FChan := i;
  if FChan < 0 then begin
    { Look for 5 times as many changes one way as the other }
    for i := 0 to High(ChannelDest) do
      if (ChannelDest[i] >= 0) and (ChangesDown[i] > 0) then begin
        Ratio := Abs(ChangesUp[i] / ChangesDown[i]);
        if (Abs(Ratio) >= 5) or (Abs(Ratio) <= 0.2) then
          if FChan >= 0 then begin
            ForcePrompt := True;
            Break; { more than one monotonic channel }
          end else
            FChan := i;
          end;
  end;
  if FChan < 0 then
    if Length(ChannelDest) > 1 then
      ForcePrompt := True
    else if (Length(ChannelDest) = 1) and (nValues[0] = 1) then
      FChan := 0; { Single value must be frequency }
  DChan := 1;
  for i := 0 to High(ChannelDest) do
    if i = FChan then
      ChannelDest[i] := 0 { Frequency channel }
    else if ChannelDest[i] >= 0 then begin
      ChannelDest[i] := DChan; { Default data channel is first non-frequency channel }
      DChan := -1;
    end;
end;

function TExperimentalPlotLoader.DoLoad(var TheLog: TextFile; DoAskAboutWeights: Boolean): Boolean;
begin
  Log := @TheLog;
  AskAboutWeights := DoAskAboutWeights;
  Result := Setup;
  if Result then
    Result := Load;
  if not Result then
    Rewind;
end;

procedure TExperimentalPlotLoader.CheckAvailableChannels(N, nPoints: Integer; AskAboutWeights: Boolean);
begin
  Connector.CheckAvailableChannels(N,nPoints,AskAboutWeights);
end;

function TExperimentalPlotLoader.Setup: Boolean;
begin
  Result := True;
end;

procedure TExperimentalPlotLoader.SetupChannels(nPlots, nPoints: Integer; HasIndependentFrequencyScales:Boolean);
var
  i: Integer;
  HasFrequency, HasIntensity, Weighted: Boolean;
begin
  IndependentFrequencyScales := HasIndependentFrequencyScales;
  SetLength(ChannelDest,Max(Length(ChannelDest),nPlots));
  HasFrequency := False;
  HasIntensity := False;
  Weighted := False;
  for i := 0 to High(ChannelDest) do
    if ChannelDest[i] = 0 then
      HasFrequency := True
    else if ChannelDest[i] >= 1000 then
      Weighted := True
    else if ChannelDest[i] > 0 then
      HasIntensity := True;
  Connector.nPlots := nPlots;
  Connector.SetLength(-1,nPoints,Weighted,HasFrequency,HasIntensity);
end;

constructor TExperimentalPlotLoader.Create(AConnector: TPlotLoaderConnector; AStream: TStream; AExt: string; DefaultUnits: TResonanceUnits; DefaultIntensityUnits: TIntensityUnits);
begin
  FStream := AStream;
  InitialPosition := Stream.Position;
  FExt := LowerCase(AExt);
  LineNumber := 0;
  FrequencyFactor := 1;
  ForcePrompt := False;
  LittleEndian := True;
  FConnector := AConnector;
  SetnChannels(2);
  if FConnector <> nil then begin
    FConnector.ExperimentalPlotLoader := Self;
    Connector.FReportErrors := True;
    Connector.SetUnits(-1,DefaultUnits);
    Connector.SetIntensityUnits(DefaultIntensityUnits);
  end;
end;

procedure TExperimentalPlotLoader.FileErr(Msg: string);
begin
  if Connector.ReportErrors then begin
    if LineNumber <> 0 then
      Msg := Format('%s on line %d', [Msg, LineNumber]);
    Connector.ReportError(Msg,False);
  end;
end;

function TExperimentalPlotLoader.FindNode(Node: TDOMNode; const NodeName: string): TDOMNode;
var
  Child: TDomNode;
begin
  Result := Node.FindNode(NodeName);
  Child := Node.FirstChild;
  while not Assigned(Result) and Assigned(Child) do begin
    if Child is TDOMNode_WithChildren then
      Result := FindNode(TDOMNode_WithChildren(Child),NodeName);
    Child := Child.NextSibling;
  end;
end;

procedure TExperimentalPlotLoader.FixEndian(var Value; Length: Integer);
var
  i: Integer;
  s, e: PAnsiChar;
  t: AnsiChar;
begin
  if not SwitchingEnds then
    Exit;
  if {$IFNDEF FPC_BIG_ENDIAN} not {$ENDIF} FLittleEndian then begin
    s := @Value;
    e := s + (Length - 1);
    for i := 0 to Length div 2 - 1 do begin
      t := s^;
      s^ := e^;
      e^ := t;
      Inc(s);
      Dec(e);
    end;
  end;
  if SwapWords and (Length = 4) then begin
    { Odd mode found in .spc files - words, but not bytes swapped }
    i := PInteger(@Value)^;
    i := ((i and $ffff) shl 16) or ((i shr 16) and $ffff);
    PInteger(@Value)^ := i;
  end;
end;

procedure TExperimentalPlotLoader.Read(var a; Count: Integer; From: TStream);
begin
  if not Assigned(From) then
    From := Stream;
  if From.Read(a, Count) <> Count then
    Connector.ReportError('Unexpected end of file',True);
end;

function TExperimentalPlotLoader.NextInt16: Word;
begin
  Read(Result, 2);
  FixEndian(Result, 2);
end;

function TExperimentalPlotLoader.NextInt32: Integer;
begin
  Read(Result, 4);
  FixEndian(Result, 4);
end;

function TExperimentalPlotLoader.NextSingle: Single;
begin
  Read(Result, 4);
  FixEndian(Result, 4);
end;

function TExperimentalPlotLoader.NextDouble: Double;
begin
  Read(Result, 8);
  FixEndian(Result, 8);
end;

procedure TExperimentalPlotLoader.ReadAndCopyArray(Plot: Integer;
  ArrayIndex: Integer; ArrayType: TArrayType; nRead: Longint; Scale: Double;
  Offset: Double; Stream: TStream);
begin
  { Plot number -1 => take from ArrayIndex }
  { Plot number -2 => Frequency scale, copy to all }
  Connector.ReadAndCopyArray(Plot, ArrayIndex,ArrayType,nRead,Scale,Offset,Stream);
end;

procedure TExperimentalPlotLoader.AddChange(Channel: Integer; Value: Double);
begin
  if IsNanOrInf(Value) then
    Exit;
  if IsNan(LastValue[Channel]) then begin
    Limits[Channel].X := Value;
    Limits[Channel].Y := Value;
  end else begin
    if LastValue[Channel] > Value then
      Inc(ChangesDown[Channel])
    else if LastValue[Channel] < Value then
      Inc(ChangesUp[Channel]);
    if Value < Limits[Channel].X then
      Limits[Channel].X := Value;
    if Value > Limits[Channel].Y then
      Limits[Channel].Y := Value;
  end;
  LastValue[Channel] := Value;
  Inc(nValues[Channel]);
end;

procedure TExperimentalPlotLoader.AddIncrementCountFor(Channel: Integer; nPoints: Integer; ArrayType: TArrayType);
var
  A: TDoubleArray;
  i: Integer;
begin
  { Read array of given size and type, and add the count of net changes to a list }
  SetLength(A, nPoints);
  ReadArray(A, ArrayType);
  if Length(ChangesUp) <= Channel then
    SetnChannels(Channel+1);
  for i := 0 to High(A) do
    AddChange(Channel,A[i]);
end;

procedure TExperimentalPlotLoader.ReadArray(var a: TDoubleArray;
  ArrayType: TArrayType; Scale: Double; Offset: Double; From: TStream);
var
  i: Integer;
  PSingle: ^Single;
  PInt64: ^Int64;
  PLongint: ^Longint;
  PSmallint: ^Smallint;
  PShortint: ^Shortint;
  PByte: ^Byte;
begin
  case ArrayType of
    tyDouble: begin
                Read(a[0], Length(a)*Sizeof(Double), From); { Easy case }
                if SwitchingEnds then
                  for i := 0 to Length(a) - 1 do
                    FixEndian(a[i],8);
                if (Scale <> 1) or (Offset <> 0) then
                  for i := 0 to Length(a) - 1 do
                    a[i] := a[i]*Scale + Offset;
              end;
    tySingle: begin
               Read(a[0], Length(a)*Sizeof(Single), From); { Read in place and re-pack }
               PSingle := Pointer(@a[0]);
               Inc(PSingle, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 FixEndian(PSingle^, 4);
                 a[i] := PSingle^*Scale + Offset;
                 Dec(PSingle);
               end;
             end;
    tyI64:   begin
               Read(a[0], Length(a)*Sizeof(Int64), From); { Read in place and convert }
               PInt64 := Pointer(@a[0]);
               for i := 0 to Length(a)-1 do begin
                 FixEndian(PInt64^, 8);
                 a[i] := PInt64^*Scale + Offset;
                 Inc(PInt64);
               end;
             end;
    tyI32:   begin
               Read(a[0], Length(a)*Sizeof(Longint), From); { Read in place and re-pack }
               PLongint := Pointer(@a[0]);
               Inc(PLongint, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 FixEndian(PLongint^, 4);
                 a[i] := PLongint^*Scale + Offset;
                 Dec(PLongint);
               end;
             end;
    tyI16:   begin
               Read(a[0], Length(a)*Sizeof(Smallint), From); { Read in place and re-pack }
               PSmallint := Pointer(@a[0]);
               Inc(PSmallint, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 FixEndian(PSmallint^, 2);
                 a[i] := PSmallint^*Scale + Offset;
                 Dec(PSmallint);
               end;
             end;
    tyI8:    begin
               Read(a[0], Length(a)*Sizeof(Shortint), From); { Read in place and re-pack }
               PShortint := Pointer(@a[0]);
               Inc(PShortint, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 a[i] := PShortint^*Scale + Offset;
                 Dec(PShortint);
               end;
             end;
    tyU8:    begin
               Read(a[0], Length(a)*Sizeof(Byte), From); { Read in place and re-pack }
               PByte := Pointer(@a[0]);
               Inc(PByte, Length(a)-1);
               for i := Length(a)-1 downto 0 do begin
                 a[i] := PByte^*Scale + Offset;
                 Dec(PByte);
               end;
             end;
  end;
end;

procedure TExperimentalPlotLoader.ResetConnector;
var
  NewConnector: TObject;
begin
  { Reset Channel steering info }
  NewConnector := Connector.Another(Connector.nPlots);
  Connector.Free;
  FConnector := NewConnector as TPlotLoaderConnector;
  FConnector.ExperimentalPlotLoader := Self;
  ChangesUp := nil;
  SetnChannels(2);
  ChannelDest := nil;
end;

procedure TExperimentalPlotLoader.Rewind;
begin
  Stream.Position := InitialPosition;
end;

procedure TExperimentalPlotLoader.SetLittleEndian(const Value: Boolean);
begin
  FLittleEndian := Value;
  FSwitchingEnds := {$IFNDEF FPC_BIG_ENDIAN} not {$ENDIF} FLittleEndian or SwapWords;
end;

procedure TExperimentalPlotLoader.SetNameN(Plot, ArrayIndex: Integer; AName: string);
begin
  if MapIndex(Plot, ArrayIndex) and (Plot >= 0) and (ArrayIndex = 1) then
    Connector.SetNameN(Plot+1, AName);
end;

procedure TExperimentalPlotLoader.SetnChannels(nChannels: Integer);
var
  Lastn, i: Integer;
begin
  Lastn := Length(ChangesUp);
  SetLength(ChannelDest, nChannels);
  SetLength(ChangesUp, nChannels);
  SetLength(ChangesDown, nChannels);
  SetLength(LastValue, nChannels);
  SetLength(Limits, nChannels);
  SetLength(nValues, nChannels);
  for i := Lastn to High(ChangesUp) do begin
    ChannelDest[i] := i; { Default }
    ChangesUp[i] := 0;
    ChangesDown[i] := 0;
    nValues[i] := 0;
    SetNan(LastValue[i]);
    SetNan(Limits[i].X);
    SetNan(Limits[i].Y);
  end;
end;

procedure TExperimentalPlotLoader.SetSwapWords(const Value: Boolean);
begin
  FSwapWords := Value;
  LittleEndian := LittleEndian; { Recalc SwitchingEnds }
end;

destructor TExperimentalPlotLoader.Destroy;
begin
  FreeAndNil(FConnector);
  FreeAndNil(XMLDocument);
  inherited Destroy;
end;

function TExperimentalPlotLoader.MapIndex(var Plot, ArrayIndex: Integer): Boolean;
begin
  Result := True;
  if Plot = -1 then begin
    ArrayIndex := ChannelDest[ArrayIndex];
    if ArrayIndex < 0 then
      Result := False
    else if ArrayIndex = 0 then
      Plot := -2
    else if ArrayIndex < 1000 then begin
      Plot := ArrayIndex - 1;
      ArrayIndex := 1;
    end else begin
      Plot := ArrayIndex - 1000;
      ArrayIndex := 2;
    end;
  end;
end;

procedure TExperimentalPlotLoader.SetValue(Plot, ArrayIndex, Row: Integer; Value: Double);
begin
  if MapIndex(Plot, ArrayIndex) then
    Connector.SetValue(Plot, ArrayIndex, Row, Value);
end;

{ TPlotLoaderConnector }

function TPlotLoaderConnector.GetComment: string;
begin
  Result := '';
end;

function TPlotLoaderConnector.QueryPickDest(
  var AChannelDest: TIntegerArray; var AnWeights: Integer; Limits: TDPointArray): Integer;
begin
  Result := -1; { Not implemented }
end;

procedure TPlotLoaderConnector.ReportError(Msg: string; Fatal: Boolean);
begin
  if Fatal then
    raise Exception.Create(Msg)
  else
    WriteLn(ErrOutput, Msg);
end;

procedure TPlotLoaderConnector.SetBaseName(const AValue: string);
begin
  { Ignore by default }
end;

procedure TPlotLoaderConnector.SetColour(Plot: Integer; Colour: Integer);
begin
  { Ignore by default }
end;

procedure TPlotLoaderConnector.SetComment(const AValue: string);
begin
  { Ignore by default }
end;

procedure TPlotLoaderConnector.SetCommentN(Plot: Integer; const AValue: string);
begin
  if Plot = 1 then
    Comment := AValue;
end;

constructor TPlotLoaderConnector.Create(AnObject: TObject; AStreamName: string);
begin
  inherited Create;
  FStreamName := AStreamName;
  FnPlots := 1;
end;

procedure TPlotLoaderConnector.SetLineList(Plot: Integer; IsLineList: Boolean);
begin
  { Ignore by default }
end;

procedure TPlotLoaderConnector.SetName(const AValue: string);
begin
  { Ignore by default }
end;

procedure TPlotLoaderConnector.SetNameN(Plot: Integer; const AValue: string);
begin
  if Plot = 1 then
    SetName(AValue);
end;

procedure TPlotLoaderConnector.SetnPlots(const Value: Integer);
begin
  FnPlots := Max(Value,1);
end;

procedure TPlotLoaderConnector.SetIntensityUnits( IntensityUnits: TIntensityUnits);
begin
  { Ignore by default }
end;

procedure TPlotLoaderConnector.SetUnits(Plot: Integer; Units: TResonanceUnits);
begin
  { Ignore by default }
end;

procedure TPlotLoaderConnector.SetScale(Scale: Double);
begin
  { Ignore by default }
end;

{ TExperimentalPlotLoaderConnector }

procedure TExperimentalPlotLoaderConnector.CheckAvailableChannels(N, nPoints: Integer; AskAboutWeights: Boolean);
var
  nPlts, nWeights, nAct, i: Integer;
begin
  if N < 1 then
    ReportError(Format('Only %d channels of data', [N]),True);
  if LoadAll then begin
    { Simple defaults - all unused channels assumed to be intensity }
    nPlts := 0;
    for i := 0 to N-1 do
      nPlts := Max(nPlts,ExperimentalPlotLoader.ChannelDest[i]);
    for i := 0 to N-1 do
      if (ExperimentalPlotLoader.ChannelDest[i] < 0) and (ExperimentalPlotLoader.ChannelDest[i] > -1000) then begin
        Inc(nPlts);
        ExperimentalPlotLoader.ChannelDest[i] := nPlts;
      end;
  end else begin
    nWeights := 0;
    nPlts := 1;
    nAct := 0;
    for i := 0 to N-1 do
      if ExperimentalPlotLoader.nValues[i] > 0 then
        Inc(nAct);
    if not AskAboutWeights and not ExperimentalPlotLoader.ForcePrompt then begin
      if nAct > 3 then
        ReportError(Format('Too many channels:%d', [N]),True)
      else if nAct = 3 then begin
        for i := 0 to N-1 do
          if (ExperimentalPlotLoader.ChannelDest[i] < 0) and (ExperimentalPlotLoader.ChannelDest[i] > -1000) then
            ExperimentalPlotLoader.ChannelDest[i] := 1000; { Switch unused channel to weights }
        nWeights := 1;
      end;
    end else if (nAct > 2) or ExperimentalPlotLoader.ForcePrompt then begin
      if High(ExperimentalPlotLoader.ChannelDest) >= 0 then begin
        nPlts := QueryPickDest(ExperimentalPlotLoader.ChannelDest, nWeights, ExperimentalPlotLoader.Limits);
        if nPlts = 0 then
          Abort;
      end;
      if nPlts < 0 then
        ReportError(Format('Too many (%d) channels of data', [N]),True);
    end;
  end;
  ExperimentalPlotLoader.SetupChannels(nPlts, nPoints, False);
end;

constructor TExperimentalPlotLoaderConnector.Create(AnObject: TObject; AStreamName: string);
begin
  inherited;
  System.SetLength(Plots, 1);
  Plots[0] := Self;
  LoadAll := True;
end;

destructor TExperimentalPlotLoaderConnector.Destroy;
var
  i: Integer;
begin
  for i := 1 to High(Plots) do
    FreeAndNil(Plots[i]);
  inherited;
end;

procedure TExperimentalPlotLoaderConnector.ReadAndCopyArray(Plot: Integer; ArrayIndex: Integer; ArrayType: TArrayType; nRead: Integer; Scale, Offset: Double; Stream: TStream );
var
  A: TDoubleArray;
  i: Integer;
begin
  if Plot = -1 then
    { Work the destination out from ArrayIndex }
    if ArrayIndex = 0 then
      Plot := -2
    else if ArrayIndex < 1000 then begin
      Plot := ArrayIndex-1;
      ArrayIndex := 1; { Intensity }
    end else begin
      Plot := ArrayIndex-1000;
      ArrayIndex := 2; { Weights }
    end;
  if Plot = -2 then begin
    { Frequency, copy to all }
    if ArrayIndex <> 0 then
      raise Exception.CreateFmt('TExperimentalPlotLoaderConnector.ReadAndCopyArray %d',[ArrayIndex]);
    A := Frequency
  end else
    case ArrayIndex of
      0: A := Plots[Plot].Frequency;
      1: A := Plots[Plot].Intensity;
      2: A := Plots[Plot].Weights;
    else
      raise Exception.CreateFmt('TExperimentalPlotLoaderConnector.ReadAndCopyArray %d',[ArrayIndex]);
    end;
  if (nRead > 0) and (nRead <> Length(A)) then
    if ArrayIndex = 0 then
      { Note that setlength must not be called for Frequency channel, as this creates an unwanted copy }
      raise Exception.Create('Frequency channel must have the same number of points as other channels')
    else
      System.SetLength(A, nRead);
  ExperimentalPlotLoader.ReadArray(A, ArrayType, Scale, Offset, Stream);
  if plot = -2 then
    for i := 1 to High(Plots) do
      Move(A[0], Plots[i].Frequency[0], Length(A)*SizeOf(Double));
end;

procedure TExperimentalPlotLoaderConnector.SetCommentN(Plot: Integer; const AValue: string);
begin
  if (Plot > 1) and (Plot <= Length(Plots)) then
    Plots[Plot-1].SetCommentN(1,AValue)
  else
   inherited;
end;

procedure TExperimentalPlotLoaderConnector.SetLength(Plot: Integer; Length: Integer; Weighted, HasFrequency, HasIntensity: Boolean);
var
  i, j, s, e: Integer;
begin
  if Plot < 0 then begin
    s := 0;
    e := High(Plots)
  end else begin
    s := Plot;
    e := Plot;
  end;
  for i := s to e do begin
    System.SetLength(Plots[i].Frequency, Length);
    System.SetLength(Plots[i].Intensity, Length);
    if Weighted then
      System.SetLength(Plots[i].Weights, Length)
    else
      System.SetLength(Plots[i].Weights, 0);
    if not HasFrequency then
      for j := 0 to Length-1 do
        Plots[i].Frequency[j] := j+1;
    if not HasIntensity then begin
      SetLineList(i,True);
      for j := 0 to Length-1 do
        Plots[i].Intensity[j] := 1;
    end;
  end;
end;

procedure TExperimentalPlotLoaderConnector.SetNameN(Plot: Integer; const AValue: string);
begin
  if (Plot > 1) and (Plot <= Length(Plots)) then
    Plots[Plot-1].SetNameN(1,AValue)
  else
    inherited;      
end;

procedure TExperimentalPlotLoaderConnector.SetnPlots(const Value: Integer);
var
  OldValue, i: Integer;
begin
  OldValue := nPlots;
  inherited;
  for i := nPlots to OldValue - 1 do
    Plots[i].Free;
  System.SetLength(Plots, nPlots);
  for i := Max(1,OldValue) to nPlots-1 do
    Plots[i] := Another(i) as TExperimentalPlotLoaderConnector;
  Plots[0] := Self;
end;

procedure TExperimentalPlotLoaderConnector.SetValue(Plot, ArrayIndex, Row: Integer; Value: Double);
var
  i: Integer;
begin
  if ((ArrayIndex = 0) and not ExperimentalPlotLoader.IndependentFrequencyScales)
     or (ArrayIndex = -2) then
    for i := 0 to High(Plots) do
      Plots[i].Frequency[Row] := Value
  else if ArrayIndex = 0 then
    Plots[Plot].Frequency[Row] := Value
  else if ArrayIndex = 1 then
    Plots[Plot].Intensity[Row] := Value
  else if ArrayIndex = 2 then
    Plots[Plot].Weights[Row] := Value
  else
    raise Exception.CreateFmt('TExperimentalPlotLoaderConnector.SetValue %d %d',[Plot, ArrayIndex]);
end;

end.

