unit FITSLoader;
{ Copyright (C) 2005 C M Western; see NLFit.pas for license}
interface

uses SSLoader, Classes;

type

  { TFITSFileLoader }

  TFITSFileLoader = class(TSpreadSheetFileLoader)
  public
    function Load: Boolean; override;
  end;

implementation

uses SysUtils, ExpPlotLoader, UCompat;

{ TFITSFileLoader }

function TFITSFileLoader.Load: Boolean;
var
  Rec: string[80];
  Keyword: string[8];
  Value: AnsiString;
  Comment: string;
  BITPIX: Integer;
  NAXIS, NAXIS1, NAXIS2, BLANK: Integer;
  CRPIX1, { Reference pixel }
  CRVAL1, { Coordinate at reference pixel }
  CDELT1,  { Coordinate increment per pixel }
  BSCALE, BZERO: Double;
  XSTART, XEND: Double; { X values of the first and last data point }
  xRangeGiven: Boolean; { If both of the above specified }
  //WSTART, WSTOP, DELW: Double;
  ArrayType: TArrayType;
  FrequencyScale: Boolean;
  function NextRecord: Boolean;
  var
    P, Q: Integer;
  begin
    SetLength(Rec,80);
    Result := Stream.Read(Rec[1],80) = 80;
    if Result then begin
      Keyword := Trim(Copy(Rec,1,8));
      Value := '';
      if Rec[9] = '=' then begin
        P := 10;
        while (P <= 80) and (Rec[P] = ' ') do
          Inc(P);
        if (P > 80) or (Rec[P] = '''') then
          Exit;
        Q := P;
        while (Q <= 80) and (Rec[Q] <> ' ') do
          Inc(Q);
        Value := Copy(Rec,P,Q-P);
      end;
    end;
  end;
var
  i: Integer;
begin
  Result := False;
  if (Ext = '') or (Ext = '.fits') or (Ext = '.fts') or (Ext = '.fit') then begin
    LittleEndian := False;
    if not NextRecord or (Keyword <> 'SIMPLE') or (Value <> 'T') then
      Exit;
    Comment := Trim(Rec);
    repeat
      NAXIS := -1;
      NAXIS1 := 1;
      NAXIS2 := 1;
      BSCALE := 1;
      BZERO := 0;
      BLANK := Low(Integer);
      BITPIX := -1;
      //WSTART := -1;
      //WSTOP := -1;
      //DELW := 0;
      CRPIX1 := 0;
      CRVAL1 := 0;
      CDELT1 := 1;
      XSTART := 0;
      XEND := 0;
      xRangeGiven := False;
      FrequencyScale := False;
      while NextRecord do begin
        Comment := Comment + LineEnding + Trim(Rec);
        if Keyword = 'BITPIX' then begin
          if not TryStrToInt(Value, BITPIX) then
            Exit;
        end else if Keyword = 'NAXIS' then begin
          if not TryStrToInt(Value, NAXIS) then
            Exit;
        end else if Keyword = 'NAXIS1' then begin
          if not TryStrToInt(Value, NAXIS1) then
            Exit;
        end else if Keyword = 'NAXIS2' then begin
          if not TryStrToInt(Value, NAXIS2) then
            Exit;
        end else if Keyword = 'BSCALE' then begin
          if not TryStrToFloat(Value, BSCALE) then
            BSCALE := 1;
        end else if Keyword = 'BZERO' then begin
          if not TryStrToFloat(Value, BZERO) then
            BZERO := 0;
        end else if Keyword = 'XSTART' then begin
          if not TryStrToFloat(Value, XSTART) then
            XSTART := 0
          else
            xRangeGiven := True;
        end else if Keyword = 'XEND' then begin
          if not TryStrToFloat(Value, XEND) then
            XEND := 0
          else
            xRangeGiven := True;
        end else if Keyword = 'END' then begin
          Stream.Seek(2880-((Stream.Position - InitialPosition) mod 2880), soFromCurrent);
          Break;
        //end else if Keyword = 'WSTART' then begin
        //  if not TryStrToFloat(Value, WSTART) then
        //    WSTART := -1;
        //end else if Keyword = 'WSTOP' then begin
        //  if not TryStrToFloat(Value, WSTOP) then
        //    WSTOP := -1;
        //end else if Keyword = 'DELW' then begin
        //  if not TryStrToFloat(Value, DELW) then
        //    DELW := 0;
        end else if Keyword = 'CRPIX1' then begin
          if not TryStrToFloat(Value, CRPIX1) then
            CRPIX1 := 0
          else
            FrequencyScale := True;
        end else if Keyword = 'CRVAL1' then begin
          if not TryStrToFloat(Value, CRVAL1) then
            CRVAL1 := 0
          else
            FrequencyScale := True;
        end else if Keyword = 'CDELT1' then begin
          if not TryStrToFloat(Value, CDELT1) then
            CDELT1 := 1
          else
            FrequencyScale := True;
        end;
      end;
    until NAXIS <> 0;
    Result := (NAXIS > -1) and (NAXIS < 3);
    case BITPIX of
      16: ArrayType := tyI16;
      32: ArrayType := tyI32;
      64: ArrayType := tyI64;
     -32: ArrayType := tySingle;
     -64: ArrayType := tyDouble;
    else
      Result := False;
    end;
  end;
  if not Result then
    Exit;
  ClearSpreadSheet;
  SS.Sheet := 0;
  for i := 0 to NAXIS-1 do begin
    SS.MakeAvailable(NAXIS1-1,i);
    ReadArray(SS.Sheets[0][i],ArrayType,BSCALE,BZERO);
  end;
  //if (NAXIS = 1) and (WSTART > 0) and (WSTOP > 0) and (DELW <> 0) then begin
  //  { Probably found frequency scale, but sanity check against resolution  }
  //  if (WSTART > WSTOP) xor (DELW < 0) then
  //    DELW := -DELW;
  //  if Abs(WSTART+DELW*NAXIS1-WSTOP) < ABS(DELW/5) then begin
  //    { Resolution has been calculated as (WSTOP-WSTART)/NAXIS1,
  //      but there are only NAXIS-1 steps }
  //    WSTART := WSTART + DELW/2;
  //    WSTOP := WSTOP - DELW/2;
  //  end;
  //  if Abs(WSTART+DELW*(NAXIS1-1)-WSTOP) < ABS(DELW/5) then begin
  //    SS.MakeAvailable(0,1);
  //    SS.Sheets[0][1] := SS.Sheets[0][0]; { Copy intensity channel to column 2 }
  //    SetLength(SS.Sheets[0][0], NAXIS1); { Allocate new array }
  //    for i := 0 to NAXIS1-1 do
  //      SS.Sheets[0][0][i] := WSTART + DELW*i;
  //  end;
  //end;
  if FrequencyScale or xRangeGiven then begin
    if FrequencyScale and xRangeGiven then
      Connector.ReportError('Ambiguous frequency scale',False);
    { Insert frequency scale as first channel }
    SS.MakeAvailable(0,NAXIS);
    for i := NAXIS downto 1 do
      SS.Sheets[0][i] := SS.Sheets[0][i-1];
    SetLength(SS.Sheets[0][0], NAXIS1); { Force allocation of new array }
    if FrequencyScale then
      for i := 0 to NAXIS1-1 do
        SS.Sheets[0][0][i] := (i + 1-CRPIX1)*CDELT1 + CRVAL1
    else { xRangeGiven }
      for i := 0 to NAXIS1-1 do
        SS.Sheets[0][0][i] := XSTART + i*(XEND-XSTART)/(NAXIS1-1);
  end;
  ProcessSpreadSheet;
  Connector.Comment := Comment;
end;

initialization
  RegisterPlotLoader(TFITSFileLoader, 'FITS Files', '*.fits;*.fts;*.fit');
end.

