unit ItemStrm;
{$I utils.inc}{ Copyright (C) 2016 C M Western; see NLFit.pas for license}
interface

type

  TItemStream = class
  public
    FlowToNextLine: Boolean;
    function Line:Integer; virtual; abstract;
    function Column: Integer; virtual; abstract;
    function GetFreeItem: Boolean; virtual; abstract;
    function Token: string; virtual; abstract;
    function IsIntegerToken(var Value: Integer):Boolean; virtual; abstract;
    function IsRealToken(var Value: Double):Boolean; virtual; abstract;
    function IsEmptyToken: Boolean; virtual; abstract;
    function NextLine: Boolean; virtual; abstract;
    function AtEndOfFile: Boolean; virtual; abstract;
    procedure SkipBlankLines; virtual; abstract;
    procedure Rewind; virtual; abstract;
  end;

function SetupImportStream(var FileName: string): TItemStream;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFNDEF CONSOLE} Clipbrd, WCompat, {$ENDIF}
     Classes, SysUtils, UCompat, Huge, ExpPlotLoader, SSLoader, Worker, Units;

var
  CF_Biff8: Cardinal = 0;

function SetupImportStream(var FileName: string): TItemStream;
var
  C: THugeMemoryStream;
  S: TMemoryStream;
  Stream: TStream;
  Ext: string;
  i: Integer;
  PlotLoader: TSpreadSheetFileLoader;
begin
  Result := nil;
  Stream := nil;
  if FileName = '' then
    FileName := '<clipboard>';
  try
    if Assigned(SetHandleFromClipboard) and (CompareText(FileName, '<clipboard>') = 0) then begin
      C := THugeMemoryStream.Create;
      Stream := C;
      {$IFDEF MSWINDOWS} {$IFNDEF CONSOLE}
      if CF_Biff8 = 0 then
        CF_Biff8 := RegisterClipboardFormat('Biff8');
      if Clipboard.HasFormat(CF_Biff8) then begin
        FileName := '<XLS clipboard>';
        C.SetFromClipboard(CF_Biff8)
      end else
      {$ENDIF} {$ENDIF}
        C.SetFromClipboard(CF_PREFTEXT);
    end else begin
      S := TMemoryStream.Create;
      Stream := S;
      UnlockedLoadFromFile(S, FileName);
    end;
    Ext := LowerCase(ExtractFileExt('a'+FileName)); { Allows Filename to be simply '.ext' }
    for i := High(PlotLoaders) downto 0 do
      if PlotLoaders[i].InheritsFrom(TSpreadSheetFileLoader) then begin
        PlotLoader := PlotLoaders[i].Create(nil, Stream, Ext, cm1, Arbitrary) as TSpreadSheetFileLoader;
        try
          Result := PlotLoader.MakeItemStream(Log);
          if Result <> nil then begin
            PlotLoader := nil; { Keep it }
            Exit;
          end;
        finally
          PlotLoader.Free;
        end;
      end;
  finally
    Stream.Free;
  end;
end;

end.
