unit OvrLoader;
{ Copyright (C) 2008 C M Western; see NLFit.pas for license}
interface

{ Loader for .ovr files from PGOPHER }

uses ExpPlotLoader, TextStrm, Channel, pgDom;

type

  TOvrFileLoader = class(TExperimentalPlotLoader)
  private
    nPlots: Integer;
  public
    function Load: Boolean; override;
    procedure ReadExpPlot(Child: TDOMNode);
    procedure ReadData(Child: TDOMCDATASection);
    procedure ReadPeaks(Child: TDOMCDATASection);
    destructor Destroy; override;
  end;


implementation

uses Classes, pgXMLRead, SysUtils, Math, TabbedLine, Peak;

{ TOvrFileLoader }

destructor TOvrFileLoader.Destroy;
begin

  inherited;
end;

function TOvrFileLoader.Load: Boolean;
var
  Position: Int64;
  Buffer: string;
  TryLoad: Boolean;
  Child: TDOMNode;
begin
  Result := False;
  Position := Stream.Position;
  if Ext = '' then begin
    SetLength(Buffer, 60);
    SetLength(Buffer, Stream.Read(Buffer[1],Length(Buffer)));
    TryLoad := (Pos(Buffer, '<Overlays ') > 0) or (Pos(Buffer, '<ExperimentalPlot ') > 0);
  end else
    TryLoad := Ext = '.ovr';
  Stream.Position := Position;
  if TryLoad then begin
    ReadXMLFile(XMLDocument, Stream);
    Result := True;
    nPlots := 0;
    IndependentFrequencyScales := True;
    Child := XMLDocument.FirstChild;
    if Child.NodeName = 'Overlays' then begin
      Child := Child.FirstChild;
      while Assigned(Child) and (Child.NodeName = 'ExperimentalPlot') do begin
        ReadExpPlot(Child);
        Child := Child.NextSibling;
      end;
    end else if XMLDocument.FirstChild.NodeName = 'ExperimentalPlot' then
      ReadExpPlot(Child);
  end;
end;

procedure TOvrFileLoader.ReadData(Child: TDOMCDATASection);
var
  s: AnsiString;
  P: PAnsiChar;
  nFields, nPoints, nWeights, cLength: Integer;
  Intensity, Frequency, Weights: Double;
begin
  s := Child.Data;
  P := @s[1];
  nPoints := 0;
  nWeights := 0;
  cLength := -1;
  Connector.nPlots := nPlots+1;
  Connector.SetLineList(nPlots-1,False);
  while P^ <> #0 do begin
    if nPoints >= cLength-1 then begin
      if cLength = 0 then
        cLength := Length(s) div 40 + 10 { Guess at length }
      else
        cLength := cLength + 10000;
      Connector.SetLength(nPlots,cLength,nWeights > 0, True, True);
//      SetLength(Intensity,Length(Frequency));
//      SetLength(Weights,Length(Frequency));
    end;
    Weights := 1;
    P := TryReadTabbedLine(tlAny, P, nil, [@Frequency,@Intensity,@Weights], nFields, EmptyArray);
    case nFields of
      3: Inc(nWeights);
      1: Intensity := 1;
    end;
    if nFields > 0 then begin
      Inc(nPoints);
      SetValue(nPlots,0,nPoints,Frequency);
      SetValue(nPlots,1,nPoints,Intensity);
      if nWeights > 0 then
        SetValue(nPlots,0,nPoints,Weights);
    end;
    if P = nil then
      Break
    else if not SkipLineEnding(P) then
      FileErr('Unexpected character in overlay data:'+P^);
  end;
  Inc(nPlots);
  Connector.SetLength(nPlots,nPoints,nWeights > 0, True, True);
end;

procedure TOvrFileLoader.ReadExpPlot(Child: TDOMNode);
begin
  Child := Child.FirstChild;
  while Assigned(Child) do begin
    if Child is TDOMCDATASection then begin
      ReadData(TDOMCDATASection(Child));
      Child := Child.NextSibling;
      if Child is TDOMCDATASection then
        ReadPeaks(TDOMCDATASection(Child));
      Exit;
    end;
    Child := Child.NextSibling;
  end;
end;

procedure TOvrFileLoader.ReadPeaks(Child: TDOMCDATASection);
var
  Strings: TStringList;
  i, j: Integer;
  Peak: TPeak;
begin
  Strings := TStringList.Create;
  try
    Strings.Text := TDOMCDATASection(Child).Data;
    Child.ParentNode.RemoveChild(Child);
    Connector.nPlots := nPlots+1;
    Connector.SetLineList(nPlots,True);
    j := 0;
    for i := 0 to Strings.Count-1 do
      if TryStrToPeak(Strings[i], Peak) then begin
        if not IsNan(Peak.Position) then begin
          Connector.SetValue(nPlots,0,j,Peak.Position);
          Connector.SetValue(nPlots,1,j,Peak.Actual);
          Connector.SetValue(nPlots,3,j,Peak.Residual);
          Connector.SetValue(nPlots,2,j,Peak.Weight);
          Inc(j);
        end;
      end else
        FileErr('Bad line in peak list:'+Strings[i]);
  finally
    Strings.Free;
  end;
end;

initialization
  RegisterPlotLoader(TOvrFileLoader, 'Overlay Files', '*.ovr');
end.
