unit PSLoader;
{ Copyright (C) 2007 C M Western; see NLFit.pas for license}
interface

{ Import OceanOptics .ProcSPec files }

uses ZipLoader;

type

  { TProcSpecFileLoader }

  TProcSpecFileLoader = class(TZipFileLoader)
    function Load: Boolean; override;
  end;

implementation

uses SysUtils, Math, pgDOM, pgXMLRead, MathSup, Units, ExpPlotLoader;

{ TProcSpecFileLoader }

function TProcSpecFileLoader.Load: Boolean;
  function ReadValues(Node: TDOMNode; Col: Integer):Integer;
  var
    Child, e: TDOMNode;
    s: string;
  begin
    Result := 0;
    Child := Node.FirstChild;
    while Assigned(Child) do begin
      if Col >= 0 then begin
        e := (Child as TDOMElement).FirstChild;
        s := e.NodeValue;
        SetValue(-1,Col,Result,StrToFloat(s));
      end;
      Inc(Result);
      Child := Child.NextSibling;
    end;
  end;
  function DeepFindNode(Node: TDOMNode_WithChildren; const AName: string):TDomNode;
  var
    Child: TDomNode;
  begin
    Result := Node.FindNode(AName);
    Child := Node.FirstChild;
    while not Assigned(Result) and Assigned(Child) do begin
      if Child is TDOMNode_WithChildren then
        Result := DeepFindNode(TDOMNode_WithChildren(Child),AName);
      if Assigned(Result) and (ReadValues(Result,-1) < 4) then
        Result := nil; { Ignore nodes that are just references to other nodes }
      Child := Child.NextSibling;
    end;
  end;
var
  XNode, YNode: TDOMNode_WithChildren;
begin
  Result := False;
  if (Ext = '') or (Ext = '.procspec') then begin
    FirstEntry;
    while NextEntry do begin
      if (CompareText(Copy(CurrentFileName,1,2), 'ps') = 0) or (CentralHdr.Uncompressed_Size > 3000) then begin
        PrepareStream;
        Break;
      end;
    end;
    if DStream = nil then
      Exit;
    pgXMLRead.ReadXMLFile(XMLDocument, DStream, CurrentFileName, StreamSize);
    XNode := DeepFindNode(XMLDocument, 'channelWavelengths') as TDOMNode_WithChildren;
    Connector.SetUnits(-1,nmAir);
    YNode := DeepFindNode(XMLDocument, 'processedPixels') as TDOMNode_WithChildren;
    if not Assigned(YNode) then
      YNode := DeepFindNode(XMLDocument, 'pixelValues') as TDOMNode_WithChildren;
    if not (Assigned(XNode) and Assigned(YNode)) then
      Exit;
    Result := True;
    SetupChannels(1,ReadValues(XNode,-1),False);
    ReadValues(XNode, 0);
    ReadValues(YNode, 1);
  end else
    Exit; { Wrong extension }
end;

initialization
  RegisterPlotLoader(TProcSpecFileLoader, 'ProcSpec Files', '*.ProcSpec');
end.

