unit Peak;
{ Copyright (C) 2008 C M Western; see NLFit.pas for license}
interface

uses Math;

type

  TPeak = record
    Position: Double;
    Actual:   Double;
    Residual: Double;
    Weight:   Double;
  end;

function PeakToString(Peak: TPeak):string;
function TryStrToPeak(const s: AnsiString; out Peak: TPeak):Boolean;

const
  EmptyPeak: TPeak = (Position:Nan; Actual:Nan; Residual: Nan; Weight:1.0);

implementation

uses SysUtils, MathSup, TabbedLine;

function PeakToString(Peak: TPeak):string;
  function Add(const V: Double): string;
  begin
    Result := '';
    if not IsNan(V) then
      Result := FloatToStr(V);
  end;
var
  NonDefaultWeight: Boolean;
begin
  with Peak do begin
    if IsNan(Weight) then
      Weight := 0;
    Result := Add(Position);
    NonDefaultWeight := Weight <> 1;
    if not IsNan(Actual) or NonDefaultWeight then begin
      Result := Result + #9 + Add(Actual);
      if NonDefaultWeight then
        Result := Result + #9 + Add(Weight);
    end;
  end;
end;

function TryStrToPeak(const s: AnsiString; out Peak: TPeak):Boolean;
var
  nFields: Integer;
begin
  Peak := EmptyPeak;
  Result := nil = TryReadTabbedLine(tlAny, PAnsiChar(s),nil,[@Peak.Position, @Peak.Actual, @Peak.Weight],nFields,EmptyArray);
  if Result and (Peak.Weight = 0) then
    SetNan(Peak.Weight);
end;

end.
