unit WindowsXLSLoader;
{$I utils.inc}{ Copyright (C) 2016 C M Western; see NLFit.pas for license}
interface

uses Windows, ActiveX,
     ExpPlotLoader, Classes, SSLoader;

type

  TLockBytes = class(TInterfacedObject, ILockBytes)
    Stream: TStream;
    InitialPosiition: Int64;
    constructor Create(TheStream: TStream);
    function ReadAt(ulOffset: Largeint; pv: Pointer; cb: Longint; pcbRead: PLongint): HResult; stdcall;
    function WriteAt(ulOffset: Largeint; pv: Pointer; cb: Longint; pcbWritten: PLongint): HResult; stdcall;
    function Flush: HResult; stdcall;
    function SetSize(cb: Largeint): HResult; stdcall;
    function LockRegion(libOffset: Largeint; cb: Largeint; dwLockType: Longint): HResult; stdcall;
    function UnlockRegion(libOffset: Largeint; cb: Largeint; dwLockType: Longint): HResult; stdcall;
    function Stat(out statstg: TStatStg; grfStatFlag: Longint): HResult; stdcall;
  end;

  TXLSFileLoader = class(TSpreadSheetFileLoader)
    type
      TSpreadsheetFormat = (sfUnknwown=0, sfExcel2, sfExcel5, sfExcel8);
  private
    SpreadsheetFormat: TSpreadsheetFormat;
    XLSStream: TStream;
    LockBytes: TLockBytes;
    Storage: IStorage;
    OLEStream: TStream;
    EOF: Boolean;
    BIFFversion: Integer;
    function FindStream: TStream;
    function NextWord: Word;
    function NextDWORD: DWORD;
    function NextByte: Byte;
    function NextDouble: Double;
    function NextRK: Double;
    function NextString(ByteCount: Boolean):string;
    procedure ProcessStream;
  public
    destructor Destroy; override;
    function Load: Boolean; override;
  end;

implementation

uses SysUtils, Math, UCompat, Worker;

type

  TIStream = class(TStream)
    Stream: IStream;
    constructor Create(AStream: IStream);
    function Read(var Buffer; Count: Longint): Longint; override;
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;
  end;

{ TXLSFileLoader }

const
  BIFF2header: string[4] = #9#0#4#0;

destructor TXLSFileLoader.Destroy;
begin
  OLEStream.Free;
  inherited;
end;

function TXLSFileLoader.FindStream: TStream;
var
  Header: string[4];
  Position: Int64;
  ElementsRead: Integer;
  StatStg: TStatStg;
  IEnum: IEnumStatStg;
  IStrm: IStream;
  EName: string;
begin
  Result := nil;
  Position := Stream.Position;
  if Stream.Read(Header[1], 4) = 4 then begin
    if BIFF2header = Header then begin
      SpreadsheetFormat := sfExcel2;
      Result := Stream;
      Exit;
    end;
  end;
  Stream.Position := Position;
  LockBytes := TLockBytes.Create(Stream);
  if StgOpenStorageOnILockBytes(LockBytes, nil, STGM_DIRECT_SWMR or STGM_READ or STGM_SHARE_DENY_NONE, nil, 0, Storage) = S_OK then begin
    if Storage.EnumElements(0, nil, 0, IEnum) = S_OK then begin
      repeat
        IEnum.Next(1, StatStg, @ElementsRead);
        if ElementsRead = 1 then begin
          EName := WideCharToString(StatStg.pwcsName);
          // WriteLn(Log, StatStg.dwType, ' ', StatStg.cbSize, ' ', EName);
          if EName = 'Book' then
            SpreadsheetFormat := sfExcel5
          else if EName = 'Workbook' then
            SpreadsheetFormat := sfExcel8;
          if (SpreadsheetFormat <> sfUnknwown) and (StatStg.dwType = STGTY_STREAM) then begin
            Storage.OpenStream(StatStg.pwcsName, nil, STGM_SHARE_EXCLUSIVE or STGM_READ {or STGM_SHARE_DENY_NONE}, 0, IStrm);
            if IStrm <> nil then begin
              OLEStream := TIStream.Create(IStrm);
              Result := OLEStream;
              Exit;
            end;
          end;
          SpreadsheetFormat := sfUnknwown;
        end;
      until ElementsRead = 0;
    end;
  end else
    FreeAndNil(LockBytes);
end;

function TXLSFileLoader.Load: Boolean;
begin
  Result := False;
  if (Ext = '') or (Ext = '.xls') then begin
    XLSStream := FindStream;
    if XLSStream = nil then
      Exit;
    Result := True;
    ClearSpreadSheet;
    ProcessStream;
    ProcessSpreadSheet;
  end;
end;

function TXLSFileLoader.NextByte: Byte;
begin
  EOF := XLSStream.Read(Result, SizeOf(Result)) <> SizeOf(Result);
  if EOF then
    Result := 0;
end;

function TXLSFileLoader.NextDouble: Double;
begin
  EOF := XLSStream.Read(Result, SizeOf(Result)) <> SizeOf(Result);
  if EOF then
    Result := 0;
end;

function TXLSFileLoader.NextDWORD: DWORD;
begin
  EOF := XLSStream.Read(Result, SizeOf(Result)) <> SizeOf(Result);
  if EOF then
    Result := 0;
end;

function TXLSFileLoader.NextRK: Double;
var
  D: DWORD;
begin
  D := NextDWORD;
  if D and 2 = 0 then begin
    Result := 0;
    PDWORD(PAnsiChar(@Result)+4)^ := D and not 3;
  end else
    Result := Integer(D and not 3) div 4; { Integer }
  if D and 1 = 1 then
    Result := Result / 100;
end;

function TXLSFileLoader.NextString(ByteCount: Boolean): string;
var
  l: Integer;
  A: AnsiString;
  Flags: Byte;
begin
  if ByteCount then
    l := NextByte
  else
    l := NextWord;
  if BIFFversion < 8 then begin
    SetLength(A,l);
    XLSStream.Read(A[1],l);
    Result := A
  end else begin
    Flags := NextByte;
    if Flags and 8 <> 0 then
      NextWord;
    if Flags and 4 <> 0 then
      NextDWORD;
    if Flags and 1 = 0 then begin
      SetLength(A,l);
      XLSStream.Read(A[1],l);
      Result := A;
    end else begin
      {$IFDEF UNICODE}
        SetLength(Result, l);
        XLSStream.Read(Result[1], 2*l);
      {$ELSE}
        SetLength(A,2*l);
        XLSStream.Read(A[1],2*l);
        Result := WideCharLenToString(PWideChar(@A[1]),l);
      {$ENDIF}
    end;
  end;
end;

function TXLSFileLoader.NextWord: Word;
begin
  EOF := XLSStream.Read(Result, SizeOf(Result)) <> SizeOf(Result);
  if EOF then
    Result := 0;
end;

procedure TXLSFileLoader.ProcessStream;
var
  recType: Word;
  recSize: Word;
  recStart: Int64;
  InSheet: Boolean;
  nSheets, iV, nSharedStrings: Integer;
  SheetStart: array of DWORD;
  RowIndex, ColIndex: Integer;
  fV: Double;
  sV: string;
  i: Integer;
  procedure ShowRec(Fmt: string; Values: array of const);
  begin
    WriteLn(Log^, recStart-4:10, ' ', IntToHex(recType,4), recSize:6, ' ', Format(Fmt, Values));
  end;
  procedure ShowCell(const Fmt: string; Values: array of const);
  var
    s: string;
  begin
    if Verbose then
      ShowRec('[%d,%d] %s', [RowIndex, ColIndex, Format(Fmt, Values)]);
    if Length(Values) = 0 then
      Exit;
    with Values[0] do
     case VType of
      vtChar:       s := VChar;
      vtWideChar:   s := VWideChar;
      vtString:     s := VString^;
      vtAnsiString: s := PAnsiChar(VAnsiString);
      {$IFDEF UNICODE}
      vtUnicodeString: S := PWideChar(VUnicodeString);
      {$ENDIF}
      vtExtended:   begin SS.AddValue(RowIndex, ColIndex, VExtended^); Exit; end;
     else
       raise EReportException.Create('Help');
     end;
     SS.AddString(RowIndex, ColIndex, s);
  end;
  procedure ReadRowCol;
  begin
    RowIndex := NextWord;
    ColIndex := NextWord;
    NextWord; { Skip formatting }
    if BIFFversion = 2 then
      NextByte;
  end;
  function NextRecord:Boolean;
  begin
    XLSStream.Position := recStart + recSize;
    recType := NextWord;
    recSize := NextWord;
    recStart := XLSStream.Position;
    Result := not EOF
  end;
  function ErrStr(Index: Integer):string;
  begin
    case Index of
      $00: Result := '#NULL!';
      $07: Result := '#DIV/0!';
      $0F: Result := '#VALUE!';
      $17: Result := '#REF!';
      $1D: Result := '#NAME?';
      $24: Result := '#NUM!';
      $2A: Result := '#N/A';
    else
      Result := '???';
    end;
  end;
begin
  nSheets := 0;
  nSharedStrings := 0;
  recStart := XLSStream.Position;
  recSize := 0;
  while NextRecord do begin
    case recType of
      $0009, $0209, $0409, $0809:
        begin
          iV := NextWord;
          if BIFFversion = 0 then
            case recType of
              $9:    BIFFversion := 2;
              $0209: BIFFversion := 3;
              $0409: BIFFversion := 4;
              $0809: if iV = $0500 then
                       BIFFversion := 5
                     else
                       BIFFversion := 8;
            end;
          InSheet := NextWord = $10;
          if InSheet then begin
            SS.Sheet := 0; { Default }
            for i := 0 to nSheets - 1 do
              if SheetStart[i] = recStart-4 then begin
                SS.Sheet := i;
                Break;
              end;

          end;
          if Verbose then
            ShowRec('BOF %d %s of sheet %d', [BIFFversion,BoolToStr(InSheet,True),SS.Sheet]);
        end;
      $000A: if Verbose then
               ShowRec('EOF',[]);
      $0085:
        begin
          SetLength(SheetStart,nSheets+1);
          SheetStart[nSheets] := NextDWORD;
          NextByte;
          SS.Sheet := nSheets; { Trigger allocation }
          if NextByte = 0 then begin
            SS.SheetNames[nSheets] := NextString(True);
            if Verbose then
              ShowRec('SHEET %d %s',[SheetStart[nSheets], SS.SheetNames[nSheets]]);
          end else if Verbose then
            ShowRec('SHEET',[]);
          Inc(nSheets);
        end;
      $0002: begin; ReadRowCol; ShowCell('INTEGER %d',[NextWord]); end;
      $0003, $203:
        begin
          ReadRowCol;
          ShowCell('NUMBER %g',[NextDouble]);
        end;
      $027E:
        begin
          ReadRowCol;
          ShowCell('RK %g',[NextRK]);
        end;
      $00BD:
        begin
          ReadRowCol;
          for i := 0 to (recSize - 6) div 6 - 1 do begin
            ShowCell('RK %g',[NextRK]);
            NextWord;
            Inc(ColIndex);
          end;
        end;
      $0004, $204:
        begin
          ReadRowCol;
          sV := NextString(recType = $0004);
          ShowCell('LABEL %s', [sV]);
        end;
      $00FC:
        begin
          NextDWORD;
          nSharedStrings := NextDWORD;
          SetLength(SS.SharedStrings, nSharedStrings);
          for i := 0 to nSharedStrings - 1 do
            SS.SharedStrings[i] := NextString(False);
          if Verbose then          
            ShowRec('SST %d', [nSharedStrings]);
        end;
      $00FD:
        begin
          ReadRowCol;
          iV := NextDWORD;
          if iV < nSharedStrings then
            SS.AddSharedString(RowIndex, ColIndex, iv)
          else
            ShowCell('LABELSST %s', ['???']);
        end;
      $00D6:
        begin
          ReadRowCol;
          sV := NextString(True);
          ShowCell('RSTRING %s', [sV]);
        end;
      $0005, $205:
        begin
          ReadRowCol;
          iV := NextByte;
          if NextByte = 0 then
            ShowCell('BOOLERR %s', [BoolToStr(iV <> 0, True)])
          else
            ShowCell('BOOLERR %s', [ErrStr(iV)]);
        end;
      $0006, $206, $406:
        begin
          ReadRowCol;
          fV := NextDouble;
          if (BIFFversion > 2) and (PWORD(PAnsiChar(@fv)+6)^ = $FFFF) then
            case PAnsiChar(@fv)^ of
              #0: { String record following }
                  while NextRecord do
                    case recType of
                      $0007, $0207:
                        begin
                          sV := NextString(recType = $0007);
                          ShowCell('FORMULA %s', [sV]);
                          Break;
                        end;
                      $0021, $0221: { ARRAY } ;
                      $04BC: { SHAREDFMLA } ;
                    else
                      { Putback record }
                      XLSStream.Position := recStart - 4;
                      recSize := 0;
                      Break;
                    end;
              #1: ShowCell('FORMULA %s',[BoolToStr((PAnsiChar(@fv)+2)^ <> #0,True)]); { Boolean }
              #2: ShowCell('FORMULA %s',[ErrStr(Ord((PAnsiChar(@fv)+2)^))]);{ Error }
              #3: ShowCell('FORMULA',[]) { Empty Cell }
            end
          else
            ShowCell('FORMULA %g',[fV]);
        end;
    end;
  end;
  Inc(recStart,4);
  if Verbose then
    ShowRec('(End)',[]);
end;

{ TLockBytes }

constructor TLockBytes.Create(TheStream: TStream);
begin
  inherited Create;
  Stream := TheStream;
  InitialPosiition := Stream.Position;
end;

function TLockBytes.Flush: HResult;
begin
  Result := S_OK;
end;

function TLockBytes.LockRegion(libOffset, cb: Largeint; dwLockType: Integer): HResult;
begin
  Result := S_OK;
end;

function TLockBytes.ReadAt(ulOffset: Largeint; pv: Pointer; cb: Integer; pcbRead: PLongint): HResult;
var
  n: Longint;
begin
  Result := S_OK;
  Stream.Position := ulOffset + InitialPosiition;
  n := Stream.Read(pv^, cb);
  if pcbRead <> nil then
    pcbRead^ := n;
end;

function TLockBytes.SetSize(cb: Largeint): HResult;
begin
  Result := STG_E_ACCESSDENIED;
end;

function TLockBytes.Stat(out statstg: TStatStg; grfStatFlag: Integer): HResult;
begin
  FillChar(statstg, SizeOf(statstg), 0);
  statstg.dwType := STGTY_LOCKBYTES;
  statstg.cbSize := Stream.Size - InitialPosiition;
  Result := S_OK;
end;

function TLockBytes.UnlockRegion(libOffset, cb: Largeint; dwLockType: Integer): HResult;
begin
  Result := S_OK;
end;

function TLockBytes.WriteAt(ulOffset: Largeint; pv: Pointer; cb: Integer; pcbWritten: PLongint): HResult;
begin
  Result := E_FAIL;
end;

{ TIStream }

constructor TIStream.Create(AStream: IStream);
begin
  inherited Create;
  Stream := AStream;
end;

function TIStream.Read(var Buffer; Count: Integer): Longint;
begin
  Stream.Read(@Buffer, Count, Pointer(@Result));
end;

function TIStream.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
begin
  Stream.Seek(Offset, Ord(Origin), {$IFDEF UNICODE}LargeUInt(Result){$ELSE}Result{$ENDIF})
end;

initialization
  RegisterPlotLoader(TXLSFileLoader, 'Excel Files', '*.xls');
end.
