unit XLSLoader;
{$I utils.inc}{ Copyright (C) 2016 C M Western; see NLFit.pas for license}
interface

uses ExpPlotLoader, Classes, SSLoader;

type

  TXLSFileLoader = class(TSpreadSheetFileLoader)
  public
    function Load: Boolean; override;
  end;

implementation

uses xlsbiff2, xlsbiff5, xlsbiff8, xlsxooxml,
     fpsopendocument, { TODO: Horribly slow; speed up or use custom reader }
     fpspreadsheet, fpsTypes, fpsReaderWriter, SysUtils, Math, Worker;

type

  { TWorkbook }

  TWorkbook = class(TsWorkbook)
    function FindFormatID(Stream: TStream; const Ext: string):TsSpreadFormatID;
  end;

function TWorkbook.FindFormatID(Stream: TStream; const Ext: string): TsSpreadFormatID;
var
  Formats: TsSpreadFormatIDArray;
  Position: Int64;
begin
  Result := -1;
  if Stream.Size < 8 then
    Exit;
  if (Ext <> '') then begin
    Formats := GetSpreadFormatsFromFileName(faRead,'a'+Ext);
    if Length(Formats) = 0 then
      Exit;
    if Length(Formats) = 1 then begin
      Result := Formats[0];
      Exit;
    end;
  end;
  Position := Stream.Position;
  if not GetFormatFromFileHeader(Stream, Result) then begin
    Result := -1;
  end;
  Stream.Position := Position;
end;

{ TXLSFileLoader }

function TXLSFileLoader.Load: Boolean;
var
  WorkBook: TWorkbook;
  Worksheet: TsWorksheet;
  FormatID: TsSpreadFormatID;
  i: Integer;
  Row, Col: Cardinal;
  Cell: PCell;
begin
  Result := False;
  if (Ext = '') or (Ext = '.xls') or (Ext = '.ods') then begin
    Workbook := TWorkbook.Create;
    try
      FormatID := WorkBook.FindFormatID(Stream, Ext);
      if FormatID < 0 then
        Exit;
      Workbook.ReadFromStream(Stream,FormatID);
      if Workbook.GetWorksheetCount() < 1 then
        Exit;
      Result := True;
      ClearSpreadSheet;
      SS.Sheet := Workbook.GetWorksheetCount()-1;
      for i := 0 to Workbook.GetWorksheetCount() - 1 do begin
        Worksheet := Workbook.GetWorksheetByIndex(i);
        SS.Sheet := i;
        SS.SheetNames[i] := Worksheet.Name;
        SS.SetDim(Worksheet.GetLastRowIndex+1, Worksheet.GetLastColIndex+1);
        if Verbose then
          WriteLn(Log^, 'Worksheet ',i,' ', Worksheet.Name);
        for Row := 0 to Worksheet.GetLastRowIndex do
          for Col := 0 to Worksheet.GetLastColIndex do begin
            Cell := Worksheet.FindCell(Row, Col);
            if Cell <> nil then begin
              case Cell^.ContentType of
                cctFormula, cctNumber:  SS.AddValue(Row,Col,Worksheet.ReadAsNumber(Cell));
              else
                SS.AddString(Row, Col, Worksheet.ReadAsText(Cell));
              end;
            end;
          end;
      end;
    finally
      WorkBook.Free;
    end;
    ProcessSpreadSheet;
  end;
end;

initialization
  RegisterPlotLoader(TXLSFileLoader, 'Spreadsheet Files', '*.xls;*.ods');
end.
