unit XLSXLoader;
{$I utils.inc}{ Copyright (C) 2016 C M Western; see NLFit.pas for license}

interface

uses Classes, MathSup, ZipLoader;

type

  TXLSXFileLoader = class(TZipFileLoader)
  public
    function Load: Boolean; override;
  end;

implementation

uses Math, SysUtils, pgDOM, pgXMLRead, ExpPlotLoader;

{ TXLSXFileLoader }

function TXLSXFileLoader.Load: Boolean;
var
  i: Integer;
  nSheets: Integer;
  function GetAttribute(Node: TDOMNode; const Name: AnsiString; const Default: string = ''):string;
  var
    Attr: TDOMAttr;
  begin
    Result := Default;
    if not (Node is TDOMElement) then
      Exit;
    Attr := TDOMElement(Node).GetAttributeNode(Name);
    if Attr <> nil then
      Result := Attr.Value;
  end;
  function GetRowCol(var s: string; out Row, Col: Integer):Boolean;
  begin
    Result := False;
    Row := 0;
    Col := 0;
    while (s <> '') and (s[1] in ['A'..'Z']) do begin
      Col := Col*26 + Ord(s[1]) - Ord('A');
      Delete(S,1,1);
      Result := True;
    end;
    if not Result then
      Exit;
    Result := False;
    while (s <> '') and (s[1] in ['0'..'9']) do begin
      Row := Row*10 + Ord(s[1]) - Ord('0');
      Delete(S,1,1);
      Result := True;
    end;
    if Result then
      Dec(ROw);
  end;
  function ReadSharedStrings:Boolean;
  var
    Child, C: TDOMNode;
  begin
    Result := False;
    if not ReadXMLFile('xl/sharedStrings.xml') then
      Exit;
    Child := XMLDocument.FirstChild;
    if (Child.NodeName = 'sst') and (Child is TDOMNode_WithChildren) then begin
      Child := TDOMNode_WithChildren(Child).FirstChild;
      while Child <> nil do begin
        C := Child;
        while C is TDOMNode_WithChildren do
          C := TDOMNode_WithChildren(C).FirstChild;
        if C is TDOMText then
          SS.AddShared(TDOMText(C).Data)
        else
          SS.AddShared(' ');
        Child := Child.NextSibling;
      end;
    end;
    Result := True;
  end;
  function ReadSheetIDs: Boolean;
  var
    Child: TDOMNode;
  begin
    Result := False;
    nSheets := 0;
    if not ReadXMLFile('xl/workbook.xml') then
      Exit;
    Child := FindNode(XMLDocument, 'sheets');
    if (Child = nil) or not (Child is TDOMNode_WithChildren) then
      Exit;
    Child := TDOMNode_WithChildren(Child).FirstChild;
    while Child <> nil do begin
      SS.Sheet := nSheets;
      SS.SheetNames[nSheets] := GetAttribute(Child, 'name');
      Child := Child.NextSibling;
      Inc(nSheets);
    end;
    Result := True;
  end;
  function ReadSheet(SheetID: Integer):Boolean;
  var
    Child, C, V: TDOMNode;
    s, CellType, CellValue: string;
    Row, Col: Integer;
  begin
    Result := False;
    SS.Sheet := SheetID;
    if not ReadXMLFile('xl/worksheets/sheet'+IntToStr(i+1)+'.xml') then
      Exit;
    Result := True;
    Child := FindNode(XMLDocument, 'dimension');
    if Child <> nil then begin
      s := GetAttribute(Child, 'ref');
      if GetRowCol(s,Row,Col) and (Copy(s,1,1) = ':') then begin
        Delete(s,1,1);
        if GetRowCol(s,Row,Col) then
          SS.SetDim(Row+1,Col+1);
      end;
    end;
    Child := FindNode(XMLDocument, 'sheetData');
    if (Child = nil) or not (Child is TDOMNode_WithChildren) then
      Exit;
    Child := TDOMNode_WithChildren(Child).FirstChild;
    while Child <> nil do begin
      if (Child.NodeName = 'row') and  (Child is TDOMNode_WithChildren) then begin
        C := TDOMNode_WithChildren(Child).FirstChild;
        while C <> nil do begin
          s := GetAttribute(C,'r');
          CellValue := '';
          if GetRowCol(s,Row,Col) then begin
            CellType := GetAttribute(C,'t','n');
            V := C.FindNode('v');
            if V is TDOMNode_WithChildren then
              V := TDOMNode_WithChildren(V).FirstChild;
            if V is TDOMText then
              CellValue := TDOMText(V).Data;
          end;
          if CellValue <> '' then
            if CellType = 'n' then
              SS.AddValue(Row, Col, StrToFloat(CellValue))
            else if CellType = 's' then
              SS.AddSharedString(Row, Col, StrToInt(CellValue))
            else
              SS.AddString(Row, Col, CellValue);
          C := C.NextSibling;
        end;
      end;
      Child := Child.NextSibling;
    end;
  end;
begin
  Result := False;
  if (Ext = '') or (Ext = '.xlsx') then begin
    if not ReadHeader then
      Exit;
    ClearSpreadSheet;
    if not ReadSharedStrings then
      Exit;
    if not ReadSheetIDs then
      Exit;
    SS.Sheet := nSheets-1; { Allocates arrays }
    for i := 0 to nSheets-1 do
      Result := ReadSheet(i) or Result;
    ProcessSpreadSheet;
  end;
end;

initialization
  RegisterPlotLoader(TXLSXFileLoader, 'Excel Files', '*.xlsx');
end.
