unit ApplyFrm;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, MNode, Buttons, Overlay;

type

  { TApplyForm }

  TApplyForm = class(TForm)
    Label1: TLabel;
    OverlayComboBox: TComboBox;
    OKButton: TButton;
    CancelButton: TButton;
    CheckBox: TCheckBox;
    procedure FormShow(Sender: TObject);
  public
    procedure StuffComboBox(MNode: TMNode; RequiredClass: TMNodeClass; IncludeAll: Boolean; Exclude: TMNode = nil; MustBeLineList: Boolean = False);
    function UnStuffComboBox(MNode: TMNode; RequiredClass: TMNodeClass; IncludeAll: Boolean; Exclude: TMNode = nil):TMNode;
  end;

function PromptForOverlay(Owner: TCustomForm; TheCaption: string; var Flag: Boolean; FlagName: string; IncludeAll: Boolean;
                          RequiredClass: TMNodeClass = nil; Exclude: TMNode = nil; MustBeLineList: Boolean = False):TMNode;

implementation

uses Global, UCompat, ExpPlot, SizeHelpers;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TApplyForm }

procedure TApplyForm.FormShow(Sender: TObject);
begin
  {$IFDEF LCL_}
    ConstantSizeButtons(Self);
  {$ENDIF}
end;

procedure TApplyForm.StuffComboBox(MNode: TMNode; RequiredClass: TMNodeClass; IncludeAll: Boolean; Exclude: TMNode; MustBeLineList: Boolean);
var
  I: Integer;
begin
  OverlayComboBox.Clear;
  if IncludeAll then
    OverlayComboBox.Items.Add('All');
  for I := 0 to MNode.ComponentCount-1 do
    if (MNode.Components[i] <> Exclude) and (MNode.Components[i].ClassType = RequiredClass) then
      if not MustBeLineList or ( (MNode.Components[i] is TExperimentalPlot) and TExperimentalPlot(MNode.Components[i]).IsLineList )  then      
        OverlayComboBox.Items.Add(TMNode(MNode.Components[i]).Name);
  OverlayComboBox.ItemIndex := 0;
end;

function TApplyForm.UnStuffComboBox(MNode: TMNode; RequiredClass: TMNodeClass; IncludeAll: Boolean; Exclude: TMNode): TMNode;
var
  I, J: Integer;
begin
  I := OverlayComboBox.ItemIndex;
  if IncludeAll then
    if I = 0 then begin
      Result := MNode;
      Exit;
    end else
      Dec(I);
  for J := 0 to MNode.ComponentCount-1 do
    if (MNode.Components[J] <> Exclude) and (MNode.Components[J].ClassType = RequiredClass) then begin
      Dec(I);
      if I < 0 then begin
        Result := MNode.Components[J] as TMNode;
        Exit;
      end;
    end;
  Result := nil;
end;

function PromptForOverlay(Owner: TCustomForm; TheCaption: string; var Flag: Boolean; FlagName: string; IncludeAll: Boolean;
                          RequiredClass: TMNodeClass = nil; Exclude: TMNode = nil; MustBeLineList: Boolean = False):TMNode;
begin
  Result := nil;
  with TApplyForm.Create(Owner) do
    try
      with CheckBox do
        if FlagName <> '' then begin
          Visible := True;
          Caption := FlagName;
          Checked := Flag
        end else
          Visible := False;
      StuffComboBox(GlobalOverlays, RequiredClass, IncludeAll, Exclude);
      if OverlayComboBox.Items.Count = 0 then
        raise EReportException.Create('No suitable overlays');
      Caption := TheCaption;
      if ShowModal = mrOk then begin
        Result := UnStuffComboBox(GlobalOverlays, RequiredClass, IncludeAll, Exclude);
        if FlagName <> '' then
          Flag := CheckBox.Checked;
      end
    finally
      Free
    end;
end;

end.
