unit AutoFitFrm;
{$I utils.inc}{Copyright (C) 2015-6 C M Western; see commandline.pas for license}
interface

uses
  {$IFNDEF FPC} Windows, {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms, ComCtrls,
  Dialogs, SettingsForm, StdCtrls, Grids, NGrid, ExtCtrls, ShellSort, AlEdit,
  EditExtr, NewSpin, MathSup, MNode, ExpPlot, ChildForm, Mixture, SizeHelpers,
  Settings, AutoFit;

type

  { TAutoFitForm }

  TAutoFitForm = class(TChildForm)
    Panel: TPanel;
    AcceptLabel: TLabel;
    WindowEdit: TEdit;
    SearchButton: TButton;
    Grid: TNumberGrid;
    KeepLabel: TLabel;
    KeepSpin: TSpinBox;
    TrialView: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    TotalTrials: TEdit;
    Label3: TLabel;
    FailView: TEdit;
    MaxBlendsLabel: TLabel;
    MaxBlendsSpinBox: TSpinBox;
    ResetButton: TButton;
    TestButton: TButton;
    ActiveObject: TComboBox;
    procedure GridDblClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure GridDrawCellContents(Sender: TObject; ACol, ARow: Longint;
      Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
    procedure MaxBlendsSpinBoxGetIntegerValue(Sender: TObject;
      var NewValue: Longint; TheText: string; var StringValid: Boolean);
    procedure MaxBlendsSpinBoxSetIntegerValue(Sender: TObject;
      NewValue: Longint; var NewText: string);
    procedure ResetButtonClick(Sender: TObject);
    procedure SearchButtonClick(Sender: TObject);
    procedure KeepSpinChange(Sender: TObject);
    procedure MaxBlendsSpinBoxChange(Sender: TObject);
    procedure WindowEditChange(Sender: TObject);
  protected
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    procedure HandleComponent(Component: TComponent; var PName, AValue, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
    procedure ShowStatus(ReturnOnAbort: Boolean);
  private
    PanelControls: TControlArray;
    AutoFitter: TAutoFitter;
    DiffColumnShown: Boolean;
    FitRect: TGridRect;
    procedure LoadAutoFitter(TheFitter: TAutoFitter);
    procedure LoadGrid;
  public
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure StuffSelector(MNode: TMNode); override;
    procedure UpdateStatus(Sender: TObject; Trial: Int64);
  end;

var
  AutoFitForm: TAutoFitForm;

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses Math, UCompat, Obs, Global, LineFrm, PeakFinders, Profile, FormUtils,
     Progress, Fitter, Worker, Units, Main, LogUnit, MatrixVec;

function TAutoFitForm.BaseNodeClass: TClass;
begin
  Result := TMixture
end;

procedure TAutoFitForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  if MNode = CurrentMNode then
    ShowItem(CurrentMNode);
end;

procedure TAutoFitForm.LoadAutoFitter(TheFitter: TAutoFitter);
begin
  with TheFitter do begin
    aW := Abs(StrToFloatEx(LineForm.AcceptEdit.Text)); { Acceptance Window }
    //dW := Abs(StrToFloatEx(WindowEdit.Text)); { Search window }
    //KeepOrder := KeepOrderCheckBox.Checked;
    //QuietFit := not FitDetailsCheckBox.Checked;
    //LimitRange := LimitRangeCheckBox.Checked;
    //MaxBlends := MaxBlendsSpinBox.Value;
    //MaxCycles := 20;
    OriginalParameters := nil;
  end;
end;

procedure TAutoFitForm.LoadGrid;
var
  i, j, nPar: Integer;
  ShowTrialNo: Boolean;
begin
  DiffColumnShown := False;
  if AutoFitter = nil then begin
    Grid.Clear;
    Exit;
  end;
  Grid.FixedSizeNumbers := True;
  AutoFitter.SetParametersFrom(High(Integer),False); { Force valid OriginalParameters }
  Grid.RowCount := AutoFitter.nBest+1;
  if AutoFitter.nBest > 0 then
    Grid.FixedRows := 1
  else
    Grid.Clear;
  if (AutoFitter.nFitted > 0) and (AutoFitter.nBest > 0) and (Length(AutoFitter.Best[0]) > 2) then
    nPar := Length(AutoFitter.OriginalParameters)
  else
    nPar := 0;
  ShowTrialNo := Length(AutoFitter.TrialNo) > 0;
  if ShowTrialNo then begin
    Grid.ColCount := nPar+5;
    Grid.Title[Grid.ColCount-1] := IntToStr(AutoFitter.nTrials); { For autosizing }
  end else
    Grid.ColCount := nPar+4;
  for i := 0 to nPar-1 do
    with AutoFitter.OriginalParameters[i] do
      Grid.Title[i+4] := MNode.UniqueName + ' ' + MNode.ParameterNames[Index];
  with AutoFitter do
    for i := 0 to nBest-1 do begin
      Grid.Value[i+1,1] := nCheckOK[i];
      for j := 2 to nPar+3 do
        Grid.Value[i+1,j] := Best[i][j-2];
      if ShowTrialNo then
        Grid.Value[i+1,Grid.ColCount-1] := TrialNo[i];
    end;
  Grid.AutoDefaultRowHeight;
  for i := 1 to Grid.ColCount - 1 do
    Grid.AutoAdjustColumn(i);
  if ShowTrialNo then
    Grid.Title[Grid.ColCount-1] := 'Trial';
end;

procedure TAutoFitForm.ShowItem(MNode: TMNode);
begin
  if MNode <> nil then
    BaseNode := MNode.RootNode;
  CurrentMNode := MNode;
  AutoFitter := CurrentMNode as TAutoFitter;
  if AutoFitter = nil then
    Exit;
  UpdateEditControl(WindowEdit, AutoFitter.SearchWindow);
  KeepSpin.Value := AutoFitter.Keep;
  MaxBlendsSpinBox.Value := AutoFitter.MaxBlends;
  TotalTrials.Text := IntToStr(AutoFitter.nTrials);
  UpdateStatus(AutoFitter,-1);
  LoadGrid;
end;

procedure TAutoFitForm.ShowStatus(ReturnOnAbort: Boolean);
begin
  UpdateStatus(nil,-1);
end;

function TAutoFitForm.StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer;
begin
  if MNode is TAutoFitter then begin
    ActiveObject.AddItem(Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end else
    Result := 0;
end;

procedure TAutoFitForm.StuffSelector(MNode: TMNode);
begin
  StuffSelectorHelper(MNode,ActiveObject);
  FComboBoxUpdated := False;
end;

procedure TAutoFitForm.UpdateStatus(Sender: TObject; Trial: Int64);
var
  nFail, i: Integer;
begin
  if not Assigned(AutoFitter) then
    Exit;
  if Length(AutoFitter.Fitters) > 0 then begin
    Trial := AutoFitter.Fitters[0].Trial;
    nFail := 0;
    for i := 0 to High(AutoFitter.Fitters) do
      Inc(nFail, AutoFitter.Fitters[i].nFail);
  end else
    nFail := AutoFitter.nFail;
  if AutoFitter.nTrials >= High(Integer) then
    SetProgress(Round(Trial/AutoFitter.nTrials*1e6))
  else
    SetProgress(Trial);
  FailView.Text := IntToStr(nFail);
  if Trial >= 0 then  
    TrialView.Text := IntToStr(Trial);
end;

procedure TAutoFitForm.WindowEditChange(Sender: TObject);
var
  v: Double;
  NText: string;
begin
  NText := Trim(WindowEdit.Text);
  if (NText <> '') and Visible and TryStrToFloatEx(NText, V) then
    AutoFitter.SearchWindow := v;
end;

procedure TAutoFitForm.MaxBlendsSpinBoxChange(Sender: TObject);
begin
  if Visible then
    AutoFitter.MaxBlends := MaxBlendsSpinBox.Value;
end;

procedure TAutoFitForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  if MNode = CurrentMNode then
    ShowItem(CurrentMNode);
end;

procedure TAutoFitForm.SearchButtonClick(Sender: TObject);
var
  nFail, i: Integer;
  Testing: Boolean;
  Problem: Exception;
begin
  Testing := Sender = TestButton;
  GlobalMixture.CheckProperties;

  LoadAutoFitter(AutoFitter);

  { Sorted copy of active linelist (in correct units) }
  AutoFitter.SetLineList(ActiveLinelist);

  LineForm.LoadObservations('', obLine, tkAll, False);
  with LineForm do begin
    FitRect := Grid.Selection;
    with Grid.Selection do
      if (Left = Right) and (Top = Bottom) then
        { No complete lines selected - no fitting process, just search }
      else begin
        for i := Top-1 to Bottom - 1 do
          Observations[i].Flags := Observations[i].Flags + [ofTrial];
      end;
  end;
  AutoFitter.SetObservations(Observations);
  nFail := AutoFitter.ComputeSearchRange;
  if nFail > 0 then
    if Testing then
      WriteLn(Log, nFail, ' Observations with no possible matches')
    else begin
      ShowMessage(IntToStr(nFail)+ ' Observations with no possible matches - showing closest match only');
      AutoFitter.Closest;
      TotalTrials.Text := '0';
      LoadGrid;
      Exit;
    end;

  if AutoFitter.nFitted > 0 then begin
    GlobalMixture.FillParameters;
    if Length(GlobalMixture.FittingParameters) > Length(AutoFitter.Observations) then
      raise EReportException.CreateFmt('More parameters (%d) than observations (%d)', [Length(GlobalMixture.FittingParameters), Length(AutoFitter.Observations)]);
    if Testing then begin
      ReShowLogForm;
      if AutoFitter.logNtrials < 5 then
        WriteLn(Log, AutoFitter.nTrials, ' trials')
      else
        WriteLn(Log, Power(10,Frac(AutoFitter.logNtrials)):5:2, 'E', floor(AutoFitter.logNtrials), ' trials');
    end else if AutoFitter.logNtrials > 18.9 then begin
      ShowMessage(Format('10^%.1g trials is too many', [AutoFitter.logNtrials]));
      Exit;
    end else if AutoFitter.logNtrials > 5 then
      if MessageDlg(Format('10^%.1g trials - do you want to continue?', [AutoFitter.logNtrials]), mtWarning, mbOKCancel, 0) <> mrOK then
        Exit;
    TotalTrials.Text := IntToStr(AutoFitter.nTrials);
    if Testing then begin
      AutoFitter.ShowSVD(True);
      Exit;
    end;
  end else if Testing then
    ShowMessage('Select some search lines');
  FailView.Text := '0';

  if AutoFitter.Verbose or AutoFitter.FitDetails then
    ReShow(LogForm);

  AutoFitter.nBest := 0;
  AutoFitter.OnUpdateStatus := {$IFDEF FPC}@{$ENDIF}UpdateStatus;
  LoadGrid;
  if AutoFitter.nTrials >= High(Integer) then
    PushProgress('Trials', 1000000) { Scale to useful range }
  else
    PushProgress('Trials', AutoFitter.nTrials);
  try
    BeginCalculation(True);
    Problem := nil;
    AutoFitter.Search(0, 1, Problem, {$IFDEF FPC}@{$ENDIF}ShowStatus);
    if Assigned(Problem) then
      raise Problem;
    if AutoFitter.nBest = 0 then begin
      ShowMessage('All trials failed - showing closest lines to selected transitions only');
      AutoFitter.Closest; { All trials failed - simply show closest }
    end;
    LoadGrid;
    UpdateStatus(AutoFitter,AutoFitter.nTrials);
    AutoFitter.WriteInfo(Log);
  finally
    PopProgress;
    ClearObservations;
    EndCalculation;
  end;
end;

procedure TAutoFitForm.FormResize(Sender: TObject);
begin
  WrapPanel(Panel, PanelControls);
end;

procedure TAutoFitForm.GridDrawCellContents(Sender: TObject; ACol,
  ARow: Longint; Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
var
  dCol: Integer;
begin
  dCol := Grid.ColCount-1;
  if DiffColumnShown then
    Dec(dCol);
  Grid.CurrentFixedSizeNumbers := (ACol > 1) and (ACol < dCol);
  if not Grid.CurrentFixedSizeNumbers then
    Grid.CurrentCellAlignment := taRightJustify;
end;

procedure TAutoFitForm.MaxBlendsSpinBoxGetIntegerValue(Sender: TObject; var NewValue: Longint; TheText: string; var StringValid: Boolean);
begin
  if TheText = 'any' then begin
    StringValid := True;
    NewValue := -1;
  end;
end;

procedure TAutoFitForm.MaxBlendsSpinBoxSetIntegerValue(Sender: TObject; NewValue: Longint; var NewText: string);
begin
  if NewValue = -1 then
    NewText := 'any';
end;

procedure TAutoFitForm.GridDblClick(Sender: TObject);
var
  iFit, i, dCol: Integer;
begin
  if AutoFitter = nil then
    Exit;
  iFit := Grid.Row-1;
  dCol := Grid.ColCount-1;
  if not DiffColumnShown then begin
    Inc(dCol);
    Grid.ColCount := dCol+1;
    Grid.Title[dCol] := 'nDiff';
    DiffColumnShown := True;
    Grid.AutoAdjustColumn(dCol);
  end;
  for i := 0 to AutoFitter.nBest-1 do
    if i = iFit then
      Grid.Valid[i+1,dCol] := False
    else
      Grid.Value[i+1,dCol] := AutoFitter.CountDifferences(iFit,i);
  if Length(AutoFitter.Best[iFit]) > 2 then
    AutoFitter.SetParametersFrom(iFit, False);
  LineForm.LoadAutoFitAssignment(AutoFitter, iFit);
end;

procedure TAutoFitForm.HandleComponent(Component: TComponent; var PName, AValue,
  PHint: string; var Save: Boolean; var SettingType: TSettingType;
  Write: Boolean);
begin
  if Component = ActiveObject then
    inherited;
end;

procedure TAutoFitForm.KeepSpinChange(Sender: TObject);
begin
  if Visible then
    AutoFitter.Keep := KeepSpin.Value;
end;

procedure TAutoFitForm.ResetButtonClick(Sender: TObject);
begin
  if AutoFitter = nil then
    Exit;
  AutoFitter.SetParametersFrom(-1,True);
  LineForm.LoadAutoFitAssignment(AutoFitter, -1);
  LineForm.Grid.Selection := FitRect;
end;

end.
