unit AutoFitHelper;
{$I utils.inc}{Copyright (C) 2016 C M Western; see commandline.pas for license}
{ Helpers to allow a LineForm analog for command line use }

interface

uses
  MNode, Mixture, pgDOM;

type

  { TLineNoForm }

  TLineNoForm = class(TMNode)
  public
    Lines: DOMString;
    Accept: Double;
  published
    property FirstLine: Integer index 1 read GetSetting write SetSetting;
    property LastLine:  Integer index 2 read GetSetting write SetSetting;
    property AdvancedItem: Boolean Index 3 read GetFlag Write SetFlag;
  end;

function LoadLineNoForm(Mixture: TMixture):TLineNoForm;

procedure DoAutoFit(Mixture: TMixture);

implementation

uses Classes, SysUtils, Math, UCompat, MathSup, TextStrm, Worker, Global, Obs, AutoFit;

function LoadLineNoForm(Mixture: TMixture):TLineNoForm;
var
  Child: TDOMNode;
  nUnread: Integer;
begin
  Result := TLineNoForm.Create(nil);
  Result.Name := 'LineWindow';
  Child := nil;
  if Assigned(Mixture.LoadedFormSettings) then
    Child := Mixture.LoadedFormSettings.FirstChild;
  while Assigned(Child) and (ReadValue(Child, 'Name', '') <> 'LineWindow') do
    Child := Child.NextSibling;
  if Assigned(Child) then begin
    Result.Lines := ReadValue(Child, 'Lines', '');
    Result.Accept := StrToFloatEx(ReadValue(Child,'AcceptEdit','0.1'));
    nUnread := Result.TransferProperties(Child, False, False);
    nUnread := nUnread + Result.nUnreadProperties(Child);
  end;
end;

procedure DoAutoFit(Mixture: TMixture);
var
  AutoFitter: TAutoFitter;
  LNF: TLineNoForm;
  i, nFail: Integer;
  Str: TTextStream;
  Problem: Exception;
begin
  AutoFitter := Mixture.ChildWithClass(TAutoFitter) as TAutoFitter;
  if not Assigned(AutoFitter) then
    raise EReportException.Create('Must be exactly one AutoFitter object');
  AutoFitter.SetLineList(ActiveLinelist);

  LNF := LoadLineNoForm(Mixture);
  try
    Str := TTextStream.Create(TStringStream.Create(LNF.Lines));
    Str.ItemSeparators := [isTab];
    Str.CommentChars := [];
    try
      LoadObservations(Mixture, Str, LineListFileName, obLine, tkAll);
    finally
      Str.UnderlyingStream.Free;
      Str.Free;
    end;
    { The line form reader ignores blank lines completely, so force numbers to be consecutive, starting from 1 }
    for i := 0 to High(Observations) do
      Observations[i].LineNumber := i+1;
    AutoFitter.aW := Abs(LNF.Accept);
    for i := Max(LNF.FirstLine-1,0) to LNF.LastLine-1 do
      Observations[i].Flags := Observations[i].Flags + [ofTrial];
    AutoFitter.SetObservations(Observations);

    nFail := AutoFitter.ComputeSearchRange;
    if nFail > 0 then
      raise EReportException.Create(IntToStr(nFail)+ ' Observations with no possible matches');
    if AutoFitter.nFitted <= 0 then
      raise EReportException.Create('No search lines');

    GlobalMixture.FillParameters;
    if Length(GlobalMixture.FittingParameters) > Length(AutoFitter.Observations) then
      raise EReportException.CreateFmt('More parameters (%d) than observations (%d)', [Length(GlobalMixture.FittingParameters), Length(Observations)]);
    if AutoFitter.logNtrials > 18.9 then
      raise EReportException.CreateFmt('10^%.1g trials is too many', [AutoFitter.logNtrials]);
    //TotalTrials.Text := IntToStr(AutoFitter.nTrials);
    //FailView.Text := '0';

    AutoFitter.nBest := 0;
    //AutoFitter.OnUpdateStatus := {$IFDEF FPC}@{$ENDIF}UpdateStatus;
    //LoadGrid;

    Problem := nil;
    AutoFitter.Search(0, 1, Problem, nil); // {$IFDEF FPC}@{$ENDIF}ShowStatus);
    if Assigned(Problem) then
      raise Problem;
    if AutoFitter.nBest = 0 then begin
      WriteLn(Log, 'All trials failed - showing closest lines to selected transitions only');
      AutoFitter.Closest; { All trials failed - simply show closest }
    end;
    //LoadGrid;
    //UpdateStatus(AutoFitter,AutoFitter.nTrials);
    AutoFitter.WriteInfo(Log);
  finally
    ClearObservations;
    LNF.Free;
  end;
end;

initialization

end.

