unit BaselineFrm;

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, OverlayCalcFrm, Menus, ComCtrls, ScrlCtrl, GraphI, GraphM, Graphs,
  Split, StdCtrls, NewSpin, AlEdit, EditExtr, Buttons, ExtCtrls, ExpPlot,
  MathSup;

type

  { TBaselineForm }

  TBaselineForm = class(TOverlayCalcForm)
    MakeLinelistItem: TMenuItem;
    AppendLinelisttoItem: TMenuItem;
    AbsorbanceItem: TMenuItem;
    OpticalDepthItem: TMenuItem;
    procedure OverwriteItemClick(Sender: TObject);
    procedure ApplytoNewItemClick(Sender: TObject);
    procedure MakeLinelistItemClick(Sender: TObject);
    procedure AppendLinelisttoItemClick(Sender: TObject);
    procedure BaselineButtonClick(Sender: TObject);
    procedure ToolsPopupMenuPopup(Sender: TObject);
  public
    procedure CopyToPlot(NewPlot: TExperimentalPlot; NewBaseline: TDoubleArray; Method: Integer);
    procedure PlotData; override;
    procedure PlotUpper; override;
  end;

var
  BaselineForm: TBaselineForm;

implementation

uses Math, Main, ApplyFrm, UCompat, Overlay, PeakFinders, Worker, Global;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

procedure TBaselineForm.AppendLinelisttoItemClick(Sender: TObject);
var
  NewPlot: TExperimentalPlot;
  Replace, HasIntensity: Boolean;
  OldN, i: Integer;
begin
  Replace := False;
  NewPlot := PromptForOverlay(Self,'Add lines to',Replace,'Erase Current Region',False,TExperimentalPlot,Plot,True) as TExperimentalPlot;
  if NewPlot = nil then
    Exit;
  if not NewPlot.IsLineList then
    raise EReportException.Create(NewPlot.Name + ' is not a linelist');
  HasIntensity := (Length(NewPlot.Frequency) = 0) or (Length(NewPlot.Intensity) > 0);
  StartPeakScan;
  while MoveToNextPeak do
    ;
  if Replace then begin
    with NewPlot do
      for i := 0 to High(Frequency) do
        if (Frequency[i] >= xMin) and (Frequency[i] <= xMax) then
          SetNan(Frequency[i]);
    NewPlot.Sort; { This discards invalid points }
  end;
  OldN := Length(NewPlot.Frequency);
  SetLength(NewPlot.Frequency, OldN+nPeaks);
  for i := 0 to nPeaks - 1 do
    NewPlot.Frequency[i+OldN] := PeakFrequency[i];
  if HasIntensity then begin
    SetLength(NewPlot.Intensity, OldN+nPeaks);
    for i := 0 to nPeaks - 1 do
      NewPlot.Intensity[i+OldN] := PeakIntensity[i];
  end;
  NewPlot.LastPoint := -1;
  NewPlot.Sort;
  MainForm.ConsiderReplot;
end;

const
  MethodNames: array [0..2] of string = ('baseline', 'absorbance', 'optical depth');

procedure TBaselineForm.ApplytoNewItemClick(Sender: TObject);
var
  NewPlot: TExperimentalPlot;
  MTag: Integer;
begin
  if Length(fBaseline) = 0 then
    Exit;
  MTag := (Sender as TComponent).Tag;
  NewPlot := MakeNewPlot(Plot.Name+'-'+MethodNames[MTag], False);
  NewPlot.Comment := Plot.Comment + MethodStatus;
  CopyToPlot(NewPlot, fBaseline, MTag);
end;

procedure TBaselineForm.BaselineButtonClick(Sender: TObject);
begin
  if BaselineWindow.Value > 0 then begin
    BaselineChanged := True;
    fBaseline := MakeFloatingBaseline(Noise, y, BaselineWindow.Value, NoiseMult.Value, NoiseWindow.Value, DenseItem.Checked, InvertedItem.Checked, VerboseItem.Checked);
    StatusBar.Panels[1].Text := DtosM(Noise,10,-1);
    UpdateStatus;
  end else
    inherited;
end;

procedure TBaselineForm.CopyToPlot(NewPlot: TExperimentalPlot; NewBaseline: TDoubleArray; Method: Integer);
var
  i: Integer;
  V: Double;
begin
  NewPlot.DiscardCalibration;
  NewPlot.Frequency := Copy(x);
  SetLength(NewPlot.Intensity, nPoints);
  for i := 0 to nPoints - 1 do begin
    case Method of
      0: V := y[i]-NewBaseline[i];
      1: V := Log10(NewBaseline[i]/y[i]);
      2: V := Ln(NewBaseline[i]/y[i]);
    end;
    NewPlot.Intensity[i] := V;
  end;
  NewPlot.Weights := nil;
  NewPlot.ClearCache(True);
  NewPlot.ClearSelection;
  NewPlot.Modified := True;
  MainForm.ConsiderReplot;
end;

procedure TBaselineForm.MakeLinelistItemClick(Sender: TObject);
var
  NewPlot: TExperimentalPlot;
begin
  StartPeakScan;
  while MoveToNextPeak do
    ;
  SetLength(PeakFrequency, nPeaks);
  SetLength(PeakIntensity, nPeaks);
  NewPlot := MakeNewPlot(Plot.Name + '-linelist', True);
  NewPlot.Comment := Plot.Comment + MethodStatus;
  NewPlot.IntensityUnits := Plot.IntensityUnits;
  NewPlot.Units := Plot.Units;
  NewPlot.Frequency := PeakFrequency;
  NewPlot.Intensity := PeakIntensity;
  MainForm.ConsiderReplot;
  SetLength(PeakFrequency, 0);
  SetLength(PeakIntensity, 0);
end;

procedure TBaselineForm.OverwriteItemClick(Sender: TObject);
var
  NewBaseline: TDoubleArray;
begin
  if Length(fBaseline) = 0 then
    Exit;
  NewBaseline := fBaseline; { as next line clears it }
  Plot.Name := Plot.Name+'-baseline';
  CopyToPlot(Plot, NewBaseline, 0);
end;

procedure TBaselineForm.PlotData;
var
  V: Double;
  i: Integer;
begin
  inherited;
  if (Length(fBaseline) > 0) and (Noise > 0) then begin
    V := Noise*NoiseMult.Value;
    G.Style(4);
    G.ClipPoint(hMin, fBaseLine[ViewLeft]+V);
    for i := ViewLeft+1 to ViewRight do
      G.Join(x[i], fBaseLine[i]+V);
  end;
end;

procedure TBaselineForm.PlotUpper;
var
  i: Integer;
  V: Double;
begin
  if Length(fBaseline) = 0 then
    Exit;
  if (nVscales = 0) or BaselineChanged then begin
    rMin := y[ViewLeft]-fBaseline[ViewLeft];
    rMax := rMin;
    for i := ViewLeft+1 to ViewRight do begin
      V := y[i]-fBaseline[i];
      rMin := Min(rMin, V);
      rMax := Max(rMax, V);
    end;
    BaselineChanged := False;
  end;
  if LiveUpdateItem.Checked then begin
    StartPeakScan;
    Verbose := False;
    LastPos := Max(ViewLeft-100,0);
    while MoveToNextPeak and (LastPos < ViewRight+100) do
      ;
  end;
  Residual.Start;
  PlotVscale(Residual, rMin, rMax);
  Residual.Scales('Frequency', 'I-Baseline');
  Residual.ClipPoint(x[ViewLeft], y[ViewLeft]-fBaseline[ViewLeft]);
  for i := ViewLeft to ViewRight do
    Residual.ClipJoin(x[i], y[i]-fBaseline[i]);
  if LiveUpdateItem.Checked then begin
    Residual.Style(4);
    for i := 0 to nPeaks - 1 do
      if (PeakFrequency[i] > hMin) and (PeakFrequency[i] < hMax) then begin
        Residual.Point(PeakFrequency[i],rMin);
        Residual.Join(PeakFrequency[i],rMax);
      end;
  end;
  Residual.Finish;
end;

procedure TBaselineForm.ToolsPopupMenuPopup(Sender: TObject);
var
  OK: Boolean;
  i: Integer;
begin
  OK := Length(fBaseline) > 0;
  OverwriteItem.Enabled := OK;
  ApplytoNewItem.Enabled := OK;
  MakeLinelistItem.Enabled := OK;
  if OK then begin
    OK := False;
    with GlobalOverlays do
      for i := 0 to Count-1 do
        if (Plots[i] is TExperimentalPlot) and TExperimentalPlot(Plots[i]).IsLineList then begin
          OK := True;
          Break;
        end;
  end;
  AppendLinelisttoItem.Enabled := OK;
end;

end.
