unit CalFrm;
{$I utils.inc}{Copyright (C) 20014 C M Western; see commandline.pas for license}

interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
     Dialogs, PeakFitFrm, Menus, Grids, NGrid, StdCtrls, ComCtrls, AlEdit, EditExtr, NewSpin,
     Buttons, ExtCtrls, Units, MNode, Settings, Overlay, Peaks, Gendlg;

type

  { TCalForm }

  TCalForm = class(TPeakFitForm)
    CopyFrequencyItem: TMenuItem;
    ApplyItem: TMenuItem;
    RangeItem: TMenuItem;
    EtalonItem: TMenuItem;
    AssignButton: TSpeedButton;
    procedure CopyFrequencyItemClick(Sender: TObject);
    procedure ApplyButtonClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure RangeItemClick(Sender: TObject);
    procedure EtalonItemClick(Sender: TObject);
    procedure GridClick(Sender: TObject);
    procedure GridDblClick(Sender: TObject);
    procedure GridDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
    procedure GridSetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
    procedure AssignButtonClick(Sender: TObject);
    procedure FitButtonClick(Sender: TObject);
    procedure PolyOrderChange(Sender: TObject);
    procedure NLfitButtonClick(Sender: TObject);
    procedure SplineFitItemClick(Sender: TObject);
    procedure ResetItemClick(Sender: TObject);
  private
    FLastNPeaks: Integer;
    FAssigningPeak: Integer;
    EtalonSpacing: Double;
    LastSortCol: Integer;
    procedure SetAssigningPeak(Value: Integer);
  protected
    function DefaultFunctionText(Order: Integer):string;
    procedure AddExtraVariables(AFormSettings: TFormSettings); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    procedure WriteExtraVariables(AFormSettings: TFormSettings); override;
  public
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure ShowPeak(AIndex: Integer); override;
    procedure PeakChanged(PeakNumber:Integer); override;
    procedure PeakMeasured(Source: TMNode; Index: Integer); override;
    function Peaks: TPeaks; override;
    function HasDataSource: Boolean; override;
    property AssigningPeak: Integer read FAssigningPeak write SetAssigningPeak;
  end;

var
  CalForm: TCalForm;

function CalibratingPlot:TPlot;

implementation

uses Math, MathSup, Peak, Global, ExpPlot, PlotDataSource, Main, ApplyFrm, DataSource;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function CalibratingPlot:TPlot;
begin
  Result := nil;
  with CalForm do
    if Visible and (CurrentMNode is TPlot) then
      Result := CurrentMNode as TPlot;
end;

{ TCalForm }

procedure TCalForm.AddExtraVariables(AFormSettings: TFormSettings);
var
  Index: Integer;
begin
  if EtalonSpacing = 0 then
    EtalonSpacing := 1;
  inherited;
  Index := AFormSettings.IndexOfParameter('EtalonSpacing');
  if Index < 0 then
    Index := AFormSettings.AddParameter('EtalonSpacing', 1);
  AFormSettings.SetParameter(Index, EtalonSpacing);
end;

procedure TCalForm.ApplyButtonClick(Sender: TObject);
var
  Plot: TMNode;
  NotUsed: Boolean;
  i: Integer;
begin
  Plot := PromptForOverlay(Self,'Apply Calibration To',NotUsed,'',True,TExperimentalPlot);
  if Assigned(Plot) then begin
    if Plot is TPlot then
      (CurrentMNode as TExperimentalPlot).ApplyCalibration(Plot as TExperimentalPlot,Sender = nil)
    else begin { Apply to all }
      for i := 0 to GlobalOverlays.ComponentCount - 1 do
        if (GlobalOverlays.Components[i] <> CurrentMNode)
           and (GlobalOverlays.Components[i] is TExperimentalPlot) then
            (CurrentMNode as TExperimentalPlot).ApplyCalibration(GlobalOverlays.Components[i] as TExperimentalPlot,Sender = nil);
      { Doing the current one last }
      (CurrentMNode as TExperimentalPlot).ApplyCalibration(CurrentMNode as TExperimentalPlot,Sender = nil);
    end;
    MainForm.ConsiderReplot(False,False);
  end;
end;

procedure TCalForm.FormResize(Sender: TObject);
begin
  Constraints.MinWidth := OtherButton.Left + OtherButton.Width + 2;
  Panel.Height := AvgErrEdit.Top + AvgErrEdit.Height + 2;
end;

procedure TCalForm.AssignButtonClick(Sender: TObject);
begin
  if AssigningPeak >= 0 then
    AssigningPeak := -1
  else
    AssigningPeak := Grid.Row-1;
end;

function TCalForm.BaseNodeClass: TClass;
begin
  Result := TOverlays;
end;

procedure TCalForm.CopyFrequencyItemClick(Sender: TObject);
begin
  ApplyButtonClick(nil);
end;

function TCalForm.DefaultFunctionText(Order: Integer): string;
var
  i: Integer;
begin
  case Order of
    0: Result := 'x+c'; { Simple offset }
    1: Result := 'm*x+c';
  else
    Result := 'poly(x,c,m';
    for i := 2 to Order do
      Result := Result + ',a'+IntToStr(i);
    Result := Result + ')';
  end;
end;

procedure TCalForm.EtalonItemClick(Sender: TObject);
var
  s: string;
begin
  s := FloatToStr(EtalonSpacing);
  if Dialogs.InputQuery('Find equally spaced peaks', 'Enter Spacing', s) then begin
    EtalonSpacing := StrToFloatEx(s);
    try
      Grid.BeginUpdate;
      DisableReplot := True;
      (CurrentMNode as TExperimentalPlot).Etalon(EtalonSpacing);
    finally
      DisableReplot := False;
      Grid.EndUpdate;
    end;
    AssigningPeak := -1;
    MainForm.ConsiderReplot(False,False);
  end;
end;

procedure TCalForm.FitButtonClick(Sender: TObject);
var
  OldLinearFit, OldUseSVD: Boolean;
begin
  if CurrentMNode is TExperimentalPlot then begin
    with (CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource, DataSourceFitter do
      try
        { Higher order polynomials give "sensible" results with these settings }
        OldLinearFit := LinearFit;
        LinearFit := True;
        OldUseSVD := UseSVD;
        UseSvd := False;
        NLFitDlg.FitOnceClick(nil);
      finally
        LinearFit := OldLinearFit;
        UseSVD := OldUseSVD;
      end;
    MainForm.ResidualsItemClick(nil);
  end;
end;

procedure TCalForm.GridClick(Sender: TObject);
begin
  AssigningPeak := -1;
end;

procedure TCalForm.GridDblClick(Sender: TObject);
begin
  if LastCell.Y = 0 then begin
    { Double click on column heading to sort on that column }
    if CurrentMNode is TPlot then
      (CurrentMNode as TPlot).Peaks.SortPeaks(LastCell.X-1, LastCell.X <> LastSortCol);
    if LastSortCol = LastCell.X then
      LastSortCol := 0
    else
      LastSortCol := LastCell.X; 
  end else if LastCell.X = 2 then
    AssigningPeak := LastCell.Y-1
end;

procedure TCalForm.GridDrawCell(Sender: TObject; ACol, ARow: Integer;  Rect: TRect; State: TGridDrawState);
begin
  if (AssigningPeak >=0) and (ACol = 2) and (AssigningPeak = ARow-1) then
    ColouredTextInGrid(Grid.Canvas, Rect, 'Measure', clRed);
end;

procedure TCalForm.GridSetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
var
  P: TPeak;
  function Update(var OldValue: Double; NewValue: Double): Boolean;
  begin
    Result := (IsNanOrInf(OldValue) xor IsNanOrInf(NewValue))
              or (OldValue <> NewValue);
    if Result then
      OldValue := NewValue;
  end;
begin
  if not Valid then
    SetNan(v);
  if CurrentMNode is TPlot then begin
    if (Row >= Peaks.nPeaks) and not IsNanOrInf(v) then
      Peaks.nPeaks := Row;
    P := Peaks.Peaks[Row-1];
    case Col of
      1: Update(P.Position,v);
      2: Update(P.Actual,v);
      4: Update(P.Weight,v);
    end;
    try
      FChanging := True;
      Peaks.Peaks[Row-1] := P;
    finally
      FChanging := False;
    end;
  end;
  AssigningPeak := -1;
end;

function TCalForm.HasDataSource: Boolean;
begin
  Result := False;
  if not Visible or (csDestroying in Application.ComponentState) then
    Exit;
  if CurrentMNode is TExperimentalPlot then
    if TExperimentalPlot(CurrentMNode).HasDataSource then begin
      DataSource := TExperimentalPlot(CurrentMNode).DataSource as TInteractiveDataSource;
      Result := True;
    end;
end;

procedure TCalForm.NLfitButtonClick(Sender: TObject);
begin
  ReShow(((CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource).NLFitDlg);
end;

procedure TCalForm.PeakChanged(PeakNumber: Integer);
var
  AutoAssign: Boolean;
begin
  CurrentMNode.Modified := True;
  AutoAssign := FLastNPeaks = Peaks.nPeaks-1;
  inherited;
  if AutoAssign then
    AssigningPeak := Peaks.nPeaks-1;
end;

procedure TCalForm.PeakMeasured(Source: TMNode; Index: Integer);
var
  Peak: TPeak;
begin
  if (AssigningPeak >= 0) and (CurrentMNode <> Source) then
    with CurrentMNode as TPlot do begin
      Peak := Peaks.Peaks[AssigningPeak];
      Peak.Actual := Convert(LastPeakPosition, LastPeakUnits, Units);
      Peak.Residual := Peak.Actual-ConvertPeakPosition(Peak.Position,Units);
      Peaks.Peaks[AssigningPeak] := Peak;
    end;
end;

function TCalForm.Peaks: TPeaks;
begin
  if CurrentMNode is TPlot then
    Result := (CurrentMNode as TPlot).Peaks
  else
    Result := DummyPeaks;
end;

procedure TCalForm.PolyOrderChange(Sender: TObject);
begin
  if (PolyOrder.Value >= 0) and (CurrentMNode is TExperimentalPlot) then
    with ((CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource).NLFitDlg do begin
      FunctionText := DefaultFunctionText(PolyOrder.Value);
      PlotFunctionClick(nil);
    end;
end;

procedure TCalForm.RangeItemClick(Sender: TObject);
var
  Peak: TPeak;
begin
  if CurrentMNode is TExperimentalPlot then
    with CurrentMNode as TExperimentalPlot do begin
      Peak := EmptyPeak;
      Peak.Weight := 1000;
      if Length(OriginalFrequency) > 0 then begin
        Peak.Position := OriginalFrequency[0];
        Peaks.AddPeak(Peak);
        Peak.Position := OriginalFrequency[High(OriginalFrequency)];
      end else begin
        Peak.Position := Frequency[0];
        Peaks.AddPeak(Peak);
        Peak.Position := Frequency[High(Frequency)];
      end;
      Peaks.AddPeak(Peak);
    end;
  AssigningPeak := -1;
end;

procedure TCalForm.ResetItemClick(Sender: TObject);
begin
  if CurrentMNode is TExperimentalPlot then
    with CurrentMNode as TExperimentalPlot do begin
      if Length(OriginalFrequency) > 0 then
        Frequency := OriginalFrequency;
        OriginalFrequency := nil;
        FrequencyScale := 1;
        FrequencyOffset := 0;
      MainForm.ConsiderReplot(False,False);
    end;
end;

procedure TCalForm.SetAssigningPeak(Value: Integer);
begin
  if not (CurrentMNode is TPlot) then
    Value := -1
  else
    Value := Min(Value,(CurrentMNode as TPlot).nPeaks-1);
  AssignButton.Down := Value >= 0;
  if Value = AssigningPeak then
    Exit;
  FAssigningPeak := Value;
  Grid.Invalidate;
end;

procedure TCalForm.ShowItem(MNode: TMNode);
var
  i, n, poly: Integer;
  ft: string;
begin
  {$IFDEF LCL_}
  AssignButton.Width := OtherButton.Width; { Autosize fails for speed buttons }
  {$ENDIF}
  { Update Node display }
  Grid.EditorMode := False;
  SetItemIndex(ActiveObject,MNode);
  CurrentMNode := MNode;
  AssigningPeak := -1;
  FLastNPeaks := 0;
  UpdateStatus;
  Caption := 'Calibrating';
  poly := 1;
  if MNode is TPlot then begin
    Caption := CurrentMNode.Name + ' - ' + Caption;
    n := Peaks.nPeaks;
    if MNode is TExperimentalPlot then
      with TExperimentalPlot(MNode) do
        if HasDataSource then begin
          poly := -1;
          ft := LowerCase(Trim((DataSource as TInteractiveDataSource).NLFitDlg.FunctionText));
          for i := PolyOrder.MinValue+1 to PolyOrder.MaxValue do
            if ft = DefaultFunctionText(i) then
              poly := i;
        end;
  end else
    n := 0;
  if n = 0 then begin
    Grid.RowCount := 2;
    Grid.Clear;
    Grid.Enabled := False;
  end else begin
    Grid.Enabled := True;
    Grid.RowCount := n+1;
    for i := 0 to n-1 do
      ShowPeak(i);
  end;
  with PolyOrder do
    if poly <> Value then begin
      OnChange := nil; { Disable function update }
      Value := poly;
      OnChange := {$IFDEF FPC}@{$ENDIF}PolyOrderChange;
    end;
  if not DisableReplot then
    MainForm.ConsiderReplot(False,False);
end;

procedure TCalForm.ShowPeak(AIndex: Integer);
begin
  AssigningPeak := -1;
  FLastNPeaks := -2;
  if CurrentMNode is TPlot then
    with Peaks[Aindex] do begin
      FLastNPeaks := Peaks.nPeaks;
      Grid[AIndex+1,1] := Position;
      Grid[AIndex+1,2] := Actual;
      Grid[AIndex+1,3] := Residual;
      Grid[AIndex+1,4] := Weight;
    end;
end;

procedure TCalForm.SplineFitItemClick(Sender: TObject);
begin
  ((CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource).NLFitDlg.SplineItemClick(nil);
end;

function TCalForm.StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer;
begin
  if MNode is TExperimentalPlot then begin
    ActiveObject.AddItem(Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end else
    Result := 0;
end;

procedure TCalForm.WriteExtraVariables(AFormSettings: TFormSettings);
var
  Index: Integer;
begin
  inherited;
  Index := AFormSettings.IndexOfParameter('EtalonSpacing');
  if Index > 0 then
    EtalonSpacing := AFormSettings.Parameters[Index];
end;

end.
