unit CheckOverlay;

interface

uses Classes, ExpPlot, Overlay;

function CheckOverlays(Overlays: TOverlays; S: TStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; SortBoth, Interpolate: Boolean; Ext: string):Double;

implementation

uses Math, Units, MathSup, SysUtils, Worker, CheckImport;

function CheckOverlays(Overlays: TOverlays; S: TStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; SortBoth, Interpolate: Boolean; Ext: string):Double;
var
  i, p, junk: Integer;
  Ref, Test: TExperimentalPlot;
  RefO: TOverlays;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then
      WriteLn(Log, Test.Name, ' Point Number ', i, ':', Name);
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string);
  var
    Err: Double;
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else begin
      if tol < 0 then begin
        if Abs(n1-n2) < 1e-15 then
          Err := Abs(n1-n2) { Permit small absolute errors }
        else
          Err := 2*Abs(n1-n2)/(Abs(n1)+Abs(n2));
      end else
        Err := Abs(n1-n2);
      if Err > Abs(tol) then
        Fail(Format('%s out of tolerance: %g %g %g', [Name, n1, n2, Err]));
      if (Err > Result) and (Failures >= 0) then
        Result := Err;
    end;
  end;
begin
  nTried := 0;
  Result := 0;
  RefO := TOverlays.Create(nil);
  try
    if Ext = '.ovr' then begin
      RefO.LoadFromStream(S);
      Ref := RefO.Plots[0] as TExperimentalPlot;
    end else begin
      Ref := TExperimentalPlot.Create(RefO);
      Ref.WeightedLoadFromStream(S, cm1, Ext, False);
      RefO.Units := Overlays.Units;
    end;
    for p := 0 to Min(Overlays.nPlots,RefO.nPlots) -1 do begin
      Test := Overlays.Items[p] as TExperimentalPlot;
      Ref := RefO.Items[p] as TExperimentalPlot;
      if Ref.Units <> Test.Units then begin
        Ref.ScaleParameters(Ref.Units, Test.Units);
        Ref.Units := Test.Units;
      end;
      if SortBoth or Interpolate then begin
        // Make sure both going in the same direction
        Ref.Sort(Test.Frequency[0] < Test.Frequency[High(Test.Frequency)]);
      end;
      { Copy some values across, as these will never be identical }
      Ref.Name := Test.Name;
      Test.Colour := Ref.Colour;
      if Ref.LastPoint < 0 then
        if Interpolate then
          Ref.LastPoint := Test.LastPoint { Supress otherwise trivial warning }
        else
          Ref.LastPoint := High(Ref.Frequency);
      if Ref.Comment = Ext then
        Ref.Comment := Test.Comment;
    end;
    RefO.Name := Overlays.Name;
    CompareNode(Overlays, RefO, Failures, Result, Tolerance);
    for p := 0 to Min(Overlays.nPlots,RefO.nPlots)-1 do begin
      Test := Overlays.Items[p] as TExperimentalPlot;
      Ref := RefO.Items[p] as TExperimentalPlot;
      if not Interpolate then begin
        i := -1;
        CheckNo(High(Ref.Frequency), High(Test.Frequency), Tolerance, 'x nPoints-1');
        CheckNo(High(Ref.Intensity), High(Test.Intensity), Tolerance, 'y nPoints-1');
        CheckNo(High(Ref.Weights), High(Test.Weights), Tolerance, 'weights');
        for i := 0 to Min(High(Ref.Frequency), High(Test.Frequency)) do begin
          CheckNo(Ref.Frequency[i], Test.Frequency[i], Tolerance, 'Frequency');
          CheckNo(Ref.Intensity[i], Test.Intensity[i], Tolerance, 'Intensity');
          if (High(Ref.Weights) > 0) and (High(Test.Weights) > 0) then
            CheckNo(Ref.Weights[i], Test.Weights[i], Tolerance, 'Weights');
          Inc(nTried);
        end;
      end else begin
        for i := 0 to High(Test.Frequency) do begin
          CheckNo(Ref.IntensityAt(Convert(Test.Frequency[i],Test.Units,cm1),junk), Test.Intensity[i], Tolerance, 'Intensity');
          Inc(nTried);
        end;
      end;
    end;
  finally
    RefO.Free;
  end;
end;

end.
