unit DeGlitchFrm;

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, OverlayCalcFrm, Menus, ComCtrls, ScrlCtrl, GraphI, GraphM, Graphs,
  Split, StdCtrls, NewSpin, AlEdit, EditExtr, Buttons, ExtCtrls, MathSup,
  ExpPlot;

type
  TDeGlitchForm = class(TOverlayCalcForm)
    procedure BaselineButtonClick(Sender: TObject);
    procedure OverwriteItemClick(Sender: TObject);
    procedure ApplytoNewItemClick(Sender: TObject);
  private
    D: TDoubleArray;
    PeakPosition: TIntegerArray;
  public
    procedure PlotData; override;
    procedure PlotUpper; override;
    procedure FindGlitches;
    procedure RemoveGlitches(Interpolate: Boolean);
    procedure CopyToPlot(NewPlot: TExperimentalPlot);
  end;

var
  DeGlitchForm: TDeGlitchForm;

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses Math, Main;

{ TDeGlitchForm }

procedure TDeGlitchForm.BaselineButtonClick(Sender: TObject);
begin
  inherited;
  FindGlitches;
  RemoveGlitches(True);
  UpdateStatus;
end;

procedure TDeGlitchForm.CopyToPlot(NewPlot: TExperimentalPlot);
begin
  NewPlot.DiscardCalibration;
  NewPlot.Frequency := Copy(x);
  NewPlot.Intensity := D;
  NewPlot.Weights := nil;
  NewPlot.ClearCache(True);
  NewPlot.ClearSelection;
  NewPlot.RemoveNullPoints(True);
  NewPlot.Sort;
  NewPlot.Modified := True;
  MainForm.ConsiderReplot;
end;

procedure TDeGlitchForm.FindGlitches;
begin
  StartPeakScan;
  PeakThreshold := yMin + BaselineWindow.Value/100*(yMax-yMin);
  while MoveToNextPeak do begin
    SetLength(PeakPosition, Length(PeakFrequency));
    PeakPosition[nPeaks-1] := LastPeak;
  end;
end;

procedure TDeGlitchForm.ApplytoNewItemClick(Sender: TObject);
var
  NewPlot: TExperimentalPlot;
begin
  if Length(fBaseline) = 0 then
    Exit;
  RemoveGlitches(False);
  NewPlot := MakeNewPlot(Plot.Name+'-deglitched', False);
  NewPlot.Comment := Plot.Comment + MethodStatus;
  CopyToPlot(NewPlot);
end;

procedure TDeGlitchForm.OverwriteItemClick(Sender: TObject);
begin
  if Length(fBaseline) = 0 then
    Exit;
  RemoveGlitches(False);
  Plot.Name := Plot.Name+'-deglitched';
  CopyToPlot(Plot);
end;

procedure TDeGlitchForm.RemoveGlitches(Interpolate: Boolean);
var
  i, j, S, E: Integer;
  T: Double;
begin
  T := yMin + BaselineWindow.Value/200*(yMax-yMin);
  D := Copy(y);
  for i := 0 to nPeaks - 1 do begin
    S := PeakPosition[i]-1;
    E := S+1;
    while (S >= 0) and (y[S]-fBaseline[S] >= T) do
      Dec(S);
    S := Max(S-1,0);
    while (E < Length(D)) and (y[E]-fBaseline[S] >= T) do
      Inc(E);
    E := Min(E+1, High(D));
    for j := S+1 to E - 1 do
      if Interpolate then
        D[j] := y[S] + (y[E]-y[S])/(E-S)*(j-S)
      else
        SetNan(D[j]);
  end;
end;

procedure TDeGlitchForm.PlotData;
var
  V: Double;
  i: Integer;
begin
  inherited;
  V := yMin + BaselineWindow.Value/100*(yMax-yMin);
  G.Style(4);
  G.ClipPoint(hMin, V);
  G.Join(hMax, V);
end;

procedure TDeGlitchForm.PlotUpper;
var
  i: Integer;
begin
  if Length(fBaseline) = 0 then
    Exit;
  if LiveUpdateItem.Checked then begin
    FindGlitches;
    RemoveGlitches(True);
  end;
  if nVscales = 0 then
    Limits(D, rMin, rMax, ViewLeft, ViewRight);
  Residual.Start;
  PlotVscale(Residual, rMin, rMax);
  Residual.Scales('Frequency', 'DeGlitched(I)');
  Residual.ClipPoint(x[ViewLeft], D[ViewLeft]);
  for i := ViewLeft to ViewRight do
    Residual.ClipJoin(x[i], D[i]);
  Residual.Style(4);
  for i := 0 to nPeaks - 1 do
    if (PeakFrequency[i] > hMin) and (PeakFrequency[i] < hMax) then begin
      Residual.Point(PeakFrequency[i],rMax);
      Residual.Join(PeakFrequency[i],rMax+(rMax-rMin)*0.1)
    end;
  Residual.Finish;
end;

end.
