unit Global;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Forms, {$ENDIF}
     Mixture, ExpPlot, Overlay, MNode, TextStrm, Units, UCompat;

function DecodePgopherVersion(VersionName: string):Integer;
var
  VersionInfo: string;
  nExprFailures: Integer;

{$IFNDEF CONSOLE}
procedure ReShow(Form: TCustomForm);
function GlobalMixture: TMixture;
function GlobalOverlays: TOverlays;
function GlobalLineList: TLineList;

procedure LoadGeneralFile(AFileName: string; Merge, PreferOverlay: Boolean; DefaultUnits: TResonanceUnits);
procedure LoadAnyObservations;
{$ELSE}
var
  GlobalMixture: TMixture;
  GlobalOverlays: TOverlays;
{$ENDIF}
function ActiveLinelist(ReturnNil: Boolean = False): TExperimentalPlot;
procedure AddOverlay(AFileName: string; Merge:Boolean; DefaultUnits: TResonanceUnits);

function BeginCalculation(Fitting: Boolean): Boolean;
function CalculationInProgress: Boolean;
function FittingInProgress: Boolean;
procedure EndCalculation;
procedure BeginGlobalUpdate;
procedure EndGlobalUpdate(Postpone: Boolean = false);

procedure PeakMeasured(Position: Double; Units: TResonanceUnits; Lab: string; Source: TMNode; PlotTag: Integer; EstErr: Double = -1);
var
  LastPeakPosition: Double;
  LastPeakUnits: TResonanceUnits;
  LastPeakErr: Double;

type
  TMixtureReader = function (Stream: TTextStream):TMixture;
  TFileType = (fiParameters, fiObservations, fiOverlay);

function FileType(AFileName: string; PreferOverlay: Boolean; out Reader: TMixtureReader): TFileType;
function ImportMixture(DefaultProc: TMixtureReader; AFileName: string): TMixture;
function IsMixtureFile(AFileName: string):Boolean;

procedure ShowInfo(const Msg: string);
procedure ReShowLogForm;

type

  TPlottedPoint = record
    X, Y:  Double;  { Plot Position (must use double to allow for very small plot ranges) }
    Index: Integer; { Observation number }
    Extra: {$IFDEF FPC}PtrInt{$ELSE}Integer{$ENDIF}; { Information }
  end;

  TPlottedPoints = array of TPlottedPoint;

  TOnGetPointInfo = function (Index: Integer; Extra: PtrInt):string of object;

implementation

uses {$IFNDEF CONSOLE}
        CalFrm, Huge, Clipbrd, LineFrm, Main, PlotFrm, Dialogs, LogUnit, ChildForm, WCompat,
        {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$ENDIF}
     SysUtils, Classes, Hitran, ImportCALPGM, ImportSpecView, ImportSTROTAB,
     ImportJB95, ImportACDD, D3hImport, PgopherImport, ImportVIBCA, ExoMol, Worker;

function DecodePgopherVersion(VersionName: string):Integer;
var
  p: Integer;
  InNumber: Boolean;
  LastNumber: Integer;
  nNumbers: Integer;
begin
  Result := 0;
  InNumber := False;
  LastNumber := 0;
  VersionName := VersionName + ' ';
  nNumbers := 0;
  for p := 1 to Length(VersionName) do
    if VersionName[p] in ['0'..'9'] then begin
      InNumber := True;
        LastNumber := LastNumber*10 + Ord(VersionName[p])-Ord('0')
    end else if InNumber then begin
      InNumber := False;
      Result := Result*1000 + LastNumber;
      LastNumber := 0;
      Inc(nNumbers);
      if nNumbers >=3 then
        Exit;
    end;
  for p := nNumbers+1 to 3 do
    Result := Result*1000;
end;

procedure AddOverlay(AFileName: string; Merge:Boolean; DefaultUnits: TResonanceUnits);
begin
  {$IFNDEF CONSOLE}
    if Assigned(MainForm) then
      MainForm.AddOverlay(AFileName, Merge, DefaultUnits);
  {$ELSE}
    AddToOverlay(GlobalOverlays, AFileName, Merge, DefaultUnits, False, Arbitrary);
  {$ENDIF}
end;

{$IFNDEF CONSOLE}
procedure LoadGeneralFile(AFileName: string; Merge, PreferOverlay: Boolean; DefaultUnits: TResonanceUnits);
begin
  if Assigned(MainForm) then
    MainForm.LoadGeneralFile(AFileName, Merge, PreferOverlay, DefaultUnits);
end;

procedure LoadAnyObservations;
begin
  if Assigned(MainForm) then
    MainForm.LoadAnyObservations;
end;

function GlobalMixture: TMixture;
begin
  if Assigned(MainForm) then
    Result := MainForm.Mixture
  else
    Result := nil;
end;

function GlobalOverlays: TOverlays;
begin
  if Assigned(MainForm) then
    Result := MainForm.Overlays
  else
    Result := nil;
end;

function GlobalLineList: TLineList;
var
  Mixture: TMixture;
begin
  Mixture := GlobalMixture;
  if Assigned(Mixture) then
    Result := Mixture.LineList
  else
    Result := nil;
end;

procedure ReShow(Form: TCustomForm);
begin
  if not Form.Visible then
    Form.Show;
  if Form.WindowState in [wsMinimized, wsMaximized] then
    Form.WindowState := wsNormal;
  Form.BringToFront;
end;

{$ENDIF}

function IsMixtureFile(AFileName: string):Boolean;
var
  Buffer: string[100];
begin
  Result := False;
  try
    with TFileStream.Create(AFileName,fmOpenRead or DefaultShareMode) do
      try
        SetLength(Buffer, Read(Buffer[1], 100));
        Result := Pos('<Mixture ',Buffer) > 1;
      finally
        Free;
      end;
  except
  end;
end;

function FileType(AFileName: string; PreferOverlay: Boolean; out Reader: TMixtureReader): TFileType;
var
  Ext: string;
begin
  Reader := nil;
  Ext := LowerCase(ExtractFileExt(AFileName));
  if (((Ext = '.pgo') or (Ext = '.xml')) and IsMixtureFile(AFileName)) then
    Result := fiParameters
  else if ((Ext = '.par') and not IsHITRANfile(AFileName))
          or (Ext = '.var') then begin
    Result := fiParameters;
    Reader := {$IFDEF FPC}@{$ENDIF}ImportCALPGMFile;
  end else if Ext = '.svw' then begin
    Result := fiParameters;
    Reader := {$IFDEF FPC}@{$ENDIF}ImportSpecViewFile;
  end else if Ext = '.jbw' then begin
    Result := fiParameters;
    Reader := {$IFDEF FPC}@{$ENDIF}ImportJB95File;
  end else if Ext = '.sta' then begin
    Result := fiParameters;
    Reader := {$IFDEF FPC}@{$ENDIF}ImportSTROTABFile;
  end else if Ext = '.vib' then begin
    Result := fiParameters;
    Reader := {$IFDEF FPC}@{$ENDIF}ImportVIBCAFile;
  end else if Ext = '.states' then begin
    Result := fiParameters;
    Reader := {$IFDEF FPC}@{$ENDIF}ImportExoMolFile;
  end else if (Ext = '.lin') or (Ext = '.obs') or (((Ext = '.cat') or (Ext = '.mrg')) and not PreferOverlay) then
      Result := fiObservations
  else
    Result := fiOverlay;
end;

function ImportMixture(DefaultProc: TMixtureReader; AFileName: string): TMixture;
var
  Stream: TFreeFormatStream;
  FStream: TFileStream;
  Line: string;
begin
  Result := nil;
  FStream := TFileStream.Create(AFileName, fmOpenRead or DefaultShareMode);
  SetCurrentDir(ExtractFileDir(AFileName));
  Stream := nil;
  try
    Stream := TFreeFormatStream.Create(FStream, False);
    Stream.StreamName := AFileName;
    Line := LowerCase(Stream.RestOfLine);
    Stream.Position := 0;
    if Pos('this is a d3h ', Line) > 0 then
      DefaultProc := {$IFDEF FPC}@{$ENDIF}ImportD3hFile
    else if Pos(' pgopher ', Line) > 0 then
      DefaultProc := {$IFDEF FPC}@{$ENDIF}ImportPgopherFile
    else if Pos('idn', Line) > 0 then
      DefaultProc := {$IFDEF FPC}@{$ENDIF}ImportACDDFile;
    Result := DefaultProc(Stream);
  finally
    Stream.Free;
    FStream.Free;
  end;
end;

{$IFNDEF CONSOLE}
var
  CF_Measurement: DWORD;
{$ENDIF}

procedure PeakMeasured(Position: Double; Units: TResonanceUnits; Lab: string; Source: TMNode; PlotTag: Integer; EstErr: Double);
{$IFNDEF CONSOLE}
var
  Stream: THugeMemoryStream;
  ASource: AnsiString;
{$ENDIF}
begin
  LastPeakPosition := Position;
  LastPeakUnits := Units;
  LastPeakErr := EstErr;
  {$IFNDEF CONSOLE}
  if CF_Measurement = 0 then
    CF_Measurement := RegisterClipboardFormat('Measurement');
  Stream := THugeMemoryStream.Create;
  try
    (* Mimic the following C structure:
      typedef struct {
        double value;
        char signature[4];    /* Normally MEAS */
        char filename[1];     /* Null terminated file name; can be longer */
      } ClipboardMeasurement ;
    *)
    Stream.Write(Position, Sizeof(Double));
    Stream.Write('MEAS', 4);
    if Lab <> '' then begin
      ASource := Lab;
      Stream.Write(ASource[1], Length(ASource)+1);
    end else
      Stream.Write(#0#0, 1);
    { Copy to clipboard }
    Clipboard.Open;
    try
      Clipboard.AsText := FloatToStr(Position);
      Stream.CopyToClipboard(CF_Measurement);
    finally
      Clipboard.Close;
    end;
  finally
    Stream.Free;
  end;
  LineForm.AssignLines(Position, Units, Lab, Source, False, EstErr);
  {$ENDIF}
  PerformMNodeNotification(Source, mnPeakMeasured, PlotTag);
end;

procedure ShowInfo(const Msg: string);
begin
  {$IFDEF CONSOLE}
  WriteLn(Log,Msg);
  {$ELSE}
  ShowMessage(Msg);
  {$ENDIF}
end;

procedure ReShowLogForm;
begin
  {$IFNDEF CONSOLE}
  ReShow(LogForm);
  {$ENDIF}
end;

function ActiveLinelist(ReturnNil: Boolean): TExperimentalPlot;
var
  i: Integer;
begin
  Result := nil;
  if GlobalOverlays <> nil then
    with GlobalOverlays do
      for i := 0 to nPlots-1 do
        if (Plots[i] is TExperimentalPlot) and (TExperimentalPlot(Plots[i]).IsLineList) then begin
          if (Result = nil) or not (Plots[i] as TExperimentalPlot).Hide then
            Result := Plots[i] as TExperimentalPlot;
          if not Result.Hide then
            Break;
        end;
  if not ReturnNil and (Result = nil) then
    raise EReportException.Create('No overlay linelist');
end;

var
  FCalculationInProgress: Boolean;
  FFitting: Boolean;

function CalculationInProgress: Boolean;
begin
  Result := FCalculationInProgress;
end;

function BeginCalculation(Fitting: Boolean): Boolean;
begin
  Result := CalculationInProgress;
  if not Result then begin
    FCalculationInProgress := True;
    FFitting := Fitting;
  end;
end;

procedure EndCalculation;
begin
  FCalculationInProgress := False;
end;

function FittingInProgress: Boolean;
begin
  Result := FCalculationInProgress and FFitting;
end;

procedure BeginGlobalUpdate;
begin
  {$IFNDEF CONSOLE}
    ChildForm.BeginGlobalUpdate;
  {$ENDIF}
end;

procedure EndGlobalUpdate(Postpone: Boolean = false);
begin
  {$IFNDEF CONSOLE}
    ChildForm.EndGlobalUpdate(Postpone);
  {$ENDIF}
end;

end.
