unit ImportFrm;
{$I utils.inc}{Copyright (C) 2016 C M Western; see commandline.pas for license}
interface

uses
  {$IFDEF VCL_} Messages, {$ENDIF} SysUtils, Variants, Classes, Graphics, Controls,
  {$IFNDEF LCL_} FileNameEdit, {$ELSE} EditBtn, {$ENDIF} Forms, WhatsThis,
  Dialogs, StdCtrls, ScrlCtrl, LogControl, AlEdit, EditExtr, MNode, MakeHelp;

type

  TImportForm = class(THelpForm)
    FileNameEdit: TFileNameEdit;
    LogControl: TLog;
    EditButton: TButton;
    TestButton: TButton;
    ApplyButton: TButton;
    CancelButton: TButton;
    Overwrite: TCheckBox;
    CreateCheck: TCheckBox;
    procedure EditButtonClick(Sender: TObject);
    procedure TestButtonClick(Sender: TObject);
    procedure ApplyButtonClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    Template: TMNode;
    {$IFDEF VCL_}
    procedure WMDropFiles(var Message: TMessage); message WM_DROPFILES;
    {$ENDIF}
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
  public
    OK: Boolean;
    InitialText: string;
    procedure RunImport(Apply: Boolean);
  end;

var
  ImportForm: TImportForm;

function ExecuteImportMNode(MNode: TMNode):Boolean;

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses WCompat, SlaveEdit, Importer, TextAssign, SizeHelpers, Global, Worker, ItemStrm;

procedure TImportForm.EditButtonClick(Sender: TObject);
begin
  LaunchEditor(FileNameEdit.FileName);
end;

procedure TImportForm.FormCreate(Sender: TObject);
begin
  DragAcceptFiles(Self, True);
  {$IFNDEF VCL_}
  OnDropFiles := @FormDropFiles;
  {$ENDIF}
  InitialText := LogControl.Lines.Text;
end;

procedure TImportForm.FormDropFiles(Sender: TObject; const FileNames: array of String);
begin
  if Length(FileNames) > 0 then
    FileNameEdit.FileName := FileNames[0];
end;

procedure TImportForm.FormShow(Sender: TObject);
begin
  {$IFDEF LCL}
    ConstantSizeButtons(Self);
  {$ENDIF}
  LogControl.Lines.Text := InitialText;
  LogControl.MakeVisible(Point(0,0));
end;

procedure TImportForm.RunImport(Apply: Boolean);
var
  S: TItemStream;
  FileName: string;
  F: System.Text;
begin
  LogControl.Clear;
  AssignStrings(F, LogControl.Lines);
  Rewrite(F);
  BeginGlobalUpdate;
  S := nil;
  try
    FileName := FileNameEdit.FileName;
    S := SetupImportStream(FileName);
    if Apply then
      OK := ImportMNodes(Log, Template, S, FileName, Overwrite.Checked, CreateCheck.Checked, True)
    else
      OK := ImportMNodes(F, Template, S, FileName, False, False, True)
  finally
    S.Free;
    CloseFile(F);
    EndGlobalUpdate;
  end;
end;

procedure TImportForm.ApplyButtonClick(Sender: TObject);
begin
  RunImport(True);
end;

procedure TImportForm.TestButtonClick(Sender: TObject);
begin
  RunImport(False);
end;

{$IFDEF VCL_}
procedure TImportForm.WMDropFiles(var Message: TMessage);
begin
  HandleWMDropFiles(Message, FormDropFiles);
end;
{$ENDIF}

function ExecuteImportMNode(MNode: TMNode):Boolean;
begin
  Result := False;
  with ImportForm do begin
    Template := MNode;
    if ShowModal = mrOK then
      Result := OK;
  end;
end;

end.
