unit LevelFrm;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ScrlCtrl, Graphs, StdCtrls, ExtCtrls,
  WhatsThis, ChildForm, MNode, Mixture, EditExtr, NewSpin,
  AlEdit, Buttons, FormUtils, Menus, PlotOC, Settings, QuantumSpin, GraphI,
  GraphM, Global, UCompat;

type

  { TLevelForm }

  TLevelForm = class(TChildForm)
    Panel: TPanel;
    Label1: TLabel;
    ActiveObject: TComboBox;
    ExcludePerturbations: TCheckBox;
    LG: TGraphD;
    StatusBar: TStatusBar;
    Ymax: TFixedSpin;
    Label2: TLabel;
    Label3: TLabel;
    Ymin: TFixedSpin;
    XMin: TQuantumSpin;
    Label4: TLabel;
    Label5: TLabel;
    XMax: TQuantumSpin;
    Label6: TLabel;
    Reduction: TAlignedEdit;
    Label7: TLabel;
    AllButton: TSpeedButton;
    ZoomButton: TSpeedButton;
    ZoomOutButton: TSpeedButton;
    CalcPoints: TSpeedButton;
    CalcLine: TSpeedButton;
    CalcBoth: TSpeedButton;
    JoinHighest: TCheckBox;
    ObsUpper: TSpeedButton;
    ObsLower: TSpeedButton;
    ReloadButton: TSpeedButton;
    JoinObs: TSpeedButton;
    AnotherButton: TButton;
    ListButton: TButton;
    TrackState: TCheckBox;
    TrackQuanta: TCheckBox;
    LabelsStyle: TComboBox;
    Symmetry: TSpinBox;
    Label10: TLabel;
    Label11: TLabel;
    ErrorMultiplier: TFixedSpin;
    XQuantumLabel: TLabel;
    XQuantumNumber: TSpinBox;
    Label13: TLabel;
    IgnoreSpins: TSpinBox;
    GPopupMenu: TPopupMenu;
    EditPointItem: TMenuItem;
    ASelect: TQuantumSpin;
    ALabel: TLabel;
    BSelect: TQuantumSpin;
    BLabel: TLabel;
    ShowLineItem: TMenuItem;
    ExternalField: TCheckBox;
    FieldPanel: TPanel;
    FieldLabel: TLabel;
    MinField: TFixedSpin;
    Label15: TLabel;
    MaxField: TFixedSpin;
    FieldSamples: TSpinBox;
    Label16: TLabel;
    Label17: TLabel;
    ShiftMult: TFixedSpin;
    PlotTypeGroup: TGroupBox;
    EScaleGroup: TGroupBox;
    SelectGroup: TGroupBox;
    QuantumNumberGroup: TGroupBox;
    ObservedGroup: TGroupBox;
    ExcludeGroup: TGroupBox;
    LabelsGroup: TGroupBox;
    Label8: TLabel;
    ESweep: TRadioButton;
    BSweep: TRadioButton;
    AltPlot: TCheckBox;
    RefreshPlot: TButton;
    SummaryButton: TButton;
    LimitJ: TCheckBox;
    ShowStateItem: TMenuItem;
    procedure YminChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure GPopupMenuPopup(Sender: TObject);
    procedure EditPointItemClick(Sender: TObject);
    procedure IgnoreSpinsSetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: string);
    procedure IgnoreSpinsGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: string; var StringValid: Boolean);
    procedure XQuantumNumberSetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: string);
    procedure XQuantumNumberGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: string; var StringValid: Boolean);
    procedure ErrorMultiplierIncrement(Sender: TObject; var NewValue: Double; LastValue:Double; n: Integer);
    procedure ListButtonClick(Sender: TObject);
    procedure ExcludePerturbationsClick(Sender: TObject);
    procedure LGMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure AllButtonClick(Sender: TObject);
    procedure ZoomButtonClick(Sender: TObject);
    procedure ZoomOutButtonClick(Sender: TObject);
    procedure ReloadButtonClick(Sender: TObject);
    procedure AnotherButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure ASelectSetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: string);
    procedure ASelectGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: string; var StringValid: Boolean);
    procedure ShowLineItemClick(Sender: TObject);
    procedure MaxFieldIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
    procedure MinFieldIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
    procedure YminIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
    procedure RefreshPlotClick(Sender: TObject);
    procedure SummaryButtonClick(Sender: TObject);
    procedure ShowStateItemClick(Sender: TObject);
  private
    EigenvaluesValid: Boolean;
    FormUpdating: Boolean;
    { For plotted points save memory by encoding Manifold, Symmetry and J in Index }
    { Bit pack: Manifold index (8 bits); Symmetry (8 bits); F (16 bits) }
    ResidualPlotter: TObsNumberPlotter;
    YmaxIndex, YminIndex, LastPointNumber: Integer;
    LastExtra: PtrInt;
    function OnGetPointInfo(Index: Integer; Extra: PtrInt):string;
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer); override;
    function ReadValues(ShowError: Boolean): Boolean;
//    function ReadVal(ValText, ValName: string; var Value: Double; ShowError: Boolean):Boolean;
    procedure CheckValue(Component: TComponent; PName, Value, PHint: string; SettingType: TSettingType; Data: TObject);
  protected
    PlottedPoints: TPlottedPoints;
    nPlottedPoints: Integer;
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    procedure HandleComponent(Component: TComponent; var PName, AValue, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    function FormSettingsName: string; override;
  public
    procedure Loaded; override;
    destructor Destroy; override;
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); overload; override;
    procedure ShowItem(MNode: TMNode; Log, ShowErrors, ForcePlot, FieldSummary: Boolean); overload;
    procedure StuffSelector(MNode: TMNode); override;
    property ETop: Double index 1 read GetParameter write SetParameter;
    property EBottom: Double index 2 read GetParameter write SetParameter;
  end;

var
  LevelForm: TLevelForm;

implementation

uses Obs, Gmarks, GStyle, MathSup, LogUnit, SlaveEdit,
     QuantumNumber, QuantumMath, MNodeColour, Math, Worker, Progress,
     SettingsForm, LevelPlotter, Main, StateInfo;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TLevelForm }

function TLevelForm.BaseNodeClass: TClass;
begin
  Result := TMixture;
end;

procedure TLevelForm.YminChange(Sender: TObject);
begin
  if CalculationInProgress then
    Exit;
  if Lockout > 0 then
    ReStuff
  else
    ShowItem(CurrentMNode);
end;

procedure TLevelForm.YminIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
begin
  NewValue := LastValue + n*(Ymax.Value-Ymin.Value)/10;
end;

destructor TLevelForm.Destroy;
begin
  FreeAndNil(ResidualPlotter);
  inherited;
end;

procedure TLevelForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  YminChange(nil);
end;

procedure TLevelForm.SetParameter(AIndex: Integer; const Value: Double);
begin
  inherited;
  if AIndex = YminIndex then
    YMin.Value := Value
  else if AIndex = YmaxIndex then
    YMax.Value := Value
end;

procedure TLevelForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  YminChange(nil);
end;

procedure TLevelForm.ShowItem(MNode: TMNode; Log, ShowErrors, ForcePlot, FieldSummary: Boolean);
begin
  if (csLoading in ComponentState) or not Visible or FormUpdating then
    Exit;
  try
    FormUpdating := True;
    FieldPanel.Visible := ExternalField.Checked;
    Constraints.MinWidth := Max(ExcludeGroup.Left + ExcludeGroup.Width, LabelsGroup.Left + LabelsGroup.Width) + 2;
    { Force display of settings as text }
    try
      if Assigned(BaseNode) then
        BaseNode.CheckProperties;
      UpdateSymmetrySpin(Symmetry, MNode, False);
      UpdateQnoBox(MNode, ALabel, ASelect, qtA);
      UpdateQnoBox(MNode, BLabel, BSelect, qtB);
      ErrorMultiplier.ValueAB;
      IgnoreSpins.Value := IgnoreSpins.ValueAB;
      XMin.ValueAB;
      XMax.ValueAB;
      if FieldPanel.Visible then begin
        MinField.ValueAB;
        MaxField.ValueAB;
        XQuantumNumber.Visible := False;
        XQuantumLabel.Caption := 'M range:';
        if ESweep.Checked then
          FieldLabel.Caption := 'field V/m:'
        else
          FieldLabel.Caption := 'field T:'
      end else begin
        XQuantumLabel.Caption := 'x Axis Type';
        XQuantumNumber.Value := XQuantumNumber.ValueAB;
        XQuantumNumber.Visible := True;
      end;
      Reduction.Enabled := not FieldPanel.Visible;
      JoinHighest.Enabled := not FieldPanel.Visible;
    except
      on E: Exception do begin
        StatusBar.SimpleText := E.Message;
        if ShowErrors then
          MessageDlg(E.Message, mtError, [mbOk], 0);
        Exit;
      end;
    end;
    if not ReadValues(ShowErrors) then
      Exit;
    { Update Node display }
    if ExternalField.Checked then
      if ESweep.Checked then
        Caption := 'Electric Field Plot'
      else
        Caption := 'Magnetic Field Plot'
    else
      Caption := 'Energy Level Plot';
    nPlottedPoints := 0;
    if not Assigned(MNode) then Exit;
    Caption := MNode.UniqueName + ' - ' + Caption;
    SetItemIndex(ActiveObject,MNode);
    ResidualPlotter.SetValid(True);
    BeginGlobalUpdate;
    try
      if Assigned(BaseNode) and ((BaseNode as TMixture).AutoReplot or ForcePlot) then
        (Self as TLevelPlotter).PlotItem(LG,MNode,Log,FieldSummary);
      CurrentMNode := MNode;
      EigenvaluesValid := True;
    finally
      EndGlobalUpdate;
    end;
  finally
    FormUpdating := False;
  end;
end;

procedure TLevelForm.ShowItem(MNode: TMNode);
begin
  ShowItem(MNode, False, False, False, False);
end;

function TLevelForm.StuffObject(const Name1, Name2: string; MNode: TMNode;
  var ParentNode: TTreeNode): Integer;
begin
  Result := 0;
  if    ( (MNode is TMixture) and (TMixture(MNode).Count >1) )
     or ( (MNode is TSpecies) and (TSpecies(MNode).Count >1) )
     or ( (MNode is TMolecule) and (TMolecule(MNode).nManifolds >1) )
     or ( (MNode is TManifold) and (TManifold(MNode).nStates <>1) )
     or ( MNode is TState ) then begin
    ActiveObject.AddItem(Name1 + Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end;
end;

procedure TLevelForm.StuffSelector(MNode: TMNode);
begin
  StuffSelectorHelper(MNode,ActiveObject);
  FComboBoxUpdated := False;
end;

procedure TLevelForm.ExcludePerturbationsClick(Sender: TObject);
begin
  if Lockout > 0 then
    ReStuff
  else
    ShowItem(CurrentMNode,False,True,False,False);
end;

procedure TLevelForm.LGMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  ResidualPlotter.nMouseLines := 0;
  LastPointNumber := High(LastPointNumber);
  StatusBar.SimpleText := ShowGraphInfo(Sender as TGraphD, PlottedPoints, {$IFDEF FPC}@{$ENDIF}OnGetPointInfo);
end;

procedure TLevelForm.GPopupMenuPopup(Sender: TObject);
begin
  EditPointItem.Enabled := ResidualPlotter.SaveMouseLines;
  ShowLineItem.Enabled := EditPointItem.Enabled;
  ShowStateItem.Enabled := (LastPointNumber <> High(LastPointNumber)) and EigenvaluesValid;
end;

procedure TLevelForm.HandleComponent(Component: TComponent; var PName, AValue,
  PHint: string; var Save: Boolean; var SettingType: TSettingType;
  Write: Boolean);
begin
  inherited;
  if (Component = Reduction) or (Component = ErrorMultiplier)
     or (Component = Ymax) or (Component = Ymin)
     or (Component = ShiftMult) or (Component = MinField) or (Component = MaxField) then
    SettingType := stDouble;
end;

procedure TLevelForm.AllButtonClick(Sender: TObject);
var
  d: Double;
begin
  BeginGlobalUpdate;
  try
    with Self as TLevelPlotter do begin
      d := (EMax-Emin)/20;
      if d = 0 then d := 1;
      ETop := EMax+d;
      EBottom := EMin-d;
    end;
  finally
    EndGlobalUpdate;
  end;
  ExcludePerturbationsClick(nil);
end;

procedure TLevelForm.ZoomButtonClick(Sender: TObject);
var
  d, b, t: Double;
begin
  BeginGlobalUpdate;
  try
    if LG.AreaSelected then begin
      d := Abs(LG.Ymax-LG.Ymin)/LG.Height;
      t := RoundToIncrement(LG.SelectedMathArea.Top,d);
      b := RoundToIncrement(LG.SelectedMathArea.Bottom,d);
      InOrder(b, t);
      EBottom := b;
      ETop := t;
    end else if ReadValues(True) then begin
      d := (ETop - EBottom)/4;
      Etop := Etop - d;
      EBottom := EBottom + d;
    end;
  finally
    EndGlobalUpdate;
  end;
  ExcludePerturbationsClick(nil);
end;

procedure TLevelForm.CheckValue(Component: TComponent; PName, Value, PHint: string; SettingType: TSettingType; Data: TObject);
var
  T: Double;
begin
  if SettingType = stDouble then
    if not TryStrToFloatEx(Trim(Value), T) then begin
      Value := 'Bad value for ' + PName + ':' + Value;
      StatusBar.SimpleText := Value;
      if Assigned(Data) then
        MessageDlg(Value, mtError, [mbOK], 0);
      Abort;
    end;
end;

function TLevelForm.ReadValues(ShowError: Boolean): Boolean;
begin
  Result := False;
  try
    ForAllControls({$IFDEF FPC}@{$ENDIF}CheckValue,{$IFDEF FPC}TObject(PtrUInt{$ELSE}(Pointer{$ENDIF}(ShowError)));
  except
    on EAbort do
      Exit;
  end;
  WriteSettings(FormSettings,False,True);
  if (FormSettings.IndexOfParameter('Ymax') <> 1)
     or (FormSettings.IndexOfParameter('Ymin') <> 2) then
    raise NotYet;
  Result := True;
end;

procedure TLevelForm.RefreshPlotClick(Sender: TObject);
begin
  ShowItem(CurrentMNode, False, False, True, False);
end;

procedure TLevelForm.ZoomOutButtonClick(Sender: TObject);
var
  d: Double;
begin
  BeginGlobalUpdate;
  try
    if ReadValues(True) then begin
      d := (ETop - EBottom)/2;
      Etop := Etop + d;
      EBottom := EBottom - d;
    end;
  finally
    EndUpdate(False);
  end;
  ExcludePerturbationsClick(nil);
end;

procedure TLevelForm.ReloadButtonClick(Sender: TObject);
begin
  if not SaveAllWindows then Exit;
  LoadAnyObservations;
  ExcludePerturbationsClick(nil);
end;

procedure TLevelForm.AnotherButtonClick(Sender: TObject);
var
  NewLevelForm: TLevelForm;
begin
  NewLevelForm := TLevelPlotter.Create(Application);
  NewLevelForm.Show;
  NewLevelForm.BaseNode := BaseNode;
end;

procedure TLevelForm.ASelectGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: string; var StringValid: Boolean);
begin
  if SameText(TheText, 'all') or (TheText = '') then begin
    NewValue := (Sender as TSpinBox).MinValue;
    StringValid := True;
  end;
end;

procedure TLevelForm.ASelectSetIntegerValue(Sender: TObject; NewValue: Integer;
  var NewText: string);
begin
  if NewValue = (Sender as TSpinBox).MinValue then
    NewText := 'all';
end;

procedure TLevelForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  if Self <> LevelForm then
    CloseAction := caFree;
end;

procedure TLevelForm.FormCreate(Sender: TObject);
begin
  LG.StuffPopupMenu(GPopupMenu);
end;

function TLevelForm.FormSettingsName: string;
begin
  Result := 'LevelWindow'
end;

procedure TLevelForm.ListButtonClick(Sender: TObject);
begin
  ShowItem(CurrentMNode, True, True, True, False);
  ReShow(LogForm)
end;

procedure TLevelForm.SummaryButtonClick(Sender: TObject);
begin
  ShowItem(CurrentMNode, False, True, True, True);
  ReShow(LogForm)
end;

procedure TLevelForm.Loaded;
begin
  inherited;
  XQuantumNumber.MaxValue := High(QnoExprNames)-1;
  ResidualPlotter := TObsNumberPlotter.Create('');
  UpdateSymmetrySpin(Symmetry, nil, False);
  YmaxIndex := FormSettings.IndexOfParameter('Ymax');
  YminIndex := FormSettings.IndexOfParameter('Ymin');
end;

procedure TLevelForm.MaxFieldIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
begin
  NewValue := SpinScaleBy(n,2,LastValue);
end;

procedure TLevelForm.MinFieldIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
begin
  NewValue := LastValue + n*(MaxField.Value-LastValue)/10;
end;

procedure TLevelForm.MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
begin
  { TODO: Respond to mnParameterChange, mnSettingChange only if Autoreplot set, and not during fitting }
  inherited;
  if (Change = mnCacheCleared) and (Sender.RootNode is TMixture) then
    EigenvaluesValid := False;
end;

function TLevelForm.OnGetPointInfo(Index: Integer; Extra: PtrInt): string;
var
  Ev: TEigenvalues;
begin
  LastPointNumber := Index;
  LastExtra := Extra;
  if Index >= 0 then
    Result := ResidualPlotter.OnGetPointInfo(Index, Extra)
  else if EigenvaluesValid then begin
    Ev := TEigenvalues(Extra);
    if -Index <= Ev.Size then
      Result := ' ' + Ev.DominantLabel(-Index,GlobalMixture.QuantumNumberFormat);
    Result := Result + ':' + IntToStr(-Index);
  end else
    Result := ' ' + IntToStr(-Index);
end;

procedure TLevelForm.EditPointItemClick(Sender: TObject);
begin
  ResidualPlotter.EditLastPoint;
end;

procedure TLevelForm.ShowLineItemClick(Sender: TObject);
begin
  ResidualPlotter.ShowLastPoint;
end;

procedure TLevelForm.ShowStateItemClick(Sender: TObject);
  procedure ShowState(Manifold: TManifold; J: TQuantumNumber; Sym, Index: Integer);
  begin
    if not Assigned(Manifold) then
      Exit;
    ReShow(StatesForm);
    StatesForm.J.Value := J;
    StatesForm.Symmetry.Text := Manifold.Molecule.JAdjustedSymName(J, Sym);
    StatesForm.BaseNode := Manifold.RootNode;
    StatesForm.ShowItem(Manifold);
    StatesForm.EigGrid.SetFocus;
    if (Index > 0) and (Index <= StatesForm.EigGrid.RowCount) then
      StatesForm.EigGrid.Row := Index;
  end;
var
  Ev: TEigenvalues;
  OT: TObsType;
begin
  if not EigenValuesValid then
    Exit;
  if (LastPointNumber < 0) and EigenvaluesValid then begin
    Ev := TEigenvalues(LastExtra);
    ShowState(Ev.Manifold, Ev.J, Ev.Sym, -LastPointNumber);
  end else if ResidualPlotter.ObservationsValid and (LastPointNumber >=0) and (LastPointNumber < Length(Observations)) then
    with Observations[LastPointNumber] do begin
      if ObsType = obLine then
        if LastExtra = 0 then
          OT := obEUpper
        else
          OT := obELower
      else
        OT := ObsType;
      case OT of
        obELower, obCDLower: ShowState(ManifoldL, JL, SymL, AIndexL);
        obEUpper, obCDUpper: ShowState(ManifoldU, JU, SymU, AIndexU);
      end;
    end;
end;

procedure TLevelForm.ErrorMultiplierIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
begin
  if LastValue > 0 then
    NewValue := Max(1, SpinScaleBy(n,10,LastValue))
  else
    NewValue := SpinScaleBy(n,10,LastValue);
end;

procedure TLevelForm.XQuantumNumberGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: string; var StringValid: Boolean);
var
  i: Integer;
begin
  for i := 0 to High(QnoExprNames)-1 do
    if QnoExprNames[i] = TheText then begin
      NewValue := i;
      StringValid := True;
      Exit;
    end;
end;

procedure TLevelForm.XQuantumNumberSetIntegerValue(Sender: TObject;
  NewValue: Integer; var NewText: string);
begin
  NewText := QnoExprNames[NewValue];
end;

procedure TLevelForm.IgnoreSpinsGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: string; var StringValid: Boolean);
begin
  if TheText = 'all' then begin
    StringValid := True;
    NewValue := 0;
  end;
end;

procedure TLevelForm.IgnoreSpinsSetIntegerValue(Sender: TObject;
  NewValue: Integer; var NewText: string);
begin
  if NewValue = 0 then
    NewText := 'all';
end;

end.
