unit LineFrm;
{$I utils.inc}{Copyright (C) 2005,2016 C M Western; see commandline.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     SysUtils, Variants, Classes, Graphics, Controls, Forms,
     Dialogs, Grids, NGrid, ExtCtrls, Units, MNode, SizeHelpers,
     Mixture, StdCtrls, Buttons, Menus, QuantumNumber, Obs, Settings, SettingsForm,
     AlEdit, EditExtr, NewSpin, QuantumSpin, AutoFit;

type

  TStateDef = record
    Manifold: string;
    J:        TQuantumNumber;
    Sym:      Integer;
  end;

  TLineW = record
    Molecule:  string;
    Upper:     TStateDef;
    Lower:     TStateDef;
    Weight:    Double;
    Line:      TLine;  { Note that eigenvalues not always set }
    IObs:      Double; { Observed intensity of line; 0 to flag not present }
  end;

  TLineSortFunction = function (const A, B: TLineW): Integer;

  TLineListType = (llStandard, llBranch, llSimple, llBranchExpr);

  { TLineForm }

  TLineForm = class(TSettingsForm)
    Panel: TPanel;
    Grid: TNumberGrid;
    ClearButton: TButton;
    AssignButton: TButton;
    AllButton: TButton;
    NoneButton: TButton;
    CopyButton: TSpeedButton;
    CutButton: TSpeedButton;
    DeleteButton: TSpeedButton;
    DeleteUnassigned: TButton;
    MinusButton: TButton;
    Difference: TLabel;
    MoreButton: TButton;
    SortPopupMenu: TPopupMenu;
    FrequencySortItem: TMenuItem;
    UpperStateSortItem: TMenuItem;
    LowerStateSortItem: TMenuItem;
    StrengthSortItem: TMenuItem;
    WidthSortItem: TMenuItem;
    WeightSortItem: TMenuItem;
    Auto: TCheckBox;
    BranchItem: TMenuItem;
    TestFitButton: TButton;
    FitButton: TButton;
    UndoFitButton: TButton;
    SortonItem: TMenuItem;
    OptionsItem: TMenuItem;
    CopyBranchTableItem: TMenuItem;
    UpperStateCDFitItem: TMenuItem;
    LowerStateCDfitItem: TMenuItem;
    SavetofileItem: TMenuItem;
    SaveDialog: TSaveDialog;
    GridPopupMenu: TPopupMenu;
    CopySimpleFormatItem: TMenuItem;
    InStandardFormatItem: TMenuItem;
    InBranchFormatItem: TMenuItem;
    InSimpleFormatItem: TMenuItem;
    AltFitButton: TSpeedButton;
    FitPopupMenu: TPopupMenu;
    BranchFitItem: TMenuItem;
    SimpleFitItem: TMenuItem;
    NameSortItem: TMenuItem;
    InBranchExprFormatItem: TMenuItem;
    CopyBranchExprFormatItem: TMenuItem;
    PanelEx: TPanel;
    AcceptLabel: TLabel;
    IncLabel: TLabel;
    JDownButton: TButton;
    JUpButton: TButton;
    AcceptEdit: TEdit;
    Increment: TQuantumSpin;
    ShowButton: TButton;
    BumpKa: TCheckBox;
    BumpKc: TCheckBox;
    UpdateButton: TButton;
    AdvancedItem: TMenuItem;
    SuppressButton: TButton;
    MoveUpButton: TSpeedButton;
    MoveDownButton: TSpeedButton;
    DiscardItem: TMenuItem;
    OutsidePlotRange: TMenuItem;
    Outsidelinelistrange1: TMenuItem;
    LowIntensity1: TMenuItem;
    CDplot: TButton;
    MoveTopButton: TSpeedButton;
    NearestButton: TButton;
    ResetButton: TButton;
    CheckDerivativesItem: TMenuItem;
    procedure FormShow(Sender: TObject);
    procedure UndoFitButtonClick(Sender: TObject);
    procedure BranchItemClick(Sender: TObject);
    procedure GridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
    procedure ClearButtonClick(Sender: TObject);
    procedure AllButtonClick(Sender: TObject);
    procedure NoneButtonClick(Sender: TObject);
    procedure CutButtonClick(Sender: TObject);
    procedure CopyButtonClick(Sender: TObject);
    procedure DeleteButtonClick(Sender: TObject);
    procedure CloseButtonClick(Sender: TObject);
    procedure AssignButtonClick(Sender: TObject);
    procedure GridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: String);
    procedure GridChangeCellQuery(Sender: TObject; Col, Row: Integer; var CanChange: Boolean);
    procedure GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
    procedure DeleteUnassignedClick(Sender: TObject);
    procedure MinusButtonClick(Sender: TObject);
    procedure StrengthSortItemClick(Sender: TObject);
    procedure UpperStateSortItemClick(Sender: TObject);
    procedure FrequencySortItemClick(Sender: TObject);
    procedure LowerStateSortItemClick(Sender: TObject);
    procedure WidthSortItemClick(Sender: TObject);
    procedure WeightSortItemClick(Sender: TObject);
    procedure NameSortItemClick(Sender: TObject);
    procedure MoreButtonClick(Sender: TObject);
    procedure FitButtonClick(Sender: TObject);
    procedure OptionsItemClick(Sender: TObject);
    procedure SavetofileItemClick(Sender: TObject);
    procedure GridDrawCellContents(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
    procedure GridSetCell(Sender: TObject; ACol, ARow: Integer; var v: Double; var Valid: Boolean);
    procedure AltFitButtonClick(Sender: TObject);
    procedure UpdateButtonClick(Sender: TObject);
    procedure ShowButtonClick(Sender: TObject);
    procedure JUpButtonClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure AdvancedItemClick(Sender: TObject);
    procedure SuppressButtonClick(Sender: TObject);
    procedure MoveUpButtonClick(Sender: TObject);
    procedure MoveDownButtonClick(Sender: TObject);
    procedure OutsidePlotRangeClick(Sender: TObject);
    procedure Outsidelinelistrange1Click(Sender: TObject);
    procedure LowIntensity1Click(Sender: TObject);
    procedure CDplotClick(Sender: TObject);
    procedure MoveTopButtonClick(Sender: TObject);
    procedure GridDblClick(Sender: TObject);
    procedure NearestButtonClick(Sender: TObject);
    procedure ResetButtonClick(Sender: TObject);
  private
    NotEmpty: Boolean;
    nPendingLines: Integer;
    FLongestNameLength: Integer;
    FnLinesToShow: Integer;
    FMinIntensity, FMinAssignedIntensity: Double;
    FShowZeroIntensityLines: Boolean;
    FCurrentLineListType: TLineListType;
    DummyReader: TObservationsReaderBase;
    PanelControls, PanelExControls: TControlArray;
    MyAutoFitter: TAutoFitter;
    HasMeasuredIntensity: Boolean;
    function nVisibleLines: Integer; { Number of entries in Lines[] that have actually been displayed - normally equal to nLines }
    procedure LoadOptionsDialog;
    procedure UnLoadOptionsDialog;
    procedure SetLineByIndex(var Line: TLineW; const Value: string; Index: Integer);
  protected
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer); override;
    procedure UpdateActions; override;
    procedure AddExtraVariables(AFormSettings: TFormSettings); override;
    procedure WriteExtraVariables(AFormSettings: TFormSettings); override;
    procedure HandleComponent(Component: TComponent; var PName, Value, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
    procedure WriteSettings(AFormSettings: TFormSettings; DoPositions: Boolean; DoSettings: Boolean); override;
    function FormSettingsName: string; override;
    function StringRep(ARow, ACol: Integer; var buffer: string): Boolean;
  public
    { First three should be protected }
    Lines: array of TLineW;
    nLines: Integer; { Number of valid entries in Lines - not necessarily all displayed }
    procedure UpdateGridLines(First, Last: Integer);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddLine(const ALine: TLine; AUnits: TResonanceUnits);
    procedure ArrangeForm(Screen: TRect); override;
    function ShowLines(Simulation: TSimulation; FilterLines, DiscardDuplicates: Boolean): Integer;
    procedure BringUp;
    procedure ClearTLines;
    function HasObservations:Boolean;
    procedure HighlightRow(ARow: Integer);
    procedure LoadObservations(const Source: string; ObsType: TObsType; TakeWeight: TTakeWeight; Selected: Boolean);
    procedure SortLines(SortFunction: TLineSortFunction; First, Last: Integer);
    procedure CopyToStream(Stream: TStream; Row1, Row2: Integer; All: Boolean; LineListType: TLineListType; PreserveLineNumbers: Boolean; PlatformDefaultCodePage: Boolean);
    procedure UpdateAssignment(ARow: Integer; AFrequency, AWeight: Double; SourceName: string; Intensity: Double = 0);
    procedure AssignLines(APosition: Double; Units: TResonanceUnits; SourceName: string; Source: TMNode; Force: Boolean; AErr: Double = -1);
    procedure ConsiderRename(MNode: TMNode; OldName: string);
    function MakeObs(var Obs:TObservation; Row: Integer; AlwaysCalculate: Boolean): Boolean;
    function FindBranch(ABranch: string; nSearch: Integer):Integer;
    procedure FinishDiscard;
    function IsUnassigned(Index: Integer):Boolean;
    procedure UnAssignLine(Index: Integer);
    procedure LoadAutoFitAssignment(AutoFitter: TAutoFitter; iFit: Integer);
  end;

var
  LineForm: TLineForm;

const
  LabelCol = 15;

procedure SplitCachedName(var Line: TLine; out Branch, TxName: string);

implementation

uses {$IFDEF LCL_} Types, {$ENDIF}
     Math, MathSup, TextAssign, Huge, ClipBrd, PeakFinders, ShellSort,
     SlaveEdit, UCompat, LogUnit, Global, Main, LineOpt, ResidFrm, TextStrm,
     WCompat, FixFnt, Worker, Fitter, Overlay, QuantumMath, HundA, ExpPlot,
     LoomisWood, Peak;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

var
  CF_Measurement, CF_Difference: DWORD;

{ TLineForm }

procedure TLineForm.AddExtraVariables(AFormSettings: TFormSettings);
var
  Index: Integer;
  Stream: TStringStream;
  R: TGridRect;
begin
  inherited;
  HandleComponents(LineFrmOptDlg, {$IFDEF FPC}@{$ENDIF}AddValue, AFormSettings);
  Index := AFormSettings.AddSetting('QuantumNumberFormat', stString, 'J', True);
  AFormSettings.SetSettingFromStr(Index, QuantumNumberFormatNames[Ord(GlobalMixture.QuantumNumberFormat)]);
  R := Grid.Selection;
  with R do
    if (Left = Right) and (Top = Bottom) then begin
      Top := -1; { Force no write }
      Bottom := -1;
    end;
  Index := AFormSettings.AddSetting('FirstLine', stInteger, '-1', True);
  AFormSettings.SetSetting(Index, R.Top);
  Index := AFormSettings.AddSetting('LastLine', stInteger, '-1', True);
  AFormSettings.SetSetting(Index, R.Bottom);
  Index := AFormSettings.AddSetting('Lines', stString, '', True);
  if nVisibleLines > 0 then begin
    Stream := TStringStream.Create(''{$IFDEF UNICODE},TEncoding.Unicode{$ENDIF});
    try
      Grid.SaveToStream(Stream);
      AFormSettings.SetSettingFromStr(Index,sLineBreak+Stream.DataString);
    finally
      Stream.Free;
    end;
  end else
    AFormSettings.SetSettingFromStr(Index,'');
end;

procedure TLineForm.AddLine(const ALine: TLine; AUnits: TResonanceUnits);
var
  Rect: TGridRect;
  function StateDef(Eigenvalues: TEigenvalues; AIndex: Integer):TStateDef;
  begin
    with Result do
      if Assigned(Eigenvalues) then begin
        Manifold := Eigenvalues.Manifold.Name;
        J := Eigenvalues.J;
        Sym := Eigenvalues.Sym;
      end else begin
        Manifold := '';
        J := Low(TquantumNumber);
        Sym := Low(Integer);
      end;
  end;
begin
  { Reset scrolling so that first column is in view }
  Rect := Grid.Selection;
  Rect.Left := 1;
  Rect.Right := 1;
  Grid.Selection := Rect;
  if nLines >= Length(Lines) then
    SetLength(Lines, nLines+50);
  with Lines[nLines] do begin
    Line := ALine;
    ConvertLine(Line, AUnits, GlobalMixture.Simulations[0].PlotUnits);
    if Assigned(ALine.EigenvaluesL) then
      Molecule := ALine.EigenvaluesL.Manifold.Molecule.Name;
    Upper := StateDef(ALine.EigenvaluesU, ALine.IndexU);
    Lower := StateDef(ALine.EigenvaluesL, ALine.IndexL);
    SetNan(Weight);
    if GlobalMixture.ShowEstUnc and Assigned(ALine.EigenvaluesL) then begin
      Weight := EstErr(ALine.EigenvaluesU, ALine.EigenvaluesL, ALine.IndexU, ALine.IndexL);
      if Weight <= 0 then
        SetNan(Weight) { No errors set }
      else
        Weight := -ConvertWidth(Weight, ALine.Frequency, AUnits, GlobalMixture.Simulations[0].PlotUnits)
    end;
  end;
  Inc(nLines);
  FreeAndNil(MyAutoFitter);
end;

procedure TLineForm.AdvancedItemClick(Sender: TObject);
begin
  PanelEx.Visible := AdvancedItem.Checked;
  FormResize(nil);
end;

procedure TLineForm.SortLines(SortFunction: TLineSortFunction; First, Last: Integer);
var
  N, I, J, Offset: Integer;
  IsInOrder: Boolean;
begin
  Grid.EditorMode := False;
  { ShellSort }
  N := Last-First+1;
  Offset := N;
  while Offset > 1 do begin
    Offset := Offset shr 1;
    repeat
      IsInOrder := True;
      for J := First to First + N - Offset - 1 do begin
        I := J + Offset;
        if SortFunction(Lines[I], Lines[J]) > 0 { I > J? } then begin
          IsInOrder := False;
          SwapValues(Lines[I], Lines[J], SizeOf(TLineW));
        end;
      end;
    until IsInOrder;
  end;
  UpdateGridLines(First, Last);
end;

procedure SplitCachedName(var Line: TLine; out Branch, TxName: string);
var
  p: Integer;
begin
  TxName := MakeTxName(Line);
  p := Pos(':',TxName);
  if Assigned(Line.EigenvaluesL) and not Assigned(Line.EigenvaluesU) then
    p := 0; { One state only - no branch to extract }
  if p > 0 then begin
    Branch := Copy(TxName,1,p-1);
    if (Length(Branch) > 0) and (Branch[Length(Branch)] = ' ') then
      Delete(Branch,Length(Branch),1);
    Delete(TxName,1,p);
    if (TxName <> ' ') and (TxName[1] = ' ') then
      Delete(TxName,1,1);
  end else
    Branch := '';
end;

function ByIntensity(const A, B: TLineW): Integer;
begin
 Result := Sign(A.Line.Strength - B.Line.Strength);
end;

function ByFrequency(const A, B: TLineW): Integer;
begin
 Result := -Sign(A.Line.Frequency - B.Line.Frequency);
end;

function ByWidth(const A, B: TLineW): Integer;
begin
 Result := Sign(A.Line.Width - B.Line.Width);
end;

function ByWeight(const A, B: TLineW): Integer;
begin
 if IsNanOrInf(A.Weight) then
   if IsNanOrInf(B.Weight) then
     Result := 0
   else
     Result := -1
 else if IsNanOrInf(B.Weight) then
   Result := 1
 else
   Result := Sign(A.Weight - B.Weight);
end;

function ByEigenvalues(const A, B: TStateDef): Integer;
begin
  Result := CompareText(A.Manifold, B.Manifold);
  if Result = 0 then
    Result := A.J - B.J;
  if Result = 0 then
    Result := A.Sym - B.Sym;
end;

function ByUpperState(const A, B: TLineW): Integer;
begin
 Result := CompareText(A.Molecule, B.Molecule);
 if Result = 0 then
   Result := ByEigenvalues(A.Upper, B.Upper);
 if Result = 0 then
   Result := A.Line.IndexU - B.Line.IndexU;
 Result := -Result;
end;

function ByLowerState(const A, B: TLineW): Integer;
begin
 Result := CompareText(A.Molecule, B.Molecule);
 if Result = 0 then
   Result := ByEigenvalues(A.Lower, B.Lower);
 if Result = 0 then
   Result := A.Line.IndexL - B.Line.IndexL;
 Result := -Result;
end;

function ByName(const A, B: TLineW): Integer;
var
  NameA, NameB, Junk: string;
  Line: TLine;
begin
 Line := A.Line; SplitCachedName(Line, Junk, NameA);
 Line := B.Line; SplitCachedName(Line, Junk, NameB);
 Result := -CompareText(NameA, NameB);
end;

function ByBranch(const A, B: TLineW): Integer;
var
  BranchA, BranchB: string;
  function RemoveBrackets(const L: TLineW):string;
  { Branch label without (J) }
  var
    Line: TLine;
    Junk: string;
    s, e: Integer;
  begin
    Line := L.Line;
    SplitCachedName(Line, Result, Junk);
    s := Pos('(',Result);
    e := Pos(')',Result);
    if (s > 0) and (e > s) then
      Delete(Result, s, e-s+2);
    Result := Trim(Result);
  end;
begin
  BranchA := RemoveBrackets(A);
  BranchB := RemoveBrackets(B);
  Result := CompareText(BranchB, BranchA);
  if Result = 0 then
    Result := B.Lower.J - A.Lower.J;
end;

procedure TLineForm.ShowButtonClick(Sender: TObject);
var
  i: Integer;
  V, Mi, Mx: Double;
begin
  with Grid.Selection do
    if nLines >= Top then
      if Top >= Bottom then
        GridDblClick(nil)
      else begin
        V := Lines[Top-1].Line.Frequency;
        Mi := V;
        Mx := Mi;
        for i := Top to Bottom - 1 do begin
          V := Lines[i].Line.Frequency;
          Mi := Min(V, Mi);
          Mx := Max(V, Mx)
        end;
        if Mi < Mx then
          with MainForm.Simulation do begin
            V := (Mx-Mi)/20;
            Fmin := Mi-V;
            Fmax := Mx+V;
            MainForm.ShowSimulation;
          end
        else
          GridDblClick(nil)
      end;
end;

function TLineForm.ShowLines(Simulation: TSimulation; FilterLines, DiscardDuplicates: Boolean): Integer;
{ Show the lines that have not been shown yet }
var
  i, j, f, VisibleRows: Integer;
  Rect: TGridRect;
  Problem: Exception;
  ABranch: string;
begin
  if nLines <= nVisibleLines then Exit;
  Grid.BeginUpdate;
  if DiscardDuplicates then begin
    { Initially just look for match in branch }
    i := nVisibleLines;
    while i < nLines do begin
      if StringRep(i+1,14,ABranch) and (FindBranch(ABranch, nVisibleLines) >= 0) then begin
        for j := i to nLines-2 do
          Lines[j] := Lines[j+1];
        Dec(nLines);
      end else
        Inc(i);
    end;
  end;
  f := nVisibleLines;
  Problem := nil;
  for i := f to nLines - 1 do
    with Lines[i].Line do
      if Assigned(EigenvaluesL) then
        Strength := Simulation.LineIntensity(Lines[i].Line, Simulation.PlotUnits, Problem);
  Problem.Free; { Easiest to ignore problems }
  if FilterLines then
    SortLines({$IFDEF FPC} @ {$ENDIF}ByIntensity, nVisibleLines, nLines-1)
  else
    SortLines({$IFDEF FPC} @ {$ENDIF}ByFrequency, nVisibleLines, nLines-1);
  if FilterLines then begin
    nLines := f + Min(nLines-f,FnLinesToShow);
    for i := nLines-1 downto f do
      if ( not FShowZeroIntensityLines and (Lines[i].Line.Strength = 0) )
        or (Lines[i].Line.Strength < FMinIntensity*Lines[f].Line.Strength) then
        nLines := i;
  end;
  Result := nLines - f;
  BringUp;
  if nLines = 0 then
    ClearButtonClick(nil)
  else begin
    Grid.RowCount := nLines+1;
    nPendingLines := 0;
    Grid.Row := nLines;
    Rect.Top := f+1;
    Rect.Bottom := Grid.RowCount-1;
    Rect.Left := 1;
    Rect.Right := Grid.ColCount-1;
    Grid.Selection := Rect;
    VisibleRows := {$IFDEF LCL_Off} Grid.Height div (Grid.DefaultRowHeight + Grid.GridLineWidth) - 1 {$ELSE} Grid.VisibleRowCount {$ENDIF};
    if Rect.Bottom > VisibleRows then
      Grid.TopRow := Min(Rect.Bottom - VisibleRows + 1, Rect.Top);
  end;
  Grid.EndUpdate;
end;

procedure TLineForm.UpdateGridLines(First, Last: Integer);
var
  i, dn: Integer;
  procedure MakeTitle(U, L: TEigenvalues);
  begin
    if Assigned(U) and Assigned(L) and (Length(U.Basis) > 1) then
      Grid.Title[14] := U.Basis[1].TxNameKey(U.Basis[1],L.Basis[1])
    else
      Grid.Title[14] := 'Branch';
    if Assigned(L) and (Length(L.Basis) > 1) then begin
      Grid.Title[LabelCol] := L.Basis[1].StateLabelKey;
      if Assigned(U) and (Length(U.Basis) > 1) then
        Grid.Title[LabelCol] := U.Basis[1].StateLabelKey + ' - ' + Grid.Title[15];
    end else
      Grid.Title[LabelCol] := 'Name';
  end;
  procedure Load(r, c, Value: Integer);
  begin
    if Value <> Low(Value) then
      Grid[r,c] := Value
    else
      Grid.Valid[r,c] := False;
  end;
  procedure LoadQ(r, c: Integer; Value: TQuantumNumber);
  begin
    if Value <> Low(Value) then
      Grid[r,c] := QuantumNumberToDisplayDouble(Value, GlobalMixture.QuantumNumberFormat)
    else
      Grid.Valid[r,c] := False;
  end;
begin
  if nLines <= 0 then begin
    ClearButtonClick(nil);
    Exit;
  end;
  Grid.BeginUpdate;
  Grid.RowCount := nLines+1;
  NotEmpty := True;
  for i := First+1 to Last+1 do
    with Lines[i-1] do begin
      Load(i,3,Upper.J);
      Load(i,4,Upper.Sym);
      Load(i,5,Line.IndexU);
      Load(i,7,Lower.J);
      Load(i,8,Lower.Sym);
      Load(i,9,Line.IndexL);
      Grid[i,10] := Line.Frequency;
      Grid[i,11] := Weight;
      Grid[i,12] := Line.Strength;
      Grid[i,13] := Line.Width;
      if HasMeasuredIntensity then
        Grid[i,16] := IObs
      else if not IsNanOrInf(IObs) and (IObs > 0) and AdvancedItem.Checked then begin
        HasMeasuredIntensity := True;
        Grid.ColCount := LabelCol+2;
        Grid.Title[16] := 'Iobs';
        Grid[i,16] := IObs;
        Grid.AutoAdjustColumn(16);
      end;
    end;
  if First = 0 then begin
    MakeTitle(Lines[0].Line.EigenvaluesU,Lines[0].Line.EigenvaluesL);
    Grid.AutoAdjustColumn(12);
    Grid.AutoAdjustColumn(14,False,True);
    Grid.AutoAdjustColumn(0);
  end;
  Grid.Invalidate;
  Grid.EndUpdate;
end;

function TLineForm.IsUnassigned(Index: Integer): Boolean;
begin
  Result := IsNanOrInf(Lines[Index].Weight) or (Lines[Index].Weight <= 0);
end;

type
  TLineReader = class(TObservationsReaderBase)
    function ReadQuantumNumber(var a: Integer; const si: string): Boolean; override;
  end;

function TLineReader.ReadQuantumNumber(var a: Integer; const si: string): Boolean;
begin
  Result := not TryStrToQuantumNumber(si, a, GlobalMixture.QuantumNumberFormat);
end;

procedure TLineForm.JUpButtonClick(Sender: TObject);
var
  LineW: TLineW;
  Branch, OrgBranch, TxName, RG: string;
  Obs: TObservation;
  StateL, StateU: TState;
  Calc, W: Double;
  FoundPeak: TFoundPeak;
  i, iLine, p: Integer;
  BumpAssignments: Boolean;
  Plot: TPlot;
begin
  if not TryStrToFloatEx(AcceptEdit.Text,W) then
    raise EReportException.Create('Bad accept window:'+AcceptEdit.Text);
  W := Abs(W);
  BumpAssignments := Grid.Selection.Bottom > Grid.Selection.Top;
  for iLine := Grid.Selection.Top to Grid.Selection.Bottom do begin
    if not MakeObs(Obs, iLine, False) then
      Exit;
    LineW := Lines[iLine-1];
    SplitCachedName(LineW.Line, OrgBranch, TxName);
    Branch := Trim(OrgBranch);
    if Branch = '' then
      Continue;
    StateL := LineW.Line.EigenvaluesL.DominantBasis[Obs.AIndexL].State;
    StateU := LineW.Line.EigenvaluesU.DominantBasis[Obs.AIndexU].State;
    Branch := StateL.BumpBranchQno(Branch, (Sender as TComponent).Tag*Increment.Value, qtJ);
    if Branch = '' then
      Continue;
    if BumpKa.Checked then
      Branch := StateL.BumpBranchQno(Branch, (Sender as TComponent).Tag*Increment.Value, qtA);
    if Branch = '' then
      Continue;
    if BumpKc.Checked then
      Branch := StateL.BumpBranchQno(Branch, (Sender as TComponent).Tag*Increment.Value, qtB);
    if Branch = '' then
      Continue;
    if DummyReader = nil then begin
      DummyReader := TLineReader.Create;
      DummyReader.BranchRank := 2;
      DummyReader.DefaultSymmetry := -1;
    end;
    if StateL is TLinear then begin
      if Obs.SymU = Obs.SymL then
        DummyReader.BranchRank := 0
      else
        DummyReader.BranchRank := 2;
      if Length(Branch) < Length(OrgBranch) then
        Branch := StringOfChar(' ', Length(OrgBranch)-Length(Branch)-1)+Branch;
    end;
    if not StateL.ReadBranchFormat(Branch, Obs, StateU, DummyReader) then
      Continue;
    if BumpAssignments then begin
      { Copy back updated fields }
      if LineW.Lower.J >= 0 then
        LineW.Lower.J := Obs.JL;
      if LineW.Lower.Sym >= 0 then
        LineW.Lower.Sym := Obs.SymL;
      if LineW.Upper.J >= 0 then
        LineW.Upper.J := Obs.JU;
      if LineW.Upper.Sym >= 0 then
        LineW.Upper.Sym := Obs.SymU;
      if LineW.Line.IndexL > 0 then
        LineW.Line.IndexL := Obs.AIndexL;
      if LineW.Line.IndexU > 0 then
        LineW.Line.IndexU := Obs.AIndexU;
      p := 0;
      if not IsNanOrInf(LineW.Weight) and (LineW.Weight > 0) then begin
        { Preserve assignment name }
        p := Pos(':', LineW.Line.CachedName);
        if p > 0 then begin
          p := Pos(':', LineW.Line.CachedName, p+1);
        end;
      end;
      if p > 0 then
        LineW.Line.CachedName := Obs.Name + ' ' + Copy(LineW.Line.CachedName, p, Length(LineW.Line.CachedName))
      else
        LineW.Line.CachedName := Obs.Name;
      Lines[iLine-1] := LineW;
    end else begin
      if not CalculateObsEx(Obs, Calc, RG, False, obIntensity, LineW.Line.EigenvaluesU, LineW.Line.EigenvaluesL, LineW.Line) then
        Continue;
      Calc := LineW.Line.Frequency;
      MainForm.CentrePlot(Calc, GlobalMixture.Units);
      { Search for new branch in current line list }
      i := FindBranch(Branch, nLines);
      if i >= 0 then begin
        { shift to original measurement }
        HighlightRow(i+1);
        if not IsNanOrInf(Lines[i].Weight) and (Lines[i].Weight > 0) then
          Exit;
      end else begin
        { New line }
        AddLine(LineW.Line, GlobalMixture.Units);
        ShowLines(GlobalMixture.Simulations[0],True,False);
      end;
      if GlobalOverlays = nil then
        Continue;
      Plot := nil;
      with GlobalOverlays do
        for i := 0 to nPlots - 1 do
          if not Plots[i].Hide then begin
            Plot := Plots[i];
            Break;
          end;
      if (Plot <> nil) and Plot.FindPeak(Calc-W,Calc+W, FoundPeak, GlobalMixture.Simulations[0].PlotUnits) then begin
        { If linelist, require single line found by peak finder }
        if not ((Plot is TExperimentalPlot) and TExperimentalPlot(Plot).IsLineList and (FoundPeak.nAbove > 1)) then
          Plot.AddPeak(FoundPeak.Position, GlobalMixture.Simulations[0].PlotUnits, 0);
      end;
    end;
  end;
  if BumpAssignments then
    UpdateGridLines(Grid.Selection.Top-1, Grid.Selection.Bottom-1);
  FitButtonClick(TestFitButton);
end;

procedure TLineForm.BringUp;
begin
  if WindowState in [wsMinimized, wsMaximized] then
    WindowState := wsNormal;
  if not Visible then
    Show
  else
    BringToFront;
end;

procedure TLineForm.CDplotClick(Sender: TObject);
var
  LW: TLoomisWood;
  Plot: TPlot;
  i, n: Integer;
  Peak: TPeak;
begin
  if GlobalOverlays = nil then
    Exit;
  LW := GlobalOverlays.ChildWithClass(TLoomisWood) as TLoomisWood;
  if LW = nil then begin
    Plot := nil;
    with GlobalOverlays do
      for i := 0 to nPlots - 1 do
        if not Plots[i].Hide then begin
          if Plot = nil then
            Plot := Plots[i];
          LW := Plots[i].ChildWithClass(TLoomisWood)  as TLoomisWood;
          if LW <> nil then
            Break;
        end;
    if LW = nil then begin
      if Plot = nil then
        Exit;
      LW := TLoomisWood.Create(Plot);
      LW.Name := 'CD';
    end;
  end;
  n := Grid.Selection.Bottom - Grid.Selection.Top + 1;
  LW.nPlots := n+1;
  LW.ShowProduct := True;
  LW.Active := True;
  LW.Peaks.Clear;
  for i := Grid.Selection.Top-1 to Grid.Selection.Bottom-1 do begin
    Peak.Position := Lines[i].Line.Frequency;
    Peak.Actual := i-Grid.Selection.Top+1;
    Peak.Weight := 1;
    LW.Peaks.AddPeak(Peak);
  end;
  LW.ShowLoomisWood(nil);
end;

procedure TLineForm.ClearButtonClick(Sender: TObject);
var
  i, nAssigned: Integer;
begin
  if Assigned(Sender) then begin
    nAssigned := 0;
    for i := 0 to nVisibleLines-1 do
      if not IsUnassigned(i) then
        Inc(nAssigned);
    if (nAssigned > 0)
       and (MessageDlg(Format('%d assigned lines - are you sure you want to delete them?', [nAssigned]),
                      mtWarning, mbOKCancel, 0) <> mrOK) then
      Exit;
  end;
  Grid.RowCount := 2;
  NotEmpty := False;
  nLines := 0;
  Grid.ColCount := 16;
  HasMeasuredIntensity := False;
  for i := 1 to Grid.ColCount-1 do
    Grid.Valid[1,i] := False;
  Grid.ColWidths[LabelCol] := 150;
  Grid.Invalidate;
end;

procedure TLineForm.GridDblClick(Sender: TObject);
begin
  if nLines >= Grid.Selection.Top then
    MainForm.CentrePlot(Lines[Grid.Selection.Top-1].Line.Frequency, MainForm.Simulation.PlotUnits);
end;

procedure TLineForm.GridDrawCellContents(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
var
  buffer: string;
  Pitch: TFontPitch;
  FontName: TFontName;
  L: Integer;
  v: Double;
  Alignment: TAlignment;
begin
  Drawn := StringRep(ARow, ACol, buffer);
  if not Drawn then begin
    Grid.CurrentFixedSizeNumbers := ACol in [10,12,16];
    if ACol in [4,5,8,9] then
      Grid.CurrentCellAlignment := taRightJustify;
  end else if ACol = LabelCol then begin
    Pitch := Grid.Canvas.Font.Pitch;
    FontName := Grid.Canvas.Font.Name;
    Grid.Canvas.Font.Name := DefaultFixedName;
    {$IFDEF MSWINDOWS}
    Grid.Canvas.Font.Size := Grid.Canvas.Font.Size + 1;
    {$ENDIF}
    Grid.Canvas.Font.Pitch := fpFixed;
    case Lines[ARow-1].Line.Colour of
      clWhite, clDefault, clNone: ;
    else
      if not (gdSelected in State) then
        Grid.Canvas.Font.Color := Lines[ARow-1].Line.Colour;
    end;
    L := Grid.Canvas.TextWidth(buffer) + 10;
    if FLongestNameLength < L then
      FLongestNameLength := L;
    TextInGrid(Grid.Canvas, Rect, buffer);
    {$IFDEF MSWINDOWS}
    Grid.Canvas.Font.Size := Grid.Canvas.Font.Size - 1;
    {$ENDIF}
    Grid.Canvas.Font.Name := FontName;
    Grid.Canvas.Font.Pitch := Pitch;
  end else begin
    if ACol in [3,7] then
      Alignment := taRightJustify
    else
      Alignment := taLeftJustify;
    TextInGrid(Grid.Canvas, Rect, buffer, Alignment)
  end;
end;

procedure TLineForm.AllButtonClick(Sender: TObject);
var
  Rect: TGridRect;
begin
  Rect.Top := 1;
  Rect.Bottom := Grid.RowCount-1;
  Rect.Left := 1;
  Rect.Right := Grid.ColCount-1;
  Grid.Selection := Rect;
end;

procedure TLineForm.ArrangeForm(Screen: TRect);
var
  R, MainRect: TRect;
begin
  inherited;
  MainRect := ScreenBounds(MainForm);
  R := ScreenBounds(Self);
  R.Left := MainRect.Left;
  R.Right := MainRect.Right;
  OffsetRect(R,0,MainRect.Bottom-R.Top);
  if R.Bottom > Screen.Bottom then
    OffsetRect(R,0,Screen.Bottom-R.Bottom);
  if LogForm.Visible then
    Dec(R.Right,25);
  SetScreenBounds(Self,R);
end;

procedure TLineForm.NoneButtonClick(Sender: TObject);
begin
  Grid.CancelSelection;
end;

procedure TLineForm.CutButtonClick(Sender: TObject);
begin
  CopyButtonClick(Sender);
  DeleteButtonClick(Sender);
end;

procedure TLineForm.CopyToStream(Stream: TStream; Row1, Row2: Integer; All: Boolean;
                                 LineListType: TLineListType; PreserveLineNumbers: Boolean;
                                 PlatformDefaultCodePage: Boolean);
var
  F: TextFile;
  i: Integer;
  buffer, LastBranch: string;
  LastNQN: Integer;
  BranchRank: Integer;
  procedure ProcessColumn(j: Integer);
  begin
    case j of
    3,7: begin
           StringRep(i, j, buffer);
           if buffer = '' then
             Write(F, ' - ')
           else
             Write(F, buffer:3, ' ');
         end;
    4,8: if not Grid.Valid[i,j] then
           Write(F, '-1 ')
         else
           Write(F, FloatToStr(Grid[i,j]):2, ' ');
     10: begin
           buffer := FloatToStr(Grid[i,j]);
           Write(F, ' ', buffer, ' ':20-Length(buffer));
         end;
     11: if Grid.Valid[i,j] then
           Write(F, FloatToStr(Grid[i,j]), ' ')
         else
           Write(F, '- ');
     12, 13: if Grid.Valid[i,j] then
               Write(F, dtosm(Grid[i,j],6,-1,True):6, ' ') { Limit Strength and width to 6 characters }
             else
               Write(F, '       ');
     14: begin StringRep(i,j,buffer); Write(F,': ', buffer); end;
     LabelCol: begin StringRep(i,j,buffer); WriteLn(F,' : ', buffer); end;
    else
      if StringRep(i, j, buffer) then
        Write(F, buffer:3, ' ')
      else if not Grid.Valid[i,j] then
        Write(F, ' - ')
      else
        Write(F, FloatToStr(Grid[i,j]):2, ' ');
    end;
  end;
  procedure MakeStandard;
  var
    col: Integer;
  begin
    for col := 1 to LabelCol do { Omit Iobs for now }
      ProcessColumn(col);
  end;
  function ExtractSimple(s: string; Manifold: TManifold):Boolean;
  var
    p, NQN: Integer;
    InWord: Boolean;
  begin
    Result := False;
    if not Assigned(Manifold) then
      Exit;
    { Given the part transition name that specifies a state, extract the "Simple Format" quantum numbers }
    { Strip molecule if present }
    if Pos(Manifold.Molecule.Name+' ',s) = 1 then
      Delete(s,1,Length(Manifold.Molecule.Name)+1);
    { Strip manifold if present }
    if Pos(Manifold.Name+' ',s) = 1 then
      Delete(s,1,Length(Manifold.Name)+1);
    { Strip state name; logic here allows state name to include spaces }
    for p := 1 to Length(s)-1 do { -1 as blank string not allowed }
      if (s[p] = ' ') and Assigned(Manifold.StateWithName(Copy(s,1,p-1))) then begin
        s := Copy(s,p+1,Length(s));
        Result := true;
        Break;
      end;
    if not Result and (Manifold.nStates <> 1) then
      Exit;
    Result := True;
    Manifold.Molecule.FixSimpleFormat(s);
    buffer := buffer + s + ' ';
    { Check NQN }
    InWord := False;
    NQN := 0;
    for p := 1 to Length(s) do
      if (s[p] <> ' ') <> Inword then begin
        Inword := not Inword;
        if InWord then
          Inc(NQN);
      end;
    if LastNQN <> NQN then begin
      WriteLn(F, 'NQN ', NQN);
      if PreserveLineNumbers then
        WriteLn(F, 'linenumber ',i);
      LastNQN := NQN;
    end;
  end;
  procedure MakeTail;
  var
    p: Integer;
  begin
    ProcessColumn(10); { Position }
    ProcessColumn(11); { Weight }
    ProcessColumn(12); { Intensity }
    StringRep(i,LabelCol,buffer); { Comment }
    { strip TxName part }
    p := Pos(':',buffer);
    if p > 0 then
      Delete(buffer, 1, p);
    if Trim(buffer) <> '' then
      WriteLn(F,' : ', Trim(buffer))
    else
      WriteLn(F);
  end;
  procedure DetermineBranchRank(MNode: TMNode);
  var
    i: Integer;
  begin
    if MNode is TTransitionMoments then begin
      with TTransitionMoments(MNode) do
        for i := 0 to ComponentCount-1 do
          if Components[i] is TTransitionMoment then
            with TTransitionMoment(Components[i]) do
              case BranchRank of
                -1: BranchRank := Rank mod 4;
                -2: { Clash };
              else
                if Minus(Rank-BranchRank) < 0 then
                  BranchRank := -2;
              end;
    end else
      for i := 0 to MNode.ComponentCount-1 do
        DetermineBranchRank(MNode.Components[i] as TMNode);
  end;
  procedure MakeSimple;
  var
    Pre, Upper, Lower, Post: string;
    Molecule: TMolecule;
  begin
    Molecule := GlobalMixture.FindMolecule(Lines[i-1].Molecule);
    if not Assigned(Molecule) then
      if (GlobalMixture.nMolecules = 1) and (GlobalMixture.Species[0].Count = 1) then
        Molecule := GlobalMixture.Species[0].Isotopomer[0]
      else
        Molecule := nil; { Need to know molecule }
    if Assigned(Molecule) then begin
      StringRep(i,LabelCol,buffer);
      SplitObsName(buffer, Pre, Upper, Lower, Post);
      buffer := '';
    end;
    if not (Assigned(Molecule)
            and ExtractSimple(Upper, Molecule.ManifoldWithName(Lines[i-1].Upper.Manifold))
            and ExtractSimple(Lower, Molecule.ManifoldWithName(Lines[i-1].Lower.Manifold))) then begin
      Write(F, '#Unconverted: ');
      MakeStandard;
      Exit;
    end;
    Write(F, buffer);
    MakeTail;
  end;
  procedure MakeBranch;
  begin
    StringRep(i,14,buffer);
    if Trim(buffer) = '' then begin
      WriteLn(F); { Preserve line numbers }
      Exit; { No branch - skip }
    end;
    Write(F, buffer, ' ');
    MakeTail;
  end;
  procedure MakeBranchExpr;
  var
    J, Branch: string;
    p: Integer;
  begin
    StringRep(i,14,buffer);
    Branch := Trim(Buffer);
    p := Pos('(',Branch);
    if p <= 0 then begin
      WriteLn(F); { Preserve line numbers }
      Exit; { No branch - skip }
    end;
    { Decompose branch into J and general branch }
    J := Copy(Branch,p+1,Length(Branch)-p-1);
    Delete(Branch,p,Length(Branch));
    Branch := Branch + '(J)';
    if Branch <> LastBranch then begin
      LastBranch := Branch;
      WriteLn(F, 'BranchExpr ',LastBranch);
    end;
    Write(F, J:3);
    MakeTail;
  end;
begin
  AssignStream(F, Stream, PlatformDefaultCodePage);
  Append(F);
  BranchRank := -1;
  DetermineBranchRank(GlobalMixture);
  if (LineListType in [llBranch,llBranchExpr]) and (BranchRank = 0) then
    WriteLn(F, 'BranchRank 2');
  LastNQN := 0;
  LastBranch := '';
  for i := Row1 to Row2 do begin
    if All or Grid.Valid[i,11] then
      case LineListType of
        llStandard:   MakeStandard;
        llBranch:     MakeBranch;
        llSimple:     MakeSimple;
        llBranchExpr: MakeBranchExpr;
      end;
  end;
  System.Close(F);
end;

constructor TLineForm.Create(AOwner: TComponent);
begin
  FnLinesToShow := 10;
  FMinIntensity := 0.05;
  FShowZeroIntensityLines := False;
  FMinAssignedIntensity := 0.25;
  inherited;
end;

procedure TLineForm.CopyButtonClick(Sender: TObject);
var
  s: THugeMemoryStream;
  LineListType: TLineListType;
begin
  if not NotEmpty then Exit;
  if Sender = CopyBranchTableItem then
    LineListType := llBranch
  else if Sender = CopySimpleFormatItem then
    LineListType := llSimple
  else if Sender = CopyBranchExprFormatItem then
    LineListType := llBranchExpr
  else
    LineListType := llStandard;
  s := THugeMemoryStream.Create;
  try
    CopyToStream(s, Grid.Selection.Top, Grid.Selection.Bottom, True, LineListType, False, True);
    s.CopyToClipboard(CF_PREFTEXT);
  finally
    s.Free;
  end;
end;

procedure TLineForm.DeleteButtonClick(Sender: TObject);
var
  i, d, n, e: Integer;
begin
  for i := Grid.Selection.Top-1 to Grid.Selection.Bottom-1 do
    SetNan(Lines[i].Line.Frequency);
  FinishDiscard;
end;

procedure TLineForm.CloseButtonClick(Sender: TObject);
begin
  Close;
end;

type
  ClipboardMeasurement = packed record
    Value: Double;
    Signature: array [0..3] of AnsiChar;
    Filename: AnsiChar;
  end;
  PClipboardMeasurement = ^ClipboardMeasurement;

procedure TLineForm.AssignButtonClick(Sender: TObject);
var
  Value: Double;
  CHandle: TGHandle;
  Clip: PClipboardMeasurement;
  FileName: string;
begin
  if ClipBoard.HasFormat(CF_TEXT) then begin
    if CF_Measurement = 0 then
      CF_Measurement := RegisterClipboardFormat('Measurement');
    if not TryStrToFloatEx(Trim(ClipBoard.AsText),Value) then
      raise EReportException.Create(ClipBoard.AsText+' is not a valid number');
    if ClipBoard.HasFormat(CF_Measurement) then begin
      CHandle := TGHandle.Create;
      try
        CHandle.SetFromClipboard(CF_Measurement);
        Clip := PClipboardMeasurement(CHandle.Base);
        if Clip^.Signature = 'MEAS' then begin
          Value := Clip^.Value;
          FileName := ExtractFileName(PAnsiChar(@Clip^.FileName));
        end;
      finally
        CHandle.Free;
      end;
    end else
      FileName := '';
    AssignLines(Value, MainForm.Simulation.PlotUnits, Filename, nil, True);
  end;
end;

procedure TLineForm.AssignLines(APosition: Double; Units: TResonanceUnits; SourceName: string;
  Source: TMNode; Force: Boolean; AErr: Double);
var
  StrMax, E: Double;
  s, Branch: string;
  i, p: Integer;
begin
  if not (NotEmpty and Visible) then
    Exit;
  if not Force and (not Auto.Checked or (SourceName = 'Calculated')) then
    Exit;
  StrMax := Lines[Grid.Selection.Top-1].Line.Strength;
  for i := Grid.Selection.Top to Grid.Selection.Bottom do begin
    { Avoid assignment from calculation of same transition }
    if Assigned(Source) then
      with Lines[i-1].Line do begin
        if Assigned(EigenValuesU) and Source.IsAncestorOf(EigenvaluesU.Manifold) then
          Exit;
        if Assigned(EigenValuesL) and Source.IsAncestorOf(EigenvaluesL.Manifold) then
          Exit;
      end;
    if StrMax < Lines[i-1].Line.Strength then
      StrMax := Lines[i-1].Line.Strength;
  end;
  for i := Grid.Selection.Top to Grid.Selection.Bottom do
    if Lines[i-1].Line.Strength > StrMax*FMinAssignedIntensity then begin
      if AErr > 0 then
        E := ConvertWidth(AErr, APosition, Units, MainForm.Simulation.PlotUnits)
      else
        E := 1;
      UpdateAssignment(i, Convert(APosition,Units,MainForm.Simulation.PlotUnits), E, ExtractFileName(SourceName));
    end;
  UpdateGridLines(Grid.Selection.Top-1, Grid.Selection.Bottom-1);
end;

procedure TLineForm.UnAssignLine(Index: Integer);
begin
  Lines[Index-1].Weight := 0;
  UpdateGridLines(Index-1, Index-1);
end;

procedure TLineForm.ConsiderRename(MNode: TMNode; OldName: string);

  function Sweep(MakeChange: Boolean): Boolean;
  var
    i, P, Pc: Integer;
    Branch, TxName, OldNameEx, NewName: string;
    procedure CheckName(var s: string);
    begin
      if s = OldName then begin
        Result := True;
        if MakeChange then
          s := NewName;
      end;
    end;
  begin
    Result := False;
    if MNode is TState then begin
      OldNameEx := TState(MNode.Owner).Name + ' ' + OldName;
      NewName := TState(MNode.Owner).Name + ' ' + MNode.Name;
    end else begin
      OldNameEx := OldName;
      NewName := MNode.Name;
    end;
    for i := 1 to nVisibleLines do begin
      SplitCachedName(Lines[i-1].Line, Branch, TxName);
      if MNode is TMolecule then
        CheckName(Lines[i-1].Molecule)
      else begin
        if MNode is TManifold then begin
          CheckName(Lines[i-1].Upper.Manifold);
          CheckName(Lines[i-1].Lower.Manifold);
        end;
        if Pos(OldNameEx+' ', TxName) = 1 then begin
          Result := True;
          if MakeChange then begin
            TxName := NewName + Copy(TxName, Length(OldNameEx)+1, Length(TxName));
            SetLineByIndex(Lines[i-1], TxName, LabelCol);
          end;
        end;
        P := pos('- '+OldNameEx, TxName);
        if P < 1 then
          P := pos('-'#9+OldNameEx, TxName);
        { Avoid replacing text in filenames or other comments }
        Pc := pos(':', TxName);
        if Pc < 1 then
          Pc := Length(TxName);
        if (P > 0) and (P < Pc) then begin
          Result := True;
          if MakeChange then begin
            TxName := Copy(TxName, 1, P+1) + NewName + Copy(TxName, P+2+Length(OldNameEx), Length(TxName));
            SetLineByIndex(Lines[i-1], TxName, LabelCol);
          end;
        end;
      end;
    end;
  end;

begin
  { Consider updating linelist to eflect name change }
  if not (((MNode is TMolecule) or (MNode is TManifold) or (MNode is TState)) and Sweep(False)) then
    Exit;
  if mrYes <> MessageDlg('Update linelist to reflect name change?', mtConfirmation, [mbYes, mbNo],0) then
    Exit;
  Sweep(True);
  Grid.Invalidate;
end;

function TLineForm.StringRep(ARow, ACol: Integer; var buffer: string): Boolean;
var
  Junk: string;
  function JValue(J: TQuantumNumber):string;
  begin
    if J <> Low(TQuantumNumber) then
      Result := GlobalMixture.QuantumNumberToStr(J)
    else
      Result := '';
  end;
begin
  Result := NotEmpty and (ARow > 0);
  if Result then
    case ACol of
      1: buffer := Lines[ARow-1].Molecule;
      2: buffer := Lines[ARow-1].Upper.Manifold;
      3: buffer := JValue(Lines[ARow-1].Upper.J);
      6: buffer := Lines[ARow-1].Lower.Manifold;
      7: buffer := JValue(Lines[ARow-1].Lower.J);
     14: begin
           SplitCachedName(Lines[ARow-1].Line, buffer, Junk);
           buffer := Trim(buffer);
         end;
     LabelCol: SplitCachedName(Lines[ARow-1].Line, Junk, buffer);
    else
      Result := False;
    end;
end;

procedure TLineForm.SuppressButtonClick(Sender: TObject);
var
  LineList: TExperimentalPlot;
  i, p: Integer;
  f, dW: Double;
begin
  LineList := ActiveLineList;
  LineList.LastPoint := -1;
  LineList.Sort;
  dW := StrToFloatEx(AcceptEdit.Text);
  for i := Grid.Selection.Top-1 to Grid.Selection.Bottom-1 do
    if Lines[i].Weight > 0 then begin
      f := LineList.DeAdjustFrequency(Convert(Lines[i].Line.Frequency,GlobalMixture.Units,LineList.Units));
      p := ClosestPoint(f, LineList.Frequency);
      if Abs(Convert(LineList.LocalFrequency(p),LineList.Units,GlobalMixture.Units)-Lines[i].Line.Frequency) <= dW then begin
        SetNan(LineList.Frequency[p]);
        LineList.Sort;
      end;
    end;
  MainForm.ConsiderReplot;
end;

procedure TLineForm.GridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: String);
begin
  StringRep(ARow, ACol, Value);
end;

procedure TLineForm.GridChangeCellQuery(Sender: TObject; Col, Row: Integer; var CanChange: Boolean);
begin
  // Was : CanChange := CanChange and (Col in [10,11,14]);
end;

procedure TLineForm.SavetofileItemClick(Sender: TObject);
var
  Stream: TFileStream;
  r: Integer;
  LineListType: TLineListType;
begin
  if Sender = InBranchFormatItem then
    LineListType := llBranch
  else if Sender = InSimpleFormatItem then
    LineListType := llSimple
  else if Sender = InBranchExprFormatItem then
    LineListType := llBranchExpr
  else
    LineListType := llStandard;
  if SaveDialog.Execute then begin
    if FileExists(SaveDialog.FileName) then
      r := MessageDlg('File exists - append to file instead of overwriting?', mtWarning, [mbYes, mbNo, mbCancel], 0)
    else
      r := mrNo;
    case r of
      mrYes: begin; Stream := TFileStream.Create(SaveDialog.FileName, fmOpenReadWrite); Stream.Seek(0, soFromEnd); end;
      mrNo:  Stream := TFileStream.Create(SaveDialog.FileName, fmCreate);
    else
      Exit;
    end;
    try
      CopyToStream(Stream, 1, nLines, True, LineListType, False, False);
    finally
      Stream.Free;
    end;
    LogForm.ObsFileName.FileName := SaveDialog.FileName;
  end;
end;

procedure TLineForm.SetLineByIndex(var Line: TLineW; const Value: string; Index: Integer);
{ This copy back is required for loading grid from .pgo files  }
{ Note that quantum numbers are doubled }
var
  v: Double;
  Branch, TxName: string;
  iv: Integer;
begin
  if Index in [10,11,12,13,16] then
    if not TryStrToFloatEx(Value, v) then
      SetNan(v);
  if Index in [3,4,5,7,8,9] then
    if not TryStrToInt(Value,iv) then
      if Index in [5,9] then
        iv := Low(TIndex)
      else
        iv := Low(Integer);
  case Index of
     1: Line.Molecule := Value;
     2: Line.Upper.Manifold := Value;
     3: Line.Upper.J := iv;
     4: Line.Upper.Sym := iv;
     5: Line.Line.IndexU := iv;
     6: Line.Lower.Manifold := Value;
     7: Line.Lower.J := iv;
     8: Line.Lower.Sym := iv;
     9: Line.Line.IndexL := iv;
    10: Line.Line.Frequency := v;
    11: Line.Weight := v;
    12: Line.Line.Strength := v;
    13: Line.Line.Width := v;
    14,LabelCol:
      begin
        SplitCachedName(Line.Line, Branch, TxName);
        if Index = 14 then
          Branch := Value
        else
          TxName := Value;
        TxName := Branch + ' : ' + TxName;
        Line.Line.CachedName := TxName;
      end;
    16: Line.IObs := v;
  end;
end;

procedure TLineForm.GridSetCell(Sender: TObject; ACol, ARow: Integer; var v: Double; var Valid: Boolean);
var
  buffer: string;
begin
  { This allows the delete key to act on cell ranges }
  if not Valid and (ARow <= Length(Lines)) then begin
    SetLineByIndex(Lines[ARow-1],'',ACol);
    Grid.Changed(ACol, ARow);
  end;
end;

procedure TLineForm.GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
var
  Q: TQuantumNumber;
begin
  if ACol in [3,7] then begin
    if TryStrToQuantumNumber(Value, Q, GlobalMixture.QuantumNumberFormat) then
      case ACol of
        3: Lines[ARow-1].Upper.J := Q;
        7: Lines[ARow-1].Lower.J := Q;
      end;
  end else
    SetLineByIndex(Lines[ARow-1],Value,ACol);
end;

function TLineForm.nVisibleLines: Integer;
begin
  if NotEmpty then
    Result := Grid.RowCount-1
  else
    Result := 0
end;

procedure TLineForm.OptionsItemClick(Sender: TObject);
begin
  LoadOptionsDialog;
  if LineFrmOptDlg.ShowModal = mrOk then
    UnLoadOptionsDialog;
end;

procedure TLineForm.Outsidelinelistrange1Click(Sender: TObject);
var
  LL: TExperimentalPlot;
  i: Integer;
  Mx, Mi: Double;
begin
  LL := ActiveLinelist;
  Mi := Convert(LL.HMin, LL.Units, GlobalMixture.Units);
  Mx := Convert(LL.HMax, LL.Units, GlobalMixture.Units);
  InOrder(Mi, Mx);
  for i := 0 to nLines-1 do
    if IsUnassigned(i) and ((Lines[i].Line.Frequency < Mi) or (Lines[i].Line.Frequency > Mx)) then
      SetNan(Lines[i].Line.Frequency);
  FinishDiscard;
end;

procedure TLineForm.OutsidePlotRangeClick(Sender: TObject);
var
  i: Integer;
  Mx, Mi: Double;
begin
  Mi := MainForm.G.Xmin;
  Mx := MainForm.G.Xmax;
  InOrder(Mi, Mx);
  for i := 0 to nLines-1 do
    if IsUnassigned(i) and ((Lines[i].Line.Frequency < Mi) or (Lines[i].Line.Frequency > Mx)) then
      SetNan(Lines[i].Line.Frequency);
  FinishDiscard;
end;

procedure TLineForm.ResetButtonClick(Sender: TObject);
begin
  if Assigned(MyAutoFitter) then
    LoadAutoFitAssignment(MyAutoFitter, -1);
end;

procedure TLineForm.LowIntensity1Click(Sender: TObject);
var
  i: Integer;
  Mx: Double;
begin
  Mx := 0;
  for i := 0 to nLines-1 do
    if IsUnassigned(i) then
      Mx := Max(Lines[i].Line.Strength, Mx);
  Mx := Mx * FMinIntensity;
  for i := 0 to nLines-1 do
    if IsUnassigned(i) and (Lines[i].Line.Strength < Mx) then
      SetNan(Lines[i].Line.Frequency);
  FinishDiscard;
end;

procedure TLineForm.DeleteUnassignedClick(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to nLines-1 do
    if IsUnassigned(i) then
      SetNan(Lines[i].Line.Frequency);
  FinishDiscard;
end;

destructor TLineForm.Destroy;
begin
  FreeAndNil(DummyReader);
  LineForm := nil;
  FreeAndNil(MyAutoFitter);
  inherited;
end;

function TLineForm.MakeObs(var Obs: TObservation; Row: Integer; AlwaysCalculate: Boolean): Boolean;
var
  Stream: TStream;
  SavedObservations: TObservationArray;
  Calc: Double;
  RG: string;
  Ev: TEigenvalues;
  LineW: TLineW;
  NewLine: TLine;
begin
  Result := True;
  LineW := Lines[Row-1];
  if (LineW.Line.EigenvaluesL = nil) or (LineW.Line.EigenvaluesU = nil) then begin
    { Regenerate full transition info if not already available }
    SavedObservations := Observations;
    Stream := TTextStream.Create(nil,False);
    try
      CopyToStream(Stream, Row, Row, True, llStandard, True, False);
      Observations := nil;
      Stream.Position := 0;
      AddToObservations(GlobalMixture, '', Stream, obLine, tkAll);
      Obs := Observations[0];
    finally
      Observations := SavedObservations;
      Stream.Free;
    end;
    AlwaysCalculate := True;
  end else
    with Obs do begin
      IntensityUnits:= GlobalMixture.Simulations[0].IntensityUnits;
      ExField := GlobalMixture.Simulations[0].ExField;
      RemoveSpins := 0;
      Colour := LineW.Line.Colour;
      Ev := LineW.Line.EigenvaluesL;
      SymL := Ev.Sym;
      JL := Ev.J;
      ManifoldL := Ev.Manifold;
      IndexL := Max(0,LineW.Line.IndexL);
      AIndexL := IndexL;
      Ev := LineW.Line.EigenvaluesU;
      SymU := Ev.Sym;
      JU := EV.J;
      ManifoldU := Ev.Manifold;
      IndexU := Max(0,LineW.Line.IndexU);
      AIndexU := IndexU;
      QuantumNumberFormat := GlobalMixture.QuantumNumberFormat;
      MNode := nil;
    end;
  if AlwaysCalculate then begin
    Result := CalculateObsEx(Obs, Calc, RG, True, obLine, Lines[Row-1].Line.EigenvaluesU, Lines[Row-1].Line.EigenvaluesL, NewLine);
    if Result then begin
      if IsNanOrInf(LineW.Weight) or (LineW.Weight <= 0) then begin
        Lines[Row-1].Line.Frequency := NewLine.Frequency;
        Lines[Row-1].Line.CachedName := NewLine.CachedName;
      end;
      Lines[Row-1].Line.Strength := NewLine.Strength;
    end;
  end;
end;

procedure TLineForm.MinusButtonClick(Sender: TObject);
var
  d: Double;
  Stream: THugeMemoryStream;
begin
  if Grid.Selection.Top >= Grid.Selection.Bottom then begin
    MessageDlg('The button takes the difference between the first and last selected line', mtInformation, [mbOK], 0);
    Exit;
  end;
  if CF_Difference = 0 then
    CF_Difference := RegisterClipboardFormat('Difference');
  with Grid.Selection do
    d := Abs(Grid[Bottom,10] - Grid[Top,10]);
  Difference.Caption := dtosm(d, 12, 4);
  Stream := THugeMemoryStream.Create;
  Clipboard.Open;
  try
    Clipboard.AsText := FloatToStr(d);
    Stream.Write(d, Sizeof(Double));
    Stream.CopyToClipboard(CF_Difference);
  finally
    Clipboard.Close;
    Stream.Free;
  end;
  BroadcastName('WM_New_Difference');
end;

procedure TLineForm.MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
begin
  inherited;
  if (Change = mnCacheCleared) and (Sender.RootNode is TMixture) then
    ClearTLines;
end;

procedure TLineForm.StrengthSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByIntensity, 0, nLines-1);
end;

procedure TLineForm.UpperStateSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByUpperState, 0, nLines-1);
end;

procedure TLineForm.BranchItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByBranch, 0, nLines-1);
end;

procedure TLineForm.FrequencySortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByFrequency, 0, nLines-1);
end;

procedure TLineForm.LowerStateSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByLowerState, 0, nLines-1);
end;

procedure TLineForm.WidthSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByWidth, 0, nLines-1);
end;

procedure TLineForm.NameSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByName, 0, nLines-1);
end;

procedure TLineForm.NearestButtonClick(Sender: TObject);
var
  Problem: Exception;
begin
  GlobalMixture.CheckProperties;
  if not Assigned(MyAutoFitter) then
    MyAutoFitter := TAutoFitter.Create(nil);
  with MyAutoFitter do begin
    Mixture := GlobalMixture;
    MaxBlends := -1;
    aW := Abs(StrToFloatEx(AcceptEdit.Text));
    { Sorted copy of active linelist (in correct units) }
    SetLineList(ActiveLinelist);
    LoadObservations('', obLine, tkAll, True);
    SetObservations(Obs.Observations);
    ComputeSearchRange;
    try
      BeginCalculation(True);
      Problem := nil;
      Search(0, 1, Problem);
      if Problem <> nil then
        raise Problem;
    finally
      ClearObservations;
      EndCalculation;
    end;
  end;
  LoadAutoFitAssignment(MyAutoFitter, 0);
end;

procedure TLineForm.WriteExtraVariables(AFormSettings: TFormSettings);
var
  Index, i, P, l: Integer;
  Strings: TStringList;
  Line: TLineW;
  Buffer, Item: string;
  QuantumNumberFormat: TQuantumNumberFormat;
  Qno: TQuantumNumber;
  R: TGridRect;
begin
  inherited;
  HandleComponents(LineFrmOptDlg, {$IFDEF FPC}@{$ENDIF}SetValue, AFormSettings);
  UnLoadOptionsDialog;
  Index := AFormSettings.IndexOfSetting('Lines');
  if Index > 0 then begin
    ClearButtonClick(nil);
    Strings := TStringList.Create;
    Strings.Text := AFormSettings.SettingToStr(Index);
    { Determine quantum number format used for writing }
    QuantumNumberFormat := qftJ;
    Index := AFormSettings.IndexOfSetting('QuantumNumberFormat');
    if Index > 0 then begin
      Buffer := AFormSettings.SettingToStr(Index);
      for Index := 0 to High(QuantumNumberFormatNames) - 1 do
        if CompareText(Buffer, QuantumNumberFormatNames[Index]) = 0 then
          QuantumNumberFormat := TQuantumNumberFormat(Index);
    end;
    Line.Line.EigenvaluesU := nil;
    Line.Line.EigenvaluesL := nil;
    Line.Line.ColourAndType := clDefault;
    SetNan(Line.IObs);
    for l := 0 to Strings.Count-1 do begin
      Buffer := Strings[l];
      if Buffer <> '' then begin
        for i := 1 to 16 do begin
          P := Pos(#9, Buffer);
          if P > 0 then begin
            Item := Copy(Buffer, 1, P-1);
            Buffer := Copy(Buffer, P+1, Length(Buffer));
          end else
            Item := Buffer;
          if (i in [3,7]) and TryStrToQuantumNumber(Item, Qno, QuantumNumberFormat) then
            { Correct quantum numbers }
            Item := IntToStr(Qno);
          SetLineByIndex(Line, Item, i);
        end;
        if nLines >= Length(Lines) then
          SetLength(Lines, nLines+50);
        Lines[nLines] := Line;
        Inc(nLines);
      end;
    end;
    Strings.Free;
  end;
  UpdateGridLines(0,nLines-1);
  Index := AFormSettings.IndexOfSetting('FirstLine');
  if Index > 0 then
    R.Top := AFormSettings.Settings[Index]
  else
    R.Top := -1;
  Index := AFormSettings.IndexOfSetting('LastLine');
  if Index > 0 then
    R.Bottom := AFormSettings.Settings[Index]
  else
    R.Bottom := -1;
  if R.Bottom >= 0 then begin
    if R.Top < 0 then
      R.Top := 1;
    R.Left := 1;
    R.Right := Grid.ColCount-1;
    Grid.Selection := R;
  end;
end;

procedure TLineForm.WriteSettings(AFormSettings: TFormSettings; DoPositions: Boolean; DoSettings: Boolean);
begin
  LoadOptionsDialog;
  inherited;
end;

procedure TLineForm.WeightSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByWeight, 0, nLines-1);
end;

procedure TLineForm.MoreButtonClick(Sender: TObject);
begin
  SortPopupMenu.PopupComponent := MoreButton;
  SortPopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TLineForm.MoveDownButtonClick(Sender: TObject);
var
  i: Integer;
  L: TLineW;
  R: TGridRect;
begin
  R := Grid.Selection;
  if R.Bottom >= nLines then
    Exit;
  L := Lines[R.Bottom];
  for i := R.Bottom downto R.Top do
    Lines[i] := Lines[i-1];
  Lines[R.Top-1] := L;
  Inc(R.Top);
  Inc(R.Bottom);
  UpdateGridLines(0,nLines-1);
  Grid.Selection := R;
end;

procedure TLineForm.MoveTopButtonClick(Sender: TObject);
var
  R: TGridRect;
  L: array of TLineW;
  i, n: Integer;
begin
  R := Grid.Selection;
  if R.Top < 2 then
    Exit;
  Dec(R.Top);
  Dec(R.Bottom);
  n := R.Bottom-R.top+1;
  SetLength(L, n);
  for i := R.Top to R.Bottom do
    L[i-R.Top] := Lines[i];
  for i := R.Top-1 downto 0 do
    Lines[i+n] := Lines[i];
  for i := 0 to n - 1 do
    Lines[i] := L[i];
  UpdateGridLines(0,R.Bottom);
  R.Top := 1;
  R.Bottom := n;
  Grid.Selection := R;
end;

procedure TLineForm.MoveUpButtonClick(Sender: TObject);
var
  i: Integer;
  L: TLineW;
  R: TGridRect;
begin
  R := Grid.Selection;
  if R.Top < 2 then
    Exit;
  L := Lines[R.Top-2];
  Dec(R.Top);
  Dec(R.Bottom);
  for i := R.Top to R.Bottom do
    Lines[i-1] := Lines[i];
  Lines[R.Bottom] := L;
  UpdateGridLines(0,nLines-1);
  Grid.Selection := R;
end;

procedure TLineForm.AltFitButtonClick(Sender: TObject);
begin
  FitPopupMenu.PopupComponent := AltFitButton;
  FitPopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TLineForm.FormResize(Sender: TObject);
begin
  SizeHelpers.WrapPanel(Panel, PanelControls, -1, -1, -1, 1);
  if PanelEx.Visible then
    SizeHelpers.WrapPanel(PanelEx, PanelControls, -1, -1, -1, 1)
  else begin
    HasMeasuredIntensity := False;
    Grid.ColCount := 16;
  end;
  Grid.AutoDefaultRowHeight;
  Constraints.MinWidth := UndoFitButton.Left + UndoFitButton.Width + 2;
end;

function TLineForm.FormSettingsName: string;
begin
  Result := 'LineWindow'; 
end;

function TLineForm.FindBranch(ABranch: string; nSearch: Integer): Integer;
var
  ThisBranch: string;
begin
  ABranch := Trim(ABranch);
  Result := 0;
  while Result < nSearch do begin
    if StringRep(Result+1,14,ThisBranch) and (ABranch = Trim(ThisBranch)) then
      Exit;
    Inc(Result);
  end;
  Result := -1;
end;

procedure TLineForm.FinishDiscard;
var
  i, j: Integer;
begin
  i := 0;
  j := 0;
  while i < nLines do begin
    if not IsNanOrInf(Lines[i].Line.Frequency) then begin
      Lines[j] := Lines[i];
      Inc(j);
    end;
    Inc(i);
  end;
  if j <> nLines then
    FreeAndNil(MyAutoFitter);
  nLines := j;
  UpdateGridLines(0, nLines-1);
end;

procedure TLineForm.FitButtonClick(Sender: TObject);
var
  ObsType: TObsType;
  i: Integer;
  CheckDeriv: Double;
begin
  if not NotEmpty then Exit;
  MainForm.LastObsWasLineForm := True;
  MainForm.ClearAllPeaksItemClick(nil);
  CheckDeriv := 0;
  if Sender <> TestFitButton then
    ReShow(LogForm);
  ReShow(ResidualForm);
  FCurrentLineListType := llStandard;
  ObsType := obLine;
  case (Sender as TComponent).Tag of
    1: ObsType := obCDUpper;
    2: ObsType := obCDLower;
    3: FCurrentLineListType := llBranch;
    4: FCurrentLineListType := llSimple;
  end;
  if (Sender <> TestFitButton) and CheckDerivativesItem.Checked and AskForDerivativesStepScale then
    CheckDeriv := StrToFloatEx(DerivativesStepScale);
  for i := 0 to nLines-1 do
    if Lines[i].Weight > 0 then
      Lines[i].Line.ColourAndType := clNone;
  MainForm.DoLineFit({$IFDEF FPC}@{$ENDIF}LoadObservations,'',ObsType,Sender = TestFitButton, tkObs, CheckDeriv);
  for i := 0 to High(Observations) do
    if Observations[i].FileName = LineListFileName then
      Lines[Observations[i].LineNumber-1].Line.ColourAndType := FindObservationColour(Observations[i], True);
  Grid.Invalidate;
end;

procedure TLineForm.ClearTLines;
var
  i: Integer;
begin
  for i := 0 to High(Lines) do
    with Lines[i] do begin
      Line.EigenvaluesU := nil;
      Line.EigenvaluesL := nil;
    end;
end;

procedure TLineForm.HandleComponent(Component: TComponent; var PName, Value,
  PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean);
begin
  inherited;
  if (Component = LineFrmOptDlg.MinIntensity)
     or (Component = LineFrmOptDlg.MinAssignedIntensity) then
    SettingType := stDouble;
end;

function TLineForm.HasObservations: Boolean;
var
  i: Integer;
begin
  Result := False;
  if not NotEmpty then Exit;
  for i := 0 to Grid.RowCount-2 do
    if not (IsNan(Lines[i].Weight)) and (Lines[i].Weight > 0) then begin
      Result := True;
      Exit;
    end;
end;

procedure TLineForm.HighlightRow(ARow: Integer);
var
  R: TGridRect;
begin
  BringUp;
  Grid.Row := ARow;
  R.Left := Grid.ColCount-1;
  R.Right := 1;
  R.Top := ARow;
  R.Bottom := ARow;
  Grid.Selection := R;
  Grid.SetFocus;
end;

procedure TLineForm.LoadAutoFitAssignment(AutoFitter: TAutoFitter; iFit: Integer { -1 implies reset } );
var
  Lmin, Lmax, i: Integer;
  W, Intensity: Double;
begin
  Lmin := High(Integer);
  Lmax := 0;
  for i := 0 to High(AutoFitter.FitIndices[0]) do
    with AutoFitter.Obs0[i] do
      if (Weight <= 0) and (FileName = LineListFileName) and (LineNumber >= 0) then begin
        Lmin := Min(Lmin, LineNumber);
        Lmax := Max(Lmax, LineNumber);
        if iFit < 0 then
          LineForm.UpdateAssignment(LineNumber, Frequency, 0, '')
        else if AutoFitter.FitIndices[iFit][i] >= 0 then begin
          if Length(AutoFitter.Weights) > 0 then
            W := AutoFitter.Weights[AutoFitter.FitIndices[iFit][i]]
          else
            W := 1;
          if Length(AutoFitter.Intensity) > 0 then
            Intensity := AutoFitter.Intensity[AutoFitter.FitIndices[iFit][i]]
          else
            Intensity := 0;
          UpdateAssignment(LineNumber, AutoFitter.f[AutoFitter.FitIndices[iFit][i]], W, AutoFitter.SourceName, Intensity) ;
        end else
          UpdateAssignment(LineNumber, Frequency, -1, '');
      end;
  UpdateGridLines(Lmin-1, Lmax-1);
  TestFitButton.Click;
end;

procedure TLineForm.LoadObservations(const Source: string; ObsType: TObsType; TakeWeight: TTakeWeight; Selected: Boolean);
var
  s: TTextStream;
  i: Integer;
begin
  s := TTextStream.Create(nil,False);
  try
    if Selected then
      CopyToStream(s, Grid.Selection.Top, Grid.Selection.Bottom, True, FCurrentLineListType, True, False)
    else
      CopyToStream(s, 1, Grid.RowCount-1, True, FCurrentLineListType, True, False);
    s.Position := 0;
    Obs.LoadObservations(GlobalMixture, s, LineListFileName, ObsType, TakeWeight);
    if Selected then
      for i := 0 to High(Observations) do
        with Observations[i] do
          Inc(LineNumber, Grid.Selection.Top-1);
  finally
    s.Free;
  end;
end;

procedure TLineForm.LoadOptionsDialog;
begin
  with LineFrmOptDlg do begin
    ShowLines.Value := FnLinesToShow;
    MinIntensity.Text := FloatToStr(FMinIntensity);
    DiscardZeroIntensity.Checked := not FShowZeroIntensityLines;
    MinAssignedIntensity.Text := FloatToStr(FMinAssignedIntensity);
  end;
end;

procedure TLineForm.UndoFitButtonClick(Sender: TObject);
begin
  if GlobalMixture <> nil then
    GlobalMixture.UndoFit;
end;

procedure TLineForm.FormShow(Sender: TObject);
begin
  PanelEx.Visible := AdvancedItem.Checked;
end;

procedure TLineForm.UnLoadOptionsDialog;
begin
  with LineFrmOptDlg do begin
    if not TryStrToFloatEx(MinIntensity.Text, FMinIntensity) then
      raise EReportException.Create('Bad number:'+MinIntensity.Text)
    else if not TryStrToFloatEx(MinAssignedIntensity.Text, FMinAssignedIntensity) then
      raise EReportException.Create('Bad number:'+MinAssignedIntensity.Text)
    else begin
      FnLinesToShow := ShowLines.Value;
      FShowZeroIntensityLines := not DiscardZeroIntensity.Checked;
    end;
  end;
end;

procedure TLineForm.UpdateActions;
begin
  inherited UpdateActions;
  if FLongestNameLength > Grid.ColWidths[LabelCol] then
    Grid.ColWidths[LabelCol] := FLongestNameLength;
  FLongestNameLength := 0;
  UndoFitButton.Enabled := (GlobalMixture <> nil) and GlobalMixture.CanUndoFit
end;

procedure TLineForm.UpdateAssignment(ARow: Integer; AFrequency, AWeight: Double; SourceName: string; Intensity: Double);
var
  Branch, s: string;
  p: Integer;
begin
  SplitCachedName(Lines[ARow-1].Line, Branch, s);
  p := Pos(':', s);
  if p > 0 then begin
    if (p > 1) and (s[p-1] = ' ') then
      Dec(p); { Trim trailing space }
    Delete(s,p,Length(s));
  end;
  if SourceName <> '' then
    s := s + ' : ' + SourceName;
  if Branch <> '' then
    s := Branch + ' : ' + s;
      Lines[ARow-1].Line.CachedName := s;
  Lines[ARow-1].Line.Frequency := AFrequency;
  if AWeight <= 0 then
    SetNan(AWeight);
  Lines[ARow-1].Weight := AWeight;
  if Intensity <= 0 then
    SetNan(Intensity);
  Lines[ARow-1].IObs := Intensity;
end;

procedure TLineForm.UpdateButtonClick(Sender: TObject);
var
  i: Integer;
  Obs: TObservation;
begin
  for i := Grid.Selection.Top to Grid.Selection.Bottom do
    if not Grid.Valid[i,11] or (Grid.Value[i,11] <= 0) then
      MakeObs(Obs, i, True);
  UpdateGridLines(Grid.Selection.Top-1, Grid.Selection.Bottom-1);
end;

procedure TLineForm.GridGetCellLabel(Sender: TObject; Col, Row: Integer;
  var s: string; var IsText: Boolean);
begin
  if (Row > 0) and (Col > 0) then
    IsText := StringRep(Row, Col, s);
end;

end.
