unit LmatrixBaseFrm;
{$I utils.inc}{Copyright (C) 2014 C M Western; see commandline.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     SysUtils, Variants, Classes, Graphics, Controls, Forms,
     Dialogs, VibPlotFrm, Menus, StdCtrls, NewSpin,
     ExtCtrls, GraphI, GraphM, Split, Vib, MathSup, ComCtrls, AlEdit, EditExtr,
  ScrlCtrl, Graphs;

type

  TDisplacementSelect = (dsZeroPoint=0, dslMatrix=1, dsADM=2);

  TLmatrixBaseForm = class(TVibPlotForm)
    ShowTRCheckBox: TCheckBox;
    DispSelect: TComboBox;
  public
   procedure PlotAtoms(State:TElectronicState; { PlotPlane, } Mode: Integer;
                       FindLimits, PlotAxes, PlotLastPoints, AvoidDoubleLabels, { ShowTR, } InPrincipal: Boolean;
                       { DisplacementSelect: TDisplacementSelect;
                       Magnification: Double; } nMark: Integer; var LastPoints, Labels: TDPointArray;
                       var xMin, xMax, yMin, yMax: Double; var LimitsSet: Boolean);
  end;

function MassWeight(State: TElectronicState; DisplacementSelect: TDisplacementSelect; nNucleus, nMode: Integer): Double;

implementation

uses Math, MatrixVec, GStyle, Gmarks, InternalCoords, zMatrix, Units;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function MassWeight(State: TElectronicState; DisplacementSelect: TDisplacementSelect; nNucleus, nMode: Integer): Double;
begin
  case DisplacementSelect of
    dsADM:       Result := 1;
    dslMatrix:   Result := Sqrt((State.Nucleus(nNucleus) as TMolecularNucleus).Mass);
    dsZeroPoint: if nMode >= 0 then
                   // Was: Result := 1/Sqrt(2*Pi*ReduceFrequency*State.Mode(nMode).Omega) ( Differs by Sqrt(2*pi) }
                   Result := 1/Sqrt(gammaConversion*Sqrt(lambdaConversion)*State.Mode(nMode).Omega)
                 else
                   Result := 1;
  end;
end;

{ TLmatrixBaseForm }

procedure TLmatrixBaseForm.PlotAtoms(State: TElectronicState; Mode: Integer;
  FindLimits, PlotAxes, PlotLastPoints, AvoidDoubleLabels, InPrincipal: Boolean;
  nMark: Integer; var LastPoints, Labels: TDPointArray; var xMin, xMax, yMin,
  yMax: Double; var LimitsSet: Boolean);
var
  nLabel: Integer;
  chDist: Double;
  DisplacementSelect: TDisplacementSelect;
  procedure AddLabel(x, y: Double; Lab: string);
  begin
    Inc(nLabel);
    if not AvoidDoubleLabels then begin
      { Keep track of label positions }
      if High(Labels) <= nLabel then
        SetLength(Labels, nLabel+10);
      Labels[nLabel].X := x;
      Labels[nLabel].Y := y;
    end else
      if (nLabel < Length(Labels))
         and (Sqr(Labels[nLabel].X-x)+Sqr(Labels[nLabel].Y-y) < chDist) then
        Exit;
    G.Text(x,y,Lab);
  end;
  procedure Limit(x,y:Double);
  begin
    if not FindLimits then
      Exit;
    if not LimitsSet then begin
//      xMin := x;
//      xMax := x;
//      yMin := y;
//      yMax := y;
      xMin := -0.5; xMax := -xMin;
      yMin := xMin; yMax := -yMin;
      LimitsSet := True;
    end;
    if not IsNanOrInf(x) then begin
      xMin := Min(xMin,x);
      xMax := Max(xMax,x);
    end;
    if not IsNanOrInf(y) then begin
      yMin := Min(yMin,y);
      yMax := Max(yMax,y);
    end;
  end;
var
  i, j, ix, iy, ia, M, M1, M2: Integer;
  Gx, Gy, Dx, Dy: Double;
  cx, cy, Q: Double;
  P: TDPoint;
  xyz, dxyz: Txyz;
begin
  PrepRotation;
  case PlotPlaneGroup.ItemIndex of
    0: begin ix := 1; iy := 2; end;
    1: begin ix := 2; iy := 3; end;
    2: begin ix := 1; iy := 3; end;
  end;
  DisplacementSelect := TDisplacementSelect(DispSelect.ItemIndex);
  if PlotAxes then begin
    G.MapLin(xMin-(Xmax-xMin)/10,xMax+(xMax-xMin)/10,yMin-(yMax-yMin)/10,yMax+(yMax-yMin)/10);
    with State.Nucleus(0) do
      G.Scales(ParameterNames[IndexOfParameter('x')+ix-1],ParameterNames[IndexOfParameter('x')+iy-1]);
  end;
  nLabel := -1;
  G.Chlen(cx,cy);
  chDist := Sqr(cx)+Sqr(cy);
  if FindLimits then
    State.PrepareMatrices(False,True)
  else if not InPrincipal then begin
    if nMark = 4 then
      G.LineStyle.Pattern := lpDot;
    ShowPrincipalAxes(State.zMatrix.CofM[ix], State.zMatrix.CofM[iy], ix, iy, State.zMatrix.Itrans);
    G.LineStyle.Pattern := lpSolid;
  end else begin
    G.MapArg(Dx, Gx, Dy, Gy);
    G.ClipPoint(Dx,0);
    G.ClipJoin(Gx,0);
    G.ClipPoint(0,Dy);
    G.ClipJoin(0,Gy);
  end;
  if not PlotLastPoints then
    SetLength(LastPoints, State.Molecule.nNuclei);
  ia := -1; { Separate index needed to skip dummy atoms }
  for i := 0 to State.Molecule.nNuclei-1 do begin
    State.zMatrix.NextRealAtom(ia);
    if InPrincipal then
      xyz := State.zMatrix.Atoms[ia].pxyz
    else
      xyz := State.zMatrix.Atoms[ia].xyz;
    Rotate(xyz, ix);
    Gx := xyz[ix];
    Gy := xyz[iy];
    Limit(Gx,Gy);
    M1 := 0;
    M2 := Mode;
    if FindLimits then
      M2 := State.adm.ColCount-1
    else if Mode >= 0 then
      M1 := Mode;
    for M := M1 to M2 do begin
      if DisplacementSelect = dsZeroPoint then
        if (M < State.Molecule.nModes) and (Abs(State.Mode(M).Omega) > 1e-5) then begin
          // { Force constant in cm-1 Angstrom^-2 amu }
          // keff := 2*Pi*ReduceFrequency*Sqr(State.Mode(M).Omega);
          // { Value of normal co-ordinate at zero point energy - not}
          // Was Q := Sqrt(State.Mode(M).Omega/keff);
          Q := MassWeight(State, DisplacementSelect, 0 {Not used }, M)
        end else
          { Keep sensible scale for translations and rotationa }
          Q := 0.1 // Was: Q := 1/400/ReduceFrequency;
      else
        Q := 0.1;
      if DisplacementSelect = dslMatrix then
        Q := Q*Sqrt((State.Nucleus(i) as TMolecularNucleus).Mass);
      Q := Q*MagnificationSpin.Value;
      if InPrincipal then begin
        Q := Q/Sqrt((State.Nucleus(i) as TMolecularNucleus).Mass);
        for j := 1 to 3 do
          dxyz[j] := State.lMatrix[3*i+j,M+1]*Q;
      end else
        for j := 1 to 3 do
          dxyz[j] := State.adm[3*i+j,M+1]*Q;
      Rotate(dxyz, ix);
      Dx := Gx + dxyz[ix];
      Dy := Gy + dxyz[iy];
      Limit(Dx,Dy);
    end;
    if not FindLimits then begin
      G.Symbol(Gx,Gy,nMark);
      AddLabel(Gx+cx,Gy-cy,State.Nucleus(i).Name);
      if Mode >= 0 then begin
        G.Arrow(Gx,Gy,Dx,Dy,arHead);
        P.X := Dx;
        P.Y := Dy;
      end else begin
        P.X := Gx;
        P.Y := Gy;
      end;
      if PlotLastPoints then begin
        G.LineStyle.Pattern := lpDot;
        with LastPoints[i] do
          G.Point(X, Y);
        G.Join(P.X, P.Y);
        G.LineStyle.Pattern := lpSolid;
      end else
        LastPoints[i] := P;
    end;
  end;
end;

end.
