unit LoomisWoodFrm;
{$I utils.inc}{Copyright (C) 14 C M Western; see commandline.pas for license}

interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, PeakFitFrm, Menus, Grids, NGrid, StdCtrls, ComCtrls, AlEdit, EditExtr,
  NewSpin, Buttons, ExtCtrls, MNode, Units, Peaks, LoomisWood;

type

  { TLoomisWoodForm }

  TLoomisWoodForm = class(TPeakFitForm)
    SortItem: TMenuItem;
    MoveUpButton: TSpeedButton;
    MoveDownButton: TSpeedButton;
    mOffset: TSpinBox;
    mOffsetLabel: TLabel;
    ActiveCheckBox: TCheckBox;
    CopyforBranchExprItem: TMenuItem;
    CentreButton: TSpeedButton;
    nPlotsSpin: TSpinBox;
    Label4: TLabel;
    procedure FitButtonClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure PolyOrderChange(Sender: TObject);
    procedure NLfitButtonClick(Sender: TObject);
    procedure SplineFitItemClick(Sender: TObject);
    procedure SortItemClick(Sender: TObject);
    procedure MoveUpButtonClick(Sender: TObject);
    procedure mOffsetChange(Sender: TObject);
    procedure ActiveCheckBoxClick(Sender: TObject);
    procedure GridSetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
    procedure ResetItemClick(Sender: TObject);
    procedure CopyforBranchExprItemClick(Sender: TObject);
    procedure CentreButtonClick(Sender: TObject);
    procedure nPlotsSpinChange(Sender: TObject);
  private
    LoomisWood: TLoomisWood;
  protected
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
  public
    destructor Destroy; override;
    function BaseNodeClass: TClass; override;
    function Peaks: TPeaks; override;
    function HasDataSource: Boolean; override;
    procedure PeakMeasured(Source: TMNode; Index: Integer); override;
    procedure ShowItem(MNode: TMNode); override;
    procedure ShowPeak(AIndex: Integer); override;
    function DefaultFunctionText(Order: Integer):string;
    procedure LoadExpressionAndVariables;
    procedure UnLoadExpressionAndVariables;
  end;

var
  LoomisWoodForm: TLoomisWoodForm;

procedure ConnectLoomisWoodForm(LoomisWood: TLoomisWood);
function ActiveLoomisWood: TLoomisWood;

implementation

uses Math, MathSup, Clipbrd, Huge, TextAssign, Peak, ChildForm, Overlay, Global,
     Main, Gendlg, OvrlyFrm, UCompat;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

procedure ConnectLoomisWoodForm(LoomisWood: TLoomisWood);
var
  Stuff: Boolean;
begin
  Stuff := not LoomisWoodForm.Visible or (LoomisWoodForm.CurrentMNode <> LoomisWood);
  ReShow(LoomisWoodForm);
  if Stuff then begin
    LoomisWoodForm.BaseNode := LoomisWood.Owner as TMNode;
    LoomisWoodForm.ShowItem(LoomisWood);
    MainForm.ConsiderReplot(False, False);
  end;
end;

function ActiveLoomisWood: TLoomisWood;
begin
  Result := nil;
  if Assigned(LoomisWoodForm) and LoomisWoodForm.Visible then
    Result := LoomisWoodForm.CurrentMNode as TLoomisWood;
end;

{ TLoomisWoodForm }

procedure TLoomisWoodForm.ActiveCheckBoxClick(Sender: TObject);
begin
  if LoomisWood.Active = ActiveCheckBox.Checked then
    Exit;
  LoomisWood.Active := ActiveCheckBox.Checked;
  if (not ActiveCheckBox.Checked) and (LoomisWood.Owner is TOverlays) then
    MainForm.SinglePlotItemClick(nil);
  MainForm.ConsiderReplot(False, False);
end;

function TLoomisWoodForm.BaseNodeClass: TClass;
begin
  Result := TPlot;
end;

procedure TLoomisWoodForm.CentreButtonClick(Sender: TObject);
var
  i, m: Integer;
  NewCentre: Double;
begin
  if Lockout > 0 then
    Exit;
  m := Peaks.nPeaks;
  if Peaks.nPeaks > 0 then begin
    { Find nearest m value to mOffset }
    with Peaks do begin
      m := 0;
      while not Valid(m) and (m < nPeaks) do
        Inc(m);
      for i := m+1 to nPeaks - 1 do
        if Valid(i) and (Abs(Actual[i]-LoomisWood.mOffset) < Abs(Actual[m]-LoomisWood.mOffset)) then
          m := i;
    end;
  end;
  if m < Peaks.nPeaks then begin
    with LoomisWood do
      NewCentre := Peaks[m].Position - xShift(Peaks.Actual[m]-mOffset) + xShift(0);
    MainForm.CentrePlot(NewCentre, MainForm.Simulation.PlotUnits);
  end else
    MainForm.ConsiderReplot(False, False);
end;

procedure TLoomisWoodForm.CopyforBranchExprItemClick(Sender: TObject);
var
  S: THugeMemoryStream;
  i: Integer;
  F: TextFile;
  Lab: string;
begin
  Lab := ExtractFileName((LoomisWood.Owner as TMNode).Name);
  if LoomisWood.Owner is TPlot then begin
    Lab := ExtractFileName((LoomisWood.Owner as TMNode).Name);
    if Lab = '1' then
      Lab := ''
    else
      Lab := ' - '+Lab;
    Lab := ExtractFileName((LoomisWood.Owner.Owner as TMNode).Name)+Lab;
  end;
  S := THugeMemoryStream.Create;
  AssignStream(F, S);
  Append(F);
  try
    for i := Grid.Selection.Top-1 to Grid.Selection.Bottom-1 do
      with Peaks do
        if Valid(i) then
          with Peaks[i] do
            WriteLn(F, FloatToStr(Actual):3, '  ', FloatToStr(Position):15, ' ', FloatToStr(Weight):3, ' : ', Lab);
    System.Close(F);
    S.CopyToClipboard(CF_PREFTEXT);
  finally
    S.Free;
  end;
end;

function TLoomisWoodForm.DefaultFunctionText(Order: Integer): string;
begin
  Result := LoomisWood.DefaultFunctionText(Order);
end;

destructor TLoomisWoodForm.Destroy;
begin
  if LoomisWoodForm = Self then
    LoomisWoodForm := nil;
  inherited;
end;

procedure TLoomisWoodForm.FitButtonClick(Sender: TObject);
begin
  if PolyOrder.Value > 0 then
    PolyOrderChange(nil);
  LoadExpressionAndVariables;
  DataSource.NLFitDlg.FitOnceClick(nil);
  MainForm.ResidualsItemClick(nil);
end;

procedure TLoomisWoodForm.FormResize(Sender: TObject);
begin
  Constraints.MinWidth := Max(OtherButton.Left + OtherButton.Width, ActiveCheckBox.Left + ActiveCheckBox.Width) + 2;
  Panel.Height := nPlotsSpin.Top + nPlotsSpin.Height + 2;
end;

procedure TLoomisWoodForm.GridSetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
var
  P: TPeak;
  function Update(var OldValue: Double; NewValue: Double): Boolean;
  begin
    Result := (IsNanOrInf(OldValue) xor IsNanOrInf(NewValue))
              or (OldValue <> NewValue);
    if Result then
      OldValue := NewValue;
  end;
begin
  if not Valid then
    SetNan(v);
  if (Row >= Peaks.nPeaks) and not IsNanOrInf(v) then
    Peaks.nPeaks := Row;
  P := Peaks.Peaks[Row-1];
  case Col of
    2: Update(P.Position,v);
    1: Update(P.Actual,v);
    4: Update(P.Weight,v);
  end;
  try
    FChanging := True;
    Peaks.Peaks[Row-1] := P;
  finally
    FChanging := False;
  end;
end;

function TLoomisWoodForm.HasDataSource: Boolean;
begin
  Result := Assigned(LoomisWood) and LoomisWood.HasDataSource;
  if Result then
    DataSource := LoomisWood.DataSource as TInteractiveDataSource;
end;

procedure TLoomisWoodForm.LoadExpressionAndVariables;
var
  i, j: Integer;
begin
  DataSource := LoomisWood.DataSource as TInteractiveDataSource;
  DataSource.DataSourceFitter.Expression := LoomisWood.Comment;
  LoomisWood.Floating[LoomisWood.IndexOfParameter('mLowest')] := False;
  for i := LoomisWood.FirstDynamicParameter to LoomisWood.nParameters do begin
    j := DataSource.DataSourceFitter.Variables.IndexOf(LoomisWood.ParameterNames[i]);
    if j >= 0 then begin
      DataSource.DataSourceFitter.Floating[j] := LoomisWood.Floating[i];
      if LoomisWood.StdDev[i] > 0 then
        DataSource.DataSourceFitter.LastError[j] := LoomisWood.StdDev[i]
      else
        SetNaN(DataSource.DataSourceFitter.LastError[j]);
      DataSource.DataSourceFitter.Variables.Objects[j].Value := LoomisWood.Parameters[i];
    end;
  end;
  DataSource.NLFitDlg.LoadFromFitter;
end;

procedure TLoomisWoodForm.mOffsetChange(Sender: TObject);
begin
  LoomisWood.mOffset := mOffset.Value;
  CentreButtonClick(nil);
end;

procedure TLoomisWoodForm.MoveUpButtonClick(Sender: TObject);
var
  i: Integer;
  mShift: Integer;
begin
  LoomisWood.CheckProperties;
  mShift := (Sender as TComponent).Tag;
  for i := 0 to Peaks.nPeaks - 1 do
    Peaks.Actual[i] := Peaks[i].Actual + mShift;
  MainForm.ConsiderReplot(False, False);
end;

procedure TLoomisWoodForm.NLfitButtonClick(Sender: TObject);
begin
  LoadExpressionAndVariables;
  ReShow(DataSource.NLFitDlg);
end;

procedure TLoomisWoodForm.nPlotsSpinChange(Sender: TObject);
begin
  if LoomisWood = nil then
    Exit;
  LoomisWood.nPlots := nPlotsSpin.Value;
  if Lockout <= 0 then
    MainForm.ConsiderReplot(False, False);
end;

procedure TLoomisWoodForm.UnLoadExpressionAndVariables;
var
  i, j: Integer;
begin
  LoomisWood.Comment := DataSource.DataSourceFitter.Expression;
  LoomisWood.CheckVariables.Free;
  for i := 0 to DataSource.DataSourceFitter.Variables.Count-1 do begin
    j := LoomisWood.IndexOfParameter(DataSource.DataSourceFitter.Variables[i]);
    if j >= 0 then begin
      LoomisWood.Parameters[j] := DataSource.DataSourceFitter.Variables.Objects[i].Value;
      LoomisWood.Floating[j] := DataSource.DataSourceFitter.Floating[i];
      LoomisWood.StdDev[j] := DataSource.DataSourceFitter.LastError[i];
    end;
  end;
end;

procedure TLoomisWoodForm.PeakMeasured(Source: TMNode; Index: Integer);
var
  Peak: TPeak;
  i: Integer;
  ForThis: Boolean;
begin
  if CurrentMNode.Owner is TOverlays then
    ForThis := Source.Owner = CurrentMNode.Owner
  else
    ForThis := Source = CurrentMNode.Owner;
  if ForThis then begin
    { Move new peak into Loomis Wood object }
    with Source as TPlot do begin
      Peak := Peaks[nPeaks-1];
      Peaks.DeletePeak(nPeaks-1);
    end;
    Peak.Actual := Index div 1000 + LoomisWood.mOffset;
    for i := 0 to Peaks.nPeaks - 1 do
      if Round(Peaks[i].Actual) = Peak.Actual then begin
        Peaks[i] := Peak;
        Exit;
      end;
    Peaks.AddPeak(Peak);
  end;
end;

function TLoomisWoodForm.Peaks: TPeaks;
begin
  if CurrentMNode is TLoomisWood then
    Result := LoomisWood.Peaks
  else
    Result := DummyPeaks;
end;

procedure TLoomisWoodForm.PolyOrderChange(Sender: TObject);
var
  i, ii: Integer;
begin
  if Visible and (PolyOrder.Value > 0) then
    with LoomisWood do begin
      Comment := DefaultFunctionText(PolyOrder.Value);
      CheckVariables.Free;
      { Make sure all coefficients in polynomial are floating }
      with Expr.Lex.Functions do
        for i := 0 to Count-1 do begin
          ii := IndexOfParameter(Strings[i]);
          if ii > 1 then
            Floating[ii] := True;
        end;
      if HasDataSource then
        LoadExpressionAndVariables;
    end;
end;

procedure TLoomisWoodForm.ResetItemClick(Sender: TObject);
var
  i: Integer;
begin
  BeginGlobalUpdate;
  try
    PolyOrder.Value := 1;
    with LoomisWood do
      for i := FirstDynamicParameter to nParameters do
        Parameters[i] := 0;
    mOffset.Value := 0;
  finally
    EndGlobalUpdate(False);
  end;
  MainForm.ConsiderReplot(False, False);
end;

procedure TLoomisWoodForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  if MNode = LoomisWood then begin
    mOffset.Value := LoomisWood.mOffset;
    ActiveCheckBox.Checked := LoomisWood.Active;
    nPlotsSpin.Value := LoomisWood.nPlots;
  end;
end;

procedure TLoomisWoodForm.ShowItem(MNode: TMNode);
var
  i, n: Integer;
begin
  { Update Node display }
  Grid.EditorMode := False;
  SetItemIndex(ActiveObject,MNode);
  CurrentMNode := MNode;
  LoomisWood := MNode as TLoomisWood;
  UpdateStatus;
  Caption := 'Loomis Wood';
  n := 0;
  if Assigned(LoomisWood) then begin
    if LoomisWood.Owner is TPlot then
      Caption := TPlot(LoomisWood.Owner).Name + ' - ' + Caption;
    n := Peaks.nPeaks;
    SettingChanged(LoomisWood, -1);
  end else
    Close;
  if n = 0 then begin
    Grid.RowCount := 2;
    Grid.Clear;
    Grid.Enabled := False;
  end else begin
    Grid.Enabled := True;
    Grid.RowCount := n+1;
    for i := 0 to n-1 do
      ShowPeak(i);
  end;
end;

procedure TLoomisWoodForm.ShowPeak(AIndex: Integer);
begin
  with Peaks[AIndex] do begin
    Grid[AIndex+1,1] := Actual;
    Grid[AIndex+1,2] := Position;
    Grid[AIndex+1,3] := Residual;
    Grid[AIndex+1,4] := Weight;
  end;
end;

procedure TLoomisWoodForm.SortItemClick(Sender: TObject);
begin
  Peaks.SortPeaks(1, True);
end;

procedure TLoomisWoodForm.SplineFitItemClick(Sender: TObject);
begin
  LoadExpressionAndVariables;
  DataSource.NLFitDlg.SplineItemClick(nil);
end;

function TLoomisWoodForm.StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer;
begin
  if MNode is TLoomisWood then begin
    ActiveObject.AddItem(Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end else
    Result := 0;
end;

end.
