﻿unit Main;
{$I utils.inc}
{
  PGOPHER - a program for simulating rotational structure
  Copyright (C) 2005-2013 C M Western
  School of Chemistry, University of Bristol, Bristol, UK
  For further information please see http://pgopher.chm.bris.ac.uk/
  or contact help-pgopher@bristol.ac.uk

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}

interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, PlotFrm, ExtCtrls, ScrlCtrl, Graphs, ComCtrls, Buttons, ToolWin,
  StdCtrls, PrintPre, ExActn, ActnList, ProjAbout, Menus,
  ToolBarEx, Split, Mixture, MNode, Obs, Overlay, TextStrm, MatrixVec, AlEdit,
  SampleMols, Settings, GraphI, GraphM, Units, Global, Recent;

{$IFNDEF FPC}
{$IF CompilerVersion >= 18.0}
{$DEFINE SHOWMEMUSAGE}
{$ELSE}
{$IFDEF FASTMM4} {$DEFINE SHOWMEMUSAGE} {$ENDIF}
{$IFEND}
{$ENDIF}

type

  { TMainForm }

  TMainForm = class(TPlotForm)
    MainMenu: TMainMenu;
    FileMenu: TMenuItem;
    angVacItem: TMenuItem;
    angAirItem: TMenuItem;
    AddOverlayItem: TMenuItem;
    NewItem: TMenuItem;
    EmptyMixtureItem: TMenuItem;
    NewLinearMoleculeItem: TMenuItem;
    OpenItem: TMenuItem;
    ImportItem: TMenuItem;
    OldPgopherItem: TMenuItem;
    OldD3hItem: TMenuItem;
    CALPGMfileItem: TMenuItem;
    BinaryfileItem: TMenuItem;
    ExportItem: TMenuItem;
    ExportLineListItem: TMenuItem;
    ExportPlotItem: TMenuItem;
    LoadOverlayItem: TMenuItem;
    MergeItem: TMenuItem;
    N6: TMenuItem;
    SaveItem: TMenuItem;
    SaveAsItem: TMenuItem;
    N2: TMenuItem;
    PrintItem: TMenuItem;
    PrintSetupItem: TMenuItem;
    FileExitSeparator: TMenuItem;
    ExitItem: TMenuItem;
    EditMenu: TMenuItem;
    CopyItem: TMenuItem;
    PasteItem: TMenuItem;
    CopyPictureItem: TMenuItem;
    PasteOverlay1Item: TMenuItem;
    PlotMenu: TMenuItem;
    LeftItem: TMenuItem;
    RightItem: TMenuItem;
    CompressItem: TMenuItem;
    ExpandItem: TMenuItem;
    PlotSelectionItem: TMenuItem;
    AllItem: TMenuItem;
    N3: TMenuItem;
    SimulateItem: TMenuItem;
    NextPlotstyleItem: TMenuItem;
    ShowPartsItem: TMenuItem;
    ShowSumItem: TMenuItem;
    FortratItem: TMenuItem;
    ShowFortratItem: TMenuItem;
    UpperStateItem: TMenuItem;
    UnitsItem: TMenuItem;
    WavenumbersItem: TMenuItem;
    MHzItem: TMenuItem;
    KelvinItem: TMenuItem;
    eVItem: TMenuItem;
    nmVacuumItem: TMenuItem;
    nmAirItem: TMenuItem;
    N5: TMenuItem;
    MagnifyPlotItem: TMenuItem;
    ShrinkplotItem: TMenuItem;
    ResetZoomItem: TMenuItem;
    OverlaysMenu: TMenuItem;
    PasteOverlayItem: TMenuItem;
    PasteAnotherItem: TMenuItem;
    SelectRegionItem: TMenuItem;
    ClearSelectionItem: TMenuItem;
    InvertItem: TMenuItem;
    ClearOverlayItem: TMenuItem;
    OverlayPropertiesItem: TMenuItem;
    N1: TMenuItem;
    DeleteLastPeakItem: TMenuItem;
    ClearAllPeaksItem: TMenuItem;
    ViewMenu: TMenuItem;
    ConstantsItem: TMenuItem;
    LineListItem: TMenuItem;
    LogItem: TMenuItem;
    OverlaysItem: TMenuItem;
    StatesItem: TMenuItem;
    TransitionsItem: TMenuItem;
    LevelsItem: TMenuItem;
    SpectrumItem: TMenuItem;
    TimingItem: TMenuItem;
    ClearLogItem: TMenuItem;
    ClearCacheItem: TMenuItem;
    ClearObservedItem: TMenuItem;
    HelpMenu: TMenuItem;
    IndexItem: TMenuItem;
    SearchItem: TMenuItem;
    ProceduresItem: TMenuItem;
    WhatIsItem: TMenuItem;
    UsingHelpItem: TMenuItem;
    AboutItem: TMenuItem;
    DumpItem: TMenuItem;
    OpenDialog: TOpenDialog;
    ImportDialog: TOpenDialog;
    ExportLineListDialog: TSaveDialog;
    OpenOverlayDialog: TOpenDialog;
    Project: TProject;
    ActionList: TActionList;
    ExCopy: TExCopy;
    ExPaste: TExPaste;
    GraphPrintPreview: TGraphPrintPreview;
    SaveDialog: TSaveDialog;
    CalibrateItem: TMenuItem;
    N4: TMenuItem;
    ArrangeFormsItem: TMenuItem;
    CalibrationItem: TMenuItem;
    I2Item: TMenuItem;
    ThreadsItem: TMenuItem;
    AsymmetricTopItem: TMenuItem;
    VelocityItem: TMenuItem;
    ResidualsItem: TMenuItem;
    SaveArrangementItem: TMenuItem;
    SavePositionsItem: TMenuItem;
    FileOptionsItem: TMenuItem;
    SaveSettingsItem: TMenuItem;
    LoadPositionsItem: TMenuItem;
    LoadSettingsItem: TMenuItem;
    SaveOverlayItem: TMenuItem;
    SaveOverlayAsItem: TMenuItem;
    SpecViewfileItem: TMenuItem;
    JB95FileItem: TMenuItem;
    CopyTopPlotItem: TMenuItem;
    TopPlotItem: TMenuItem;
    SymmetricTopItem: TMenuItem;
    HitranFileItem: TMenuItem;
    STROTABFileItem: TMenuItem;
    PlotOptionsItem: TMenuItem;
    FontItem: TMenuItem;
    VariableCharSizeItem: TMenuItem;
    LimitLineSegmentsItem: TMenuItem;
    LmatrixImportITEM: TMenuItem;
    VibrationalSpectrumItem: TMenuItem;
    PromptOnImportItem: TMenuItem;
    OnImportItem: TMenuItem;
    DefaulttoArbitraryUnitsItem: TMenuItem;
    N7: TMenuItem;
    MagnifyFromBottomItem: TMenuItem;
    MagnifyFromTopItem: TMenuItem;
    AutoVerticalScaleItem: TMenuItem;
    AllWindowsOnTaskBarItem: TMenuItem;
    WindowsonTopItem: TMenuItem;
    RecentFiles: TRecentFiles;
    VIBCAImportItem: TMenuItem;
    ExoMolFileItem: TMenuItem;
    SinglePlotItem: TMenuItem;
    CalculatorItem: TMenuItem;
    AnotherItem: TMenuItem;
    UsePeakFitterItem: TMenuItem;
    ColouredItem: TMenuItem;
    LineWidthItem: TMenuItem;
    AutofitItem: TMenuItem;
    ShowLinelistItem: TMenuItem;
    procedure AddOverlayItemClick(Sender: TObject);
    procedure I2ItemClick(Sender: TObject);
    procedure ArrangeFormsItemClick(Sender: TObject);
    procedure CalibrateItemClick(Sender: TObject);
    procedure ShowSumButtonClick(Sender: TObject);
    procedure SaveAsItemClick(Sender: TObject);
    procedure SaveItemClick(Sender: TObject);
    procedure DumpItemClick(Sender: TObject);
    procedure AboutItemClick(Sender: TObject);
    procedure UsingHelpItemClick(Sender: TObject);
    procedure WhatIsItemClick(Sender: TObject);
    procedure ProceduresItemClick(Sender: TObject);
    procedure SearchItemClick(Sender: TObject);
    procedure IndexItemClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure ClearObservedItemClick(Sender: TObject);
    procedure ClearCacheItemClick(Sender: TObject);
    procedure ClearLogItemClick(Sender: TObject);
    procedure TimingItemClick(Sender: TObject);
    procedure SpectrumItemClick(Sender: TObject);
    procedure LevelsItemClick(Sender: TObject);
    procedure TransitionsItemClick(Sender: TObject);
    procedure StatesItemClick(Sender: TObject);
    procedure LogItemClick(Sender: TObject);
    procedure LineListItemClick(Sender: TObject);
    procedure ConstantsItemClick(Sender: TObject);
    procedure ClearAllPeaksItemClick(Sender: TObject);
    procedure DeleteLastPeakItemClick(Sender: TObject);
    procedure OverlayPropertiesItemClick(Sender: TObject);
    procedure ClearOverlayItemClick(Sender: TObject);
    procedure InvertItemClick(Sender: TObject);
    procedure ClearSelectionItemClick(Sender: TObject);
    procedure SelectRegionItemClick(Sender: TObject);
    procedure WavenumbersItemClick(Sender: TObject);
    procedure ShowFortratItemClick(Sender: TObject);
    procedure ShowSumItemClick(Sender: TObject);
    procedure ShowPartsItemClick(Sender: TObject);
    procedure PasteAnotherItemClick(Sender: TObject);
    procedure UpperStateItemClick(Sender: TObject);
    procedure PasteOverlayItemClick(Sender: TObject);
    procedure CopyPictureItemClick(Sender: TObject);
    procedure PrintSetupItemClick(Sender: TObject);
    procedure PrintItemClick(Sender: TObject);
    procedure ExitItemClick(Sender: TObject);
    procedure MRUFileListMRUItemClick(Sender: TObject; AFilename: string);
    procedure MRUFileListGetDisplayName(Sender: TObject; AFilename: string; var ADisplayName: string);
    procedure LoadOverlayItemClick(Sender: TObject);
    procedure ExportPlotItemClick(Sender: TObject);
    procedure ExportLineListItemClick(Sender: TObject);
    procedure BinaryfileItemClick(Sender: TObject);
    procedure CALPGMfileItemClick(Sender: TObject);
    procedure OldD3hItemClick(Sender: TObject);
    procedure OldPgopherItemClick(Sender: TObject);
    procedure OpenItemClick(Sender: TObject);
    procedure NewLinearMoleculeItemClick(Sender: TObject);
    procedure EmptyMixtureItemClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure AsymmetricTopItemClick(Sender: TObject);
    procedure ResidualsItemClick(Sender: TObject);
    procedure SaveArrangementItemClick(Sender: TObject);
    procedure SaveOverlayItemClick(Sender: TObject);
    procedure SaveOverlayAsItemClick(Sender: TObject);
    procedure SpecViewfileItemClick(Sender: TObject);
    procedure JB95FileItemClick(Sender: TObject);
    procedure CopyTopPlotItemClick(Sender: TObject);
    procedure TopPlotItemClick(Sender: TObject);
    procedure SymmetricTopItemClick(Sender: TObject);
    procedure HitranFileItemClick(Sender: TObject);
    procedure STROTABFileItemClick(Sender: TObject);
    procedure VariableCharSizeItemClick(Sender: TObject);
    procedure PlotOptionsItemClick(Sender: TObject);
    procedure FontItemClick(Sender: TObject);
    procedure LimitLineSegmentsItemClick(Sender: TObject);
    procedure LmatrixImportITEMClick(Sender: TObject);
    procedure VibrationalSpectrumItemClick(Sender: TObject);
    procedure EditMenuClick(Sender: TObject);
    procedure ShrinkplotItemClick(Sender: TObject);
    procedure ResetZoomItemClick(Sender: TObject);
    procedure MagnifyPlotItemClick(Sender: TObject);
    procedure WindowsonTopItemClick(Sender: TObject);
    procedure AllWindowsOnTaskBarItemClick(Sender: TObject);
    procedure VIBCAImportItemClick(Sender: TObject);
    procedure ExoMolFileItemClick(Sender: TObject);
    procedure SinglePlotItemClick(Sender: TObject);
    procedure CalculatorItemClick(Sender: TObject);
    procedure AnotherItemClick(Sender: TObject);
    procedure ColouredItemClick(Sender: TObject);
    procedure LineWidthItemClick(Sender: TObject);
    procedure AutofitItemClick(Sender: TObject);
    procedure ShowLinelistItemClick(Sender: TObject);
  private
    FMixture: TMixture;
    MixtureName: string;
    PlotForms: array of TPlotForm;
    nPlotForms: Integer;
    ShowConstantsOnNextFail: Boolean;
    function MakePlotForms:Integer;
    procedure SetMixture(AMixture: TMixture);
    procedure ChangeSimulationSetting(AIndex: Integer; Sender: TObject);
    procedure UpdateToolbar;
    {$IFDEF SHOWMEMUSAGE }
    procedure MemoryItemCLick(Sender: TObject);
    {$ENDIF}
    procedure nCPUsItemCLick(Sender: TObject);
    procedure SetName(AFilename: string);
    procedure HandleException(Sender: TObject; E: Exception);
    procedure AddThreadItems;
  protected
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer); override;
  public
    Overlays: TOverlays;
    LastObsWasLineForm: Boolean;
    ShowLogForm: Boolean;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure ArrangeForm(ScreenRect: TRect); override;
    procedure BringAllToFront(Sender: TObject);
    function LastPlot(WithPeak: Boolean = False): TPlot;
    procedure UpdateActions; override;

    { File Handling }
    procedure AddOverlay(FileName: string; Merge: Boolean; DefaultUnits: TResonanceUnits);
    procedure AddSampleMixture(SampleProc: TSampleProc);
    function ClearOverlays(Prompt: Boolean): Boolean;
    procedure ImportFile(DefaultProc: TMixtureReader; DefaultMerge: Boolean = False); overload;
    procedure ImportFile(DefaultProc: TMixtureReader; AFilename: string; Merge: Boolean); overload;
    procedure MergeOrReplaceMixture(NewMixture: TMixture; AFilename: string; Merge: Boolean);
    procedure LoadAnyObservations;
    procedure LoadFile(AFilename: string; Merge: Boolean);
    procedure LoadGeneralFile(AFilename: string; Merge, PreferOverlay: Boolean; DefaultUnits: TResonanceUnits);
    function OKToChange(TheMixture: Boolean; TheOverlays: Boolean): Boolean;
    function MakePlotForm(ASimulation: TSimulation; ATag: Integer = -1): TPlotForm;
    procedure SaveMixtureTo(AFilename: string);

    { Fitting }
    procedure ContourFit(JustResiduals: Boolean; AFilename: string; FitType: TFitType; CheckDerivatives: Double = 0; nCycles: Integer = -1);
    procedure ShowContour(Spectra: TOverlays; ASimulation: TColumnVector; ObsMinusCalc: Boolean);
    procedure ShowOverlaySum(Spectra: TOverlays; ASimulation: TColumnVector; ObsMinusCalc: Boolean);
    procedure DoLineFit(ObservationsLoader: TObservationReaderProc; const Source: string; ObsType: TObsType; JustResiduals: Boolean = False;
      TakeWeight: TTakeWeight = tkObs; CheckDerivatives: Double = 0; nCycles: Integer = -1);

    { Plotting }
    procedure ConsiderReplot(All: Boolean = False; PurgeLineSelection: Boolean = True);
    procedure DoPlot(All: Boolean = False; PurgeLineSelection: Boolean = True);
    procedure ShowSimulation; override;

    property Mixture: TMixture read FMixture write SetMixture;
  end;

var
  MainForm: TMainForm;

implementation

uses
  {$IFDEF LCL_} LCLType, LCLProc, HelpIntfs, LazHelpHTML,
  {$ENDIF}
  {$IFDEF USE_LAZCHMHELP} LazChmHelp,
  {$ENDIF} {Define to use lazarus .chm viewer; not recomended at the moment as it does not handle greek letters such as &Pi;}
  {$IFDEF VCL_} PHtmlHelpViewer, {$ENDIF}
  {$IFDEF RecyclerMM} FRMMSnapShot, {$ENDIF}
  {$IFDEF SHOWMEMUSAGE} FastMMUsageTracker, {$ENDIF}
  {$IFDEF HELPHELP} MakeHelp, {$ENDIF}
  Math, Hitran, PgopherImport, D3hImport, ImportCALPGM, ImportVIBCA, FormUtils,
  ImportJB95, ImportSTROTAB, ImportLmatrix, UCompat, GStyle, WidthFrm,
  LineExportOpt, ExpPlotLoader, ExpPlot, ReadBinary, ChildForm, ExpCan,
  ExportG, ImportSpecView, Clipbrd, OverlaySumFitter, Constrain, ExoMol,
  LogUnit, Progress, Fitter, LineFrm, OvrlyFrm, ConstFrm, Worker,
  StateInfo, TranInfo, LevelFrm, Profile, ProfileForm, WhatsThis, WCompat,
  QuantumSpin, CalFrm, Atomic, CalSource, Sysinfo, PGWork, FixForm, ResidFrm,
  SettingsForm, MNodeExpr, LoomisWood, Expr, Exporter, Detach, AutoFitFrm,
  AutoFit;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}
{ Helpers }

procedure TMainForm.SetMixture(AMixture: TMixture);
var
  i: Integer;
  SaveSettingsInfo: TSaveSettingsInfo;
begin
  if (AMixture <> Mixture) and not BeginCalculation(False) then begin
    try
      if not(csDestroying in ComponentState) then
        Application.ProcessMessages; { Process any pending actions }
      FreeAndNil(FMixture);
      if not(csDestroying in ComponentState) then
        Application.ProcessMessages; { Allow time for forms to disappear }
      ForAllSettingsForms({$IFDEF FPC}@{$ENDIF}ClearSettings, nil);
      FMixture := AMixture;
    finally
      EndCalculation;
    end;
    if Assigned(AMixture) then
    begin
      FSimulation := Mixture.Simulations[0];
      FSimulation.Name := 'Simulation';
      if Mixture.nSimulations > 1 then
        for i := 1 to Mixture.nSimulations - 1 do
          MakePlotForm(Mixture.Simulations[i]);
      SetGlobalQuantumNumberFormat(Mixture.QuantumNumberFormat);
      with SaveSettingsInfo do begin
        DOMDocument := Mixture.LoadedFormSettings;
        SavePositions := LoadPositionsItem.Checked;
        SaveSettings := LoadSettingsItem.Checked;
      end;
      BeginGlobalUpdate;
      ForAllSettingsForms({$IFDEF FPC}@{$ENDIF}FinishLoading, TObject(@SaveSettingsInfo));
      EndGlobalUpdate;
      Mixture.ClearFormSettings;
    end else
      FSimulation := nil;
    PerformMNodeNotification(Mixture, mnNewBaseNode, 0);
  end;
end;

procedure TMainForm.ChangeSimulationSetting(AIndex: Integer; Sender: TObject);
begin
  if Assigned(Simulation) then
    Simulation.SetFlag(AIndex, (Sender as TSpeedButton).Down);
  ShowSimulation;
end;

procedure TMainForm.UpdateToolbar;
var
  i: Integer;
begin
  if not Assigned(Mixture) then
    Exit; { Mixture being changed }
  with Mixture.Simulations[0] do begin
    UpdateEditControl(MainForm.Temperature, Temperature);
    UpdateEditControl(MainForm.Gaussian, Gaussian);
    UpdateEditControl(MainForm.Lorentzian, Lorentzian);
    UpdateEditControl(MainForm.Foffset, Foffset);
    ShowSumButton.Down := ShowSum;
    ShowPartsButton.Down := ShowParts;
    ShowFortratButton.Down := ShowFortrat;
    ShowSumItem.Checked := ShowSum;
    ShowPartsItem.Checked := ShowParts;
    ShowFortratItem.Checked := ShowFortrat;
    UpperStateItem.Checked := UseUpper;
    for i := 0 to UnitsItem.Count - 1 do
      if UnitsItem.Items[i].Tag = Ord(PlotUnits) then
        UnitsItem.Items[i].Checked := True;
  end;
end;

function TMainForm.LastPlot(WithPeak: Boolean): TPlot;
var
  i: Integer;
begin
  Result := nil;
  if not Assigned(Overlays) then
    Exit;
  for i := 0 to Overlays.nPlots - 1 do
    if Overlays[i] = FLastPlot then
      Break
    else if i = Overlays.nPlots - 1 then
      FLastPlot := nil;
  if Assigned(FLastPlot) and (not WithPeak or (FLastPlot.nPeaks > 0)) then begin
    Result := FLastPlot;
    Exit;
  end;
  for i := Overlays.nPlots - 1 downto 0 do
    if not WithPeak or (Overlays[i].nPeaks > 0) then begin
      Result := Overlays[i];
      Exit;
    end;
end;

procedure TMainForm.UpdateActions;
begin
  inherited UpdateActions;
  { LCL seems to need this here, rather than in Form.Show }
  if Visible and ShowLogForm then begin
    Application.ProcessMessages;
    LogForm.Show; { Deferred from start up }
    ShowLogForm := False;
  end;
end;

constructor TMainForm.Create(AOwner: TComponent);
{$IFDEF LCL_}
var
  HelpFileDir, lHelpExe: string;
  UseChmFile: Boolean;
{$ENDIF}
begin
  RegisterClass(TPanel); { May be a bug, but needed in LCL and VCL }
  inherited;
  Tag := 0;
  FMixture := TMixture.Create(nil);
  FSimulation := Mixture.Simulations[0];
  UpdateToolbar;
  DefaultFormSettings := MakeFormSettings(nil);
  {$IFDEF LCL_}
    HelpFileDir := ExtractFilePath(ParamStr(0));
    lHelpExe := HelpFileDir + 'lhelp';
    if DirectoryExists(HelpFileDir + 'Help') then
      HelpFileDir := HelpFileDir + 'Help'
    else if DirectoryExists(HelpFileDir + '..' + DirectorySeparator + 'Help') then
      HelpFileDir := HelpFileDir + '..' + DirectorySeparator + 'Help'
    else
      HelpFileDir := 'Help';
    HelpFileDir := ExpandFileName(HelpFileDir);
    if not FileExists(lHelpExe) then
      lHelpExe := HelpFileDir + DirectorySeparator + 'lhelp';
    UseChmFile :=
    {$IFDEF USE_LAZCHMHELP}
        FileExists(lHelpExe) and FileExists(HelpFileDir + DirectorySeparator +
        'pgopher.chm');
    {$ELSE}
        False;
    {$ENDIF}
    with TSingleHTMLHelpDatabase.Create(Self) do begin
      AutoRegister := True;
      KeywordPrefix := 'Pgopher/';
      if not UseChmFile then
        if FileExists(HelpFileDir+DirectorySeparator+'index.htm') then
          DefaultBaseURL := 'file://' + HelpFileDir
        else
          DefaultBaseURL := 'http://pgopher.chm.bris.ac.uk/Help'
      else
        DefaultBaseURL := 'pgopher.chm://';
    end;
    if not UseChmFile then
      with THTMLBrowserHelpViewer.Create(Self) do
        AutoRegister := True;
    {$IFDEF USE_LAZCHMHELP}
      if UseChmFile then
        with TChmHelpViewer.Create(Self) do
        begin
          AutoRegister := True;
          HelpEXE := lHelpExe;
          HelpLabel := 'pgopherxxx';
          HelpFilesPath := HelpFileDir + DirectorySeparator;
        end;
    {$ENDIF}
    Application.OnException := {$IFDEF FPC}@{$ENDIF}HandleException;
    {$IFNDEF OLDLAZARUS}
      Application.Flags := Application.Flags + [AppNoExceptionMessages];
    {$ENDIF}
  {$ENDIF}
  HasBeenArranged := True; { Take initial default }
  // FixAllForms := False; { Only adjust the main form }
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  MenuItem: TMenuItem;
begin
  inherited;
  {$IFDEF LCL}
    Application.OnHint := @StatusBarHint;
    angVacItem.Caption := 'Å (Vac&uum)';
    angAirItem.Caption := 'Å (A&ir)';
  {$ENDIF}
  {$IFDEF SHOWMEMUSAGE }
    MenuItem := TMenuItem.Create(Self);
    MenuItem.Caption := 'Memory';
    MenuItem.OnClick := MemoryItemCLick;
    ViewMenu.Insert(ViewMenu.IndexOf(TimingItem) + 1, MenuItem);
  {$ENDIF}
  ToolBar.DisableAlign;
  OpenButton.Left := 0;
  ToolBar.EnableAlign;
  StuffCalibrationMenu(CalibrationItem, {$IFDEF FPC}@{$ENDIF}I2ItemClick);
  {$IFDEF USEBETA}
    DumpItem.Visible := True;
  {$ELSE}
    DumpItem.Visible := False;
  {$ENDIF}
  {$IFNDEF MSWINDOWS}
    DumpItem.Visible := False;
    WhatIsItem.Visible := False;
    {$IFNDEF DARWIN}
      MenuItem := TMenuItem.Create(Self);
      MenuItem.Caption := 'Bring All To Front';
      MenuItem.OnClick := @BringAllToFront;
      MenuItem.Hint := 'Bring all windows in front of windows from other applications';
      ViewMenu.Insert(ViewMenu.IndexOf(ArrangeFormsItem) + 1, MenuItem);
    {$ENDIF}
  {$ENDIF}
  AdjustMainMenu(MainMenu, AboutItem.OnClick);
  {$IFDEF DARWIN}
    AboutItem.Visible := False;
    ExitItem.Visible := False;
  {$ENDIF}
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  inherited;
  AddThreadItems; { Add here so that Checked item reflects number of processes }
  {$IFDEF LCL_}
    Icon := Application.Icon; // Oddly required if debugging only
  {$ENDIF}
  {$IFDEF MSWINDOWS}
    SearchItem.Visible := False; { Search is broken - http://support.microsoft.com/kb/241381 }
    UsingHelpItem.Visible := False; { No longer in help file }
    IndexItem.Caption := 'Contents';
    ProceduresItem.Caption := 'Index';
  {$ENDIF}
end;

procedure TMainForm.FontItemClick(Sender: TObject);
begin
  G.ChooseFont(nil);
  OC.Font := G.Font;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := CanClose and OKToChange(True, True);
end;

function TMainForm.OKToChange(TheMixture: Boolean; TheOverlays: Boolean): Boolean;
var
  s, E: string;
  procedure Check(MNode: TStreamableMNode; DefaultName: string);
  begin
    if Assigned(MNode) and MNode.Modified then begin
      Result := False;
      if s <> '' then begin
        s := s + ' and ';
        E := ' have';
      end;
      if MNode.Name = '' then
        s := s + DefaultName
      else
        s := s + MNode.Name;
    end;
  end;

begin
  Result := True;
  s := '';
  E := ' has';
  if TheMixture then
    Check(Mixture, 'Mixture');
  if TheOverlays then
    Check(Overlays, 'Overlay');
  if not Result then begin
    Result := MessageDlg(s + E + ' changed: discard?', mtWarning, mbOkCancel,
      0) = mrOK;
    if Result then begin
      if TheMixture and Assigned(Mixture) then
        Mixture.Modified := False;
      if TheOverlays and Assigned(Overlays) then
        Overlays.Modified := False;
    end;
  end;
end;

destructor TMainForm.Destroy;
begin
  Application.OnException := nil;
  ClearOverlays(False);
  SetMixture(nil);
  inherited;
end;

procedure TMainForm.SetName(AFilename: string);
var
  i: Integer;
  NewCaption: string;
begin
  DoDirSeparators(AFilename);
  MixtureName := AFilename;
  OpenDialog.FileName := MixtureName;
  if AFilename = '' then
    NewCaption := 'PGOPHER'
  else begin
    MixtureName := ExpandFileName(AFilename);
    SetCurrentDir(ExtractFilePath(AFilename));
    { TODO: This is not that useful, as default directory for windows open dialog is from most recent open/save dialog }
    NewCaption := ExtractFileName(AFilename) + ' - PGOPHER';
    RecentFiles.AddItem(MixtureName);
    RecentFiles.SaveToRegistry;
  end;
  for i := 0 to Screen.FormCount - 1 do
    if Screen.Forms[i] is TPlotForm then
      if Screen.Forms[i] <> Self then
        TPlotForm(Screen.Forms[i]).Caption := IntToStr(Screen.Forms[i].Tag) +
          ' - ' + NewCaption
      else
        TPlotForm(Screen.Forms[i]).Caption := NewCaption;
  Application.Title := NewCaption;
end;

procedure TMainForm.HandleException(Sender: TObject; E: Exception);
begin
{$IFDEF LCL_}
  if IsExpectedException(E.ClassType) then
    Application.MessageBox(PChar(E.Message), 'Pgopher', MB_OK + MB_ICONERROR)
  else
    try
{$IFNDEF OLDLAZARUS}
      Application.Flags := Application.Flags - [AppNoExceptionMessages];
      if ExceptObject is Exception then begin
        WriteLn(Log, 'TApplication.HandleException ', Exception(ExceptObject).Message);
      end else begin
        WriteLn(Log, 'TApplication.HandleException Strange Exception ');
      end;
      System.DumpExceptionBackTrace(Log);
{$ENDIF}
      Application.ShowException(E);
    finally
{$IFNDEF OLDLAZARUS}
      Application.Flags := Application.Flags + [AppNoExceptionMessages];
{$ENDIF}
    end;
{$ENDIF}
end;

{ Plotting }

procedure TMainForm.ShowSimulation;
var
  i, old: Integer;
  LW: TLoomisWood;
  xShift, Shift0: Double;
  ArrangeNeeded: Boolean;
begin
  LW := nil;
  if Assigned(Overlays) then
    LW := ActiveMnodeExpr(Overlays, TLoomisWood) as TLoomisWood;
  MakePlotForms;
  if Assigned(LW) and (LW.nPlots > 1) then begin
    if WindowsonTopItem.Checked then begin
      WindowsonTopItem.Checked := False;
      WindowsonTopItemClick(nil);
    end;
    for i := LW.nPlots to High(PlotForms) do
      PlotForms[i].Free; { Discard unwanted forms }
    ArrangeNeeded := (LW.nPlots < Length(PlotForms)) or (LW.LastyOffset <> LW.yOffset);
    Old := Length(PlotForms);
    SetLength(PlotForms, LW.nPlots);
    for i := Old to LW.nPlots - 1 do
      PlotForms[i] := nil;
    Shift0 := LW.xShift(0);
    for i := 1 to LW.nPlots - 1 do begin
      if not Assigned(PlotForms[i]) then begin
        PlotForms[i] := MakePlotForm(nil,i);
        ArrangeNeeded := True;
      end else begin
        PlotForms[i].Simulation.ResetOverride;
        PlotForms[i].Simulation.UpdateBaseNode(Self.Simulation);
      end;
      xShift := LW.xShift(i)-Shift0;
      PlotForms[i].ToolBar.Visible := False;
      with PlotForms[i].Simulation do begin
        Fmin := Self.Simulation.Fmin+xShift;
        Fmax := Self.Simulation.Fmax+xShift;
      end;
    end;
    if ArrangeNeeded then
      ArrangeFormsItemClick(nil);
    LW.LastyOffset := LW.yOffset;
  end else
    for i := 1 to High(PlotForms) do
      if Assigned(PlotForms[i]) then
        PlotForms[i].ToolBar.Visible := True;
  try
    ShowSimulation(Mixture, Overlays);
  except
    if ShowConstantsOnNextFail then
      ConstantsItemClick(nil);
    raise;
  end;
  ShowConstantsOnNextFail := False;
  { Update other screens - should use same linelist }
  for i := 0 to Screen.FormCount - 1 do
    if (Screen.Forms[i] <> Self) and (Screen.Forms[i] is TPlotForm) then
      TPlotForm(Screen.Forms[i]).ShowSimulation(Mixture, Overlays);
end;

procedure TMainForm.DoPlot(All: Boolean; PurgeLineSelection: Boolean);
var
  i: Integer;
begin
  if PurgeLineSelection and Assigned(Mixture.LineList) and Mixture.LineList.Partial then
    Mixture.LineList := nil;
  if All then begin
    if AllWanted then
      FindLimits(True);
    for i := 0 to Screen.FormCount - 1 do
      if Screen.Forms[i] is TPlotForm then
        with TPlotForm(Screen.Forms[i]) do
          if AllWanted then
            FindLimits(True);
  end;
  AllWanted := False;
  ShowSimulation;
end;

procedure TMainForm.ColouredItemClick(Sender: TObject);
begin
  if ColouredItem.Checked then
    G.GOptions := G.GOptions + [grColouredPens]
  else
    G.GOptions := G.GOptions - [grColouredPens];
  OC.GOptions := G.GOptions;
end;

procedure TMainForm.ConsiderReplot(All: Boolean; PurgeLineSelection: Boolean);
begin
  if not Assigned(Mixture) or Mixture.AutoReplot then begin
    ForAllSettingsForms({$IFDEF FPC}@{$ENDIF}DoUpdateActions, nil);
    { Force update of Child Windows }
    AllWanted := True;
    DoPlot(All, PurgeLineSelection);
  end;
end;

{ File Handling }

procedure TMainForm.AddOverlay(FileName: string; Merge: Boolean; DefaultUnits: TResonanceUnits);
var
  DefaultIntensityUnits: TIntensityUnits;
begin
  if not Merge and not OKToChange(False, True) then
    Abort;
  BeginGlobalUpdate;
  try
    if DefaulttoArbitraryUnitsItem.Checked then
      DefaultIntensityUnits := Arbitrary
    else
      DefaultIntensityUnits := Normalized;
    AddToOverlay(Overlays, FileName, Merge, DefaultUnits,
    PromptOnImportItem.Checked, DefaultIntensityUnits);
  finally
    EndGlobalUpdate(True);
  end;
  if FileName = '' then
    Exit { Skip }
  else if FileName[Length(FileName)] = '>' then
    RecentFiles.AddItem(FileName)
  else
    RecentFiles.AddItem(ExpandFileName(FileName));
  RecentFiles.SaveToRegistry;
end;

function TMainForm.ClearOverlays(Prompt: Boolean): Boolean;
begin
  if Prompt and not OKToChange(False, True) then begin
    Result := False;
    Exit;
  end;
  Result := True;
  BeginGlobalUpdate;
  try
    FreeAndNil(Overlays);
  finally
    EndGlobalUpdate(True);
  end;
end;

procedure TMainForm.ImportFile(DefaultProc: TMixtureReader; DefaultMerge: Boolean);
var
  Merge: Boolean;
begin
  Merge := DefaultMerge or MergeItem.Checked;
  if (Merge or CloseQuery) and ImportDialog.Execute then
    ImportFile(DefaultProc, ImportDialog.FileName, Merge);
end;

procedure TMainForm.MergeOrReplaceMixture(NewMixture: TMixture; AFilename: string; Merge: Boolean);
var
  MemoryStream: TMemoryStream;
begin
  if Merge then
    try
      MemoryStream := TMemoryStream.Create;
      try
        BeginGlobalUpdate;
        NewMixture.SaveToStream(MemoryStream);
        MemoryStream.Position := 0;
        Mixture.MergeFromStream(MemoryStream);
      finally
        EndGlobalUpdate(True);
        MemoryStream.Free;
      end;
    except
      NewMixture.Free;
    end
  else begin
    SetMixture(NewMixture);
    SetName('');
    Mixture.Comment := 'Imported from ' + AFilename;
    Mixture.CheckProperties;
    Mixture.Modified := False;
  end;
end;

procedure TMainForm.ImportFile(DefaultProc: TMixtureReader; AFilename: string; Merge: Boolean);
begin
  MergeOrReplaceMixture(ImportMixture(DefaultProc, AFilename), AFilename, Merge);
  ConsiderReplot;
end;

procedure TMainForm.I2ItemClick(Sender: TObject);
var
  MenuItem: TMenuItem;
  s: string;
  P: Integer;
begin
  MenuItem := Sender as TMenuItem;
  s := MenuItem.Caption;
  P := Pos('&', s);
  if P > 0 then
    Delete(s, P, 1);
  AddOverlay('<' + s + ' Calibration>', True, cm1);
  ConsiderReplot(False, False);
end;

procedure TMainForm.LoadAnyObservations;
begin
  if LastObsWasLineForm and LineForm.HasObservations then
    LineForm.LoadObservations('', obLine, tkObs, False)
  else
    LogForm.LoadObservations;
end;

procedure TMainForm.LoadFile(AFilename: string; Merge: Boolean);
var
  NewMixture: TMixture;
  Stream: TFileStream;
  FileDir: string;
begin
  if Merge and Assigned(Mixture) and (Mixture.nMolecules > 0) then begin
    Stream := TFileStream.Create(AFilename, fmOpenRead or DefaultShareMode);
    try
      BeginGlobalUpdate;
      Mixture.MergeFromStream(Stream, AFilename);
      SetName('');
      ShowConstantsOnNextFail := True;
    finally
      EndGlobalUpdate(True);
      Stream.Free;
    end;
  end else begin
    NewMixture := TMixture.Create(nil);
    try
      NewMixture.LoadFromFile(AFilename);
      SetMixture(NewMixture);
      SetName(AFilename);
      FileDir := ExtractFilePath(AFilename);
      if (FileDir <> '') and (LogForm.ObsFileName.FileName <> '') and (ExtractFilePath(LogForm.ObsFileName.FileName) = '') then
        LogForm.ObsFileName.FileName := FileDir + LogForm.ObsFileName.FileName;
      ShowConstantsOnNextFail := True;
    except
      NewMixture.Free;
      raise;
    end;
  end;
  ClearObservations;
end;

procedure TMainForm.LoadGeneralFile(AFilename: string; Merge, PreferOverlay: Boolean; DefaultUnits: TResonanceUnits);
var
  Reader: TMixtureReader;
begin
  case FileType(AFilename, PreferOverlay, Reader) of
    fiParameters:
      if Merge or OKToChange(True, False) then
        if not Assigned(Reader) then
          LoadFile(AFilename, Merge)
        else begin
          ImportFile(Reader, AFilename, Merge);
          if Pointer({$IFNDEF FPC}@{$ENDIF}Reader) = Pointer(@ImportJB95File) then
            AddOverlay(AFilename, Merge, DefaultUnits);
          { JBW files can also contain an overlay }
        end;
    fiObservations:
      LogForm.ObsFileName.FileName := AFilename;
    fiOverlay:
      AddOverlay(AFilename, Merge, DefaultUnits);
  end;
end;

procedure TMainForm.AddThreadItems;
var
  MenuItem: TMenuItem;
begin
  while ThreadsItem.Count < Max(Max(4, 4 * nCPUs), nWorkers) do begin
    MenuItem := TMenuItem.Create(Self);
    MenuItem.Tag := ThreadsItem.Count + 1;
    if ThreadsItem.Count = 0 then begin
      MenuItem.Caption := 'Single Process';
      MenuItem.Tag := 0;
    end else
      MenuItem.Caption := Format('%d Processes', [ThreadsItem.Count + 1]);
    MenuItem.OnClick := {$IFDEF FPC}@{$ENDIF}nCPUsItemCLick;
    MenuItem.AutoCheck := True;
    MenuItem.RadioItem := True;
    MenuItem.GroupIndex := 45;
    ThreadsItem.Insert(ThreadsItem.Count, MenuItem);
  end;
  ThreadsItem.Items[Max(nWorkers, 1) - 1].Checked := True;
end;

procedure TMainForm.AllWindowsOnTaskBarItemClick(Sender: TObject);
begin
  SetAllOnTaskBarMode(AllWindowsOnTaskBarItem.Checked);
end;

procedure TMainForm.AnotherItemClick(Sender: TObject);
begin
  Detach1(ApplicationExecutable,'');
end;

procedure TMainForm.ArrangeForm(ScreenRect: TRect);
var
  R: TRect;
  i, y, dy: Integer;
  LW: TLoomisWood;
  y2O: Double;
begin
  inherited;
  R := ScreenBounds(Self);
  if MakePlotForms < 2 then begin
    OffsetRect(R, ScreenRect.Left - R.Left, ScreenRect.Top - R.Top);
    SetScreenBounds(Self, R);
  end else begin
    dy := (ScreenRect.Bottom-ScreenRect.Top);
    y := ScreenRect.Top;
    if Assigned(Overlays) then
      LW := ActiveMnodeExpr(Overlays, TLoomisWood) as TLoomisWood;
    if Assigned(LW) and (LW.nPlots > 1) then begin
        dy := dy - 2*ToolBar.Height;
        if LW.yOffset >= 0 then
          y2O := Min(LW.yOffset,1.2)
        else
          y2O := Max(LW.yOffset,-1.2);
        dy := Trunc(dy/(1+(nPlotForms-1)*Abs(y2O)));
        if y2O < 0 then begin
          y := y-Trunc(y2O*dy)*(nPlotForms-1)+2*ToolBar.Height;
        end;
    end else begin
      dy := dy div nPlotForms;
      y2O := 1;
    end;
    y := y + dy;
    for i := High(PlotForms) downto 0 do
      if PlotForms[i] <> nil then begin
        if i = 0 then begin
          Inc(dy, 2*ToolBar.Height);
          if y2O > 0 then
            Inc(y, 2*ToolBar.Height);
        end;
        SetScreenBounds(PlotForms[i], Rect(ScreenRect.Left, y-dy, ScreenRect.Left+R.Right-R.Left, y));
        PlotForms[i].BringToFront;
        Application.ProcessMessages;
        y := y + Trunc(y2O*dy);
      end;
  end;
end;

procedure TMainForm.BringAllToFront(Sender: TObject);
begin
  WCompat.BringAllToFront;
end;

procedure TMainForm.ArrangeFormsItemClick(Sender: TObject);
var
  i: Integer;
  Rect: TRect;
begin
  Rect := ScreenAreaFor(Application.MainForm);
  for i := 0 to Screen.FormCount - 1 do
    if Screen.Forms[i] is THelpForm then
      with THelpForm(Screen.Forms[i]) do
        if Visible then
          ArrangeForm(Rect)
        else
          HasBeenArranged := False;
end;

{ Fitting }

procedure TMainForm.ContourFit(JustResiduals: Boolean; AFilename: string;
  FitType: TFitType; CheckDerivatives: Double; nCycles: Integer);
var
  i: Integer;
  NoParameters: TParameters;
begin
  if not Assigned(Overlays) then
    raise EReportException.Create('Overlays required');
  if BeginCalculation(True) then begin
    WriteLn(Log, 'Fit already in progress');
    Exit;
  end;
  try
    PushProgress('Fit', Mixture.FitCycles);
    if Overlays.ContourFile then begin
      if AFilename = '' then
        raise EReportException.Create
          ('ContourFile flag set but no external file provided');
      PushProgress('Loading');
      try
        LoadObservations(Mixture, nil, AFilename, obContour);
      finally
        PopProgress;
      end;
    end else
      ClearObservations;
    if CheckDerivatives <> 0 then
      nCycles := 1
    else if nCycles < 0 then
      nCycles := Mixture.FitCycles;
    for i := 1 to Max(nCycles,0) do begin
      SetProgress(i);
      if nCycles > 0 then
        Mixture.PrepareForFit(FitType, Overlays)
      else
        NoParameters := nil;
      if Mixture.PrintLevel > plMinimal then
        Mixture.PrettyPrint(Worker.Log, Mixture.PrintLevel);
      try
        if FitType = ftContour then
          if nCycles > 0 then
            FitContour(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowContour, Mixture.FittingParameters, JustResiduals, CheckDerivatives)
          else
            FitContour(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowContour, NoParameters, JustResiduals, CheckDerivatives)
        else begin
          if nCycles > 0 then
            FitOverlaySum(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowOverlaySum, Mixture.FittingParameters, JustResiduals, CheckDerivatives)
          else
            FitOverlaySum(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowOverlaySum, NoParameters, JustResiduals, CheckDerivatives);
          if OverlayForm.Visible then
            OverlayForm.ShowCurrentPlot;
        end;
        if nCycles > 0 then
          Mixture.PushParameters(FitType, True); { Save current set for redo fit }
      except
        Mixture.DiscardFits(1);
        raise;
      end;
    end;
    PopProgress;
  finally
    EndCalculation;
    PopProgress;
  end;
end;

procedure TMainForm.ShowContour(Spectra: TOverlays; ASimulation: TColumnVector; ObsMinusCalc: Boolean);
var
  i, j, l: Integer;
  v, f, df: Double;
  dmin, dmax: Double; { Limits of obs-calc }
  omin, omax: Double; { Limits of observed }
  XLabel: string;
  Plotted: Boolean;

  procedure ProcessPoint(v: Double);
  begin
    if l = 1 then begin
      omin := v;
      omax := v;
    end else begin
      if v < omin then
        omin := v;
      if v > omax then
        omax := v;
    end;
    if ObsMinusCalc then
      v := ASimulation[l]
    else
      v := v - ASimulation[l];
    if l = 1 then begin
      dmin := v;
      dmax := v;
    end else begin
      if v < dmin then
        dmin := v;
      if v > dmax then
        dmax := v;
    end;
    Inc(l);
  end;
  procedure ShowPoint(G: TGraphM);
  begin
    if Plotted then
      G.ClipJoin(f, v)
    else
      G.ClipPoint(f, v);
    Plotted := True;
  end;
begin
  SetLength(FortratPlotRanges, 0);
  SetLength(SpectrumPlotRanges, 0);
  Split.Visible := True;
  Split.MakeBothVisible;
  { Determine Limits on obs-calc, obs }
  l := 1;
  for i := 0 to Spectra.nPlots - 1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then
          if ConvolutionNeeded then
            with ConvolutedSpectrum do
              for j := 0 to nDF-1 do
                ProcessPoint(Plot[j])
          else
            for j := StartPoint to EndPoint do begin
              v := Intensity[j];
              if IsNan(v) then
                Continue;
              if ObsMinusCalc then
                v := AdjustIntensity(v);
              ProcessPoint(v);
            end;
  if l = 1 then
    Exit; { No points }
  { Plot obs - calc }
  OC.Start;
  OC.Map(Mixture.Simulations[0].Fmin, Mixture.Simulations[0].Fmax, dmin, dmax);
  OC.Axes('', 'Obs-Calc');
  l := 1;
  for i := 0 to Spectra.nPlots - 1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then begin
          Plotted := False;
          if ConvolutionNeeded then
            with ConvolutedSpectrum do begin
              df := (FMax-FMin)/(nDF-1);
              for j := 0 to nDF-1 do begin
                f := FMin+j*df;
                v := ASimulation[l];
                if not ObsMinusCalc then
                  v := Plot[j] - v;
                ShowPoint(OC);
                Inc(l);
              end
            end
          else
            for j := StartPoint to EndPoint do begin
              f := Convert(AdjustFrequency(Frequency[j]), Units, Mixture.Simulations[0].PlotUnits);
              v := ASimulation[l];
              if not ObsMinusCalc then
                v := Intensity[j] - v;
              ShowPoint(OC);
              Inc(l);
            end;
        end;
  OC.Finish;
  G.Start;
  G.Map(Mixture.Simulations[0].Fmin, Mixture.Simulations[0].Fmax, omin, omax);
  XLabel := TypeOfUnit(Mixture.Simulations[0].PlotUnits) + '/' + NameOfUnit(Mixture.Simulations[0].PlotUnits);
  G.Axis(Mixture.Simulations[0].Fmin, Mixture.Simulations[0].Fmax, omin, True, XLabel);
  G.Axis(G.Ymin, G.Ymax, G.Xmin, False, '');
  { Plot calculated spectrum }
  G.Style(0);
  l := 1;
  for i := 0 to Spectra.nPlots - 1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then begin
          Plotted := False;
          if ConvolutionNeeded then
            with ConvolutedSpectrum do begin
              df := (FMax-FMin)/(nDF-1);
              for j := 0 to nDF-1 do begin
                f := FMin+j*df;
                v := ASimulation[l];
                if ObsMinusCalc then
                  v := Plot[j] - v;
                ShowPoint(G);
                Inc(l);
              end;
            end
          else
            for j := StartPoint to EndPoint do begin
              f := Convert(AdjustFrequency(Frequency[j]), Units, Mixture.Simulations[0].PlotUnits);
              v := ASimulation[l];
              if ObsMinusCalc then
                v := AdjustIntensity(Intensity[j]) - v;
              ShowPoint(G);
              Inc(l);
            end;
        end;
  G.Style(1);
  { Plot observed spectrum }
  l := 1;
  for i := 0 to Spectra.nPlots - 1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then begin
          Plotted := False;
          if ConvolutionNeeded then
            with ConvolutedSpectrum do begin
              df := (FMax-FMin)/(nDF-1);
              for j := 0 to nDF-1 do begin
                f := FMin+j*df;
                v := Plot[j];
                ShowPoint(G);
                Inc(l);
              end
            end
          else
            for j := StartPoint to EndPoint do begin
              f := Convert(AdjustFrequency(Frequency[j]), Units, Mixture.Simulations[0].PlotUnits);
              v := AdjustIntensity(Intensity[j]);
              ShowPoint(G);
              Inc(l);
            end;
        end;
  G.Finish;
end;

procedure TMainForm.ShowOverlaySum(Spectra: TOverlays; ASimulation: TColumnVector; ObsMinusCalc: Boolean);
var
  i, j, l: Integer;
  v, f: Double;
  dmin, dmax: Double; { Limits of obs-calc }
  omin, omax: Double; { Limits of observed }
  XLabel: string;
begin
  SetLength(FortratPlotRanges, 0);
  SetLength(SpectrumPlotRanges, 0);
  Split.Visible := True;
  Split.MakeBothVisible;
  { Determine Limits on obs-calc, obs }
  l := 1;
  for i := 0 to Spectra.nPlots - 1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if ComponentType = Target then
          for j := StartPoint to EndPoint do
          begin
            { TODO: should probably select frequency range here also }
            v := Intensity[j];
            if l = 1 then
            begin
              omin := v;
              omax := v;
            end
            else
            begin
              if v < omin then
                omin := v;
              if v > omax then
                omax := v;
            end;
            v := v - ASimulation[l];
            if l = 1 then
            begin
              dmin := v;
              dmax := v;
            end
            else
            begin
              if v < dmin then
                dmin := v;
              if v > dmax then
                dmax := v;
            end;
            Inc(l);
          end;
  if l = 1 then
    Exit; { No points }
  { Plot obs - calc }
  OC.Start;
  OC.Map(Mixture.Simulations[0].Fmin, Mixture.Simulations[0].Fmax, dmin, dmax);
  OC.Axes('', 'Obs-Calc');
  l := 1;
  for i := 0 to Spectra.nPlots - 1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if ComponentType = Target then
        begin
          if Colour <> clNone then
            OC.StyleP(Colour, OC.LineStyle.Width, lpSolid)
          else
            OC.Style(1);
          for j := StartPoint to EndPoint do
          begin
            f := Convert(AdjustFrequency(Frequency[j]), Units,
              Mixture.Simulations[0].PlotUnits);
            v := Intensity[j] - ASimulation[l];
            if j = StartPoint then
              OC.ClipPoint(f, v)
            else
              OC.ClipJoin(f, v);
            Inc(l);
          end;
        end;
  OC.Finish;
  G.Start;
  G.Map(Mixture.Simulations[0].Fmin, Mixture.Simulations[0].Fmax, omin, omax);
  XLabel := Mixture.UnitsTypeName + '/' + NameOfUnit(Mixture.Simulations[0].PlotUnits);
  G.Axis(Mixture.Simulations[0].Fmin, Mixture.Simulations[0].Fmax, omin,  True, XLabel);
  G.Axis(G.Ymin, G.Ymax, G.Xmin, False, '');
  l := 1;
  for i := 0 to Spectra.nPlots - 1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if ComponentType = Target then
        begin
          if Colour <> clNone then
            G.StyleP(Colour, G.LineStyle.Width, lpDot)
          else
            G.Style(1);
          for j := StartPoint to EndPoint do
          begin
            f := Convert(AdjustFrequency(Frequency[j]), Units,
              Mixture.Simulations[0].PlotUnits);
            v := ASimulation[l];
            if j = StartPoint then
              G.ClipPoint(f, v)
            else
              G.ClipJoin(f, v);
            Inc(l);
          end;
        end;
  G.Style(1);
  for i := 0 to Spectra.nPlots - 1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if ComponentType = Target then
        begin
          if Colour <> clNone then
            G.StyleP(Colour, G.LineStyle.Width, lpSolid)
          else
            G.Style(1);
          for j := StartPoint to EndPoint do
          begin
            f := Convert(AdjustFrequency(Frequency[j]), Units,
              Mixture.Simulations[0].PlotUnits);
            v := Intensity[j];
            if j = StartPoint then
              G.ClipPoint(f, v)
            else
              G.ClipJoin(f, v);
          end;
        end;
  G.Finish;
end;

procedure TMainForm.DoLineFit(ObservationsLoader: TObservationReaderProc;
  const Source: string; ObsType: TObsType; JustResiduals: Boolean;
  TakeWeight: TTakeWeight; CheckDerivatives: Double; nCycles: Integer);
var
  i: Integer;
  FitStatus: string;
  Problem: Exception;
  NoFittingParameters: TParameters;
  OldSD, NewSD: Double;
begin
  Problem := Mixture.FixupProperties;
  if Assigned(Problem) then
    if Problem is EWarningException then begin
      WriteLn(Log, Problem.Message);
      Problem.Free;
    end else
      raise Problem;
  if BeginCalculation(True) then
    Exit;
  PushProgress('Fitting');
  try
    PushProgress('Loading');
    try
      ObservationsLoader(Source, ObsType, TakeWeight, False);
    finally
      PopProgress;
    end;
    if CheckDerivatives <> 0 then
      nCycles := 1
    else if nCycles < 0 then
      nCycles := Mixture.FitCycles;
    if not JustResiduals then
      for i := 1 to Max(nCycles,1) do begin
        if nCycles > 0 then
          Mixture.PrepareForFit(ftLine, nil);
        if Mixture.PrintLevel > plMinimal then
          Mixture.PrettyPrint(Worker.Log, Mixture.PrintLevel);
        try
          SetProgress(i, Max(nCycles,1));
          if nCycles > 0 then begin
            FitStatus := FitLines(Mixture.FittingParameters, Observations, Mixture.Precision, Mixture, CheckDerivatives, OldSD, NewSD);
            Mixture.PushParameters(ftLine, True);
            { Save current set for redo fit }
          end else begin
            NoFittingParameters := nil;
            FitStatus := FitLines(NoFittingParameters, Observations, Mixture.Precision, Mixture, CheckDerivatives, OldSD, NewSD);
          end;
        except
          Mixture.DiscardFits(1);
          raise;
        end;
      end
    else
      FitStatus := '';
    ResidualForm.PlotResiduals(FitStatus);
  finally
    EndCalculation;
    PopProgress;
  end;
  ConsiderReplot;
end;

{ Menu Handling }

procedure TMainForm.AddSampleMixture(SampleProc: TSampleProc);
var
  Mix: TMixture;
begin
  if OKToChange(True, False) then
  begin
    Mix := TMixture.Create(nil);
    if Assigned(SampleProc) then
      SampleProc(Mix);
    SetMixture(Mix);
    SetName('');
    if Assigned(SampleProc) then
    begin
      DoPlot;
      if Mix.Simulations[0].Fmin = 0 then
        AllButtonClick(nil);
    end;
    Mixture.Modified := False;
  end;
end;

procedure TMainForm.EditMenuClick(Sender: TObject);
begin
  CopyTopPlotItem.Enabled := Split.Visible;
  TopPlotItem.Enabled := Split.Visible;
end;

procedure TMainForm.EmptyMixtureItemClick(Sender: TObject);
begin
  AddSampleMixture(nil);
end;

procedure TMainForm.NewLinearMoleculeItemClick(Sender: TObject);
begin
  AddSampleMixture({$IFDEF FPC}@{$ENDIF}AddSampleLinearMolecule);
end;

procedure TMainForm.AsymmetricTopItemClick(Sender: TObject);
begin
  AddSampleMixture({$IFDEF FPC}@{$ENDIF}AddSampleAsymmetricTop);
end;

procedure TMainForm.AutofitItemClick(Sender: TObject);
var
  AutoFitter: TMNode;
begin
  ReShow(AutoFitForm);
  AutoFitter := GlobalMixture.ChildWithClass(TAutoFitter, 0);
  if AutoFitter = nil then begin
    AutoFitter := TAutoFitter.Create(GlobalMixture);
    AutoFitter.Name := 'AutoFitter';
  end;
  AutoFitForm.ShowItem(AutoFitter);
end;

procedure TMainForm.SymmetricTopItemClick(Sender: TObject);
begin
  AddSampleMixture({$IFDEF FPC}@{$ENDIF}AddSampleSymmetricTop);
end;

procedure TMainForm.VibrationalSpectrumItemClick(Sender: TObject);
begin
  AddSampleMixture({$IFDEF FPC}@{$ENDIF}AddSampleVibratingMolecule);
end;

procedure TMainForm.OpenItemClick(Sender: TObject);
var
  i: Integer;
  Files: array of string;
begin
  if CloseQuery and OpenDialog.Execute then
  begin
    SetLength(Files, OpenDialog.Files.Count);
    for i := 0 to OpenDialog.Files.Count - 1 do
      Files[i] := OpenDialog.Files[i];
    FormDropFiles(nil, Files);
  end;
end;

procedure TMainForm.OldPgopherItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.dat';
  ImportDialog.FilterIndex := 1;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportPgopherFile);
end;

procedure TMainForm.OldD3hItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.dat';
  ImportDialog.FilterIndex := 1;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportD3hFile);
end;

procedure TMainForm.CALPGMfileItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.par';
  ImportDialog.FilterIndex := 2;
  ImportDialog.Filter := StringReplace(ImportDialog.Filter, 'HITRAN', 'SPFIT/SPCAT', []);
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportCALPGMFile);
end;

procedure TMainForm.HitranFileItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.par';
  ImportDialog.FilterIndex := 2;
  ImportDialog.Filter := StringReplace(ImportDialog.Filter, 'SPFIT/SPCAT', 'HITRAN', []);
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportHITRANFile);
end;

procedure TMainForm.ExoMolFileItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.states';
  ImportDialog.FilterIndex := 8;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportExoMolFile);
end;

procedure TMainForm.SpecViewfileItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.svw';
  ImportDialog.FilterIndex := 3;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportSpecViewFile);
end;

procedure TMainForm.JB95FileItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.jbw';
  ImportDialog.FilterIndex := 4;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportJB95File);
end;

procedure TMainForm.STROTABFileItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.sta';
  ImportDialog.FilterIndex := 5;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportSTROTABFile);
end;

procedure TMainForm.LmatrixImportITEMClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.out';
  ImportDialog.FilterIndex := 6;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportLmatrixFile);
end;

procedure TMainForm.VIBCAImportItemClick(Sender: TObject);
begin
  ImportDialog.DefaultExt := '.vib';
  ImportDialog.FilterIndex := 7;
  ImportFile({$IFDEF FPC}@{$ENDIF}ImportVIBCAFile);
end;

procedure TMainForm.BinaryfileItemClick(Sender: TObject);
var
  ExperimenalPlot: TExperimentalPlot;
  Stream: TStream;
  BinaryFileLoader: TBinaryFileLoader;
  DefaultIntensityUnits: TIntensityUnits;
begin
  if not MergeItem.Checked then
  begin
    if not ClearOverlays(False) then
      Exit;
    Overlays := TOverlays.Create(nil);
    PerformMNodeNotification(Overlays, mnNewBaseNode, 0);
  end;
  ImportDialog.DefaultExt := '.dat';
  ImportDialog.FilterIndex := 1;
  if ImportDialog.Execute then
  begin
    Stream := TFileStream.Create(ImportDialog.FileName, fmOpenRead or DefaultShareMode);
    try
      ExperimenalPlot := TExperimentalPlot.Create(Overlays);
      try
        ExperimenalPlot.Name := ExtractFileName(ImportDialog.FileName);
        ExperimenalPlot.Comment := 'Imported from ' + ImportDialog.FileName;
        if DefaulttoArbitraryUnitsItem.Checked then
          DefaultIntensityUnits := Arbitrary
        else
          DefaultIntensityUnits := Normalized;
        BeginGlobalUpdate;
        BinaryFileLoader := nil;
        try
          BinaryFileLoader := TBinaryFileLoader.Create
            (TPGOPHERPlotLoaderConnector.Create(ExperimenalPlot,
            ImportDialog.FileName), Stream, '', Simulation.PlotUnits,
            DefaultIntensityUnits);
          if not BinaryFileLoader.DoLoad(Worker.Log, False) then
            FreeAndNil(ExperimenalPlot);
        finally
          BinaryFileLoader.Free;
          EndGlobalUpdate(True);
        end;
      except
        ExperimenalPlot.Free;
      end;
    finally
      Stream.Free;
    end;
  end;
end;

procedure TMainForm.ExportLineListItemClick(Sender: TObject);
var
  f: TextFile;
  i: Integer;
  Threshold: Double;
var
  Problem: Exception;
  ASimulation: TSimulation;
  PrintLevel: TPrintLevel;
  SS: TShower;
begin
  if not Assigned(Mixture) then
    Exit;
  Problem := nil;
  ASimulation := Mixture.Simulations[0];
  if not Assigned(Mixture.LineList) then
    ASimulation.MakeLineList(Problem);
  if not Assigned(Mixture.LineList) then
    Exit;
  if Mixture.Name <> '' then
    ExportLineListDialog.FileName := ChangeFileExt(Mixture.Name, '');
  if not ExportLineListDialog.Execute then
    Exit;
  LineExportOptions.SetDefaultType(ExportLineListDialog.FileName);
  if LineExportOptions.ShowModal = mrOK then begin
    PrintLevel := LineExportOptions.PrintLevel;
    AssignFile(f, ExportLineListDialog.FileName);
    try
      ReWrite(f);
      if PrintLevel = plLineList then
        Mixture.PrettyPrint(f, PrintLevel);
      if Assigned(Problem) then
        WriteLn(f, Problem.Message);
      with Mixture.LineList do begin
        PushProgress('Line List output', nLines);
        Threshold := LargestIntensity * LineExportOptions.ThresholdSpin.Value;
        SetShower(f, PrintLevel);
        for i := 0 to nLines - 1 do begin
          ShowLine(i, Threshold, Mixture.Units, Mixture.Simulations[0].PlotUnits,
                   LineExportOptions.FitFormatCheck.Checked, Problem);
          if i mod 100 = 0 then
            SetProgress(i);
        end;
        ShowFooter;
        PopProgress;
      end;
    finally
      CloseFile(f);
    end;
  end;
  Problem.Free;
end;

procedure TMainForm.ExportPlotItemClick(Sender: TObject);
begin
  G.ExecuteExport;
end;

procedure TMainForm.TopPlotItemClick(Sender: TObject);
begin
  OC.ExecuteExport;
end;

procedure TMainForm.LoadOverlayItemClick(Sender: TObject);
var
  i: Integer;
begin
  OpenOverlayDialog.Filter :=
    'Data File (*.dat)|*.dat|Overlay File (*.ovr)|*.ovr|' + PlotLoaderFilter
  // +GraphicFilter(TGraphic)+'|'
    + 'All Files (*.*)|' + AllFilesMask;
  if OpenOverlayDialog.Execute then begin
    for i := 0 to OpenOverlayDialog.Files.Count - 1 do
      AddOverlay(OpenOverlayDialog.Files[i], MergeItem.Checked or (i > 0) or (Sender = nil),
        Simulation.PlotUnits);
    ConsiderReplot(Mixture.nSpecies = 0);
  end;
end;

procedure TMainForm.AddOverlayItemClick(Sender: TObject);
begin
  LoadOverlayItemClick(nil);
end;

procedure TMainForm.SaveItemClick(Sender: TObject);
begin
  if MixtureName = '' then
    SaveAsItemClick(nil)
  else
    SaveMixtureTo(MixtureName);
end;

procedure TMainForm.SaveMixtureTo(AFilename: string);
var
  SaveSetingsInfo: TSaveSettingsInfo;
begin
  Mixture.ClearFormSettings;
  if SavePositionsItem.Checked or SaveSettingsItem.Checked then begin
    with SaveSetingsInfo do begin
      DOMDocument := Mixture.LoadedFormSettings;
      SavePositions := SavePositionsItem.Checked;
      SaveSettings := SaveSettingsItem.Checked;
    end;
    ForAllSettingsForms({$IFDEF FPC}@{$ENDIF}SaveSettings,
      TObject(@SaveSetingsInfo));
  end;
  Mixture.SaveToFile(AFilename);
  Mixture.ClearFormSettings;
end;

procedure TMainForm.SaveOverlayAsItemClick(Sender: TObject);
begin
  OverlayForm.SaveAsButtonClick(Overlays);
end;

procedure TMainForm.SaveOverlayItemClick(Sender: TObject);
begin
  OverlayForm.SaveButtonClick(Overlays);
end;

procedure TMainForm.SaveArrangementItemClick(Sender: TObject);
begin
  BeginGlobalUpdate;
  try
    ForAllSettingsForms({$IFDEF FPC}@{$ENDIF}ShowInMixture, nil);
  finally
    EndGlobalUpdate;
  end;
end;

procedure TMainForm.SaveAsItemClick(Sender: TObject);
begin
  if not Assigned(Mixture) then
    Exit;
  SaveDialog.FileName := MixtureName;
  if SaveDialog.Execute then begin
    if ExtractFileExt(SaveDialog.FileName) = '.ovr' then
      raise EReportException.Create
        ('.ovr files are used to save overlaid spectral data; please use Save Overlay menu item for these');
    SaveMixtureTo(SaveDialog.FileName);
    SetName(SaveDialog.FileName);
  end;
end;

procedure TMainForm.PrintItemClick(Sender: TObject);
begin
  GraphPrintPreview.Show;
end;

procedure TMainForm.PrintSetupItemClick(Sender: TObject);
begin
  GraphPrintPreview.PrinterSetup;
end;

procedure TMainForm.MRUFileListGetDisplayName(Sender: TObject;
  AFilename: string; var ADisplayName: string);
var
  P: Integer;
begin
  if ADisplayName[Length(ADisplayName)] = '>' then
    Exit; { Calibration or clipboard item }
  P := Pos(' ', ADisplayName);
  if P > 0 then
    ADisplayName := Copy(ADisplayName, 1, P)
  else
    ADisplayName := '';
  ADisplayName := ADisplayName + ExtractRelativepath(GetCurrentDir + PathDelim,
    AFilename);
end;

procedure TMainForm.MRUFileListMRUItemClick(Sender: TObject; AFilename: string);
var
  i: Integer;
begin
  if AFilename = '' then
    Exit;
  if not FileExists(AFilename) then begin
    if AFilename[1] = '<' then
      for i := 0 to CalibrationItem.Count - 1 do
        if '<' + StripHotkey(CalibrationItem[i].Caption) + ' Calibration>' = AFilename then begin
          CalibrationItem[i].Click;
          Exit;
        end;
    raise EReportException.Create(AFilename + ' does not exist');
  end;
  if LowerCase(ExtractFileExt(AFilename)) = '.pgo' then begin
    if OKToChange(True, False) then
      LoadFile(AFilename, MergeItem.Checked);
  end else
    LoadGeneralFile(AFilename, MergeItem.Checked, False, Simulation.PlotUnits);
  ConsiderReplot(Mixture.nSpecies = 0);
end;

procedure TMainForm.nCPUsItemCLick(Sender: TObject);
begin
  StartWorkers((Sender as TComponent).Tag, TPgopherWorker);
end;

procedure TMainForm.ExitItemClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.CopyPictureItemClick(Sender: TObject);
begin
  G.CopyToClipboard;
end;

procedure TMainForm.CopyTopPlotItemClick(Sender: TObject);
begin
  OC.CopyToClipboard;
end;

procedure TMainForm.ShowPartsItemClick(Sender: TObject);
begin
  ShowPartsButton.Down := not ShowPartsButton.Down;
  ChangeSimulationSetting(ShowPartsIndex, ShowPartsButton);
end;

procedure TMainForm.ShowSumItemClick(Sender: TObject);
begin
  ShowSumButton.Down := not ShowSumButton.Down;
  ChangeSimulationSetting(ShowSumIndex, ShowSumButton);
end;

procedure TMainForm.ShowFortratItemClick(Sender: TObject);
begin
  ShowFortratButton.Down := not ShowFortratButton.Down;
  ChangeSimulationSetting(ShowFortratIndex, ShowFortratButton);
end;

procedure TMainForm.ShowLinelistItemClick(Sender: TObject);
var
  SavedPrintLevel: TPrintLevel;
begin
  SavedPrintLevel := Mixture.PrintLevel;
  try
    if (SavedPrintLevel < plCSV) or (SavedPrintLevel = plDetail) then
      with Mixture do
        ModifySetting(IndexOfSetting('PrintLevel'),Ord(plLineList),False);
    Mixture.ClearCache(True);
    SimulateButtonClick(nil);
  finally
    with Mixture do
      ModifySetting(IndexOfSetting('PrintLevel'),Ord(SavedPrintLevel),False);
  end;
  LogItemClick(nil);
end;

procedure TMainForm.UpperStateItemClick(Sender: TObject);
begin
  Mixture.Simulations[0].UseUpper := not Mixture.Simulations[0].UseUpper;
  ConsiderReplot;
end;

procedure TMainForm.WavenumbersItemClick(Sender: TObject);
begin
  if not Assigned(Mixture) then
    Exit;
  Mixture.Simulations[0].ChangePlotUnits
    (TResonanceUnits((Sender as TComponent).Tag));
  UpdateToolbar;
  ConsiderReplot;
end;

procedure TMainForm.PasteOverlayItemClick(Sender: TObject);
begin
  if not MergeItem.Checked then
    if not ClearOverlays(True) then
      Exit;
  PasteAnotherItemClick(Sender);
end;

procedure TMainForm.PlotOptionsItemClick(Sender: TObject);
begin
  VariableCharSizeItem.Checked := not (grFixedCharSize in G.GOptions);
  LimitLineSegmentsItem.Checked := not (grUnlimitedSegmentLength in G.GOptions);
  ColouredItem.Checked := grColouredPens in G.GOptions;
end;

{$IFNDEF MSWINDOWS}
var
  text_uri_list: TClipboardFormat = 0;
{$ENDIF}

procedure TMainForm.PasteAnotherItemClick(Sender: TObject);
var
{$IFDEF MSWINDOWS}
  Files: TStringArray;
{$ELSE}
  Stream: TTextStream;
  Prefix: string;
  nFiles: Integer;
  s: string;
  Pass: Integer;
{$ENDIF}
  i: Integer;
  DefaultUnits: TResonanceUnits;
begin
  if not(Sender is TPlotForm) then
    Sender := Self;
  if Assigned(Sender) then
    DefaultUnits := (Sender as TPlotForm).Simulation.PlotUnits
  else
    DefaultUnits := cm1;
  {$IFDEF MSWINDOWS}
    if Clipboard.HasFormat(CF_HDROP) then begin
      Clipboard.Open;
      try
  {$IFDEF FPC}
        Windows.GetClipboardData(CF_HDROP);
  {$ELSE}
        Files := ConvertHDROP(Clipboard.GetAsHandle(CF_HDROP));
  {$ENDIF}
      finally
        Clipboard.Close;
      end;
      for i := 0 to High(Files) do
        LoadGeneralFile(Files[i], True, True, DefaultUnits);
      ConsiderReplot(GlobalMixture.nSpecies = 0);
      Exit;
    end;
  {$ELSE}
    if text_uri_list = 0 then
      text_uri_list := RegisterClipboardFormat('text/uri-list');
    Stream := nil;
    try
      Prefix := '';
      if Clipboard.HasFormat(text_uri_list) then begin
        Stream := TTextStream.CreateFromClipboard(text_uri_list);
        Prefix := 'file://';
      end else if Clipboard.HasFormat(CF_Text) then begin
        Stream := TTextStream.CreateFromClipboard(CF_PREFTEXT);
        if Stream.Size > 2500 then
          FreeAndNil(Stream); { Don't bother if it looks like too much text }
      end;
      if (Stream <> nil) and (Stream.Size > 0) then begin
        for Pass := 1 to 2 do begin
          nFiles := 0;
          Stream.Position := 0;
          while not Stream.AtEndOfFile do begin
            s := Trim(Stream.RestOfLine);
            if s <> '' then begin
              nFiles := -1; { Assume fail }
              if Prefix <> '' then
                if Pos(Prefix, s) = 1 then
                  Delete(s,1,Length(Prefix))
                else
                  Break;
              if (s = '') or (s[1] <> DirectorySeparator) or not(FileExists(s)) then
                Break;
              nFiles := 1;
              if Pass = 2 then
                LoadGeneralFile(Trim(s), True, True, DefaultUnits);
            end;
            Stream.NextLine;
          end;
          if nFiles < 1 then
            Break; { Not a list of files }
        end;
      end;
      if nFiles = 1 then begin
        ConsiderReplot(GlobalMixture.nSpecies = 0);
        Exit;
      end;
    finally
      Stream.Free;
    end;
  {$ENDIF}
  AddOverlay('', True, DefaultUnits);
  ShowSimulation;
end;

procedure TMainForm.SelectRegionItemClick(Sender: TObject);
var
  i: Integer;
begin
  if not Assigned(Overlays) or not LastGraph.AreaSelected then
    Exit;
  for i := 0 to Overlays.nPlots - 1 do
    Overlays[i].SelectFrequencyRange(LastGraph.SelectedMathArea.Left,
      LastGraph.SelectedMathArea.Right);
  ConsiderReplot(False, False);
end;

procedure TMainForm.ClearSelectionItemClick(Sender: TObject);
var
  i: Integer;
begin
  if not Assigned(Overlays) then
    Exit;
  for i := 0 to Overlays.nPlots - 1 do
    Overlays[i].ClearSelection;
  ConsiderReplot(False, False);
end;

procedure TMainForm.InvertItemClick(Sender: TObject);
begin
  if Assigned(Overlays) then
    Overlays.Invert;
  ConsiderReplot(False, False);
end;

procedure TMainForm.ClearOverlayItemClick(Sender: TObject);
begin
  if ClearOverlays(True) then
    ShowSimulation;
end;

procedure TMainForm.OverlayPropertiesItemClick(Sender: TObject);
begin
  ReShow(OverlayForm);
  OverlayForm.BaseNode := Overlays;
end;

var
  CalculatorText: string;

procedure TMainForm.CalculatorItemClick(Sender: TObject);
var
  Value: Double;
  Message: string;
begin
  if InputQuery('Calculator', 'Expression', CalculatorText) and (Trim(CalculatorText) <> '') then begin
    Message := MNodeExpr.EvaluateExpression(CalculatorText, GlobalMixture, Value);
    if Message <> '' then
      ShowInfo('Bad expression:'+CalculatorText+' - '+Message)
    else
      ShowInfo(CalculatorText+' = '+FloatToStr(Value));
  end;
end;

procedure TMainForm.CalibrateItemClick(Sender: TObject);
begin
  ReShow(CalForm);
  CalForm.BaseNode := Overlays;
  if Sender <> nil then
    ConsiderReplot(False, False);
end;

procedure TMainForm.DeleteLastPeakItemClick(Sender: TObject);
var
  Plot: TPlot;
begin
  Plot := LastPlot(True);
  if Assigned(Plot) then begin
    Plot.Peaks.DeletePeak(Plot.nPeaks - 1);
    ConsiderReplot(False, False);
  end;
end;

procedure TMainForm.ClearAllPeaksItemClick(Sender: TObject);
var
  i: Integer;
begin
  ClearObservations;
  if Assigned(Overlays) then
    for i := 0 to Overlays.nPlots - 1 do
      Overlays[i].Peaks.Clear;
  if Assigned(Sender) then
    ConsiderReplot(False, False);
end;

procedure TMainForm.ConstantsItemClick(Sender: TObject);
begin
  if not Assigned(Mixture) then
    Exit;
  ConstantsForm.BaseNode := Mixture;
  ReShow(ConstantsForm);
end;

procedure TMainForm.LineListItemClick(Sender: TObject);
begin
  ReShow(LineForm);
end;

procedure TMainForm.LineWidthItemClick(Sender: TObject);
var
  WidthUpdated: Boolean;
  i: Integer;
begin
  if not Assigned(Mixture) or Mixture.AutoReplot then
    WidthUpdated := WidthDialog(G, {$IFDEF FPC}@{$ENDIF}SimulateButtonClick)
  else
    WidthUpdated := WidthDialog(G, nil);
  if not WidthUpdated then
    Exit;
  for i := 0 to High(PlotForms) do
    if Assigned(PlotForms[i]) then begin
      PlotForms[i].G.GOptions := MainForm.G.GOptions;
      PlotForms[i].OC.GOptions := MainForm.G.GOptions;
      PlotForms[i].G.LineStyles[0].Width := MainForm.G.LineStyles[0].Width;
      PlotForms[i].OC.LineStyles[0].Width := MainForm.G.LineStyles[0].Width;
      PlotForms[i].G.LineStyles[1].Width := MainForm.G.LineStyles[0].Width;
      //PlotForms[i].G.LineStyles[4].Width := MainForm.G.LineStyles[0].Width;
    end;
  ConsiderReplot;
end;

procedure TMainForm.LogItemClick(Sender: TObject);
begin
  ReShow(LogForm);
end;

procedure TMainForm.StatesItemClick(Sender: TObject);
begin
  ReShow(StatesForm);
  StatesForm.BaseNode := Mixture;
end;

procedure TMainForm.TransitionsItemClick(Sender: TObject);
begin
  ReShow(TransitionForm);
  TransitionForm.BaseNode := Mixture;
end;

procedure TMainForm.LevelsItemClick(Sender: TObject);
begin
  ReShow(LevelForm);
  LevelForm.BaseNode := Mixture;
end;

procedure TMainForm.SpectrumItemClick(Sender: TObject);
var
  PlotForm: TPlotForm;
begin
  if not Assigned(Mixture) then
    Exit;
  PlotForm := MakePlotForm(nil);
  PlotForm.ShowSimulation(Mixture, Overlays);
end;

procedure TMainForm.TimingItemClick(Sender: TObject);
begin
  WriteTimingInfo(Worker.Log);
  ReShow(ProfilerForm);
end;

{$IFDEF SHOWMEMUSAGE }

procedure TMainForm.MemoryItemCLick(Sender: TObject);
begin
  {$IFDEF RecyclerMM}
    RMMSnapShot.Show;
  {$ENDIF}
  ShowFastMMUsageTracker;
end;
{$ENDIF}

procedure TMainForm.ClearLogItemClick(Sender: TObject);
begin
  LogForm.ClearButtonClick(nil);
end;

procedure TMainForm.ClearCacheItemClick(Sender: TObject);
begin
  Mixture.SetCachedInformationPresent; { Force clear }
  Mixture.ClearCache(True);
  if Assigned(Overlays) then begin
    Overlays.SetCachedInformationPresent;
    Overlays.ClearCache(True);
  end;
end;

procedure TMainForm.ClearObservedItemClick(Sender: TObject);
begin
  ClearObservations;
  ConsiderReplot(False, False);
end;

procedure TMainForm.IndexItemClick(Sender: TObject);
begin
{$IFDEF VCL_}
  Application.HelpSystem.ShowTableOfContents;
{$ELSE}
  Application.HelpKeyword('Index');
{$ENDIF}
end;

procedure TMainForm.SearchItemClick(Sender: TObject);
begin
{$IFDEF VCL_}
  PHtmlHelpViewer.ShowHtmlSearch;
{$ELSE}
  Application.HelpKeyword('aindex');
{$ENDIF}
end;

procedure TMainForm.ProceduresItemClick(Sender: TObject);
begin
  Application.HelpKeyword('Procedures');
end;

procedure TMainForm.ResidualsItemClick(Sender: TObject);
begin
  ReShow(ResidualForm);
end;

procedure TMainForm.WhatIsItemClick(Sender: TObject);
begin
  SetInHelp(True);
end;

procedure TMainForm.WindowsonTopItemClick(Sender: TObject);
begin
  SetChildrenOnTop(WindowsonTopItem.Checked);
end;

procedure TMainForm.UsingHelpItemClick(Sender: TObject);
begin
  Application.HelpKeyword('Using Help');
end;

procedure TMainForm.VariableCharSizeItemClick(Sender: TObject);
begin
  if grFixedCharSize in G.GOptions then
    G.GOptions := G.GOptions - [grFixedCharSize]
  else
    G.GOptions := G.GOptions + [grFixedCharSize];
  OC.GOptions := G.GOptions;
end;

procedure TMainForm.LimitLineSegmentsItemClick(Sender: TObject);
begin
  if grUnlimitedSegmentLength in G.GOptions then
    G.GOptions := G.GOptions - [grUnlimitedSegmentLength]
  else
    G.GOptions := G.GOptions + [grUnlimitedSegmentLength];
  OC.GOptions := G.GOptions;
end;

procedure TMainForm.AboutItemClick(Sender: TObject);
begin
  Project.Execute(Global.VersionInfo);
end;

procedure TMainForm.DumpItemClick(Sender: TObject);
begin
{$IFDEF HELPHELP} MakeSkeletonHelp(Screen.ActiveForm); {$ENDIF}
end;

procedure TMainForm.ShowSumButtonClick(Sender: TObject);
begin
  ChangeSimulationSetting(TComponent(Sender).Tag, Sender);
end;

function TMainForm.MakePlotForm(ASimulation: TSimulation; ATag: Integer): TPlotForm;
begin
  Result := TPlotForm.Create(Application);
  if ATag >=0 then
    Result.Tag := ATag;
  if Assigned(ASimulation) then
    Result.Simulation := ASimulation
  else
    Result.InitializeFromSimulation(Simulation);
  Simulation.Tag := Result.Tag;
  Result.Caption := IntToStr(Result.Tag) + ' - ' + Caption;
  Result.Show;
end;

function TMainForm.MakePlotForms: Integer;
var
  i, j, MaxTag: Integer;
  PlotForm: TPlotForm;
begin
  MaxTag := -1;
  nPlotForms := 0;
  Tag := 0;
  for i := 0 to Screen.FormCount - 1 do
    if Screen.Forms[i] is TPlotForm then begin
      PlotForm := TPlotForm(Screen.Forms[i]);
      repeat
        SetLength(PlotForms, Max(PlotForm.Tag+1,Length(PlotForms)));
        for j := MaxTag+1 to PlotForm.Tag do
          PlotForms[j] := nil;
        MaxTag := Max(MaxTag, PlotForm.Tag);
        if not Assigned(PlotForms[PlotForm.Tag]) then
          Break;
        PlotForm.Tag := PlotForm.Tag + 1;
      until False;
      PlotForms[PlotForm.Tag] := PlotForm;
      Inc(nPlotForms);
    end;
  SetLength(PlotForms, MaxTag+1);
  Result := nPlotForms;
end;

procedure TMainForm.MNodeChange(Sender: TMNode; Change: TMNodeChange;
  Index: Integer);
begin
  if (Change = mnPropNameChange) and Assigned(Mixture) then
    Mixture.ClearUndo;
  if Change = mnDestroy then begin
    if Assigned(Mixture) then
      Mixture.ClearUndo;
    ClearObservations;
  end;
  inherited;
  if Sender = nil then
    Exit;
  if Sender = Mixture then begin
    if (Change = mnSettingChange) and
      (Index = Mixture.IndexOfSetting('QuantumNumberFormat')) then
      SetGlobalQuantumNumberFormat(Mixture.QuantumNumberFormat);
    if Change = mnNewBaseNode then
      UpdateToolbar;
  end;
  if not Assigned(FSimulation) or (Sender <> Simulation) then
    Exit;
  if ((Change = mnParameterChange) and (Index in [TemperatureIndex,
    GaussianIndex, LorentzianIndex, FoffsetIndex])) or
    (Change = mnSettingChange) and
    (Index in [ShowSumIndex, ShowPartsIndex, ShowFortratIndex,
    ShowFortratIndex + 1]) then
    UpdateToolbar;
end;

procedure TMainForm.MagnifyPlotItemClick(Sender: TObject);
begin
  if LastGraph.AreaSelected then
    LastGraph.ZoomSelected(True);
end;

procedure TMainForm.ShrinkplotItemClick(Sender: TObject);
begin
  LastGraph.ZoomOut;
end;

procedure TMainForm.SinglePlotItemClick(Sender: TObject);
var
  i: Integer;
begin
  for i := 1 to MakePlotForms-1 do
    PlotForms[i].Free;
end;

procedure TMainForm.ResetZoomItemClick(Sender: TObject);
begin
  LastGraph.ResetZoom;
end;

end.
