unit OverlayCalcFrm;
{$I utils.inc}{Copyright (C) 2015 C M Western; see commandline.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms, Dialogs, Split,
  ScrlCtrl, GraphI, GraphM, Graphs, ExtCtrls, MNode, ChildForm, Overlay,
  Math, MathSup, ExpPlot, ComCtrls, StdCtrls, Gendlg, Expr, pgDOM, AlEdit,
  EditExtr, NewSpin, Grids, NGrid, TextStrm, Buttons, Menus, SizeHelpers;

type

  { TBaselineForm }

  TOverlayCalcForm = class(TChildForm)
    Panel: TPanel;
    GPanel: TPanel;
    Residual: TGraphD;
    G: TGraphD;
    Split: TSplit;
    StatusBar: TStatusBar;
    ActiveObject: TComboBox;
    FirstSpin: TSpinBox;
    Label1: TLabel;
    Label2: TLabel;
    LastSpin: TSpinBox;
    NoiseMult: TFixedSpin;
    BaselineWindow: TSpinBox;
    Label4: TLabel;
    Label5: TLabel;
    Label7: TLabel;
    NoiseWindow: TSpinBox;
    BaselineButton: TButton;
    LeftButton: TSpeedButton;
    RightButton: TSpeedButton;
    CompressButton: TSpeedButton;
    ExpandButton: TSpeedButton;
    XZoomButton: TSpeedButton;
    AllButton: TSpeedButton;
    ScaleFromMinButton: TSpeedButton;
    ScaleFromMaxButton: TSpeedButton;
    TogleAutoYScaleButton: TSpeedButton;
    ToolsButton: TSpeedButton;
    ToolsPopupMenu: TPopupMenu;
    LoadAllItem: TMenuItem;
    LoadVisibleItem: TMenuItem;
    OverwriteItem: TMenuItem;
    ApplytoNewItem: TMenuItem;
    VerboseItem: TMenuItem;
    DenseItem: TMenuItem;
    AlternateAreaItem: TMenuItem;
    SelectVisibleItem: TMenuItem;
    N2: TMenuItem;
    LiveUpdateItem: TMenuItem;
    N3: TMenuItem;
    InvertedItem: TMenuItem;
    procedure FormResize(Sender: TObject);
    procedure GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure GMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; AX, AY: Integer);
    procedure GContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
    procedure FirstSpinChange(Sender: TObject);
    procedure LastSpinChange(Sender: TObject);
    procedure NoiseMultChange(Sender: TObject);
    procedure LeftButtonClick(Sender: TObject);
    procedure RightButtonClick(Sender: TObject);
    procedure CompressButtonClick(Sender: TObject);
    procedure ExpandButtonClick(Sender: TObject);
    procedure XZoomButtonClick(Sender: TObject);
    procedure AllButtonClick(Sender: TObject);
    procedure ScaleFromMinButtonClick(Sender: TObject);
    procedure ScaleFromMaxButtonClick(Sender: TObject);
    procedure TogleAutoYScaleButtonClick(Sender: TObject);
    procedure ToolsButtonClick(Sender: TObject);
    procedure LoadAllItemClick(Sender: TObject);
    procedure LoadVisibleItemClick(Sender: TObject);
    procedure SelectVisibleItemClick(Sender: TObject);
    procedure BaselineButtonClick(Sender: TObject);
  private
    { Layout }
    PanelControls: TControlArray;
  protected
    { Peak finding }
    LastPos, LastPeak: Integer;
    nPeaks: Integer;
    PeakFrequency, PeakIntensity: TDoubleArray;
    Verbose: Boolean;
    PeakThreshold: Double;
    CofMThreshold: Double;

    xMin, xMax, yMin, yMax, { Limits on complete data }
    hMin, hMax, gMin, gMax, { Limits on section plotted }
    rMin, rMax: Double;
    x, y, fBaseline: TDoubleArray;
    Plot: TExperimentalPlot;
    nPoints: Integer;
    Noise: Double;
    FirstPoint, LastPoint, ViewLeft, ViewRight, nVscales: Integer;
    ViewReversed, BaselineChanged: Boolean;
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    procedure ViewShifted; virtual;
    procedure MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer); override;
  public
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure StuffSelector(MNode: TMNode); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    procedure LoadData(const TheBaseLine: array of Double);  virtual;
    procedure ChangeSelection(First, Last: Integer);
    procedure StartPlot;
    procedure PlotData; virtual;
    procedure PlotUpper; virtual;
    procedure PlotVscale(Graph: TGraphD; Mi, Mx: Double);
    procedure UpdateStatus; virtual;
    procedure UpdateData;
    function MakeNewPlot(NewName: string; IsLineList: Boolean): TExperimentalPlot;
    function MethodStatus: string;
    procedure StartPeakScan;
    function MoveToNextPeak:Boolean;
  end;

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses StrUtils, UCompat, ShellSort, Global, FormUtils, Worker, PeakFinders, SlaveEdit,
     Main, Progress, Units, ApplyFrm;

procedure TOverlayCalcForm.LeftButtonClick(Sender: TObject);
var
  d: Integer;
begin
  d := ViewRight - ViewLeft;
  ViewLeft := ViewLeft - d div 2;
  ViewRight := ViewLeft + d;
  ViewShifted;
end;

procedure TOverlayCalcForm.RightButtonClick(Sender: TObject);
var
  d: Integer;
begin
  d := ViewRight - ViewLeft;
  ViewRight := ViewRight + d div 2;
  ViewLeft := ViewRight - d;
  ViewShifted;
end;

procedure TOverlayCalcForm.CompressButtonClick(Sender: TObject);
var
  d: Integer;
begin
  d := Max((ViewRight - ViewLeft) div 2,1);
  ViewLeft := ViewLeft-d;
  ViewRight := ViewRight+d;
  ViewShifted;
end;

procedure TOverlayCalcForm.ExpandButtonClick(Sender: TObject);
var
  d: Integer;
begin
  d := Max((ViewRight - ViewLeft) div 4,1);
  ViewLeft := ViewLeft + d;
  ViewRight := ViewRight - d;
  ViewShifted;
end;

procedure TOverlayCalcForm.AllButtonClick(Sender: TObject);
begin
  ViewLeft := 0;
  ViewRight := nPoints-1;
  nVscales := 0;
  UpdateStatus;
end;

procedure TOverlayCalcForm.XZoomButtonClick(Sender: TObject);
var
  Z: TGraphD;
  R: TDRect;
begin
  if G.SignificantAreaSelected then
    Z := G
  else if Residual.SignificantAreaSelected then
    Z := Residual
  else begin
    ExpandButtonClick(nil);
    Exit;
  end;
  R := Z.SelectedMathArea;
  ViewLeft := ClosestPoint(R.Left,x);
  ViewRight := ClosestPoint(R.Right,x);
  ViewShifted;
end;

procedure TOverlayCalcForm.TogleAutoYScaleButtonClick(Sender: TObject);
begin
  nVscales := 0;
  UpdateStatus;
end;

procedure TOverlayCalcForm.ToolsButtonClick(Sender: TObject);
begin
  ToolsPopupMenu.PopupComponent := ToolsButton;
  ToolsPopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TOverlayCalcForm.ScaleFromMaxButtonClick(Sender: TObject);
begin
  Dec(nVscales);
  UpdateStatus;
end;

procedure TOverlayCalcForm.ScaleFromMinButtonClick(Sender: TObject);
begin
  Inc(nVscales);
  UpdateStatus;
end;

procedure TOverlayCalcForm.BaselineButtonClick(Sender: TObject);
var
  i: Integer;
  Baseline: Double;
begin
  BaselineChanged := True;
  Noise := EstimateNoise(y, Baseline, NoiseWindow.Value, 2, NoiseMult.Value);
  SetLength(fBaseline, nPoints);
  for i := 0 to nPoints-1 do
    fBaseline[i] := BaseLine;
  StatusBar.Panels[1].Text := DtosM(Noise,10,-1);
  UpdateStatus;
end;

function TOverlayCalcForm.BaseNodeClass: TClass;
begin
  Result := TOverlays;
end;

procedure TOverlayCalcForm.ChangeSelection(First, Last: Integer);
begin
  { Requires Plot to be monotonic, but can be increasing or decreasing }
  if Plot.LocalFrequency(LastPoint) > Plot.LocalFrequency(FirstPoint) then begin
    LastPoint := Last + FirstPoint;
    FirstPoint := First + FirstPoint;
  end else begin
    FirstPoint := LastPoint - Last;
    LastPoint := FirstPoint + Last - First;
  end;
end;

procedure TOverlayCalcForm.NoiseMultChange(Sender: TObject);
begin
  UpdateStatus;
end;

procedure TOverlayCalcForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  {}
end;

procedure TOverlayCalcForm.FormResize(Sender: TObject);
begin
  Split.ParentResized;
  WrapPanel(Panel, PanelControls, -1, -1, 150);
  AutoSizeStatusBar(StatusBar, ['Noise:', '00000000000']);
end;

procedure TOverlayCalcForm.GContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
begin
  Handled := True;
end;

procedure TOverlayCalcForm.GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  s: string;
  GG: TGraphD;
  R: TDRect;
  P: Integer;
begin
  GG := Sender as TGraphD;
  s := ShowGraphInfo(GG, nil, nil);
  P := ClosestPoint(GG.LastMouseX,Self.x);
  if ViewReversed then
    P := LastPoint - FirstPoint - P;
  s := s + Format(' Point: %d', [P+FirstPoint]);
  if GG.AreaSelected then begin
    R := GG.SelectedMathArea;
    s := s + Format(' [%d]', [ClosestPoint(R.Right,Self.x)-ClosestPoint(R.Left,Self.x)]);
  end;
  StatusBar.Panels[2].Text := s;
end;

procedure TOverlayCalcForm.GMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; AX, AY: Integer);
var
  s, e: Integer;
  PeakInfo: TPeakInfo;
begin
  if (Button = mbRight) and G.SignificantAreaSelected then begin
    s := ClosestPoint(G.SelectedMathArea.Left, Self.x);
    e := ClosestPoint(G.SelectedMathArea.Right, Self.x);
    InOrder(s, e);
    if e > s+1 then begin
      if FindCentreOfMass(PeakInfo, y, x, fBaseline, (Plot.Owner as TOverlays).PeakThreshold, s, e) then begin
        if VerboseItem.Checked then
          WriteLn(Log, PeakInfo.Position:10, PeakInfo.Centre:15:6, DTosMA(PeakInfo.Area,11,-1));
        with ActiveLinelist do begin
          if Length(Intensity) > 0 then begin
            SetLength(Intensity, Length(Intensity)+1);
            Intensity[High(Intensity)] := PeakInfo.Area;
          end;
          SetLength(Frequency, Length(Frequency)+1);
          Frequency[High(Frequency)] := PeakInfo.Centre;
          LastPoint := -1;
          Sort;
        end;
        MainForm.ConsiderReplot;
      end;
      Exit;
    end;
  end;
  if Button = mbRight then
    G.PopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

type

  TDataSorter = class(TShellSorter)
    F: TOverlayCalcForm;
    function InOrder(I,J:Integer):Boolean; override;
  end;

function TDataSorter.InOrder(I, J: Integer): Boolean;
begin
  Result := F.x[I] <= F.x[J];
  if Result then
    Exit;
  SwapValues(F.x[I], F.x[J], SizeOf(F.x[I]));
  SwapValues(F.y[I], F.y[J], SizeOf(F.y[I]));
end;

procedure TOverlayCalcForm.FirstSpinChange(Sender: TObject);
begin
  if not CanReStuff then
    Exit;
  FirstPoint := FirstSpin.Value;
  UpdateData;
end;

procedure TOverlayCalcForm.LastSpinChange(Sender: TObject);
begin
  if not CanReStuff then
    Exit;
  LastPoint := LastSpin.Value;
  UpdateData;
end;

procedure TOverlayCalcForm.LoadAllItemClick(Sender: TObject);
begin
  FirstPoint := Plot.FirstPoint;
  LastPoint := Plot.LastPoint;
  UpdateData;
end;

procedure TOverlayCalcForm.LoadVisibleItemClick(Sender: TObject);
var
  First, Last: Integer;
begin
  with MainForm do
    with Plot do begin
      IntensityAt(Convert(G.Xmin,Simulation.PlotUnits,Units),First);
      IntensityAt(Convert(G.Xmax,Simulation.PlotUnits,Units),Last);
      InOrder(First,Last);
    end;
  if Last - First > 1 then begin
    FirstPoint := First;
    LastPoint := Last;
    UpdateData;
  end;
end;

procedure TOverlayCalcForm.ViewShifted;
begin
  UpdateStatus;
end;

procedure TOverlayCalcForm.SelectVisibleItemClick(Sender: TObject);
begin
  if ViewReversed then
    FirstPoint := LastPoint - FirstPoint - ViewRight
  else
    FirstPoint := ViewLeft + FirstPoint;
  LastPoint := FirstPoint + ViewRight - ViewLeft;
  UpdateData;
end;

procedure TOverlayCalcForm.UpdateData;
begin
  if not Visible then
    Exit;
  LoadData([]);
  if LiveUpdateItem.Checked then
    BaselineButton.Click;
  UpdateStatus;
end;

procedure TOverlayCalcForm.LoadData(const TheBaseLine: array of Double);
var
  i, mxp: Integer;
  DataSorter: TDataSorter;
begin
  nVscales := 0;
  fBaseline := nil;
  BaselineChanged := True;
  if not Assigned(Plot) then
    Exit;
  DataSorter := TDataSorter.Create;
  BeginGlobalUpdate;
  try
    mxp := High(Plot.Frequency);
    FirstSpin.MaxValue := mxp-1;
    LastSpin.MaxValue := mxp;
    if FirstPoint >= mxp then
      Exit;
    FirstPoint := Max(FirstPoint, 0);
    LastPoint := Min(LastPoint, mxp);
    FirstSpin.Value := FirstPoint;
    LastSpin.Value := LastPoint;
    nPoints := LastPoint - FirstPoint + 1;
    if nPoints < 3 then
      Exit;
    ViewLeft := 0;
    ViewRight := nPoints-1;
    SetLength(x, nPoints);
    SetLength(y, nPoints);
    nPoints := 0;
    for i := 0 to High(x) do begin
      x[nPoints] := Plot.LocalFrequency(i+FirstPoint);
      y[nPoints] := Plot.AdjustIntensity(Plot.Intensity[i+FirstPoint]);
      if Length(TheBaseLine) > 0 then
        y[nPoints] := y[nPoints] - TheBaseLine[nPoints];
      if not IsNanOrInf(x[nPoints]) and not IsNanOrInf(y[nPoints]) then
        Inc(nPoints);
    end;
    SetLength(x, nPoints);
    SetLength(y, nPoints);
    ViewReversed := (nPoints > 1) and (x[0] > x[nPoints-1]);
    Limits(x, xMin, xMax);
    Limits(y, yMin, yMax);
    DataSorter.F := Self;
    DataSorter.Sort(nPoints);
  finally
    EndGlobalUpdate;
    DataSorter.Free;
  end;
  StatusBar.Panels[1].Text := '';
end;

procedure TOverlayCalcForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  {}
end;

procedure TOverlayCalcForm.ShowItem(MNode: TMNode);
begin
  G.Clear;
  Residual.Clear;
  if not Assigned(MNode) or not(MNode is TExperimentalPlot) then begin
    CurrentMNode := nil;
    Exit;
  end;
  if (BaseNode <> MNode) and (BaseNode <> MNode.Owner) then
    BaseNode := MNode.Owner as TMNode;
  Plot := MNode as TExperimentalPlot;
  CurrentMNode := Plot;
  Noise := 0;
  FirstPoint := 0;
  LastPoint := High(Plot.Frequency);
  UpdateData;
end;

function TOverlayCalcForm.MakeNewPlot(NewName: string; IsLineList: Boolean): TExperimentalPlot;
begin
  BeginGlobalUpdate;
  try
    Result := TExperimentalPlot.Create(Plot.Owner);
    Result.Name := NewName;
    Result.IsLineList := IsLineList;
    if not IsLineList then
      ActiveObject.AddItem(NewName, Result);
  finally
    FReStuffPending := False;
    EndGlobalUpdate;
  end;
end;

function TOverlayCalcForm.MethodStatus: string;
begin
  Result := LineEnding
            +'Baseline Window = '+IntToStr(BaselineWindow.Value)+' Noise Multiplier = '+DtosM(NoiseMult.Value,10,-1)
            +' Noise = '+DtosM(Noise,10,-1)+'; Noise Window = '+IntToStr(NoiseWindow.Value);
end;

procedure TOverlayCalcForm.MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer);
begin
  if (Sender is TPlot) and (Change = mnDestroy) and (Sender <> CurrentMNode) then begin
    { Defeat form reset - just update Active object }
    BeginGlobalUpdate;
    try
      StuffSelector(CurrentMNode);
      FReStuffPending := False;
    finally
      EndGlobalUpdate;
    end;
  end else
    inherited;
end;

function TOverlayCalcForm.StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer;
begin
  if (MNode is TExperimentalPlot) and not TExperimentalPlot(MNode).IsLineList then begin
    ActiveObject.AddItem(Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end else
    Result := 0;
end;

procedure TOverlayCalcForm.StuffSelector(MNode: TMNode);
begin
  StuffSelectorHelper(MNode,ActiveObject);
  FComboBoxUpdated := False;
end;

procedure TOverlayCalcForm.StartPlot;
begin
  InOrder(ViewLeft, ViewRight);
  ViewLeft := Max(0, ViewLeft);
  ViewRight := Min(nPoints-1, ViewRight);
  hMin := x[ViewLeft];
  hMax := x[ViewRight];
  if nVscales = 0 then
    Limits(y, gMin, gMax, ViewLeft, ViewRight);
  G.Start;
  PlotVscale(G, gMin, gMax);
  G.Scales('Frequency', 'I');
end;

procedure TOverlayCalcForm.PlotData;
var
  i: Integer;
  V: Double;
begin
  G.ClipPoint(x[ViewLeft], y[ViewLeft]);
  for i := ViewLeft+1 to ViewRight do
    G.ClipJoin(x[i], y[i]);
  if Length(fBaseline) > 0 then begin
    G.Style(3);
    G.ClipPoint(hMin, fBaseLine[ViewLeft]);
    for i := ViewLeft+1 to ViewRight do
      G.ClipJoin(x[i], fBaseLine[i]);
  end;
end;

procedure TOverlayCalcForm.PlotUpper;
begin
end;

procedure TOverlayCalcForm.PlotVscale(Graph: TGraphD; Mi, Mx: Double);
var
  i: Integer;
begin
  for i := nVscales to -1 do
    Mi := Mx - (Mx-Mi)/2;
  for i := 1 to nVscales do
    Mx := Mi + (Mx-Mi)/2;
  Graph.Map(hMin, hMax, Mi, Mx);
end;

procedure TOverlayCalcForm.UpdateStatus;
var
  i: Integer;
  V: Double;
begin
  if not Visible or (csDestroying in ComponentState) then
    Exit;
  if Length(fBaseline) > 0 then
    Split.MakeBothVisible
  else
    Split.OneWindowVisible(G);
  StartPlot;
  PlotData;
  G.Finish;
  PlotUpper;
end;

procedure TOverlayCalcForm.StartPeakScan;
begin
  LastPeak := -100;
  LastPos := 0;
  nPeaks := 0;
  PeakThreshold := Noise*NoiseMult.Value;
  Verbose := VerboseItem.Checked;
  CofMThreshold := (Plot.Owner as TOverlays).PeakThreshold;
end;

function TOverlayCalcForm.MoveToNextPeak: Boolean;
var
  Maxy, Last, Area: Double;
  p, Start, Edge, Extra: Integer;
  PastPeak: Boolean;
  PeakInfo: TPeakInfo;
begin
  Result := False;
  repeat
    repeat
      Start := LastPos;
      while (LastPos < nPoints) and (y[LastPos] < fBaseline[LastPos]+PeakThreshold) do
        Inc(LastPos);
      if LastPos >= nPoints then
        Exit;
      { Now on peak }
      Edge := LastPos;
      Maxy := y[LastPos];
      Area := Maxy - (fBaseline[LastPos]+PeakThreshold);
      p := LastPos+1;
      PastPeak := False;
      while p < nPoints do begin
        Area := Area + Max(y[p] - (fBaseline[LastPos]+PeakThreshold),0);
        if y[p] >= Maxy then begin
          Maxy := y[p]; { Still going up }
          PastPeak := False;
        end else if y[p] < fBaseline[p]+PeakThreshold then
          Break { Definitely off peak }
        else if y[p] < Maxy-PeakThreshold then begin
          if PastPeak and (y[p] > Last) then
            Break;
          PastPeak := True;
        end;
        Last := y[p];
        Inc(p);
      end;
      LastPos := p;
    until p-Edge > 1;
    Area := Area/(p-Edge-1);
  until Area/PeakThreshold > 0.5;
  Extra := Max(10,(p-Edge)*2); { Additional padding either side if possible }
  p := Min(p, High(y));
  Start := Edge - Min(Edge-Start,Extra);
  Extra := Min(p+Extra, High(y));
  while (p < Extra) and (y[p+1] < fBaseline[p+1]+PeakThreshold) do
    Inc(p);
  if FindCentreOfMass(PeakInfo, y, x, fBaseline, CofMThreshold, Start, p) then begin
    if Length(PeakFrequency) <= nPeaks then begin
      SetLength(PeakFrequency, Max(nPeaks*2, 100));
      SetLength(PeakIntensity, Length(PeakFrequency));
    end;
    if (AlternateAreaItem.Checked) and (PeakInfo.Area > 0) then
      Area := PeakInfo.Area;
    PeakFrequency[nPeaks] := PeakInfo.Centre;
    PeakIntensity[nPeaks] := Area;
    Inc(nPeaks);
    if Verbose then
      WriteLn(Log, nPeaks:8, PeakInfo.Position:10, Start:10, p:10, PeakInfo.Centre:15:6, DTosMA(PeakInfo.Area,11,-1), DTosMA(Area,11,-1));
    LastPeak := PeakInfo.Position;
  end;
  Result := True;
end;

end.


