unit ASCII85;

{$H+}

interface

uses
  Classes, SysUtils, PDFStream, PDFreader;

type

  { TASCII85DecompressionStream }

  TASCII85DecompressionStream = class(TPDFStream)
  private
    Raw: array [0..4] of Byte;
    Converted: array [0..3] of Byte;
    Next, Last: Integer;
    EofPending: Boolean;
    procedure FillBuffer;
  public
    constructor Create(Source: TStream; Params: TPDFDictionary);
    function NextByte: Byte; override;
  end;

implementation


{ TASCII85DecompressionStream }

procedure TASCII85DecompressionStream.FillBuffer;
var
  nRead, i: Integer;
  t: Cardinal;
  function NextNonSpace:Byte;
  begin
    repeat
      EofPending := (Stream.Read(Result,1) <> 1) or (Result = Ord('~'));
    until EofPending or (Result > Ord(' '));
  end;

begin
  Next := 0;
  Last := 3;
  if not EofPending then
    Raw[0] := NextNonSpace;
  if EofPending then begin
    Eof := True;
    Exit;
  end else if Raw[0] = Ord('z') then
    FillChar(Converted[0],4,0)
  else begin
    nRead := 1;
    while (nRead < 5) and not EofPending do begin
      Raw[nRead] := NextNonSpace;
      if not EofPending then
        Inc(nRead);
    end;
    Last := nRead-2;
    for i := nRead to 4 do
      Raw[i] := $21+84;
    t := 0;
    for i := 0 to 4 do
      t := t*85 + (Raw[i]-Cardinal($21));
    for i := 3 downto 0 do begin
      Converted[i] := Byte(t);
      t := t shr 8;
    end;
  end;
end;

constructor TASCII85DecompressionStream.Create(Source: TStream; Params: TPDFDictionary);
begin
  inherited Create(Source);
  Next := 1;
  Last := 0;
end;

function TASCII85DecompressionStream.NextByte: Byte;
begin
  while not Eof and (Next > Last) do
    FillBuffer;
  Result := Converted[Next];
  Inc(Next);
  //WriteLn(Result);
end;

end.

