unit LZW;
{$H+}
interface

uses Classes, PDFStream, PDFReader;

type

  TLZWDecompressionStream = class(TPDFStream)
  private
    Early: Integer;
    nBits, CodeLength: Integer;
    CurByte: Byte;
    CurMask: Cardinal;
    NewChar: AnsiChar;
    CurCode, NextFree, LastCode: Word;
    CodeTable: array [258..4095] of AnsiString;
    CodeLeft: Integer;
    First: Boolean;
  public
    constructor Create(Source: TStream; Params: TPDFDictionary);
    function NextCode:Word;
    function Read(var Buffer; Count: Longint): Longint; override;
    procedure ClearTable;
  end;


implementation

uses SysUtils;

{ TLZWDecompressionStream }

procedure TLZWDecompressionStream.ClearTable;
begin
  CodeLength := 9;
  CurMask := $01ff;
  NextFree := 258;
  LastCode := 0;
  First := True;
end;

constructor TLZWDecompressionStream.Create(Source: TStream; Params: TPDFDictionary);
begin
  inherited Create(Source);
  nBits := 0;
  ClearTable;
  CodeLeft := 0;
  Early := Params.FindNumber('EarlyChange',1);
end;

function TLZWDecompressionStream.NextCode: Word;
begin
  if nBits > 0 then
    Result := Word(CurByte shl (CodeLength-nBits))
  else begin
    Result := 0;
    nBits := 0;
  end;
  while nBits < CodeLength do begin
    if Stream.Read(CurByte,1) <> 1 then
      raise Exception.Create('PDF:Unexpected end of LZW stream');
    if CodeLength >= nBits+8 then
      Result := Result or (CurByte shl (CodeLength-nBits-8))
    else
      Result := Result or (CurByte shr (8-(CodeLength-nBits)));
    Inc(nBits,8);
  end;
  Dec(nBits, CodeLength);
  Result := Result and CurMask;
end;

function TLZWDecompressionStream.Read(var Buffer; Count: Integer): Longint;
var
  P: PAnsiChar;
  procedure Add(c: AnsiChar);
  begin
    P^ := c;
    Inc(P);
    Dec(Count);
    Inc(Result);
  end;
begin
  P := @Buffer;
  Result := 0;
  while not Eof and (Count > 0) do begin
    while (CodeLeft > 0) and (Count > 0) do begin
      Add(CodeTable[LastCode][Length(CodeTable[LastCode])-CodeLeft+1]);
      Dec(CodeLeft);
    end;
    if Count <= 0 then
      Break;
    CurCode := NextCode;
    case CurCode of
      256: ClearTable;
      257: Eof := True;
    else
      if CurCode < 256 then begin
        NewChar := AnsiChar(CurCode);
        Add(NewChar);
      end else if CurCode < NextFree then begin
        { Current code }
        CodeLeft := Length(CodeTable[CurCode]);
        NewChar := CodeTable[CurCode][1];
      end else if CurCode = NextFree then begin
        if LastCode < 256 then
          CodeLeft := 2
        else
          CodeLeft := Length(CodeTable[LastCode])+1;
      end else
        raise Exception.Create('PDF:Bad LZW stream');
      if First then
        First := False
      else begin
        if LastCode < 256 then
          CodeTable[NextFree] := AnsiChar(LastCode) + NewChar
        else
          CodeTable[NextFree] := CodeTable[LastCode] + NewChar;
        Inc(NextFree,1+Early);
        if (NextFree = 512) or (NextFree = 1024) or (NextFree = 2048) then begin
          Inc(CodeLength);
          CurMask := (1 shl CodeLength)-1;
        end;
        Dec(NextFree,Early);
      end;
      LastCode := CurCode;
    end;
  end;
end;

end.
