unit PeakFindFrm;
{$I utils.inc}{Copyright (C) 2015 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, BaselineFrm, Menus, ComCtrls, ScrlCtrl, GraphI, GraphM, Graphs,
  Split, StdCtrls, NewSpin, AlEdit, EditExtr, Buttons, ExtCtrls, Overlay, Grids,
  NGrid, ExpPlot, Expr, Gendlg, MathSup, TextStrm;

{$DEFINE SHOWDIAG}
type

  { TPeakFindForm }

  TPeakFindForm = class(TBaselineForm)
    PeakPanel: TPanel;
    Label8: TLabel;
    FitStatus: TLabel;
    Label11: TLabel;
    FitButton: TButton;
    MaxCyclesSpin: TSpinBox;
    Results: TNumberGrid;
    AdvancedButton: TButton;
    AssignButton: TButton;
    AutoCheckBox: TCheckBox;
    ResetButton: TButton;
    IncludeError: TCheckBox;
    LorentzianCheck: TCheckBox;
    AssignArea: TCheckBox;
    AssignWidth: TCheckBox;
    ThresholdSpin: TSpinBox;
    LogPeaks: TCheckBox;
    procedure ResetButtonClick(Sender: TObject);
    procedure AssignButtonClick(Sender: TObject);
    procedure FitButtonClick(Sender: TObject);
    procedure AdvancedButtonClick(Sender: TObject);
    procedure ThresholdSpinChange(Sender: TObject);
  private
    FDataSource: TInteractiveDataSource;
    function GetDataSource: TInteractiveDataSource;
  protected
    Centre, Area, Gaussian, Lorentzian, BaseLine: Double;
    d: TDoubleArray;
    procedure ViewShifted; override;
  public
    PeakValid: Boolean;
    {$IFDEF SHOWDIAG}
    LastStatus: string;
    DummyPlot: TExperimentalPlot;
    LogStream: TTextStream;
    LogList: TStringList;
    LogListPos, LogID: Integer;
    LastLoadedLogFile, LastLoadedInfo: string;
    HeaderWritten: Boolean;
    constructor Create(AOwner: TComponent); override;
    procedure StartFile(AFileName: string);
    function LoadNextEntry:Boolean;
    procedure TestNext(Sender: TObject);
    procedure TestPrev(Sender: TObject);
    procedure TestAll(Sender: TObject);
    procedure LoadLog(Sender: TObject);
    procedure LoadDir(Sender: TObject);
    procedure EditPeaklog(Sender: TObject);
    procedure UpdateLog(Sender: TObject);
    procedure ModTestCase(Sender: TObject);
    {$ENDIF}
    destructor Destroy; override;
    procedure ShowPeak(FoundPeak: TFoundPeak; ThePlot: TExperimentalPlot);
    procedure LoadData(const TheBaseLine: array of Double); override;
    function CofMAlgorithm:Boolean;
    procedure ClearStatus;
    procedure ShowCofMStatus;
    procedure UpdateStatus; override;
    procedure LogData(Status: string; LogFileName: string = ''; Info: string = ''; Force: Boolean = False);
    function FittedFunction(f: Double): Double;
    property DataSource: TInteractiveDataSource read GetDataSource write FDataSource;
  end;

  TPeakFindDataSource = class (TInteractiveDataSource)
  private
    FPeakFindForm: TPeakFindForm;
    procedure SetPeakFindForm(const Value: TPeakFindForm);
  public
    xVariable: TVariable;
    FitExpr: TExpr;
    AvgErr: Double;
    CyclesUsed: Integer;
    destructor Destroy; override;
    procedure AddIndependentVariables(Lex:TLex); override;
    function nRows: Integer; override;
    function GetData(r: Longint; var x, y, w: Double; SetVariableControlsState: Boolean = False; ExtraAxis: Integer = -1): Boolean; override;
    function Weighted: Boolean; override;
    function Caption: string; override;
    procedure SetFitExpr(Expr: TExpr; nPar: Integer); override;
    procedure UpdateVariables(FloatWidths: Boolean);
    function AutoFit(Verbose:Boolean = False):Boolean;
    function CycleFits(MaxCycles: Integer; Verbose: Boolean = False):Boolean;
    function BadWidths:Boolean;
    property PeakFindForm: TPeakFindForm read FPeakFindForm write SetPeakFindForm;
  end;

var
  PeakFindForm: TPeakFindForm;

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses Math, UCompat, Global, MNode, PeakFinders, DataSource, Worker, SlaveEdit,
     Main, StrUtils;

{ TPeakFindForm }

procedure TPeakFindForm.AdvancedButtonClick(Sender: TObject);
begin
  ReShow(DataSource.NLFitDlg);
end;

procedure TPeakFindForm.AssignButtonClick(Sender: TObject);
var
  E, P: Double;
begin
  if AssignArea.Checked then begin
    P := Area;
    E := (DataSource.DataSourceFitter as TInteractiveDataSourceFitter).GetStdDev('a');
  end else if AssignWidth.Checked then
    if LorentzianCheck.Checked then begin
      P := Lorentzian;
      E := (DataSource.DataSourceFitter as TInteractiveDataSourceFitter).GetStdDev('l');
    end else begin
      P := Gaussian;
      E := (DataSource.DataSourceFitter as TInteractiveDataSourceFitter).GetStdDev('g');
    end
  else begin
    P := Centre;
    E := (DataSource.DataSourceFitter as TInteractiveDataSourceFitter).GetStdDev('c');
  end;
  if not IncludeError.Checked then begin
    P := RoundToIncrement(P, E/10);
    E := -1;
  end;
  Plot.AddPeak(P, GlobalMixture.Simulations[0].PlotUnits, 10 { TODO }, E);
  MainForm.ConsiderReplot;
end;

procedure TPeakFindForm.ClearStatus;
begin
  Results.Clear;
  Results.AutoAdjustColumn(0);
end;

function TPeakFindForm.CofMAlgorithm: Boolean;
var
  PeakInfo: TPeakInfo;
  St: string;
begin
  Result := FindCentreOfMass(PeakInfo, y, x, [], ThresholdSpin.Value/100, 0, nPoints-1);
  Centre := PeakInfo.Centre;
  Area := PeakInfo.Area;
  Result := Result and (Area > 0);
  if LorentzianCheck.Checked then begin
    Lorentzian := PeakInfo.Width;
    Gaussian := 0;
  end else begin
    Lorentzian := 0;
    Gaussian := PeakInfo.Width;
  end;
  BaseLine := PeakInfo.Baseline;
  FitStatus.Caption := 'CofM';
  if not Result then
    St := 'zero area'
  else begin
    if (PeakInfo.nAbove = 1) and (yMax = y[0]) then
      St := 'single point at start'
    else if (PeakInfo.nAbove = 1) and (yMax = y[nPoints-1]) then
      St := 'single point at end'
    else
      St := '';
  end;
  Result := St = '';
  if Result then  
    FitStatus.Caption := 'OK; CofM Threshold = '+IntToStr(ThresholdSpin.Value) + '%'
  else
    FitStatus.Caption := 'CofM fail: '+St;
end;

destructor TPeakFindForm.Destroy;
begin
  FreeAndNil(FDataSource);
  {$IFDEF SHOWDIAG}
    FreeAndNil(DummyPlot);
    FreeAndNil(LogStream);
    FreeAndNil(LogList);
  {$ENDIF}
  inherited;
end;

procedure TPeakFindForm.FitButtonClick(Sender: TObject);
begin
  (DataSource as TPeakFindDataSource).AutoFit;
end;

function TPeakFindForm.FittedFunction(f: Double): Double;
begin
  (DataSource as TPeakFindDataSource).xVariable.Value := f;
  Result := DataSource.DataSourceFitter.Expr.Value;
end;

function TPeakFindForm.GetDataSource: TInteractiveDataSource;
begin
  if not Assigned(FDataSource) then begin
    FDataSource := TPeakFindDataSource.Create(nil);
    TPeakFindDataSource(FDataSource).PeakFindForm := Self;
  end;
  Result := FDataSource;
end;

procedure TPeakFindForm.LoadData(const TheBaseLine: array of Double);
begin
  inherited;
  SetLength(d, nPoints);
  if nPoints > 0 then
    FillChar(d[0], nPoints*SizeOf(Double), 0);
end;

procedure TPeakFindForm.LogData(Status, LogFileName, Info: string; Force: Boolean);
  procedure WriteInfo(var F: System.Text);
  var
    i: Integer;
  begin
    WriteLn(F, 'Peak,',Status,',',DateTimeToStr(Now));
    if Info = '' then
      WriteLn(F, FirstPoint,',',LastPoint,',',Plot.Name)
    else
      WriteLn(F, Info);
    WriteLn(F, 'Data,',Length(x));
    for i := 0 to High(x) do
      WriteLn(F, x[i],',',y[i]);
    for i := 1 to Results.RowCount-1 do
      WriteLn(F, Results.Lbl[i],',',Results.Value[i,1],',',Results.Value[i,2]);
    WriteLn(F, 'Threshold',ThresholdSpin.Value/100);
    WriteLn(F,'****');
  end;
{$IFDEF SHOWDIAG}
var
  LogFile: System.Text;
  BaseName: string;
  i: Integer;
{$ENDIF}
begin
 {$IFDEF SHOWDIAG}
  LastStatus := Status;
  if LogFileName = '' then begin
    if not Force and (not LogPeaks.Checked or (DummyPlot = Plot)) then
      Exit;
    LogFileName := ExtractFilePath(Application.ExeName) + 'peaks';
    if not DirectoryExists(LogFileName) then begin
      LogPeaks.Checked := False;
      Exit;
    end;
    LogFilename := LogFileName + PathDelim;
    if Status <> 'OK' then
      LogFilename := LogFileName + 'BAD';
    BaseName := LogFileName + FormatDateTime('yyyy-mm-dd-hh-nn-ss', Now);
    i := 0;
    repeat
      if i > 0 then
        LogFileName := BaseName + '-' + IntToStr(i)
      else
        LogFileName := BaseName;
      LogFileName := LogFileName +'.peaklog';
      Inc(i);
    until not FileExists(LogFilename);
  end;
  AssignFile(LogFile, LogFileName);
  if FileExists(LogFileName) then
    Append(LogFile)
  else
    Rewrite(LogFile);
  try
    WriteInfo(LogFile);
  finally
    System.Close(LogFile);
  end;
 {$ELSE}
  WriteInfo(Log);
 {$ENDIF}
end;

procedure TPeakFindForm.ResetButtonClick(Sender: TObject);
begin
  if Sender <> nil then
    SelectVisibleItemClick(nil)
  else
    UpdateData;
  CofMAlgorithm;
  ShowCofMStatus;
  (DataSource as TPeakFindDataSource).UpdateVariables(True);
  DataSource.NLFitDlg.PlotFunctionClick(nil);
  UpdateStatus;
end;

procedure TPeakFindForm.ShowCofMStatus;
begin
  ClearStatus;
  with Results do begin
    Value[1,1] := Centre;
    Value[2,1] := Area;
    Value[3,1] := Gaussian;
    Value[5,1] := BaseLine
  end;
  PeakValid := True;
  Results.AutoDefaultRowHeight;
end;

procedure TPeakFindForm.ShowPeak(FoundPeak: TFoundPeak; ThePlot: TExperimentalPlot);
begin
  BeginGlobalUpdate;
  try
    ClearStatus;
    BaseNode := ThePlot.Owner as TMNode;
    Plot := ThePlot;
    CurrentMNode := Plot;
    StuffSelector(BaseNode);
    Noise := 0;
    FirstPoint := FoundPeak.First;
    LastPoint := FoundPeak.Last;
    Centre := FoundPeak.Position;
    Results.Value[1,1] := Centre;
    PeakValid := False;
    ThresholdSpin.Value := Round(FoundPeak.Threshold*100);
    ResetButtonClick(nil);
    if AutoCheckBox.Checked then
      FitButtonClick(nil);
    FReStuffPending := False;
  finally
    EndGlobalUpdate;
  end;
end;

procedure TPeakFindForm.ThresholdSpinChange(Sender: TObject);
begin
  if (csLoading in ComponentState) or (BaseNode = nil) then
    Exit;
  CofMAlgorithm;
  ShowCofMStatus;
  UpdateStatus;
end;

procedure TPeakFindForm.UpdateStatus;
var
  i, nP: Integer;
  v, dMin, dMax: Double;
begin
  if not Visible or (csDestroying in ComponentState) then
    Exit;
  StartPlot;
  PlotData;
  G.Style(2);
  G.Point(Centre, yMin);
  G.Join(Centre, yMax);
  if PeakValid then begin
    { Fitted function }
    G.Style(3);
    nP := Max(G.Width-50,100);
    G.Point(xMin, FittedFunction(xMin));
    for i := 1 to nP do begin
      v := xMin + i*(xMax-xMin)/nP;
      G.Join(v, FittedFunction(v));
    end;
    { Residuals }
    Limits(d, dMin, dMax);
    Residual.Start;
    if dMax > dMin then begin
      Residual.Map(xMin, xMax, dMin, dMax);
      Residual.Axes('', 'Obs-Calc');
      for i := 0 to nPoints-1 do
        Residual.Symbol(x[i], d[i], 1);
    end;
    Residual.Finish;
  end;
  G.Finish;
end;

procedure TPeakFindForm.ViewShifted;
begin
  SelectVisibleItemClick(nil);
end;

{ TPeakFindDataSource }

procedure TPeakFindDataSource.AddIndependentVariables(Lex: TLex);
begin
  if not Assigned(xVariable) then
    xVariable := TVariable.Create('x');
  Lex.AddVariable(xVariable);
end;

function TPeakFindDataSource.AutoFit(Verbose:Boolean):Boolean;
var
  i: Integer;
begin
  Result := False;
  try
    CyclesUsed := 0;
    with DataSourceFitter as TInteractiveDataSourceFitter do begin
      UpdateVariables(False);
      SetFloat('c', False);
      CycleFits(2, Verbose); { Float area and baseline only }
      if PeakFindForm.nPoints <= 4 then begin
        Result := True; { Too few points to be useful }
      end else begin
        UpdateVariables(True);
        SetFloat('c', False);
        if CycleFits(PeakFindForm.MaxCyclesSpin.Value, Verbose) then begin
          if BadWidths then begin
            for i := 1 to CyclesUsed-2 do
              NLFitDlg.UndoClick(nil);
            UpdateVariables(False);
          end else
            UpdateVariables(True);
          SetFloat('c', True);
          if CycleFits(PeakFindForm.MaxCyclesSpin.Value, Verbose) then begin
            with PeakFindForm do
              Result := not BadWidths and (Centre >= xMin) and (Centre <= xMax);
          end;
        end;
      end;
    end;
  except
    on E: Exception do begin
      PeakFindForm.FitStatus.Caption := E.Message;
      PeakFindForm.LogData(E.Message);
      raise;
    end;
  end;
  if Result then begin
    if PeakFindForm.AutoCheckBox.Checked then
      PeakFindForm.AssignButtonClick(nil);
    PeakFindForm.FitStatus.Caption := 'OK'
  end else
    PeakFindForm.FitStatus.Caption := 'Fail';
  PeakFindForm.LogData(PeakFindForm.FitStatus.Caption);
end;

function TPeakFindDataSource.BadWidths: Boolean;
begin
  Result := (PeakFindForm.Lorentzian < 0) or (PeakFindForm.Gaussian < 0) or (PeakFindForm.Area < 0);
end;

function TPeakFindDataSource.Caption: string;
begin
  Result := PeakFindForm.CurrentMNode.Name;
end;

function TPeakFindDataSource.CycleFits(MaxCycles: Integer; Verbose: Boolean): Boolean;
var
  LastErr: Double;
begin
  LastErr := -1;
  Result := False;
  repeat
    Inc(CyclesUsed);
    NLFitDlg.FitOnceClick(nil);
    if Verbose then
      NLFitDlg.ShowFitInfo(Log);
    if BadWidths then
      Exit;
    if Abs(LastErr-AvgErr) < 1e-4 then begin
      Result := True;
      Exit; { Converged }
    end;
    LastErr := AvgErr;
  until CyclesUsed > MaxCycles;
end;

destructor TPeakFindDataSource.Destroy;
begin
  PeakFindForm.DataSource := nil;
  FreeAndNil(xVariable);
  PeakFindForm := nil;
  inherited;
end;

function TPeakFindDataSource.GetData(r: Integer; var x, y, w: Double;
  SetVariableControlsState: Boolean; ExtraAxis: Integer): Boolean;
begin
  x := PeakFindForm.x[r-1];
  y := PeakFindForm.y[r-1];
  w := 1;
  xVariable.Value := x;
  DataSourceFitter.ObservedVariable.Value := y;
  DataSourceFitter.weightVariable.Value := w;
  Result := not IsNan(y);
end;

function TPeakFindDataSource.nRows: Integer;
begin
  Result := PeakFindForm.nPoints;
end;

procedure TPeakFindDataSource.SetFitExpr(Expr: TExpr; nPar: Integer);
var
  i, nData: Integer;
  Sum, x, y, w: Double;
begin
  FitExpr := Expr;
  if not Assigned(FitExpr) or not Assigned(PeakFindForm) then
    Exit;
  { Fill in residuals }
  Sum := 0;
  nData := 0;
  for i := 0 to nRows-1 do
    if GetData(i+1,x,y,w) then begin
      PeakFindForm.d[i] := PeakFindForm.y[i] - FitExpr.Value;
      Sum := Sum + Sqr(PeakFindForm.d[i]);
      Inc(nData);
    end;
  if nData > nPar then
    AvgErr := Sqrt(Sum/(nData-nPar))
  else
    SetNan(AvgErr);
  with PeakFindForm.Results, DataSourceFitter as TInteractiveDataSourceFitter do begin
    PeakFindForm.Centre := Variable['c0']+Variable['c'];
    Value[1,1] := PeakFindForm.Centre;
    PeakFindForm.Area := Variable['a'];
    Value[2,1] := PeakFindForm.Area;
    PeakFindForm.Gaussian := Variable['g'];
    Value[3,1] := PeakFindForm.Gaussian;
    PeakFindForm.Lorentzian := Variable['l'];
    Value[4,1] := PeakFindForm.Lorentzian;
    PeakFindForm.BaseLine := Variable['b'];
    Value[5,1] := PeakFindForm.BaseLine;
    Value[6,1] := CyclesUsed;
    Value[7,1] := AvgErr;
    if nPar > 0 then begin
      Value[1,2] := GetStdDev('c');
      Value[2,2] := GetStdDev('a');
      Value[3,2] := GetStdDev('g');
      Value[4,2] := GetStdDev('l');
      Value[5,2] := GetStdDev('b');
    end else
      for i := 1 to 5 do
        Valid[i,2] := False;
  end;
  PeakFindForm.UpdateStatus;
end;

procedure TPeakFindDataSource.SetPeakFindForm(const Value: TPeakFindForm);
begin
  FPeakFindForm := Value;
  if Value <> nil then
    with DataSourceFitter do begin
      NLFitDlg.EmptyUndoList;
      NLFitDlg.FunctionText := 'a*voigt(x-c-c0,l,g)+b';
      NLFitDlg.CheckFunctionClick(nil);
      UpdateVariables(True);
    end;
end;

procedure TPeakFindDataSource.UpdateVariables(FloatWidths: Boolean);
begin
  with DataSourceFitter as TInteractiveDataSourceFitter do begin
    Variable['c0'] := PeakFindForm.Centre;
    SetStdDev('c0', 0);
    SetFloat('c0', False);
    Variable['a'] := PeakFindForm.Area;
    SetStdDev('a', PeakFindForm.Area/10);
    Variable['b'] := PeakFindForm.BaseLine;
    SetStdDev('b',(PeakFindForm.yMax-PeakFindForm.yMin)/10);
    Variable['c'] := 0;
    if PeakFindForm.LorentzianCheck.Checked then begin
      Variable['g'] := 0;
      SetStdDev('g',0);
      Variable['l'] := PeakFindForm.Lorentzian;
      SetStdDev('l',PeakFindForm.Lorentzian/10);
      SetFloat('g', False);
      SetFloat('l', FloatWidths);
      SetStdDev('c', PeakFindForm.Lorentzian/10);
    end else begin
      Variable['g'] := PeakFindForm.Gaussian;
      SetStdDev('g',PeakFindForm.Gaussian/10);
      Variable['l'] := 0;
      SetStdDev('l',0);
      SetFloat('g', FloatWidths);
      SetFloat('l', False);
      SetStdDev('c', PeakFindForm.Gaussian/10);
    end;
  end;
  PeakFindForm.PeakValid := True;
end;

function TPeakFindDataSource.Weighted: Boolean;
begin
  Result := False;
end;

{$IFDEF SHOWDIAG}
constructor TPeakFindForm.Create(AOwner: TComponent);
var
  Lft, Tp, Bt: Integer;
  procedure AddButton(TheClick: TNotifyEvent; TheCaption: string);
  begin
    with TButton.Create(Self) do begin
      Parent := PeakPanel;
      Top := Tp;
      Left := Lft + 1;
      Caption := TheCaption;
      OnClick := TheClick;
      Width := Canvas.TextExtent(Caption).cx+10;
      Lft := Left + Width;
      Bt := Top + Height;
    end;
  end;
begin
  inherited;
  Tp := AssignButton.Top + AssignButton.Height + 3;
  Lft := AssignButton.Left;
  AddButton({$IFDEF FPC}@{$ENDIF}LoadLog, 'Load');
  AddButton({$IFDEF FPC}@{$ENDIF}LoadDir, 'Dir');
  AddButton({$IFDEF FPC}@{$ENDIF}TestNext, 'Tst');
  AddButton({$IFDEF FPC}@{$ENDIF}TestPrev, 'Prv');
  AddButton({$IFDEF FPC}@{$ENDIF}TestAll, 'All');
  AddButton({$IFDEF FPC}@{$ENDIF}EditPeaklog, 'Edit');
  Lft := AssignButton.Left;
  Tp := Bt + 3;
  AddButton({$IFDEF FPC}@{$ENDIF}UpdateLog, 'Upd');
  AddButton({$IFDEF FPC}@{$ENDIF}ModTestCase, 'Mod');
//  AddButton({$IFDEF FPC}@{$ENDIF}ConvertToLineList, 'LineL');
end;

procedure TPeakFindForm.StartFile(AFileName: string);
begin
  FreeAndNil(LogStream);
  LogStream := TTextStream.Create(AFileName);
  LogStream.ItemSeparators := [isComma];
  LogStream.FlowToNextRecord := False;
  LastLoadedLogFile := AFileName;
  while not LogStream.AtEndOfFile and (LogStream.NextString('') = '') do
    LogStream.NextLine;
  LogID := 0;
end;

function TPeakFindForm.LoadNextEntry:Boolean;
var
  i: Integer;
  V: Double;
  LStatus: string;
  OldErrs: array [1..5] of Double;
begin
  Result := False;
  if LogStream = nil then
    if (LogList = nil) or (LogListPos >= LogList.Count) then
      Exit
    else begin
      StartFile(LogList[LogListPos]);
      Inc(LogListPos)
    end;
  if LogStream.Token <> 'Peak' then begin
    FreeAndNil(LogStream);
    raise EReportException.Create(LastLoadedLogFile+' is not a peak file');
    Exit;
  end;
  Inc(LogID);
  Caption := ExtractFileName(LogStream.StreamName);
  LStatus := LogStream.NextString('');
  LogStream.NextLine;
  LastLoadedInfo := LogStream.RestOfLine; // Original location
  LogStream.NextLine;
  if LogStream.NextString('') <> 'Data' then
    raise EReportException.Create('Missing Data');
  LogStream.NextIntegerVar(LastPoint);
  LogStream.NextLine;
  if not Assigned(DummyPlot) then
    DummyPlot := TExperimentalPlot.Create(nil);
  FirstPoint := 0;
  SetLength(DummyPlot.Frequency, LastPoint);
  SetLength(DummyPlot.Intensity, LastPoint);
  for i := 0 to LastPoint-1 do begin
    LogStream.NextDoubleVar(DummyPlot.Frequency[i]);
    LogStream.NextDoubleVar(DummyPlot.Intensity[i]);
    LogStream.NextLine;
  end;
  ThresholdSpin.Value := 50; { The default }
  BaseNode := DummyPlot;
  ShowItem(DummyPlot);
  LoadData([]);
  repeat
    if LogStream.GetFreeItem then
      if LogStream.Token = 'Position' then begin
        LogStream.NextDoubleVar(Centre);
        LogStream.NextDoubleVar(OldErrs[1]);
      end else if LogStream.Token = 'Area' then begin
        LogStream.NextDoubleVar(Area);
        LogStream.NextDoubleVar(OldErrs[2]);
      end else if LogStream.Token = 'Gaussian' then begin
        LogStream.NextDoubleVar(Gaussian);
        LogStream.NextDoubleVar(OldErrs[3]);
      end else if LogStream.Token = 'Lorentzian' then begin
        LogStream.NextDoubleVar(Lorentzian);
        LogStream.NextDoubleVar(OldErrs[4]);
      end else if LogStream.Token = 'BaseLine' then begin
        LogStream.NextDoubleVar(BaseLine);
        LogStream.NextDoubleVar(OldErrs[5]);
      end else if LogStream.Token = 'Threshold' then begin
        LogStream.NextDoubleVar(V);
        ThresholdSpin.Value := Round(V*100);
      end else if LogStream.Token = '****' then
        Break;
    LogStream.NextLine;
  until LogStream.AtEndOfFile;
  LorentzianCheck.Checked := Lorentzian > 0;
  LogStream.NextLine;
  while not LogStream.AtEndOfFile and (LogStream.NextString('') = '') do
    LogStream.NextLine;
  if LogStream.AtEndOfFile then
    FreeAndNil(LogStream);
  (DataSource as TPeakFindDataSource).UpdateVariables(True);
  DataSource.NLFitDlg.PlotFunctionClick(nil);
  UpdateStatus;
  for i := 1 to High(OldErrs) do
    Results.Value[i,2] := OldErrs[i];
  FitStatus.Caption := LStatus;
  Result := True;
end;

procedure TPeakFindForm.TestNext(Sender: TObject);
var
  PeakInfo: TPeakInfo;
  OldOK, NewOK, Match: Boolean;
  function Ck(a, b: Double):Boolean;
  begin
    Result := Abs(a-b)/(a+b) > 0.05;
  end;
var
  OldValues, OldErrs: array [1..7] of Double;
  i: Integer;
  OldStatus, Diff, OldS, NewS, Buf: string;
  NewVal, NewErr: Double;
const
  Widths: array [1..7] of Integer = (20, 14, 14, 14, 14, 3, 12);
begin
  if not LoadNextEntry then
    Exit;
  for i := 1 to High(OldValues) do begin
    OldValues[i] := Results.Value[i,1];
    OldErrs[i] := Results.Value[i,2];
  end;
  OldStatus := FitStatus.Caption;
  OldOK := Copy(OldStatus,1,2) = 'OK';
  {.$DEFINE TESTCOFM}
  {$DEFINE TESTAUTO}
  {$IFDEF TESTCOFM}
  if not HeaderWritten then
    WriteLn(Log, '     OK        Position                 Centre                Width              Area                Baseline');
  CofMAlgorithm;
  ShowCofMStatus;
  OldOK := Copy(FitStatus.Caption,1,2) = 'OK';
  NewOK := FindCentreOfMass(PeakInfo, y, x, Threshold);
  if OldOK <> NewOK then
    if OldOK then
      Write(Log, 'FailN ')
    else
      Write(Log, Copy(FitStatus.Caption,1,4):4, ' ')
  else if iPos <> PeakInfo.Position then
    Write(Log, 'iPos ')
  else if Abs(Centre-PeakInfo.Centre) > 1e-4 then
    Write(Log, 'Cent ')
  else if Ck(PeakInfo.Area,Area) then
    Write(Log, 'Area ')
  else if Ck(PeakInfo.Width,Gaussian+Lorentzian) then
    Write(Log, 'Widt ')
  else if Ck(PeakInfo.Baseline,BaseLine) then
    Write(Log, 'Base ')
  else
    Write(Log, 'OK   ');
  WriteLn(Log, Ord(OldOK):1, Ord(NewOK):1, iPos:10, PeakInfo.Position:10, Centre:15:4, PeakInfo.Centre:15:4,
               Gaussian+Lorentzian:8:4, PeakInfo.Width:8:4, DTosMA(Area,11,-1), DTosMA(PeakInfo.Area,11,-1),
               DTosMA(Baseline,11,-1), DTosMA(PeakInfo.Baseline,11,-1), ' ', LastLoadedLogFile, ' ', LogID);
  {$ENDIF}
  {$IFDEF TESTAUTO}
  NewOK := CofMAlgorithm;
  ShowCofMStatus;
  if NewOK then
    NewOK := (DataSource as TPeakFindDataSource).AutoFit;
  if OldOK <> NewOK then
    Diff := 'Stat'
  else
    Diff := '';
  Buf := '';
  if not HeaderWritten then
    Write(Log, '     ');
  for i := 1 to High(OldValues) do begin
    if not HeaderWritten then
      Write(Log, PadCenter(Results.Lbl[i],2*Widths[i]));
    NewVal := Results.Value[i,1];
    NewErr := Results.Value[i,2];
    Match := (IsNanOrInf(NewErr) = IsNanOrInf(OldErrs[i]))
             and (IsNanOrInf(NewVal) = IsNanOrInf(OldValues[i]));
    NewS := '';
    OldS := '';
    if Match then
      if IsNanOrInf(NewErr) then begin
        NewS := DtosM(NewVal, 10, -1);
        OldS := DtosM(OldValues[i], 10, -1);
        Match := OldS = NewS;
      end else if not IsNanOrInf(NewVal) then begin
        NewS := DtosErr(NewVal, NewErr, 2, -1);
        OldS := DtosErr(OldValues[i], OldErrs[i], 2, -1);
        Match := OldS = NewS;
      end;
    if not Match and ((Diff = '') or (Diff = 'Cycl')) then
      Diff := Copy(Results.Lbl[i],1,4);
    Buf := Buf + PadRight(NewS, Widths[i]) + PadRight(OldS, Widths[i]);
  end;
  if not HeaderWritten then
    WriteLn(Log);
  if Diff = '' then
    Diff := 'OK  ';
  WriteLn(Log, Diff, ' ', Buf, FitStatus.Caption, ' ', OldStatus, ' ', LastLoadedLogFile, ' ', LogID);
  {$ENDIF}
  HeaderWritten := True;
end;

procedure TPeakFindForm.TestPrev(Sender: TObject);
begin
  if LogListPos > 1 then begin
    Dec(LogListPos, 2);
    TestNext(nil);
  end;
end;

procedure TPeakFindForm.TestAll(Sender: TObject);
begin
  while ((LogList <> nil) and (LogListPos < LogList.Count)) or (LogStream <> nil)  do
    TestNext(nil);
end;

procedure TPeakFindForm.LoadLog(Sender: TObject);
begin
  AutoCheckBox.Checked := False;
  with TOpenDialog.Create(Self) do
    try
      DefaultExt := '.peaklog';
      Filter := 'Peak Log File|*.peaklog|All Files|'+AllFilesMask;
      Options := [ofEnableSizing];
      if Execute then begin
        StartFile(FileName);
        TestNext(nil);
      end;
    finally
      Free;
    end;
end;

procedure TPeakFindForm.LoadDir(Sender: TObject);
var
  Sr: TSearchRec;
begin
  AutoCheckBox.Checked := False;
  with TOpenDialog.Create(Self) do
    try
      DefaultExt := '.peaklog';
      Filter := 'Peak Log File|*.peaklog|All Files|'+AllFilesMask;
      Options := [ofEnableSizing];
      if Execute then begin
        if LogList = nil then
          LogList := TStringList.Create
        else
          LogList.Clear;
        LogListPos := High(Integer);
        if FindFirst(ExtractFileDir(FileName)+PathDelim+'*.peaklog', 0, Sr) = 0 then begin
            repeat
              LogList.Add(ExtractFileDir(FileName)+PathDelim+Sr.Name);
            until FindNext(Sr) <> 0;
          FindClose(sr);
          FreeAndNil(LogStream);
          LogListPos := 0;
          TestNext(nil);
        end;
      end;
    finally
      Free;
    end;
end;

procedure TPeakFindForm.EditPeaklog(Sender: TObject);
begin
  LaunchEditor(LastLoadedLogFile);
end;

procedure TPeakFindForm.UpdateLog(Sender: TObject);
begin
  if (DummyPlot <> Plot) or (LastLoadedLogFile = '') then
    Exit;
  if FileExists(LastLoadedLogFile+'.prev') or not RenameFile(LastLoadedLogFile, LastLoadedLogFile+'.prev') then
    raise EReportException.Create(LastLoadedLogFile+'.prev already exists');
  LogData(LastStatus, LastLoadedLogFile, LastLoadedInfo);
end;

procedure TPeakFindForm.ModTestCase(Sender: TObject);
begin
  if DummyPlot = Plot then
    LogData(PeakFindForm.FitStatus.Caption, '', 'Updated from '+ExtractFileName(LastLoadedLogFile), True)
  else
    LogData(PeakFindForm.FitStatus.Caption, '', '', True)
end;
{$ENDIF}

end.
