unit PlotDataSource;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Gendlg, {$ELSE} DataSource, {$ENDIF}
     ExpPlot, Expr, pgDOM, DataFitter;

type

  TPlotDataSource = class ({$IFNDEF CONSOLE} TInteractiveDataSource {$ELSE} TFitterDataSource {$ENDIF})
  private
    FPlot: TExperimentalPlot;
    procedure SetPlot(const Value: TExperimentalPlot);
  public
    xVariable: TVariable;
    FitExpr: TExpr;
    AvgErr: Double;
    destructor Destroy; override;
    procedure AddIndependentVariables(Lex:TLex); override;
    function nRows: Integer; override;
    function GetData(r: Longint; var x, y, w: Double; SetVariableControlsState: Boolean = False; ExtraAxis: Integer = -1): Boolean; override;
    function Weighted: Boolean; override;
    function Caption: string; override;
    procedure WriteXML(Result: TDOMNode);
    function ReadXML(DOMNode: TDOMNode; CheckUnread: Boolean):Integer;
    procedure SetFitExpr(Expr: TExpr; nPar: Integer); override;
    property Plot: TExperimentalPlot read FPlot write SetPlot;
  end;

implementation

uses {$IFNDEF CONSOLE} Main, CalFrm, PlotOC, LogUnit, ResidFrm, {$ENDIF}
     SysUtils, Math, Peak, Overlay, MNode, PropHelp,
     MathSup, Global, Mixture, Peaks, UCompat;

{$IFNDEF CONSOLE}
type
  TCalResidualPlotter = class(TResidualPlotter)
  protected
    ExpPlot: TExperimentalPlot;
    function GetX(Index: Integer): Double; override;
    function SkipY(var O, C, W: Double; Index: Integer):Boolean; override;
    function nObs:Integer; override;
    procedure SetXType(const Value: TOCXStyle); override;
  public
    constructor Create(AExpPlot: TExperimentalPlot);
    destructor Destroy; override;
    function ResidualAvailable(var Res, Wt: Double; Index: Integer):Boolean; override;
    function OnGetPointInfo(Index: Integer; Extra: PtrInt):string; override;
    procedure EditLastPoint; override;
    procedure ShowLastPoint; override;
    procedure RemovePoints(Area: TDRect; Outside: Boolean); override;
    procedure RemoveLastPoints; override;
  end;

{ TCalResidualPlotter }

constructor TCalResidualPlotter.Create(AExpPlot: TExperimentalPlot);
begin
  ExpPlot := AExpPlot;
  ExpPlot.ResidualPlotter:= Self;
  inherited Create('');
end;

destructor TCalResidualPlotter.Destroy;
begin
  if ExpPlot.ResidualPlotter = Self then
    ExpPlot.ResidualPlotter := nil;
  inherited;
end;

procedure TCalResidualPlotter.EditLastPoint;
begin
  if (nSavedLines > 0) and (SavedLines[0] < CalForm.Grid.RowCount) then
    CalForm.Grid.Row := SavedLines[0] + 1;
end;

procedure TCalResidualPlotter.SetXType(const Value: TOCXStyle);
begin
  inherited SetXType(Value);
  if XLabel = '' then
    SetXType(ocxNumber); { Force sensible default }
end;

procedure TCalResidualPlotter.RemoveLastPoints;
var
  i: Integer;
begin
  for i := 0 to nSavedLines-1 do
    if SavedLines[i] < CalForm.Grid.RowCount then
      CalForm.Grid.Valid[SavedLines[i]+1,4] := False;
end;

procedure TCalResidualPlotter.RemovePoints(Area: TDRect; Outside: Boolean);
var
  i: Integer;
  P: TPeak;
begin
  for i := 0 to ExpPlot.nPeaks-1 do begin
    SkipPoint := False;
    PreparePoint(i);
    if (InRange(GetX(i),Area.Left,Area.Right)
         and InRange(GetY(i),Area.Bottom,Area.Top)) xor Outside
       and not SkipPoint then
      CalForm.Grid.Valid[i+1,4] := False;
  end;
end;

procedure TCalResidualPlotter.ShowLastPoint;
begin
  if (nSavedLines > 0) and (SavedLines[0] < ExpPlot.nPeaks) then
    MainForm.CentrePlot(ExpPlot.Peaks.Peaks[SavedLines[0]].Actual, ExpPlot.Units);
end;

function TCalResidualPlotter.GetX(Index: Integer): Double;
begin
  with ExpPlot.Peaks.Peaks[Index] do begin
    SkipPoint := SkipPoint or IsNan(Weight) or IsNan(Actual) or (Weight <= 0);
    case XType of
      ocxNumber:    Result := Index+1;
      ocxFrequency: Result := Position;
      // ocxEnergy: ;
      ocxCalculated: Result := Actual-Residual;
      ocxObserved:   Result := Actual;
      ocxWeight:     Result := Weight;
      // ocxPopulation: ;
      // ocxQNO: ;
    else
      SetNan(Result);
    end;
  end;
end;

function TCalResidualPlotter.nObs: Integer;
begin
  Result := ExpPlot.nPeaks;
end;

function TCalResidualPlotter.OnGetPointInfo(Index: Integer; Extra: PtrInt): string;
begin
  if Index < ExpPlot.nPeaks then begin
    Result := ' ' + IntToStr(Index+1);
    if nMouseLines < Length(LastMouseLines) then begin
      LastMouseLines[nMouseLines] := Index;
      Inc(nMouseLines);
    end;
  end else begin
    Result := '';
    nMouseLines := 0;
  end;
end;

function TCalResidualPlotter.ResidualAvailable(var Res, Wt: Double; Index: Integer): Boolean;
begin
  Result := True;
  with ExpPlot.Peaks[Index] do begin
    Res := Residual;
    Wt := Weight;
  end;
end;

function TCalResidualPlotter.SkipY(var O, C, W: Double; Index: Integer): Boolean;
begin
  with ExpPlot.Peaks[Index] do begin
    Result := IsNan(Weight) or IsNan(Actual) or (Weight <= 0);
    O := Actual;
    C := Actual-Residual;
    W := Weight;
  end;
end;
{$ENDIF}

{ TPlotDataSource }

procedure TPlotDataSource.AddIndependentVariables(Lex: TLex);
begin
  if not Assigned(xVariable) then
    xVariable := TVariable.Create('x');
  Lex.AddVariable(xVariable);
end;

function TPlotDataSource.Caption: string;
begin
  Result := Plot.Name;
  {$IFNDEF CONSOLE}
  CalForm.UpdateStatus
  {$ENDIF}
end;

destructor TPlotDataSource.Destroy;
begin
  Plot.DataSource := nil;
  FreeAndNil(xVariable);
  inherited;
end;

function TPlotDataSource.GetData(r: Integer; var x, y, w: Double;
  SetVariableControlsState: Boolean; ExtraAxis: Integer): Boolean;
var
  Peak: TPeak;
begin
  Peak := Plot.Peaks.Peaks[r-1];
  x := Peak.Position;
  y := Peak.Actual;
  w := Peak.Weight;
  xVariable.Value := x;
  DataSourceFitter.ObservedVariable.Value := y;
  DataSourceFitter.weightVariable.Value := w;
  Result := not IsNan(Peak.Actual) and not IsNan(Peak.Weight) and (Peak.Weight > 0);
end;

function TPlotDataSource.nRows: Integer;
begin
  Result := Plot.nPeaks;
end;

procedure TPlotDataSource.SetFitExpr(Expr: TExpr; nPar: Integer);
var
  i, nData: Integer;
  x, y, w, Sum: Double;
  Peak: TPeak;
begin
  FitExpr := Expr;
  if not Assigned(FitExpr) or not Assigned(Plot) then
    Exit;
  { Fill in residuals }
  Sum := 0;
  nData := 0;
  for i := 0 to nRows-1 do
    if GetData(i+1,x,y,w) then
      with Plot do begin
        Peak := Peaks.Peaks[i];
        Peak.Residual := Peak.Actual - FitExpr.Value;
        Sum := Sum + Sqr(Peak.Residual);
        Inc(nData);
        Peaks.Peaks[i] := Peak;
      end;
  if nData > nPar then
    AvgErr := Sqrt(Sum/(nData-nPar))
  else
    SetNan(AvgErr);
  { Construct new frequency scale }
  with Plot do begin
    if Length(OriginalFrequency) = 0 then begin
      OriginalFrequency := Frequency;
      Frequency := nil; { Force allocation of new array }
      SetLength(Frequency,Length(OriginalFrequency));
    end;
    for i := 0 to High(Frequency) do begin
      xVariable.Value := OriginalFrequency[i];
      Frequency[i] := FitExpr.Value;
    end;
    CheckProperties;
  end;
  {$IFNDEF CONSOLE}
  CalForm.AvgErr := AvgErr;
  CalForm.UpdateStatus;
  ResidualForm.ResidualPlotter := TCalResidualPlotter.Create(Plot);
  if not Plot.Changing then
    MainForm.ConsiderReplot(False,False);
  {$ENDIF}
end;

procedure TPlotDataSource.SetPlot(const Value: TExperimentalPlot);
var
  WasChanging: Boolean;
begin
  FPlot := Value;
  WasChanging := FPlot.Changing;
  FPlot.Changing := True;
  try
    with DataSourceFitter do begin
      {$IFDEF CONSOLE}
      if Expression = '' then begin
        Expression := 'm*x+c';
        Variable['c'] := 0; // Plot.FrequencyOffset;
        Variable['m'] := 1; // Plot.FrequencyScale;
      end;
//      Plot.FrequencyOffset := 0; (Deleted as want to preserve theses values)
//      Plot.FrequencyScale := 1;
      SetFitExpr(Expr,0);
      Plot.CheckProperties;
      {$ELSE}
      Expression := 'm*x+c';
      Variable['c'] := 0; //Plot.FrequencyOffset;
      Variable['m'] := 1; // Plot.FrequencyScale;
//      Plot.FrequencyOffset := 0;
//      Plot.FrequencyScale := 1;
      SetFitExpr(Expr,0);
      Plot.CheckProperties;
      with NLFitDlg do begin
        CheckFunctionClick(nil);
        PlotFunctionClick(nil);
      end;
      {$ENDIF}
    end;
  finally
    FPlot.Changing := WasChanging;
  end;
end;

function TPlotDataSource.Weighted: Boolean;
begin
  Result := True;
end;

procedure TPlotDataSource.WriteXML(Result: TDOMNode);
var
  Node, VarNode: TDOMElement;
  i: Integer;
begin
  if Assigned(DataSourceFitter) then begin
    Node := Result.OwnerDocument.CreateElement('Calibration');
    Result.AppendChild(Node);
    with DataSourceFitter do begin
      if not Assigned(DataSource) then
        WriteValue(Node,'GlobalFit','true','');
      WriteValue(Node, 'Increment', FloatToStr(Increment), '1E-5');
      WriteValue(Node, 'ScaleChanges', FloatToStr(ScaleChanges), '1');
      WriteValue(Node, 'SVDThresh', FloatToStr(SVDThresh), '1e-15');
      WriteValue(Node, 'UseStdDev', BoolToStr(UseStdDev,True), 'true');
      WriteValue(Node, 'LinearFit', BoolToStr(LinearFit,True), 'false');
      WriteValue(Node, 'UseSVD', BoolToStr(UseSVD,True), 'true');
      VarNode := Result.OwnerDocument.CreateElement('Expression');
      Node.AppendChild(VarNode);
      WriteValue(VarNode,'Function',Expression,'');
      for i := 0 to Variables.Count-1 do begin
        VarNode := Result.OwnerDocument.CreateElement('Variable');
        Node.AppendChild(VarNode);
        WriteValue(VarNode, 'Name', Variables[i], '');
        WriteValue(VarNode, 'Value', FloatToStr(Variables.Objects[i].Value), '0');
        if not Floating[i] then
          WriteValue(VarNode, 'Float', 'false', '');
      end;
    end;
  end;
end;

function TPlotDataSource.ReadXML(DOMNode: TDOMNode; CheckUnread: Boolean):Integer;
var
  Child, NextChild: TDOMNode;
  nVar: Integer;
  s: string;
begin
  Result := 0;
  with DataSourceFitter do begin
    Child := DOMNode.FindNode('Expression');
    if Assigned(Child) then begin
      Expression := ReadValue(Child, 'Function', '');
      if CheckUnread then
        Result := nUnreadProperties('Expression', Child);
      DOMNode.RemoveChild(Child);
    end;
    Child := DOMNode.FirstChild;
    while Assigned(Child) do begin
      NextChild := Child.NextSibling;
      if Child.NodeName = 'Variable' then begin
        s := ReadValue(Child, 'Name', '');
        nVar := Variables.IndexOf(s);
        if nVar >= 0 then
          Variables.Objects[nVar].Value := StrToFloatEx(ReadValue(Child, 'Value', '0'));
        Floating[nVar] := StrToBoolean(ReadValue(Child, 'Float', 'true'));
        if CheckUnread then
          Result := Result + nUnreadProperties('Variable', Child);
        DOMNode.RemoveChild(Child);
      end;
      Child := NextChild;
    end;
    Increment := StrToFloatEx(ReadValue(DOMNode, 'Increment', '1e-5'));
    ScaleChanges := StrToFloatEx(ReadValue(DOMNode, 'ScaleChanges', '1'));
    SVDThresh := StrToFloatEx(ReadValue(DOMNode, 'SVDThresh', '1e-15'));
    GlobalFit := StrToBoolean(ReadValue(DOMNode, 'GlobalFit', 'false'));
    UseStdDev := StrToBoolean(ReadValue(DOMNode, 'UseStdDev', 'true'));
    LinearFit := StrToBoolean(ReadValue(DOMNode, 'LinearFit', 'false'));
    UseSVD := StrToBoolean(ReadValue(DOMNode, 'UseSVD', 'true'));
  end;
  {$IFDEF CONSOLE}
    SetPlot(FPlot);
  {$ELSE}
    with NLFitDlg do begin
      LoadFromFitter;
      PlotFunctionClick(nil);
    end;
  {$ENDIF}
end;

end.
