unit PlotFrm;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFDEF LCL_} LMessages, {$ENDIF}
     Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
     Dialogs, Buttons, ToolWin, ComCtrls, ExtCtrls, ScrlCtrl, Graphs,
     StdCtrls, SpectrumPlot, MNode, Mixture, GraphRange, Split, Overlay,
     SettingsForm, AlEdit, Menus, Settings, GraphI, GraphM, MathSup, Units;

type

  { TPlotForm }

  TPlotForm = class(TSettingsForm)
    GaussianPanel: TPanel;
    LorentzianPanel: TPanel;
    OffsetPanel: TPanel;
    PlotSelectAllItem: TMenuItem;
    SelectVisibleItem: TMenuItem;
    TemperaturePanel: TPanel;
    ToolBar: TToolBar;
    LeftButton: TSpeedButton;
    RightButton: TSpeedButton;
    CompressButton: TSpeedButton;
    ExpandButton: TSpeedButton;
    XZoomButton: TSpeedButton;
    AllButton: TSpeedButton;
    SimulateButton: TSpeedButton;
    PlotStyleButton: TSpeedButton;
    ShowPartsButton: TSpeedButton;
    ShowSumButton: TSpeedButton;
    ShowFortratButton: TSpeedButton;
    ScaleFromMinButton: TSpeedButton;
    ScaleFromMaxButton: TSpeedButton;
    TogleAutoYScaleButton: TSpeedButton;
    PasteOverlayButton: TSpeedButton;
    StatusBar: TStatusBar;
    G: TGraphD;
    OC: TGraphD;
    Split: TSplit;
    OpenButton: TSpeedButton;
    Label1: TLabel;
    Foffset: TAlignedEdit;
    Label4: TLabel;
    Temperature: TAlignedEdit;
    Label3: TLabel;
    Gaussian: TAlignedEdit;
    Label2: TLabel;
    Lorentzian: TAlignedEdit;
    OverlayPopupMenu: TPopupMenu;
    JoinToItem: TMenuItem;
    InvertPopupItem: TMenuItem;
    SmoothItem: TMenuItem;
    HideItem: TMenuItem;
    ShowAllItem: TMenuItem;
    DeleteItem: TMenuItem;
    HideOthersItem: TMenuItem;
    CalibrateThisItem: TMenuItem;
    PropertiesItem: TMenuItem;
    UndoItem: TMenuItem;
    SplitItem: TMenuItem;
    AsLineListItem: TMenuItem;
    HideOutofRangeItem: TMenuItem;
    CompressSpectrumItem: TMenuItem;
    ArbitraryUnitsItem: TMenuItem;
    SetZeroItem: TMenuItem;
    DeletePointItem: TMenuItem;
    BaselineItem: TMenuItem;
    DeletePointsItem: TMenuItem;
    DeGlitchItem: TMenuItem;
    procedure FoffsetChange(Sender: TObject);
    procedure FoffsetKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure GContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
    procedure SelectVisibleItemClick(Sender: TObject);
    procedure StatusBarHint(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure PasteOverlayButtonClick(Sender: TObject);
    procedure ShowFortratButtonClick(Sender: TObject);
    procedure ShowPartsButtonClick(Sender: TObject);
    procedure ShowSumButtonClick(Sender: TObject);
    procedure SimulateButtonClick(Sender: TObject);
    procedure TogleAutoYScaleButtonClick(Sender: TObject);
    procedure ScaleFromMaxButtonClick(Sender: TObject);
    procedure ScaleFromMinButtonClick(Sender: TObject);
    procedure ExpandButtonClick(Sender: TObject);
    procedure CompressButtonClick(Sender: TObject);
    procedure RightButtonClick(Sender: TObject);
    procedure LeftButtonClick(Sender: TObject);
    procedure XZoomButtonClick(Sender: TObject);
    procedure GMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure GMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure PlotStyleButtonClick(Sender: TObject);
    procedure AllButtonClick(Sender: TObject);
    procedure JoinToItemClick(Sender: TObject);
    procedure InvertPopupItemClick(Sender: TObject);
    procedure SmoothItemClick(Sender: TObject);
    procedure HideItemClick(Sender: TObject);
    procedure ShowAllItemClick(Sender: TObject);
    procedure DeleteItemClick(Sender: TObject);
    procedure HideOthersItemClick(Sender: TObject);
    procedure CalibrateThisItemClick(Sender: TObject);
    procedure PropertiesItemClick(Sender: TObject);
    procedure UndoItemClick(Sender: TObject);
    procedure SplitItemClick(Sender: TObject);
    procedure AsLineListItemClick(Sender: TObject);
    procedure HideOutofRangeItemClick(Sender: TObject);
    procedure ArbitraryUnitsItemClick(Sender: TObject);
    procedure DeletePointItemClick(Sender: TObject);
    procedure BaselineItemClick(Sender: TObject);
    procedure DeletePointsItemClick(Sender: TObject);
    procedure DeGlitchItemClick(Sender: TObject);
  private
    FLastGraph: TGraphD;
    FLastFortrat: TFortratPlot;
    GraphStyle: TGraphStyle; { TODO: This belongs in TSimulation }
    FLastLineInfo: string;
    OverlayStuffed: Boolean;
    function ShowYValue(GD: TgraphD; const PlotRanges: TPlotRanges; var YWidth: Double; out YRange: Double):string;
    function GetSimulation: TSimulation;
    procedure SetSimulation(const Value: TSimulation);
    function LookForSimulationLineHits(Y:Double; const PlotRanges: TPlotRanges; OnCheckLine: TOnCheckLine):Boolean;
    function CheckFortratLine(const Line: TLine):Boolean;
    type
      TYScaleType = (ysUp, ysDown, ysAutoOn, ysAutoOff);
    procedure AdjustYScale(YScaleType: TYScaleType); overload;
    procedure AdjustYScale(MNode: TStretchableMNode; YScaleType: TYScaleType); overload;
  protected
    class var FLastPlot: TPlot;
  protected
    FSimulation: TSimulation;
    SpectrumPlotRanges: TPlotRanges; { Tags: -2 => Obs-Calc tick marks; 10+n => Overlay n; -9 => Overall Simulation; -10-n => Individual simulation n }
    FortratPlotRanges: TPlotRanges;
    PlotColours: TColours;

    { Information about where popup menu launched from }
    GPopedUpAt: Double;
    GPopupOverlay: TPlot;
    FLastAreaSelected: Boolean;
    FLastSelectedMathArea: TDRect;

    { Link between FormSettings and Mixture }
    function LastGraph: TGraphD;
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer); override;
    procedure ReadTextWindow(Edit: TCustomEdit; ParameterIndex: Integer);
    {$IFDEF VCL_}
    procedure WMDropFiles(var Message: TMessage); message WM_DROPFILES;
    {$ENDIF}
    procedure PopulateGPopup(Plot: TPlot);
    function ActiveSimulation: TSimulation;
    function ActivePlotForm: TPlotForm;
    procedure HandleComponent(Component: TComponent; var PName, Value, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
  public
    AllWanted: Boolean;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function IsShortCut(var Message: {$IFDEF LCL_}TLMKey{$ELSE}TWMKey{$ENDIF}): Boolean; override;

    procedure ShowSimulation(GlobalMixture: TMixture; Overlays: TOverlays); overload;
    procedure ShowSimulation; overload; virtual;
    procedure InitializeFromSimulation(ASimulation: TSimulation);
    function FindLimits(IncludeOverlays: Boolean):Boolean;
    procedure CentrePlot(APosition: Double; TheUnits: TResonanceUnits);

    property Simulation: TSimulation read GetSimulation write SetSimulation;
  end;

implementation

uses {$IFDEF LCL_} LCLType, {$ENDIF}
     Main, WCompat, Math, Obs, Fitter, LineFrm, BaselineFrm, DeGlitchFrm,
     Global, CalFrm, MNodeColour, ApplyFrm, ExpPlot, SmoothDlg,
     OvrlyFrm, FormUtils, Worker, Plotter, UCompat, SizeHelpers;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TPlotForm }

procedure TPlotForm.ShowSimulation(GlobalMixture: TMixture; Overlays: TOverlays);
var
  AnySpectrum: Boolean;
begin
  FLastAreaSelected := False;
  AllWanted := False;
  if BeginCalculation(False) then Exit;
  try
    SetLength(FortratPlotRanges,0);
    SetLength(SpectrumPlotRanges,0);
    FLastLineInfo := '';
    FreeAndNil(FLastFortrat);
    if not Assigned(GlobalMixture) then Exit;
    if Tag > 0 then
      InitializeFromSimulation(MainForm.Simulation);
    ReadTextWindow(Foffset,FoffsetIndex);
    ReadTextWindow(Temperature,TemperatureIndex);
    ReadTextWindow(Gaussian,GaussianIndex);
    ReadTextWindow(Lorentzian,LorentzianIndex);
    AnySpectrum := Simulation.ShowSum or Simulation.ShowParts;

    Simulation.GraphStyle := GraphStyle; { TODO: GraphStyle belongs in simulation anyway }
    if Simulation.Fmin = Simulation.Fmax then begin
      AllButtonClick(nil);
      Exit;
    end;

    if AnySpectrum and Simulation.ShowFortrat then begin
      Split.MakeBothVisible;
      Split.Visible := True;
    end else begin
      if not Simulation.ShowFortrat then
        Split.OneWindowVisible(G)
      else
        Split.OneWindowVisible(OC);
      Split.Visible := False;
    end;

    if GlobalMixture.PrintLevel >= plLineList then
      GlobalMixture.PrettyPrint(Log, GlobalMixture.PrintLevel);
    G.ResetZoom;
    if G.Visible then
      FLastLineInfo := MakeLowerPlot(G,GlobalMixture,Simulation,Overlays,SpectrumPlotRanges,PlotColours);
    StatusBar.Panels[5].Text := FLastLineInfo;
    if Simulation.ShowFortrat then begin
      OC.ResetZoom;
      if not MakeUpperPlot(OC,Simulation,FLastFortrat,FortratPlotRanges) and not Split.Visible then
        Simulation.ShowSum := True; { Avoid display of blank screen }
      if Length(PlotColours) = 0 then
        FLastFortrat.GetColoursUsed(PlotColours);
    end;
    GlobalMixture.Simulations[0].HasPlotted := True;
  finally
    EndCalculation;
  end;
end;

procedure TPlotForm.ShowSimulation;
begin
  MainForm.ShowSimulation;
end;

procedure TPlotForm.InitializeFromSimulation(ASimulation: TSimulation);
begin
  Simulation.UpdateBaseNode(ASimulation);
end;

procedure TPlotForm.HandleComponent(Component: TComponent; var PName, Value,
  PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean);
begin
  if Component is TSplit then
    inherited;
end;

{ G Popup Menu Items }

procedure TPlotForm.PopulateGPopup(Plot: TPlot);
var
  i: Integer;
  IsPlot: Boolean;
begin
  GPopupOverlay := Plot;
  if Assigned(Plot) then
    GPopedUpAt := Convert(G.LastMouseX, Simulation.PlotUnits, Plot.Units);
  DeleteItem.Visible := Assigned(Plot);
  HideItem.Visible := Assigned(Plot);
  PropertiesItem.Visible := Assigned(Plot);
  InvertPopupItem.Visible := Assigned(Plot);
  HideOthersItem.Visible := Assigned(Plot) and (GlobalOverlays <> nil) and (GlobalOverlays.nPlots > 1);
  HideOutofRangeItem.Visible := HideOthersItem.Visible;
  IsPlot := Plot is TExperimentalPlot;
  JoinToItem.Visible := IsPlot;
  AsLineListItem.Visible := IsPlot;
  SplitItem.Visible := IsPlot;
  SmoothItem.Visible := IsPlot;
  DeletePointItem.Visible := IsPlot;
  DeletePointsItem.Visible := IsPlot and G.SignificantAreaSelected;
  if IsPlot then begin
    AsLineListItem.Checked := TExperimentalPlot(Plot).IsLineList;
    BaselineItem.Visible := not TExperimentalPlot(Plot).IsLineList;
    DeGlitchItem.Visible := not TExperimentalPlot(Plot).IsLineList;
  end else
    BaselineItem.Visible := False;
  CalibrateThisItem.Visible := IsPlot;
  UndoItem.Visible := Assigned(Plot);
  if Assigned(Plot) then
    UndoItem.Enabled := Plot.CanUndo;
  SelectVisibleItem.Visible := IsPlot;
  ShowAllItem.Visible := False;
  if GlobalOverlays <> nil then
    for i := 0 to GlobalOverlays.nPlots-1 do
      if GlobalOverlays[i].Hide then
        ShowAllItem.Visible := True
      else
        SelectVisibleItem.Visible := True;
  PlotSelectAllItem.Visible := SelectVisibleItem.Visible;
  if (GlobalOverlays <> nil) and (GlobalOverlays.nPlots > 0) then begin
    ArbitraryUnitsItem.Visible := True;
    ArbitraryUnitsItem.Checked := GlobalOverlays.IntensityUnits = Arbitrary;
  end else
    ArbitraryUnitsItem.Visible := False;
  CompressSpectrumItem.Visible := ArbitraryUnitsItem.Visible;
  SetZeroItem.Visible := ArbitraryUnitsItem.Visible;
  if not OverlayStuffed then
    G.StuffPopupMenu(OverlayPopupMenu);
  OverlayStuffed := True;
  OverlayPopupMenu.PopupComponent := G;
  OverlayPopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TPlotForm.PropertiesItemClick(Sender: TObject);
begin
  MainForm.OverlayPropertiesItemClick(nil);
  OverlayForm.ShowItem(GPopupOverlay);
end;

procedure TPlotForm.SelectVisibleItemClick(Sender: TObject);
var
  ActionType: Integer;
  FirstPlot: Boolean;
  procedure Process(Plot: TPlot);
  var
    First, Last: Integer;
  begin
    if Plot is TExperimentalPlot then
      with Plot as TExperimentalPlot do
        if ActionType = 1 {SelectAll} then begin
          FirstPoint := 0;
          LastPoint := -1;
        end else if ActionType = 3 {Compress} then begin
          OverlayForm.CompressPlot(Plot as TExperimentalPlot,FirstPlot);
          FirstPlot := False;
        end else begin
          IntensityAt(Convert(G.Xmin,Simulation.PlotUnits,Units),First);
          IntensityAt(Convert(G.Xmax,Simulation.PlotUnits,Units),Last);
          InOrder(First,Last);
          if Last - First > 1 then
            if ActionType = 0 {Select Visible} then begin
              FirstPoint := First;
              LastPoint := Last;
            end else
              BaseLine := AutoZero(-1,First,Last);
        end;
  end;
var
  i: Integer;
begin
  ActionType := (Sender as TComponent).Tag;
  FirstPlot := True;
  if Assigned(GPopupOverlay) then
    Process(GPopupOverlay)
  else if GlobalOverlays <> nil then
    for i := 0 to GlobalOverlays.nPlots-1 do
      if not GlobalOverlays[i].Hide then
        Process(GlobalOverlays[i]);
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.SetSimulation(const Value: TSimulation);
begin
  FSimulation := Value;
  if Tag > 0 then
    FSimulation.Name := 'Simulation ' + IntToStr(Tag)
  else
    FSimulation.Name := 'Simulation ';
 end;

procedure TPlotForm.DeleteItemClick(Sender: TObject);
begin
  if Application.MessageBox(PChar('Are you sure you want to delete ' + GPopupOverlay.UniqueName + '?'),
                           'Warning', MB_YESNO or MB_ICONQUESTION) = IDNO then
    Exit;
  GPopupOverlay.Free;
  MainForm.ConsiderReplot;
end;

destructor TPlotForm.Destroy;
begin
  FSimulation.Free; { Set to nil in MNodeChange; FreeAndNil would trigger G.Plotchanged }
  FreeAndNil(FLastFortrat);
  inherited;
end;

procedure TPlotForm.InvertPopupItemClick(Sender: TObject);
begin
  GPopupOverlay.Invert;
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.DeletePointItemClick(Sender: TObject);
begin
  (GPopupOverlay as TExperimentalPlot).Deglitch(GPopedUpAt);
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.DeletePointsItemClick(Sender: TObject);
begin
  (GPopupOverlay as TExperimentalPlot).DeletePoints(FLastSelectedMathArea.Left, FLastSelectedMathArea.Right, Simulation.PlotUnits);
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.JoinToItemClick(Sender: TObject);
var
  Plot: TPlot;
  ScalePlot: Boolean;
begin
  ScalePlot := False;
  Plot := PromptForOverlay(Self,'Join ' + GPopupOverlay.Name+ ' To',ScalePlot,'Scale plot',False,TExperimentalPlot,GPopupOverlay) as TPlot;
  if Assigned(Plot) then
    (GPopupOverlay as TExperimentalPlot).JoinTo(GPopedUpAt,Plot as TExperimentalPlot,ScalePlot,IsEnergyUnit(Simulation.PlotUnits) xor (Simulation.Fmax < Simulation.Fmin));
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.AsLineListItemClick(Sender: TObject);
begin
  with GPopupOverlay as TExperimentalPlot do
    IsLineList := not IsLineList;
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.BaselineItemClick(Sender: TObject);
begin
  ReShow(BaselineForm);
  BaselineForm.ShowItem(GPopupOverlay);
end;

procedure TPlotForm.DeGlitchItemClick(Sender: TObject);
begin
  ReShow(DeGlitchForm);
  DeGlitchForm.ShowItem(GPopupOverlay);
end;

procedure TPlotForm.SmoothItemClick(Sender: TObject);
begin
  with SmoothForm do
    if ShowModal = mrOK then
      (GPopupOverlay as TExperimentalPlot).Smooth(WidthSpinBox.Value,DerivativeSpinBox.Value,PolySpinBox.Value);
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.SplitItemClick(Sender: TObject);
begin
 (GPopupOverlay as TExperimentalPlot).Split(GPopedUpAt);
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.HideItemClick(Sender: TObject);
begin
  GPopupOverlay.Hide := True;
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.HideOthersItemClick(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to GlobalOverlays.nPlots-1 do
    if GlobalOverlays[i] <> GPopupOverlay then
      GlobalOverlays[i].Hide := True;
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.HideOutofRangeItemClick(Sender: TObject);
var
  Fi, Fx, Oi, Ox: Double;
  i: Integer;
begin
  Fi := Simulation.Fmin-Simulation.Foffset;
  Fx := Simulation.Fmax-Simulation.Foffset;
  InOrder(Fi,Fx);
  for i := 0 to GlobalOverlays.nPlots-1 do
    with GlobalOverlays[i] do begin
      Oi := Convert(HMin,Units,Simulation.PlotUnits);
      Ox := Convert(HMax,Units,Simulation.PlotUnits);
      InOrder(Oi,Ox);
      if (Ox < Fi) or (Fx < Oi) then
        Hide := True;
    end;
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.ShowAllItemClick(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to GlobalOverlays.nPlots-1 do
    GlobalOverlays[i].Hide := False;
  MainForm.ConsiderReplot;
end;

procedure TPlotForm.ShowSumButtonClick(Sender: TObject);
begin
  Simulation.ShowSum := ShowSumButton.Down;
  ShowSimulation;
end;

procedure TPlotForm.ShowPartsButtonClick(Sender: TObject);
begin
  Simulation.ShowParts := ShowPartsButton.Down;
  ShowSimulation;
end;

procedure TPlotForm.ShowFortratButtonClick(Sender: TObject);
begin
  Simulation.ShowFortrat := ShowFortratButton.Down;
  ShowSimulation;
end;

constructor TPlotForm.Create(AOwner: TComponent);
var
  i, M: Integer;
begin
  inherited;
  FLastGraph := G;
  GraphStyle := gsOffsetAll;
  M := 0;
  for i := 0 to Screen.CustomFormCount - 1 do
    if Screen.CustomForms[i] is TPlotForm then
      M := Max(Screen.CustomForms[i].Tag, M);
  Tag := M+1;
end;

function TPlotForm.ActivePlotForm: TPlotForm;
begin
  { To be called for Page Up and similar shortcuts to apply to current
    plot window only; works in conjunction with IsShortCut }
  if Screen.ActiveCustomForm is TPlotForm then
    Result := TPlotForm(Screen.ActiveCustomForm)
  else
    Result := Self;
end;

function TPlotForm.ActiveSimulation: TSimulation;
begin
  Result := ActivePlotForm.Simulation;
end;

function TPlotForm.IsShortCut(var Message: {$IFDEF LCL_}TLMKey{$ELSE}TWMKey{$ENDIF}): Boolean;
begin
  { Some short cuts should work on plot forms only - see ActiveSimulation }
  if (Message.CharCode in [VK_PRIOR, VK_NEXT]) and not (Screen.ActiveCustomForm is TPlotForm) then
    Result := False
  else
    Result := inherited IsShortCut(Message);
end;

procedure TPlotForm.AllButtonClick(Sender: TObject);
begin
  AdjustYScale(ysAutoOn);
  ActivePlotForm.AllWanted := True;
  MainForm.DoPlot(True)
end;

procedure TPlotForm.ArbitraryUnitsItemClick(Sender: TObject);
var
  i: Integer;
begin
  if GlobalOverlays <> nil then
    with GlobalOverlays do begin
      if IntensityUnits = Arbitrary then
        IntensityUnits := Normalized
      else
        IntensityUnits := Arbitrary;
      { TODO: Next lines are inelegant; top level setting should do this }
      for i := 0 to Count - 1 do
        if Components[i] is TPlot then
          (Components[i] as TPlot).IntensityUnits := IntensityUnits;
      MainForm.ConsiderReplot;
    end;
end;

procedure TPlotForm.PlotStyleButtonClick(Sender: TObject);
begin
  if GraphStyle = High(GraphStyle) then
    GraphStyle := Low(GraphStyle)
  else
    Inc(GraphStyle);
  ShowSimulation;
end;

procedure TPlotForm.XZoomButtonClick(Sender: TObject);
begin
  if not LastGraph.AreaSelected then Exit;
  Simulation.Fmin := LastGraph.SelectedMathArea.Left + Simulation.Foffset;
  Simulation.Fmax := LastGraph.SelectedMathArea.Right + Simulation.Foffset;
  ShowSimulation;
end;

procedure TPlotForm.LeftButtonClick(Sender: TObject);
var
  d: Double;
begin
  with ActiveSimulation do begin
    d := (Fmax-Fmin)/2;
    Fmin := Fmin - d;
    Fmax := Fmax - d;
  end;
  ShowSimulation;
end;

procedure TPlotForm.RightButtonClick(Sender: TObject);
var
  d: Double;
begin
  with ActiveSimulation do begin
    d := (Fmax-Fmin)/2;
    Fmin := Fmin + d;
    Fmax := Fmax + d;
  end;
  ShowSimulation;
end;

procedure TPlotForm.CalibrateThisItemClick(Sender: TObject);
begin
  MainForm.CalibrateItemClick(nil);
  CalForm.ShowItem(GPopupOverlay);
end;

procedure TPlotForm.CentrePlot(APosition: Double; TheUnits: TResonanceUnits);
var
  d: Double;
begin
  APosition := Convert(APosition, TheUnits, Simulation.PlotUnits);
  with Simulation do begin
    d := APosition - (Fmax+Fmin)/2;
    Fmin := Fmin + d;
    Fmax := Fmax + d;
  end;
  ShowSimulation;
end;

procedure TPlotForm.UndoItemClick(Sender: TObject);
begin
  if GPopupOverlay.CanUndo then begin
    GPopupOverlay.Undo;
    MainForm.ConsiderReplot(False,False);
  end;
end;

procedure TPlotForm.CompressButtonClick(Sender: TObject);
var
  d: Double;
begin
  with ActiveSimulation do begin
    d := (Fmax-Fmin)/2;
    Fmin := Fmin - d;
    Fmax := Fmax + d;
  end;
  ShowSimulation;
end;

procedure TPlotForm.ExpandButtonClick(Sender: TObject);
var
  d: Double;
begin
  with ActiveSimulation do begin
    d := (Fmax-Fmin)/4;
    Fmin := Fmin + d;
    Fmax := Fmax - d;
  end;
  ShowSimulation;
end;

procedure TPlotForm.AdjustYScale(MNode: TStretchableMNode; YScaleType: TYScaleType);
begin
  with MNode do
    case YScaleType of
      ysUp:   begin
                AutoMax := False;
                YMax := YMin + (Ymax-Ymin)/2;
              end;
      ysDown: begin
                AutoMin := False;
                YMin := YMax + (Ymin-Ymax)/2;
              end;
      ysAutoOff: begin
                   AutoMax := False;
                   AutoMin := False;
                 end;
      ysAutoOn:  begin
                   AutoMax := True;
                   AutoMin := True;
                 end;
    end;
end;

procedure TPlotForm.AdjustYScale(YScaleType: TYScaleType);
var
  OnOverlays: Boolean;
  i: Integer;
begin
  OnOverlays := (GlobalOverlays <> nil) and (GlobalOverlays.nVisible > 0);
  if OnOverlays then
    for i := 0 to GlobalOverlays.nPlots-1 do
      if not GlobalOverlays[i].Hide then
        AdjustYScale(GlobalOverlays[i], YScaleType);
  if not OnOverlays or (Simulation.GraphStyle = gsNoOffset) or (YScaleType in [ysAutoOn,ysAutoOff]) then
    AdjustYScale(Simulation, YScaleType);
  ShowSimulation;
end;

procedure TPlotForm.ScaleFromMinButtonClick(Sender: TObject);
begin
  AdjustYScale(ysUp);
end;

procedure TPlotForm.ScaleFromMaxButtonClick(Sender: TObject);
begin
  AdjustYScale(ysDown);
end;

procedure TPlotForm.TogleAutoYScaleButtonClick(Sender: TObject);
var
  AllAuto: Boolean;
  i: Integer;
begin
  AllAuto := True;
  if GlobalOverlays <> nil then
    for i := 0 to GlobalOverlays.nPlots-1 do
      if not GlobalOverlays[i].Hide then
        with GlobalOverlays[i] do
          AllAuto := AllAuto and AutoMin and AutoMax;
  with Simulation do
    AllAuto := AllAuto and AutoMin and AutoMax;
  if AllAuto then
    AdjustYScale(ysAutoOff)
  else
    AdjustYScale(ysAutoOn);
end;

function TPlotForm.CheckFortratLine(const Line: TLine): Boolean;
var
  Y: Double;
  PlotNumber: Integer;
begin
  if Assigned(FLastFortrat) then begin
    FLastFortrat.GetPlotY(Line, PlotNumber, Y);
    Result := Abs(Y-LastGraph.LastMouseY) <= SmallY*LastGraph.PixelSize(False);
  end else
    Result := False;
end;

function TPlotForm.LookForSimulationLineHits(Y:Double; const PlotRanges: TPlotRanges; OnCheckLine: TOnCheckLine):Boolean;
var
  Range: Double;
  Colour: TColor;
  FLastTag: Integer;
begin
  Result := False;
  { Possible cases: No Simulations selected; more than one selected;
    single colour selected }
  FLastTag := FindYRange(Y, LastGraph, PlotRanges, Low(Integer), -9);
  if FLastTag <=-Length(PlotRanges) then
    Exit; { No hits }
  Colour := clHide; { Assume multiple hits, or showing sum, or plot number check deferred }
  if (FLastTag >= 0) and not Assigned(OnCheckLine) then begin
    FLastTag := -PlotRanges[FLastTag].Tag-10;
    if (FLastTag >= 0) and (FLastTag < Length(PlotColours)) then
      Colour := PlotColours[FLastTag]; { Single spectrum }
  end;
  Range := Max(Simulation.Lorentzian + Simulation.Gaussian, SmallX*LastGraph.PixelSize(True));
  GlobalLineList.FindLines(LastGraph.LastMouseX+Simulation.Foffset, Range, Simulation.PlotUnits, Colour, OnCheckLine);
  LineForm.ShowLines(Simulation, True, False);
  Result := True;
end;

procedure TPlotForm.GMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  P, Off: Double;
  Area: TDRect;

  function LookForObsToCalcHits(GD: TGraphD; const PlotRanges: TPlotRanges):Boolean;
  var
    y: Double;
    i: Integer;
  begin
    Result := False;
    y := GD.LastMouseY;
    i := FindYRange(y,GD,PlotRanges,-2,-2);
    if i < 0 then
      Exit;
    for i := 0 to High(Observations) do
      if GD.CloseToX(GD.LastMouseX, Observations[i].Frequency)
         or GD.CloseToX(GD.LastMouseX, Observations[i].Calculated) then begin
        Result := True;
        EditObservation(Observations[i]);
        Exit;
      end;
  end;

  function LookForOverlayHits(Y:Double; const PlotRanges: TPlotRanges):Boolean;
  var
    LastTag: Integer;
  begin
    Result := False;
    { Possible cases: No Simulations selected; more than one selected;
      single colour selected }
    LastTag := FindYRange(Y, LastGraph, PlotRanges, 10);
    if LastTag < 0 then
      Exit; { None or multiple hits }
    { Single Overlay selected - pop up menu }
    if FLastAreaSelected then
      G.SelectedMathArea := FLastSelectedMathArea;
    PopulateGPopup(GlobalOverlays[PlotRanges[LastTag].Tag mod 1000 - 10]);
    Result := True;
  end;

  procedure MeasureSimulationPeaks;
  var
    i, LastTag, LastRange: Integer;
    sl, su, Position: Double;
    IMin, IMax: Double; { Record range of plots }
    Colour: TColor;
  begin
    Area.Left := Area.Left + Simulation.Foffset;
    Area.Right := Area.Right + Simulation.Foffset;
    sl := YFraction(Area.Bottom,G);
    su := YFraction(Area.Top,G);
    InOrder(sl,su);
    LastTag := 0;
    IMax := 0;
    IMin := 1;
    for i := 0 to High(SpectrumPlotRanges) do
      with SpectrumPlotRanges[i] do
        if (Tag <= -9) then begin
          IMin := Min(IMin, y1);
          IMax := Max(IMax, y2);
          if RangeOverlaps(sl,su,SpectrumPlotRanges[i]) then begin
            { Consider peak finding here }
            LastRange := i;
            if LastTag = 0 then
              LastTag := Tag
            else
              LastTag := -9; { Flag sum or multiple hits }
          end;
        end;
    LastTag := -LastTag-10;
    if LastTag < -1 then
      Exit { No hits }
    else if (LastTag < 0) or (LastTag >= Length(PlotColours)) then
      Colour := clHide { Multiple hits, or showing sum }
    else
      Colour := PlotColours[LastTag]; { Single spectrum }
    if GlobalLineList.FindPeak(Area.Left, Area.Right, Position, Simulation.PlotUnits, Colour) then begin
      { Found a peak, so display it }
      if Colour <> clHide then
        with SpectrumPlotRanges[LastRange] do begin
          { Just plot peak over selected spectrum }
          IMin := y1;
          IMax := y2;
        end;
      IMin := FractionToY(IMin, G);
      IMax := FractionToY(IMax, G);
      Position := Position - Simulation.Foffset;
      G.ClipPoint(Position, Imin);
      G.ClipJoin(Position, Imax);
      G.ClipPoint(Position, Imin); { Force flush of line plot }
      G.TextAngle(90);
      G.TextJ(Position, Imax, taRightJustify, FloatToStr(Position));
      G.TextAngle(0);
    end;
  end;
var
  Onp: Integer;
  FoundHit: Boolean;
begin
  if ((ssAlt in Shift) or (ssCtrl in Shift)) and (Button = mbLeft) and LastGraph.AreaSelected then begin
    Off := LastGraph.SelectedMathArea.Right - LastGraph.SelectedMathArea.Left;
    Off := RoundToIncrement(Off, LastGraph.PixelSize(True)/2);
    if CalibratingPlot <> nil then begin
      if Simulation.PlotUnits <> CalibratingPlot.Units then
        Off := Convert(LastGraph.SelectedMathArea.Right,Simulation.PlotUnits,CalibratingPlot.Units)
               - Convert(LastGraph.SelectedMathArea.Left,Simulation.PlotUnits,CalibratingPlot.Units);
      CalibratingPlot.OffsetBy(off);
    end else if TryStrToFloatEx(Foffset.Text,P) then
      Simulation.Foffset := P + Off
    else
      GlobalMixture.Simulations[0].Foffset := GlobalMixture.Simulations[0].Foffset + Off;
    ShowSimulation;
  end else if (Button = mbLeft) then begin
    if Sender = G then begin
      FLastAreaSelected := G.SignificantAreaSelected;
      FLastSelectedMathArea := G.SelectedMathArea;
    end else
      FLastAreaSelected := False;
  end else if (Button = mbRight) then
    if not LastGraph.SignificantAreaSelected then begin
      { TODO: Check for linelist overlay hits }
      FoundHit := False;
      if LastGraph.InPlotArea(LastGraph.LastMouseX, LastGraph.LastMouseY) then begin
        if GlobalLineList <> nil then
          if LastGraph = G then
            FoundHit := LookForSimulationLineHits(LastGraph.LastMouseY, SpectrumPlotRanges, nil)
          else
            FoundHit := LookForSimulationLineHits(LastGraph.LastMouseY, FortratPlotRanges, {$IFDEF FPC}@{$ENDIF}CheckFortratLine);
        if not FoundHit and (LastGraph = G) then begin
          FoundHit := LookForOverlayHits(LastGraph.LastMouseY, SpectrumPlotRanges);
        end;
      end;
      if not FoundHit and (LastGraph = G) then
        FoundHit := LookForObsToCalcHits(LastGraph, SpectrumPlotRanges);
      if not FoundHit then
        { Do Something - will show Show all and copy picture at least }
        if LastGraph = G then
          PopulateGPopup(nil)
        else
          OC.PopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
    end else if LastGraph = G then begin
      Area := G.SelectedMathArea;
      if GlobalOverlays <> nil then begin
        Area.Top := YFraction(Area.Top,G);
        Area.Bottom := YFraction(Area.Bottom,G);
        Onp := GlobalOverlays.FindPeak(Area, P, FLastPlot, Simulation.PlotUnits, SpectrumPlotRanges, Tag);
        if Onp < 0 then begin
          { Found a peak }
          ShowSimulation;
        end;
        if Onp <> 0 then
          Exit; { Scanned at least some overlays }
      end;
      if GlobalLineList <> nil then
        MeasureSimulationPeaks;
    end;
end;

procedure TPlotForm.GMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FLastGraph := Sender as TGraphD;
  if Sender = G then
    OC.AreaSelected := False
  else
    G.AreaSelected := False;
end;

function TPlotForm.LastGraph: TGraphD;
begin
  Result := FLastGraph;
  if Assigned(Result) and Result.Visible then
    Exit;
  if Result = G then
    Result := OC
  else
    Result := G;
  FLastGraph := Result;
end;

procedure TPlotForm.SimulateButtonClick(Sender: TObject);
begin
  if not (Simulation.ShowSum or Simulation.ShowParts or Simulation.ShowFortrat) then
    Simulation.ShowSum := True; { Force some display if nothing shown }
  MainForm.DoPlot;
end;

function TPlotForm.FindLimits(IncludeOverlays: Boolean): Boolean;
  procedure AddLimits(FMin, FMax: Double);
  var
    t: Double;
  begin
    if FMin > FMax then begin
      t := FMin;
      FMin := FMax;
      FMax := t;
    end;
    if not Result then begin
      Simulation.Fmin := Fmin;
      Simulation.Fmax := Fmax;
      Result := True;
    end else begin
      if Fmin < Simulation.Fmin then Simulation.Fmin := Fmin;
      if Fmax > Simulation.Fmax then Simulation.Fmax := Fmax;
    end;
  end;
begin
  Result := False;
  if not Assigned(FSimulation) then Exit;
  if (GlobalLineList <> nil) and (GlobalLineList.nLines > 0) then
    AddLimits(Convert(GlobalLineList.FMin,GlobalMixture.Units,Simulation.PlotUnits),
              Convert(GlobalLineList.FMax,GlobalMixture.Units,Simulation.PlotUnits));
  if IncludeOverlays and (GlobalOverlays <> nil) then begin
    GlobalOverlays.FixupProperties;
    if GlobalOverlays.HasLimits then
      AddLimits( Convert(GlobalOverlays.HMin,GlobalOverlays.Units,Simulation.PlotUnits)+Simulation.Foffset,
                 Convert(GlobalOverlays.HMax,GlobalOverlays.Units,Simulation.PlotUnits)+Simulation.Foffset);
  end;
  if not Result then
    if IsWavelengthUnit(Simulation.PlotUnits) then begin
      Simulation.Fmin := 1000;
      Simulation.Fmax := 100;
    end else begin
      Simulation.Fmin := Convert(0,cm1,Simulation.PlotUnits);
      Simulation.Fmax := Convert(50000,cm1,Simulation.PlotUnits)
    end
  else if Simulation.Fmin = Simulation.Fmax then begin
    Simulation.Fmin := Simulation.Fmin - 0.5;
    Simulation.Fmax := Simulation.Fmax + 0.5;
  end else begin
    Simulation.Fmin := Simulation.Fmin - (Simulation.Fmax - Simulation.Fmin)/50;
    Simulation.Fmax := Simulation.Fmax + (Simulation.Fmax - Simulation.Fmin)/50;
  end;
end;

procedure TPlotForm.PasteOverlayButtonClick(Sender: TObject);
begin
  MainForm.PasteOverlayItemClick(Self);
end;

procedure TPlotForm.FormResize(Sender: TObject);
begin
  Split.ParentResized;
  AutoSizeStatusBar(StatusBar, ['0000000000', '00000000', 'Width:', '00000000', '00000000'], True);
end;

procedure TPlotForm.FormDestroy(Sender: TObject);
begin
  if HandleAllocated then
    DragAcceptFiles(Self, False);
end;

procedure TPlotForm.FormDropFiles(Sender: TObject; const FileNames: array of String);
var
  FileType, ft: TFileType;
  Merge: Boolean;
  i: Integer;
  nObsFile: Integer; { Used in logic to load observation file last }
                     { Note that more than one obseravtion file is an error }
  Reader: TMixtureReader;
begin
  nObsFile := -1;
  { Must load overlays last, to correctly re-load correlation matrix }
  for ft := High(TFileType) downto Low(TFileType) do begin
    Merge := MainForm.MergeItem.Checked;
    for i := 0 to High(FileNames) do begin
      { Second (or subsequent) file of any given type must be loaded with merge }
      FileType := Global.FileType(FileNames[i],True,Reader);
      if FileType = ft then begin
        if FileType = fiObservations then
          nObsFile := i
        else
          LoadGeneralFile(FileNames[i],Merge,True,Simulation.PlotUnits);
        Merge := True;
      end;
    end;
  end;
  if nObsFile >= 0 then
    LoadGeneralFile(FileNames[nObsFile],False,False,Simulation.PlotUnits);
  MainForm.ConsiderReplot(GlobalMixture.nSpecies = 0);
end;

{$IFDEF VCL_}
procedure TPlotForm.WMDropFiles(var Message: TMessage);
begin
  HandleWMDropFiles(Message, FormDropFiles);
end;
{$ENDIF}

procedure TPlotForm.FormCreate(Sender: TObject);
begin
  Temperature.Tag := TemperatureIndex;
  Gaussian.Tag := GaussianIndex;
  Lorentzian.Tag := LorentzianIndex;
  FOffset.Tag := FOffsetIndex;
  {$IFDEF VCL_}
  StatusBar.AutoHint := True; { TODO: LCL equivalent }
  {$ENDIF}
  ShowSumButton.Tag := ShowSumIndex;
  ShowPartsButton.Tag := ShowPartsIndex;
  ShowFortratButton.Tag := ShowFortratIndex;
  Split.OneWindowVisible(G);
  Split.Visible := False;
  StatusBar.SimplePanel := False; { Defaults incompatible between VCL and LCL }
  {$IFNDEF VCL_}
  OnDropFiles := @FormDropFiles;
  {$ENDIF}
end;

procedure TPlotForm.GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  GD: TGraphD;
  YWidth, YRange: Double;
begin
  GD := Sender as TGraphD;
  StatusBar.Panels[0].Text := RoundedStr(GD.LastMouseX, (GD.Xmax-GD.Xmin)/Width);
  if GD.AreaSelected then begin
    YWidth := GD.SelectedMathArea.Top;
    if FLastAreaSelected and (ssLeft in Shift) and (GD.SignificantAreaSelected) then
      FLastAreaSelected := False;
  end else
    YWidth := 0;
  if GD <> G then
    StatusBar.Panels[1].Text := ShowYValue(GD, FortratPlotRanges, YWidth, YRange)
  else
    StatusBar.Panels[1].Text := ShowYValue(G, SpectrumPlotRanges, YWidth, YRange);
  if GD.AreaSelected then
    with GD.SelectedMathArea do begin
      StatusBar.Panels[3].Text := RoundedStr(Right-Left, (GD.Xmax-GD.Xmin)/Width);
      if YRange <> 0 then
        StatusBar.Panels[4].Text := RoundedStr(YWidth, YRange)
      else
        StatusBar.Panels[4].Text := '';
    end
  else begin
    StatusBar.Panels[3].Text := '';
    StatusBar.Panels[4].Text := '';
  end;
  if CalculationInProgress then
    StatusBar.Panels[2].Text := 'Busy'
  else
    StatusBar.Panels[2].Text := 'Width:'
end;

function TPlotForm.ShowYValue(GD: TgraphD; const PlotRanges: TPlotRanges; var YWidth: Double; out YRange: Double): string;
var
  Index: Integer;
  Value: Double;
  procedure ShowObservationHint;
  var
    i: Integer;
    s: string;
  begin
    s := '';
    for i := 0 to High(Observations) do
      if GD.CloseToX(GD.LastMouseX, Observations[i].Frequency)
         or GD.CloseToX(GD.LastMouseX, Observations[i].Calculated) then
        if Length(s) > 70 then begin
          s := s + '...';
          Break;
        end else
          s := s + Format(' %s:%d', [ExtractFileName(Observations[i].FileName), Observations[i].LineNumber]);
    {$IFDEF LCL_}
      Application.Hint := s;
    {$ELSE}
      G.Hint := s;
    {$ENDIF}
  end;
begin
  { Find plot relating to given YValue and display it }
  Value := GD.LastMouseY;
  YRange := 0;
  Index := FindYRange(Value, GD, PlotRanges);
  if Index > -Length(PlotRanges) then
    with PlotRanges[Abs(Index)] do
      if Tag = -2 then begin
        ShowObservationHint;
        Result := '';
        Exit;
      end else begin
        YRange := (yx-yi)/((y2-y1)*GD.Height);
        Result := RoundedStr(Value, YRange);
        YWidth := Value - FractionToValue(YFraction(YWidth,GD),GD,PlotRanges[Abs(Index)]);
      end
  else
    Result := ''; // Result := RoundedStr(GD.LastMouseY, (GD.Ymax-GD.Ymin)/GD.Height);
  G.Hint := '';
end;

procedure TPlotForm.StatusBarHint(Sender: TObject);
begin
  if CalculationInProgress then
    StatusBar.Panels[2].Text := 'Busy'
  else
    StatusBar.Panels[2].Text := 'Width:';
  if Application.Hint = '' then
    StatusBar.Panels[5].Text := FLastLineInfo
  else
    StatusBar.Panels[5].Text := Application.Hint;
end;

procedure TPlotForm.GContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
begin
  Handled := True; { Don't want default context menu }
end;

function TPlotForm.GetSimulation: TSimulation;
begin
  if not Assigned(FSimulation) then begin
    BeginGlobalUpdate;
    try
      Simulation := TSimulation.Create(GlobalMixture);
      FSimulation.Modified := False;
    finally
      EndGlobalUpdate(True);
    end;
  end;
  Result := FSimulation;
end;

procedure TPlotForm.MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
begin
  inherited;
  if (Sender = FSimulation) and (Change = mnDestroy) then begin
    FSimulation := nil;
    if Tag > 0 then
      Close;
  end;
  if (Change = mnGraphicChanged) or ((Change = mnDestroy) and (Sender is TPlot))then
    G.PlotChanged;

//  if (Sender = GlobalMixture) and (Sender <> Mixture) then begin
//    if (Change = mnParameterChange)
//       and (Pos(','+Mixture.ParameterNames[Index]+',', ',Fmin,Fmax,') <= 0) then
//      Mixture.SetParameter(Index,Sender.Parameters[Index]);
//    if (Change = mnSettingChange)
//       and (Pos(','+Mixture.SettingNames[Index]+',', ',ShowSum,ShowParts,ShowFortrat,') <= 0) then
//      Mixture.SetSetting(Index,Sender.Settings[Index]);
//  end else if (Sender = Mixture) and (Sender <> GlobalMixture) and SettingsAvailable then begin
//    if Change = mnParameterChange then begin
//      AIndex := FormSettings.IndexOfParameter(Sender.ParameterNames[Index]);
//      if AIndex >= 0 then
//        FormSettings.SetParameter(AIndex, Sender.Parameters[Index]);
//    end else if Change = mnSettingChange then begin
//      AIndex := FormSettings.IndexOfSetting(Sender.ParameterNames[Index]);
//      if AIndex >= 0 then
//        FormSettings.SetSettingFromStr(AIndex, Sender.SettingToStr(Index));
//    end;
//  end;
end;

procedure TPlotForm.FoffsetKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then begin
    Key := 0;
    ShowSimulation;
  end;
end;

procedure TPlotForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  CloseAction := caFree;
end;

procedure TPlotForm.FormShow(Sender: TObject);
begin
  DragAcceptFiles(Self, True);
end;

procedure TPlotForm.FoffsetChange(Sender: TObject);
var
  NText: string;
  Value: Double;
begin
  NText := Trim((Sender as TEdit).Text);
  SetNan(Value);
  if Assigned(Simulation) and (NText <> '') and (TryStrToFloatEx(NText, Value) or (Simulation <> GlobalMixture.Simulations[0])) then
    Simulation.SetParameter(TComponent(Sender).Tag, Value);
end;

procedure TPlotForm.ReadTextWindow(Edit: TCustomEdit; ParameterIndex: Integer);
var
  Value: Double;
begin
  if Edit.Modified and (Edit.Text <> '') then begin
    if TryStrToFloatEx(Trim(Edit.Text), Value) then
      Simulation.Parameters[ParameterIndex] := Value
    else if Simulation = GlobalMixture.Simulations[0] then
      raise EReportException.Create('Bad number: "'+Edit.Text+'"')
    else
      Simulation.Parameters[ParameterIndex] := GlobalMixture.Simulations[0].Parameters[ParameterIndex];
  end else if (Simulation = GlobalMixture.Simulations[0]) or Simulation.ParameterOverridden(ParameterIndex) then
    Edit.Text := FloatToStr(Simulation.Parameters[ParameterIndex]);
  Edit.Modified := False;
end;

end.

