unit PlotOC;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

{ Residual plot for line and line intensity fits }
interface

uses Classes, Graphs, QuantumNumber, QuantumMath, GStyle, MNode, FormUtils,
     Global, UCompat, MathSup;

  type TOCStyle = (ocWeighted, ocUnweighted, ocWeightedRelative, ocUnweightedRelative,
                   ocObserved, ocCalculated, ocObsCalc, ocWeight, ocRatio);

  TOCXStyle = (ocxNumber, ocxFrequency, ocxEnergy, ocxCalculated, ocxObserved, ocxWeight, ocxPopulation, ocxQNO,
               ocx1, ocx2, ocx3, ocx4, ocx5, ocx6, ocx7, ocx8, ocx9, ocx10, ocx11, ocx12, ocx13, oxc14, ocx15, ocx16);

  TResidualPlotter = class
  private
    FLogX: Boolean;
    FLogY: Boolean;
    procedure SetLogX(const Value: Boolean);
    procedure SetLogY(const Value: Boolean);
  protected
    FYType: TOCStyle;
    FXType: TOCXStyle;
    XLabel, YLabel: string;
    SkipPoint: Boolean;
    MarkStyle: TMarkStyle;
    nPlottedPoints: Integer;
    Title2: string;
    xv, yv: Double;
    FLastPlotFailed: Boolean;
    function GetX(Index: Integer): Double; virtual; abstract;
    function GetY(Index: Integer): Double; virtual;
    function SkipY(var O, C, W: Double; Index: Integer):Boolean; virtual; abstract;
    function IncludeYZero: Boolean; virtual;
    function EnsureCalcAvailable(Index: Integer): Boolean; virtual;
    procedure PreparePoint(Index: Integer); virtual;
    procedure AfterPoint(G: TGraphD; Index: Integer); virtual;
    function nObs:Integer; virtual; abstract;
    procedure SetYType(const Value: TOCStyle); virtual;
    procedure SetXType(const Value: TOCXStyle); virtual;
  public
    PlottedPoints: TPlottedPoints;
    FitStatus: string;
    nMouseLines: Integer;
    LastMouseLines: array [0..10] of Integer; { Lines near mouse }
    nSavedLines: Integer; { Saved after menu popup }
    SavedLines: array [0..10] of Integer;
    ShowFileNames: Boolean;
    constructor Create(AFitStatus: string); virtual;
    destructor Destroy; override;
    function ResidualAvailable(var Residual, Weight: Double; Index: Integer):Boolean; virtual; abstract;
    function OnGetPointInfo(Index: Integer; Extra: PtrInt):string; virtual; abstract;
    function SaveMouseLines: Boolean;
    procedure Plot(G: TGraphD);
    procedure SetValid(ValiD: Boolean); virtual;
    procedure EditLastPoint; virtual; abstract;
    procedure ShowLastPoint; virtual; abstract;
    procedure RemovePoints(Area: TDRect; Outside: Boolean); virtual; abstract;
    procedure RemoveLastPoints; virtual; abstract;
    property X[Index: Integer]: Double read GetX;
    property Y[Index: Integer]: Double read GetY;
    property YType: TOCStyle read FYType write SetYType;
    property XType: TOCXStyle read FXType write SetXType;
    property LogX: Boolean read FLogX write SetLogX;
    property LogY: Boolean read FLogY write SetLogY;
    property LastPlotFailed: Boolean read FLastPlotFailed;
  end;

  TResidualPlotterClass = class of TResidualPlotter;

  TObsNumberPlotter = class(TResidualPlotter)
  private
    FUpper: Boolean;
    procedure SetUpper(const Value: Boolean);
  protected
    XQNO: TQnoType;
    LastFile: string;
    nFail, LastRemove: Integer;
    function IncludeYZero: Boolean; override;
    procedure PreparePoint(Index: Integer); override;
    procedure AfterPoint(G: TGraphD; Index: Integer); override;
    function EnsureCalcAvailable(Index: Integer): Boolean; virtual;
    function SkipY(var O, C, W: Double; Index: Integer):Boolean; override;
    function nObs:Integer; override;
    procedure SetXType(const Value: TOCXStyle); override;
    function GetX(Index: Integer): Double; override;
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
  public
    ObservationsValid: Boolean;
    constructor Create(AFitStatus: string); override;
    destructor Destroy; override;
    procedure SetValid(Valid: Boolean); override;
    function ResidualAvailable(var Residual, Wt: Double; Index: Integer):Boolean; override;
    function OnGetPointInfo(Index: Integer; Extra: PtrInt):string; override;
    procedure Remove(Index: Integer);
    procedure ShowRemove;
    procedure EditLastPoint; override;
    procedure ShowLastPoint; override;
    procedure RemoveLastPoints; override;
    procedure RemovePoints(Area: TDRect; Outside: Boolean); override;
    property Upper: Boolean read FUpper write SetUpper;
  end;

//function SymmetryKey(MNode: TMNode):string;

implementation

uses SysUtils, Math, Obs, Fitter, Mixture, Gmarks, Graphics, MNodeColour,
     SlaveEdit, ResidFrm, Main, Units, SymUtils, LineFrm;

//function SymmetryKey(MNode: TMNode):string;
// { Display symmetry key of given MNode; pass Mixture for overall key }
//var
//  Molecule: TMolecule;
//  Mixture: TMixture;
//  Node: TComponent;
//  sym, i, j, LastSym, LastSymShown: Integer;
//  s1, s2: string;
//  NumbersOnly: Boolean;
//begin
//  Node := MNode;
//  Molecule := nil;
//  while Assigned(Node) do begin
//    if Node is TMolecule then
//      Molecule := Node as TMolecule;
//    Node := Node.Owner;
//  end;
//  if Assigned(Molecule) then begin
//    { Work out symmetry key for this molecule }
//    Result := 'Symmetry';
//    if Molecule.Mixture.JAdjustSym then
//      Result := Result + ' (Jeven/Jodd)';
//    Result := Result + ': ';
//    for sym := 0 to Molecule.LastSym do begin
//      Result := Result + #1 + Char(Ord(FirstMark)+sym) + #0':';
//      s1 := Molecule.JAdjustedSymName(0,sym);
//      if Molecule.Mixture.JAdjustSym then begin
//        s2 := Molecule.JAdjustedSymName(2,Molecule.JAdjustedSym(2,Sym));
//        if s2 <> s1 then
//          s1 := s1 + '/' + s2;
//        Result := Result + s1 + ' (';
//        s1 := Molecule.SymName(sym);
//        s2 := Molecule.SymName(Molecule.JAdjustedSym(2,Sym));
//        if s1 <> s2 then
//          Result := Result + s1 + '/' + s2
//        else
//          Result := Result + s1;
//        Result := Result + ')'
//      end else
//        Result := Result + s1;
//      Result := Result + ' ';
//    end;
//  end else begin
//    { Check to see of all Molecules have same symmetry name }
//    Result := '';
//    LastSym := 0;
//    LastSymShown := -1;
//    NumbersOnly := False;
//    Mixture := MNode.RootNode as TMixture;
//    for i := 0 to Mixture.nSpecies-1 do
//      with Mixture.Species[i] do
//        for j := 0 to Count-1 do begin
//          LastSym := Max(Isotopomer[j].LastSym,LastSym);
//          if not NumbersOnly then
//            if Result = '' then
//              Result := SymmetryKey(Isotopomer[j])
//            else if Result <> SymmetryKey(Isotopomer[j]) then begin
//              { Symmetry keys different - force numeric display of symmetries }
//              NumbersOnly := True;
//              Result := '';
//            end;
//          if NumbersOnly then begin
//            if Result = '' then
//              Result := 'Symmetry:';
//            while LastSymShown < LastSym do begin
//              Inc(LastSymShown);
//              Result := Result + #1 + Char(Ord(FirstMark)+LastSymShown) + #0':'
//                               + IntToStr(LastSymShown) + ' ';
//            end;
//          end;
//        end;
//  end;
//end;

{ TResidualPlotter }

procedure TResidualPlotter.AfterPoint(G: TGraphD; Index: Integer);
begin

end;

constructor TResidualPlotter.Create(AFitStatus: string);
begin
  inherited Create;
  AFitStatus := FitStatus;
  MarkStyle := TMarkStyle.Create;
end;

destructor TResidualPlotter.Destroy;
begin
  ResidualForm.DiscardResidualPlotter(Self);
  FreeAndNil(MarkStyle);
  inherited;
end;

function TResidualPlotter.EnsureCalcAvailable(Index: Integer): Boolean;
begin

end;

function TResidualPlotter.GetY(Index: Integer): Double;
var
  Observed, Calculated, Weight: Double;
begin
  SkipPoint := SkipPoint or SkipY(Observed, Calculated, Weight, Index);
  Result := 0;
  if SkipPoint then Exit;
  case YType of
    ocWeighted:           Result := (Observed - Calculated)/Abs(Weight);
    ocUnweighted:         Result := (Observed - Calculated);
    ocWeightedRelative:   if Observed <> 0 then
                            Result := (Observed - Calculated)/Abs(Weight)/Observed
                          else
                            Result := 0;
    ocUnweightedRelative: if Observed <> 0 then
                            Result := (Observed - Calculated)/Observed
                          else
                            Result := 0;
    ocObserved, ocObsCalc:Result := Observed;
    ocCalculated:         Result := Calculated;
    ocWeight:             Result := Weight;
    ocRatio:              if Observed <> 0 then
                            Result := Calculated/Observed
                          else
                            SkipPoint := True;
  end;
  if not SkipPoint and LogY then
    if Result = 0 then
      SkipPoint := True
    else
      Result := Log10(Abs(Result));
end;

function TResidualPlotter.IncludeYZero: Boolean;
begin
  Result := True;
end;

procedure TResidualPlotter.PreparePoint(Index: Integer);
begin

end;

function TResidualPlotter.SaveMouseLines: Boolean;
var
  i: Integer;
begin
  Result := nMouseLines > 0;
  for i := 0 to nMouseLines-1 do
    SavedLines[i] := LastMouseLines[i];
  nSavedLines := nMouseLines;
end;

procedure TResidualPlotter.SetLogX(const Value: Boolean);
begin
  FLogX := Value;
  SetXType(XType);
end;

procedure TResidualPlotter.SetLogY(const Value: Boolean);
begin
  FLogY := Value;
  SetYType(YType);
end;

procedure TResidualPlotter.SetValid(Valid: Boolean);
begin

end;

procedure TResidualPlotter.SetXType(const Value: TOCXStyle);
begin
  FXType := Value;
  case XType of
    ocxNumber:      XLabel := 'Observation Number';
    ocxFrequency:   XLabel := 'Frequency';
    ocxCalculated:  XLabel := 'Calculated';
    ocxObserved:    XLabel := 'Observed';
    ocxWeight:      XLabel := 'Weight';
  else
    XLabel := '';
  end;
  if LogX then
    XLabel := 'log10('+XLabel+')';
end;

procedure TResidualPlotter.SetYType(const Value: TOCStyle);
begin
  FYType := Value;
  case YType of
    ocWeighted:           YLabel := 'Weighted Obs-Calc';
    ocUnweighted:         YLabel := 'Unweighted Obs-Calc';
    ocWeightedRelative:   YLabel := 'Weighted (Obs-Calc)/Obs';
    ocUnweightedRelative: YLabel := 'Unweighted (Obs-Calc)/Obs';
    ocObserved:           YLabel := 'Observed';
    ocCalculated:         YLabel := 'Calculated';
    ocObsCalc:            YLabel := 'Observed and Calculated';
    ocWeight:             YLabel := 'Weight';
    ocRatio:              YLabel := 'Calc/Obs';
  end;
  if LogY then
    YLabel := 'log10('+YLabel+')';
end;

procedure TResidualPlotter.Plot(G: TGraphD);
var
  Xi, Xx, Yi, Yx, Resid, Wt, T: Double;
  i: Integer;
  FirstPoint, AllCalcAvailable: Boolean;
  WtSd, UnWtSD: Double;
begin
  FLastPlotFailed := False;
  SetLength(PlottedPoints, 0);
  FirstPoint := True;
  Yi := 0;
  Yx := Yi;
  WtSd := 0;
  UnWtSD := 0;
  { If no fit status available, calculate our own }
  AllCalcAvailable := FitStatus = '';
  for i := 0 to nObs-1 do begin
    SkipPoint := False;
    xv := X[i];
    yv := Y[i];
    SkipPoint := SkipPoint or IsNan(xv) or IsNan(yv);
    if not SkipPoint then
      AllCalcAvailable := AllCalcAvailable and ResidualAvailable(Resid,Wt,i);
    if not SkipPoint then begin
      if AllCalcAvailable then begin
        UnWtSD := UnWtSD + Sqr(Resid);
        WtSD := WtSD + Sqr(Resid/Wt);
      end;
      if FirstPoint then begin
        FirstPoint := False;
        Xi := xv;
        Xx := Xi;
        if IncludeYZero then
          Yi := 0
        else
          Yi := yv;
        Yx := Yi;
      end;
      if xv < Xi then Xi := xv;
      if xv > Xx then Xx := xv;
      if yv < Yi then Yi := yv;
      if yv > Yx then Yx := yv;
    end;
  end;
  if FirstPoint then begin
    G.PlotChanged;
    Exit;
  end;
  if AllCalcAvailable then begin
    FitStatus := 'Average Error: ' + FloatToStr(Sqrt(WtSd/nObs));
    if WtSd <> UnWtSD then
      FitStatus := FitStatus + '  (Unweighted: ' +  FloatToStr(Sqrt(UnWtSD/nObs)) + ')';
  end;
  if Xi = Xx then
    Xi := Xi - 1
  else
    Xi := Xi - (Xx-Xi)/40;
  if Yi = Yx then
    Yi := Yi - 1;
  T := (Yx-Yi)/10;
  Yx := Yx + T;
  Yi := Yi - T;
  G.Start;
  G.Map(Xi, Xx, Yi, Yx);
  G.Title1(FitStatus);
  G.Title2(Title2);
  G.Scales(XLabel, YLabel);
  SetLength(PlottedPoints, nObs);
  nPlottedPoints := 0;
  for i := 0 to nObs-1 do begin
    SkipPoint := False;
    xv := X[i];
    yv := Y[i];
    PreparePoint(i);
    if not SkipPoint or (MarkStyle.Colour = clHide) then
      with PlottedPoints[nPlottedPoints] do begin
        X := xv;
        Y := yv;
        Index := i;
        G.SymbolT(xv, yv, MarkStyle);
        AfterPoint(G,i);
        Inc(nPlottedPoints);
      end;
  end;
  G.Finish;
  SetLength(PlottedPoints,nPlottedPoints);
end;

{ TObsNumberPlotter }

procedure TObsNumberPlotter.AfterPoint(G: TGraphD; Index: Integer);
var
  yy, xl, xu, yl, yu: Double;
begin
  case YType of
    ocObsCalc:
      begin
        yy := Observations[Index].Calculated;
        if LogY then
          if Observations[Index].Calculated <> 0 then
            Exit
          else
            yy := Log10(Abs(yy));
        G.Point(xv, yv);
        G.Join(xv,yy);
      end;
  end;
  if ShowFileNames and (XType = ocxNumber) and ((Observations[Index].FileName <> LastFile) or (Index = nObs-1)) then begin
    G.MapArg(xl, xu, yl, yu);
    G.StyleP(clSilver, 0, lpSolid);
    G.Point(xv,yl);
    G.Join(xv,yu);
    G.TextAngle(90);
    G.Text(xv-G.Cx/2,yl,LastFile);
    G.TextAngle(0);
    LastFile := Observations[Index].FileName;
  end;
end;

constructor TObsNumberPlotter.Create(AFitStatus: string);
begin
  inherited Create(FitStatus);
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange, [mnObsCleared]);
  SetValid(True);
end;

destructor TObsNumberPlotter.Destroy;
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
  inherited;
end;

procedure TObsNumberPlotter.EditLastPoint;
begin
  if (nSavedLines > 0) and (SavedLines[0] < Length(Observations)) then
    EditObservation(Observations[SavedLines[0]])
end;

procedure TObsNumberPlotter.ShowLastPoint;
begin
  if (nSavedLines > 0) and (SavedLines[0] < Length(Observations)) then
    with Observations[SavedLines[0]] do
      if ObsType = obLine then
        MainForm.CentrePlot(Frequency, GlobalMixture.Units);
end;

procedure TObsNumberPlotter.RemoveLastPoints;
var
  i: Integer;
begin
  nFail := 0;
  LastRemove := -1;
  for i := 0 to nSavedLines-1 do
    Remove(SavedLines[i]);
  ShowRemove;
end;

procedure TObsNumberPlotter.RemovePoints(Area: TDRect; Outside: Boolean);
var
  i, nRem: Integer;
begin
  nFail := 0;
  LastRemove := -1;
  for i := 0 to nObs-1 do begin
    SkipPoint := False;
    xv := X[i];
    yv := Y[i];
    PreparePoint(i);
    if (not SkipPoint or (MarkStyle.Colour = clHide))
       and (InRange(xv, Area.Left, Area.Right)
            and InRange(yv, Area.Bottom, Area.Top))
           xor OutSide then
      Remove(i);
  end;
  ShowRemove;
end;

function TObsNumberPlotter.EnsureCalcAvailable(Index: Integer): Boolean;
var
  s:string;
begin
  with Observations[Index] do begin
    if not (ofCalculatedSet in Flags) then begin
      Result := CalculateObs(Observations[Index], Calculated, s, False, Observations[Index].ObsType);
      if Result then
        Flags := Flags + [ofCalculatedSet]
      else
        FLastPlotFailed := True;
    end;
    Result := ofCalculatedSet in Flags;
  end;
end;

function TObsNumberPlotter.GetX(Index: Integer): Double;
var
  Ev: TEigenvalues;

  function FindEv(Manifold: TManifold; J: TQuantumNumber; Sym: Shortint; Ind: Integer; IsUpper: Boolean):Integer;
  begin
    Result := 0;
    with Observations[Index] do begin
      if not Assigned(Manifold) or (ObsType in [obParameter,obExpression]) then begin
        SkipPoint := True;
        Exit;
      end;
      Result := FindEigenvaluesAndIndex(Ev, Manifold, J, Sym, Ind, Name,
                                        not IsUpper, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField);
    end;
  end;

  function FindQNO(Manifold: TManifold; J: TQuantumNumber; Sym: Shortint; Ind: Integer; IsUpper: Boolean):Double;
  begin
    Ind := FindEv(Manifold, J, Sym, Ind, IsUpper);
    if Ind <= 0 then
      Result := 0
    else if XType < ocxQNO then
      Result := Ev.E[Ind]
    else
      Result := Ev.DominantBasis[Ind].Qno(XQNO) / 2
  end;

  function FindPop(Manifold: TManifold; J: TQuantumNumber; Sym: Shortint; Ind: Integer; IsUpper: Boolean): Double;
  var
    Err: Exception;
  begin
    Result := 0;
    Ind := FindEv(Manifold, J, Sym, Ind, IsUpper);
    if Ind <= 0 then
      SkipPoint := True
    else begin
      Err := nil;
      Result := Manifold.Population(Ev, Ind, Manifold.Mixture.Simulations[0], Err, 0);
      if Assigned(Err) then begin
        SkipPoint := True;
        Err.Free;
      end;
    end;
  end;

var
  XSubType: TOCXStyle;
begin
  if not (XType in [ocxNumber,ocxObserved,ocxWeight]) then
    SkipPoint := SkipPoint or not EnsureCalcAvailable(Index);
  Result := 0;
  if SkipPoint then Exit;
  if XType < ocxQNO then
    XSubType := XType
  else
    XSubType := ocxQNO;
  with Observations[Index] do
   case XSubType of
    ocxNumber:     Result := Index+1;
    ocxFrequency:  if not (ObsType in [obEUpper, obELower]) then
                     Result := FindQNO(ManifoldU, JU, SymU, AIndexU, True) - FindQNO(ManifoldL, JL, SymL, AIndexL, False)
                   else
                     Result := Frequency; { Avoid crash }
    ocxCalculated: Result := Calculated;
    ocxObserved:   Result := Frequency;
    ocxWeight:     Result := Weight;
    ocxPopulation: Result := FindPop(ManifoldL, JL, SymL, IndexL, False);
    ocxEnergy, ocxQNO:
      if Upper and (ObsType <> obELower) then
        Result := FindQNO(ManifoldU, JU, SymU, AIndexU, True)
      else
        Result := FindQNO(ManifoldL, JL, SymL, AIndexL, False)
  else
    Result := 0;
  end;
  if not SkipPoint and LogX then
    if Result = 0 then
      SkipPoint := True
    else
      Result := Log10(Abs(Result));
end;

function TObsNumberPlotter.IncludeYZero: Boolean;
begin
  Result := YType in [ocWeighted, ocUnweighted, ocWeightedRelative, ocUnweightedRelative];
end;

procedure TObsNumberPlotter.MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
begin
  if Change = mnObsCleared then
    SetValid(False);
end;

function TObsNumberPlotter.nObs: Integer;
begin
  Result := Length(Observations);
end;

function TObsNumberPlotter.OnGetPointInfo(Index: Integer; Extra: PtrInt): string;
begin
  if ObservationsValid and (Index >=0) and (Index < Length(Observations)) then begin
    if ofCD in Observations[Index].Flags then
      Result := Observations[Index].Name
    else
      Result := Format(' %s:%d', [ExtractFileName(Observations[Index].FileName), Observations[Index].LineNumber]);
    if nMouseLines < Length(LastMouseLines) then begin
      LastMouseLines[nMouseLines] := Index;
      Inc(nMouseLines);
    end;
  end else begin
    nSavedLines := 0;
    Result := '';
  end;
end;

procedure TObsNumberPlotter.PreparePoint(Index: Integer);
begin
  inherited;
  if Index = 0 then
    LastFile := Observations[Index].FileName;
  MarkStyle.Colour := FindObservationColour(Observations[Index], Upper);
  with Observations[Index] do begin
    if ObsType in [obParameter,obExpression] then
      MarkStyle.MarkType := FirstMark
    else if (ObsType = obEUpper) or (Upper and (ObsType <> obELower)) then
      MarkStyle.MarkType := MarkTypeFor(ManifoldU.Molecule.JDeAdjustedSym(JU,SymU))
    else
      MarkStyle.MarkType := MarkTypeFor(ManifoldL.Molecule.JDeAdjustedSym(JL,SymL));
  end;
end;

procedure TObsNumberPlotter.Remove(Index: Integer);
begin
  if Index < Length(Observations) then
    with Observations[Index] do
      if (FileName <> LineListFileName) then
        Inc(nFail)
      else begin
       LineForm.UnAssignLine(LineNumber);
       if LastRemove < 0 then
         LastRemove := LineNumber;
      end;
end;

procedure TObsNumberPlotter.ShowRemove;
begin
  if nFail > 0 then
    ShowInfo(IntToStr(nFail)+' points not removed - file editing not implemented');
  if LastRemove >= 0 then begin
    with LineForm do begin
      HighlightRow(LastRemove);
      FitButtonClick(TestFitButton); { Note this will discard the current object, so must be last line in the routine }
    end;
  end;
end;


function TObsNumberPlotter.ResidualAvailable(var Residual, Wt: Double; Index: Integer): Boolean;
begin
  with Observations[Index] do begin
    Wt := Weight;
    Result := ofCalculatedSet in Flags;
    if Result then
      Residual := Frequency - Calculated;
  end;
end;

procedure TObsNumberPlotter.SetUpper(const Value: Boolean);
begin
  FUpper := Value;
  SetXType(XType);
end;

procedure TObsNumberPlotter.SetValid(Valid: Boolean);
begin
  inherited;
  ObservationsValid := Valid;
end;

procedure TObsNumberPlotter.SetXType(const Value: TOCXStyle);
var
  Prefix: string;
begin
  FXType := Value;
  if Length(Observations) > 0 then
    with Observations[0] do
     if Assigned(ManifoldL) then
       Title2 := SymmetryKey(ManifoldL,False)
     else if Assigned(ManifoldU) then
       Title2 := SymmetryKey(ManifoldU,False);
  if Upper then
    Prefix := 'Upper State '
  else
    Prefix := 'Lower State ';
  if Value >= ocxQNO then begin
    XQNO := TQnoType(Ord(Value)-Ord(ocxQNO));
    XLabel := Prefix + QnoTypeNames[Ord(XQNO)];
  end else begin
    case XType of
      ocxEnergy:     XLabel := Prefix + 'Energy';
      ocxPopulation: XLabel := 'Population';
    else
      inherited SetXType(Value);
      Exit;
    end;
  end;
  if LogX then
    XLabel := 'log10('+XLabel+')';
end;

function TObsNumberPlotter.SkipY(var O, C, W: Double; Index: Integer): Boolean;
begin
  if not (YType in [ocObserved,ocWeight]) then
    Result := not EnsureCalcAvailable(Index)
  else
    Result := False;
  if not Result then
    with Observations[Index] do begin
      O := Frequency;
      C := Calculated;
      W := Weight;
    end;
end;

end.
