unit PopFunc;
{ Methods for calculating custom populations, widths and state dependent lineshapes }

{$I utils.inc}{Copyright (C) 2010-15 C M Western; see commandline.pas for license}
interface

uses Classes, SysUtils, MNode, MNodeExpr, Mixture, Constrain, Expr, pgDOM, Units;

type

  { TCustomPopulationFunction }

  TCustomPopulationFunction = class(TPopulationFunction)
  { Custom population, using expression taken from comment field }
  public
    class function CanBeChildOf(MNode: TMNode): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure Population(var Result: Double; Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer); override;
  end;

  { TCustomWidthFunction }

  TCustomWidthFunction = class(TWidthFunction)
  public
    class function CanBeChildOf(MNode: TMNode): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure Width(var Result: Double; Eigenvalues: TEigenvalues; i: Integer); override;
  end;

  { TCustomLineshapeFunction }

  TCustomLineshapeFunction = class(TLineshapeFunction)
  public
    xVariable, StrengthVariable, FrequencyVariable: TVariable;
    class function CanBeChildOf(MNode: TMNode): Boolean; override;
    procedure SetUpVariables; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure PrepareLineshape(Eigenvalues: TEigenvalues; i: Integer); override;
    function LineshapeValue(x: Double): Double; override;
  end;

implementation

uses QuantumMath, StrUtils, Worker, QuantumNumber;

{ TCustomLineshapeFunction }

class function TCustomLineshapeFunction.CanBeChildOf(MNode: TMNode): Boolean;
begin
  Result := MNode is TState;
end;

procedure TCustomLineshapeFunction.SetUpVariables;
begin
  inherited SetUpVariables;
  V('x', xVariable);
  V('Strength', StrengthVariable);
  V('Frequency',FrequencyVariable);
end;

function TCustomLineshapeFunction.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'If true, use this to calculate the lineshape of each level'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TCustomLineshapeFunction.PrepareLineshape(Eigenvalues: TEigenvalues; i: Integer);
var
  NotUsed: Double;
begin
  SetVariables(NotUsed, Eigenvalues, i);
end;

function TCustomLineshapeFunction.LineshapeValue(x: Double): Double;
begin
  xVariable.Value := x;
  Result := Evaluate;
end;

{ TCustomWidthFunction }

class function TCustomWidthFunction.CanBeChildOf(MNode: TMNode): Boolean;
begin
  Result := MNode is TState;
end;

function TCustomWidthFunction.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'If true, use this to calculate the width of each level'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TCustomWidthFunction.Width(var Result: Double; Eigenvalues: TEigenvalues; i: Integer);
begin
  SetVariables(Result, Eigenvalues, i);
  Result := Evaluate;
end;

{ TCustomPopulationFunction }

function TCustomPopulationFunction.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'If true, use this to generate the partition function'
  else
    Result:=inherited HelpOnSetting(AIndex);
end;

class function TCustomPopulationFunction.CanBeChildOf(MNode: TMNode): Boolean;
begin
  Result := MNode is TManifold;
end;

procedure TCustomPopulationFunction.Population(var Result: Double;
  Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation;
  var Overflow: Exception; IncludeTspin: Integer);
begin
  SetVariables(Result, Eigenvalues, i);
  Result := Evaluate;
end;

initialization
  RegisterClass(TCustomPopulationFunction);
  TTable.RegisterClass(TCustomPopulationFunction);
  RegisterClass(TCustomWidthFunction);
  TTable.RegisterClass(TCustomWidthFunction);
  RegisterClass(TCustomLineshapeFunction);
  TTable.RegisterClass(TCustomLineshapeFunction);
end.
