unit Progress;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

{$IFNDEF CONSOLE}
uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, Buttons, Worker;

type

  { TProgressForm }

  TProgressForm = class(TForm)
    WasCalculating: TLabel;
    ProgressBar: TProgressBar;
    Button: TButton;
    Calculating: TLabel;
    ValueLabel: TLabel;
    procedure ButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    ShouldBeVisible: Boolean;
  public
    procedure CheckVisibility(Sender: TObject);
  end;

var
  ProgressForm: TProgressForm;

procedure CheckMemoryLimit(Extra: Int64 = 0);

{$ENDIF}

function CheckForAbort:Boolean;

procedure ResetProgress;
procedure PushProgress(Calculating: string; AFinalPosition: Integer = -1);
procedure PopProgress;
procedure SetProgress(Position: Integer; AFinalPosition: Integer = -1; DoCheckForAbort: Boolean = True);
function CurrentProgress: Integer;

{$IFDEF TRIGGERABORT}
procedure TriggerAbort;

var
  TriggerAbortCount: Integer = 0;
  InitialnTriggers: Integer = 1000+20*747;
  StepnTriggers: Integer = 747;
{$ENDIF}


implementation

uses {$IFNDEF CONSOLE}
       {$IFDEF LCL_} {$ENDIF}
       {$IFDEF VCL_STACK} StackTrace, {$ENDIF}
       {$IFDEF MSWINDOWS} Huge, {$ENDIF}
     {$ENDIF}
     Clock, Sysinfo;

{$IFNDEF CONSOLE}
{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

var
  Level: Integer = -1;
  FinalPosition: array of Integer;
  CalcName: array of string;

procedure TProgressForm.ButtonClick(Sender: TObject);
{$IFDEF VCL_STACK}
var
  St: TCallStack;
{$ENDIF}
begin
  if ShouldBeVisible then begin
    {$IFDEF VCL_STACK}
    SaveThreadStatus;
    FillCallStack(St,8);
    WriteLn(Log,CallStackTextualRepresentation(St,''));
    ShowThreadStatus(Log);
    {$ENDIF}
    Aborted := True
  end else
    CheckVisibility(nil);
end;

procedure TProgressForm.CheckVisibility(Sender: TObject);
begin
  {$IFDEF MSWINDOWS}
  if HandleAllocated and (ShouldBeVisible <> IsWindowVisible(Handle)) then begin
    { This is needed as hide is ineffective if the application is minimized }
    Visible := not ShouldBeVisible;
    Visible := ShouldBeVisible;
  end;
  {$ENDIF}
end;

procedure TProgressForm.FormDestroy(Sender: TObject);
begin
  {$IFDEF MSWINDOWS}
  Application.OnRestore := nil;
  {$ENDIF}
end;

{$ENDIF}

var
  NextUpdate: Int64;
  ShortDelay: Integer;
  FCurrentProgress: Integer;

function CurrentProgress: Integer;
begin
  Result := FCurrentProgress;
end;

procedure ResetProgress;
begin
  {$IFNDEF CONSOLE}
  Aborted := False;
  Level := -1;
  if GetCurrentThreadId = MainThreadID then begin
    ProgressForm.ShouldBeVisible := False;
    ProgressForm.Hide;
    ProgressForm.Button.Enabled := True;
  end;
  {$ENDIF}
end;

{$IFNDEF CONSOLE}
var
  LargeOK: Boolean = False;

procedure CheckMemoryLimit(Extra: Int64);
var
  Cur: Int64;
begin
  if LargeOK or Aborted or (GetCurrentThreadId <> MainThreadID) then
    Exit;
  Cur := MemoryInUse+Extra;
  if Cur > MemoryLimit then
    if MessageDlg(Format('Currently using %.1fG of memory. Continue calculation?',[Cur/1024/1024/1024]), mtInformation, [mbYes, mbAbort], 0) = mrYes then
      LargeOK := True
    else begin
      Aborted := True;
      CheckForAbort;
    end;
end;

{$ENDIF}

function CheckForAbort:Boolean;
begin
  {$IFNDEF CONSOLE}
  Result := Aborted;
  if Result then
    if GetCurrentThreadId <> MainThreadID then
      Abort
    else begin
      if (nWorkers > 0) and (nFreeWorkers < nWorkers) then begin
        Level := 0;
        CalcName[0] := 'Aborting';
        FinalPosition[0] := nWorkers;
        ProgressForm.Button.Enabled := False;
        while nFreeWorkers < nWorkers do begin
          FCurrentProgress := nFreeWorkers;
          WaitAllWorkersFree;
        end;
      end;
      ResetProgress;
      WriteLn(Log,'Abort triggered in '+ThreadName);
      AbortCalculation;
    end;
  {$ELSE}
  Result := False;
  {$ENDIF}
end;

procedure PushProgress(Calculating: string; AFinalPosition: Integer = -1);
begin
  {$IFNDEF CONSOLE}
  CheckForAbort;
  if GetCurrentThreadId <> MainThreadID then
    Exit;
  Inc(Level);
  if Level > High(FinalPosition) then begin
    SetLength(FinalPosition, Level+10);
    SetLength(CalcName, Level+10);
  end;
  FinalPosition[Level] := AFinalPosition;
  CalcName[Level] := Calculating;
  if Level = 0 then begin
    ShortDelay := Round(ClockFrequency/5);
    NextUpdate := Round(ClockFrequency) + Clocks();
  end;
  {$ENDIF}
end;

procedure PopProgress;
begin
  {$IFNDEF CONSOLE}
  CheckForAbort;
  if GetCurrentThreadId <> MainThreadID then
    Exit;
  Dec(Level);
  if Level < 0 then
    ResetProgress;
  {$ENDIF}
end;

procedure SetProgress(Position: Integer; AFinalPosition: Integer; DoCheckForAbort: Boolean);
var
  CurrentClock: Int64;
begin
  {$IFNDEF CONSOLE}
  if DoCheckForAbort then
    CheckForAbort;
  if GetCurrentThreadId <> MainThreadID then
    Exit;
  FCurrentProgress := Position;
  if Level < 0 then
    Exit;
  if AFinalPosition > 0 then
    FinalPosition[Level] := AFinalPosition;
  if Position > FinalPosition[Level] then
    FinalPosition[Level] := FinalPosition[Level] + FinalPosition[Level] div 10 + 2;
  if FinalPosition[Level] > 0 then begin
    CurrentClock := Clocks();
    if CurrentClock > NextUpdate then begin
      CheckMemoryLimit;
      NextUpdate := CurrentClock + ShortDelay;
      if not ProgressForm.Visible then begin
        {$IFDEF MSWINDOWS}
        Application.OnRestore := {$IFDEF FPC}@{$ENDIF}ProgressForm.CheckVisibility;
        {$ENDIF}
        ProgressForm.Show;
      end;
      ProgressForm.ShouldBeVisible := True;
      if Screen.ActiveForm <> ProgressForm then
        ProgressForm.BringToFront;
      ProgressForm.Calculating.Caption := CalcName[Level];
      if Level > 0 then
        ProgressForm.WasCalculating.Caption := CalcName[Level-1]
      else
        ProgressForm.WasCalculating.Caption := '';
      ProgressForm.ProgressBar.Max := FinalPosition[Level];
      ProgressForm.ProgressBar.Position := Position;
      ProgressForm.ValueLabel.Caption := IntToStr(Position);
      Application.ProcessMessages;
    end;
    {$IFDEF TRIGGERABORT}
      if TriggerAbortCount > 0 then begin
        Dec(TriggerAbortCount);
        if TriggerAbortCount = 0 then
          Aborted := True;
      end;
    {$ENDIF}
  end;
  {$ENDIF}
end;

{$IFDEF TRIGGERABORT}
procedure TriggerAbort;
  { Triggers abort after a determined number of SetProgress calls }
  TriggerAbort := InitialnTriggers;
  InitialnTriggers := InitialnTriggers+ StepnTriggers;
end;
{$ENDIF}

end.
