unit ReadBinary;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, Messages,  {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms, Buttons, Worker,
  Dialogs, StdCtrls, AlEdit, EditExtr, NewSpin, ExtCtrls, ExpPlotLoader;

type
  TBinaryImportDlg = class(TForm)
    BytesPerValueSpin: TSpinBox;
    Label1: TLabel;
    MSBCheck: TCheckBox;
    OKButton: TButton;
    CancelButton: TButton;
    Label2: TLabel;
    SkipSpin: TSpinBox;
    SignedCheck: TCheckBox;
    TypeRadioGroup: TRadioGroup;
  end;

var
  BinaryImportDlg: TBinaryImportDlg;

type

  { TBinaryFileLoader }

  TBinaryFileLoader = class(TExperimentalPlotLoader)
  private
    nBytes: Integer;
    NumberType: Integer;
    SkipBytes: Integer;
    procedure ReadBody;
  public
    function Load: Boolean; override;
  end;


implementation

uses MathSup, UCompat, MNode;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

type
  Tm = record
    case Integer of
      0: (i64: Int64);
      1: (r4: Single);
      2: (r6: Real48);
      3: (r8: Double);
      4: (r10: Extended);
      5: (u8: Byte);
      6: (i8: Shortint);
      7: (u16: Word);
      8: (i16: Smallint);
      9: (u32: LongWord);
     10: (i32: Longint);
    end;

{ TBinaryFileLoader }

procedure TBinaryFileLoader.ReadBody;
var
  n, d, l, nVal: Integer;
  v: Tm;
  p, pi: ^Byte;
  Signed: Boolean;
  Exponent: Integer;
  RV: Double;
begin
  with BinaryImportDlg do begin
    Stream.Position := SkipBytes;
    Signed := SignedCheck.Checked;
    pi := Pointer(@v);
    d := 1;
    if {$IFDEF FPC_BIG_ENDIAN} not {$ENDIF} MSBCheck.Checked then begin
      Inc(pi, nBytes-1);
      d := -1;
    end;
    l := 0;
    n := 0;
    p := pi;
    v.i64 := 0;
    nVal := 0;
    while Stream.Read(p^,1) = 1 do begin
      Inc(n);
      Inc(p,d);
      if n = nBytes then begin
        case NumberType of
        0: { Integer }
          if Signed then
            case nBytes of
              1: RV := v.i8;
              2: RV := v.i16;
              3,4: RV := v.i32;
              5..8: RV := v.i32;
            end
          else
            case nBytes of
              1: RV := v.u8;
              2: RV := v.u16;
              3,4: RV := v.u32;
              5..8: RV := v.u32;
            end;
        1: { IEEE Float }
          case nBytes of
            4: RV := v.r4;
            6: RV := v.r6;
            8: RV := v.r8;
           10: RV := v.r10;
          end;
        2: { Other Float}
          case nBytes of
            4: { Restuff to IEEE }
               if v.u32 = 0 then
                 RV := 0
               else begin
                 Exponent := (v.i32 shr 23) - 15;
                 v.u32 := (v.u32 and not $7f800000) or ((Exponent and $ff) shl 23);
                 RV := v.r4;
               end;
          else
            raise Exception.Create('Only 4 bytes implemented for other float');
          end;
        end;
        if not IsNanOrInf(RV) then begin
          SetValue(-1, 0, nVal, l);
          SetValue(-1, 1, nVal, RV);
          Inc(nVal);
        end;
        Inc(l);
        n := 0;
        p := pi;
        v.i64 := 0;
      end;
    end;
  end;
  Connector.SetLength(0,nVal,False,True,True);
end;

function TBinaryFileLoader.Load: Boolean;
var
  N: Integer;
begin
  Result := False;
  if BinaryImportDlg.ShowModal <> mrOk then
    Exit;
  with BinaryImportDlg do begin
    nBytes := BytesPerValueSpin.Value;
    NumberType := TypeRadioGroup.ItemIndex;
    if (NumberType = 1) and not (nBytes in [4,6,8,10]) then
      raise EReportException.Create('Floating point values must be 4, 6, 8 or 10 bytes');
    SkipBytes := SkipSpin.Value;
  end;
  N := (Stream.Size - SkipBytes) div nBytes;
  SetLength(ChannelDest,2);
  ChannelDest[0] := 0; { Frequency channel }
  ChannelDest[1] := 1; { data channel }
  CheckAvailableChannels(Length(ChannelDest), N, AskAboutWeights);
  ReadBody;
  Result := True;
End;

end.
