unit ResidFrm;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms, ActnList,
  Dialogs, ExtCtrls, ScrlCtrl, LogControl, Buttons, StdCtrls, AlEdit, EditExtr,
  NewSpin, MNode, SettingsForm, Split, Graphs, PlotOC, ComCtrls, Menus,
  Obs, Mixture, GraphI, GraphM, MathSup;

type

  { TLogForm }

  { TResidualForm }

  TResidualForm = class(TSettingsForm)
    G: TGraphD;
    OCPanel: TPanel;
    YAxis: TComboBox;
    Label2: TLabel;
    Label3: TLabel;
    XAxis: TComboBox;
    Upper: TCheckBox;
    ReloadButton: TSpeedButton;
    ZoomButton: TSpeedButton;
    ZoomOutButton: TSpeedButton;
    ResetZoomButton: TSpeedButton;
    StatusBar: TStatusBar;
    GPopupMenu: TPopupMenu;
    EditPointItem: TMenuItem;
    LoadAll: TCheckBox;
    ShowTransitionItem: TMenuItem;
    LogY: TCheckBox;
    LogX: TCheckBox;
    ShowLabels: TCheckBox;
    ShowandEditItem: TMenuItem;
    RemovePointsOutsideItem: TMenuItem;
    RemovePointsInsideItem: TMenuItem;
    RemovePointsItem: TMenuItem;
    procedure UpperClick(Sender: TObject);
    procedure PopupMenuPopup(Sender: TObject);
    procedure EditPointItemClick(Sender: TObject);
    procedure GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure ResetZoomButtonClick(Sender: TObject);
    procedure ZoomOutButtonClick(Sender: TObject);
    procedure ZoomButtonClick(Sender: TObject);
    procedure YAxisChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ReloadButtonClick(Sender: TObject);
    procedure ShowTransitionItemClick(Sender: TObject);
    procedure ShowandEditItemClick(Sender: TObject);
    procedure GMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure GMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure RemovePointsOutsideItemClick(Sender: TObject);
    procedure RemovePointsInsideItemClick(Sender: TObject);
    procedure RemovePointsItemClick(Sender: TObject);
  private
    FResidualPlotter: TResidualPlotter;
    FLastAreaSelected: Boolean;
    FLastSelectedMathArea: TDRect;
    procedure SetResidualPlotter(const Value: TResidualPlotter);
  public
    procedure ArrangeForm(Screen: TRect); override;
    procedure PlotResiduals(FitStatus: string);
    procedure DiscardResidualPlotter(const Value: TResidualPlotter);
    procedure Loaded; override;
    property ResidualPlotter: TResidualPlotter read FResidualPlotter write SetResidualPlotter;
  end;

var
  ResidualForm: TResidualForm;

implementation

uses TextAssign, Fitter, Huge, ClipBrd, WCompat, SlaveEdit, UCompat,
     Global, Main, QuantumNumber, QuantumMath, FormUtils, CalFrm, Math, CommandLine,
     LogUnit, Types;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

procedure TResidualForm.EditPointItemClick(Sender: TObject);
begin
  if Assigned(ResidualPlotter) then
    ResidualPlotter.EditLastPoint
end;

procedure TResidualForm.RemovePointsItemClick(Sender: TObject);
begin
  if Assigned(ResidualPlotter) then
    ResidualPlotter.RemoveLastPoints;
end;

procedure TResidualForm.ShowandEditItemClick(Sender: TObject);
begin
  ShowTransitionItemClick(nil);
  EditPointItemClick(nil);
end;

procedure TResidualForm.ShowTransitionItemClick(Sender: TObject);
begin
  if Assigned(ResidualPlotter) then
    ResidualPlotter.ShowLastPoint
end;

procedure TResidualForm.DiscardResidualPlotter(const Value: TResidualPlotter);
begin
  { Called when residual plotter being deleted }
  if FResidualPlotter = Value then begin
    FResidualPlotter := nil;
    UpperClick(nil);
  end
end;

procedure TResidualForm.ArrangeForm(Screen: TRect);
var
  R, MainRect: TRect;
begin
  inherited;
  { Bottom left corner }
  MainRect := ScreenBounds(MainForm);
  R := Rect(MainRect.Right, MainRect.Bottom, Screen.Right, Screen.Bottom);
  with R do begin
    if Right - Left < 200 then
      Left := Right - 200;
    if Right - Left > 700 then
      Right := Left + 700;
    if Bottom-Top < 200 then
      Top := Bottom - 200;
    if Bottom-Top > 700 then
      Bottom := Top + 700;
  end;
  SetScreenBounds(Self,R);
end;

procedure TResidualForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(FResidualPlotter);
end;

procedure TResidualForm.SetResidualPlotter(const Value: TResidualPlotter);
begin
  FreeAndNil(FResidualPlotter);
  FResidualPlotter := Value;
  UpperClick(nil);
end;

procedure TResidualForm.UpperClick(Sender: TObject);
begin
  if ResidualPlotter is TObsNumberPlotter then
    TObsNumberPlotter(ResidualPlotter).Upper := Upper.Checked;
  YAxisChange(Sender);
end;

procedure TResidualForm.PlotResiduals(FitStatus: string);
begin
  ResidualPlotter := TObsNumberPlotter.Create(FitStatus);
  if ResidualPlotter.LastPlotFailed then begin
    ReShow(ResidualForm);
    raise EReportException.Create('Failure in plotting residuals - see log for details');
  end;
end;

procedure TResidualForm.PopupMenuPopup(Sender: TObject);
begin
  if Assigned(ResidualPlotter) then
    EditPointItem.Enabled := ResidualPlotter.SaveMouseLines
  else
    EditPointItem.Enabled := False;
  ShowTransitionItem.Enabled := EditPointItem.Enabled;
  ShowandEditItem.Enabled := EditPointItem.Enabled;
  RemovePointsOutsideItem.Enabled := FLastAreaSelected;
  RemovePointsInsideItem.Enabled := FLastAreaSelected;
  if FLastAreaSelected then
    G.SelectedMathArea := FLastSelectedMathArea;
end;

procedure TResidualForm.YAxisChange(Sender: TObject);
begin
  FLastAreaSelected := False;
  if Assigned(ResidualPlotter) and Visible then begin
    ResidualPlotter.XType := TOCXStyle(XAxis.ItemIndex);
    ResidualPlotter.YType := TOCStyle(YAxis.ItemIndex);
    ResidualPlotter.LogX := LogX.Checked;
    ResidualPlotter.LogY := LogY.Checked;
    ResidualPlotter.ShowFileNames := ShowLabels.Checked;
    ResidualPlotter.Plot(G);
  end else
    G.PlotChanged;
end;

procedure TResidualForm.ZoomButtonClick(Sender: TObject);
begin
  G.ZoomSelected(True);
end;

procedure TResidualForm.ZoomOutButtonClick(Sender: TObject);
begin
  G.ZoomOut;
end;

procedure TResidualForm.ReloadButtonClick(Sender: TObject);
var
  TakeWeight: TTakeWeight;
begin
  if not SaveAllWindows then Exit;
  if LoadAll.Checked then
    TakeWeight := tkAll
  else
    TakeWeight := tkObs;
  LogForm.DoFit(LogForm.ObsFileName.FileName, True, TakeWeight, 0, 1);
end;

procedure TResidualForm.RemovePointsInsideItemClick(Sender: TObject);
begin
  if FLastAreaSelected and Assigned(ResidualPlotter) then
    ResidualPlotter.RemovePoints(FLastSelectedMathArea, False);
end;

procedure TResidualForm.RemovePointsOutsideItemClick(Sender: TObject);
begin
  if FLastAreaSelected and Assigned(ResidualPlotter) then
    ResidualPlotter.RemovePoints(FLastSelectedMathArea, True);
end;

procedure TResidualForm.ResetZoomButtonClick(Sender: TObject);
begin
  G.ResetZoom;
end;

procedure TResidualForm.GMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  G.AreaSelected := False;
end;

procedure TResidualForm.GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(ResidualPlotter) then begin
    ResidualPlotter.nMouseLines := 0;
    StatusBar.SimpleText := ShowGraphInfo(G, ResidualPlotter.PlottedPoints, {$IFDEF FPC}@{$ENDIF}ResidualPlotter.OnGetPointInfo);
  end;
  if FLastAreaSelected and (ssLeft in Shift) and (G.SignificantAreaSelected) then
    FLastAreaSelected := False;
end;

procedure TResidualForm.GMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) then begin
    FLastAreaSelected := G.SignificantAreaSelected;
    FLastSelectedMathArea := G.SelectedMathArea;
  end;
end;

procedure TResidualForm.Loaded;
var
  i: Integer;
begin
  for i := 0 to High(QnoTypeNames)-1 do
    XAxis.Items.Add(QnoTypeNames[i]);
  G.StuffPopupMenu(GPopupMenu);
  inherited;
end;

end.

