unit Skeleton;

interface

uses QuantumNumber;

procedure SkeletonFunction(Param2: string);

implementation

uses SysUtils, MNode, HundA, Mixture, Overlay, JPLCat, Vib, MatrixElements, SampleMols,
     { ExpBitmap, } ExpPlot, OAsymT, SymT, Worker, Global, UCompat, PartFunc, OscFields,
     zMatrix, InternalCoords, VibPart, LoomisWood, nJ, CalSource, MNodeExpr, ExpBitmap,
     PopFunc, TranFunc, ExoMol, Settings, ExternalList, SpectrumPlot, I2;

var
  IndFile: Text;
  Context: string;

function HundTransformab(Jp, J, Omp, Om, Lamp, Lam, Np, N, k, S: TQuantumNumber):Double;
begin
    Result := Minus(Jp-Omp) / threej(Jp, -Omp, k, Omp-Om, J, Om)
            * Minus(Np-N+S+Omp-Om+J+Lamp+k) * (N+1) * (Np+1)
            * threej(Jp, Omp, S, -(Omp-Lamp), Np, -Lamp)
            * threej(J,  Om,  S, -(Om-Lam),   N,  -Lam)
            * sixj(Np, Jp, S, J, N, k)
            * threej(Np, -Lamp, k, Lamp-Lam, N, Lam);
WriteLn(Log, Jp:4, J:4, Omp:4, Om:4, Lamp:4, Lam:4, Np:4, N:4, k:4, Omp-Om:4, S:4, ' ', Result);
end;

procedure SkeletonFunctionOff(Param2: string);
var
  Jp, J, Omp, Om, Lamp, Lam, Np, N, k, q, S: TQuantumNumber;
  Sum, ME: Double;
begin
  Jp := 5;
  J := 3;
  S := 1;
  Lamp := 2;
  Lam := 2;
  k := 2;
  Omp := Lamp-S;
  repeat
    Om := Lam-S;
    repeat
      Sum := 0;
      N := J-S;
      repeat
        Np := Jp-S;
        repeat
          if Np = 6 then
            if N = 4 then
              ME := 0.009025740
            else
              ME := 0
          else if NP = 4 then
            if N = 4 then
              ME := 0.009791540
            else if N = 2 then
              ME := 0.009276870
            else
              ME := 0
          else
            ME := 0;
          if (ME = 0) and (Abs(Np-N) <= 2) then
            raise Exception.Create('Help');
          Sum := Sum + HundTransformab(Jp, J, Omp, Om, Lamp, Lam, Np, N, k, S)*ME;
          Inc(Np,2);
        until Np > Jp+S;
        Inc(N,2);
      until N > J+S;
      WriteLn(Log, Omp:4, Om:4, ' ', FloatToStr(Sum));
      Inc(Om, 2);
    until Om > 3;
    Inc(Omp,2);
  until Omp > 3;
end;

procedure SkeletonFunctionA(Param2: string);
var
  S, Lam, Om, J, N: TQuantumNumber;
  V: Double;
begin
  S := 1;
  Lam := 2;
  J := 3;
  repeat
    Om := Lam-S;
    repeat
      N := J-S;
      repeat
        V := Minus(N-S+Om) * Sqrt((N+1))
             * threej(J, Om, S, -(Om-Lam), N, -Lam);
        WriteLn(Log, J:4, Om:4, N:4, ' ', FloatToStr(V));
        Inc(N,2);
      until N > J+S;
      Inc(Om,2);
    until Om > Lam+S;
    WriteLn(Log);
    Inc(J,2);
  until J > 5;
end;

procedure WriteSkeleton(MNodeClass: TMNodeClass; StandAlone: Boolean = True; AltFileName: string = '');
var
  Name, LocalContext: string;
  MNode: TMNode;
  i: Integer;
begin
  MNode := nil;
  if (GlobalMixture <> nil) and not StandAlone then
    MNode := GlobalMixture.ComponentWithName('*',MNodeClass,nil,True);
  if not Assigned(MNode) and StandAlone then
    try
      MNode := MNodeClass.Create(nil);
    except
      Exit;
    end;
  if not Assigned(MNode) then
    Exit;
  Name := Copy(MNode.ClassType.ClassName,2,255);
  if AltFileName = '' then
    AltFileName := LowerCase(Name) + '.htm';
  AssignFile(Output,'skel/' + AltFileName);
  Rewrite(Output);
  Writeln('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">');
  Writeln('<html>');
  Writeln('<head>');
  Writeln('<title>', Name, '</title>');
  Writeln('<body>');
  Writeln('<h1>', Name, '</h1>');
  if MNode.nSettings > 0 then begin
    Writeln('<h2>Settings</h2>');
    Writeln('<table border="1" cellpadding="2" cellspacing="2"><tbody>');
    for i := 1 to MNode.nSettings do begin
      Writeln('<tr><td><a name="', MNode.SettingNames[i], '"></a>', MNode.SettingNames[i], '</td><td>', MNode.HelpOnSetting(i),'</td>');
    end;
    Writeln('</tbody></table>');
  end;
  if MNode.nParameters > 0 then begin
    Writeln('<h2>Parameters</h2>');
    Writeln('<table border="1" cellpadding="2" cellspacing="2"><tbody>');
    for i := 1 to MNode.nParameters do begin
      Writeln('<tr><td><a name="', MNode.ParameterNames[i], '"></a>', MNode.ParameterNames[i], '</td><td>', MNode.HelpOnParameter(i),'</td>');
    end;
    Writeln('</tbody></table>');
  end;
  Writeln('</body>');
  Writeln('</html>');
  CloseFile(Output);
  {$IFDEF OFF}
  WriteLn(IndFile, #9'<LI> <OBJECT type="text/sitemap">');
  WriteLn(IndFile, #9#9'<param name="Name" value="', Name, '">');
  WriteLn(IndFile, #9#9'<param name="Local" value="', LowerCase(Name), '">');
  WriteLn(IndFile, #9'</OBJECT>');

  for i := 1 to MNode.nSettings do begin
    WriteLn(IndFile, #9'<LI> <OBJECT type="text/sitemap">');
    WriteLn(IndFile, #9#9'<param name="Name" value="', MNode.SettingNames[i], '">');
    WriteLn(IndFile, #9#9'<param name="Local" value="', LowerCase(Name), '.htm#', MNode.SettingNames[i], '">');
    WriteLn(IndFile, #9'</OBJECT>');
  end;
  for i := 1 to MNode.nParameters do begin
    WriteLn(IndFile, #9'<LI> <OBJECT type="text/sitemap">');
    WriteLn(IndFile, #9#9'<param name="Name" value="', MNode.ParameterNames[i], '">');
    WriteLn(IndFile, #9#9'<param name="Local" value="', LowerCase(Name), '.htm#', MNode.ParameterNames[i], '">');
    WriteLn(IndFile, #9'</OBJECT>');
  end;
  {$ENDIF}
  LocalContext := Context;
  if LocalContext = '' then begin
    LocalContext := Name;
    for i := Length(Name) downto 1 do
      if LocalContext[i] in ['A'..'Z'] then
        Insert(' ',LocalContext,i);
    LocalContext := LocalContext + #9'Main'
  end;
  WriteLn(IndFile,Name,#9,AltFileName,#9,LocalContext);
  for i := 1 to MNode.nSettings do
    WriteLn(IndFile,MNode.SettingNames[i],#9,AltFileName,'#',MNode.SettingNames[i],#9,LocalContext);
  for i := 1 to MNode.nParameters do
    WriteLn(IndFile,MNode.ParameterNames[i],#9,AltFileName,'#',MNode.ParameterNames[i],#9,LocalContext);
end;

{$IFDEF OFF}
procedure MakeSymTables;
var
  Mixture: TMIxture;
  Species: TSpecies;
  Molecule: TAsymmetricMolecule;
  Manifold: TAsymmetricManifold;
  State: TAsymmetricTop;
  PointGroup: TPointGroup;
  Sym: Integer;
  zAx, xAx: TabcAxis;
  Rep: TRepresentation;
begin
  Mixture := TMixture.Create(nil);
  Species := TSpecies.Create(Mixture);
  Molecule := TAsymmetricMolecule.Create(Species);
  Manifold := TAsymmetricManifold.Create(Molecule);
  State := TAsymmetricTop.Create(Manifold);
  for PointGroup := Low(PointGroup) to High(PointGroup) do begin
    Molecule.PointGroup := PointGroup;
    for Rep := Low(TRepresentation) to High(TRepresentation) do begin
      Molecule.Representation := Rep;
      for zAx := aAxis to cAxis do begin
        for xAx := cAxis downto aAxis do begin
          if zAx <> xAx then begin
            for Sym := 0 to Molecule.LastSym do begin
              Molecule.C2zAxis := zAx;
              Molecule.C2xAxis := xAx;
              State.Symmetry := Sym;
              { TODO: Molecule.PseudoC2v := True; }
              Mixture.CheckProperties;
              State.SymmetryTable(Output);
            end;
            if Molecule.PointGroup in [C1, Ci, C2, Cs, C2h] then Break;
          end;
        end;
        if Molecule.PointGroup in [C1, Ci] then Break;
      end;
    end;
  end;
end;

procedure MakeAsymLineLists(DirName: string);
var
  Mixture: TMixture;
  Species: TSpecies;
  Molecule: TAsymmetricMolecule;
  ManifoldL: TAsymmetricManifold;
  StateL, StateZ, StateX, StateY: TAsymmetricTop;
  PointGroup: TPointGroup;
  Sym: Integer;
  zAx, xAx: TabcAxis;
  Rep: TRepresentation;
  FileName: string;
begin
  Mixture := GlobalMixture;
  Species := Mixture.Species[0];
  Molecule := Species.Isotopomer[0] as TAsymmetricMolecule;
  ManifoldL := Molecule.Manifold[0] as TAsymmetricManifold;
  StateL := ManifoldL.State[0] as TAsymmetricTop;
  StateZ := Molecule.Manifold[1].State[0] as TAsymmetricTop;
  StateX := Molecule.Manifold[2].State[0] as TAsymmetricTop;
  StateY := Molecule.Manifold[3].State[0] as TAsymmetricTop;
  for PointGroup := Low(PointGroup) to High(PointGroup) do begin
    Molecule.PointGroup := PointGroup;
    for Rep := Low(TRepresentation) to High(TRepresentation) do begin
      Molecule.Representation := Rep;
      for zAx := aAxis to cAxis do begin
        for xAx := cAxis downto aAxis do begin
          if zAx <> xAx then begin
            for Sym := 0 to Molecule.LastSym do begin
              Molecule.C2zAxis := zAx;
              Molecule.C2xAxis := xAx;
              StateL.Symmetry := Sym;
              case PointGroup of
                C1: begin StateZ.Symmetry := Sym; StateX.Symmetry := Sym; StateY.Symmetry := Sym; end;
                Ci: begin StateZ.Symmetry := Sym xor 1; StateX.Symmetry := Sym xor 1; StateY.Symmetry := Sym xor 1; end;
                C2: begin StateZ.Symmetry := Sym; StateX.Symmetry := Sym xor 1; StateY.Symmetry := Sym xor 1; end;
                Cs: begin StateZ.Symmetry := Sym xor 1; StateX.Symmetry := Sym; StateY.Symmetry := Sym; end;
                D2: begin StateZ.Symmetry := Sym xor 1; StateX.Symmetry := Sym xor 3; StateY.Symmetry := Sym xor 2; end;
                C2v: begin StateZ.Symmetry := Sym; StateX.Symmetry := Sym xor 2; StateY.Symmetry := Sym xor 3; end;
                C2h: begin StateZ.Symmetry := Sym xor 2; StateX.Symmetry := Sym xor 3; StateY.Symmetry := Sym xor 3; end;
                D2h: begin StateZ.Symmetry := Sym xor 5; StateX.Symmetry := Sym xor 7; StateY.Symmetry := Sym xor 6; end;
              end;
              { TODO: Molecule.PseudoC2v := True; }
              Mixture.CheckProperties;
              FileName := DirName + PathDelim
                          + Molecule.SettingToStr(Molecule.IndexOfSetting('PointGroup')) + '_'
                          + Molecule.SettingToStr(Molecule.IndexOfSetting('Representation')) + '_'
                          + Molecule.SettingToStr(Molecule.IndexOfSetting('C2zAxis')) + '_'
                          + Molecule.SettingToStr(Molecule.IndexOfSetting('C2xAxis')) + '_'
                          + StateL.SettingToStr(StateL.IndexOfSetting('Symmetry'));
              case FileName[Length(FileName)] of
                '''': FileName[Length(FileName)] := 'p';
                '"': begin FileName[Length(FileName)] := 'p'; FileName := FileName +'p'; end;
              end;
              Mixture.SaveToFile(FileName+'.pgo');
              WriteLineList(Mixture, FileName);
            end;
            if Molecule.PointGroup in [C1, Ci, C2, Cs, C2h] then Break;
          end;
        end;
        if Molecule.PointGroup in [C1, Ci] then Break;
      end;
    end;
  end;
end;
{$ENDIF}

{$IFDEF OFF}
procedure WriteLineList(Mixture: TMixture; Name: string);
var
  Problem: Exception;
  F: Text;
  i, n0: Integer;
begin
  Problem := nil;
  Mixture.Simulations[0].MakeLineList(Problem);
  if Assigned(Problem) then
    raise Problem;
  if not Assigned(Mixture.LineList) then
    Exit;
  n0 := 0;
  with Mixture.LineList do
    for i := 0 to nLines-1 do begin
      if Line[i].EigenvaluesL.Sym = 0 then
        Inc(n0);
      if Line[i].EigenvaluesU.Sym = 0 then
        Inc(n0);
    end;
  WriteLn(Log, Name:30, Mixture.LineList.nLines:5, n0:5);
  if n0 = 0 then begin
    WriteLn(Log, '%PGO% >>dotests.out --testpgo ', Name, '.pgo nolines.out');
    raise Exception.Create('Help');
  end;
  WriteLn(Log, '%PGO% >>dotests.out --testpgo ', Name, '.pgo ', Name, '.out');
  AssignFile(F, Name + '.out');
  try
    ReWrite(F);
    with Mixture.LineList do begin
      HeadingPrinted := False;
      for i := 0 to nLines-1 do begin
        ShowLine(i,F,Mixture.PrintLevel,Mixture.Simulations[0].OThreshold,False,Problem);
        if Assigned(Problem) then
          raise Problem;
      end;
      ShowFooter(F,Mixture.PrintLevel);
    end;
  finally
    CloseFile(F);
  end;
  Problem.Free;
end;
{$ENDIF}

procedure MakeSymDataFiles(DirName: string; UseAlt: Boolean);
var
  Mixture: TMixture;
  Species: TSpecies;
  Molecule: TSymmetricMolecule;
  ManifoldL, ManifoldU: TSymmetricManifold;
  TransitionMoment: TSphericalTransitionMoment;
  StateL, StateU: TSymmetricTop;
  PointGroup: TPointGroup;
  FileName, s: string;
  SymL: Integer;
  SymU: Integer;
  Component: Integer;
  i: Integer;
  rU, rL: Double;
begin
  Mixture := GlobalMixture;
  Species := Mixture.Species[0];
  Molecule := Species.Isotopomer[0] as TSymmetricMolecule;
  Molecule.AlternateBasis := UseAlt;
  ManifoldL := Molecule.Manifold[0] as TSymmetricManifold;
  ManifoldU := Molecule.Manifold[1] as TSymmetricManifold;
  StateL := ManifoldL.State[0] as TSymmetricTop;
  StateU := ManifoldU.State[0] as TSymmetricTop;
  rL := StateL.r;
  rU := StateU.r;
  TransitionMoment := Molecule.GetTransitionMoments(0).TransitionMoment[0] as TSphericalTransitionMoment;
  for PointGroup := Low(PointGroup) to High(PointGroup) do begin
    Molecule.PointGroup := PointGroup;
    for i := WTI+1 to WTI+MaxnStatWeight - 1 do
      Molecule.Settings[i] := 11+i-WTI;
    for Component := 0 to 1 do begin
      for SymL := 0 to Molecule.LastSym do begin
        StateL.Symmetry := SymL;
        if PointGroup in [D3, D3d, D5, D5d, D7, D7d] then
          StateL.r := 0
        else
          StateL.r := rL;
        for SymU := 0 to Molecule.LastSym do begin
          StateU.Symmetry := SymU;
          if PointGroup in [D3, D3d, D5, D5d, D7, D7d] then
            StateU.r := 0
          else
            StateU.r := rL;
          TransitionMoment.Component := 2*Component;
          Mixture.CheckProperties;
          FileName := DirName + PathDelim
                      + Molecule.SettingToStr(Molecule.IndexOfSetting('PointGroup'));
          if Molecule.AlternateBasis then
            FileName := FileName + 'Alt';
          FileName := FileName + '_'
                      + StateU.SettingToStr(StateU.IndexOfSetting('Symmetry')) + '_'
                      + StateL.SettingToStr(StateL.IndexOfSetting('Symmetry')) + '_'
                      + IntToStr(Component);
          for i := 1 to Length(FileName) do
            case FileName[i] of
              '''': FileName[i] := 'p';
              '"': FileName[i] := 'd';
            end;
          //Mixture.SaveToFile(FileName+'.pgo');
          //s := FileName;
          //if Molecule.VibronicallyAllowed(TransitionMoment) = vaYes then
          //  WriteLineList(Mixture, FileName)
          //else
          //  s := 'nolines';
          //WriteLn('%PGO% >>dotests.out --testpgo ', FileName, '.pgo ', s, '.out');
          if Molecule.VibronicallyAllowed(TransitionMoment) = vaYes then
            Mixture.SaveToFile(FileName+'.pgo');
        end;
      end;
    end;
  end;
end;

procedure WriteHelpTemplate;
begin
  // Run against skeleton.pgo for objects that need a parent
  // --misc Help/scripts/skeleton.pgo dummy
  AssignFile(IndFile,'skel/base.hhk');
  Rewrite(IndFile);
//
  Context := '';
  WriteSkeleton(TPolarization);
  WriteSkeleton(TInterpolatedPartitionFunction);
  WriteSkeleton(TMixture);
  WriteSkeleton(TSpecies);
  WriteSkeleton(TManifold);
  WriteSkeleton(TMolecule);
  WriteSkeleton(TPerturbation);
  WriteSkeleton(TSpecies);
  WriteSkeleton(TState);
  WriteSkeleton(TTransitionMoment);
  WriteSkeleton(TTransitionMoments);
  WriteSkeleton(TNucleus);
  WriteSkeleton(TSimulation);
  WriteSkeleton(TOverlays);
  WriteSkeleton(TPlot);
  WriteSkeleton(TBitmapPlot);
  WriteSkeleton(TExperimentalPlot);
  WriteSkeleton(TLoomisWood);
  WriteSkeleton(TVibrationalPartitionFunction);
  WriteSkeleton(TTable);
  WriteSkeleton(TVariables);
  WriteSkeleton(TCorrelations,False);
  WriteSkeleton(TCustomPopulationFunction);
  WriteSkeleton(TCustomTransitionFunction);
  WriteSkeleton(TCustomWidthFunction);
  WriteSkeleton(TCustomLineshapeFunction);
  WriteSkeleton(TLineshape);
// Not documented  WriteSkeleton(TLineListSource);
// Not documented  WriteSkeleton(TI2);
// Not documented  WriteSkeleton(TIntensityCorrection);
// Not documented  WriteSkeleton(TLineshape);
// Not documented  WriteSkeleton(TFormSettings);
//
  Context := 'Asymmetric Top';
  WriteSkeleton(TAsymmetricTop,False);
  WriteSkeleton(TAsymmetricTopPerturbation);
// No unique settings  WriteSkeleton(TAsymmetricManifold);
  WriteSkeleton(TAsymmetricMolecule);
  WriteSkeleton(TAsymmetricTopNucleus,False);
  WriteSkeleton(TCartesianTransitionMoment);
  WriteSkeleton(TCartesianTransitionMoment,True,'asymcartesiantransitionmoment.htm');
  WriteSkeleton(TSphericalTransitionMoment,True,'asymsphericaltransitionmoment.htm');
  WriteSkeleton(TAsymmetricTopOrbitalTransitionMoment);
  WriteSkeleton(TAsymmetricTopRotationalTransitionMoment);
  WriteSkeleton(TNuclearMagneticMoment,True,'asymnucleartransitionmoment.htm');
  WriteSkeleton(TAsymmetricTopSpinSpin);
  WriteSkeleton(TAsymmetricTopIntensityPerturbation,False);
  WriteSkeleton(TAxisSwitch);
//
  Context := 'Symmetric Top';
  WriteSkeleton(TSymmetricTop);
  WriteSkeleton(TSymmetricTopPerturbation);
  WriteSkeleton(TSymmetricTopNucleus,False);
// No unique settings  WriteSkeleton(TSymmetricManifold);
  WriteSkeleton(TSymmetricMolecule);
  WriteSkeleton(TSphericalTransitionMoment,True,'symsphericaltransitionmoment.htm');
  WriteSkeleton(TCartesianTransitionMoment,True,'symcartesiantransitionmoment.htm');
//

  Context := 'Linear Molecule';
  WriteSkeleton(TLinear);
  WriteSkeleton(TLinearPerturbation);
// No unique settings  WriteSkeleton(TLinearManifold);
  WriteSkeleton(TLinearMolecule);
  WriteSkeleton(TLinearNucleus,False);
  WriteSkeleton(TSphericalTransitionMoment,True,'linsphericaltransitionmoment.htm');
  WriteSkeleton(TOrbitalTransitionMoment);
  WriteSkeleton(TSpinTransitionMoment);
  WriteSkeleton(TRotationalTransitionMoment);
  WriteSkeleton(TLinearSpinSpin);
  WriteSkeleton(TNuclearMagneticMoment,True,'linnucleartransitionmoment.htm');
  WriteSkeleton(TForbiddenTransitionMoment);
//

// Vib
  Context := 'Vibrating Molecule';
  WriteSkeleton(TVibratingMolecule);
// No unique settings  WriteSkeleton(TVibrationalManifold);
  WriteSkeleton(TElectronicState,False);
  WriteSkeleton(TVibrationalMode,False);
  WriteSkeleton(TMolecularNucleus,False);
  WriteSkeleton(TVibrationalPerturbation);
  WriteSkeleton(TMultidimensionalFCF,False);
  WriteSkeleton(TVibrationalModeChange,False);
  WriteSkeleton(TzMatrix);
  WriteSkeleton(TInternalCoords);
  WriteSkeleton(TCartesianDisplacements,True,'cartesiancoords.htm');
  WriteSkeleton(TSymmetryCoords,True,'symmetrycoordsform.htm');

// No unique settings  WriteSkeleton(TExoMolMolecule);
// No unique settings  WriteSkeleton(TExoMolManifold);
// No unique settings  WriteSkeleton(TExoMolTransitions);

  Context := 'HITRAN';
  WriteSkeleton(THITRANMolecule,False,'hitran.htm');
  // No unique settings  WriteSkeleton(THITRANManifold);
  CloseFile(IndFile);
end;

  // MakeAsymLineLists(Param2);
  // MakeSymTables;
  // MakeSymLineLists(Param2,False);
  // MakennTable(StrToInt(Param2),GlobalMixture.Species[0].Molecule[0]);
  //WriteHelpTemplate;
  //Context := 'Symmetric Top';
  //WriteSkeleton(TSymmetricTop);
  // DumpSymTables;

type
  TMakeState = procedure (Mixture: TMixture);

procedure CheckSymbolic(MakeState: TMakeState);
var
  State: TState;
  Mixture: TMixture;
  i: Integer;
begin
  Mixture := TMixture.Create(nil);
  MakeState(Mixture);
  Mixture.FixupProperties;
  State := Mixture.Species[0].Isotopomer[0].Manifold[0].State[0];
  if State is TLinear then
    with State as TLinear do begin
      Lambda := 2;
      S := 2;
    end;
  for i := 1 to State.nParameters do
    try
      if State.SymbolicRotationalMatrixElement(i,-1) = nil then
        WriteLn(State.ClassName,':',State.ParameterNames[i])
      else
        WriteLn(State.ParameterNames[i]);
    except
      WriteLn(State.ClassName,'::::',State.ParameterNames[i]);
    end;
end;

procedure CheckAllSymbolic;
begin
  CheckSymbolic(@AddSampleLinearMolecule);
  CheckSymbolic(@AddSampleSymmetricTop);
  CheckSymbolic(@AddSampleAsymmetricTop);
end;

procedure SkeletonFunction(Param2: string);
begin
  WriteHelpTemplate;
  //MakeSymDataFiles(Param2, False);
end;

initialization
//  MakeSymTables;
  WriteHelpTemplate;
  //WriteSkeleton(TCustomLineshapeFunction);
//  CheckAllSymbolic;
end.
