unit StateInfo;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, Split, Grids, NGrid, StdCtrls, AlEdit, EditExtr, NewSpin,
  ExtCtrls, ChildForm, MNode, Menus, Mixture, QuantumSpin;

type

  TStatesForm = class(TChildForm)
    EigLabel: TLabel;
    LabelPanel: TPanel;
    ActiveObject: TComboBox;
    J: TQuantumSpin;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Symmetry: TSpinBox;
    HamGrid: TNumberGrid;
    HamiltonianLabel: TLabel;
    HamPanel: TPanel;
    Split: TSplit;
    EigPanel: TPanel;
    EigenValuesLabel: TLabel;
    EigGrid: TNumberGrid;
    Perturbations: TCheckBox;
    HamLabel: TLabel;
    Label4: TLabel;
    IgnoreSpins: TSpinBox;
    EigenvaluesPopupMenu: TPopupMenu;
    AddtopopulationselectionItem: TMenuItem;
    RemovefrompopulationselectionItem: TMenuItem;
    CopyItem: TMenuItem;
    nBasisLabel: TLabel;
    nBasisSpinBox: TSpinBox;
    Blocked: TCheckBox;
    procedure RemovefrompopulationselectionItemClick(Sender: TObject);
    procedure AddtopopulationselectionItemClick(Sender: TObject);
    procedure IgnoreSpinsSetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: string);
    procedure IgnoreSpinsGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: string; var StringValid: Boolean);
    procedure EigGridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
    procedure JChange(Sender: TObject);
    procedure CopyItemClick(Sender: TObject);
  protected
    BasisLabels: array of string;
    IsComplex: Boolean;
    LastMolecule: TMNode;
    procedure SetBaseNode(const Value: TMNode); override;
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
  public
    procedure Loaded; override;
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure StuffSelector(MNode: TMNode); override;
  end;

var
  StatesForm: TStatesForm;

implementation

uses Math, MatrixVec, FormUtils, QuantumNumber, ComplexMatrix, StaticField;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TStatesForm }

function TStatesForm.BaseNodeClass: TClass;
begin
  Result := TMixture;
end;

procedure TStatesForm.CopyItemClick(Sender: TObject);
begin
  EigGrid.CopyToClipboard(True);
end;

procedure TStatesForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  JChange(nil);
end;

procedure TStatesForm.SetBaseNode(const Value: TMNode);
begin
  LastMolecule := nil;
  inherited;
end;

procedure TStatesForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  JChange(nil);
end;

procedure TStatesForm.ShowItem(MNode: TMNode);
var
  Manifold: TManifold;
  State: TState;
  EigenValues: TEigenvalues;
  i, ii, Base, RowBase: Integer;
  Sym: Integer;
  JValue: TQuantumNumber;
  HasFields: Boolean;
  {$IFNDEF USECACHE}
  Matrices: TComplexMatrices;
  {$ENDIF}
begin
  if (csLoading in ComponentState) or not Visible then
    Exit;
  try
    IgnoreSpins.Value := IgnoreSpins.ValueAB;
    J.ValueAB;
    if Assigned(BaseNode) then begin
      BaseNode.CheckProperties;
      HasFields := (BaseNode as TMixture).Simulations[0].HasFields;
    end;
    UpdateSymmetrySpin(Symmetry,MNode,HasFields);
    nBasisSpinBox.Value := nBasisSpinBox.ValueAB;
    with ActiveObject.Items do
      for i := 0 to Count - 1 do
        if MNode = Objects[i] then begin
          ActiveObject.ItemIndex := i;
          Break;
        end;
  except
    on E: Exception do begin
      HamGrid.Blank;
      EigGrid.Blank;
      HamLabel.Caption := E.Message;
      HamLabel.Font.Style := [fsBold];
      Exit;
    end;
  end;
  HamLabel.Font.Style := [];
  Caption := 'State Details';
  if not Assigned(MNode) then Exit;
  Caption := MNode.UniqueName + ' - Details';
  BeginGlobalUpdate;
  try
    State := nil;
    if MNode is TState then begin
      Manifold := TState(MNode).Manifold;
      if Manifold.nStates > 1 then
        State := TState(MNode); { Need to select State }
    end else
      Manifold := MNode as TManifold;
    Manifold.CheckProperties;
    Manifold.RootNode.FixupProperties.Free; { Check Mixture also to acount for e.g. spin changes }
    CurrentMNode := MNode;

    { Force correct even/oddness of J }
    if HasFields then begin
      J.MinValue := -J.MaxValue;
      J.Value := (J.Value and not 1) or Manifold.HalfSpin(-IgnoreSpins.Value);
      Sym := -1;
    end else begin
      J.MinValue := 0; { Avoid getting range check }
      J.Value := (Abs(J.Value) and not 1) or Manifold.HalfSpin(-IgnoreSpins.Value);
      J.MinValue := Manifold.HalfSpin(-IgnoreSpins.Value);
      { Force symmetry in range }
      Symmetry.MaxValue := Symmetry.Value; { Avoid getting range check }
      if Symmetry.Value >= Manifold.nSym then
        Symmetry.Value := Manifold.nSym - 1;
      Symmetry.MaxValue := Manifold.nSym - 1;
      Sym := Manifold.Molecule.JAdjustedSym(J.Value, Symmetry.Value);
    end;
    { Do not (normally) attempt to use cache, to avoid problems with different settings }
    {$IFNDEF USECACHE}
      EigenValues := TEigenvalues.Create(Manifold, J.Value, Sym, (BaseNode as TMixture).Simulations[0].ExField, Perturbations.Checked, State, -IgnoreSpins.Value);
    {$ELSE}
      Eigenvalues := Manifold.Eigenvalues(JSpinBox.Value, Sym, [meHrot], (BaseNode as TMixture).Simulations[0].ExField, True, -10, State, PerturbationsCheckBox.Checked, -IgnoreSpins.Value);
    {$ENDIF}
    JValue := J.Value; { Avoid scoping problems }
    try
      {$IFNDEF USECACHE}
        EigenValues.MakeBasis(Manifold.Mixture.BasisOrder);
      {$ENDIF}
      with EigenValues do
        if Size > 0 then begin
          {$IFNDEF USECACHE}
            MakeHamiltonian(Matrices, Blocked.Checked);
          {$ENDIF}
          HamGrid.RowCount := Size+1;
          if Matrices[0].Mode = cmReal then
            HamGrid.ColCount := Size+1
          else
            HamGrid.ColCount := 2*Size+1;
          for i := 1 to Size do begin
            HamGrid.Title[i] := Basis[i].BasisLabel;
            HamGrid.Lbl[i] := Basis[i].BasisLabel;
          end;
          if Length(Matrices) = 1 then
            ShowComplexMatrixInGrid(Matrices[0], HamGrid, 0, HamGrid.Width < HamPanel.Width, False, False, 0)
          else begin
            HamGrid.Clear;
            Base := HamGrid.ColCount-1;
            for i := High(Matrices) downto 0 do begin { Fill backwards so that copying up of complex matrix labels works }
              Dec(Base,Matrices[i].RowCount);
              RowBase := Base;
              if Matrices[i].Mode <> cmReal then begin
                Dec(Base,Matrices[i].RowCount);
                for ii := Matrices[i].RowCount downto 1 do
                  HamGrid.Title[Base + ii] := Basis[Base div 2 + ii].BasisLabel;
                RowBase := Base div 2;
              end;
              ShowComplexMatrixInGrid(Matrices[i], HamGrid, Base, HamGrid.Width < HamPanel.Width, False, False, RowBase);
            end;
          end;
          HamiltonianLabel.Caption := 'Hamiltonian matrix for Sym = ' + Manifold.Molecule.BothSymNames(JValue, Sym);
          if Length(Matrices) > 1 then
            HamiltonianLabel.Caption := HamiltonianLabel.Caption + ' ('+IntToStr(Length(Matrices))+' blocks)';
          // HamGrid.Title[0] := '<' + Basis[1].BasisLabelKey + '|';
          HamGrid.AutoAdjustColumn(0);
          HamLabel.Caption := '| ' + Basis[1].BasisLabelKey + ' >';
          if HasFields then
            Label2.Caption := 'M'
          else if Length(Basis[1].HyperfineState) > 0 then
            Label2.Caption := 'F'
          else
            Label2.Caption := 'J';
          {$IFNDEF USECACHE}
             Diagonalize(Matrices, True, Manifold.EigenSearch, Manifold.LimitSearch); { TODO: Make checkboxes }
          {$ENDIF}
          if Manifold.HasWidth then
            MakeWidths(plNone);
          if nBasisSpinBox.Value > 0 then begin
            SetLength(BasisLabels,Size);
            for i := 1 to Size do
              BasisLabels[i-1] := Basis[i].BasisLabel;
          end else
            SetLength(BasisLabels,0);
          IsComplex := IsComplexMatrix(Eigenvalues.C);
          EigenvaluesLabel.Caption := ShowEigenvaluesInGrid(Eigenvalues,EigGrid,EigLabel,nBasisSpinBox.Value);
          nBasisLabel.Left := EigenValuesLabel.Left	+ EigenValuesLabel.Width + 4;
          nBasisSpinBox.Left := nBasisLabel.Left + nBasisLabel.Width + 2;
        end else begin
          { Empty basis - show blank matrices }
          HamiltonianLabel.Caption := 'No states';
          EigenvaluesLabel.Caption := 'No states';
          HamGrid.Blank;
          EigGrid.Blank;
        end;
    finally
      {$IFNDEF USECACHE}
      for i := 0 to High(Matrices) do
        Matrices[i].Free;
      EigenValues.Free;
      {$ENDIF}
    end;
  finally
    EndGlobalUpdate;
  end;
  EigGrid.Invalidate; { Make sure row labels refresh }
  LabelPanel.Height := ActiveObject.Top + ActiveObject.Height + 2;
end;

procedure TStatesForm.StuffSelector(MNode: TMNode);
begin
  StuffSelectorHelper(MNode, ActiveObject);
  MNode := CurrentMNode;
  while Assigned(MNode) and not (MNode is TMolecule) do
    MNode := MNode.Owner as TMNode;
  if (MNode is TMolecule) and (MNode <> LastMolecule) then begin
    LastMolecule := MNode;
    Blocked.Checked := TMolecule(MNode).BlockMatrix;
  end;
  FComboBoxUpdated := False;
end;

function TStatesForm.StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer;
begin
  Result := 0;
  if ((MNode is TManifold) and (TManifold(MNode).nStates > 1)) or (MNode is TState) then begin
    ActiveObject.AddItem(Name1 + Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end;
end;

procedure TStatesForm.JChange(Sender: TObject);
begin
//  if Lockout > 0 then
//    ReStuff
//  else
//    ShowItem(CurrentMNode);
  FReStuffPending := True;
end;

procedure TStatesForm.Loaded;
begin
  inherited;
  UpdateSymmetrySpin(Symmetry, nil, False);
  EigGrid.CopyLabels := True;
  HamGrid.CopyLabels := True;
  {$IFDEF FPC}
    {$IFDEF MSWINDOWS}
      Blocked.Top := 1;
      HamGrid.Top := Blocked.Top + Blocked.Height + 2;
    {$ENDIF}
  {$ENDIF}
  EigGrid.FixedSizeNumbers := True;
  HamGrid.FixedSizeNumbers := True;
end;

procedure TStatesForm.EigGridGetCellLabel(Sender: TObject; Col, Row: Integer;
  var s: string; var IsText: Boolean);
var
  Base, nCof, bl: Integer;
begin
  if Row > 0 then
    if Col = 0 then
      s := IntToStr(Row)
    else if Col = 1 then begin
      s := EigGrid.Lbl[Row];
      IsText := True;
    end else if Length(BasisLabels) > 0 then begin
      { Check for basis state label, rather than value }
      if IsComplex then
        nCof := 3
      else
        nCof := 2;
      Base := EigGrid.ColCount - nCof*Min(nBasisSpinBox.Value,Length(BasisLabels));
      if (Col > Base) and ((Col-Base) mod nCof = nCof-1) then begin
        bl := Round(EigGrid.Value[Row,Col])-1;
        if (bl >= 0) and (bl < Length(BasisLabels)) then        
          s := BasisLabels[bl]
        else
          s := '?'; { Should only happen during update }
        IsText := True;
      end;
    end;
end;

procedure TStatesForm.IgnoreSpinsGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: string; var StringValid: Boolean);
begin
  if TheText = 'all' then begin
    StringValid := True;
    NewValue := 0;
  end;
end;

procedure TStatesForm.IgnoreSpinsSetIntegerValue(Sender: TObject;
  NewValue: Integer; var NewText: string);
begin
  if NewValue = 0 then
    NewText := 'all';
end;

procedure TStatesForm.AddtopopulationselectionItemClick(Sender: TObject);
var
  Manifold: TManifold;
  HasFields, Adding: Boolean;
  r, i: Integer;
begin
  if not Assigned(CurrentMNode) then Exit;
  if CurrentMNode is TState then
    Manifold := (CurrentMNode as TState).Manifold
  else
    Manifold := CurrentMNode as TManifold;
  HasFields := Manifold.Mixture.Simulations[0].HasFields;
  Adding := Sender <> nil;
  BeginGlobalUpdate;
  try
    for r := EigGrid.Selection.Top to EigGrid.Selection.Bottom do begin
      if HasFields then
        i := Manifold.PopulationIndex(J.Value, -1, r, Adding)
      else
        i := Manifold.PopulationIndex(J.Value, Manifold.Molecule.JDeAdjustedSym(J.Value, Symmetry.Value), r, Adding);
      if not Adding and (i > 0) then
        Manifold.DeleteParameter(i);
    end;
  finally
    EndGlobalUpdate;
  end;
end;

procedure TStatesForm.RemovefrompopulationselectionItemClick(Sender: TObject);
begin
  AddtopopulationselectionItemClick(nil);
end;

end.
