﻿unit TranInfo;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ChildForm, StdCtrls, AlEdit, EditExtr, NewSpin, ExtCtrls, ComCtrls,
  MNode, Split, Grids, NGrid, Buttons, QuantumSpin, QuantumNumber, Mixture,
  Settings;

type

  { TTransitionForm }

  TTransitionForm = class(TChildForm)
    LabelPanel: TPanel;
    Label1: TLabel;
    JLabel: TLabel;
    Label3: TLabel;
    UpperObject: TComboBox;
    UpperJ: TQuantumSpin;
    UpperSymmetry: TSpinBox;
    LowerObject: TComboBox;
    LowerSymmetry: TSpinBox;
    LowerJ: TQuantumSpin;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    ActiveObject: TComboBox;
    OrgPanel: TPanel;
    OriginalCaption: TLabel;
    OrgGrid: TNumberGrid;
    Split: TSplit;
    TranPanel: TPanel;
    Label8: TLabel;
    TranGrid: TNumberGrid;
    Square: TCheckBox;
    Label9: TLabel;
    PlotLines: TCheckBox;
    Label10: TLabel;
    JChange: TQuantumSpin;
    Perturbations: TCheckBox;
    OriginalLabel: TLabel;
    TransformedLabel: TLabel;
    SimulateButton: TSpeedButton;
    ExcitationType: TComboBox;
    Label11: TLabel;
    ExcitationRank: TQuantumSpin;
    Label12: TLabel;
    UpperIndex: TSpinBox;
    LowerIndex: TSpinBox;
    AllButton: TSpeedButton;
    OrgHonlLondon: TCheckBox;
    TranHonlLondon: TCheckBox;
    SymChange: TSpinBox;
    IndexChange: TSpinBox;
    QnoAU: TQuantumSpin;
    ALabel: TLabel;
    QnoAD: TSpinBox;
    QnoAL: TQuantumSpin;
    BLabel: TLabel;
    QnoBU: TQuantumSpin;
    QnoBD: TSpinBox;
    QnoBL: TQuantumSpin;
    FullCalculation: TCheckBox;
    AddButton: TButton;
    ReplaceButton: TButton;
    Filter: TCheckBox;
    FilterStatus: TLabel;
    InPlotRange: TCheckBox;
    StrengthCheck: TCheckBox;
    Fraction: TEdit;
    Label2: TLabel;
    DiscardDuplicates: TCheckBox;
    AutoRange: TCheckBox;
    AssignNearest: TButton;
    procedure FormResize(Sender: TObject);
    procedure OrgGridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
    procedure AllButtonClick(Sender: TObject);
    procedure UpperJChange(Sender: TObject);
    procedure LowerSymmetrySetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: String);
    procedure UpperSymmetryGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: String; var StringValid: Boolean);
    procedure JChangeGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: String; var StringValid: Boolean);
    procedure JChangeSetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: String);
    procedure SimulateButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure UpperJIncrement(Sender: TObject; var NewValue: Integer; LastValue, n: Integer);
    procedure SymChangeSetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: string);
    procedure SymChangeGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: string; var StringValid: Boolean);
    procedure IndexChangeIncrement(Sender: TObject; var NewValue: Integer; LastValue, n: Integer);
    procedure LowerIndexGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: string; var StringValid: Boolean);
    procedure LowerIndexSetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: string);
    procedure ReplaceButtonClick(Sender: TObject);
    procedure AddButtonClick(Sender: TObject);
    procedure AssignNearestClick(Sender: TObject);
  private
    FSymChange, FIndexChange, FUpperIndex, FLowerIndex, FAChange, FBChange: Integer;
    FQnoAU, FQnoAL, FQnoBU, FQnoBL: TQuantumNumber;
    FPlotMin, FPlotMax: array of Double; { For each simulation }
    FStateU, FStateL: TState;
    function OnAddLine(const ALine: TLine): Boolean;
    function ReadJSpinBox(SpinBox: TQuantumSpin):TQuantumNumber;
    function OutsideRange(Frequency: Double):Boolean;
  protected
    procedure FormConstrainedResize(Sender: TObject; var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    function ShowTransitions(PlotAllSelected, AddLinesToLinelist:Boolean): Integer;
    procedure CheckExtraVisible;
    procedure HandleComponent(Component: TComponent; var PName, AValue, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
  public
    procedure Loaded; override;
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure StuffSelector(MNode: TMNode); override;
  end;

var
  TransitionForm: TTransitionForm;

implementation

uses Math, MathSup, MatrixVec, Types, UCompat, Global, LineFrm, Units, DRintensity,
     Main, FormUtils, SettingsForm, Profile, ComplexMatrix, QuantumMath,
     MNodeColour;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TTransitionForm }

function TTransitionForm.BaseNodeClass: TClass;
begin
  Result := TMixture;
end;

procedure TTransitionForm.CheckExtraVisible;
begin
  if Filter.Checked then
    LabelPanel.Height := Fraction.Top + Fraction.Height + 2
  else
    LabelPanel.Height := Fraction.Top;
  AssignNearest.Visible := LineForm.Visible and LineForm.PanelEx.Visible;
end;

procedure TTransitionForm.HandleComponent(Component: TComponent; var PName,
  AValue, PHint: string; var Save: Boolean; var SettingType: TSettingType;
  Write: Boolean);
var
  Q: TQuantumNumber;
  OK: Boolean;
begin
  if Component = JChange then
    { Special case handling of offsets from (dynamic) MinValue; use large fixed value when saving }
    if Write then
      JChange.MinValue := -1000 { Will be updated on first use of form }
    else if JChange.Value = JChange.MinValue+2 then begin
      AValue := IntToStr(-1000+2);
      Save := True;
      Exit;
    end;
  inherited HandleComponent(Component, PName, AValue, PHint, Save, SettingType, Write);
end;

procedure TTransitionForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  ComboBoxChange(ActiveObject);
end;

function TTransitionForm.ReadJSpinBox(SpinBox: TQuantumSpin): TQuantumNumber;
begin
  Result := SpinBox.Value;
  if Result <= SpinBox.MinValue then
    Result := Low(Integer);
end;

function TTransitionForm.OutsideRange(Frequency: Double): Boolean;
var
  i: Integer;
begin
  Result := True;
  for i := 0 to High(FPlotMin) do
    if (Frequency >= FPlotMin[i]) and (Frequency <= FPlotMax[i]) then begin
      Result := False;
      Break;
    end;
  Result := Result and (Length(FPlotMin) > 0);
end;

procedure TTransitionForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  ComboBoxChange(ActiveObject);
  if (MNode is TMolecule) and (MNode.IndexOfSetting('JAdjustSym') = Index) then
    ReStuff;
end;

procedure TTransitionForm.ShowItem(MNode: TMNode);
var
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  i: Integer;
  Change: TQuantumNumber;
  HasFields: Boolean;
  PrevUpper, PrevLower: TObject;
  procedure StuffManifold(Manifold: TManifold; ComboBox: TComboBox);
  var
    i: Integer;
  begin
    if (Manifold.nStates > 0) and (ComboBox.Items.IndexOfObject(Manifold.State[0]) >= 0) then
      Exit;
    if Manifold.nStates <> 1 then begin
      if ComboBox.Items.IndexOfObject(Manifold) >= 0 then { Note External molecule has no states }
        Exit;
      ComboBox.AddItem(Manifold.UniqueName, Manifold);
    end;
    for i := 0 to Manifold.nStates-1 do
      ComboBox.AddItem(Manifold.State[i].UniqueName, Manifold.State[i]);
  end;
  procedure StuffTransitionMoments(TransitionMoments: TTransitionMoments);
  var
    Bra, Ket: TManifold;
  begin
    Bra := TransitionMoments.Bra as TManifold;
    Ket := TransitionMoments.Ket as TManifold;
    StuffManifold(Bra, UpperObject);
    StuffManifold(Ket, LowerObject);
//    if Bra.Initial then
//      StuffManifold(Bra, LowerObject)
//    else
//      StuffManifold(Bra, UpperObject);
//    if Ket.Initial then
//      StuffManifold(Ket, LowerObject)
//    else
//      StuffManifold(Ket, UpperObject);
//    if Bra.Initial and Ket.Initial then begin
//      { Ensure upper state box contains all sensible final states }
//      StuffManifold(Ket, UpperObject);
//      StuffManifold(Bra, UpperObject);
//    end;
  end;
  procedure CheckForMultipleManifolds(ComboBox: TComboBox; JSpinBox: TQuantumSpin; SymSpinBox: TSpinBox);
  var
    i: Integer;
    nManifolds: Integer;
    CurManifold, Manifold: TManifold;
  begin
    nManifolds := 0;
    CurManifold := nil;
    for i := 0 to ComboBox.Items.Count-1 do begin
      if ComboBox.Items.Objects[i] is TManifold then
        Manifold := TManifold(ComboBox.Items.Objects[i])
      else if ComboBox.Items.Objects[i] is TState then
        Manifold := TState(ComboBox.Items.Objects[i]).Manifold
      else
        raise Exception.Create('CheckForMultipleManifolds');
      if Manifold <> CurManifold then begin
        CurManifold := Manifold;
        Inc(nManifolds);
      end;
    end;
    if nManifolds = 0 then
      Exit;
    if nManifolds > 1 then
      ComboBox.Items.InsertObject(0, '(All)', nil);
    ComboBox.ItemIndex := 0;
    if ComboBox.Items.Objects[0] is TManifold then begin
      Manifold := TManifold(ComboBox.Items.Objects[0]);
      { Force symmetry in range }
      UpdateSymmetrySpin(SymSpinBox,Manifold,HasFields);
    end else if ComboBox.Items.Objects[0] is TState then begin
      Manifold := TState(ComboBox.Items.Objects[0]).Manifold;
      UpdateSymmetrySpin(SymSpinBox,Manifold,HasFields);
    end else
      UpdateSymmetrySpin(SymSpinBox,Manifold.Mixture,HasFields);
    i := ReadJSpinBox(JSpinBox); { Save previous value to preserve 'all' status }
    { Force correct even/oddness of J }
    if Manifold.Mixture.Simulations[0].HasFields then
      JSpinBox.MinValue := -JSpinBox.MaxValue - Manifold.HalfSpin
    else
      JSpinBox.MinValue := Manifold.HalfSpin - 2;
    if i = Low(Integer) then
      i := JSpinBox.MinValue;
    i := (Max(i,JSpinBox.MinValue) and not 1) or Manifold.HalfSpin;
    JSpinBox.Value := i;
  end;
  procedure ClearObjects(ComboBox: TComboBox; out CurrentObject: TObject);
  begin
    CurrentObject := nil;
    if ComboBox.ItemIndex >= 0 then
      CurrentObject := ComboBox.Items.Objects[ComboBox.ItemIndex];
    { Invalid objects OK, as specifically checked later on }
    ComboBox.Clear;
  end;
begin
  if not Assigned(MNode) then Exit;
  BeginGlobalUpdate;
  try
    if MNode is TMolecule then begin
      Molecule := TMolecule(MNode);
      if Molecule.nTransitionMoments = 1 then
        TransitionMoments := Molecule.GetTransitionMoments(0)
      else
        TransitionMoments := nil;
    end else begin
      if MNode is TTransitionMoment then
        TransitionMoments := MNode.Owner as TTransitionMoments
      else
        TransitionMoments := MNode as TTransitionMoments;
      Molecule := TransitionMoments.Molecule;
    end;
    HasFields := Molecule.Mixture.Simulations[0].HasFields;
    ClearObjects(UpperObject, PrevUpper);
    ClearObjects(LowerObject, PrevLower);
    if Assigned(TransitionMoments) then
      StuffTransitionMoments(TransitionMoments)
    else if Molecule.nTransitionMoments = 0 then { No transition moments - appropriate for PopDist }
      for i := 0 to Molecule.nManifolds - 1 do
        StuffManifold(Molecule.Manifold[i],LowerObject)
    else
      for i := 0 to Molecule.nTransitionMoments-1 do
        StuffTransitionMoments(Molecule.GetTransitionMoments(i));
    CheckForMultipleManifolds(UpperObject, UpperJ, UpperSymmetry);
    CheckForMultipleManifolds(LowerObject, LowerJ, LowerSymmetry);
    UpdateComboBox(UpperObject, PrevUpper);
    UpdateComboBox(LowerObject, PrevLower);
    UpdateQnoBox(Molecule, ALabel, QnoAL, qtA);
    UpdateQnoBox(Molecule, BLabel, QnoBL, qtB);
    UpdateQnoBox(Molecule, ALabel, QnoAU, qtA);
    UpdateQnoBox(Molecule, BLabel, QnoBU, qtB);
    FComboBoxUpdated := False;
    { Update possible changes using MaxRank, preserving 'all <>' if required }
    Change := JChange.Value;
    if Change <= JChange.MinValue+2 then
      Change := -Molecule.MaxRank-4+(Change-JChange.MinValue);
    JChange.MinValue := -Molecule.MaxRank-4;
    JChange.MaxValue := Molecule.MaxRank;
    JChange.Value := Max(Min(Change,JChange.MaxValue), JChange.MinValue);
    { Force display of all if required }
    UpperIndex.Value := UpperIndex.Value;
    LowerIndex.Value := LowerIndex.Value;
    SymChange.Value := SymChange.Value;
    IndexChange.Value := IndexChange.Value;
    QnoAD.Value := QnoAD.Value;
    QnoBD.Value := QnoBD.Value;
    CurrentMNode := MNode;
    if Molecule.Mixture.AutoReplot then
      ShowTransitions(False, False);
  finally
    EndGlobalUpdate;
  end;
end;

procedure TTransitionForm.SimulateButtonClick(Sender: TObject);
begin
  ShowTransitions(False, False);
end;

procedure TTransitionForm.AddButtonClick(Sender: TObject);
begin
  ShowTransitions(False, True);
end;

procedure TTransitionForm.ReplaceButtonClick(Sender: TObject);
begin
  LineForm.DeleteUnassignedClick(Self);
  AddButtonClick(nil);
end;

procedure TTransitionForm.AllButtonClick(Sender: TObject);
begin
  ShowTransitions(True, False);
end;

procedure TTransitionForm.AssignNearestClick(Sender: TObject);
begin
  CheckExtraVisible;
  if not AssignNearest.Visible then
    Exit;
  if ShowTransitions(False, True) = 0 then
    Exit;
  LineForm.NearestButtonClick(Self);
  LineForm.FitButton.Click;
  LineForm.TestFitButton.Click;
end;

procedure TTransitionForm.UpperJChange(Sender: TObject);
begin
  if (Lockout > 0) or (BaseNode = nil) or not (BaseNode is TMixture) then Exit;
  CheckExtraVisible;
  if (BaseNode as TMixture).AutoReplot then
    ShowTransitions(False, False);
end;

function TTransitionForm.ShowTransitions(PlotAllSelected, AddLinesToLinelist:Boolean): Integer;
var
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  TransitionMoment: TTransitionMoment;
  ManifoldU, ManifoldL: TManifold;
  EigenValuesU, EigenValuesL: TEigenvalues;
  SymU, SymL: Integer;
  CalcStateU, CalcStateL: TState;
  LineList: TLineList;
  NoUpper: Boolean;
  DeltaJ: TQuantumNumber;
  Excitation: TExcitation;
  Simulation: TSimulation;
  FilterNeeded: Boolean;
  function Extract(var Manifold: TManifold; var State, CalcState: TState; Sym: Integer; ComboBox: TComboBox;
                   JSpinBox: TQuantumSpin; OtherJ, DeltaJ: TQuantumNumber): TEigenvalues;
  var
    MNode: TMNode;
    J: TQuantumNumber;
    Matrices: TComplexMatrices;
    I: Integer;
  begin
    Result := nil;
    State := nil;
    Manifold := nil;
    CalcState := nil;
    if ComboBox.Items.Count = 0 then
      Exit;
    I := ComboBox.ItemIndex;
    if I < 0 then
      if ComboBox.Items.Count = 1 then
        I := 0 { Only one item, so take that }
      else
        Exit;
    MNode := ComboBox.Items.Objects[I] as TMNode;
    if MNode is TManifold then
      Manifold := TManifold(MNode)
    else if MNode is TState then begin
      State := TState(MNode);
      Manifold := State.Manifold;
      if Manifold.nStates < 2 then
        State := nil { No state selection necessary }
      else if FullCalculation.Checked then
        FilterNeeded := True { Need to reject unwanted states }
      else
        CalcState := State; { Restrict calculation }
    end;
    if not Simulation.HasFields and (Sym < 0) then Exit;
    J := ReadJSpinBox(JSpinBox);
    if (J = Low(Integer)) and (OtherJ <> Low(Integer)) and (Abs(DeltaJ) < 100) then
      J := OtherJ + DeltaJ;
    if (J = Low(Integer)) or (not Simulation.HasFields and (J < 0)) then
      Exit;
    if Assigned(Manifold) then begin
      if Odd(J+Manifold.HalfSpin) then Exit;
      Result := TEigenvalues.Create(Manifold, J, Manifold.Molecule.JAdjustedSym(J,Sym), Simulation.ExField, True, CalcState, 0);
      if Assigned(Result) then
        try
          Result.MakeBasis(Manifold.Mixture.BasisOrder);
          if Result.Size = 0 then
            FreeAndNil(Result)
          else begin
            Result.MakeHamiltonian(Matrices, Manifold.Molecule.BlockMatrix); { TODO: Make checkbox }
            Result.Diagonalize(Matrices, True, Manifold.EigenSearch, Manifold.LimitSearch);
          end;
        except
          for I := 0 to High(Matrices) do
            Matrices[I].Free;
          Result.Free;
          raise;
        end
      else
        FreeAndNil(Result);
    end;
  end;
  procedure ShowMatrix(D: TMatrix; Grid: TNumberGrid; ShowBasis, Square: Boolean; Scale: Double);
  var
    i: Integer;
    function SLabel(i: Integer; Ev: TEigenValues): string;
    begin
      if ShowBasis then
        Result := Ev.Basis[i].BasisLabel
      else
        Result := Ev.DominantBasis[i].StateLabel(Ev.Manifold.Mixture.QuantumNumberFormat);
    end;
  begin
    Grid.RowCount := Max(EigenValuesL.Size,1)+1;
    Grid.ColCount := Max(EigenValuesU.Size,1)+1;
    for i := 1 to EigenValuesU.Size do begin
      Grid.Title[i] := SLabel(i, EigenValuesU);
      if EigenValuesL.Size = 0 then
        Grid.Valid[1, i] := False;
    end;
    if EigenValuesL.Size = 0 then
      Grid.Lbl[1] := ' ';
    for i := 1 to EigenValuesL.Size do begin
      Grid.Lbl[i] := SLabel(i, EigenValuesL);
      if EigenValuesU.Size = 0 then
        Grid.Valid[i, 1] := False;
    end;
    ShowComplexMatrixInGrid(D, Grid, 0, Grid.Width < OrgPanel.Width, Square, False, 0, Scale);
    if EigenValuesU.Size = 0 then
      Grid.Title[1] := ' ';
    if (EigenValuesU.Sym < 0) or (EigenValuesL.Sym < 0) then
      Grid.Caption := Format('<%s||%s>', [EigenValuesL.Manifold.UniqueName, EigenValuesU.Manifold.UniqueName])
    else
      Grid.Caption := Format('<%s||%s>', [Molecule.BothSymNames(EigenValuesL.J, EigenvaluesL.Sym),
                                          Molecule.BothSymNames(EigenValuesU.J, EigenvaluesU.Sym)]);
    Grid.AutoAdjustColumn(0);
  end;
  procedure ShowMatrices;
  var
    D: TMatrix;
    F: TComplexMatrix;
    s: string;
    Factor: Double;
    OnlyRank: TQuantumNumber;
    Ds: TTransitionMatrices;
  begin
    F := nil;
    Ds.Clear(False);
    Factor := 1;
    if (EigenValuesL.Size > 0) and (EigenValuesU.Size >0)
       and (EigenValuesL.HasFields
            or (EigenValuesL.SpinSpecies(1) = EigenValuesU.SpinSpecies(1)))
      then begin
        if Assigned(TransitionMoment) then
          OnlyRank := TransitionMoment.Rank
        else if Assigned(TransitionMoments) and (TransitionMoments.nRanks = 1) then
          OnlyRank := TransitionMoments.MaxRank { = rank of all selected transition moments }
        else
          OnlyRank := -1;
        if OnlyRank >= 0 then begin
          F := TransitionMoments.OriginalTransitionMatrix(EigenValuesL, EigenValuesU, OnlyRank, TransitionMoment, True, True, True, Simulation);
          OrgHonlLondon.Enabled := not EigenValuesL.HasFields;
          TranHonlLondon.Enabled := not EigenValuesL.HasFields;
          Factor := Sqrt(OnlyRank+1);
        end else if Assigned(TransitionMoments) then begin
          Ds := TransitionMoments.TransitionMatrix(EigenValuesL, EigenValuesU, True, False, nil, Simulation, TransitionMoments.Mixture.PrintLevel);
          D := Ds.D[ttDipole];
          if D = nil then
            D := Ds.D[ttQuadrupole];
          { TODO: Should toggle display of both if quadrupole and dipole present }
        end;
      end;
    try
      ShowMatrix(F, OrgGrid, True, False, IfThen(OrgHonlLondon.Checked,Factor,1));
      s := '< Lower ';
      if EigenValuesL.Size > 0 then
        s := s + EigenValuesL.Basis[1].BasisLabelKey;
      s := s + ' |T| Upper ';
      if EigenValuesU.Size > 0 then
        s := s + EigenValuesU.Basis[1].BasisLabelKey;
      s := s + ' >';
      OriginalLabel.Caption := s;
      OriginalCaption.Caption := 'Original Transition Matrix';
      if Assigned(F) then begin
        EigenValuesL.TransformTransitionMatrix(F, EigenValuesU);
        Square.Enabled := True;
        s := '< Lower ';
        if EigenValuesL.Size > 0 then
          s := s + EigenValuesL.Basis[1].StateLabelKey;
        s := s + ' |T| Upper ';
        if EigenValuesU.Size > 0 then
          s := s + EigenValuesU.Basis[1].StateLabelKey;
        s := s + ' >';
        TransformedLabel.Caption := s;
        ShowMatrix(F, TranGrid, False, Square.Checked, IfThen(TranHonlLondon.Enabled and TranHonlLondon.Checked,Factor,1));
      end else begin
        Square.Enabled := False;
        Square.Checked := True;
        { D matrix is already squared, so don't square it again }
        ShowMatrix(D, TranGrid, False, False, IfThen(TranHonlLondon.Enabled and TranHonlLondon.Checked,Factor,1));
      end;
    finally
      F.Free;
      Ds.Clear(True);
    end;
  end;
  procedure CalculateMatrices; { If transition selection sufficiently limited,
                                 show transition matrixes }
  var
    DeltaJ: TQuantumNumber;
    procedure ShowMatrix(Eigenvalues: TEigenvalues; Name: string);
    begin
      Split.OneWindowVisible(OrgPanel);
      Split.Visible := False;
      OrgGrid.Caption := '';
      OrgGrid.OnGetCellLabel := {$IFDEF FPC}@{$ENDIF}OrgGridGetCellLabel;
      OriginalCaption.Caption := Name + ShowEigenvaluesInGrid(Eigenvalues, OrgGrid, OriginalLabel);
      OrgGrid.Invalidate; { Make sure row labels refresh }
    end;
  begin
    DeltaJ := JChange.Value;
    if DeltaJ < JChange.MinValue+4 then
      DeltaJ := -100-(DeltaJ-JChange.MinValue);
    OrgHonlLondon.Enabled := False;
    TranHonlLondon.Enabled := False;
    FSymChange := SymChange.Value;
    if Simulation.HasFields then begin
      SymU := -1;
      SymL := -1;
    end else begin
      SymU := UpperSymmetry.Value;
      SymL := LowerSymmetry.Value;
      if (SymU >=0) xor (SymL >=0) then
        if FSymChange = 0 { equal symmetries } then begin
          { If one symmetry specified, force the other }
          SymL := Max(SymL, SymU);
          SymU := SymL;
        end else if (FSymChange = 1) { unequal symmetries } and (Molecule.LastSym = 2) then
          { Only two symmetries }
          if SymL < 0 then
            SymL := 1-SymU
          else
            SymU := 1-SymL;
    end;
    EigenvaluesU := Extract(ManifoldU, FStateU, CalcStateU, SymU, UpperObject, UpperJ, ReadJSpinBox(LowerJ), DeltaJ);
    try
      EigenvaluesL := Extract(ManifoldL, FStateL, CalcStateL, SymL, LowerObject, LowerJ, ReadJSpinBox(UpperJ), -DeltaJ);
      try
        if Assigned(EigenvaluesU) and Assigned(EigenvaluesL) then begin
          if not Assigned(TransitionMoments) then
            TransitionMoments := Molecule.FindTransitionMoments(EigenValuesU.Manifold, EigenValuesL.Manifold);
          Split.Visible := True;
          Split.MakeBothVisible;
          OrgGrid.OnGetCellLabel := nil;
          ShowMatrices;
        end else if Assigned(EigenValuesU) then
          ShowMatrix(EigenValuesU, 'Upper')
        else if Assigned(EigenValuesL) then
          ShowMatrix(EigenValuesL, 'Lower')
        else begin
          OrgGrid.Blank;
          TranGrid.Blank;
          Split.OneWindowVisible(OrgPanel);
        end;
      finally
        EigenValuesL.Free;
      end;
    finally
      EigenValuesU.Free;
    end;
  end;

  procedure LoadInteger(var FValue: Integer; SpinBox: TSpinBox; Scale: Integer);
  begin
    FValue := SpinBox.Value;
    if FValue = SpinBox.MinValue then
      FValue := Low(Integer)
    else begin
      FilterNeeded := True;
      FValue := FValue*Scale;
    end;
  end;

  procedure LoadQno(var FValue: TQuantumNumber; SpinBox: TQuantumSpin);
  begin
    FValue := SpinBox.Value;
    if FValue = SpinBox.MinValue then
      FValue := Low(Integer)
    else
      FilterNeeded := True;
  end;

  procedure ComputeRanges(RangeUnits: TResonanceUnits);
  var
    i: Integer;
  begin
    if Filter.Checked and InPlotRange.Checked then begin
      FilterNeeded := True;
      with Molecule.Mixture do begin
        SetLength(FPlotMin, nSimulations);
        SetLength(FPlotMax, nSimulations);
        for i := 0 to nSimulations-1 do begin
          FPlotMin[i] := Convert(Simulations[i].Fmin,Simulations[i].PlotUnits,RangeUnits);
          FPlotMax[i] := Convert(Simulations[i].Fmax,Simulations[i].PlotUnits,RangeUnits);
          InOrder(FPlotMin[i], FPlotMax[i]);
        end;
      end;
    end else begin
      FPlotMin := nil;
      FPlotMax := nil;
    end;
  end;

var
  Problem: Exception;
  i, nLines, nShown: Integer;
  Strengths: array of Double;
  MinStrength, MaxStrength, Mi, Mx: Double;
begin
  Result := 0;
  if (csLoading in ComponentState) or not Visible then
    Exit;
  Caption := 'Selected Transitions';
  if CurrentMNode = nil then Exit;
  CheckExtraVisible;
  TransitionMoment := nil;
  if CurrentMNode is TMolecule then begin
    Molecule := TMolecule(CurrentMNode);
    if Molecule.nTransitionMoments = 1 then
      TransitionMoments := Molecule.GetTransitionMoments(0)
    else
      TransitionMoments := nil;
  end else begin
    if CurrentMNode is TTransitionMoment then begin
      TransitionMoment := TTransitionMoment(CurrentMNode);
      TransitionMoments := CurrentMNode.Owner as TTransitionMoments;
    end else
      TransitionMoments := CurrentMNode as TTransitionMoments;
    Molecule := TransitionMoments.Molecule;
  end;
  Simulation := Molecule.Mixture.Simulations[0];
  if Simulation.HasFields then
    JLabel.Caption := '&M'
  else
    JLabel.Caption := '&J';
  try
    if Assigned(BaseNode) then
      BaseNode.CheckProperties;
    JChange.ValueAB;
    ExcitationRank.ValueAB;
    UpperJ.ValueAB;
    UpperSymmetry.Enabled := not Simulation.HasFields;
    UpperSymmetry.ValueAB;
    UpperIndex.ValueAB;
    LowerJ.ValueAB;
    LowerSymmetry.Enabled := not Simulation.HasFields;
    LowerSymmetry.ValueAB;
    LowerIndex.ValueAB;
    IndexChange.ValueAB;
    SymChange.Enabled := not Simulation.HasFields;
    SymChange.ValueAB;
    QnoAD.ValueAB;
    QnoBD.ValueAB;
  except
    on E: Exception do begin
      OriginalLabel.Caption := E.Message;
      Exit;
    end;
  end;
  Caption := CurrentMNode.UniqueName + ' - ' + Caption;
  BeginGlobalUpdate;
  try
    Molecule.CheckProperties;
    Molecule.RootNode.FixupProperties.Free; { Check Mixture also to acount for e.g. spin changes }
    FilterNeeded := False;
    FPlotMin := nil;
    LoadInteger(FIndexChange, IndexChange, 1);
    LoadQno(FQnoAU, QnoAU);
    LoadQno(FQnoAL, QnoAL);
    LoadQno(FQnoBU, QnoBU);
    LoadQno(FQnoBL, QnoBL);
    LoadInteger(FAChange, QnoAD, QnoAU.Increment);
    LoadInteger(FBChange, QnoBD, QnoBU.Increment);
    FUpperIndex := UpperIndex.Value;
    FLowerIndex := LowerIndex.Value;
    FilterNeeded := FilterNeeded or (FLowerIndex <> 0) or (FUpperIndex <> 0);
    { Note on handling lines exclued by range:
      PlotAllSelected = False AddLinesToLinelist = False: Simple plot; linelist should contain excluded transitions, though strength threshold on transitions in range
      PlotAllSelected = True  AddLinesToLinelist = False: Simple plot; no lines excluded by range
                                                          AutoRange is the same
      PlotAllSelected = False AddLinesToLinelist = True:  Do not plot; exclude lines
    }
    if AddLinesToLinelist then
      ComputeRanges(Molecule.Mixture.Units);
    CalculateMatrices;
    NoUpper := UpperObject.Items.Count = 0;
    if ((ReadJSpinBox(LowerJ) <> Low(Integer)) and Odd(LowerJ.Value + Molecule.HalfSpin))
       or (not NoUpper and ((ReadJSpinBox(UpperJ) <> Low(Integer)) and Odd(UpperJ.Value + Molecule.HalfSpin))) then
      Exit; { Invalid J }
    if AddLinesToLinelist or PlotLines.Checked or PlotAllSelected then begin
      LineList := TLineList.Create(Simulation.PlotUnits,IsPopDist(Simulation.IntensityUnits));
      if (FSymChange >= 0) or FilterNeeded then
        LineList.OnAddLine := {$IFDEF FPC}@{$ENDIF}OnAddLine;
      try
        DeltaJ := JChange.Value;
        if NoUpper or (DeltaJ < JChange.MinValue+4) then
          DeltaJ := -10000 - (DeltaJ-JChange.MinValue) div 2;
        Excitation := RandomExcitation;
        Excitation.Rank := ExcitationRank.Value;
        Excitation.Simple := ExcitationType.ItemIndex <= 0;
        if not Excitation.Simple then begin
          { TODO : Make more general }
          Excitation.DeltaJ := ExcitationType.ItemIndex div 2;
          if Odd(ExcitationType.ItemIndex) then
            Excitation.DeltaJ := -Excitation.DeltaJ;
        end;
        Problem := nil;
        Molecule.AddToSpectrum(LineList, plNone, Excitation, Problem, Simulation,
                               ReadJSpinBox(LowerJ), SymL,
                               ReadJSpinBox(UpperJ), SymU,
                               ManifoldL, ManifoldU,
                               DeltaJ, TransitionMoments, TransitionMoment,
                               CalcStateL, CalcStateU, Perturbations.Checked);
        nLines := LineList.nLines;
        nShown := nLines;
        if Filter.Checked and StrengthCheck.Checked then begin
          if not TryStrToFloatEx(Fraction.Text, MinStrength) then
            raise EReportException.Create('Bad fraction:'+Fraction.Text);
          if not PlotAllSelected and not AutoRange.Checked then
            ComputeRanges(Simulation.PlotUnits)
          else
            FPlotMin := nil; { Selection by frequency already done, or not required }
          SetLength(Strengths, LineList.nLines);
          MaxStrength := 0;
          for i := 0 to LineList.nLines - 1 do
            with LineList.PLine[i]^ do begin
              Strengths[i] := Simulation.LineIntensity(LineList.PLine[i]^, Simulation.PlotUnits, Problem);
              if not OutsideRange(Frequency) then
                MaxStrength := Max(Strengths[i], MaxStrength);
            end;
          MinStrength := MinStrength * MaxStrength;
          nLines := 0;
          nShown := 0;
          for i := 0 to LineList.nLines - 1 do
            if Strengths[i] >= MinStrength then begin
              if i > nLines then
                LineList.PLine[nLines]^ := LineList.PLine[i]^;
              if not OutsideRange(LineList.PLine[i]^.Frequency) then
                Inc(nShown);
              Inc(nLines);
            end;
          LineList.SetnLines(nLines);
        end else if Filter.Checked and InPlotRange.Checked and not AddLinesToLinelist then begin
          { Simple count only }
          ComputeRanges(Simulation.PlotUnits);
          nShown := 0;
          for i := 0 to LineList.nLines - 1 do
            if not OutsideRange(LineList.PLine[i]^.Frequency) then
              Inc(nShown);
        end;

        FilterStatus.Caption := IntToStr(nShown)+ ' lines';
        if Assigned(Problem) then begin
          Caption := Problem.Message;
          Problem.Free;
        end;
        if AddLinesToLinelist then begin
          LineList.FindLines(0, 1e99, LineList.Units);
          Result := LineForm.ShowLines(Molecule.Mixture.Simulations[0],False,Filter.Checked and DiscardDuplicates.Checked);
        end;
        SetFocus;
      except
        LineList.Free;
        raise;
      end;
      if not AddLinesToLinelist and ((PlotLines.Checked) or PlotAllSelected) then begin
        LineList.Partial := True;
        Molecule.Mixture.LineList := LineList;
        if not PlotAllSelected and AutoRange.Checked then begin
          { Adjust plot range if necessary }
          with MainForm.Simulation do begin
            Mi := Convert(Fmin, PlotUnits, LineList.Units);
            Mx := Convert(Fmax, PlotUnits, LineList.Units);
          end;
          InOrder(Mi, Mx);
          PlotAllSelected := (LineList.FMin < Mi) or (LineList.FMax > Mx);
        end;
        if PlotAllSelected then
          MainForm.FindLimits(nLines = 0);
        MainForm.ShowSimulation;
      end else
        FreeAndNil(LineList);
    end;
  finally
    EndGlobalUpdate;
  end;
end;

function TTransitionForm.StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer;
begin
  Result := 0;
  if (MNode is TTransitionMoment)
     or ((MNode is TTransitionMoments) { and (TTransitionMoments(MNode).Count > 1)} ) { ExternalMolecule transition moments have no states }
     or ((MNode is TMolecule) and (TMolecule(MNode).nTransitionMoments <>1)) then begin
    ActiveObject.AddItem(Name1 + Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end;
end;

procedure TTransitionForm.StuffSelector(MNode: TMNode);
begin
  {$IFNDEF LCL_}
  OnConstrainedResize := FormConstrainedResize;
  {$ENDIF}
  UpperObject.Clear;
  LowerObject.Clear;
  StuffSelectorHelper(MNode,ActiveObject);
end;

procedure TTransitionForm.UpperJIncrement(Sender: TObject; var NewValue: Integer; LastValue, n: Integer);
var
  Mv: Integer;
begin
  Mv := (Sender as TSpinBox).MinValue;
  if (Mv < -2) and (NewValue <= Mv+2) then
    NewValue := (2 - (Mv and 1))*n; { Stepping from all when negative values are allowed starts from 0 }
  if Odd(NewValue+Mv) then
    Inc(NewValue); { Force reasonable spin }
end;

procedure TTransitionForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  SaveComboBox(UpperObject);
  SaveComboBox(LowerObject);
end;

procedure TTransitionForm.FormConstrainedResize(Sender: TObject;
  var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);
begin
  MinHeight := ClientToScreen(Point(0,0)).Y - Top + LabelPanel.Height;
end;

procedure TTransitionForm.IndexChangeIncrement(Sender: TObject;
  var NewValue: Integer; LastValue, n: Integer);
begin
  if (LastValue = (Sender as TSpinBox).MinValue) then
    if n > 0 then
      NewValue := 0
    else
      NewValue := -1
  else if ((LastValue = 0) and (n < 0)) or ((LastValue = -1) and (n > 0)) then
    NewValue := (Sender as TSpinBox).MinValue;
end;

procedure TTransitionForm.Loaded;
begin
  inherited;
  UpdateSymmetrySpin(UpperSymmetry, nil, False);
  UpdateSymmetrySpin(LowerSymmetry, nil, False);
  OrgGrid.CopyLabels := True;
  TranGrid.CopyLabels := True;
  OriginalLabel.Caption := '';
  TransformedLabel.Caption := '';
  OrgGrid.FixedSizeNumbers := True;
  TranGrid.FixedSizeNumbers := True;
  {$IFDEF FPC}
    {$IFDEF MSWINDOWS}
    { Unicode fix - may not be necessary long term }
    OrgHonlLondon.Caption := 'Hönl-London';
    OrgHonlLondon.Top := 1;
    OrgGrid.Top := OrgHonlLondon.Top + OrgHonlLondon.Height + 2;
    TranHonlLondon.Caption := 'Hönl-London';
    TranHonlLondon.Top := 1; { Seems slightly oversize on windows }
    Square.Top := 1;
    TranGrid.Top := TranHonlLondon.Top + TranHonlLondon.Height + 2;
    {$ENDIF}
  {$ENDIF}
end;

procedure TTransitionForm.LowerIndexGetIntegerValue(Sender: TObject; var NewValue: Integer; TheText: string; var StringValid: Boolean);
begin
  if SameText(TheText, 'all') or (TheText = '') then begin
    NewValue := 0;
    StringValid := True;
  end;
end;

procedure TTransitionForm.LowerIndexSetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: string);
begin
  if NewValue = 0 then
    NewText := 'all';
end;

procedure TTransitionForm.LowerSymmetrySetIntegerValue(Sender: TObject; NewValue: Integer; var NewText: String);
begin
  with Sender as TSpinBox do
    if NewValue <= MinValue + Increment div 2 then
      NewText := 'all';
end;

procedure TTransitionForm.UpperSymmetryGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: String; var StringValid: Boolean);
begin
  if SameText(TheText, 'all') or (TheText = '') then begin
    NewValue := (Sender as TSpinBox).MinValue;
    StringValid := True;
  end;
end;

procedure TTransitionForm.JChangeGetIntegerValue(
  Sender: TObject; var NewValue: Integer; TheText: String;
  var StringValid: Boolean);
var
  Change: Integer;
begin
  if SameText(TheText, 'all') or (TheText = '') then begin
    NewValue := (Sender as TSpinBox).MinValue;
    StringValid := True;
  end else if TheText = '<>' then begin
    NewValue := (Sender as TSpinBox).MinValue+2;
    StringValid := True;
  end else if (Length(TheText) = 1) then begin
    Change := (Ord(LowerCase(TheText)[1])-Ord('q'))*2;
    if Abs(Change) <= 18 then begin
      NewValue := Change;
      StringValid := True;
    end;
  end;
end;

procedure TTransitionForm.JChangeSetIntegerValue(
  Sender: TObject; NewValue: Integer; var NewText: String);
begin
  case NewValue - (Sender as TSpinBox).MinValue of
     0: NewText := 'all';
     2: NewText := '<>';
  else
    NewText := Char((NewValue div 2) + Ord('Q'));
  end;
end;

procedure TTransitionForm.SymChangeGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: string; var StringValid: Boolean);
begin
  if SameText(TheText, 'all') or (TheText = '') then begin
    NewValue := (Sender as TSpinBox).MinValue;
    StringValid := True;
  end else if TheText = '=' then begin
    NewValue := 0;
    StringValid := True;
  end else if (TheText = '<>') or (TheText = '!=') then begin
    NewValue := 1;
    StringValid := True;
  end;
end;

procedure TTransitionForm.SymChangeSetIntegerValue(Sender: TObject;
  NewValue: Integer; var NewText: string);
begin
  case NewValue of
    -1: NewText := 'all';
     0: NewText := '=';
     1: NewText := '<>';
  end;
end;

function TTransitionForm.OnAddLine(const ALine: TLine): Boolean;
  function DiscardIndex(Filter, Index: Integer; Eigenvalues: TEigenvalues):Boolean;
  begin
    Result := False; { Assume OK }
    if (Filter = 0) or (Eigenvalues = nil) then
      Exit; { No filter }
    if Filter > 0 then begin
      if Index = Filter then
        Exit;
    end else begin
      { Filter < 0 - count from top }
      if Index - Eigenvalues.Size - 1 = Filter then
        Exit;
    end;
    Result := True;
  end;
  function DiscardQno(Filter: TQuantumNumber; Index: Integer; Eigenvalues: TEigenvalues; QnoType: TQnoType):Boolean;
  begin
    Result := False; { Assume OK }
    if (Filter = Low(TQuantumNumber)) or (Eigenvalues = nil) then
      Exit; { No filter }
    Result := Eigenvalues.DominantBasis[Index].Qno(QnoType) <> Filter
  end;
  function DiscardQnoChange(Filter: Integer; const ALine: TLine; QnoType: TQnoType):Boolean;
  begin
    Result := False; { Assume OK }
    if Filter = Low(Integer) then
      Exit; { No filter }
    with ALine do
      Result := EigenvaluesU.DominantBasis[IndexU].Qno(QnoType) - EigenvaluesL.DominantBasis[IndexL].Qno(QnoType) <> Filter;
  end;
  function DiscardState(State: TState; Index: Integer; Eigenvalues: TEigenvalues):Boolean;
  begin
    Result := False; { Assume OK }
    if (State = nil) or (Eigenvalues = nil) then
      Exit; { No filter }
    Result := Eigenvalues.DominantBasis[Index].State <> State;
  end;
var
  Equal: Boolean;
  Molecule: TMolecule;
  i: Integer;
begin
  with ALine do begin
    { Filter by frequency first }
    Result := False;
    if OutsideRange(Frequency) then
      Exit;
    Result := False; { Assume discard }
    if FIndexChange <> Low(Integer) then
      if IndexU-IndexL <> FIndexChange then
        Exit;
    if DiscardIndex(FUpperIndex, IndexU, EigenValuesU) then
      Exit;
    if DiscardIndex(FLowerIndex, IndexL, EigenvaluesL) then
      Exit;
    if DiscardQno(FQnoAU, IndexU, EigenvaluesU, qtA)
       or DiscardQno(FQnoBU, IndexU, EigenvaluesU, qtB)
       or DiscardQno(FQnoAL, IndexL, EigenvaluesL, qtA)
       or DiscardQno(FQnoBL, IndexL, EigenvaluesL, qtB) then
      Exit;
    if DiscardState(FStateU, IndexU, EigenvaluesU)
       or DiscardState(FStateL, IndexL, EigenvaluesL) then
      Exit;
    if (EigenvaluesU <> nil) and (EigenvaluesL <> nil) then begin
      if DiscardQnoChange(FAChange, ALine, qtA)
         or DiscardQnoChange(FBChange, ALine, qtB) then
        Exit;
      if (FSymChange >= 0) then begin
        Molecule := EigenvaluesU.Manifold.Molecule;
        if Molecule.JAdjustSym then
          Equal := Molecule.JDeAdjustedSym(EigenvaluesU.J, EigenvaluesU.Sym)
                   = Molecule.JDeAdjustedSym(EigenvaluesL.J, EigenvaluesL.Sym)
        else
          Equal := EigenvaluesU.Sym = EigenvaluesL.Sym;
        if Equal <> (FSymChange=0) then
          Exit;
      end;
    end;
  end;
  Result := True; { Keep line }
end;

procedure TTransitionForm.OrgGridGetCellLabel(Sender: TObject; Col,
  Row: Integer; var s: string; var IsText: Boolean);
begin
  if Row > 0 then
    if Col = 0 then
      s := IntToStr(Row)
    else if Col = 1 then begin
      s := OrgGrid.Lbl[Row];
      IsText := True;
    end;
end;

procedure TTransitionForm.FormResize(Sender: TObject);
begin
  CheckExtraVisible;
  Constraints.MinWidth := PlotLines.Left + PlotLines.Width + Width - Label11.Left + 4;
end;

end.
