unit zMatrixFrm;

interface

uses
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Menus, VibGridFrm, StdCtrls,
  ExtCtrls, Grids, NGrid, Split, MNode, zMatrix, AlEdit, EditExtr,
  NewSpin, ScrlCtrl, GraphI, GraphM, Graphs, ComCtrls;

type

  TZMatrixForm = class(TVibGridForm)
    CheckItem: TMenuItem;
    ApplyItem: TMenuItem;
    LoadItem: TMenuItem;
    PrintItem: TMenuItem;
    procedure CheckItemClick(Sender: TObject);
    procedure PrintItemClick(Sender: TObject);
    procedure ApplyItemClick(Sender: TObject);
    procedure LoadItemClick(Sender: TObject);
  private
    zMatrix: TzMatrix;
  protected
    procedure SetActiveObject(const Value: TMNode); override;
  public
    procedure PopulateForm(ACol:Integer); overload; override;
    procedure Replot(ACol:Integer = -1); override;
    procedure ExtraPlot(AzMatrix: TzMatrix; ACol, ix, iy: Integer; Original: Boolean); override;
  end;

function ConnectZmatrixForm(zMatrix: TzMatrix): TComponent;


implementation

uses Math, Mixture, Global, Vib, VibPlotFrm;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function ConnectZmatrixForm(zMatrix: TzMatrix): TComponent;
begin
  zMatrix.FindPrincipalAxes(False);
  Result := ConnectVibFrm(zMatrix, TZMatrixForm);
end;

procedure TZMatrixForm.ApplyItemClick(Sender: TObject);
begin
  PrintItemClick(Sender);
  zMatrix.ApplyToStateEx(True,False,nil);
end;

procedure TZMatrixForm.CheckItemClick(Sender: TObject);
begin
  zMatrix.FindPrincipalAxes(False);
  PopulateForm;
end;

procedure TZMatrixForm.ExtraPlot(AzMatrix: TzMatrix; ACol, ix, iy: Integer; Original: Boolean);
  procedure JoinP(n1, n2: Integer; Style: Integer = 1);
  var
    p1, p2: Txyz;
  begin
    with AzMatrix do
      if Original then begin
        p1 := Atoms[n1].xyz;
        p2 := Atoms[n2].xyz;
      end else begin
        p1 := Atoms[n1].pxyz;
        p2 := Atoms[n2].pxyz;
      end;
    Rotate(p1, ix);
    Rotate(p2, ix);
    G.Style(Style);
    G.Point(p1[ix], p1[iy]);
    G.Join(p2[ix], p2[iy]);
    G.Style(0);
  end;
var
  An: array [1..4] of Integer;
  nAtoms, i, ind: Integer;
begin
  with AzMatrix do
    if (ACol > 1) and (ACol <= Length(A)) and (Length(A[ACol-1]) > 1) and (Trim(A[ACol-1][1]) <> '') then begin
      nAtoms := 0;
      for i := 1 to 4 do begin
        An[i] := -1;
        ind := Max(i*2-3,0);
        if ind <= High(A[ACol-1]) then
          An[i] := FindAtom(A[ACol-1][ind]);
        if An[i] >= 0 then
          nAtoms := i
        else
          Break;
      end;
      if nAtoms > 1 then begin
        JoinP(An[1], An[2]);
        if nAtoms > 2 then begin
          JoinP(An[2], An[3]);
          if nAtoms > 3 then begin
            JoinP(An[3], An[4], 2);
            JoinP(An[2], An[4], 2);
          end;
        end;
      end;
    end;
end;

procedure TZMatrixForm.PrintItemClick(Sender: TObject);
begin
  ReShowLogForm;
  zMatrix.FindPrincipalAxes(True);
  PopulateForm;
end;

procedure TZMatrixForm.LoadItemClick(Sender: TObject);
begin
  zMatrix.SetFromState(zMatrix.Owner as TElectronicState, False);
  PopulateForm;
end;

procedure TZMatrixForm.PopulateForm(ACol:Integer);
var
  i: Integer;
begin
  if csLoading in ComponentState then
    Exit;
  Caption := zMatrix.Name + ' for ' + (zMatrix.Owner as TState).UniqueName;
  TrimArray(zMatrix.A);
  Grid.RowCount := Length(zMatrix.A)+2;
  ColsHeadroom := 0;
  inherited;
  for i := 8 to Grid.ColCount-1 do
    Grid.Title[i] := ' '; { Room for a comment... }
  Replot(ACol);
end;

procedure TZMatrixForm.Replot(ACol: Integer);
begin
  if ACol < 0 then
    ACol := Grid.Row;
  PlotZmatrix(zMatrix, zMatrix.LastError, ACol, nil, 0, True, ACol-1);
end;

procedure TZMatrixForm.SetActiveObject(const Value: TMNode);
begin
  inherited;
  zMatrix := Value as TzMatrix;
  RowSelect := True;
end;

end.
