unit TextForm;
{
Slave - simple text editor that saves files when requested by other applications
Copyright (C) 1999-2007,2013 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}

interface
{$I utils.inc}
uses 
  {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
  {$IFNDEF CLX_} Graphics, Controls, Forms, Dialogs, ImgList,
                 Menus, ExtCtrls, ActnList, StdCtrls, ComCtrls, ToolWin,
  {$ELSE} QGraphics, QControls, QForms, QDialogs, QImgList, QActnList,
                 QMenus, QExtCtrls, QStdCtrls, QComCtrls, QTypes, {$ENDIF}
  {$IFDEF UNIX} CompatMsg, {$ENDIF}
  SysUtils, Classes, TextWin, FxFntDlg, LogFind,
  LogControl, EditLog, TextActn, ProjAbout, ToolBarEx,
  ExActn, LogActn, PrintPre, TextPre, ScrlCtrl, Recent;

type

  { TMainText }

  TMainText = class(TTextChild)
    ToolBar: TToolBarEx;
    Help1: TMenuItem;
    About: TMenuItem;
    Project: TProject;
    NewButton: TToolButtonEx;
    TextFind: TLogFind;
    TextFindNext: TLogFindNext;
    TextNew: TTextNew;
    TextOpen: TTextOpen;
    TextSave: TTextSave;
    OpenButton: TToolButtonEx;
    SaveButton: TToolButtonEx;
    Spacer1: TToolButtonEx;
    CutButton: TToolButtonEx;
    CopyButton: TToolButtonEx;
    PasteButton: TToolButtonEx;
    UndoButton: TToolButtonEx;
    RedoButton: TToolButtonEx;
    Spacer2: TToolButtonEx;
    FindButton: TToolButtonEx;
    FindNextButton: TToolButtonEx;
    Another: TMenuItem;
    SortUpButton: TToolButtonEx;
    SortDownButton: TToolButtonEx;
    Slave: TMenuItem;
    PasteSpecialItem: TMenuItem;
    RecentFiles: TRecentFiles;
    procedure AboutClick(Sender: TObject);
    procedure AnotherClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SlaveClick(Sender: TObject);
    procedure FormLoaded(Sender: TObject);
    procedure MRUFileListMRUItemClick(Sender: TObject; AFilename: String);
    procedure FormDestroy(Sender: TObject);
    procedure PasteSpecialItemClick(Sender: TObject);
  private
    {$IFDEF UNIX}
    SaveAllMsg, GotoLineMsg: TCompatMsg;
    {$ENDIF}
    procedure DisplayHint(Sender: TObject);
    procedure TryAndSave(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
    procedure GotoLine(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
    procedure PasteSpecial(Sender: TObject);
  public
    {$IFDEF MSWINDOWS}
    procedure WndProc(var Msg: TMessage); override;
    {$ENDIF}
    {$IFDEF UNIX}
    procedure UpdateStatus(Sender: TObject; var Done: Boolean);
    {$ENDIF}
    procedure SetFileName(FileName:string; OldFileName: string); override;
  end;

var
  MainText: TMainText;
  WM_Save_All: Word;
  WM_GOTO_LINE: Word;

implementation

uses {$IFDEF LCL_} LResources, {$ENDIF}
     Clipbrd, Huge, Detach, WCompat, UCompat, ImageMod;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TMainText }

procedure TMainText.AboutClick(Sender: TObject);
begin
  Project.Execute;
end;

procedure TMainText.AnotherClick(Sender: TObject);
var
  Ext: string;
begin
  if OpenDialog.DefaultExt <> '' then
    Ext := Concat('.', OpenDialog.DefaultExt);
  Detach1(Application.Exename, Ext);
end;

procedure TMainText.DisplayHint(Sender: TObject);
begin
  StatusBar.Panels[4].Text := Application.Hint;
end;

procedure TMainText.FormCreate(Sender: TObject);
begin
  inherited;
  {$IFDEF MSWINDOWS}
  WM_Save_All := RegisterWindowMessage('WM_Save_All');
  WM_GOTO_LINE := RegisterWindowMessage('WM_GOTO_LINE');
  Application.OnActivate := {$IFDEF FPC}@{$ENDIF}FormActivate;
  {$ENDIF}
  {$IFDEF UNIX}
    SaveAllMsg := TCompatMsg.Create(Self);
    with SaveAllMsg do begin
      MessageName := 'WM_Save_All';
      OnReceived := @TryAndSave;
    end;
    GotoLineMsg := TCompatMsg.Create(Self);
    with GotoLineMsg do begin
      MessageName := 'WM_GOTO_LINE';
      OnReceived := @GotoLine;
    end;
    Application.OnIdle := {$IFDEF FPC}@{$ENDIF}UpdateStatus;
  {$ENDIF}
  {$IFDEF FPC}
  Print.Visible := False;
  PrintPreview.Visible := False; { TODO: Make it work }
  {$ENDIF}
  AllowCtrlC;
  Application.OnHint := {$IFDEF FPC}@{$ENDIF}DisplayHint;
  AdjustMainMenu(MainMenu, About.OnClick);
end;

procedure TMainText.SetFileName(FileName: string; OldFileName: string);
begin
  if OldFileName = '' then
    OldFileName := OpenDialog.FileName;
  if OldFileName <> '' then
    RecentFiles.AddItem(OldFileName);
  inherited;
  Application.Title := ExtractFileName(Caption);
end;

procedure TMainText.TryAndSave(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
var
  Unused: Boolean;
begin
  if Memo.Modified and Slave.Checked then
    if OpenDialog.Filename = '' then
      Handled := False { Name not set, so can't save }
    else
      try
        SaveClick(Self);
        Handled := True; { OK }
      except
        Handled := False; { Fail during save }
        raise;
      end
  else
    Handled := True; { No action required }
  {$IFDEF UNIX}
  UpdateStatus(nil, Unused);
  {$ENDIF}
end;

procedure TMainText.GotoLine(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
begin
  try
    Memo.ChangeCaretPos(Memo.CaretPos.X, lParam-1, False);
    Handled := True;
  except
  end
end;

{$IFDEF MSWINDOWS}
procedure TMainText.WndProc(var Msg: TMessage);
var
  Handled: Boolean;
begin
  if Msg.Msg = WM_GOTO_LINE then
    GotoLine(Self, Msg.wParam, Msg.lParam, Handled)
  else if Msg.Msg = WM_Save_All then begin
    Msg.result := 0;
    try
      TryAndSave(Self, 0, 0, Handled);
    finally
      if not Handled then
        if InSendMessage then begin
          ReplyMessage(272727);
          Yield;
        end;     end;
  end else
    inherited WndProc(Msg);       { dispatch normally }
end;
{$ENDIF}

{$IFDEF UNIX} 
procedure TMainText.UpdateStatus(Sender: TObject; var Done: Boolean);
begin
  if Assigned(SaveAllMsg) then
    SaveAllMsg.Status := Ord(Memo.Modified and Slave.Checked);
end;
{$ENDIF}

procedure TMainText.SlaveClick(Sender: TObject);
begin
  inherited;
  Slave.Checked := not Slave.Checked;
end;

procedure TMainText.FormLoaded(Sender: TObject);
begin
  inherited;
{$IFDEF UNIX}
  Application.Icon := Icon;
{$ENDIF}
end;

procedure TMainText.MRUFileListMRUItemClick(Sender: TObject; AFilename: String);
var
  CanClose: Boolean;
begin
  CanClose := True;
  FormCloseQuery(Self, CanClose);
  if CanClose then
    LoadFile(AFilename, OpenDialog.FileName);
end;

procedure TMainText.PasteSpecial(Sender: TObject);
var
  GHandle: TGHandle;
begin
  GHandle := TGHandle.Create;
  try
    GHandle.SetFromClipboard(Clipboard.Formats[(Sender as TComponent).Tag]);
    Memo.SetSelTextBuf(GHandle.Item[0]);
  finally
    GHandle.Free;
  end;
end;

procedure TMainText.PasteSpecialItemClick(Sender: TObject);
var
  i, n: Integer;
  MenuItem: TMenuItem;
  Buffer: string;
begin
  while PasteSpecialItem.Count > Clipboard.FormatCount do
    PasteSpecialItem.Delete(PasteSpecialItem.Count-1);
  while PasteSpecialItem.Count < Clipboard.FormatCount do begin
    MenuItem := TMenuItem.Create(Self);
    PasteSpecialItem.Add(MenuItem);
  end;
  for i := 0 to Clipboard.FormatCount - 1 do begin
    SetLength(Buffer,255);
    {$IFDEF MSWINDOWS}
    n := GetClipboardFormatName(Clipboard.Formats[i],PChar(Buffer),Length(Buffer));
    if n = 0 then
      case Clipboard.Formats[i] of
        CF_TEXT: Buffer := 'TEXT';
        CF_BITMAP: Buffer := 'BITMAP';
        CF_METAFILEPICT: Buffer := 'METAFILEPICT';
        CF_SYLK: Buffer := 'SYLK';
        CF_DIF: Buffer := 'DIF';
        CF_TIFF: Buffer := 'TIFF';
        CF_OEMTEXT: Buffer := 'OEMTEXT';
        CF_DIB: Buffer := 'DIB';
        CF_PALETTE: Buffer := 'PALETTE';
        CF_PENDATA: Buffer := 'PENDATA';
        CF_RIFF: Buffer := 'RIFF';
        CF_WAVE: Buffer := 'WAVE';
        CF_UNICODETEXT: Buffer := 'UNICODETEXT';
        CF_ENHMETAFILE: Buffer := 'ENHMETAFILE';
        CF_HDROP: Buffer := 'HDROP';
        CF_LOCALE: Buffer := 'LOCALE ';
        CF_DIBV5: Buffer := 'DIBV5';
        CF_OWNERDISPLAY: Buffer := 'OWNERDISPLAY';
        CF_DSPTEXT: Buffer := 'DSPTEXT';
        CF_DSPBITMAP: Buffer := 'DSPBITMAP';
        CF_DSPMETAFILEPICT: Buffer := 'DSPMETAFILEPICT';
        CF_DSPENHMETAFILE: Buffer := 'DSPENHMETAFILE';
      else
        Buffer := '';
      end
    else
      SetLength(Buffer,n);
    {$ELSE}
    Buffer := '';
    {$ENDIF}
    with PasteSpecialItem.Items[i] do begin
      Tag := i;
      OnClick := {$IFDEF FPC}@{$ENDIF}PasteSpecial;
      Caption := Format('%s'#9'%d',[Buffer, Clipboard.Formats[i]])
    end;
  end;
end;

procedure TMainText.FormDestroy(Sender: TObject);
begin
  if OpenDialog.Filename <> '' then
    RecentFiles.AddItem(OpenDialog.Filename);
  inherited;
end;

end.
