unit IntVec;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses UCompat, Huge;

type

  TIntVector = class(TGVectorN)
  private
    function GetValue(i:Integer):Longint;
    procedure SetValue(i:Integer; v:Longint);
    procedure SetSize(n:Integer);
  public
    constructor Create(ASize: Integer);
    procedure Clear;
    procedure Move(Old, ANew:Integer);
    property Size read GetSize write SetSize;
    property Value[i:Integer]:Longint read GetValue write SetValue; default;
    procedure ClearRange(First, Count:Integer);
    function FirstItem: PInteger;
    function IndexOf(AValue: Longint): Integer;
  end;

  TIntColumnVector = class(TIntVector) { As TVector, but first index is 1 }
  public
    constructor Create(ASize: SizeUInt);
  end;

implementation

uses SysUtils;

constructor TIntVector.Create(ASize: Integer);
begin
  inherited Create(SizeOf(Longint));
  SetSize(ASize);
end;

constructor TIntColumnVector.Create(ASize: SizeUInt);
begin
  inherited Create(ASize);
  Offset := 1;
end;

function TIntVector.FirstItem: PInteger;
begin
  Result := Base;
end;

procedure TIntVector.ClearRange(First, Count:Integer);
begin
  FillChar(Plongint(Self.Item[First])^, (Count-First)*SizeOf(Longint), 0);
end;

procedure TIntVector.SetSize(n:Integer);
var
  Old: SizeUInt;
begin
  Old := Size;
  inherited Size := n;
  if Size > Old then
    ClearRange(Old+Offset, Size-Old);
end;

procedure TIntVector.Clear;
begin
  ClearRange(Offset, Size);
end;

function TIntVector.GetValue(i:Integer):Longint;
begin
  Result := PLongint(Item[i])^;
end;

procedure TIntVector.SetValue(i:Integer; v:Longint);
begin
  PLongint(Item[i])^ := v;
end;

procedure TIntVector.Move(Old, ANew:Integer);
var
  i: Integer;
  step: Integer;
  vold: Integer;
begin
  if Old > ANew then
    step := -1
  else
    step := 1;
  i := Old;
  vold := Value[i];
  while i <> ANew do begin
    Value[i] := Value[i+step];
    i := i + step;
  end;
  Value[ANew] := vold
end;

function TIntVector.IndexOf(AValue: Integer): Integer;
var
  p: PInteger;
  n: Integer;
begin
  n := Size;
  p := FirstItem;
  while (n > 0) and (p^ <> AValue) do begin
    Inc(p);
    Dec(n);
  end;
  if n > 0 then
    Result := Size - n + Offset
  else
    Result := Offset - 1;
end;

end.
