unit LineStrm;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes;

type

  { Helper class for reading a stream line by line }

  TLineStream = class
  private
    S: TStream;
    Buffer: AnsiString;
    FLastLine: Integer;
    EOF: Boolean;
  public
    constructor Create(Stream: TStream);
    function NextLine(var Line: AnsiString): Boolean;
    property LastLine: Integer read FLastLine;
  end;

implementation

uses Math;

{ TLineStream }

constructor TLineStream.Create(Stream: TStream);
begin
  S := Stream;
  FLastLine := 0;
  EOF := False;
end;

function TLineStream.NextLine(var Line: AnsiString): Boolean;
var
  nl, cr, eolpos: Integer;
begin
  repeat
    nl := Pos(#10,Buffer);
    cr := Pos(#13,Buffer);
    if nl > 0 then
      if cr > 0 then
        eolpos := Min(nl,cr)
      else
        eolpos := nl
    else if cr > 0 then
      eolpos := cr
    else
      eolpos := Length(Buffer)+1;
    if eolpos >= Length(Buffer) then begin
      { Line terminator at end of line or no line teminator found - read some more }
      eolpos := Length(Buffer);
      if not EOF then begin
        SetLength(Buffer, Length(Buffer)+10000);
        SetLength(Buffer, S.Read((PAnsiChar(Buffer)+eolpos)^,10000)+eolpos);
      end;
      if Length(buffer) = eolpos then begin
        { no more characters read - End of file }
        if EOF and (eolpos = 0) then begin
          { Read every line }
          Result := False;
          Exit;
        end;
        Inc(eolpos);
        EOF := True;
      end else
        eolpos := -1; { Force find of end of line char again }
    end;
  until eolpos > 0;
  Line := Copy(Buffer, 1, eolpos-1);
  if (eolpos < Length(Buffer)) and (Buffer[eolpos] <> Buffer[eolpos+1]) and (Buffer[eolpos+1] in [#10,#13]) then
    Inc(eolpos); { Discard two line terminators: CR+LF or LF+CR }
  Delete(Buffer, 1, eolpos);
  Inc(FLastLine);
  Result := True;
end;

end.
