unit NormRand;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

{ Normally distributed random numbers (standard deviation 1) }

function NRandom: Double;

implementation

{ Polar form of Box-Muller transformation;
  see e.g. http://www.taygeta.com/random/gaussian.html }

{ Algorithm generates pairs of numbers, so save second one }
var
  NextRandom: Double;
  NextAvailable: Boolean = False;

function NRandom: Double;
var
  w, x1, x2: Double;
begin
  if NextAvailable then begin
    Result := NextRandom;
    NextAvailable := False;
    Exit;
  end;
  repeat
    x1 := 2*Random-1;
    x2 := 2*Random-1;
    w := Sqr(x1) + Sqr(x2);
  until w < 1;
  w := Sqrt(-2*Ln(w)/w);
  NextRandom := x2*w;
  NextAvailable := True;
  Result := x1*w;
end;

end.
