﻿unit ShellSort;
{$I utils.inc}{ Copyright (C) 2005, 2015 C M Western; see UCompat.pas for license}
interface

type
  TShellSorter = class
    function InOrder(I,J:Integer):Boolean; virtual; abstract;
    procedure Sort(N: Integer); { Indexed 0..N-1 }
  end;

var
  ShellSortOffsets: array of Integer;

procedure SwapValues(var A, B; Size: Integer);

procedure Sort(var A: array of Integer; nItems: Integer = -1); overload;
procedure Sort(var A: array of Double; nItems: Integer = -1); overload;

{ Sort array A, making the same changes in array B }
procedure Sort2(var A, B: array of Double; nItems: Integer = -1);

implementation

uses Math, Worker;

//procedure Sort(var A: array of Integer; nItems: Integer);
//var
//  I, J, Offset, t: Integer;
//  IsInOrder: Boolean;
//begin
//  if nItems < 0 then
//    nItems := Length(A);
//  Offset := nItems;
//  while Offset > 1 do begin
//    Offset := Offset shr 1;
//    repeat
//      IsInOrder := True;
//      for J := 0 to nItems - Offset - 1 do begin
//        I := J + Offset;
//        if A[J] > A[I] then begin
//          IsInOrder := False;
//          t := A[I];
//          A[I] := A[J];
//          A[J] := t;
//        end;
//      end;
//    until IsInOrder;
//  end;
//end;

//procedure Sort(var A: array of Double; nItems: Integer);
//var
//  I, J, Offset: Integer;
//  IsInOrder: Boolean;
//  t: Double;
//begin
//  if nItems < 0 then
//    nItems := Length(A);
//  Offset := nItems;
//  while Offset > 1 do begin
//    Offset := Offset shr 1;
//    repeat
//      IsInOrder := True;
//      for J := 0 to nItems - Offset - 1 do begin
//        I := J + Offset;
//        if A[J] > A[I] then begin
//          IsInOrder := False;
//          t := A[I];
//          A[I] := A[J];
//          A[J] := t;
//        end;
//      end;
//    until IsInOrder;
//  end;
//end;

{ TShellSorter }

//procedure TShellSorter.Sort(N: Integer);
//var
//  I, J, Offset: Integer;
//  IsInOrder: Boolean;
//begin
//  { ShellSort }
//  Offset := N;
//  while Offset > 1 do begin
//    Offset := Offset shr 1;
//    repeat
//      IsInOrder := True;
//      for J := 0 to N - Offset - 1 do begin
//        I := J + Offset;
//        IsInOrder := InOrder(J,I) and IsInOrder;
//      end;
//    until IsInOrder;
//  end;
//end;

procedure Sort(var A: array of Integer; nItems: Integer);
var
  I, J, Offset, OffsetIndex, t: Integer;
  IsInOrder: Boolean;
begin
  if nItems < 0 then
    nItems := Length(A);
  OffsetIndex := 0;
  while ShellSortOffsets[OffsetIndex] < nItems do
    Inc(OffsetIndex);
  while OffsetIndex > 0 do begin
    Dec(OffsetIndex);
    Offset := ShellSortOffsets[OffsetIndex];
    repeat
      IsInOrder := True;
      for J := 0 to nItems - Offset - 1 do begin
        I := J + Offset;
        if A[J] > A[I] then begin
          IsInOrder := False;
          t := A[I];
          A[I] := A[J];
          A[J] := t;
        end;
      end;
    until IsInOrder;
  end;
end;

procedure Sort(var A: array of Double; nItems: Integer);
var
  I, J, Offset, OffsetIndex: Integer;
  IsInOrder: Boolean;
  t: Double;
begin
  if nItems < 0 then
    nItems := Length(A);
  OffsetIndex := 0;
  while ShellSortOffsets[OffsetIndex] < nItems do
    Inc(OffsetIndex);
  while OffsetIndex > 0 do begin
    Dec(OffsetIndex);
    Offset := ShellSortOffsets[OffsetIndex];
    repeat
      IsInOrder := True;
      for J := 0 to nItems - Offset - 1 do begin
        I := J + Offset;
        if A[J] > A[I] then begin
          IsInOrder := False;
          t := A[I];
          A[I] := A[J];
          A[J] := t;
        end;
      end;
    until IsInOrder;
  end;
end;

procedure Sort2(var A, B: array of Double; nItems: Integer);
var
  I, J, Offset, OffsetIndex: Integer;
  IsInOrder: Boolean;
  t: Double;
begin
  if nItems < 0 then
    nItems := Length(A);
  OffsetIndex := 0;
  while ShellSortOffsets[OffsetIndex] < nItems do
    Inc(OffsetIndex);
  while OffsetIndex > 0 do begin
    Dec(OffsetIndex);
    Offset := ShellSortOffsets[OffsetIndex];
    repeat
      IsInOrder := True;
      for J := 0 to nItems - Offset - 1 do begin
        I := J + Offset;
        if A[J] > A[I] then begin
          IsInOrder := False;
          t := A[I];
          A[I] := A[J];
          A[J] := t;
          t := B[I];
          B[I] := B[J];
          B[J] := t;
        end;
      end;
    until IsInOrder;
  end;
end;

procedure TShellSorter.Sort(N: Integer);
var
  I, J, Offset, OffsetIndex: Integer;
  IsInOrder: Boolean;
begin
  OffsetIndex := 0;
  while ShellSortOffsets[OffsetIndex] < N do
    Inc(OffsetIndex);
  while OffsetIndex > 0 do begin
    Dec(OffsetIndex);
    Offset := ShellSortOffsets[OffsetIndex];
    repeat
      IsInOrder := True;
      for J := 0 to N - Offset - 1 do begin
        I := J + Offset;
        IsInOrder := InOrder(J,I) and IsInOrder;
      end;
    until IsInOrder;
  end;
end;

procedure SwapValues(var A, B; Size: Integer);
var
  PA, PB: PAnsiChar;
  T: Integer;
  C: AnsiChar;
begin
  PA := @A;
  PB := @B;
  while Size >= SizeOf(Integer) do begin
    T := PInteger(PA)^;
    PInteger(PA)^ := PInteger(PB)^;
    PInteger(PB)^ := T;
    Dec(Size, SizeOf(Integer));
    Inc(PA, SizeOf(Integer));
    Inc(PB, SizeOf(Integer));
  end;
  while Size > 0 do begin
    C := PA^;
    PA^ := PB^;
    PB^ := C;
    Dec(Size, 1);
    Inc(PA, 1);
    Inc(PB, 1);
  end;
end;

procedure SetupOffsets;
var
  V: Double;
  I: Integer;
begin
  { Compute offsets based on:
    Tokuda, Naoyuki (1992). "An Improved Shellsort".
    In van Leeuven, Jan. Proceedings of the IFIP 12th World Computer Congress on Algorithms,
    Software, Architecture. Amsterdam: North-Holland Publishing Co. pp. 449–457. ISBN 0-444-89747-X.
  }
  SetLength(ShellSortOffsets, 25);
  ShellSortOffsets[0] := 1;
  V := 1;
  for I := 1 to High(ShellSortOffsets) do begin
    V := 2.25*V + 1;
    ShellSortOffsets[I] := Ceil(V);
  end;
end;

initialization
  SetupOffsets;
end.
