unit f2c;
{$I utils.inc}
interface

uses UCompat;

type
  FtnLen = Integer;
  LogBool = LongBool;

{ Notes:
  Linux - math routines provided by system libm library; no chkstk routines needed
          fortran adds underscore to routine name; fpc alias required to prevent name mangling;
  Windows with fpc:
          math routines must be provided
    Win32: fortran adds both leading and trailing underscore
    Win64: fortran only adds a trailing underscore by default
  Windows with Delphi:
          math routines must be provided
    Win32: gfortran does not produce object files compatible with Delphi, compile
           to assembler and used scripted edit to produce .asm file compatible with tasm32
           fortran adds both leading and trailing underscore; edit as above leaves only trailing underscore.
    Win64: gfortran produces files that are in principle comaptible with delphi,
           but seems to be a specific bug with relocating rip relative constants
           fortran only adds a trailing underscore by default
  MacOSX:  math routines provided by system libm library; no chkstk routines needed
           fortran adds both leading and trailing underscore
}
{$IFDEF MSWINDOWS}
  {$IFDEF WIN64}
    {$IFDEF FPC}
      procedure __chkstk_ms; cdecl; external name '___chkstk_ms';
      {$LINK mschkstk.o}
    {$ELSE}
      procedure ___chkstk_ms; cdecl;
    {$ENDIF}
  {$ELSE}
    procedure __chkstk_ms; cdecl;
  {$ENDIF}

  function _sqrt(a: Double): Double; cdecl;
  function _log(a: Double): Double; cdecl;
  function _logf(a: Single): Double; cdecl;
  function _log10(a: Double): Double; cdecl;
  function _exp(a: Double): Double; cdecl;
  function _sin(a: Double): Double; cdecl;
  function _cos(a: Double): Double; cdecl;
  function _pow(a, b: Double): Double; cdecl;
  function _lroundf(a:Double): Integer; cdecl;

  {$IFDEF FPC}
    procedure _memcpy(var dest; const src; n: SizeInt); cdecl;
    procedure _memmove(var dest; const src; n: SizeInt); cdecl;
    procedure _memset(var s; c: Integer; n: SizeInt); cdecl;
    function _memcmp(const s1; const s2; n: SizeInt):Integer; cdecl;
  {$ELSE}
    {$IFDEF WIN64}
      procedure memmove(var dest; const src; n: SizeInt); cdecl;
      function memcmp(const s1; const s2; n: SizeInt):Integer; cdecl;
    {$ELSE}
      procedure _memmove(var dest; const src; n: SizeInt); cdecl;
    {$ENDIF}
  {$ENDIF}
{$ENDIF}

{$IFDEF FPC}
  {$DEFINE UND}
  {$IFDEF UNIX}
    {$DEFINE NOUND}
    {$LINKLIB m}
  {$ENDIF}
{$ENDIF}

function __powidf2(a:Double; p:Integer): Double; cdecl;
function _gfortran_pow_i4_i4(base:Integer; exponent:Integer): Integer; cdecl;
function _gfortran_compare_string(la: FtnLen; a: PAnsiChar; lb: FtnLen; b: PAnsiChar): Integer; cdecl;
procedure _gfortran_concat_string(destlen: FtnLen; dest: PAnsiChar; len1: FtnLen; s1: PAnsiChar; len2: FtnLen; s2: PAnsiChar); cdecl;
procedure _gfortran_runtime_error_at(Where: PAnsiChar; Message: PAnsiChar); cdecl;

function s_cat(lp: PAnsiChar; rpp: PPAnsiChar; rnp: PInteger; var np: Integer; ll: FtnLen):Integer; cdecl;
function s_cmp(a, b: PAnsiChar; la, lb: FtnLen): Integer; cdecl;
procedure s_copy(a, b: PAnsiChar; la, lb: FtnLen); cdecl;

function pprint_(var tag, n: Integer; var d: Double):Integer; cdecl;

function errstop_(s: PAnsiChar; l: Integer):Integer; cdecl;
function errstopf_(fmt: PAnsiChar; args: Pointer { ... } ):Integer; cdecl;
function clog(s: PAnsiChar):Integer; cdecl;

function disnan_(var a:Double):Integer; cdecl;

implementation

uses { Compat, } Math, Sysutils, Worker;

{$IFDEF FPC}
  {$MACRO ON}
  {$IFDEF DARWIN}
    {$DEFINE alias_:=alias: #95}
  {$ELSE}
    {$IFDEF WIN32}
      {$DEFINE alias_:=alias: #95}
    {$ELSE}
      {$DEFINE alias_:=alias:}
    {$ENDIF}
  {$ENDIF}
{$ENDIF}

{$IFDEF MSWINDOWS}
function _sqrt(a: Double): Double; cdecl; {$IFDEF FPC} alias_'sqrt'; {$ENDIF}
begin
  Result := Sqrt(a);
end;

function _log(a: Double): Double; cdecl; {$IFDEF FPC} alias_'log'; {$ENDIF}
begin
  Result := Ln(a);
end;

function _logf(a: Single): Double; cdecl; {$IFDEF FPC} alias_'logf'; {$ENDIF}
begin
  Result := Ln(a);
end;

function _log10(a: Double): Double; cdecl; {$IFDEF FPC} alias_'log10'; {$ENDIF}
begin
  Result := Log10(a);
end;

function _exp(a: Double): Double; cdecl; {$IFDEF FPC} alias_'exp'; {$ENDIF}
begin
  Result := Exp(a);
end;

function _sin(a: Double): Double; cdecl; {$IFDEF FPC} alias_'sin'; {$ENDIF}
begin
  Result := Sin(a);
end;

function _cos(a: Double): Double; cdecl; {$IFDEF FPC} alias_'cos'; {$ENDIF}
begin
  Result := Cos(a);
end;

function _pow(a, b: Double): Double; cdecl; {$IFDEF FPC} alias_'pow'; {$ENDIF}
begin
  Result := Power(a,b);
end;

function _lroundf(a:Double): Integer; cdecl; {$IFDEF FPC} alias_'lroundf'; {$ENDIF}
begin
  Result := Round(a);
end;

{$IFDEF FPC}
procedure _memcpy(var dest; const src; n: SizeInt); cdecl; {$IFDEF FPC} alias_'memcpy'; {$ENDIF}
begin
  Move(src, dest, n);
end;

procedure _memmove(var dest; const src; n: PtrInt); cdecl; {$IFDEF FPC} alias_'memmove'; {$ENDIF}
begin
  Move(src, dest, n);
end;

procedure _memset(var s; c: Integer; n: PtrInt); cdecl; {$IFDEF FPC} alias_'memset'; {$ENDIF}
begin
  FillByte(s, n, Byte(c));
end;

function _memcmp(const s1; const s2; n: SizeInt):Integer; cdecl; {$IFDEF FPC} alias_'memcmp'; {$ENDIF}
var
  q: Byte;
  p1, p2: PByte;
  i: Integer;
begin
  p1 := @s1;
  p2 := @s2;
  Result := 0;
  for i := 0 to n-1 do begin
    Result := p1^ - p2^;
    if Result <> 0 then
      Exit;
  end;
end;

{$ELSE}
{$IFDEF WIN64}
procedure memmove(var dest; const src; n: PtrInt); cdecl;
begin
  Move(src, dest, n);
end;

function memcmp(const s1; const s2; n: SizeInt):Integer; cdecl;
var
  q: Byte;
  p1, p2: PByte;
  i: Integer;
begin
  p1 := @s1;
  p2 := @s2;
  Result := 0;
  for i := 0 to n-1 do begin
    Result := p1^ - p2^;
    if Result <> 0 then
      Exit;
  end;
end;
{$ELSE}
procedure _memmove(var dest; const src; n: SizeInt); cdecl;
begin
  Move(src, dest, n);
end;
{$ENDIF}
{$ENDIF}

{$IFDEF FPC}
{$ASMMODE ATT}
{$IFDEF WIN64}
{$IFDEF NOTUSED}
Generates spurious(?) sub, 8, rsp at start, so use separate .s file
procedure __chkstk_ms; cdecl; assembler; {$IFDEF FPC} alias_'__chkstk_ms'; {$ENDIF}
asm
	pushq	%rcx			// save temps
	pushq	%rax
	cmpq	$0x1000, %rax		// > 4k
	leaq	24(%rsp), %rcx		// point past return addr
	jb	.L2

.L1:	subq	$0x1000, %rcx  		// yes, move pointer down 4k
	orq	$0x0, (%rcx)   		// probe there
	subq	$0x1000, %rax  	 	// decrement count
	cmpq	$0x1000, %rax
	ja	.L1			// and do it again

.L2:	subq	%rax, %rcx
	orq	$0x0, (%rcx)		// less than 4k, just peek here

	popq	%rax
	popq	%rcx
	ret
end;
{$ENDIF}
{$ELSE}
procedure __chkstk_ms; cdecl; assembler; {$IFDEF FPC} alias: '___chkstk_ms'; {$ENDIF}
asm
	pushl	%ecx			//
	pushl	%eax
	cmpl	$0x1000, %eax		// > 4k
	leal	12(%esp), %ecx		//
	jb	.L2

.L1:	subl	$0x1000, %ecx  		// yes, move pointer down 4k
	orl	$0x0, (%ecx)   		// probe there
	subl	$0x1000, %eax  	 	// decrement count
	cmpl	$0x1000, %eax
	ja	.L1			// and do it again

.L2:	subl	%eax, %ecx
	orl	$0x0, (%ecx)		// less than 4k, just peek here

	popl	%eax
	popl	%ecx
	ret
end;
{$ENDIF}

{$ELSE}

{$IFDEF WIN32}
procedure __chkstk_ms; cdecl;
begin
  asm
    pop edx  { Return address }
    { Grow stack in 4k chunks so that automatic stack growth works }
@2: cmp eax, 4096
    jb @1
    sub esp, 4092
    push eax
    sub eax, 4096
    jmp @2
@1: push edx  { Apparently quicker than jmp edx on modern processors }
    ret
  end;
end;
{$ELSE}
procedure ___chkstk_ms; cdecl; assembler;
asm
    push	rcx
  	push	rax
	  cmp   rax, $1000
  	lea   rcx,24(rsp)
	  jb     @2

@1:	sub 	rcx, $1000
	  or	  [rcx],0
	  sub   rax, $1000
	  cmp	  rax, $1000
	  ja	  @1

@2:	sub	  rcx, rax
	  or	  [rcx],0

   	pop	  rax
  	pop	rcx
  	ret
end;
{$ENDIF}
{$ENDIF}

{$ENDIF}

function s_cat(lp: PAnsiChar; rpp: PPAnsiChar; rnp: PInteger; var np: Integer; ll: FtnLen):Integer; cdecl; {$IFDEF FPC} alias: '_s_cat'; {$ENDIF}
var
  n, i, nc: Integer;
  rp: PAnsiChar;
  { Copy of f2c routine assuming NO_OVERWRITE defined, so lhs must not contain rhs }
begin
  n := np;
  for i := 0 to n-1 do begin
    nc := ll;
    if rnp^ < nc then
      nc := rnp^;
    Inc(rnp);
    Dec(ll, nc);
    rp := rpp^;
    Inc(rpp);
    Move(rp^, lp^, nc);
  end;
  FillChar(lp^, ll, ' ');
  Result := 0;
end;

function s_cmp(a, b: PAnsiChar; la, lb: FtnLen): Integer; cdecl; {$IFDEF FPC} alias: '_s_cmp'; {$ENDIF}
var
  aend, bend: PAnsiChar;
begin
  aend := a + la;
  bend := b + lb;
  if la <= lb then begin
    while a < aend do
      if a^ <> b^ then begin
        Result := Ord(a^) - Ord(b^);
        Exit;
      end else begin
        Inc(a);
        Inc(b);
      end;
    while b < bend do
      if b^ <> ' ' then begin
        Result := Ord(' ') - Ord(b^);
        Exit;
      end else
        Inc(b);
  end else begin
    while b < bend do
      if a^ <> b^ then begin
        Result := Ord(a^) - Ord(b^);
        Exit;
      end else begin
        Inc(a);
        Inc(b);
      end;
    while a < aend do
      if a^ <> ' ' then begin
        Result := Ord(a^) - Ord(' ');
        Exit;
      end else
        Inc(a);
  end;
  Result := 0;
end;

procedure s_copy(a, b: PAnsiChar; la, lb: FtnLen); cdecl; {$IFDEF UND} alias: '_s_copy'; {$ENDIF}
{ Copy of f2c routine assuming NO_OVERWRITE defined, so lhs must not contain rhs }
begin
  if la <= lb then
    Move(b^, a^, la)
  else begin
    Move(b^, a^, lb);
    FillChar((a+lb)^, la-lb, ' ');
  end;
end;

function __powidf2(a:Double; p:Integer): Double; cdecl; {$IFDEF FPC} alias_'__powidf2'; {$ENDIF}
begin
  Result := IntPower(a,p);
end;

function _gfortran_pow_i4_i4(base:Integer; exponent:Integer): Integer; cdecl; {$IFDEF FPC} alias_'_gfortran_pow_i4_i4'; {$ENDIF}
begin
  Result := 1;
  if exponent = 0 then
    Exit
  else if exponent < 0 then
    if base = 1 then
      Exit
    else if base = -1 then begin
      if Odd(exponent) then
        Result := -1;
    end else if base = 0 then
      Result := 1 div base { Force divide by zero }
    else
      Result := 0
  else begin
    while exponent > 0 do begin
      while (exponent and 1) = 0 do begin
        exponent := exponent shr 1;
        base := base*base;
      end;
      Dec(exponent);
      Result := Result*base;
    end;
  end;
end;

function _gfortran_compare_string(la: FtnLen; a: PAnsiChar; lb: FtnLen; b: PAnsiChar): Integer; cdecl; {$IFDEF FPC} alias_'_gfortran_compare_string'; {$ENDIF}
begin
  Result := s_cmp(a, b, la, lb);
end;

procedure _gfortran_concat_string(destlen: FtnLen; dest: PAnsiChar; len1: FtnLen; s1: PAnsiChar; len2: FtnLen; s2: PAnsiChar); cdecl; {$IFDEF FPC} alias_'_gfortran_concat_string'; {$ENDIF}
begin
  if len1 <= destlen then
    Move(s1^, dest^, destlen)
  else begin
    Move(s1^, dest^, len1);
    Inc(dest,len1);
    Dec(destlen,len1);
    if len2 <= destlen then
      Move(s2, dest, destlen)
    else begin
      Move(s2^, dest^, len2);
      FillChar(dest^, destlen-len2, ' ');
    end;
  end;
end;

procedure _gfortran_runtime_error_at(Where: PAnsiChar; Message: PAnsiChar); cdecl; {$IFDEF FPC} alias_'_gfortran_runtime_error_at'; {$ENDIF}
begin
  { Message is actually printf format string, but just raise exacption for now }
  raise Exception.Create(Where + ' ' + Message);
end;

function pprint_(var tag, n: Integer; var d: Double):Integer; cdecl; {$IFDEF FPC} alias_'pprint_'; {$ENDIF}
var
  i: Integer;
  p: PDouble;
begin
  Write(Worker.Log,'PPRINT (', tag, ')', n);
  p := @d;
  for i := 0 to n - 1 do begin
    Write(Worker.Log, ' ', p^);
    Inc(p);
  end;
  WriteLn(Worker.Log);
end;

function errstop_(s: PAnsiChar; l: Integer):Integer; cdecl; {$IFDEF FPC} alias_'errstop_'; {$ENDIF}
var
  str: string;
begin
  SetString(str, s, l);
  raise Exception.Create(str);
end;

function errstopf_(fmt: PAnsiChar; args: Pointer { ... } ):Integer; cdecl; {$IFDEF FPC} alias_'errstopf_'; {$ENDIF}
var
  w: string;
  p: PAnsiChar;
begin
  w := '';
  p := fmt;
  while p^ <> #0 do begin
    case p^ of
     '\': begin Inc(p); w := w + p^; end;
     '#', #0: Break;
     '%': begin
            Inc(p);
            case p^ of
              'd': w := w + IntToStr(PInteger(args)^);
              'g': w := w + FloatToStr(PSingle(args)^);
              'l': begin Inc(p); w := w + FloatToStr(PDouble(args)^); end;
            end;
            Inc(PAnsiChar(args),SizeOf(PAnsiChar));
          end;
    else
      w := w + p^;
    end;
    Inc(p);
  end;
  raise Exception.Create(w);
  Floor(0.1);
end;

function clog(s: PAnsiChar):Integer; cdecl; {$IFDEF FPC} alias_'clog'; {$ENDIF}
begin
  WriteLn(Worker.Log,ThreadName,':',s);
end;

function disnan_(var a:Double):Integer; cdecl;  {$IFDEF FPC} alias_'disnan_'; {$ENDIF}
begin
  if IsNan(a) then
    Result := 1
  else
    Result := 0;
end;

end.
