library plapack;

{$mode objfpc}{$H+}

uses
  Math, MathSup,
  Lapack in '..\Lapack.pas'
  ;

type
  TxerblaHandler = function (SRNAME: PAnsiChar; var Info: Integer; len: FtnLen):Integer;

var
  xerblaHandler: TxerblaHandler = nil;

function SetXerblaHandler(AxerblaHandler: TxerblaHandler):TxerblaHandler;
begin
  Result := xerblaHandler;
  xerblaHandler := AxerblaHandler;
end;

procedure __chkstk_ms; cdecl; external name '___chkstk_ms';
{$LINK mschkstk.o}

function _sqrt(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'sqrt'; {$ENDIF}
begin
  Result := Sqrt(a);
end;

function _xerbla(SRNAME: PAnsiChar; var Info: Integer; len: FtnLen): Integer; cdecl; {$IFDEF FPC} alias: 'xerbla_'; {$ENDIF}
begin
  if Assigned(xerblaHandler) then
    Result := xerblaHandler(SRNAME, Info, len)
  else
    Result := 0;
end;

function _log(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'log'; {$ENDIF}
begin
  Result := Ln(a);
end;

function _logf(a: Single): Double; cdecl; {$IFDEF FPC} alias: 'logf'; {$ENDIF}
begin
  Result := Ln(a);
end;

function _log10(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'log10'; {$ENDIF}
begin
  Result := Log10(a);
end;

function _cos(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'cos'; {$ENDIF}
begin
  Result := Cos(a);
end;

function _pow(a, b: Double): Double; cdecl; {$IFDEF FPC} alias: 'pow'; {$ENDIF}
begin
  Result := Power(a,b);
end;

function _lroundf(a:Double): Integer; cdecl; {$IFDEF FPC} alias: 'lroundf'; {$ENDIF}
begin
  Result := Round(a);
end;

function ___powidf2(a:Double; p:Integer): Double; cdecl; {$IFDEF FPC} alias: '__powidf2'; {$ENDIF}
begin
  Result := Math.IntPower(a,p);
end;

function __gfortran_pow_i4_i4(base:Integer; exponent:Integer): Integer; cdecl; {$IFDEF FPC} alias: '_gfortran_pow_i4_i4'; {$ENDIF}
begin
  Result := 1;
  if exponent = 0 then
    Exit
  else if exponent < 0 then
    if base = 1 then
      Exit
    else if base = -1 then begin
      if Odd(exponent) then
        Result := -1;
    end else if base = 0 then
      Result := 1 div base { Force divide by zero }
    else
      Result := 0
  else begin
    while exponent > 0 do begin
      while (exponent and 1) = 0 do begin
        exponent := exponent shr 1;
        base := base*base;
      end;
      Dec(exponent);
      Result := Result*base;
    end;
  end;
end;

function s_cmp(a, b: PAnsiChar; la, lb: FtnLen): Integer; cdecl; {$IFDEF FPC} alias: '_s_cmp'; {$ENDIF}
var
  aend, bend: PAnsiChar;
begin
  aend := a + la;
  bend := b + lb;
  if la <= lb then begin
    while a < aend do
      if a^ <> b^ then begin
        Result := Ord(a^) - Ord(b^);
        Exit;
      end else begin
        Inc(a);
        Inc(b);
      end;
    while b < bend do
      if b^ <> ' ' then begin
        Result := Ord(' ') - Ord(b^);
        Exit;
      end else
        Inc(b);
  end else begin
    while b < bend do
      if a^ <> b^ then begin
        Result := Ord(a^) - Ord(b^);
        Exit;
      end else begin
        Inc(a);
        Inc(b);
      end;
    while a < aend do
      if a^ <> ' ' then begin
        Result := Ord(a^) - Ord(' ');
        Exit;
      end else
        Inc(a);
  end;
  Result := 0;
end;

function __gfortran_compare_string(la: FtnLen; a: PAnsiChar; lb: FtnLen; b: PAnsiChar): Integer; cdecl; {$IFDEF FPC} alias: '_gfortran_compare_string'; {$ENDIF}
begin
  Result := s_cmp(a, b, la, lb);
end;

procedure _gfortran_concat_string(destlen: FtnLen; dest: PAnsiChar; len1: FtnLen; s1: PAnsiChar; len2: FtnLen; s2: PAnsiChar); cdecl; {$IFDEF FPC} alias: '_gfortran_concat_string'; {$ENDIF}
begin
  if len1 <= destlen then
    Move(s1^, dest^, destlen)
  else begin
    Move(s1^, dest^, len1);
    Inc(dest,len1);
    Dec(destlen,len1);
    if len2 <= destlen then
      Move(s2, dest, destlen)
    else begin
      Move(s2^, dest^, len2);
      FillChar(dest^, destlen-len2, ' ');
    end;
  end;
end;

function disnan_(var a:Double):Integer; cdecl;  {$IFDEF FPC} alias: 'disnan_'; {$ENDIF}
begin
  if IsNan(a) then
    Result := 1
  else
    Result := 0;
end;

exports
 SetXerblaHandler, ieeeck_,

 dasum_, daxpy_, dcopy_, lsame_, dgemm_, dgemv_, dger_, dnrm2_, dscal_, dswap_, dsymv_, dsyr2_, dsyr2k_, dtrmm_, dtrsm_, dtrmv_, idamax_, ddot_, drot_,
 ztrmv_, zaxpy_, zcopy_, zdotc_, zdscal_, zgemm_, zgemv_, zgerc_, zhemv_, zher2_, zher2k_, zscal_, zswap_, ztrmm_, dznrm2_,

 dsytrd_, dlaebz_, dlagtf_, dlagts_, dlamch_, dlanst_, dlansy_, dlapy2_, dlar1v_, dlarf_, dlarfb_, dlarfg_, dlarft_, dlarnv_, dlarrb_, dlarre_, dlarrf_, dlarrv_, dlaruv_, dlascl_, dlaset_, dlasq2_, dlasq5_, dlasq6_, dlasrt_, dlassq_, dlatrd_, dorm2l_, dorm2r_, dormtr_, dstebz_, dstein_, dsterf_, dsyevr_, dsytd2_, dlae2_, ilaenv_, dlartg_,
 dgetrf_, dgetf2_, dlaswp_, dtrtri_, dtrti2_, dgetri_, dgetrs_, dpotrf_,
 zheevr_, zunmtr_, dladiv_, dlapy3_, ilazlc_, zhetd2_, zhetrd_, zlacgv_, zladiv_, zlansy_, zlar1v_, zlarf_, zlarfb_, zlarfg_, zlarrv_, zlaset_, zlassq_, zlatrd_, zstein_, zstemr_, zunm2l_, zunm2r_, zlartg_,
 dgesdd_, dormql_, dormqr_, zunmql_, zunmqr_, dtrtrs_, dgels_, dgelsd_;

end.

