unit nj;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses
  Classes, SysUtils;

function sixj(j1, j2, j3, j4, j5, j6: Integer):Double;
function j6jRange(j2, j3, l1, l2, l3: Integer; out j1min, j1max: Integer):Boolean;
procedure j6j(j2, j3, l1, l2, l3: Integer; var d6j: array of Double);
function threej(j1d, m1d, j2d, m2d, j3d, m3d: Integer): Double;
function cg(j1d, m1d, j2d, m2d, j3d, m3d: Integer): Double;
function ninej(j11, j12, j13,
               j21, j22, j23,
               j31, j32, j33: Integer):Double;

implementation

uses Math, MathSup, QuantumNumber, Worker;

threadvar
  sj1, sj2, sj3: array of Double;

function ninej(j11, j12, j13,
               j21, j22, j23,
               j31, j32, j33: Integer):Double;
var
  l1, l1m, l2, l2m, l3, l3m:Integer; { Lowest and highest value of Kappa for each case }
  Kappa, KappaMin, KappaMax: Integer;
  Range: Integer; { Largest work size needed }
begin
  Result := 0;
  if not ( j6jRange(j33, j11, j31, j21, j32, l1, l1m)
           and j6jRange(j23, j12, j22, j32, j21, l2, l2m)
           and j6jRange(j11, j33, j13, j23, j12, l3, l3m)) then
    Exit;
  KappaMin := Max(l1, Max(l2, l3));
  KappaMax := Min(l1m, Min(l2m, l3m));
  if KappaMin > KappaMax then
    Exit;
  { Check for zero by symmetry }
  if Minus(j11+j12+J13+J21+j22+j23+j31+j32+j33) < 0 then begin
    if (J11=j21) and (j12=j22) and (j13=j23) then
      Exit;
    if (J11=j31) and (j12=j32) and (j13=j33) then
      Exit;
    if (J21=j31) and (j22=j32) and (j23=j33) then
      Exit;
    if (J11=j12) and (j21=j22) and (j31=j32) then
      Exit;
    if (J11=j13) and (j21=j23) and (j31=j33) then
      Exit;
    if (J12=j13) and (j22=j23) and (j32=j33) then
      Exit;
  end;
  { allocate workspace }
  Range := Max(l1m-l1, Max(l2m-l2, l3m-l3)) div 2 + 1;
  if Range > Length(sj1) then begin
    Range := Max(Range*2, 20);
    SetLength(sj1, Range);
    SetLength(sj2, Range);
    SetLength(sj3, Range);
  end;
  j6j(j33, j11, j31, j21, j32, sj1);
  j6j(j23, j12, j22, j32, j21, sj2);
  j6j(j11, j33, j13, j23, j12, sj3);
  Kappa := KappaMin;
  while Kappa <= KappaMax do begin
    Result := Result + (Kappa+1) * sj1[(Kappa-l1) div 2] * sj2[(Kappa-l2) div 2] * sj3[(Kappa-l3) div 2];
    Inc(Kappa, 2);
  end;
  if Odd(KappaMin) then
    Result := -Result;
end;

function threej(j1d, m1d, j2d, m2d, j3d, m3d: Integer): Double;
{ Three J symbol optimised for the smallest J <= 1
  Arguments are doubled integers - ok for J up to 180 with 16 bit integers }
var
  phase, i, m32, jmin, jmax, j1, j2, j3, m1, m2, m3, jsum, l, iu: Integer;
  j23, j23p1, m1sq, ibc, jjp3: Longint; { Selective use of long ints to allow larger J }
  UP, J1LOWER: Boolean;
  fl, fu, flm1, fup1, flm2, fup2, anl, anu, alastl, alastu, x, ratio: Double;
  { propagation functions }
  function A(JA:Double):Double;
  begin
    Result := Sqrt((JA-J23)*(J23P1-JA)*(JA-M1SQ));
  end;
  function B(J:Double):Double;
  begin
    Result := -(J+1)*(IBC-J*(J+2)*M32);
  end;
begin
  { Check for 0 result }
  Result := 0;
  IF (m1d + m2d + m3d <> 0) then Exit;
  IF (ABS(m1d) > j1d) then Exit;
  IF (ABS(m2d) > j2d) then Exit;
  IF (ABS(m3d) > j3d) then Exit;
  IF (j1d < ABS(j2d-j3d)) or (j1d > j2d+j3d) then Exit;
  j3 := Min(j1d, j2d);
  j3 := Min(j3, j3d);
  { Permute to give smallest J as J3 }
  if j3 = j3d then begin
   j2 := j2d;
   j1 := j1d;
   m1 := m1d;
   m2 := m2d;
   m3 := m3d;
  end ELSE IF j2d = j3 THEN begin
   j2 := j1d;
   j1 := j3d;
   m3 := m2d;
   m2 := m1d;
   m1 := m3d;
  end ELSE { j1d = j3 } begin
   j2 := j3d;
   j1 := j2d;
   m3 := m1d;
   m2 := m3d;
   m1 := m2d;
  end;
  { if smallest J > 1 call standard three j routine }
  if j3 <= 2 then begin
    { ensure J1 >= J2 - permute if necessary }
    phase := 0;
    jsum := j1d + j2d + j3d;
    if j1 < j2 then begin
      i := j1;
      j1 := j2;
      j2 := i;
      i := m1;
      m1 := m2;
      m2 := i;
      phase := phase + jsum;
    end;
    { likewise ensure M3 >= 0 }
    if m3 < 0 then begin
     m3 := -m3;
     m2 := -m2;
     m1 := -m1;
     phase := phase + jsum;
    end;
    { Switch on smallest J }
    case j3 of
     { ( J  J 0 )
       ( M -M 0 ) }
      0: begin
           phase := phase + j1 - m1;
           Result := 1 / sqrt( j1 + One );
         end;
     { ( J+1/2      J 1/2 )
       ( M     -M-1/2 1/2 ) }
      1: begin
           phase := phase + j2 + m2;
           Result := sqrt( (j1 - m1) / ( (j2 + 2) * (j2 + 1) * Two ) );
         end;
      { j MIN = 1 }
      2: if m3 = 0 then begin
           if j1 = j2 then begin
            { ( J  J 1 )
              ( M -M 0 ) }
             phase := phase + j1 + m2;
             Result := m1 * sqrt( 1 / ( (j1 + 1) * j1 * (j1 + Two) ) );
           end else begin
             { ( J+1  J 1 )
               (   M -M 0 ) }
             phase := phase + j2 - m1 - 2;
             Result := sqrt( (j1 + m1) * (j1 + m2)
                             /( 2 * (j2 + 3) * (j2 + 2) * (j2 + One) ) );
           end
         end else if j1 = j2 then begin
           { ( J    J 1 )
             ( M -M-1 1 ) }
           phase := phase + j1 - m1;
           Result := sqrt( (j1 - m1) * (j1 - m2) / ( 2 * (j1 + 1) * j1 * (j1 + Two) ) );
         end else begin
           { ( J+1    J 1 )
             (   M -M-1 1 ) }
           phase := phase + j2 + m2;
           Result := sqrt( (j2 - m1) * (j1 - m1) / ( 4 * (j2 + 3) * (j2 + 2) * (j2 + One)) );
         end;
    end;
    if ABS(phase) mod 4 <> 0 then
      Result := -Result;
  end else begin
{ Recursive evaluation of 3j coefficients

         ( j1  j2  j3 )
         ( m1  m2  m3 )

 from J.Math.Phys. 16 1961 1975   Klaus Schulten and Roy Gorden.
}
    jmin:=MAX(ABS(m1),ABS(j2-j3));
    jmax:=j2+j3;
    j23:=j2-j3;
    j23:=j23*j23;
    j23p1:=j2+j3+2;
    j23p1:=j23p1*j23p1;
    m1sq:=m1;
    m1sq:=m1*m1sq;
    ibc := j2 + 2;
    ibc := ibc * j2;
    jjp3 := j3;
    jjp3 := jjp3 * (j3+2);
    ibc := (ibc - jjp3)*m1;
    m32:=m3-m2;
    l:=jmin;
    iu:=jmax;
    j1lower:=True;
    up:=True;
    fl:=1.0;
    fu:=1.0;
    if ABS(j2-j3-m1) mod 4 <> 0 then
      fu:=-1.0;
    flm1:=0.0;
    fup1:=0.0;
    flm2:=0.0;
    fup2:=0.0;
    anl:=0.0;
    anu:=0.0;
    alastl:=0.0;
    alastu:=0.0;
    while l-iu < 4 do begin
      if up then begin
        if l < jmax then begin
          IF l<>0 then begin
            x:=-(2*b(l)*fl+(l+2)*alastl*flm1);
            alastl:=a((l+2)*(l+2));
            x:=x/(l*alastl);
          end else begin
            alastl:=a(4);
            x:=-4*m32/alastl;
          end;
        end else
          x:=0;
        if l<=iu then
          anl:=anl+(l+1)*fl*fl;
        flm2:=flm1;
        flm1:=fl;
        fl:=x;
        IF (l=j1) and j1lower then
          Result := flm1;
        l:=l+2;
      end else begin
        if iu > jmin then begin
          x:=-(iu*alastu*fup1+2*b(iu)*fu);
          alastu:=a(iu*iu);
          x:=x/((iu+2)*alastu);
        end else
          x:=0;
        if iu >= l then
          anu:=anu+(iu+1)*fu*fu;
        fup2:=fup1;
        fup1:=fu;
        fu:=x;
        if iu=j1 then begin
          j1lower:=False;
          Result:=fup1;
        end;
        iu:=iu-2
      end;
      up:= not up;
    end;
    ratio:=(fl*fup2+flm1*fup1+flm2*fu)/(fl*fl+flm1*flm1+flm2*flm2);
    anl:=anl*ratio*ratio;
    if j1lower then
       Result:=Result*ratio;
    Result:=Result/sqrt(anl+anu);
  end;
end;

threadvar
  Work: array of Double;

//var
//  sixjtime: string = 'sixj';

function sixj(j1, j2, j3, j4, j5, j6: Integer):Double;
{ Six-J symbol:
   ( j1 j2 j3 )
   ( j4 j5 j6 )
 The input quantum numbers **** DOUBLED **** }

 function Callj6j(j1, j2, j3, j4, j5, j6: Integer): Double;
  var
    j1min, j1max: Integer;
  begin
    Result := 0;
    if not j6jRange(j2, j3, j4, j5, j6, j1min, j1max) then
      Exit;
    if Odd(j1+j1min) then
      raise Exception.Create('Bad Integer/half Integer mix in sixj');
    if (j1 >= j1min) and (j1 <= j1max) then begin
      case j6 of { Special cases automatically generated, with some manual tweking }
                 { These have a relatively small effect on the matrix element time - maybe 20% improvement }
        0: Result := 1/Sqrt((j1+1)*(j2+One));
               { j4-j2 j5-j1}
        1:
          if j5 < j1 then
               {  0.5 -0.5}
               Result := sqrt((-j1+j2+j3+2)*(j1-j2+j3))/(2*sqrt(j1*(j1+1)*(j2+One)*(j2+2)))
          else
               {  0.5  0.5}
               Result := ((-j1-j2-j3-4)*sqrt(2*(j1+j2-j3+2)))/(2*sqrt(2*(j1+1)*(j1+2)*(j1+j2+j3+Four)*(j2+1)*(j2+2)));
         2:
          case j5-j1+(j4-j2)*4 of
            0: {    0    0}
               Result := ((-j1-j2+j3)*(j1+j2+j3+4)+(-j1+j2+j3)*(j1-j2+j3))/(4*sqrt(j1*(j1+1)*(j1+Two)*j2*(j2+1)*(j2+2)));

            6: {    1   -1}
               Result := sqrt((-j1+j2+j3+2)*(j1-j2+j3)*(-j1+j2+j3+Four)*(j1-j2+j3-2)/(16*j1*(j1-1)*(j1+One)*(j2+1)*(j2+2)*(j2+3)));

            8: {    1    0}
               Result := (-j1-j2-j3-4)*sqrt(2*(-j1+j2+j3+2)*(j1-j2+j3)*(j1+j2-j3+Two)/(16*j1*(j1+1)*(j1+Two)*(j1+j2+j3+4)*(j2+1)*(j2+2)*(j2+3)));

            10: {    1    1}
               Result := (j1+j2+j3+4)*(j1+j2+j3+6)*sqrt((j1+j2-j3+2)*(j1+j2-j3+Four)/(16*(j1+1)*(j1+Two)*(j1+3)*(j1+j2+j3+4)*(j1+j2+j3+6)*(j2+1)*(j2+2)*(j2+3)));

          end;
         3:
          case j5-j1+(j4-j2)*4 of
            3: {  0.5 -0.5}
               Result := ((-j1+j2+j3)*(j1-j2+j3-2)-2*(j1+j2-j3)*(j1+j2+j3+4))*sqrt((-j1+j2+j3+Two)*(j1-j2+j3)/(64*j1*(j1-1)*(j1+One)*(j1+2)*j2*(j2+1)*(j2+2)*(j2+3)));

            5: {  0.5  0.5}
               Result := (-2*(-j1-j2+j3)*(j1+j2+j3+6)-4*(-j1+j2+j3)*(j1-j2+j3))*(j1+j2+j3+Four)*sqrt(2*(j1+j2-j3+Two)/(512*j1*(j1+One)*(j1+2)*(j1+3)*(j1+j2+j3+4)*j2*(j2+1)*(j2+2)*(j2+3)));

            9: {  1.5 -1.5}
               Result := sqrt((-j1+j2+j3+2)*(-j1+j2+j3+4)*(-j1+j2+j3+Six)*(j1-j2+j3)*(j1-j2+j3-2)*(j1-j2+j3-4)/(64*j1*(j1-1)*(j1-Two)*(j1+1)*(j2+1)*(j2+2)*(j2+3)*(j2+4)));

            11: {  1.5 -0.5}
               Result := (-j1-j2-j3-4)*sqrt(6*(-j1+j2+j3+2)*(-j1+j2+j3+Four)*(j1-j2+j3)*(j1-j2+j3-2)*(j1+j2-j3+2)/(128*j1*(j1-1)*(j1+One)*(j1+2)*(j1+j2+j3+4)*(j2+1)*(j2+2)*(j2+3)*(j2+4)));

            13: {  1.5  0.5}
               Result := (j1+j2+j3+4)*(j1+j2+j3+6)*sqrt(3*(-j1+j2+j3+2)*(j1-j2+j3)*(j1+j2-j3+Two)*(j1+j2-j3+4)/(64*j1*(j1+1)*(j1+Two)*(j1+3)*(j1+j2+j3+4)*(j1+j2+j3+6)*(j2+1)*(j2+2)*(j2+3)*(j2+4)));

            15: {  1.5  1.5}
               Result := (-j1-j2-j3-4)*(j1+j2+j3+8)*(j1+j2+j3+Six)*sqrt(2*(j1+j2-j3+2)*(j1+j2-j3+Four)*(j1+j2-j3+6)/(128*(j1+1)*(j1+Two)*(j1+3)*(j1+4)*(j1+j2+j3+4)*(j1+j2+j3+6)*(j1+j2+j3+8)*(j2+1)*(j2+2)*(j2+3)*(j2+4)));

          end;
         4:
          case j5-j1+(j4-j2)*4 of
             0: {    0    0}
               begin
                 Result := (j1+j2+j3+4) div 2;
                 Result :=  (((-j1+j2+j3-2)*(j1-j2+j3-2) - 8*(j1+j2-j3)*Result)*(-j1+j2+j3)*(j1-j2+j3)
                             +4*(j1+j2-j3)*(j1+j2-j3-2)*Result*(Result+1))
                            / (16*sqrt(j1*(j1-1)*(j1+One)*(j1+2)*(j1+3)*j2*(j2-1)*(j2+1)*(j2+2)*(j2+3)));
               end;

             6: {    1   -1}
               Result := ((-j1+j2+j3)*(j1-j2+j3-4)-3*(j1+j2-j3)*(j1+j2+j3+4))*sqrt((-j1+j2+j3+2)*(j1-j2+j3)*(-j1+j2+j3+Four)*(j1-j2+j3-2)/(256*j1*(j1-1)*(j1-Two)*(j1+1)*(j1+2)*j2*(j2+1)*(j2+2)*(j2+3)*(j2+4)));

             8: {    1    0}
               Result := ((j1-j2-j3)*(j1-j2+j3-2)+(j1+j2-j3)*(j1+j2+j3+6))*(j1+j2+j3+4)*sqrt(6*(-j1+j2+j3+2)*(j1-j2+j3)*(j1+j2-j3+2)/(256*j1*(j1-1)*(j1+One)*(j1+2)*(j1+3)*(j1+j2+j3+4)*j2*(j2+1)*(j2+2)*(j2+3)*(j2+4)));

            10: {    1    1}
               Result := ((-j1-j2+j3)*(j1+j2+j3+8)+3*(-j1+j2+j3)*(j1-j2+j3))*(j1+j2+j3+Four)*(j1+j2+j3+6)*sqrt((j1+j2-j3+2)*(j1+j2-j3+4)/(256*j1*(j1+1)*(j1+Two)*(j1+3)*(j1+4)*(j1+j2+j3+4)*(j1+j2+j3+6)*j2*(j2+1)*(j2+2)*(j2+3)*(j2+4)));

            12: {    2   -2}
               Result := sqrt((-j1+j2+j3+2)*(-j1+j2+j3+4)*(-j1+j2+j3+Six)*(-j1+j2+j3+8)*(j1-j2+j3)*(j1-j2+j3-2)*(j1-j2+j3-4)*(j1-j2+j3-6)/(256*j1*(j1-1)*(j1-Two)*(j1-3)*(j1+1)*(j2+1)*(j2+2)*(j2+3)*(j2+4)*(j2+5)));

            14: {    2   -1}
               Result := (-j1-j2-j3-4)*sqrt(2*(-j1+j2+j3+2)*(-j1+j2+j3+4)*(-j1+j2+j3+Six)*(j1-j2+j3)*(j1-j2+j3-2)*(j1-j2+j3-4)*(j1+j2-j3+2)/(128*j1*(j1-1)*(j1-Two)*(j1+1)*(j1+2)*(j1+j2+j3+4)*(j2+1)*(j2+2)*(j2+3)*(j2+4)*(j2+5)));

            16: {    2    0}
               Result := (j1+j2+j3+4)*(j1+j2+j3+6)*sqrt(3*(-j1+j2+j3+2)*(-j1+j2+j3+Four)*(j1-j2+j3)*(j1-j2+j3-2)*(j1+j2-j3+2)*(j1+j2-j3+4)/(128*j1*(j1-1)*(j1+One)*(j1+2)*(j1+3)*(j1+j2+j3+4)*(j1+j2+j3+6)*(j2+1)*(j2+2)*(j2+3)*(j2+4)*(j2+5)));

            18: {    2    1}
               Result := (-j1-j2-j3-4)*(j1+j2+j3+8)*(j1+j2+j3+Six)*sqrt(2*(-j1+j2+j3+2)*(j1-j2+j3)*(j1+j2-j3+Two)*(j1+j2-j3+4)*(j1+j2-j3+6)/(128*j1*(j1+1)*(j1+Two)*(j1+3)*(j1+4)*(j1+j2+j3+4)*(j1+j2+j3+6)*(j1+j2+j3+8)*(j2+1)*(j2+2)*(j2+3)*(j2+4)*(j2+5)));

            20: {    2    2}
               Result := (j1+j2+j3+10)*(j1+j2+j3+4)*(j1+j2+j3+Six)*(j1+j2+j3+8)*sqrt((j1+j2-j3+2)*(j1+j2-j3+4)*(j1+j2-j3+Six)*(j1+j2-j3+8)/(256*(j1+1)*(j1+2)*(j1+Three)*(j1+4)*(j1+5)*(j1+j2+j3+10)*(j1+j2+j3+4)*(j1+j2+j3+6)*(j1+j2+j3+8)*(j2+1)*(j2+2)*(j2+3)*(j2+4)*(j2+5)));

          end;

      else
        if Length(work) < (j1max-j1min+2) div 2 then
          SetLength(work, Max(j1max-j1min+2,20));
        j6j(j2, j3, j4, j5, j6, work);
        Result := work[(j1-j1min) div 2];
        Exit;
      end;
      Result := Result * Minus(j1+j2+j3);
    end;
  end;
  procedure Swap(var a, b: Integer); inline;
  var
    t: Integer;
  begin
    t := a;
    a := b;
    b := t;
  end;
begin
//  TStart(sixjtime);
 { To minimise computation, the smallest quantum number (which determines
   the range of the recursion relation) is in the last two columns. }
  if Min(j2,j5) > Min(j1,j4) then begin
    Swap(j1,j2);
    Swap(j4,j5);
  end;
  if Min(j2,j5) < Min(j3,j6) then begin
    Swap(j3,j2);
    Swap(j6,j5);
  end;
  { For the special case computations, ensure j6 is the smallest }
  if j6 > j3 then begin
    Swap(j3,j6);
    Swap(j2,j5);
  end;
  { ... and for j4=j2+d, j5=j1+d' permute so that
    d >= 0, and |d'| <= d to minimize the number of
    special cases required }
  if j6 <= 4 then { This permute only required for special cases }
    if Abs(j5-j1) > Abs(j4-j2) then begin
      if j5 > j1 then begin
        Swap(j1,j2);
        Swap(j4,j5);
      end else begin
        Swap(j1,j4);
        Swap(j2,j5);
      end;
    end else if j4 < j2 then begin
      Swap(j2,j4);
      Swap(j1,j5);
    end;
  Result := Callj6j(j1, j2, j3, j4, j5, j6);
  //{ mi := MIN(j1, j2, j3, j4, j5, j6) }
  //mi := Min(j1,j2);
  //if j3 < mi then mi := j3;
  //if j4 < mi then mi := j4;
  //if j5 < mi then mi := j5;
  //if j6 < mi then mi := j6;
  //if j1 = mi then
  //  Result := Callj6j(j2, j6, j4, j5, j3, j1)
  //else if j2 = mi then
  //  Result := Callj6j(j1, j6, j5, j4, j3, j2)
  //else if j3 = mi then
  //  Result := Callj6j(j1, j5, j6, j4, j2, j3)
  //else if j4 = mi then
  //  Result := Callj6j(j3, j2, j1, j6, j5, j4)
  //else if j5 = mi then
  //  Result := Callj6j(j1, j3, j2, j4, j6, j5)
  //else
  //  Result := Callj6j(j1, j2, j3, j4, j5, j6);
//  TEnd(sixjtime);
end;

function j6jRange(j2, j3, l1, l2, l3: Integer; out j1min, j1max: Integer):Boolean;
{ Calculate j1 range that j6j will calculate }
begin
  if Odd(l1+j2+l3) or Odd(l1+l2+j3) then
    raise Exception.Create('Bad Integer/Half integer mix in j6jrange');
  j1min := MAX(ABS(j2-j3),ABS(l2-l3));
  j1max := MIN(j2+j3,l2+l3);
  { also test for other triangular inequalites }
  Result := (j1max >= j1min)
            and not ( (j2 > l1+l3) or (j2 < ABS(l1-l3)) or  (j3 > l1+l2)  or  (j3 < ABS(l1-l2)) ) ;
end;

procedure j6j(j2, j3, l1, l2, l3: Integer; var d6j: array of Double);
{ Calculate all possible six-j symbols:
   ( ?  j2 j3 )
   ( l1 l2 l3 )

 Input quantum numbers ***** DOUBLED ***** }
var
  ival, j1min, j1max: Integer;
  l, li, u, ui, i: Integer;
  j1, j1s, jjp1, jjp2, jjp3, llp1, llp2, llp3: Integer;
  j2mj3sq, j2pj3p1sq, l2ml3sq, l2pl3p1sq, l7: Integer;
  aconst, f1, norm, scale, sum: Double;
  el, eu, e1: Double; {Previous e() }
  l6, u6: Double; { Previous 6-j }
  going_up, switching: Boolean;
{ This constant controls where the crossover happens
  If the routine stops 'Bad j6j match' this constant can be decreased,
  though it will cause some loss in accuracy }
const
  thresh: Double = 0.25e-2;
  
  function e(j1s:Double):Double;
  begin
    Result := Sqrt((j1s-j2mj3sq) * (j2pj3p1sq-j1s)
                   * (j1s-l2ml3sq) * (l2pl3p1sq-j1s));
  end;
  function f(j1: Integer; jjp1:Double):Double;
  begin
    Result := (j1+1)*(jjp1*(l7-jjp1)+aconst);
  end;
  procedure j6jerror(t: Double);
  begin
    raise Exception.CreateFmt('Bad j6j match (%g): %d %d %d %d %d %d', [t, j2, j3, l1, l2, l3]);
  end;

{ This routine calculates the 6-j coefficients for all permissible
  values of j1 for fixed values of j2, j3, l1, l2, and l3 using the
  recursive algorithm of K. Schulten and R. G. Gordon, J. math. phys.
  vol. 16, p. 1961, (1975).
  Programmed by d. e. fitz, 10/22/79 }
begin
  if not j6jRange(j2, j3, l1, l2, l3, j1min, j1max) then Exit;
  ival := (j1max-j1min+2) div 2 - 1;
  
  { Initialise upwards recursion }
  l := j1min;
  l6 := 1;
  d6j[0] := 1;
  el := 0;
  li := 1;

  { Initialize downwards }
  u := j1max;
  u6 := 1;
  if ((j2+j3+l2+l3) mod 4) <> 0 then
    u6 := -1;
  d6j[ival] := u6;
  eu := 0;
  ui := ival-1;

  if ival > 0 then begin { 20 }
    jjp2 := j2+2;
    jjp2 := jjp2*j2;
    jjp3 := j3+2;
    jjp3 := jjp3*j3;
    llp1 := l1+2;
    llp1 := llp1*l1;
    llp2 := l2+2;
    llp2 := llp2*l2;
    llp3 := l3+2;
    llp3 := llp3*l3;
    j2mj3sq := j2-j3;
    j2mj3sq := j2mj3sq*j2mj3sq;
    j2pj3p1sq := j2+j3+2;
    j2pj3p1sq := j2pj3p1sq * j2pj3p1sq;
    l2ml3sq := l2-l3;
    l2ml3sq := l2ml3sq * l2ml3sq;
    l2pl3p1sq := l2+l3+2;
    l2pl3p1sq := l2pl3p1sq * l2pl3p1sq;
    aconst := llp2-llp3;
    aconst := aconst * (jjp2-jjp3);
    l7 := jjp2+jjp3+llp2+llp3-2*llp1;

    { Second point upwards }
    jjp1 := l;
    jjp1 := jjp1 * (l+2);
    f1 := f(l, jjp1);
    l := l + 2;
    j1s := l;
    j1s := j1s * l;
    el := e(j1s);
    if l > 2 then
      l6 := -2*f1/(el*j1min)
    else
      l6 :=-(llp2+jjp2-llp1)/Sqrt(Four*jjp2*llp2);
    d6j[1] := l6;
    li := 2;
    if ival > 1 then begin
      { Second downward point }
      jjp1 := u + 2;
      jjp1 := jjp1 * u;
      f1 := f(u, jjp1);
      j1s := u;
      j1s := j1s * u;
      eu := e(j1s);
      u6 := -2*f1*u6 / (eu*(u+2));
      u := u-2;
      d6j[ival-1] := u6;
      ui := ival-2;
      going_up := True;
      switching := True;
      { Now complete recursion }
      while True do begin
        if l = u  then begin
        { At a suitable match point? }
          if ABS(l6) > thresh then Break;
          { NO - step back }
          l := l - 2;
          li := li - 1;
          l6 := d6j[li-1];
          d6j[ui+1] := u6;
          switching := False;
          going_up := False;
        end;
        if going_up then begin
          { Upward recursion }
          jjp1 := l + 2;
          jjp1 := jjp1 * l;
          f1 := f(l, jjp1);
          l := l+2;
          e1 := el;
          j1s := l;
          j1s := j1s * l;
          el := e(j1s);
          l6 := -(2*f1*l6+l*e1*d6j[li-2])/(el*(l-2));
          d6j[li] := l6;
          li := li + 1;
          if ABS(l6) < thresh then begin
            if ABS(d6j[li-2]) < thresh then begin
              if not switching then
                j6jerror(d6j[li-2]);
              switching := False;
              going_up := False;
              li := li - 2;
              l6 := d6j[li-1];
              l := l - 4;
              d6j[ui+1] := u6;
            end;
          end;
        end else begin
         { downward recursion. }
          e1 := eu;
          j1s := u;
          j1s := j1s * u;
          jjp1 := u + 2;
          jjp1 := jjp1 * u;
          eu := e(j1s);
          f1 := f(u, jjp1);
          u6 := -(u*e1*d6j[ui+2]+2*f1*u6) / (eu*(u+2));
          u := u-2;
          d6j[ui] := u6;
          ui := ui - 1;
          if ABS(u6) < thresh then begin
            if ABS(d6j[ui+2]) < thresh then begin
              if not switching then
                j6jerror(d6j[ui+2]);
              switching := False;
              going_up := True;
              ui := ui + 2;
              u6 := d6j[ui+1];
              u := u + 4;
              d6j[li-1] := l6;
            end;
          end;
        end;
        going_up := going_up xor switching;
      end;
    end;
  end;
  { match upward and downward recursive results by scaling. }
  scale := u6/l6;
  { Ensure matchpoint set by downward recursion }
  d6j[ui+1] := u6;
  { Normalisation - lower half }
  sum := 0;
  j1 := j1min;
  for i := 0 to li-2 do begin
    sum := sum + (j1+1)*Sqr(d6j[i]);
    j1 := j1 + 2;
  end;
  sum := sum * Sqr(scale);
  { and upper }
  for i := ui+1 to ival do begin
    sum := sum + (j1+1)*Sqr(d6j[i]);
    j1 := j1 + 2;
  end;
  norm := 1/Sqrt(sum*(l1+1));
  { Now scale both halves }
  for i := ui+1 to ival do
    d6j[i] := d6j[i] * norm;
  norm := norm * scale;
  for i := 0 to li-2 do
    d6j[i] := d6j[i] * norm;
end;

function cg(j1d, m1d, j2d, m2d, j3d, m3d: Integer): Double;
begin
  Result := threej(j1d, m1d, j2d, m2d, j3d, -m3d)*Sqrt(j3d+One);
  if Odd((j1d - j2d + m3d) div 2) then
    Result := - Result;
end;

procedure Cleansjs;
begin
  sj1 := nil;
  sj2 := nil;
  sj3 := nil;
  Work := nil;
end;

initialization
  AtThreadEnd({$IFDEF FPC}@{$ENDIF}Cleansjs);
finalization
  Cleansjs;
end.

