unit ComboFix;
{ Provide Undo/Redo for ComboBox }

{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFNDEF CLX_} Messages, StdCtrls {$ELSE} QStdCtrls {$ENDIF};

procedure ComboBoxUndo(ComboBox: TCustomComboBox);
procedure ComboBoxRedo(ComboBox: TCustomComboBox);
function ComboBoxCanUndo(ComboBox: TCustomComboBox):Boolean;
function ComboBoxCanRedo(ComboBox: TCustomComboBox):Boolean;

implementation

{$IFDEF VCL_}
type
  TOpenCustomComboBox = class(TCustomComboBox)
  public
    property EditHandle {$IFDEF VER80 } : HWND  read FEditHandle {$ENDIF};
  end;

procedure ComboBoxUndo(ComboBox: TCustomComboBox);
begin
  SendMessage(TOpenCustomComboBox(ComboBox).EditHandle, EM_UNDO, 0, 0);
end;

procedure ComboBoxRedo(ComboBox: TCustomComboBox);
begin
  SendMessage(TOpenCustomComboBox(ComboBox).EditHandle, EM_UNDO, 0, 0); { Note UNDO/REDO are the same }
end;

function ComboBoxCanUndo(ComboBox: TCustomComboBox):Boolean;
begin
  Result := SendMessage(TOpenCustomComboBox(ComboBox).EditHandle, EM_CANUNDO, 0, 0) <> 0;
end;

function ComboBoxCanRedo(ComboBox: TCustomComboBox):Boolean;
begin
  Result := SendMessage(TOpenCustomComboBox(ComboBox).EditHandle, EM_CANUNDO, 0, 0) <> 0;
end;

{$ENDIF}

{$IFDEF CLX_}

uses Qt;

procedure FakeCtrlLetter(KeyCode: Word; ComboBox: TCustomComboBox);
var
  E: QKeyEventH;
  KeyText: WideString;
begin
  KeyText := '';
  E := QKeyEvent_create(QEventType_KeyPress, KeyCode, KeyCode,
          Integer(ButtonState_ControlButton), @KeyText, False, 1);
  try
    QOpenWidget_keyPressEvent(QOpenWidgetH(QComboBox_lineEdit(ComboBox.Handle)), E)
  finally
    QKeyEvent_destroy(E);
  end;
end;

procedure ComboBoxUndo(ComboBox: TCustomComboBox);
begin
  FakeCtrlLetter(Key_Z, ComboBox);
end;

procedure ComboBoxRedo(ComboBox: TCustomComboBox);
begin
  FakeCtrlLetter(Key_Y, ComboBox);
end;

function ComboBoxCanUndo(ComboBox: TCustomComboBox):Boolean;
begin
  Result := True; { can't find out }
end;

function ComboBoxCanRedo(ComboBox: TCustomComboBox):Boolean;
begin
  Result := True; { can't find out }
end;

{$ENDIF}

{$IFDEF LCL_}
procedure ComboBoxUndo(ComboBox: TCustomComboBox);
begin
  { TDOD :  ComboBoxUndo }
end;

procedure ComboBoxRedo(ComboBox: TCustomComboBox);
begin
  { TDOD :  ComboBoxRedo }
end;

function ComboBoxCanUndo(ComboBox: TCustomComboBox):Boolean;
begin
  Result := True; { can't find out }
end;

function ComboBoxCanRedo(ComboBox: TCustomComboBox):Boolean;
begin
  Result := True; { can't find out }
end;
{$ENDIF}

end.
