{ Provides a interprogram system that can be used on both Windows and X11 }
unit CompatMsg;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFDEF UNIX}
       {$IFDEF FPC} Graphics, LCLType,
         {$IFDEF XBASED} X, XLib, {$IFDEF LCLgtk3} LazGdk3, {$ELSE} {$IFDEF LCLgtk2} gdk2, {$ELSE} gdk, {$ENDIF} {$ENDIF}{$ENDIF}
         {$IFDEF LCLcarbon} CarbonDef,  {$ENDIF}
         {$IFDEF LCLcocoa} CocoaAll,  {$ENDIF}
       {$ENDIF}
     {$ENDIF}
     SysUtils, Classes;

type

{ Typedef to help compatibility - the X Window/Windows Window Type}
{$IFDEF MSWINDOWS}
  TWnd = HWND;
{$ENDIF}
{$IFDEF UNIX}
  TWnd = {$IFDEF LCLcarbon} TCarbonWidget {$ENDIF}
         {$IFDEF LCLcocoa} THandle {$ENDIF}
         {$IFDEF XBASED}    TWindow {$ENDIF}
         ;
{$ENDIF}
  TMessageID = {$IFDEF MSWINDOWS} UINT {$ENDIF}
               {$IFDEF XBASED} TAtom {$ENDIF}
               {$IFDEF LCLcarbon} Integer {$ENDIF}
               {$IFDEF LCLcocoa} Integer {$ENDIF};

  TOnCompatMsg = procedure (Sender: TObject; wParam, lParam: Integer; var Handled: Boolean) of object;

  TCompatMsg = class(TComponent)
  private
    FMessageName: string;
    FMessageID: TMessageID;
    {$IFDEF XBASED}{$IFNDEF LCLgtk1}
    FGDKMessageID: TGdkAtom;
    {$ENDIF}{$ENDIF}
    FOnResultReceived: TOnCompatMsg;
    FOnReceived: TOnCompatMsg;
    FStatus: Integer;
    procedure SetMessageName(const Value: string);
    procedure SetOnReceived(const Value: TOnCompatMsg);
    procedure SetOnResultReceived(const AValue: TOnCompatMsg);
    procedure SetStatus(const AValue: Integer);
  protected
    FInternalHook: Boolean;
    procedure DoOnReceived(wParam, lParam: Integer; var Handled: Boolean); virtual;
//    procedure DoOnResultReceived(var Handled: Boolean); virtual;
  public
    procedure CheckHooked;
    destructor Destroy; override;
    property MessageID: TMessageID read FMessageID;
  published
    property Status: Integer read FStatus write SetStatus;
    property MessageName: string read FMessageName write SetMessageName;
    property OnReceived: TOnCompatMsg read FOnReceived write SetOnReceived;
    property OnResultReceived: TOnCompatMsg read FOnResultReceived write SetOnResultReceived;
  end;

{$IFDEF XBASED}
function RootWindow:TWnd;
{ Handler for clipboard when type = XA_PIXMAP }
function LoadBitmapFromClipboard(Bitmap: TBitmap):Boolean;
function X11Display:PDisplay;
function XHandle(Handle: HWND):TWnd;
function GetRootWindowString(Atom: TAtom): AnsiString;
function GetWindowInteger(Window: TWnd; Atom: TAtom):Integer;
{$ENDIF}

function MainWindow:TWnd;

implementation

{$IFDEF MSWINDOWS}
uses Forms;

function MainWindow: TWnd;
begin
  if Assigned(Application.MainForm) and Application.MainForm.HandleAllocated then
    Result := Application.MainForm.Handle
  else
    Result := 0;
end;
{$ENDIF}

{$IFDEF UNIX}
uses
  Clipbrd,
  {$IFDEF LCL_}
    {$IFDEF LCLcarbon} Forms; {$ENDIF}
    {$IFDEF LCLcocoa} Forms; {$ENDIF}
    {$IFDEF LCLgtk3} LazGtk3, LazGLib2, {$ENDIF}
    {$IFDEF LCLgtk2} gtk2, glib2, gdk2x, Gtk2Proc, Gtk2Def, {$ENDIF}
    {$IFDEF LCLgtk1} gtk, glib, GTKProc, GTKDef, {$ENDIF}
    {$IFDEF XBASED} Xatom, Forms, UnixType; {$ENDIF}
  {$ENDIF}
{$ENDIF}

var
  CompatMsgList: TList;

{$IFDEF XBASED}
{$IFDEF LCL_}
{$IFDEF LCLgtk3}
{ This block should be in gtk3 bindings }
function gdk_x11_get_default_xdisplay:PDisplay; cdecl; external;
function gdk_x11_window_get_xid(win : PGdkWindow):TWnd; cdecl; external;
function GDK_WINDOW_XID(win : PGdkWindow) : TXID;
begin
   Result:=gdk_x11_window_get_xid(win);
end;
{$ENDIF}

function X11Display:PDisplay;
begin
  Result := {$IFDEF LCLgtk3} gdk_x11_get_default_xdisplay {$ELSE} gdk_display {$ENDIF};
end;

{$IFDEF LCLgtk3}
function XHandle(Handle: HWND):TWnd;
var
  GdkWindow: PGdkWindow;
begin
  Result := 0;
  if Handle = 0 then Exit;
  GdkWindow := PGTKWidget(Handle)^.window;
  if GdkWindow = nil then Exit;
  Result := GDK_WINDOW_XID(GdkWindow);
end;
{$ENDIF}
{$IFDEF LCLgtk2}
function XHandle(Handle: HWND):TWnd;
var
  GdkWindow: PGdkWindow;
begin
  Result := 0;
  if Handle = 0 then Exit;
  GdkWindow := PGTKWidget(Handle)^.window;
  if GdkWindow = nil then Exit;
  Result := GDK_WINDOW_XWINDOW(GdkWindow);
end;
{$ENDIF}
{$IFDEF LCLgtk1}
function XHandle(Handle: HWND):TWnd;
var
  GdkWindowPrivate: PGdkWindowPrivate;
begin
  Result := 0;
  if Handle = 0 then Exit;
  GdkWindowPrivate := PGdkWindowPrivate(PGTKWidget(Handle)^.Window);
  if GdkWindowPrivate = nil then Exit;
  Result := GdkWindowPrivate^.xwindow;
end;
{$ENDIF}

function MainWindow: TWnd;
begin
  if Assigned(Application.MainForm) and Application.MainForm.HandleAllocated then
    Result := XHandle(Application.MainForm.Handle)
  else
    Result := 0;
end;
{$ENDIF}

var
  RootWindowSave: TWnd;

function RootWindow:TWnd;
begin
  if RootWindowSave = 0 then
    RootWindowSave := XDefaultRootWindow(X11Display);
  Result := RootWindowSave;
end;

function GetRootWindowString(Atom: TAtom): AnsiString;
var
  nItems, BytesLeft: UnixType.culong;
  ActualType: TAtom;
  ActualFormat: Integer;
  Items: PAnsiChar;
begin
  Result := '';
  Items := nil;
  if (XGetWindowProperty(X11Display, RootWindow, Atom, 0, not 0, Ord(False), AnyPropertyType, // XA_STRING,
                         @ActualType, @ActualFormat, @nItems, @BytesLeft, Pointer(@Items)) = Success)
     and Assigned(Items) then begin
    SetString(Result, Items, nItems);
    XFree(Items);
  end;
end;

function GetWindowInteger(Window: TWnd; Atom: TAtom):Integer;
var
  nItems, BytesLeft: UnixType.culong;
  ActualType: TAtom;
  ActualFormat: Integer;
  Items: PInteger;
begin
  Result := 0;
  Items := nil;
  if (XGetWindowProperty(X11Display, Window, Atom, 0, 1, Ord(False), XA_INTEGER,
                         @ActualType, @ActualFormat, @nItems, @BytesLeft, Pointer(@Items)) = Success)
     and Assigned(Items) then begin
    Result := Items^;
    XFree(Items);
  end;
end;

{$IFDEF LCL_}
function ClientMessageHook(xevent:PGdkXEvent; event:PGdkEvent; data:gpointer):TGdkFilterReturn;cdecl;
var
  i: Integer;
  Handled: Boolean;
  Evnt: TGdkEventType;
begin
  Handled := False;
  Evnt := PGdkEventAny(xevent)^.
    {$IFDEF LCLgtk3}
      type_;
    {$ENDIF}
    {$IFDEF LCLgtk2}
      _type;
    {$ENDIF}
    {$IFDEF LCLgtk1}
      thetype;
    {$ENDIF}
  if (Evnt = ClientMessage) and (Xlib.PXEvent(xevent)^.xclient.window = MainWindow)
     and Assigned(CompatMsgList) then
    { Must search through list as there is no procedure to remove a filter,
      and also to allow for multiple hooks on same message }
    for i := 0 to CompatMsgList.Count - 1 do
      with TCompatMsg(CompatMsgList.Items[i]) do
        if MessageID = Xlib.PXEvent(xevent)^.xclient.message_type then begin
          DoOnReceived(PXClientMessageEvent(xevent)^.data.l[0], PXClientMessageEvent(xevent)^.data.l[1], Handled);
          if Handled then
            Break;
        end;
   if Handled then
     Result := GDK_FILTER_REMOVE
   else
     Result := GDK_FILTER_CONTINUE; { or possibly GDK_FILTER_TRANSLATE ? }
end;
{$ENDIF}

var
  AtomsMonitored: array of TAtom;
{$ENDIF}

{$IFDEF LCLcarbon}
function MainWindow: TWnd;
begin
  if Assigned(Application.MainForm) and Application.MainForm.HandleAllocated then
    Result := TCarbonWidget(Application.MainForm.Handle)
  else
    Result := TWnd(0);
end;
{$ENDIF}

{$IFDEF LCLcocoa}
function MainWindow: TWnd;
begin
  if Assigned(Application.MainForm) and Application.MainForm.HandleAllocated then
    Result := Application.MainForm.Handle
  else
    Result := TWnd(0);
end;
{$ENDIF}

procedure Hook(CompatMsg: TCompatMsg);
{$IFDEF LCL_}
var
  i: Integer;
{$ENDIF}
begin
  if not Assigned(CompatMsgList) then begin
    CompatMsgList := TList.Create;
    // WriteLn(MainWindow);
  end;
  if CompatMsgList.IndexOf(CompatMsg) < 0 then
    CompatMsgList.Add(CompatMsg);
  {$IFDEF XBASED} {$IFDEF LCL_}
  { This logic is required because there is no gdk_remove_client_message function,
    so we make sure we call the add function only when absolutely necessary }
  for i := 0 to High(AtomsMonitored) do
    if AtomsMonitored[i] = CompatMsg.MessageID then
      Exit;
  SetLength(AtomsMonitored, Length(AtomsMonitored)+1);
  AtomsMonitored[High(AtomsMonitored)] := CompatMsg.MessageID;
  { TODO: LCLgtk3 }
  {$IFDEF LCLgtk2}
    gdk_add_client_message_filter(CompatMsg.FGDKMessageID, @ClientMessageHook, nil);
  {$ENDIF}
  {$IFDEF LCLgtk1}
    gdk_add_client_message_filter(CompatMsg.MessageID, @ClientMessageHook, nil);
  {$ENDIF}
  {$ENDIF} {$ENDIF}
end;

procedure UnHookAll;
begin
  if Assigned(CompatMsgList) then begin
    FreeAndNil(CompatMsgList);
  end;
end;

procedure UnHook(CompatMsg: TCompatMsg);
begin
  if Assigned(CompatMsgList) and (CompatMsgList.IndexOf(CompatMsg) >= 0) then begin
    CompatMsgList.Remove(CompatMsg);
    CompatMsgList.Pack;
    if CompatMsgList.Count = 0 then
      UnHookAll;
  end;
end;

{ TCompatMsg }

destructor TCompatMsg.Destroy;
begin
  FInternalHook := False;
  OnReceived := nil;
  OnResultReceived := nil;
  inherited;
end;

procedure TCompatMsg.DoOnReceived(wParam, lParam: Integer; var Handled: Boolean);
begin
  if Assigned(FOnReceived) then begin
    {$IFDEF UNIX}
    try
      Status := -1; { Processing }
      FOnReceived(Self, wParam, lParam, Handled);
      if Status = -1 then
        Status := Ord(Handled);
    except
      Status := -2; { Flag crash }
      Application.HandleException(Self);
    end;
    {$ENDIF}
  end;
end;

//procedure TCompatMsg.DoOnResultReceived(var Handled: Boolean);
//begin
  //if Assigned(FOnResultReceived) then
    //FOnResultReceived(Self, Handled);
//end;

procedure TCompatMsg.CheckHooked;
begin
  if not (csDesigning in ComponentState) then begin
    if not Assigned(OnReceived) and not Assigned(OnResultReceived) and not (FInternalHook) then
      UnHook(Self)
    else
      Hook(Self);
  end;
end;

procedure TCompatMsg.SetMessageName(const Value: string);
begin
  FMessageName := Value;
  if not (csDesigning in ComponentState) then begin
    {$IFDEF MSWINDOWS}
    FMessageID := RegisterWindowMessage(PChar(Value));
    {$ENDIF}
    {$IFDEF XBASED}
    FMessageID := XInternAtom(X11Display,PChar(Value), False);
      {$IFNDEF LCLgtk1}
      FGDKMessageID := gdk_atom_intern(PChar(Value), False);
      {$ENDIF}
    {$ENDIF}
  end;
end;

procedure TCompatMsg.SetOnReceived(const Value: TOnCompatMsg);
begin
  FOnReceived := Value;
  CheckHooked;
end;

procedure TCompatMsg.SetOnResultReceived(const AValue: TOnCompatMsg);
begin
  FOnResultReceived := AValue;
  CheckHooked;
end;

procedure TCompatMsg.SetStatus(const AValue: Integer);
begin
  if (FStatus=AValue) or (MainWindow = TWnd(0)) or (MessageID = 0) then exit;
  FStatus:=AValue;
  {$IFDEF XBASED}
  XChangeProperty(X11Display, MainWindow, MessageID, XA_INTEGER, 32, PropModeReplace, Pointer(@FStatus), 1);
  XFlush(X11Display);
  {$ENDIF}
end;

{$IFDEF XBASED}
{$IFDEF LCLgtk3}
function LoadBitmapFromClipboard(Bitmap: TBitmap):Boolean;
begin
  Result := False; { TODO }
end;
{$ENDIF}
{$IFDEF LCLgtk2}
{ Handler for clipboard when type = XA_PIXMAP }
function LoadBitmapFromClipboard(Bitmap: TBitmap):Boolean;
var
  MemStream: TMemoryStream;
  PixmapHandle: guint32;
  Pixmap: PGdkPixmap;
  w, h: Integer;
begin
  Result := False;
  MemStream:=TMemoryStream.Create;
  BeginGDKErrorTrap;
  try
    if Clipboard.GetFormat(XA_PIXMAP,MemStream) and (MemStream.Size > 0) then begin
      MemStream.Position := 0;
      MemStream.Read(PixmapHandle,SizeOf(TDrawable));
      Pixmap := gdk_pixmap_foreign_new(PixmapHandle);
      gdk_window_get_size(Pixmap, @w, @h);
      with Bitmap do begin
        Width := w;
        Height := h;
        gdk_draw_pixmap(TGtkDeviceContext(Canvas.Handle).Drawable, TGtkDeviceContext(Canvas.Handle).GC, Pixmap, 0, 0, 0, 0, -1, -1);
      end;
      gdk_pixmap_unref(Pixmap);
      Result := True;
    end;
  finally
    MemStream.Free;
    EndGDKErrorTrap;
  end;
end;
{$ENDIF}
{$ENDIF}

initialization

finalization
  UnHookAll;
end.
