unit FormGrid;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} { WinTypes, WinProcs, } Graphics, Controls, Forms, Dialogs, Grids,
     {$ELSE} QGraphics, QControls, QForms, QDialogs, QGrids, {$ENDIF}
     SysUtils, Classes, NGrid, Expr;

type

  TFormulaGrid = class(TNumberGrid)
  private
    FColumnVariableNames: TCaseStringList;
    FFormulae: TCaseStringList;
    FLex: TLex;
    FCurrentCol, FCurrentRow: Longint; { Cell currently being evaluated }
    FLexOptions: TLexOptions;
    CurrentRowVariable: TVariable;
    function GetColumnVariableName(i:Longint):string;
    procedure SetColumnVariableName(i:Longint; s:string);
    procedure SetColumnVariableNames(l:TCaseStringList);
    function ColumnValue(Sender: TObject; var a:array of Double):Double;
    function GetFormula(i:Longint):string;
    procedure SetFormula(i:Longint; s:string);
    procedure SetFormulae(l:TCaseStringList);
    function GetRowVariable: string;
    procedure SetRowVariable(rv: string);
    procedure SetLexOptions(LexOptions: TLexOptions);
  protected
    procedure CheckSizes; override;
    function CanChange(ARow, ACol:Longint):Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Changed(ACol, ARow: Longint); override;
    procedure ReparseNeeded;
    function GetCell(ARow, ACol: Longint; out v:Double):Boolean; override;
    property ColumnVariableName[i:Longint]: string read GetColumnVariableName write SetColumnVariableName;
    property Formula[i:Longint]: string read GetFormula write SetFormula;
    // property Valid read GetValid write SetValid;
    // property Value read GetValue write SetValue;
  published
    property ColumnVariableNames: TCaseStringList read FColumnVariableNames write SetColumnVariableNames;
    property Formulae: TCaseStringList read FFormulae write SetFormulae;
    property RowVariable: string read GetRowVariable write SetRowVariable;
    property LexOptions: TLexOptions read FLexOptions write SetLexOptions
             default [ xGlobalVariables, xGlobalFunctions, xFundamentalConstants, xBinary, xOctal, xCHex, xPascalHex ];
  end;

implementation

uses MathSup;

{$IFDEF SINGLE}
{$IFDEF WIN32}
{$R FormGrid.d32}
{$ELSE}
{$R FormGrid.d16}
{$ENDIF}
{$ENDIF}

type
  TOnEvalFunctionEvent = function (Sender: TObject; var a:array of Double):Double of object;

  TColumnFunction = class (TFunction)
  public
    Col: Longint;
    OnEvalFunction: TOnEvalFunctionEvent;
    function Value(var a:array of Double):Double; override;
    constructor Create(ACol:Longint; s:String; AOnEvalFunction:TOnEvalFunctionEvent);
  end;

constructor TColumnFunction.Create(ACol:Longint; s:String; AOnEvalFunction:TOnEvalFunctionEvent);
begin
  inherited Create(s, 0);
  Col := Acol;
  OnEvalFunction := AOnEvalFunction;
end;

function TColumnFunction.Value(var a:array of Double):Double;
begin
  Result := OnEvalFunction(Self, a);
end;

constructor TFormulaGrid.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLexOptions := [ xGlobalVariables, xGlobalFunctions, xFundamentalConstants, xBinary, xOctal, xCHex, xPascalHex ];
  FColumnVariableNames := TCaseStringList.Create;
  FFormulae := TCaseStringList.Create;
  ReparseNeeded;
  CurrentRowVariable := TVariable.Create('');
end;

destructor TFormulaGrid.Destroy;
begin
  FreeStringList(FFormulae);
  FLex.Free;
  Flex := nil;
  FreeStringList(FColumnVariableNames); { This frees up TVariables }
  CurrentRowVariable.Free;
  inherited Destroy;
end;

function TFormulaGrid.GetRowVariable: string;
begin
  Result := CurrentRowVariable.Name;
end;

procedure TFormulaGrid.SetRowVariable(rv: string);
begin
  CurrentRowVariable.Name := rv;
  ReparseNeeded;
end;

procedure TFormulaGrid.ReparseNeeded;
  procedure FreeObjects(sl: TStringList);
  var
    i: Longint;
  begin
    for i := 0 to sl.Count - 1 do begin
      sl.Objects[i].Free;
      sl.Objects[i] := nil;
    end;
    if sl.Count >= ColCount then
      for i := sl.Count-1 downto ColCount-1 do
        sl.Delete(i);
  end;
begin
  if Flex <> nil then begin
    FreeObjects(FFormulae);
    FreeObjects(FColumnVariableNames);
    Flex.Free;
    Flex := nil;
  end;
end;

procedure TFormulaGrid.CheckSizes;
begin
  ReparseNeeded;
  if Assigned(FColumnVariableNames) then
    while FColumnVariableNames.Count >= ColCount do
      FColumnVariableNames.Delete(FColumnVariableNames.Count-1);
  inherited CheckSizes;
end;


function TFormulaGrid.GetColumnVariableName(i:Longint):string;
begin
  if (i > 0) and (i <= FColumnVariableNames.Count) then
    Result := FColumnVariableNames.Strings[i-1]
  else
    Result := '';
end;

procedure TFormulaGrid.SetColumnVariableName(i:Longint; s:string);
begin
  if (i > FColumnVariableNames.Count) and (i <= ColCount-1) then
    while FColumnVariableNames.Count < i do
      FColumnVariableNames.Add('');
  if (i > 0) and (i <= FColumnVariableNames.Count) then
    FColumnVariableNames[i-1] := s;
  ReparseNeeded;
end;

procedure TFormulaGrid.SetColumnVariableNames(l:TCaseStringList);
var
  i: Integer;
begin
  for i := 0 to l.Count - 1 do
    SetColumnVariableName(i+1, l[i]);
end;

function TFormulaGrid.ColumnValue(Sender: TObject; var a:array of Double):Double;
begin
  if not GetCell(FCurrentRow, TColumnFunction(Sender).Col, Result) then
    Abort;
end;

function TFormulaGrid.GetFormula(i:Longint):string;
begin
  if (i > 0) and (i <= FFormulae.Count) then
    Result := FFormulae.Strings[i-1]
  else
    Result := '';
end;

procedure TFormulaGrid.SetFormula(i:Longint; s:string);
begin
  if (i > FFormulae.Count) and (i <= ColCount-1) then
    while FFormulae.Count < i do
      FFormulae.Add('');
  if (i > 0) and (i <= FFormulae.Count) and (FFormulae[i-1] <> s) then begin
    ReparseNeeded;
    FFormulae[i-1] := s;
    Changed(i, -1);
  end;
end;

procedure TFormulaGrid.SetLexOptions(LexOptions: TLexOptions);
begin
  if LexOptions <> FLexOptions then begin
    FLexOptions := LexOptions;
    ReparseNeeded;
  end;
end;

procedure TFormulaGrid.SetFormulae(l:TCaseStringList);
var
  i: Integer;
begin
  for i := 0 to l.Count - 1 do
    SetFormula(i-1, l[i]);
end;

function TFormulaGrid.CanChange(ARow, ACol:Longint):Boolean;
begin
  Result := (ACol <= 0) or (ACol > FFormulae.Count) or (Length(FFormulae[ACol-1]) = 0);
  if Result then
    Result := inherited CanChange(ARow, ACol);
end;

procedure TFormulaGrid.Changed(ACol, ARow: Longint);
var
  i: Longint;
begin
  if (FUpdateCount = 0) and (ARow > 0) then
    { One Column or cell changed - ensure formula columns invalidated }
    for i := 0 to FFormulae.Count -1 do
      if Length(FFormulae[i]) <> 0 then
        InvalidateCell(i+1, ARow);
  inherited Changed(ACol, ARow);
end;

function TFormulaGrid.GetCell(ARow, ACol: Longint; out v:Double):Boolean;
var
  ColCur, RowCur: Longint; { Cell previously being evaluated }
  i:Integer;
begin
  if (ACol <= 0) or (ACol > FFormulae.Count) or (Length(FFormulae[ACol-1]) = 0) then
    Result := inherited GetCell(ARow, ACol, v)
  else begin
    if FLex = nil then begin
      FLex := TLex.Create;
      Flex.Options := LexOptions;
      Flex.AddVariable(CurrentRowVariable);
      for i := 0 to FColumnVariableNames.Count-1 do begin
        if FColumnVariableNames.Objects[i] = nil then
          FColumnVariableNames.Objects[i] := TColumnFunction.Create(i+1,'',{$IFDEF FPC}@{$ENDIF}ColumnValue);
        TFunction(FColumnVariableNames.Objects[i]).Name := FColumnVariableNames[i];
        FLex.AddFunction(TFunction(FColumnVariableNames.Objects[i]));
      end;
      for i := 0 to FFormulae.Count-1 do
        if Length(FFormulae[i]) > 0 then begin
          if FFormulae.Objects[i] = nil then
            FFormulae.Objects[i] := TExpr.Create(FLex);
          TExpr(FFormulae.Objects[i]).Text := FFormulae[i];
        end;
    end;
    ColCur := FCurrentCol;
    RowCur := FCurrentRow;
    FCurrentCol := ACol;
    FCurrentRow := ARow;
    CurrentRowVariable.Value := ARow;
    try
      v := TExpr(FFormulae.Objects[ACol-1]).Value;
      Result := not IsNanOrInf(v);
    except
      Result := False;
    end;
    FCurrentCol := ColCur;
    FCurrentRow := RowCur;
  end;
end;

end.

