unit FrameGrp;

{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses {$IFNDEF CLX_} { WinTypes, WinProcs, } Messages, Controls, Graphics,
                    Forms, Dialogs, StdCtrls, Menus, ExtCtrls, Buttons,
     {$ELSE} QControls, QGraphics, QForms, QDialogs, QStdCtrls, QMenus,
                    QExtCtrls, QButtons, Types, {$ENDIF}
     SysUtils, Classes;

{ This class allows the use of (selected) Forms as a component on a form }
type

  { TFrameGroupBox }

  TFrameGroupBox = class(TCustomGroupBox)
  protected
    {$IFDEF VCL_}
    procedure ReadInteger(Reader: TReader);
    procedure ReadBoolean(Reader: TReader);
    {$ELSE}
    procedure SkipValue(Reader: TReader);
    {$ENDIF}
    {$IFDEF LCL_}
    procedure PropertyNotFound(Reader: TReader; Instance: TPersistent; var PropName: string; IsPath: boolean; var Handled, Skip: Boolean);
    procedure ReadState(Reader: TReader); override;
    {$ENDIF}
  public
    constructor Create(AOwner: TComponent); override;
    procedure DefineProperties(Filer: TFiler); override;
  {$IFDEF FPC}
  published
    property AutoSize;
  {$ENDIF}
  end;

implementation

{ TFrameGroupBox }

{ Below are a fudge to allow editing of control as form }

{$IFDEF LCL_}

uses LResources;

procedure TFrameGroupBox.PropertyNotFound(Reader: TReader; Instance: TPersistent;
    var PropName: string; IsPath: boolean; var Handled, Skip: Boolean);
begin
  // WriteLn(Message);
  Handled := True;
  Skip := True;
end;

procedure TFrameGroupBox.ReadState(Reader: TReader);
begin
  Reader.OnPropertyNotFound := @PropertyNotFound;
  inherited ReadState(Reader);
end;
{$ENDIF}

{$IFNDEF VCL_}
type
  TOpenReader = class(TReader)
  function GetPropname: string;
  end;

{ TOpenReader }

function TOpenReader.GetPropname: string;
begin
  Result := PropName;
end;

procedure TFrameGroupBox.SkipValue(Reader: TReader);
begin
  {$IFDEF CLX_}
  TOpenReader(Reader).SkipValue;
  {$ENDIF}
  {$IFDEF LCL_}
  Reader.Driver.SkipValue;
  {$ENDIF}
end;
{$ENDIF}

{$IFDEF VCL_}
procedure TFrameGroupBox.ReadInteger(Reader: TReader);
begin
  {ignore result} Reader.ReadInteger;
end;

procedure TFrameGroupBox.ReadBoolean(Reader: TReader);
begin
  {ignore result} Reader.ReadBoolean;
end;
{$ENDIF}

procedure TFrameGroupBox.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  {$IFDEF VCL_}
  Filer.DefineProperty('TextHeight', ReadInteger, nil, False);
  Filer.DefineProperty('PixelsPerInch', ReadInteger, nil, False);
  Filer.DefineProperty('OldCreateOrder', ReadBoolean, nil, False);
  Filer.DefineProperty('ClientHeight', ReadInteger, nil, False);
  Filer.DefineProperty('ClientWidth', ReadInteger, nil, False);
  {$ELSE}
  { Discard any properties not read }
  if (Filer is TReader) and (Length(TOpenReader(Filer).GetPropName) > 0) then
    Filer.DefineProperty(TOpenReader(Filer).GetPropName, {$IFDEF FPC}@{$ENDIF}SkipValue, nil, False);
  {$ENDIF}
end;

constructor TFrameGroupBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {$IFNDEF LCL_}
  ReadComponentRes(Self.ClassName, Self);
  {$ELSE}
  InitResourceComponent(Self, TFrameGroupBox);
  {$ENDIF}
end;

end.

