unit FxFntDlg;

interface
{$I utils.inc}

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  {$IFNDEF CLX_} Messages, Graphics, Controls, Forms, Dialogs,
  {$ELSE} QGraphics, QControls, QForms, QDialogs, Qt, {$ENDIF}
  {$IFNDEF USECOMMON} FixDlg, {$ENDIF}
  {$IFDEF LCL_} LCLType, {$ENDIF}
  SysUtils, Classes, FixFnt;

type
  TFixedFontDialog = class(TComponent)
  private
    FFixedFont: TFixedFont;
    FTitle: string;
    FLocal: Boolean;
    FOnChange: TNotifyEvent;
    procedure SetFixedFont(const Value: TFixedFont);
//    function GetHandle: {$IFNDEF CLX_} HFont {$ELSE} QFontH {$ENDIF};
    function GetFixedFont: TFixedFont;
    procedure SetLocal(const Value: Boolean);
  protected
    procedure Changed(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute:Boolean;
    procedure Loaded; override;
//    property Handle: {$IFNDEF CLX_} HFont {$ELSE} QFontH {$ENDIF} read GetHandle;
  published
    property FixedFont:TFixedFont read GetFixedFont write SetFixedFont;
    property Title:string read FTitle write FTitle;
    property Local:Boolean read FLocal write SetLocal default False;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

implementation

{ TFixedFontDialog }

{$IFDEF SINGLE}
{$R FxFntDlg.d32}

procedure Register;
begin
  RegisterComponents('CMW', [TFixedFontDialog]);
end;
{$ENDIF}

var
  CommonFixedFontDialogs: TList; { List of synchronized TFixedFontDialog's }

constructor TFixedFontDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

destructor TFixedFontDialog.Destroy;
begin
  Local := True; { Force Disconnect from CommonFixedFontDialogs }
  FFixedFont.Free;
  FFixedFont := nil;
  inherited Destroy;
end;

type
  TFntDlg = {$IFDEF USECOMMON} TFontDialog {$ELSE} TFixedFontForm {$ENDIF};
  
function TFixedFontDialog.Execute: Boolean;
var
  Dialog: TFntDlg;
begin
  Dialog := nil;
  try
    Dialog := TFntDlg.Create(Application);
    {$IFDEF USECOMMON}
    if Title <> '' then Dialog.Title := Title;
    Dialog.Options := Dialog.Options + [fdFixedPitchOnly];
    FixedFont.CopyTo(Dialog.Font);
    Result := Dialog.Execute;
    {$ELSE}
    if Title <> '' then Dialog.Caption := Title;
    Dialog.FixedFont := FixedFont;
    Result := Dialog.ShowModal = mrOK;
    {$ENDIF}
    if Result then
      {$IFDEF USECOMMON}
      FixedFont.Assign(Dialog.Font);
      {$ELSE}
      FixedFont := Dialog.FixedFont;
      {$ENDIF}
    finally
      Dialog.Free;
    end;
end;

procedure TFixedFontDialog.SetFixedFont(const Value: TFixedFont);
begin
  FixedFont.Assign(Value);
end;

//function TFixedFontDialog.GetHandle:{$IFNDEF CLX_} HFont {$ELSE} QFontH {$ENDIF};
//begin
//  Result := FixedFont.Handle;
//end;

function TFixedFontDialog.GetFixedFont: TFixedFont;
begin
  if Local or (csLoading in ComponentState) then begin
    if FFixedFont = nil then begin
      FFixedFont := TFixedFont.Create;
      FFixedFont.OnChange := {$IFDEF FPC}@{$ENDIF}Changed;
    end;
    Result := FFixedFont;
  end else begin
    if FixFnt.FixedFont = nil then
      FixFnt.FixedFont := TFixedFont.Create;
    Result := FixFnt.FixedFont;
  end;
end;

procedure TFixedFontDialog.Loaded;
begin
  inherited Loaded;
  if not Local then begin
    if (FFixedFont <> nil) and (FixFnt.FixedFont = nil) then
      FixedFont := FFixedFont; { Set global fixed font }
    FLocal := True; { Force SetLocal call }
    Local := False;
  end;
end;

procedure TFixedFontDialog.SetLocal(const Value: Boolean);
begin
  if Value <> Local then begin
    FLocal := Value;
    if not (csLoading in ComponentState) then
      if Local then begin
        if CommonFixedFontDialogs <> nil then
          CommonFixedFontDialogs.Remove(Self);
        if FixFnt.FixedFont <> nil then
          with FixFnt.FixedFont do begin
            if (CommonFixedFontDialogs <> nil) and (CommonFixedFontDialogs.Count > 0) then
              OnChange := {$IFDEF FPC}@{$ENDIF}TFixedFontDialog(CommonFixedFontDialogs.Items[0]).Changed
            else
              OnChange := nil;
          end;
      end else begin
        FFixedFont.Free;
        FFixedFont := nil;
        if CommonFixedFontDialogs = nil then
          CommonFixedFontDialogs := TList.Create;
        CommonFixedFontDialogs.Add(Self);
        FixedFont.OnChange := {$IFDEF FPC}@{$ENDIF}Changed;
      end;
  end;
end;

var
  ChangedLockout: Integer;

procedure TFixedFontDialog.Changed(Sender: TObject);
var
  i: Integer;
begin
  if not Local and (CommonFixedFontDialogs <> nil) and (ChangedLockout = 0) then
    try
      ChangedLockout := 1;
      for i := 0 to CommonFixedFontDialogs.Count-1 do
        TFixedFontDialog(CommonFixedFontDialogs.Items[i]).Changed(Self);
    finally
      ChangedLockout := 0;
    end
  else
    if Assigned(FOnChange) then FOnChange(Sender);
end;

{$F+}
procedure Finalize;
begin
  FreeAndNil(CommonFixedFontDialogs);
end;

initialization
  ChangedLockout := 0;
finalization
  Finalize;
end.
