unit GPenDlg;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} { WinTypes, WinProcs, } Messages, Graphics, Controls, Forms, Dialogs,
     {$ELSE} QGraphics, QControls, QForms, QDialogs, {$ENDIF}
     SysUtils, Classes, GStyle, Graphs, GraphI, GraphPen, PenChse;

type
  TGraphPenDialog = class(TComponent)
  private
    FLineStyle: TLineStyle;
    FGraph: TGraphB;
    FTitle: string;
    FPen: Integer;
    FEditAllPens: Boolean;
    Dialog: TPenDialog;
    FEnabledItems: TPenChooseItems;
    FDefaultLineStyle: TLineStyle;
    FOnSetDefaults: TReplotEvent;
    procedure ReadFromDialog;
    procedure LoadDialog;
    procedure PenNumberChange(Sender: TObject);
    procedure SetLineStyle(Style:TLineStyle);
    procedure SetEnabledItems(const Value: TPenChooseItems);
    procedure ReadGOptions;
    procedure EnableItems;
    procedure SetDefaultLineStyle(const Value: TLineStyle);
    procedure DefaultClick;
    procedure DefaultsClick;
  protected
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Apply;
    procedure ApplyAll;
    function Execute:Boolean;
    property DefaultLineStyle:TLineStyle read FDefaultLineStyle write SetDefaultLineStyle;
  published
    property Title:string read FTitle write FTitle;
    property Graph:TGraphB read FGraph write FGraph;
    property Pen:Integer read FPen write FPen default 0;
    property EditAllPens: Boolean read FEditAllPens write FEditAllPens default False;
    property LineStyle:TLineStyle read FLineStyle write SetLineStyle;
    property EnabledItems: TPenChooseItems read FEnabledItems write SetEnabledItems;
    property OnSetDefaults: TReplotEvent read FOnSetDefaults write FOnSetDefaults;
  end;

implementation

constructor TGraphPenDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLineStyle := TLineStyle.Create;
  FDefaultLineStyle := TLineStyle.Create;
  FEnabledItems := [pcColour, pcWidth, pcStyle];
end;

destructor TGraphPenDialog.Destroy;
begin
  FLineStyle.Free;
  FDefaultLineStyle.Free;
  inherited Destroy;
end;

procedure TGraphPenDialog.SetLineStyle(Style:TLineStyle);
begin
  FLineStyle.Assign(Style);
  if Assigned(Dialog) then
    LoadDialog;
end;

procedure TGraphPenDialog.LoadDialog;
begin
  Dialog.PenChoose.LineStyle := LineStyle;
  Dialog.PenChoose.EnabledItems := EnabledItems;
  if Assigned(FGraph) then begin
    Dialog.AllowColour.Checked := grColouredPens in Graph.GOptions;
    Dialog.AllowWidth.Checked := grPensHaveWidth in Graph.GOptions;
    Dialog.AllowStyle.Checked := grPensHaveStyle in Graph.GOptions;
  end;
  Dialog.PenChoose.ChangedItems := [];
end;

procedure TGraphPenDialog.ReadFromDialog;
begin
  ReadGOptions;
  LineStyle := Dialog.PenChoose.LineStyle;
  if EditAllPens then
    Pen := Dialog.Pen.Value;
end;

function TGraphPenDialog.Execute:Boolean;
begin
  Dialog := TPenDialog.Create(Application);
  try
    if Title <> '' then Dialog.Caption := Title;
    if not Assigned(FGraph) or (Pen < 0) or (Pen > MaxPens) then begin
      with Dialog do begin
        DisableAlign;
        PenLabel.Visible := False;
        Pen.Visible := False;
        ApplyToAll.Visible := False;
        AllowColour.Visible := False;
        AllowWidth.Visible := False;
        AllowStyle.Visible := False;
        DefaultsBtn.Visible := False;
        {$IFDEF FPC}
          DefaultBtn.Anchors := DefaultBtn.Anchors - [akRight];
          DefaultBtn.AnchorSideLeft.Control := PenChoose;
        {$ELSE}
          ClientHeight := PenChoose.Top + PenChoose.Height + PenChoose.Left;
          Apply.Left := PenChoose.ClientWidth + 2*PenChoose.Left;
          OK.Left := Apply.Left;
          Cancel.Left := Apply.Left;
          DefaultBtn.Left := Apply.Left;
          ClientWidth := Apply.Left + Apply.Width + PenChoose.Left;
        {$ENDIF}
        EnableAlign;
      end;
      { DefaultLineStyle.Assign(LineStyle); }
    end else begin
      with Dialog do begin
        DisableAlign;
        AllowColour.Visible := True;
        AllowWidth.Visible := True;
        AllowStyle.Visible := True;
        DefaultsBtn.Visible := True;
        {$IFDEF FPC}
          DefaultBtn.Anchors := DefaultBtn.Anchors + [akRight];
          DefaultBtn.AnchorSideLeft.Control := AllowColour;
        {$ELSE}
          ClientHeight := ApplyToAll.Top + ApplyToAll.Height + PenChoose.Left;
          Apply.Left := AllowColour.Left + AllowColour.Width + PenChoose.Left;
          OK.Left := Apply.Left;
          Cancel.Left := Apply.Left;
          DefaultBtn.Left := Apply.Left;
          DefaultsBtn.Left := Apply.Left;
          ClientWidth := Apply.Left + Apply.Width + PenChoose.Left;
        {$ENDIF}
      end;
      LineStyle := Graph.LineStyles[Pen];
      EnableItems;
      Dialog.Pen.MaxValue := MaxPens-1;
      Dialog.Pen.Value := Pen;
      { Graph.SetToDefaultStyle(Pen, FDefaultLineStyle); }
      if not EditAllPens then begin
        Dialog.Pen.Visible := False;
        Dialog.PenNumber.Visible := True;
        if Pen = MaxPens then
          Dialog.PenNumber.Caption := 'Axis'
        else
          Dialog.PenNumber.Caption := IntToStr(Pen);
        Dialog.ApplyToAll.Visible := False;
      end else begin
        Dialog.Pen.Visible := True;
        Dialog.PenNumber.Visible := False;
        Dialog.PenLabel.Visible := True;
        Dialog.ApplyToAll.Visible := True;
      end;
      Dialog.DoApply := {$IFDEF FPC}@{$ENDIF}Apply;
      Dialog.DoApplyAll := {$IFDEF FPC}@{$ENDIF}ApplyAll;
      Dialog.EnableAlign;
    end;
    Dialog.DoDefaults := {$IFDEF FPC}@{$ENDIF}DefaultsClick;
    Dialog.DoDefault := {$IFDEF FPC}@{$ENDIF}DefaultClick;
    LoadDialog;
    Dialog.Pen.OnChange := {$IFDEF FPC}@{$ENDIF}PenNumberChange;
    Result := Dialog.ShowModal = mrOK;
    if Result then
      ReadFromDialog;
  finally
    Dialog.Free;
    Dialog := nil;
  end;
end;

procedure TGraphPenDialog.Apply;
begin
  if Assigned(FGraph) and (Pen >= 0) and (Pen <= MaxPens) then begin
    if Assigned(Dialog) then
      ReadFromDialog;
    FGraph.Linestyles[Pen] := LineStyle;
  end;
  if Assigned(Dialog) then
    LoadDialog;
end;

procedure TGraphPenDialog.PenNumberChange(Sender: TObject);
var
  Old: Integer;
begin
  if Pen <> Dialog.Pen.Value then begin
    Old := Pen;
    ReadFromDialog;
    FGraph.Linestyles[Old] := LineStyle;
    FLineStyle.Assign(FGraph.Linestyles[Pen]);
    LoadDialog;
  end;
end;

procedure TGraphPenDialog.SetEnabledItems(const Value: TPenChooseItems);
begin
  FEnabledItems := Value;
  if Assigned(Dialog) then
    LoadDialog;
end;

procedure TGraphPenDialog.ApplyAll;
var
  i: Integer;
begin
  if Assigned(FGraph) then begin
    ReadGOptions;
    FLineStyle.Assign(Dialog.PenChoose.LineStyle);
    for i := 0 to MaxPens-1 do begin
      if pcColour in Dialog.PenChoose.ChangedItems then
        FGraph.Linestyles[i].Colour := LineStyle.Colour;
      if pcWidth in Dialog.PenChoose.ChangedItems then
        FGraph.Linestyles[i].Width := LineStyle.Width;
      if pcStyle in Dialog.PenChoose.ChangedItems then
        FGraph.Linestyles[i].Pattern := LineStyle.Pattern;
    end;
    FLineStyle.Assign(FGraph.Linestyles[Pen]);
    LoadDialog;
  end;
  Dialog.PenChoose.ChangedItems := [];
end;

procedure TGraphPenDialog.ReadGOptions;
  procedure CkOpt(Opt: TGraphOption; Value: Boolean);
  begin
    if Value then
      Graph.GOptions := Graph.GOptions + [Opt]
    else
      Graph.GOptions := Graph.GOptions - [Opt];
  end;
begin
  if Assigned(FGraph) and Dialog.AllowColour.Visible then begin
    CkOpt(grColouredPens, Dialog.AllowColour.Checked);
    CkOpt(grPensHaveWidth, Dialog.AllowWidth.Checked);
    CkOpt(grPensHaveStyle, Dialog.AllowStyle.Checked);
    EnableItems;
  end;
end;

procedure TGraphPenDialog.EnableItems;
begin
  FEnabledItems := [];
  if grColouredPens in Graph.GOptions then
    FEnabledItems := FEnabledItems + [pcColour];
  if grPensHaveWidth in Graph.GOptions then
    FEnabledItems := FEnabledItems + [pcWidth];
  if grPensHaveStyle in Graph.GOptions then
    FEnabledItems := FEnabledItems + [pcStyle];
end;

procedure TGraphPenDialog.SetDefaultLineStyle(const Value: TLineStyle);
begin
  FDefaultLineStyle.Assign(Value);
end;

procedure TGraphPenDialog.DefaultClick;
begin
  LineStyle.Assign(DefaultLineStyle);
  LoadDialog;
end;

procedure TGraphPenDialog.DefaultsClick;
var
  i: Integer;
begin
  ReadGOptions;
  if Assigned(FGraph) then begin
    for i := 0 to MaxPens do begin
      Graph.SetToDefaultStyle(i, Graph.LineStyles[i]);
      if i = Pen then LineStyle.Assign(Graph.LineStyles[i]);
    end;
    if Assigned(FOnSetDefaults) then
      FOnSetDefaults(Self, Graph);
  end;
  LoadDialog;
end;

procedure TGraphPenDialog.Loaded;
begin
  inherited Loaded;
  DefaultLineStyle.Assign(LineStyle);
end;

end.

