unit LogFiDlg;

interface
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  {$IFNDEF CLX_} Graphics, Forms, Controls, Buttons,
                 StdCtrls, ExtCtrls, Dialogs, Menus,
  {$ELSE} QGraphics, QForms, QControls, QButtons, QStdCtrls, QExtCtrls,
          QDialogs, QMenus, QTypes, {$ENDIF}
  {$IFDEF LCL_} SrchStrs, {$ENDIF}
  Classes;

type

  { TLogFindDlg }

  TLogFindDlg = class(TForm)
    Label1: TLabel;
    FindNext: TButton;
    FindText: TMemo;
    Cancel: TButton;
    WholeWord: TCheckBox;
    MatchCase: TCheckBox;
    Direction: TRadioGroup;
    NewLine: TButton;
    Tab: TButton;
    Label2: TLabel;
    ReplaceText: TMemo;
    Replace: TButton;
    ReplaceAll: TButton;
    LocalPopupMenu: TPopupMenu;
    PasteItem: TMenuItem;
    CopyItem: TMenuItem;
    CutItem: TMenuItem;
    N3: TMenuItem;
    UndoItem: TMenuItem;
    DeleteItem: TMenuItem;
    N1: TMenuItem;
    SelectAllItem: TMenuItem;
    N2: TMenuItem;
    NewlineItem: TMenuItem;
    TabItem: TMenuItem;
    PopupItem: TMenuItem;
    procedure FormResize(Sender: TObject);
    procedure NewLineClick(Sender: TObject);
    procedure TabClick(Sender: TObject);
    procedure CancelClick(Sender: TObject);
    procedure FindNextClick(Sender: TObject);
    procedure ReplaceClick(Sender: TObject);
    procedure ReplaceAllClick(Sender: TObject);
    procedure FindTextEnter(Sender: TObject);
    procedure CopyItemClick(Sender: TObject);
    procedure CutItemClick(Sender: TObject);
    procedure PasteItemClick(Sender: TObject);
    procedure UndoItemClick(Sender: TObject);
    procedure SelectAllItemClick(Sender: TObject);
    procedure PopupItemClick(Sender: TObject);
    procedure DeleteItemClick(Sender: TObject);
    procedure PopupMenuPopup(Sender: TObject);
  private
    FIsReplace: Boolean;
    OrgCancelPos: Integer;
    procedure SetIsReplace(const Value: Boolean);
  public
    OnFindNext: TNotifyEvent;
    OnReplace: TNotifyEvent;
    LastOption: TFindOption;
    LastText: TMemo;
    property IsReplace: Boolean read FIsReplace write SetIsReplace;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DoShow; override;
  end;

var
  LogFindDlg: TLogFindDlg;

implementation

uses {$IFDEF MSWINDOWS} Messages, {$ENDIF} WCompat, UCompat;

{$IFNDEF LCL_}
{$R *.DFM}
{$ELSE}
{$R *.lfm}
{$ENDIF}

constructor TLogFindDlg.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  LastText := FindText;
  FIsReplace := True;
  OrgCancelPos := Cancel.Top;
  Left := Screen.Width-Width;
  Top := Screen.Height - Height * 12 div 10;
end;

procedure TLogFindDlg.NewLineClick(Sender: TObject);
begin
  LastText.SelText := LineEnd;
  LastText.SetFocus;
end;

procedure TLogFindDlg.TabClick(Sender: TObject);
begin
  LastText.SelText := #9;
  LastText.SetFocus;
end;

procedure TLogFindDlg.CancelClick(Sender: TObject);
begin
  Close;
end;

procedure TLogFindDlg.FindNextClick(Sender: TObject);
begin
  if FindText.GetTextLen = 0 then Exit;
  if Assigned(OnFindNext) then OnFindNext(Self);
end;

procedure TLogFindDlg.ReplaceClick(Sender: TObject);
begin
  LastOption := frReplace;
  if Assigned(OnReplace) then OnReplace(Self);
  FindNextCLick(Self);
end;

procedure TLogFindDlg.ReplaceAllClick(Sender: TObject);
begin
  LastOption := frReplaceAll;
  if Assigned(OnReplace) then OnReplace(Self);
end;

procedure TLogFindDlg.FindTextEnter(Sender: TObject);
begin
  LastText := Sender as TMemo;
end;

procedure TLogFindDlg.CopyItemClick(Sender: TObject);
begin
  LastText.CopyToClipboard;
end;

procedure TLogFindDlg.CutItemClick(Sender: TObject);
begin
  LastText.CutToClipboard;
end;

procedure TLogFindDlg.PasteItemClick(Sender: TObject);
begin
  LastText.PasteFromClipboard;
end;

procedure TLogFindDlg.UndoItemClick(Sender: TObject);
begin
  {$IFDEF MSWINDOWS}
  SendMessage(LastText.Handle, EM_UNDO, 0, 0);
  {$ENDIF}
  {$IFDEF UNIX}
  // TODO: send undo char to control
  {$ENDIF}
end;

procedure TLogFindDlg.SelectAllItemClick(Sender: TObject);
begin
  LastText.SelectAll;
end;

procedure TLogFindDlg.PopupItemClick(Sender: TObject);
begin
  LastText.PopupMenu.PopUp(Left+FindText.Left+10, Top+FindText.Top+10);
end;

procedure TLogFindDlg.DeleteItemClick(Sender: TObject);
begin
  with LastText do begin
    if SelLength = 0 then
      SelLength := 1;
    ClearSelection;
  end;
end;

procedure TLogFindDlg.PopupMenuPopup(Sender: TObject);
begin
  {$IFDEF MSWINDOWS}
  UndoItem.Enabled := SendMessage(LastText.Handle, EM_CANUNDO, 0, 0) <> 0;
  {$ENDIF}
  {$IFDEF UNIX}
  // TODO: send undo char to control
  UndoItem.Enabled := True;
  {$ENDIF}
  CutItem.Enabled := LastText.SelLength > 0;
  CopyItem.Enabled := CutItem.Enabled;
  PasteItem.Enabled := ClipboardHasText;
  SelectAllItem.Enabled := LastText.GetTextLen > 0;
end;

procedure TLogFindDlg.SetIsReplace(const Value: Boolean);
var
  i, sign, NewTop: Integer;
begin
  if Value <> FIsReplace then begin
    DisableAlign;
    FIsReplace := Value;
    if IsReplace then begin
      Caption := 'Replace';
      sign := 1;
    end else begin
      Caption := 'Find';
      sign := -1;
    end;
    Replace.Visible := IsReplace;
    ReplaceAll.Visible := IsReplace;
    ReplaceText.Visible := IsReplace;
    Label2.Visible := IsReplace;
    {$IFDEF LCL_}
    MatchCase.Tag := 0;
    WholeWord.Tag := 0;
    Cancel.Tag := 0;
    if IsReplace then begin
      Cancel.AnchorSideTop.Control := ReplaceAll;
      FindNext.Anchors := FindNext.Anchors  + [akLeft];
      FindText.Anchors := FindText.Anchors  - [akBottom];
    end else begin
      Cancel.AnchorSideTop.Control := FindNext;
      FindNext.Anchors := FindNext.Anchors  - [akLeft];
      FindText.Anchors := FindText.Anchors  + [akBottom];
    end;
    {$ENDIF}
    {$IFDEF VCL_}
      if IsReplace then
        Cancel.Top := OrgCancelPos
      else begin
        Cancel.Top := Replace.Top;
        FindText.Height := WholeWord.Top - FindText.Top - 2;
      end;
    {$ENDIF}
    EnableAlign;
    FormResize(nil); { Recompute constraints }
    {$IFDEF VCL_}
      ClientHeight := Cancel.Top + Cancel.Height + 4;
    {$ELSE VCL_}
      if Visible then
        Height := Cancel.Top + Cancel.Height + 4;
    {$ENDIF}
    NewTop := Screen.Height - Height * 12 div 10;
    if NewTop < Top then Top := NewTop;
  end;
end;

procedure TLogFindDlg.FormResize(Sender: TObject);
begin
  Constraints.MinWidth := Direction.Left + Direction.Width + Width - Cancel.Left + 4 + Width - ClientWidth;
  Constraints.MinHeight := Cancel.Top + Cancel.Height + 6 + Height - ClientHeight;
  Constraints.MaxHeight := 0;
  if IsReplace then begin
    DisableAlign;
    Label2.Top := (Label1.Top + WholeWord.Top) div 2;
    FindText.Height := Label2.Top - FindText.Top - 2;
    ReplaceText.Top := Label2.Top + Label2.Height + 2;
    ReplaceText.Height := WholeWord.Top - ReplaceText.Top - 2;
    ReplaceText.Left := FindText.Left;
    ReplaceText.Width := FindText.Width;
    EnableAlign;
  end
  {$IFDEF LCL_}
    else { without this, height change sometimes ignored }
      Constraints.MaxHeight := Constraints.MinHeight
  {$ENDIF}
    ;
end;

destructor TLogFindDlg.Destroy;
begin
  if Self = LogFindDlg then
    LogFindDlg := nil;
  inherited Destroy;
end;

procedure TLogFindDlg.DoShow;
begin
  inherited DoShow;
  {$IFDEF LCL_}
    { Assignment before Show is ineffective }
    FormResize(nil); { Recompute constraints }
    ClientHeight := Cancel.Top + Cancel.Height + 4;
  {$ENDIF}
end;

end.
