unit MDIChildForm;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface
{ Use TMDIChildForm as a base class for forms that can (optionally) be used as
  a MDIForm or as a MDIChild form, switched by the use of MDIAppStyle;
  Forms designed this way should have FormStyle set to fsMDI... }

uses {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
     {$IFDEF LCL_} LCLType, {$ENDIF}
     Types, Classes, Controls, Forms, FixForm;

type

  TMDIChildForm = class(TFixForm)
  private
    FInitialHeight: Integer;
    FMaximisedHeight: Integer;
    {$IFDEF MSWINDOWS}
    procedure WMMouseActivate(var Message: TMessage); message WM_MOUSEACTIVATE;
    {$ENDIF}
    function GetMDIChildCount: Integer;
    function GetMDIChildren(i: Integer): TForm;
    function GetActiveMDIChild: TForm;
  protected
    procedure Activate; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure Resize; override;
    {$IFNDEF LCL_}
    procedure Resizing(State: TWindowState); override;
    {$ENDIF}
  public
    destructor Destroy; override;
    function CloseQuery: Boolean; override;
    procedure SetActiveForm(Form: TForm);
    property InitialHeight:Integer read FInitialHeight write FInitialHeight;
    property MaximisedHeight: Integer read FMaximisedHeight;
    property ActiveMDIChild: TForm read GetActiveMDIChild;
    property MDIChildCount: Integer read GetMDIChildCount;
    property MDIChildren[i: Integer]: TForm read GetMDIChildren;
  end;

var
  MDIAppStyle: Boolean = {$IFDEF LCL_}False{$ELSE}True{$ENDIF};

var
  MDICaptionSize: Integer; { The Caption size }
  MDIFrameWidth: Integer; { The Frame width }

{ Helper functions with switchable functionality }
function MDIWorkspaceSize(Form: TMDIChildForm):TSize;
procedure SetMdiChildBounds(Form: TCustomForm; ALeft, ATop, AWidth, AHeight: Integer);
function MainForm: TMDIChildForm;

implementation

uses {$IFDEF LCL_} LCLIntf, {$ENDIF}
     MDIFix, Math, WCompat;

function MainForm: TMDIChildForm;
begin
  Result := Application.MainForm as TMDIChildForm;
end;

var
  LastMDIChildForm: TMDIChildForm = nil;

{ TMDIChildForm }

destructor TMDIChildForm.Destroy;
begin
  if LastMDIChildForm = Self then
    LastMDIChildForm := nil;
  inherited;
end;

function TMDIChildForm.GetActiveMDIChild: TForm;
begin
  {$IFNDEF LCL_}
  if MDIAppStyle then
    Result := inherited ActiveMDIChild
  else
  {$ENDIF}
  if (Screen.ActiveForm <> Application.MainForm) and (Screen.ActiveForm is TMDIChildForm) then
    Result := Screen.ActiveForm
  else
    Result := LastMDIChildForm;
end;

function TMDIChildForm.GetMDIChildCount: Integer;
var
  i: Integer;
begin
  {$IFNDEF LCL_}
  if MDIAppStyle then
    Result := inherited MDIChildCount
  else
  {$ENDIF}
  begin
    Result := 0;
    for i := 0 to Screen.FormCount - 1 do
      if (Screen.Forms[i] <> Application.MainForm) and (Screen.Forms[i] is TMDIChildForm) then
        Inc(Result);
  end;
end;

function TMDIChildForm.GetMDIChildren(i: Integer): TForm;
var
  j: Integer;
begin
  {$IFNDEF LCL_}
  if MDIAppStyle then
    Result := inherited MDIChildren[i]
  else
  {$ENDIF}
  begin
    Result := nil;
    for j := 0 to Screen.FormCount - 1 do
      if (Screen.Forms[j] <> Application.MainForm) and (Screen.Forms[j] is TMDIChildForm) then
        if i = 0 then begin
          Result := Screen.Forms[j];
          Exit;
        end else
          Dec(i);
  end;
end;

procedure TMDIChildForm.Resize;
begin
  {$IFDEF LCL_}
  if not Visible and (fsFirstShow in FormState) then
    FInitialHeight := Max(Height,FInitialHeight);
  {$ENDIF}
  { Record Maximized height, as it can be difficult to work out }
  if Height > Screen.Width div 2 then
    FMaximisedHeight := Height;
  inherited;
end;

procedure TMDIChildForm.SetActiveForm(Form: TForm);
begin
  if (ActiveMDIChild <> nil) and (ActiveMDIChild <> Form) then
    {$IFDEF MSWINDOWS}
      SendMessage(ClientHandle, WM_MDIACTIVATE, Form.Handle, 0);
    {$ENDIF}
    {$IFDEF LINUX}
      Form.SetFocus;
    {$ENDIF}
end;

procedure TMDIChildForm.Activate;
begin
  if Self <> Application.MainForm then
    LastMDIChildForm := Self;
  {$IFDEF MSWINDOWS}
  if MDIAppStyle and (FormStyle = fsMDIChild) then
    { MDI bug fix }
    SetWindowLong(Handle, GWL_STYLE,
                  GetWindowLong(Handle, GWL_STYLE) or WS_SYSMENU);
  {$ENDIF}
  inherited;
end;

{$IFDEF MSWINDOWS}

{$IFDEF VCL_}
procedure TMDIChildForm.Resizing(State: TWindowState);
begin
  inherited;
  if not Visible and (fsCreating in FormState) then
    FInitialHeight := Height
end;
{$ENDIF}

procedure TMDIChildForm.WMMouseActivate(var Message: TMessage);
begin
  { MDI bug fix }
  if FormStyle = fsMDIChild then
    {$IFDEF VER80}WinProcs{$ELSE}Windows{$ENDIF}.SetFocus(Handle);
  inherited;
end;

type TOpenCustomForm = class(TCustomForm);
{$ENDIF}

{ Workspace size including scroll bars }
function MDIWorkspaceSize(Form: TMDIChildForm):TSize;
{$IFDEF MSWINDOWS}
var
  Rect: TRect;
  function ScrollBarVisible(fnBar: Integer; Mask: Integer):Boolean;
  var
    {$IFDEF VER80}
    min, max: Integer;
    {$ELSE}
    si: tagSCROLLINFO; // Windows.TScrollInfo;
    {$ENDIF}
  begin
    Result := (GetWindowLong({$IFDEF VCL_}TOpenCustomForm{$ENDIF}(Form).ClientHandle, GWL_STYLE) and Mask) <> 0;
    if Result then begin
      {$IFDEF VER80}
      GetScrollRange(TOpenCustomForm(Form).ClientHandle, fnBar, min, max);
      Result := max > min;
      {$ELSE}
      si.cbSize := SizeOf(si);
      si.fMask := SIF_RANGE or SIF_PAGE;
      GetScrollInfo({$IFDEF VCL_}TOpenCustomForm{$ENDIF}(Form).ClientHandle, fnBar, si);
      Result := (si.nMax > si.nMin) and (si.nMax > si.nMin + Integer(si.nPage));
      {$ENDIF}
    end;
  end;
{$ENDIF}
begin
  if MDIAppStyle then begin
  {$IFDEF MSWINDOWS}
    GetClientRect({$IFDEF VCL_}TOpenCustomForm{$ENDIF}(Form).ClientHandle, Rect);
    Result.cx := Rect.Right;
    Result.cy := Rect.Bottom;
    if ScrollBarVisible(SB_HORZ, WS_HSCROLL) then
      Inc(Result.cy, GetSystemMetrics(SM_CYVSCROLL));
    if ScrollBarVisible(SB_VERT, WS_VSCROLL) then
      Inc(Result.cx, GetSystemMetrics(SM_CXHSCROLL));
  {$ENDIF}
  end else begin
    {$IFDEF UNIX}
    Result.cx := MDIChildWidth(Form);
    {$ELSE}
      {$IFDEF LCL_}
        GetWindowRect(Form.Handle,Rect);
        Result.cx := Rect.Right - Rect.Left;
      {$ELSE}
        Result.cx := Form.Width;
      {$ENDIF}
    {$ENDIF}
    if Form.WindowState = wsMaximized then
      if Form.MaximisedHeight > 0 then
        Result.cy := Form.MaximisedHeight
      else
        Result.cy := Screen.Height
    else
      Result.cy := Form.InitialHeight;
    Dec(Result.cy, Form.Height);
  end;
end;

procedure SetMdiChildBounds(Form: TCustomForm; ALeft, ATop, AWidth, AHeight: Integer);
{$IFNDEF MSWINDOWS}
var
  Rect, MainRect: TRect;
{$ENDIF}
begin
  if MDIAppStyle then begin
    {$IFDEF MSWINDOWS}
      Form.SetBounds(ALeft, ATop, AWidth, AHeight);
    {$ENDIF}
  end else begin
    { Mainform is assumed to be essentially a menu + toolbar only;
      position windows below it }
    {$IFDEF XBASED}
      Rect:= ReadNetFrame(Form);
      MainRect := ReadNetFrame(Application.MainForm);
      with Application.MainForm do
        Form.SetBounds(ALeft+Left+Rect.Left-MainRect.Left,
                       ATop+Top+Height+MainRect.Bottom+Rect.Top,
                       AWidth-Rect.Left-Rect.Right, AHeight-Rect.Top-Rect.Bottom);
    {$ELSE}
      {$IFDEF DARWIN}
        with Application.MainForm do
          Form.SetBounds(ALeft+Left+MDIFrameWidth,
                         ATop+Top+Height+MDICaptionSize,
                         AWidth-2*MDIFrameWidth, AHeight-MDICaptionSize);
      {$ELSE}
        {$IFDEF LCL_}
          with Application.MainForm do
            Form.SetBounds(ALeft+Left,
                           ATop+Top+Height+MDICaptionSize,
                           AWidth-2*MDIFrameWidth, AHeight-MDICaptionSize);
        {$ELSE}
          with Application.MainForm do
          Form.SetBounds(ALeft+Left, ATop+Top+Height, AWidth, AHeight);
        {$ENDIF}
      {$ENDIF}
    {$ENDIF}
  end;
end;

function TMDIChildForm.CloseQuery: Boolean;
var
  i: Integer;
begin
  Result := False;
  if not MDIAppStyle and (Self = MainForm) then
    for i := 0 to MDIChildCount - 1 do
      if not MDIChildren[i].CloseQuery then
        Exit;
  Result := inherited CloseQuery;
end;

procedure TMDIChildForm.CreateParams(var Params: TCreateParams);
begin
  if not MDIAppStyle and (FormStyle in [fsMDIChild, fsMDIForm])  then
    FormStyle := fsNormal;
  { Allow designtime fsMDIChild, but auto switch to fsNormal at runtime if no main window;
    should be overwritten on loading resorce, but seems to create window before loading finished }
  if (FormStyle = fsMDIChild) and not (csDesigning in ComponentState)
     and ( (Application.MainForm = nil) or (Application.MainForm.ClientHandle = 0) ) then begin
    FormStyle := fsNormal;
  end;
  inherited;
end;

initialization
  MDIFrameWidth := GetSystemMetrics(SM_CXSIZEFRAME);
  MDICaptionSize :=  GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYSIZEFRAME);
  if MDICaptionSize = 0 then
    MDICaptionSize := 25; { Unimplmented by widget set - currently Darwin }
end.

