unit MDIFix;
{Updated Actions for MDI Child windows }
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses Forms, StdActns, ActnList, {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Classes;

type

  {$IFDEF LCL_}
  TWindowAction = class(TAction)
    function GetForm(Target: TObject): TForm;
  end;
  {$ENDIF}

  TExWindowAction = class(TWindowAction)
    function HandlesTarget(Target: TObject): Boolean; override;
    procedure UpdateTarget(Target: TObject); override;
  end;

  {$IFDEF LCL_}
  TWindowMinimizeAll = class(TWindowAction)
  end;
  {$ENDIF}

  TExWindowCascade = class(TExWindowAction)
  public
    procedure ExecuteTarget(Target: TObject); override;
    constructor Create(AOwner: TComponent); override;
  end;

  TExWindowTileHorizontal = class(TExWindowAction)
    procedure ExecuteTarget(Target: TObject); override;
    constructor Create(AOwner: TComponent); override;
  end;

  TExWindowTileVertical = class(TExWindowAction)
    procedure ExecuteTarget(Target: TObject); override;
    constructor Create(AOwner: TComponent); override;
  end;

  TExWindowArrange = class(TExWindowAction)
    procedure ExecuteTarget(Target: TObject); override;
    constructor Create(AOwner: TComponent); override;
  end;

function MDIChildWidth(Form: TCustomForm):Integer; { Width of MDI child form including frame }
function MDIChildHeight(Form: TCustomForm):Integer; { Height of MDI child form including frame }

implementation

uses {$IFDEF XBASED} X, XLib, xatom, CompatMsg, {$ENDIF}
     Types, MDIChildForm, SysUtils, WCompat;

function ArrangeMDIIcons(Form: TForm): Integer;
begin
  Result := 0;
end;

procedure TileMDIWindows(AForm: TCustomForm; Vertical: Boolean);
var
  nNormal: Integer;
  ch, rows, cols, row, col: Integer;
  i, w, h, x, y: Integer;
  Size: TSize;
  Form: TMDIChildForm;
begin
  if not (AForm is TMDIChildForm) then
    Exit;
  Form := TMDIChildForm(AForm);
  { Algorithm taken from Wine }
  nNormal := 0;
  for i := 0 to Form.MDIChildCount-1 do begin
    if Form.MDIChildren[i].WindowState = wsMaximized then
      Form.MDIChildren[i].WindowState := wsNormal;
    if Form.MDIChildren[i].WindowState = wsNormal then
      Inc(nNormal);
  end;
  if nNormal = 0 then Exit;
  Size := MDIWorkspaceSize(Form);
  ch := Size.cy;
  if ArrangeMDIIcons(Form) > 0 then
    ch := ch - MDICaptionSize;
  rows := 1;
  while rows*rows <= nNormal do Inc(rows);
  Dec(rows);
  cols := nNormal div rows;
  if not Vertical then begin
    i := rows;
    rows := cols;
    cols := i;
  end;
  i := -1;
  h := Size.cy div rows;
  w := Size.cx div cols;
  x := 0;
  for col := 1 to cols do begin
    if col = cols then begin
      rows := nNormal - rows*(cols-1);
      h := ch div rows;
    end;
    y := 0;
    for row := 1 to rows do begin
      repeat Inc(i)
      until (i >= Form.MDIChildCount) or (Form.MDIChildren[i].WindowState = wsNormal);
      if i >= Form.MDIChildCount then
        Break;
      SetMdiChildBounds(Form.MDIChildren[i], x, y, w, h);
      y := y + h;
    end;
    x := x + w;
  end;
end;

{ TExWindowAction }

function TExWindowAction.HandlesTarget(Target: TObject): Boolean;
begin
  if MDIAppStyle then
    Result := inherited HandlesTarget(Target)
end;

procedure TExWindowAction.UpdateTarget(Target: TObject);
begin
  if MDIAppStyle then
    inherited;
end;

{ TExWindowCascade }

constructor TExWindowCascade.Create(AOwner: TComponent);
begin
  inherited;
  Hint := 'Cascade|Arrange windows in a cascade';
  ImageIndex := 17;
  Caption := '&Cascade';
end;

procedure TExWindowCascade.ExecuteTarget(Target: TObject);
var
  i, h, w, x: Integer;
  nStagger: Integer;
  Size: TSize;
  Form: TMDIChildForm;
begin
  {$IFDEF VCL_}
    if MDIAppStyle then begin
      GetForm(Target).Cascade;
      Exit;
    end;
  {$ENDIF}
  if not (GetForm(Target) is TMDIChildForm) then
    Exit;
  Form := TMDIChildForm(GetForm(Target));
  Size := MDIWorkspaceSize(Form);
  with Form do begin
    if not MDIAppStyle and (MDIChildCount = 1) then begin
      SetMdiChildBounds(MDIChildren[0], 0, 0, Size.cx, Size.cy);
      Exit;
    end;
    h := Size.cy;
    if ArrangeMDIIcons(Form) > 0 then
      h := h - MDICaptionSize;
    nStagger := h div (3*MDICaptionSize);
    h := h - nStagger*MDICaptionSize;
    w := Size.cx - nStagger*MDICaptionSize;
    x := 0;
    for i := 0 to MDIChildCount-1 do begin
      if MDIChildren[i].WindowState = wsMaximized then
        MDIChildren[i].WindowState := wsNormal;
      if MDIChildren[i].WindowState = wsNormal then begin
        SetMdiChildBounds(MDIChildren[i], x, x, w, h);
        x := x + MDICaptionSize;
        if x > MDICaptionSize*nStagger then
          x := 0;
      end;
    end;
  end;
end;

{ TExWindowTileVertical }

constructor TExWindowTileVertical.Create(AOwner: TComponent);
begin
  inherited;
  Hint := 'Stack|Tile windows vertically';
ImageIndex := 16;
  Caption := '&Stack';
end;

procedure TExWindowTileVertical.ExecuteTarget(Target: TObject);
begin
  {$IFDEF VCL_}
    if MDIAppStyle then
      with GetForm(Target) do begin
        TileMode := tbVertical;
        Tile;
      end
    else
  {$ENDIF}
      TileMDIWindows(GetForm(Target), True);
end;

{ TExWindowTileHorizontal }

constructor TExWindowTileHorizontal.Create(AOwner: TComponent);
begin
  inherited;
  Hint := 'Tile|Tile windows horizontally';
  ImageIndex := 15;
  Caption := '&Tile';
end;

procedure TExWindowTileHorizontal.ExecuteTarget(Target: TObject);
begin
  {$IFDEF VCL_}
    if MDIAppStyle then
      with GetForm(Target) do begin
        TileMode := tbHorizontal;
        Tile;
      end
    else
  {$ENDIF}
      TileMDIWindows(GetForm(Target), False);
end;

{ TExWindowArrange }

constructor TExWindowArrange.Create(AOwner: TComponent);
begin
  inherited;
  Hint := 'Arrange Icons';
  Caption := '&Arrange';
end;

procedure TExWindowArrange.ExecuteTarget(Target: TObject);
begin
  {$IFDEF VCL_}
    if MDIAppStyle then
      GetForm(Target).ArrangeIcons
    else
  {$ENDIF}
      ArrangeMDIIcons(GetForm(Target));
end;

function MDIChildWidth(Form: TCustomForm):Integer; { Width of MDI child form including frame }
{$IFDEF XBASED}
var
  Rect: TRect;
{$ENDIF}
begin
{$IFDEF XBASED}
  Rect := ReadNetFrame(Form);
  Result := Form.Width + Rect.Left + Rect.Right;
{$ELSE}
  Result := Form.Width;
{$ENDIF}
end;

function MDIChildHeight(Form: TCustomForm):Integer; { Height of MDI child form including frame }
{$IFDEF XBASED}
var
  Rect: TRect;
{$ENDIF}
begin
{$IFDEF XBASED}
  Rect := ReadNetFrame(Form);
  Result := Form.Height + Rect.Top + Rect.Bottom;
{$ELSE}
  Result := Form.Height;
{$ENDIF}
end;

{$IFDEF LCL_}

{ TWindowAction }

function TWindowAction.GetForm(Target: TObject): TForm;
begin
  Result := Target as TForm;
end;

{$ENDIF}

end.
