unit NewSpin;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, {$IFNDEF LCL_}{$DEFINE MUSCROLL} {$ENDIF}{$ENDIF}
  {$IFNDEF CLX_} Messages, Graphics, Controls, Forms, Dialogs, StdCtrls, ComCtrls,
  {$ELSE} QGraphics, QControls, QForms, QDialogs, QStdCtrls, Types, {$ENDIF}
  {$IFDEF MUSCROLL} MuScroll, {$ENDIF}
  SysUtils, Classes, EditExtr;

type

  TOnIncrementInteger = procedure (Sender: TObject; var NewValue:Longint; LastValue:Longint; n:Integer) of object;
  TOnGetIntegerValue =  procedure (Sender: TObject; var NewValue:Longint; TheText: string; var StringValid: Boolean) of object;
  TOnSetIntegerValue =  procedure (Sender: TObject; NewValue:Longint; var NewText: string) of object;

  TCustomSpinBox = class(TCustomEditExtra)
  private
    {$IFDEF MUSCROLL}
    MicroScroll: TMicroScroll;
    procedure MicroScrollDownClick(Sender: TObject);
    procedure MicroScrollUpClick(Sender: TObject);
    {$ELSE}
    UpDown: TUpDown; { Not useable under windows, as limits make use awkward }
    procedure UpDownClick(Sender: TObject; Button: TUDBtnType);
    {$ENDIF}
  protected
    FLastErr: string;
    function SetAsText(s:string):Boolean;
    function CreateExtraControl:TControl; override;
    {$IFNDEF VER80}
    function DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
      {$IFDEF CLX_} const {$ENDIF} MousePos: TPoint): Boolean; override;
    {$ENDIF}
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor Create(AOwner: TComponent); override;
    function CheckCurrentValue(Report:Boolean; NullOK: Boolean): Boolean;
    function CheckText(s:string; NullOK: Boolean):Boolean; virtual; abstract;
    procedure ChangeClick(n:LongInt);
    function IncrementBy(n: LongInt):Boolean; virtual; abstract;
    property LastErr: string read FLastErr;
  published
    property Text stored False;
    property ExtraWidth default 15;
  end;

  { TSpinBox }

  TSpinBox = class(TCustomSpinBox)
  private
    FIncrement: LongInt;
    FOnIncrement: TOnIncrementInteger;
    FMaxValue: LongInt;
    FMinValue: LongInt;
    LastValue: LongInt;
    procedure SetValue(const AValue: LongInt);
    function GetValue: LongInt;
  protected
    procedure DoGetIntegerValue(var NewValue:Longint; TheText: string; var StringValid: Boolean); virtual;
    procedure DoSetIntegerValue(NewValue: Integer; var NewText: string); virtual;
  public
    FOnGetIntegerValue: TOnGetIntegerValue;
    FOnSetIntegerValue: TOnSetIntegerValue;
    constructor Create(AOwner: TComponent); override;
    function CheckText(s:string; NullOK: Boolean):Boolean; override;
    function IncrementBy(n: LongInt):Boolean; override;
    function ValueAB: LongInt;
  published
    property MinValue: LongInt read FMinValue write FMinValue default 0;
    property MaxValue: LongInt read FMaxValue write FMaxValue default 0;
    property Value: LongInt read GetValue write SetValue default 0;
    property Increment: LongInt read FIncrement write FIncrement default 1;
    property OnIncrement: TOnIncrementInteger read FOnIncrement write FOnIncrement;
    property OnGetIntegerValue: TOnGetIntegerValue read FOnGetIntegerValue write FOnGetIntegerValue;
    property OnSetIntegerValue: TOnSetIntegerValue read FOnSetIntegerValue write FOnSetIntegerValue;
  end;

  TOnIncrementFixed = procedure (Sender: TObject; var NewValue: Double; LastValue:Double; n:Integer) of object;

  { TFixedSpin }

  TFixedSpin = class(TCustomSpinBox)
  private
    FMinValue: Double;
    FMaxValue: Double;
    FIncrement: Double;
    LastValue: Double;
    FPlaces: Integer;
    FOnIncrement: TOnIncrementFixed;
    function GetValue: Double;
    procedure SetPlaces(NewPlaces: Integer);
    procedure SetValue(const NewValue: Double);
  public
    constructor Create(AOwner: TComponent); override;
    function CheckText(s:string; NullOK: Boolean):Boolean; override;
    function IncrementBy(n: LongInt):Boolean; override;
    function ValueAB: Double;
  published
    property Increment: Double read FIncrement write FIncrement;
    property MaxValue: Double read FMaxValue write FMaxValue;
    property MinValue: Double read FMinValue write FMinValue;
    property Places: Integer read FPlaces write SetPlaces default 2;
    property Value: Double read GetValue write SetValue;
    property OnIncrement: TOnIncrementFixed read FOnIncrement write FOnIncrement;
  end;

{ Helper function to use in OnIncrement handler }
function SpinScaleBy(n: Integer; ScaleFactor, LastValue: Double): Double;

implementation

uses WCompat, {$IFDEF LCL_} LCLType, {$ENDIF} MathSup;

{ TCustomSpinBox }

constructor TCustomSpinBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {$IFDEF SUBFRAME}with Edit do{$ENDIF} begin
    Alignment := taRightJustify;
    MaxLength := 50;
    {$IFDEF SUBFRAME}
    OnKeyDown := EditKeyDown;
    OnMouseWheel := EditMouseWheel;
    {$ENDIF}
  end;
end;

{$IFNDEF MUSCROLL}
procedure TCustomSpinBox.UpDownClick(Sender: TObject; Button: TUDBtnType);
begin
  if Button = btNext then
    ChangeClick(1)
  else
    ChangeClick(-1);
  if TabStop and (Screen.ActiveControl <> Self) and CanFocus then
    SetFocus;
end;
{$ENDIF}

function TCustomSpinBox.SetAsText(s: string): Boolean;
var
  SelEnd: Integer;
begin
  Result := CheckText(s, False);
  if Result then {$IFDEF SUBFRAME}with Edit do{$ENDIF}
    if HandleAllocated then begin { CMW }
      SelEnd := SelStart + SelLength;
      if Text <> s then begin
        Text := s;
        {$IFDEF LCLcarbon}
        Change; { Bug work around - change not triggered }
        {$ENDIF}
      end;
      SelStart := SelEnd;
    end else
      Text := s;
end;

{$IFDEF MUSCROLL}
procedure TCustomSpinBox.MicroScrollDownClick(Sender: TObject);
begin
  ChangeClick(-1);
end;

procedure TCustomSpinBox.MicroScrollUpClick(Sender: TObject);
begin
  ChangeClick(1);
end;
{$ENDIF}

procedure TCustomSpinBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_UP:    ChangeClick(1);
    VK_DOWN:  ChangeClick(-1);
    VK_PRIOR: ChangeClick(10);
    VK_NEXT:  ChangeClick(-10);
  else
    inherited KeyDown(Key, Shift);
    Exit;
  end;
  Key := 0;
end;

procedure TCustomSpinBox.ChangeClick(n: LongInt);
begin
  if not CheckText({$IFDEF SUBFRAME}Edit.{$ENDIF}Text, False) then
    n := 0;
  if IncrementBy(n) then
    {$IFDEF SUBFRAME}Edit.{$ENDIF}Modified := True;
end;

function TCustomSpinBox.CheckCurrentValue(Report:Boolean; NullOK: Boolean): Boolean;
begin
  Color := clWindow;
  Result := not CheckText(Text, NullOK);
  if Result then begin
    Color := clRed;
    if (csDesigning in ComponentState) or Report then
      MessageDlg(Lasterr, mtWarning, [mbOk], 0);
  end
end;

{$IFNDEF VER80}
function TCustomSpinBox.DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
   {$IFDEF CLX_} const {$ENDIF} MousePos: TPoint): Boolean;
begin
  ChangeClick(WheelDelta div WHEEL_DELTA);
  Result := True;
end;
{$ENDIF}

function TCustomSpinBox.CreateExtraControl:TControl;
begin
  {$IFDEF MUSCROLL}
  MicroScroll := TMicroScroll.Create(Self);
  {$IFDEF SUBFRAME}
  MicroScroll.Parent := Self;
  {$ENDIF}
  with MicroScroll do begin
    {$IFDEF SUBFRAME}
    Align := alRight;
    {$ENDIF}
    Width := 15;
    TabStop := False;
    OnDownClick := MicroScrollDownClick;
    OnUpClick := MicroScrollUpClick;
  end;
  MicroScroll.FocusControl := {$IFDEF SUBFRAME}Edit{$ELSE}Self{$ENDIF};
  Result := MicroScroll;
  {$ELSE}
  UpDown := TUpDown.Create(Self);
  with UpDown do begin
    Width := 15;
    TabStop := False;
    OnClick := {$IFDEF FPC}@{$ENDIF}UpDownClick;
    Wrap := True; { Disable limit checks }
  end;
  Result := UpDown;
  {$ENDIF}
end;

{ TSpinBox }

constructor TSpinBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMinValue := 0;
  FMaxValue := 0;
  FIncrement := 1;
  Value := 0;
end;

procedure TSpinBox.DoGetIntegerValue(var NewValue: Longint; TheText: string;
  var StringValid: Boolean);
begin
  if Assigned(FOnGetIntegerValue) and not (csDesigning in ComponentState) then
    FOnGetIntegerValue(Self, LastValue, TheText, StringValid);
end;

procedure TSpinBox.DoSetIntegerValue(NewValue: Integer; var NewText: string);
begin
  if Assigned(FOnSetIntegerValue) then
    FOnSetIntegerValue(Self, NewValue, NewText);
end;

function TSpinBox.CheckText(s: string; NullOK: Boolean): Boolean;
begin
  Result := NullOK;
  s := Trim(s);
  if (s = '') or (s = '-') or (s = '.') or (s = '-.') then
    LastValue := 0
  else
    Result := TryStrToInt(s, LastValue);
  DoGetIntegerValue(LastValue, s, Result);
  if not Result then begin
    LastValue := FMinValue;
    FLastErr := Format('Bad Number: %s', [s]);
  end;
  if Result and (FMaxValue > FMinValue) then
    if (LastValue < FMinValue) or (LastValue > FMaxValue) then begin
      Result := False;
      if LastValue < FMinValue then
        LastValue := FMinValue
      else
        LastValue := FMaxValue;
      FLastErr := Format('%s out of range (%d to %d)', [s, FMinValue, FMaxValue]);
    end;
end;

function TSpinBox.IncrementBy(n: LongInt): Boolean;
var
  NewValue, PrevValue: Longint;
begin
  PrevValue := LastValue;
  NewValue := LastValue+n*FIncrement;
  if FMaxValue > FMinValue then
    if NewValue < FMinValue then
      NewValue := FMinValue
    else if NewValue > FMaxValue then
      NewValue := FMaxValue;
  if Assigned(FOnIncrement) then FOnIncrement(Self, NewValue, LastValue, n);
  FLastErr := '';
  SetValue(NewValue);
  if LastErr <> '' then
    SetValue(LastValue);
  Result := PrevValue <> LastValue;
end;

function TSpinBox.ValueAB: LongInt;
begin
  { Aborts on bad value, and leaves window red }
  { Consider replacing GetValue with this }
  if CheckCurrentValue(False, False) then
    raise EAbort.Create(LastErr)
  else
    Result := LastValue;
end;

function TSpinBox.GetValue: LongInt;
begin
  if CheckCurrentValue(True, True) then
    Value := LastValue;
  Result := LastValue;
end;

procedure TSpinBox.SetValue(const AValue: LongInt);
var
  NewText: string;
begin
  NewText := IntToStr(AValue);
  DoSetIntegerValue(AValue, NewText);
  SetAsText(NewText);
end;

{ TFixedSpin }

constructor TFixedSpin.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {$IFDEF SUBFRAME}Edit.{$ENDIF}Text := '0.00';
  FIncrement := 0.01;
  FPlaces := 2;
end;

function TFixedSpin.CheckText(s: string; NullOK: Boolean): Boolean;
begin
  Result := NullOK;
  s := Trim(s);
  if (s = '') or (s = '-') or (s = '.') or (s = '-.') then
    LastValue := 0
  else
    Result := TryStrToFloatEx(s, LastValue);
  if not Result then begin
    LastValue := FMinValue;
    FLastErr := Format('Bad Number: %s', [Text]);
  end;
  if Result and (FMaxValue > FMinValue) then
    if (LastValue < FMinValue) or (LastValue > FMaxValue) then begin
      Result := False;
      if LastValue < FMinValue then
        LastValue := FMinValue
      else
        LastValue := FMaxValue;
      FLastErr := Format('%s out of range (%g to %g)', [Text, FMinValue, FMaxValue]);
    end;
end;

function TFixedSpin.ValueAB: Double;
begin
  if CheckCurrentValue(False, False) then
    raise EAbort.Create(LastErr)
  else
    Result := LastValue;
end;

function TFixedSpin.GetValue: Double;
begin
  if CheckCurrentValue(True, True) then
    Value := LastValue;
  Result := LastValue
end;

function TFixedSpin.IncrementBy(n: LongInt): Boolean;
var
  NewValue, PrevValue: Double;
begin
  PrevValue := LastValue;
  NewValue := LastValue+n*FIncrement;
  if Assigned(FOnIncrement) then FOnIncrement(Self, NewValue, LastValue, n);
  FLastErr := '';
  SetValue(NewValue);
  if LastErr <> '' then
    SetValue(LastValue);
  Result := PrevValue <> LastValue;
end;

procedure TFixedSpin.SetPlaces(NewPlaces: Integer);
begin
  if NewPlaces <> Places then begin
    FPlaces := NewPlaces;
    CheckText(Text, True); { Notionally SetValue(Value), but without exceptions on loading }
    SetValue(LastValue);
    if not (csLoading in ComponentState) then begin
      FIncrement := 1;
      if NewPlaces > 18 then NewPlaces := 18;
      while NewPlaces > 0 do begin
        FIncrement := FIncrement / 10;
        Dec(NewPlaces);
      end;
    end;
  end;
end;

procedure TFixedSpin.SetValue(const NewValue: Double);
begin
  if Places >= 0 then
    SetAsText(FloatToStrF(NewValue, ffFixed, 15, Places))
  else
    SetAsText(DtosM(NewValue, {$IFDEF SUBFRAME}Edit.{$ENDIF}nVisibleDigits-1, -1));
end;

function SpinScaleBy(n: Integer; ScaleFactor, LastValue: Double): Double;
begin
  Result := LastValue;
  try
    while (n > 0) do begin
      Result := Result * ScaleFactor;
      Dec(n);
    end;
    while n < 0 do begin
      Result := Result / ScaleFactor;
      Inc(n);
    end;
  except
  end;
end;

end.
