{ Loosely based on Z_Prof tool by A. Baars (antonie.baars@wxs.nl) }

unit ProfileForm;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Messages, Classes, Graphics, Controls, Forms, Dialogs,
     Grids, Buttons, ExtCtrls, StdCtrls, NGrid, GridEx, ComCtrls;

type

  { TProfilerForm }

  TProfilerForm = class(TForm)
    Clearbutton: TButton;
    Panel: TPanel;
    RadioGroup: TRadioGroup;
    Clock: TLabel;
    OverheadE: TLabel;
    NG: TNumberGrid;
    Label1: TLabel;
    Label2: TLabel;
    OverheadI: TLabel;
    Label4: TLabel;
    RecalibrateButton: TButton;
    TabControl: TTabControl;
    Method: TComboBox;
    MethodLabel: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure ClearbuttonClick(Sender: TObject);
    procedure RadioGroupClick(Sender: TObject);
    procedure Recalibrate(Sender: TObject);
    procedure NGGetCellLabel(Sender: TObject; Col, Row: Integer; var s: String; var IsText: Boolean);
    procedure TabControlChange(Sender: TObject);
    procedure MethodChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure NGDrawCellContents(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
  public
    destructor Destroy; override;
    procedure UpdateActions; override;
    procedure PopulateTable;
    procedure MenuHandler(Sender: TObject);
  private
    CPUclock: Double;
  end;

  function ProfilerForm: TProfilerForm;

implementation

uses Clock, Math, MathSup, SysUtils, Profile, UCompat;

{$IFNDEF LCL_}
{$R *.DFM}
{$ELSE}
{$R *.lfm}
{$ENDIF}

var
  TheProfilerForm: TProfilerForm;

function ProfilerForm: TProfilerForm;
begin
  if not Assigned(TheProfilerForm) then
    Result := TProfilerForm.Create(Application)
  else
    Result := TheProfilerForm;
end;

procedure TProfilerForm.FormCreate(Sender: TObject);
var
  i: Integer;
  s: string;
begin
  if not Assigned(TheProfilerForm) then
    TheProfilerForm := Self;
  Method.Items.Clear;
  i := 0;
  repeat
    s := TimingMethodName(i);
    if s <> '' then
      Method.Items.Add(s);
    Inc(i);
  until s = '';
  Method.ItemIndex := GetTimingMethod;
  Recalibrate(Self);
end;

procedure TProfilerForm.FormShow(Sender: TObject);
begin
  PopulateTable;
end;

procedure TProfilerForm.MethodChange(Sender: TObject);
begin
  if Method.ItemIndex = GetTimingMethod then
    Exit;
  ClearbuttonClick(nil);
  SetTimingMethod(Method.ItemIndex);
  Recalibrate(nil);
end;

procedure TProfilerForm.ClearbuttonClick(Sender: TObject);
begin
  ClearTiming;
  MarkLevel := 0;
  PopulateTable;
end;

procedure TProfilerForm.RadioGroupClick(Sender: TObject);
begin
  PopulateTable;
end;

procedure TProfilerForm.Recalibrate(Sender: TObject);
  function FormatTime(Cycles: Double): string;
  var
    o: Double;
  begin
    o := Cycles/CPUclock*1e6;
    if o < 0.9 then
      Result := Format('%f ns, %f Cycles', [o*1e3, Cycles])
    else
      Result := Format('%f s, %f Cycles', [o, Cycles]);
  end;
begin
  CPUclock := ClockFrequency;
  if CPUclock >= 1e6 then
    Clock.caption:=Format('Clock: %f MHz', [CPUclock/1e6])
  else
    Clock.caption:=Format('Clock: %f Hz', [CPUclock]);
  CalculateOverhead(500);
  OverheadI.Caption := FormatTime(OverheadCycles);
  OverheadE.Caption := FormatTime(ExternalOverheadCycles);
  PopulateTable;
end;

procedure TProfilerForm.TabControlChange(Sender: TObject);
begin
  if not (csDestroying in ComponentState) then
    PopulateTable;
end;

destructor TProfilerForm.Destroy;
begin
  if TheProfilerForm = Self then
    TheProfilerForm := nil;
  inherited;
end;

procedure TProfilerForm.UpdateActions;
begin
  inherited;
  if TimeChanged then begin
    TimeChanged := False;
    PopulateTable;
  end;
end;

procedure TProfilerForm.PopulateTable;
var
  Units: string;
  Divisor: Double;
  Digits, i, j, EffectiveTabIndex: Integer;
  procedure SetCol(Col, Dig: Integer; V: Double);
  begin
//    V := RoundTo(V, -Dig);
    NG.Value[i,Col-1] := V;
//    Dig := NG.Canvas.TextWidth(DTosM(V, -1, Dig))+10;
//    if NG.ColWidths[Col-1] < Dig then
//      Ng.ColWidths[Col-1] := Dig;
  end;
begin
  EffectiveTabIndex := TabControl.TabIndex;
  ComputeTotalTimings;
  j := Length(ThreadNames);
  if j = 2 then begin
    j := 1; { If only one thread, only show totals }
    EffectiveTabIndex := 1;
  end;
  with TabControl.Tabs do begin
    while Count < j do Add('');
    while Count > j do Delete(Count-1);
    for i := 0 to Count-1 do
      Strings[i] := ThreadNames[i];
  end;
  for i := 0 to High(TimingNames) do
    if TimingNames[i] = '' then begin
      NG.RowCount := Max(i+1,2);
      Break;
    end;
  Digits := 2;
  case RadioGroup.ItemIndex of
    0: begin Units := '/s';  Divisor :=1/CPUclock;   end;
    1: begin Units := '/ms'; Divisor :=1e3/CPUclock; end;
    2: begin Units := {$IFDEF VCL_}'/s'{$ELSE}'/1e-6s'{$ENDIF}; Divisor :=1e6/CPUclock; end;
    3: begin Units := '';    Divisor := 1; Digits := 0; end;
  end;
  with NG do begin
    if IsTrackingMemory then begin
      ColCount := 8;
      Title[5] := 'Allocs';
      Title[6] := 'Frees';
    end else
      ColCount := 6;
    Title[0] := 'Passes';
    Title[1] := 'Time' + Units;
    Title[2] := 'Mean' + Units;
    Title[3] := 'Last' + Units;
    Title[4] := 'Fastest'+ Units;
    Title[ColCount-1] := 'Routine';
    for i := 0 to ColCount-1 do
      AutoAdjustColumn(i);
    for i := 1 to Min(Length(TimingNames), RowCount-1) do
      with Marks[EffectiveTabIndex][i-1] do begin
        if Passes > 0 then begin
          SetCol(1, 0, Passes);
          SetCol(2, Digits, TotalTime*Divisor);
          SetCol(3, Digits, TotalTime/Passes*Divisor);
          SetCol(4, Digits, PrevTime*Divisor);
          SetCol(5, Digits, Fastest*Divisor);
          if IsTrackingMemory then begin
            SetCol(6, 0, Allocs);
            SetCol(7, 0, Frees);
          end;
        end else
          for j := 0 to ColCount-1 do
            Valid[i,j] := False;
      end;
    AutoAdjustColumn(ColCount-1,False,True);
    AutoDefaultRowHeight;
    Invalidate;
  end;
  Constraints.MinWidth := OverheadE.Left + OverheadE.Width + 6;
end;

procedure TProfilerForm.MenuHandler(Sender: TObject);
begin
  Show;
end;

procedure TProfilerForm.NGDrawCellContents(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
begin
  if ARow > 0 then
    if ACol = 0 then
      NG.CurrentCellAlignment := taRightJustify
//Exit;
//  Drawn := True;
//  if (ARow > 0) and (ARow <= Length(TimingNames)) then
//    if ACol = NG.ColCount-1 then
//      TextInGrid(NG.Canvas, Rect, TimingNames[ARow-1])
//    else if NG.Valid[ARow,ACol+1] then
//      TextInGrid(NG.Canvas, Rect, DtosM(NG.Value[ARow,ACol+1], -1, -1), taRightJustify)
//    else
//      Drawn := False
//  else
//    Drawn := False;
end;

procedure TProfilerForm.NGGetCellLabel(Sender: TObject; Col, Row: Integer; var s: String; var IsText: Boolean);
begin
  if (Col = NG.ColCount-1) and (Row > 0) then begin
    IsText := True;
    s := TimingNames[Row-1];
  end;
end;

initialization
  ShowTimingAtEnd := False;
end.
