unit ProjAbout;
{ Automatic version number control. This is based on TProject by
  T S Eriksen, though almost all of the code has been replaced by
  writing text to a .rc file. }
{$I utils.inc}{ Copyright (C) 2005, 2016 C M Western; see PcontrolsReg.pas for license}

interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  StdCtrls, Buttons, Controls, ExtCtrls, Forms, Classes;

type
  TFileType = (ftProgram,ftLibrary);

{CMW}

  { TAboutBox }

  TAboutBox = class(TForm)
    Panel: TPanel;
    ProgramIcon: TImage;
    ProductName: TLabel;
    Description: TLabel;
    Details: TLabel;
    OKButton: TButton;
    procedure FormShow(Sender: TObject);
  end;

  TProject = class(TComponent)
  private
    { user defined properties }
    FResFile        : string; { Resourcefile name }
    FCompany        : string; { stored in Ini-file }
    FCopyright      : string; { stored in Ini-file }
    FComment        : string; { default : <comments about this file/product> }
    FDescription    : string; { default : <Description of file> }
    FSpecialBuild   : string; { default : <How this build differs from a standard build> }
    FPrivateBuild   : string; { default : <Private information> }
    FProductname    : string; { default : <name of product this file ships with> }
    FProductVersion : string; { default : <0.0.0.0> }
    FOriginalName   : string; { default : <name of .exe file > }
    FTrademarks     : string; { default : <Legal trademarks used in this build> }
    FLanguage       : string; { default : UK_English }
    FCharset        : string; { default : Multilinqual }
    FFileType       : TFileType; { default : ffProgram }
    { Auto defined properties }
    { Internalname, }         { TProject.name }
    { Fileversion , }         { TProject.version + TProject.build }
    { FileFlags }             { ffprerease when version < 1.0, special/private build }
    { property data buffers etc }
    FAutoNaming   : Boolean;  { Styrer automatisk navngivning af project }
    FBuild        : Integer;  { build nr, incr. ved component write state}
    FVersion      : string;   { Version number }
    FBuildDate    : string;
    FAutoIncrement: Boolean;  { Increment build number automatically }
    FProjectPath  : string;   { Not stored - just used to show path }
  protected
    FirstWrite: Boolean;
    procedure   Loaded; override;
    procedure   WriteState(Writer:TWriter); override;
  public
    procedure   Execute(VersionInfo: string = '');
    procedure   WriteRC(FileName: string);
    function    CheckFile(FileName: string):Boolean;
    constructor Create(AOwner: TComponent); override;
  published
    property AutoNaming    : Boolean   read FAutoNaming    write FAutoNaming;
    property AutoIncrement : Boolean   read FAutoIncrement write FAutoIncrement;
    property Build         : Integer   read FBuild         write FBuild default 0;
    property Version       : string    read FVersion       write FVersion;
    { version resource properties }
    property ProjectPath   : string read FProjectPath      write FProjectPath stored False;
    property Company       : string read FCompany          write FCompany;
    property Copyright     : string read FCopyright        write FCopyright;
    property Comment       : string read FComment          write FComment;
    property Description   : string read FDescription      write FDescription;
    property SpecialBuild  : string read FSpecialBuild     write FSpecialBuild;
    property PrivateBuild  : string read FPrivateBuild     write FPrivateBuild;
    property Productname   : string read FProductName      write FProductname;
    property ProductVersion: string read FProductVersion   write FProductVersion;
    property OriginalName  : string read FOriginalName     write FOriginalName;
    property Trademarks    : string read FTrademarks       write FTrademarks;
    property Language      : string read FLanguage         write FLanguage;
    property Charset       : string read FCharset          write FCharset;
    property FileType      : TFileType read FFileType      write FFileType;
    property ResourceFile  : string    read FResFile       write FResFile;
    property BuildDate     : string    read FBuildDate     write FBuildDate;
  end;

implementation

Uses
  Math, SysUtils, StrUtils, Sysinfo, UCompat, Graphics, Types;

{CMW}
{$IFNDEF LCL_}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function CheckStr(const s:string):string;
begin
  if (s = '') or (s[1] = '<') then
    Result := ''
  else
    Result := s;
end;

function TProject.CheckFile(FileName: string):Boolean;
var
  S: TMemoryStream;
begin
  Result := False;
  DoDirSeparators(FileName);
  { As a safety check, require the resource file to exist, and to contain the original filename string }
  if not FileExists(FileName) then
    Exit;
  S := TMemoryStream.Create;
  try
    UnlockedLoadFromFile(S, FileName);
    if Pos(OriginalName, PAnsiChar(S.Memory)) < 1 then
      Exit;
  finally
    S.Free;
  end;
  Result := True;
end;

constructor TProject.Create(AOwner: TComponent);
begin
  inherited;
  FileType      := ftProgram;
  ResourceFile  := 'Version.rc';
  Version       := '0.1';
end;

procedure TProject.Execute(VersionInfo: string); {CMW}
var
  About:TAboutBox;
  ProdName:string;
  procedure Add(s: string);
  var
    c: string;
    {$IFDEF FPC}
    i: Integer;
    {$ENDIF}
  begin
    if s <> '' then begin
      {$IFDEF FPC}
      {Strip UTF8 characters for the time being}
      for i := 1 to Length(s) do
        if Ord(S[i]) > 128 then
          s[i] := ' ';
      {$ENDIF}
      c := About.Details.Caption;
      if c <> '' then
        About.Details.Caption := Concat(c, LineEnd, s)
      else
        About.Details.Caption := s;
    end;
  end;
var
  Icon: TIcon;
  wh: Integer;
begin
  About := TAboutBox.Create(Application);
  try
    ProdName := CheckStr(ProductName);
    if ProdName = '' then
      ProdName := Application.Title;
    About.ProductName.Caption := ProdName;
    if not Assigned(Application.Icon) or Application.Icon.Empty then begin
      if Assigned(Application.MainForm) and Assigned(TForm(Application.MainForm).Icon) and not TForm(Application.MainForm).Icon.Empty then
      Icon := TForm(Application.MainForm).Icon;
    end else
      Icon := Application.Icon;
    About.ProgramIcon.Picture.Icon := Icon;
    {$IFDEF FPC}
    with About.ProgramIcon do begin
      wh := Min(Width,Height);
      with Picture.Icon do
        Current := GetBestIndexForSize(Size(wh,wh));
    end;
    {$ENDIF}
    if VersionInfo = '' then
      VersionInfo := Concat(Version, '.', IntToStr(Build), ' ', BuildDate);
    About.Description.Caption := Concat(CheckStr(Description), LineEnd, VersionInfo);
    About.Details.Caption := '';
    Add(CheckStr(Comment));
    Add(CheckStr(Copyright));
    Add(CheckStr(Company));
    Add(Concat(LineEnd, SysInfo.DetermineCpu(True)));
    About.ShowModal;
  finally
    About.Free;
  end;
end;

{$IFNDEF MSWINDOWS}
const
  VS_VERSION_INFO = 1;
  VOS__WINDOWS32 = $00000004;
  VOS_NT = $40000;
  VFT_APP = $1;
  VFT_DLL = $2;
{$ENDIF}

procedure TProject.WriteRC(FileName: string);
var
  F: Text;
begin
  AssignFile(F, FileName);
  Rewrite(F);
  try
    {$IFDEF FPC}
    SetTextLineEnding(F,#13#10);
    {$ENDIF}
    WriteLn(F,VS_VERSION_INFO, ' VERSIONINFO');
    WriteLn(F,'FILEVERSION ', ReplaceStr(Version,'.',','), ',', Build, ',0');
    if ProductVersion = '' then
      WriteLn(F,'PRODUCTVERSION ', ReplaceStr(Version,'.',','), ',', Build, ',0')
    else
      WriteLn(F,'PRODUCTVERSION ', ProductVersion);
    WriteLn(F,'FILEOS 0x', IntToHex(VOS__WINDOWS32 or VOS_NT,8));
    case FileType of
      ftProgram: WriteLn(F,'FILETYPE ', VFT_APP);
      ftLibrary: WriteLn(F,'FILETYPE ', VFT_DLL);
    end;
    WriteLn(F,'{');
    WriteLn(F,' BLOCK "StringFileInfo"');
    WriteLn(F,' {');
    WriteLn(F,'  BLOCK "040904E4"'); { US English, Multilingual }
    WriteLn(F,'  {');
    WriteLn(F,'   VALUE "CompanyName", "', Company, '"');
    WriteLn(F,'   VALUE "Comments", "', Comment, '"');
    WriteLn(F,'   VALUE "FileDescription", "', Description, '"');
    WriteLn(F,'   VALUE "FileVersion", "', Version + '.' + IntToStr(Build), '"');
    WriteLn(F,'   VALUE "LegalCopyright", "', Copyright, '"');
    WriteLn(F,'   VALUE "OriginalFilename", "', OriginalName, '"');
    WriteLn(F,'   VALUE "ProductName", "', Productname, '"');
    if ProductVersion = '' then
      WriteLn(F,'   VALUE "ProductVersion", "', Version + '.' + IntToStr(Build), '"')
    else
      WriteLn(F,'   VALUE "ProductVersion", "', ProductVersion, '"');
    { PrivateBuild and SpecialBuild ignored for now }
    WriteLn(F,'  }');
    WriteLn(F,' }');
    WriteLn(F,' BLOCK "VarFileInfo"');
    WriteLn(F,' {');
    WriteLn(F,'  VALUE "Translation", 0x0409, 0x04E4'); { US English, Multilingual }
    WriteLn(F,' }');
    WriteLn(F,'}');
  finally
    Close(F);
  end;
end;

procedure TProject.WriteState(Writer:TWriter);
var
  s: string;
  OK: Boolean;
begin
  { When something on a form has changed all components (on it) will
    be saved when next compiled - before that save, we'll just
    increase the Build number }
  if csDesigning in ComponentState then begin
    if AutoIncrement and not FirstWrite then
      Inc(FBuild);
    FirstWrite := False;
    DateTimeToString(FBuildDate, 'd mmm yyyy h:nn', Now);
    s := ProjectPath+PathDelim+ResourceFile;
    OK := CheckFile(s);
    if not OK then begin
      GetDir(0, s);
      if s <> ProjectPath then begin
        ProjectPath := s;
        s := ProjectPath+PathDelim+ResourceFile;
        OK := CheckFile(s);
      end
    end;
    if OK then    
      WriteRC(s);
  end;
  Inherited WriteState(Writer);
end;

procedure TProject.Loaded;
var
  s : string;
  a : integer;
begin
  inherited Loaded;
  If (csDesigning in ComponentState) then begin
    { Denne blok udfres kun ved
      1) Load af projekt
      2) Visning af form/modul hvor TProject er lagret  }
    if AutoIncrement then
      Inc(FBuild);
    FirstWrite := True;
    GetDir(0, FProjectPath);
    FProjectPath := ExpandFileName(FProjectPath);
  end else begin
    { Denne blok udfrers kun ved
      1) Start af applikation }
    If (AutoNaming) then begin
      Application.Title := Name+' ver. '+Version+'.'+
                           IntToStr(Build);
      s := Application.title;
      for a:=1 to length(s) do
        if s[a]='_' then
          s[a]:=' ';
      Application.title := s;
    end;
  end;
end;

{ TAboutBox }

procedure TAboutBox.FormShow(Sender: TObject);
var
  Extra: Integer;
begin
  Description.Width := Panel.Width - Description.Left - 6;
  Description.AutoSize := True;
  Details.Top := Max(ProgramIcon.Top+ProgramIcon.Height, Description.Top+Description.Height)+5;
  Details.Anchors := [akLeft, akTop];
  {$IFNDEF FPC}
    Details.AutoSize := True;
  {$ENDIF}
  Extra := Details.Top + Details.Height + 8 - Panel.Height;
  if Extra > 0 then
    Height := Height + Extra;
  Details.Anchors := Details.Anchors + [akBottom];
end;

end.


