unit QuantumSpin;
{$I utils.inc}{ Copyright (C) 2007 C M Western; see PcontrolsReg.pas for license}
interface

uses Classes, NewSpin, QuantumNumber;

type

  { TQuantumSpin }

  TQuantumSpin = class (TSpinBox)
  private
    FFormat: TQuantumNumberFormat;
    FUseGlobalFormat: Boolean;
    function GetValue: TQuantumNumber;
    procedure SetFormat(const AValue: TQuantumNumberFormat);
    procedure SetValue(const AValue: TQuantumNumber);
    procedure SetUseGlobalFormat(const AValue: Boolean);
  protected
    procedure DoGetIntegerValue(var NewValue:Longint; TheText: string; var StringValid: Boolean); override;
    procedure DoSetIntegerValue(NewValue: Integer; var NewText: string); override;
  public
    constructor Create(AOwner: TComponent); override;
    function ValueAB: TQuantumNumber;
  published
    property Value: TQuantumNumber read GetValue write SetValue default 0;
    property Format: TQuantumNumberFormat read FFormat write SetFormat default qftJ;
    property UseGlobalFormat: Boolean read FUseGlobalFormat write SetUseGlobalFormat default False;
    property Increment default 2;
  end;

function GlobalQuantumNumberFormat: TQuantumNumberFormat;
procedure SetGlobalQuantumNumberFormat(QuantumNumberFormat: TQuantumNumberFormat);

implementation

uses Forms, Controls, SysUtils;

var
  FGlobalQuantumNumberFormat: TQuantumNumberFormat = qftJ;

{ TQuantumSpin }

constructor TQuantumSpin.Create(AOwner: TComponent);
begin
  inherited;
  Increment := 2;
end;

procedure TQuantumSpin.DoGetIntegerValue(var NewValue: Longint;
  TheText: string; var StringValid: Boolean);
begin
  if (Format <> qft2J) and not ( (TheText = '') or (TheText = '-') or (TheText = '.') or (TheText = '-.') ) then
    StringValid := TryStrToQuantumNumber(TheText, NewValue, Format);
  inherited;
end;

procedure TQuantumSpin.DoSetIntegerValue(NewValue: Integer;
  var NewText: string);
begin
  if Format <> qft2J then
    NewText := QuantumNumberToStr(NewValue, Format);
  inherited;
end;

function TQuantumSpin.GetValue: TQuantumNumber;
begin
  Result := inherited Value;
end;

function TQuantumSpin.ValueAB: TQuantumNumber;
begin
  Result := inherited ValueAB;
end;

procedure TQuantumSpin.SetFormat(const AValue: TQuantumNumberFormat);
var
  OldValue: TQuantumNumber;
begin
  if FFormat <> AValue then begin
    OldValue := Value;
    FFormat := AValue;
    SetValue(OldValue);
  end;
end;

procedure TQuantumSpin.SetUseGlobalFormat(const AValue: Boolean);
begin
  FUseGlobalFormat := AValue;
  if UseGlobalFormat then
    Format := GlobalQuantumNumberFormat;
end;

procedure TQuantumSpin.SetValue(const AValue: TQuantumNumber);
begin
  inherited Value := AValue;
end;

function GlobalQuantumNumberFormat: TQuantumNumberFormat;
begin
  Result := FGlobalQuantumNumberFormat;
end;

procedure SetGlobalQuantumNumberFormat(QuantumNumberFormat: TQuantumNumberFormat);
  procedure HandleControl(Control: TControl);
  var
    i: Integer;
  begin
    if Control is TQuantumSpin then
      with TQuantumSpin(Control) do begin
        if UseGlobalFormat then
          Format := GlobalQuantumNumberFormat;
      end
    else if Control is TWinControl then
      with TWinControl(Control) do
        for i := 0 to ControlCount - 1 do
          HandleControl(Controls[i]);
  end;
var
  i: Integer;
begin
  if QuantumNumberFormat <> GlobalQuantumNumberFormat then begin
    { Propagte change to all controls }
    FGlobalQuantumNumberFormat := QuantumNumberFormat;
    for i := 0 to Screen.CustomFormCount - 1 do
      HandleControl(Screen.CustomForms[i]);
  end;
end;

end.
