unit Split;

{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  {$IFNDEF CLX_} Messages, Graphics, Controls, Forms, Dialogs,
  {$ELSE} QGraphics, QControls, QForms, QDialogs, Qt, Types, {$ENDIF}
  SysUtils, Classes;

type

  TCanResizeEvent = procedure(Sender: TObject; var NewSize: Integer;
    var Accept: Boolean) of object;

  TResizeStyle = (rsNone, rsLine, rsUpdate);

  { TSplit }

  TSplit = class({$IFDEF LCL_}TCustomControl{$ELSE}TGraphicControl{$ENDIF})
  private
    {$IFDEF MSWINDOWS}
    FLineDC: HDC;
    {$ENDIF}
    FDownPos: TPoint;
    FSplit: Integer;
    FMinSize: Integer;
    FMaxSize: Integer;
    FResizeStyle: TResizeStyle;
    FControl: TControl;
    FNewSize: Integer;
    FOldSize: Integer;
    FActiveControl: TWinControl;
    FOldKeyDown: TKeyEvent;
    FBeveled: Boolean;
    FLineVisible: Boolean;
    FOnCanResize: TCanResizeEvent;
    FOnMoved: TNotifyEvent;
    FOnPaint: TNotifyEvent;
    FTopLeftWindow, FBottomRightWindow: TControl;
    FAutoHide: Boolean;
    procedure AllocateLineDC;
    procedure DrawLine;
    procedure ReleaseLineDC;
    procedure UpdateControlSize;
    procedure CalcSplitSize(X, Y: Integer; out NewSize, Split: Integer);
    procedure UpdateSize(X, Y: Integer);
    procedure FocusKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure SetBeveled(Value: Boolean);
    procedure SetAlign(AAlign: TAlign);
    procedure SetTopLeftWindow(AControl: TControl);
    procedure SetBottomRightWindow(AControl: TControl);
    procedure UpdateBothVisible(OtherSplit: TSplit);
    procedure UpdatePositions;
    function FindControl: TControl;
    function FindOtherSplit: TSplit;
  protected
    function CanResize(var NewSize: Integer): Boolean; reintroduce; virtual;
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure StopSizing; dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: integer); override;
    procedure ParentResized; { Call from parent.OnResize for full autohide behaviour }
    procedure OneWindowVisible(VisibleWindow: TControl);
    procedure MakeBothVisible;
    procedure CycleVisible;
    procedure Flip(TopLeftSize: Integer = 0; BottomRightSize: Integer = 0);
  published
    property TopLeftWindow: TControl read FTopLeftWindow write SetTopLeftWindow;
    property BottomRightWindow: TControl read FBottomRightWindow write SetBottomRightWindow;
    property AutoHide: Boolean read FAutoHide write FAutoHide;

    property Align write SetAlign default alLeft;
    property ResizeStyle: TResizeStyle read FResizeStyle write FResizeStyle default rsLine;
    property Beveled: Boolean read FBeveled write SetBeveled default True;
    property Color;
    property MinSize: Integer read FMinSize write FMinSize default 30;
    property ParentColor;
    property Visible;
    property OnCanResize: TCanResizeEvent read FOnCanResize write FOnCanResize;
    property OnMoved: TNotifyEvent read FOnMoved write FOnMoved;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
  end;

{$IFDEF SINGLE}
{$IFDEF VER80}
{$R Split.d16}
{$ELSE}
{$R Split.d32}
{$ENDIF}
procedure Register;
implementation

procedure Register;
begin
  RegisterComponents('CMW', [TSplit]);
end;

{$ELSE}
implementation
  {$IFDEF LCL_}
    uses
      {$IFDEF LCLgtk1} gtk, gdk, GTKDef, {$DEFINE USEGDK} {$ENDIF}
      {$IFDEF LCLgtk2} gtk2, gdk2, Gtk2Def, {$DEFINE USEGDK} {$ENDIF}
      LCLIntf, LCLType, lMessages;
  {$ENDIF}
{$ENDIF}



type
  THack = class(TWinControl);

{ TSplit }

constructor TSplit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Align := alLeft;
  Width := 3;
  Cursor := crHSplit;
  FMinSize := 30;
  FBeveled := True;
  FResizeStyle := rsLine;
end;

procedure TSplit.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
end;

procedure TSplit.AllocateLineDC;
begin
  {$IFDEF MSWINDOWS}
  FLineDC := GetDCEx(Parent.Handle, 0, DCX_CACHE or DCX_CLIPSIBLINGS
    or DCX_LOCKWINDOWUPDATE);
  {$ENDIF}
end;

procedure TSplit.DrawLine;
var
  P: TPoint;
  {$IFDEF CLX_}
  Painter: QPainterH;
  ParentWidget: QOpenWidgetH;
  {$ENDIF}
  {$IFDEF USEGDK}
  DC: TGtkDeviceContext;
  ParentForm: TCustomForm;
  WinControl: TWinControl;
  PS: TPaintStruct;
  {$ENDIF}
begin
  FLineVisible := not FLineVisible;
  P := Point(Left, Top);
  if Align in [alLeft, alRight] then
    P.X := Left + FSplit
  else
    P.Y := Top + FSplit;
  {$IFDEF MSWINDOWS}
  with P do PatBlt(FLineDC, X, Y, Width, Height, PATINVERT);
  {$ENDIF}
  {$IFDEF CLX_}
  if (Parent is TCustomForm) then
    with TCustomForm(Parent) do  begin
      if (Menu <> nil) and (Menu.ParentWidget = Handle) and
        QWidget_isVisible(Menu.Handle) then
      Inc(P.Y, QWidget_height(Menu.Handle));
    end;
  ParentWidget := QOpenWidgetH(Parent.Handle);
  QOpenWidget_setWFlags(ParentWidget,
    QOpenWidget_getWFlags(ParentWidget) or Integer(WidgetFlags_WPaintUnclipped));
  try
    Painter := QPainter_create;
    try
      QPainter_begin(Painter, QWidget_to_QPaintDevice(ParentWidget));
      try
        QPainter_setRasterOp(Painter, RasterOp_NotROP);
        Canvas.Brush.Style := bsSolid;
        Canvas.Brush.Color := clBlack;
        QPainter_fillRect(Painter, P.X, P.Y, Width, Height, Canvas.Brush.Handle);
      finally
        QPainter_end(Painter);
      end;
    finally
      QPainter_destroy(Painter);
    end;
  finally
    QOpenWidget_setWFlags(ParentWidget,
      QOpenWidget_getWFlags(ParentWidget) and not Integer(WidgetFlags_WPaintUnclipped));
  end;
  {$ENDIF}
  {$IFDEF USEGDK}
  WinControl := Parent;
  while Assigned(WinControl) and not (WinControl is TCustomForm) do
    WinControl := WinControl.Parent;
  if not Assigned(WinControl) then Exit;
  P := Parent.ClientToScreen(P);
  P := WinControl.ScreenToClient(P);
  ParentForm := WinControl as TCustomForm;
  // DC := TGtkDeviceContext(ParentForm.Canvas.Handle);
  // DC := TGtkDeviceContext(GetDC(ParentForm.Handle));
  DC := TGtkDeviceContext(BeginPaint(ParentForm.Handle, PS));
  gdk_gc_set_subwindow(DC.GC, GDK_INCLUDE_INFERIORS);
  // LPtoDP(ParentForm.Canvas.Handle,P,1);
  // ParentForm.Canvas.FillRect(P.X, P.Y, P.X+Width, P.Y+Height);
  // WriteLn(P.X,  ' ', P.Y,  ' ', P.X+Width, ' ', P.Y+Height);
  gdk_gc_set_function(DC.GC, GDK_INVERT);
  gdk_draw_rectangle(DC.Drawable, DC.GC, 1, P.X, P.Y, Width, Height);
  gdk_gc_set_function(DC.GC, GDK_COPY);
  gdk_gc_set_subwindow(DC.GC, GDK_CLIP_BY_CHILDREN);
  EndPaint(ParentForm.Handle, PS);
  {$ENDIF}
  {$IFDEF LCLGTK3}
  { TODO: }
  {$ENDIF}
end;

procedure TSplit.ReleaseLineDC;
begin
{$IFDEF MSWINDOWS}
  ReleaseDC(Parent.Handle, FLineDC);
{$ENDIF}
end;

procedure TSplit.SetAlign(AAlign: TAlign);
var
  OldAlign: TAlign;
begin
  if Align = alClient then
    raise Exception.Create('Help');
  if AAlign <> Align then begin
    OldAlign := Align;
    inherited Align := alNone;
    if (OldAlign in [alTop, alBottom]) = (AAlign in [alRight, alLeft]) then
      if OldAlign in [alTop, alBottom] then begin
        if Width > Height then Width := Height;
        Constraints.MaxHeight := 0; { Seems to be reqired for LCLgtk2 to avoid widgetset resize }
        Constraints.MaxWidth := Width;
      end else begin
        if Height > Width then Height := Width;
        Constraints.MaxWidth := 0;
        Constraints.MaxHeight := Height;
      end;
    inherited Align := AAlign;
    if (Cursor = crVSplit) and (Align in [alLeft, alRight]) then
      Cursor := crHSplit
    else if (Cursor = crHSplit) and (Align in [alBottom, alTop]) then
      Cursor := crVSplit;
  end;
end;

procedure TSplit.SetTopLeftWindow(AControl: TControl);
begin
  if (AControl = TopLeftWindow) or (AControl = BottomRightWindow) or (AControl = Self) then exit;
  FTopLeftWindow := AControl;
  if (TopLeftWindow <> nil) and TopLeftWindow.Visible and (TopLeftWindow.Align in [alTop, alLeft]) then begin
    Parent.DisableAlign;
    Align := TopLeftWindow.Align;
    // if BottomRightWindow <> nil then BottomRightWindow.Align := alClient;
    if TopLeftWindow.Align = alTop then
      Top := TopLeftWindow.Top + TopLeftWindow.Height
    else
      Left := TopLeftWindow.Left + TopLeftWindow.Width;
    Parent.EnableAlign;
  end;
end;

procedure TSplit.SetBottomRightWindow(AControl: TControl);
begin
  if (AControl = TopLeftWindow) or (AControl = BottomRightWindow) or (AControl = Self) then exit;
  FBottomRightWindow := AControl;
  if (BottomRightWindow <> nil) and BottomRightWindow.Visible and (BottomRightWindow.Align in [alBottom, alRight]) then begin
    Parent.DisableAlign;
    Align := BottomRightWindow.Align;
    // if TopLeftWindow <> nil then TopLeftWindow.Align := alClient;
    if BottomRightWindow.Align = alBottom then
      Top := BottomRightWindow.Top - Height
    else
      Left := BottomRightWindow.Left - Width;
    Parent.EnableAlign;
  end;
end;

procedure TSplit.Paint;
var
  {$IFDEF MSWINDOWS}
  FrameBrush: HBRUSH;
  {$ENDIF}
  R: TRect;
  T: array [0..2] of TPoint;
  C, D: Integer;
begin
  R := ClientRect;
  Canvas.Brush.Color := Color;
  Canvas.Brush.Style := bsSolid;
  Canvas.FillRect(ClientRect);
  if Beveled then begin
    if Align in [alLeft, alRight] then
      InflateRect(R, -1, 2)
    else
      InflateRect(R, 2, -1);
    OffsetRect(R, 1, 1);
    {$IFDEF MSWINDOWS}
    FrameBrush := CreateSolidBrush(ColorToRGB(clBtnHighlight));
    FrameRect(Canvas.Handle, R, FrameBrush);
    DeleteObject(FrameBrush);
    OffsetRect(R, -2, -2);
    FrameBrush := CreateSolidBrush(ColorToRGB(clBtnShadow));
    FrameRect(Canvas.Handle, R, FrameBrush);
    DeleteObject(FrameBrush);
    {$ENDIF}
    {$IFDEF UNIX}
    Canvas.Brush.Style := bsClear;
    Canvas.Pen.Color := clBtnHighlight;
    Canvas.Rectangle(R);
    OffsetRect(R, -2, -2);
    Canvas.Pen.Color := clBtnShadow;
    Canvas.Rectangle(R);
    {$ENDIF}
  end;
  // OffsetRect(R, 1, 1);
  Canvas.Brush.Color := clBlack;
  Canvas.Brush.Style := bsSolid;
  if Align in [alLeft, alRight] then begin
    C := (R.Top + R.Bottom) div 2;
    D := (R.Right - R.Left) div 2;
    T[0].X := R.Left;
    T[1].X := R.Right;
    T[2].X := R.Right;
    T[0].Y := C + D + 1;
    T[1].Y := C + 1;
    T[2].Y := C + 2*D + 1;
    Canvas.Polygon(T);
    T[2].X := R.Left;
    T[0].Y := C - 1;
    T[1].Y := C - D - 1;
    T[2].Y := C - 2*D - 1;
    Canvas.Polygon(T);
  end else begin
    C := (R.Right + R.Left) div 2;
    D := (R.Bottom - R.Top) div 2;
    T[0].Y := R.Top;
    T[1].Y := R.Bottom;
    T[2].Y := R.Bottom;
    T[0].X := C + D + 1;
    T[1].X := C + 1;
    T[2].X := C + 2*D + 1;
    Canvas.Polygon(T);
    T[2].Y := R.Top;
    T[0].X := C - 1;
    T[1].X := C - D - 1;
    T[2].X := C - 2*D - 1;
    Canvas.Polygon(T);
  end;
  if Assigned(FOnPaint) then FOnPaint(Self);
end;

function TSplit.CanResize(var NewSize: Integer): Boolean;
begin
  Result := True;
  if Assigned(FOnCanResize) then FOnCanResize(Self, NewSize, Result);
end;

function TSplit.FindControl: TControl;
begin
  Result := nil;
  if not (Align in [alLeft, alRight, alTop, alBottom]) then Exit;
  if TopLeftWindow = nil then
    Result := BottomRightWindow
  else if (BottomRightWindow = nil) or not BottomRightWindow.Visible then
    Result := TopLeftWindow
  else if not TopLeftWindow.Visible or (BottomRightWindow.Align <> alClient) then
    Result := BottomRightWindow
  else
    Result := TopLeftWindow;
  if (Result <> nil) and not Result.Visible then
    Result := nil;
end;

{$IFDEF LCL_}
type
  TFakeMouseCapture = class
  public
    Control: TControl;
    procedure OnUserInput(Sender: TObject; Msg: Cardinal);
    procedure CancelCapture;
  end;

  TOpenControl = class(TControl)
  end;
  
var
  FakeMouseCapture: TFakeMouseCapture;

{ Fake capture process that is currently not implemented in LCL
  Though some filtering is applied, the control should be written
  so as to be able to handle duplicate events }
procedure EnforceMouseCapture(AControl: TControl);
begin
  if not assigned(FakeMouseCapture) then
    FakeMouseCapture := TFakeMouseCapture.Create;
  with FakeMouseCapture do begin
    CancelCapture;
    Control := AControl;
    Application.AddOnUserInputHandler(@OnUserInput);
  end;
end;

{ TFakeMouseCapture }

procedure TFakeMouseCapture.OnUserInput(Sender: TObject; Msg: Cardinal);
var
  p:TPoint;
begin
  if Msg = LM_LBUTTONUP then
    Application.RemoveOnUserInputHandler(@OnUserInput);
  p := Mouse.CursorPos;
  p := Control.ScreenToClient(p);
  if (p.x < 0) or (p.y < 0) or (p.x > Control.Width) or (p.y > Control.Height) then
    if Msg = LM_MOUSEMOVE then
      TOpenControl(Control).MouseMove([], p.x, p.y)
    else if Msg = LM_LBUTTONUP then
      TOpenControl(Control).MouseUp(mbLeft, [], p.x, p.y)
end;

procedure TFakeMouseCapture.CancelCapture;
begin
  { Note: Harmless if not already on handler list }
  Application.RemoveOnUserInputHandler(@OnUserInput);
end;
{$ELSE}
procedure EnforceMouseCapture(AControl: TControl);
begin
end;
{$ENDIF}

procedure TSplit.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  I: Integer;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Button = mbLeft then begin
    EnforceMouseCapture(Self);
    FControl := FindControl;
    FDownPos := Point(X, Y);
    if Assigned(FControl) then
    begin
      if Align in [alLeft, alRight] then
      begin
        FMaxSize := Parent.ClientWidth;
        FOldSize := FControl.Width;
        if not AutoHide then
          Dec(FMaxSize, FMinSize);
        for I := 0 to Parent.ControlCount - 1 do
          with Parent.Controls[I] do
            if Align in [alLeft, alRight] then Dec(FMaxSize, Width);
        Inc(FMaxSize, FControl.Width);
      end
      else
      begin
        FMaxSize := Parent.ClientHeight;
        FOldSize := FControl.Height;
        if not AutoHide then
          Dec(FMaxSize, FMinSize);
        for I := 0 to Parent.ControlCount - 1 do
          with Parent.Controls[I] do
            if Align in [alTop, alBottom] then Dec(FMaxSize, Height);
        Inc(FMaxSize, FControl.Height);
      end;
      UpdateSize(X, Y);
      AllocateLineDC;
      {$IFNDEF LCL_} { TODO: LCL_ also ? }
      with ValidParentForm(Self) do
        if ActiveControl <> nil then
        begin
          FActiveControl := ActiveControl;
          FOldKeyDown := THack(FActiveControl).OnKeyDown;
          THack(FActiveControl).OnKeyDown := FocusKeyDown;
        end;
      {$ENDIF}
      if ResizeStyle = rsLine then DrawLine;
    end;
  end;
end;

procedure TSplit.OneWindowVisible(VisibleWindow: TControl);
var
  OtherSplit: TSplit;
  OtherWindow: TControl;
begin
  OtherSplit := FindOtherSplit;
  if VisibleWindow = TopLeftWindow then
    OtherWindow := BottomRightWindow
  else
    OtherWindow := TopLeftWindow;
  if VisibleWindow.Visible and (VisibleWindow.Align = alClient)
     and (not Assigned(OtherWindow) or not OtherWindow.Visible) then Exit;
  Parent.DisableAlign;
  if AutoHide then
    Beveled := True;
  VisibleWindow.Visible := True;
  VisibleWindow.Align := alClient;
  if VisibleWindow = TopLeftWindow then begin
    BottomRightWindow.Visible := False;
    if Align in [alLeft, alRight] then begin
      Align := alRight;
      if Assigned(OtherSplit) and (OtherSplit.TopLeftWindow = BottomRightWindow) then
        Left := OtherSplit.Left - Width
      else
        Left := BottomRightWindow.Left + BottomRightWindow.Width - Width;
      { Next may Change order }
      if Left + Width > Parent.ClientWidth then Left := Parent.ClientWidth - Width;
    end else begin
      Align := alBottom;
      if Assigned(OtherSplit) and (OtherSplit.TopLeftWindow = BottomRightWindow) then
        Top := OtherSplit.Left - Height
      else
        Top := BottomRightWindow.Top + BottomRightWindow.Height - Height;
      if Top + Height > Parent.ClientHeight then Top := Parent.ClientHeight - Height;
    end
  end else begin
    TopLeftWindow.Visible := False;
    if Align in [alLeft, alRight] then begin
      Align := alLeft;
      if Assigned(OtherSplit) and (OtherSplit.BottomRightWindow = TopLeftWindow) then
        Left := OtherSplit.Left + OtherSplit.Width
      else
        Left := TopLeftWindow.Left;
      BottomRightWindow.Left := Left + Width;
    end else begin
      Align := alTop;
      if Assigned(OtherSplit) and (OtherSplit.BottomRightWindow = TopLeftWindow) then
        Top := OtherSplit.Top + OtherSplit.Height
      else
        Top := TopLeftWindow.Top;
      BottomRightWindow.Top := Top + Height;
    end;
  end;
  UpdatePositions;
  Parent.EnableAlign;
end;

procedure TSplit.UpdatePositions;
var
  OtherSplit: TSplit;
begin
  Parent.DisableAlign;
  OtherSplit := FindOtherSplit;
  case Align of
    alLeft, alRight:
      begin
        if TopLeftWindow.Visible then begin
          if Assigned(OtherSplit) and (OtherSplit.BottomRightWindow = TopLeftWindow) then
            TopLeftWindow.Left := OtherSplit.Left + OtherSplit.Width;
          TopLeftWindow.Width := Left - TopLeftWindow.Left;
        end;
        if BottomRightWindow.Visible then begin
          if Assigned(OtherSplit) and (OtherSplit.TopLeftWindow = BottomRightWindow) then
            BottomRightWindow.Width := OtherSplit.Left - (Left + Width)
          else
            BottomRightWindow.Width := BottomRightWindow.Width + BottomRightWindow.Left - Left - Width;
          BottomRightWindow.Left := Left + Width;
        end;
      end;
    alTop, alBottom:
      begin
        if TopLeftWindow.Visible then begin
          if Assigned(OtherSplit) and (OtherSplit.BottomRightWindow = TopLeftWindow) then
            TopLeftWindow.Top := OtherSplit.Top + OtherSplit.Height;
          TopLeftWindow.Height := Top - TopLeftWindow.Top;
        end;
        if BottomRightWindow.Visible then begin
          if Assigned(OtherSplit) and (OtherSplit.TopLeftWindow = BottomRightWindow) then
            BottomRightWindow.Height := OtherSplit.Top - (Top + Height)
          else
            BottomRightWindow.Height := BottomRightWindow.Height + BottomRightWindow.Top - Top - Height;
          BottomRightWindow.Top := Top + Height;
        end;
      end;
  end;
  Parent.EnableAlign;
end;

procedure TSplit.UpdateBothVisible(OtherSplit: TSplit);
{$IFDEF FPC}
var
  OldBounds: TRect;
{$ENDIF}
begin
  Parent.DisableAlign;
  if AutoHide then
    Beveled := False;
  TopLeftWindow.Visible := True;
  BottomRightWindow.Visible := True;
  {$IFDEF FPC}
  { Setting Aling to none reverts the bounds to the BaseBounds, which defeats relative
    resizing below }
  OldBounds := TopLeftWindow.BoundsRect;
  TopLeftWindow.Align := alNone;
  TopLeftWindow.BoundsRect := OldBounds;
  OldBounds := BottomRightWindow.BoundsRect;
  BottomRightWindow.Align := alNone;
  BottomRightWindow.BoundsRect := OldBounds;
  {$ENDIF}
  UpdatePositions;
  if Assigned(OtherSplit) then
    if OtherSplit.TopLeftWindow = BottomRightWindow then
      Align := alLeft
    else
      Align := alRight;
  if Align in [alLeft, alTop] then begin
    TopLeftWindow.Align := Align;
    BottomRightWindow.Align := alClient;
  end else begin
    BottomRightWindow.Align := Align;
    TopLeftWindow.Align := alClient;
  end;
  Parent.EnableAlign;
end;

procedure TSplit.MakeBothVisible;
begin
  if TopLeftWindow.Visible and BottomRightWindow.Visible then Exit;
  FControl := FindControl;
  Parent.DisableAlign;
  if Align in [alLeft, alRight] then
    if TopLeftWindow.Visible then
      Left := TopLeftWindow.Left + TopLeftWindow.Width div 2
    else
      Left := BottomRightWindow.Left + BottomRightWindow.Width div 2
  else
    if TopLeftWindow.Visible then
      Top := TopLeftWindow.Top + TopLeftWindow.Height div 2
    else
      Top := BottomRightWindow.Top + BottomRightWindow.Height div 2;
  UpdateBothVisible(FindOtherSplit);
  Parent.EnableAlign;
  FControl := nil;
end;

procedure TSplit.UpdateControlSize;
var
  OtherControl: TControl;
  OtherNew: Integer;
  OtherSplit: TSplit;
  Shift: Integer;
begin
  if FNewSize <> FOldSize then begin
    OtherSplit := FindOtherSplit;
    OtherControl := TopLeftWindow;
    if OtherControl = FControl then OtherControl := BottomRightWindow;
    Parent.DisableAlign;
    try
      if (FNewSize < FMinSize) and AutoHide then begin
        { Hide current window - simple placement }
        OneWindowVisible(OtherControl);
      end else if OtherControl.Visible or (FOldSize - FNewSize > FMinSize) then begin
        { Both Windows visible or to be made so }
        Shift := FOldSize - FNewSize;
        if FControl = TopLeftWindow then
          Shift := -Shift;
        { First stab at moving }
        if Align in [alLeft, alRight] then
          Left := Left + Shift
        else
          Top := Top + Shift;
        { Determine if OtherControl winds up hidden }
        if Align in [alLeft, alRight] then
          OtherNew := OtherControl.Width - (FNewSize - FOldSize)
        else
          OtherNew := OtherControl.Height - (FNewSize - FOldSize);
        if (OtherNew < FMinSize) and AutoHide then begin
          OneWindowVisible(FControl);
        end else
          { Both windows remain visible }
          UpdateBothVisible(OtherSplit);
      end;
    finally
      Parent.EnableAlign;
    end;
    if Assigned(FOnMoved) then FOnMoved(Self);
  end;
end;

procedure TSplit.CalcSplitSize(X, Y: Integer; out NewSize, Split: Integer);
var
  S: Integer;
begin
  if Align in [alLeft, alRight] then begin
    Split := X - FDownPos.X;
    S := FControl.Width;
  end else begin
    Split := Y - FDownPos.Y;
    S := FControl.Height;
  end;
  if FControl = TopLeftWindow then
    Inc(S, Split)
  else
    Dec(S, Split);
  NewSize := S;
  if (S < FMinSize) and not AutoHide then
    NewSize := FMinSize
  else if NewSize < 0 then
    NewSize := 0
  else if S > FMaxSize then
    NewSize := FMaxSize;
  if S <> NewSize then
  begin
    if Align in [alRight, alBottom] then
      S := S - NewSize else
      S := NewSize - S;
    Inc(Split, S);
  end;
end;

procedure TSplit.UpdateSize(X, Y: Integer);
begin
  CalcSplitSize(X, Y, FNewSize, FSplit);
end;

procedure TSplit.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  NewSize, Split: Integer;
begin
  inherited MouseMove(Shift, X, Y);
  if Assigned(FControl) then
  begin
    CalcSplitSize(X, Y, NewSize, Split);
    if CanResize(NewSize) then
    begin
      if ResizeStyle = rsLine then DrawLine;
      FNewSize := NewSize;
      FSplit := Split;
      if ResizeStyle = rsUpdate then UpdateControlSize;
      if ResizeStyle = rsLine then DrawLine;
    end;
  end;
end;

procedure TSplit.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if Assigned(FControl) then
  begin
    if ResizeStyle = rsLine then DrawLine;
    UpdateControlSize;
    StopSizing;
  end;
end;

procedure TSplit.FocusKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if Key = VK_ESCAPE then
    StopSizing
  else if Assigned(FOldKeyDown) then
    FOldKeyDown(Sender, Key, Shift);
end;

procedure TSplit.SetBeveled(Value: Boolean);
begin
  FBeveled := Value;
  Repaint;
end;

procedure TSplit.StopSizing;
begin
  if Assigned(FControl) then
  begin
    if FLineVisible then DrawLine;
    FControl := nil;
    ReleaseLineDC;
    if Assigned(FActiveControl) then
    begin
      THack(FActiveControl).OnKeyDown := FOldKeyDown;
      FActiveControl := nil;
    end;
  end;
  if Assigned(FOnMoved) then
    FOnMoved(Self);
end;

procedure TSplit.ParentResized; { Call from parent.OnResize for full autohide behaviour }
var
 alClientControl, Other: TControl;
begin
  if AutoHide and (TopLeftWindow <> nil) and (BottomRightWindow <> nil)
     and TopLeftWindow.Visible and BottomRightWindow.Visible then begin
     alClientControl := TopLeftWindow;
     if alClientControl.Align <> alClient then begin
       alClientControl := BottomRightWindow;
       Other := TopLeftWindow;
     end else
       Other := BottomRightWindow;
     if ((Align in [alLeft, alRight]) and (alClientControl.Width < MinSize))
        or ((Align in [alTop, alBottom]) and (alClientControl.Height < MinSize)) then
       OneWindowVisible(Other);
  end;
end;

function TSplit.FindOtherSplit: TSplit;
var
  i: Integer;
begin
  { Look for another Split Window }
  for i := 0 to Parent.ControlCount-1 do
    if (Parent.Controls[i] is TSplit) and (Parent.Controls[i] <> Self) then begin
      Result := TSplit(Parent.Controls[i]);
      if (Result.TopLeftWindow = BottomRightWindow)
         or (Result.BottomRightWindow = TopLeftWindow) then
        Exit;
    end;
  Result := nil;
end;

procedure TSplit.Flip(TopLeftSize: Integer = 0; BottomRightSize: Integer = 0);
  { If only TopLeftSize > 0, take it as total width and do a 50:50 split }
  function FlipAlign(AAlign: TAlign):TAlign;
  begin
    case AAlign of
      alTop: Result := alLeft;
      alBottom: Result := alRight;
      alLeft: Result := alTop;
      alRight: Result := alBottom;
    else
      Result := AAlign;
    end;
  end;
begin
  if (TopLeftWindow = nil) or (BottomRightWindow = nil) then
    Exit;
  if TopLeftSize = 0 then
    if Align in [alLeft, alRight] then
      TopLeftSize := TopLeftWindow.Height
    else
      TopLeftSize := TopLeftWindow.Width;
  if BottomRightSize = 0 then begin
    TopLeftSize := TopLeftSize div 2;
    BottomRightSize := TopLeftSize;
  end;
  Parent.DisableAlign;
  TopLeftWindow.Align := FlipAlign(TopLeftWindow.Align);
  BottomRightWindow.Align := FlipAlign(BottomRightWindow.Align);
  Align := FlipAlign(Align);
  { Adjust position; if only one window visible keep that }
  if TopLeftWindow.Visible then
    if BottomRightWindow.Visible then begin
      { ... otherwise go for 50:50 split }
      { Need to set all bounds to avoid order flipping problems }
      if Align in [alLeft, alRight] then begin
        with TopLeftWindow do
          SetBounds(Left, Top, TopLeftSize, Height);
        Left := TopLeftWindow.Left + TopLeftWindow.Width;
        BottomRightWindow.SetBounds(Left+Width, Top, BottomRightSize, Height);
      end else begin
        with TopLeftWindow do
          SetBounds(Left, Top, Width, TopLeftSize);
        Top := TopLeftWindow.Top + TopLeftWindow.Height;
        BottomRightWindow.SetBounds(Left, Top+Height, Width, BottomRightSize);
      end;
    end else
      OneWindowVisible(TopLeftWindow)
  else
    OneWindowVisible(BottomRightWindow);
  Parent.EnableAlign;
end;

procedure TSplit.CycleVisible;
begin
  if not Assigned(BottomRightWindow) or not Assigned(TopLeftWindow) then
    Exit;
  if BottomRightWindow.Visible and TopLeftWindow.Visible then
    OneWindowVisible(TopLeftWindow)
  else if TopLeftWindow.Visible then
    OneWindowVisible(BottomRightWindow)
  else
    MakeBothVisible;
end;

{$IFDEF LCL_}
finalization
  FakeMouseCapture.Free;
{$ENDIF}
end.
